;;; $Id: anorm.scm,v 1.4 1996/08/28 17:47:39 queinnec Exp $

;;; This file defines the A-normalization, a transformation described
;;; in PLDI 93 (pp 237--247) by Flanagan, Sabry, Duba and Felleisen.
;;; It is a useful transformation before generating instructions since
;;; it avoids let forms in non-tail position (that would have to be
;;; generated as let-expressions (something that does not belong to C,
;;; nor to Java ...)).

;;; Exportation:
;;;			(a-normalize Program)
;;; No new classes.
;;; End of exportations.

;;; This transformation must be used before the box transformation
;;; (this saves some methods). 

;;;====================================================================
;;; There are three main generic functions.

(define (a-normalize o)
  (a-normalize! o (lambda (o) o)) )

;;; By default, return the expression to the continuation.
;;; This will be the case for References, Null-Argument, Box-Read ...

(define-generic (a-normalize! (o Program) k)
  (k o) )

;;; References are taken care of by default method.

(define-method (a-normalize! (o Assignment) k)
  (a-normalize-name!
   (Assignment-form o)
   (lambda (f) 
     (k (let ((oo (clone o)))
          (set-Assignment-form! oo f)
          oo )) ) ) )
;;; NOTE: Using clone above is better written in MeroonV3 with the
;;; update form. This form allows to take care of any subclass of
;;; Assignment.

(define-method (a-normalize! (o Function) k)
  (set-Function-body! o (a-normalize (Function-body o)))
  (call-next-method) )

(define-method (a-normalize! (o Alternative) k)
  (a-normalize-name!
   (Alternative-condition o)
   (lambda (c)
     (k (make-Alternative
         c
         (a-normalize (Alternative-consequent o))
         (a-normalize (Alternative-alternant o)) )) ) ) )

;(define-method (a-normalize! (o Sequence) k)
;  (a-normalize-name!
;   (Sequence-first o)
;   (lambda (f)
;     (a-normalize-name! 
;      (Sequence-last o)
;      (lambda (l)
;        (k (make-Sequence f l)) ) ) ) ) )

(define-method (a-normalize! (o Sequence) k)
  (a-normalize-name!
   (Sequence-first o)
   (lambda (f)
     (k (make-Sequence
         f (a-normalize (Sequence-last o)) )) ) ) )

;;; NOTE: Another analyze should take care of simplifying Sequences with 
;;; a first error-free term. Pay attention that (uninitialized-let (x) x 2)
;;; should provoke an error. See Analyzes/rmseq.scm

;;; Constant are taken care of by default method.
                      
(define-method (a-normalize! (o Predefined-Application) k)
  (a-normalize! (Predefined-Application-arguments o)
                (lambda (a) 
                  (k (make-Predefined-Application
                      a
                      (Predefined-Application-variable o) )) ) ) )

(define-method (a-normalize! (o Regular-Application) k)
  (a-normalize-name! (Regular-Application-function o)
                     (lambda (f)
                       (a-normalize!
                        (Regular-Application-arguments o)
                        (lambda (a)
                          (k (make-Regular-Application a f)) ) ) ) ) )

;;; Let forms of more than one variable are split into elementary ones.
;;; This is not problematic since variables are already differentiated.

(define-method (a-normalize! (o Fix-Let) k)
  (let ((vars (Fix-Let-variables o))
        (args (Fix-Let-arguments o)) )
    (if (pair? vars)
        (if (pair? (cdr vars))
            ;; split the let form
            (a-normalize!
             (Arguments-first args)
             (lambda (f)
               (make-Fix-Let
                (list (car vars))
                (make-Arguments f (make-No-Argument))
                (a-normalize! (make-Fix-Let
                               (cdr vars)
                               (Arguments-others args)
                               (Fix-Let-body o) )
                              k ) ) ) )
            ;; let with only one local variable
            (a-normalize!
             (Arguments-first args)
             (lambda (f)
               (make-Fix-Let
                vars
                (make-Arguments f (make-No-Argument))
                (a-normalize! (Fix-Let-body o) k) ) ) ) )
        ;; Let without local variable
        (Fix-Let-body o) ) ) )

;;; Null-Argument taken care of by default method.

(define-method (a-normalize! (o Arguments) k)
  (a-normalize-name!
   (Arguments-first o)
   (lambda (a)
     (a-normalize! (Arguments-others o)
                   (lambda (a*)
                     (k (make-Arguments a a*)) ) ) ) ) )

(define-method (a-normalize! (o Monitoring) k)
  (a-normalize-name!
   (Monitoring-handler o)
   (lambda (h)
     (a-normalize! (Monitoring-body o)
                   (lambda (a)
                     (k (make-Monitoring h a)) ) ) ) ) )

(define-method (a-normalize! (o Escaper) k)
  (a-normalize-name!
   (Escaper-body o)
   (lambda (a)
     (k (make-Escaper (Escaper-variable o) a)) ) ) )

;;;====================================================================
;;; This function introduces new local (immutable) variables to hold
;;; temporary results. The intuitive meaning is: Compute o, bind its
;;; value to a new fresh variable and give it to the continuation K.

(define a-normalize-index 0)

(define (a-normalize-name! o k)
  (a-normalize! 
   o
   (lambda (a)
     (if (a-normalized? a)
         (k a)
         (let ((v (make-Renamed-Local-Variable 
                   'tmp #f #f a-normalize-index )))
           (set! a-normalize-index (+ a-normalize-index 1))
           (make-Fix-Let (list v)
                         (make-Arguments a (make-No-Argument))
                         (k (make-Local-Reference v)) ) ) ) ) ) )

;;;====================================================================
;;; This generic predicate recognizes simple expressions

(define-generic (a-normalized? (o Program))
  #f )

(define-method (a-normalized? (o Reference))
  #t )

(define-method (a-normalized? (o Constant))
  #t )

(define-method (a-normalized? (o Function))
  #t )

;(define-method (a-normalized? (o Box-Read))
;  #t )

;;; end of anorm.scm
