;;; $Id: box.scm,v 1.5 1995/06/26 12:16:21 queinnec Exp $

;;; This file defines the box transformation. Its goal is to remove
;;; mutable local variables in favor of mutable boxes. After this
;;; transformation local variables are always immutable and
;;; substitutions can be applied on them without problems.

;;; Exportation: 
;;;                (insert-box! Program)
;;;                Box-* classes and related functions.
;;; End of exportations.

;;; Box transformation. Basically transform:
;;;   (lambda (... x ...)         (lambda (... x ...)
;;;     ... (set! x ...)     into   (let ((x (new-box x))...)
;;;         x ... )                   ... (box-set! x ...)
;;;                                       (box-ref x) ... )
;;; This is easy since the mutability of variables is already set up by
;;; the objectify function. This file is changed after the book: the 
;;; Box-Creation now has a body field to avoid using Sequence. The effect
;;; is the same as introducing a new special form saying that:
;;;    (let ((x (new-box x))) body) is (Box-Creation x body).

;;;====================================================================
;;; These classes are needed by the box transformation. The
;;; Local-Assignment class is no longer needed, it is replaced by
;;; the Box-Write class.

(define-class Box-Read Form (reference))

(define-class Box-Write Form (reference form))

(define-class Box-Creation Form (variable body))

;;;====================================================================
;;; By default, recurse on program instances.

(define-generic (insert-box! (o Program))
  (update-walk! insert-box! o) )

;;; a reference to X becomes (box-ref X). Don't recurse.

(define-method (insert-box! (o Local-Reference))
  (if (Local-Variable-mutable? (Local-Reference-variable o))
      (make-Box-Read o) 
      o ) )

;;; (set! X ...) becomes (box-set! X ...), but recurse on the inner form.

(define-method (insert-box! (o Local-Assignment))
  (make-Box-Write (Local-Assignment-reference o)
                  (insert-box! (Local-Assignment-form o)) ) )

;;; Let forms or functions may introduce mutable bindings then wrap
;;; their body within a sequence of box-creation to patch these bindings
;;; and make them to pass through a box.

(define-method (insert-box! (o Fix-Let))
  (set-Fix-Let-arguments! o (insert-box! (Fix-Let-arguments o)))
  (set-Fix-Let-body! 
   o (insert-box!
      (boxify-mutable-variables (Fix-Let-body o) (Fix-Let-variables o)) ) )
  o )

(define-method (insert-box! (o Function))
  (set-Function-body!
   o (insert-box!
      (boxify-mutable-variables (Function-body o) (Function-variables o)) ) )
  o )

;;; The variable of a bind-exit form may be mutable.

(define-method (insert-box! (o Escaper))
  (set-Escaper-body!
   o (insert-box!
      (boxify-mutable-variables (Escaper-body o)
                                (list (Escaper-variable o)) ) ))
  o )

;;; For each mutable variable, allocate a new box to hold its value.
;;; If X and Y are mutable, then the body is transformed into 
;;;     (let ((X (new-box X))) (let ((Y (new-box Y))) body) 
;;; but with the new Box-Creation class, it is rather:
;;;     (Box-creation X (Box-creation Y body)).

(define (boxify-mutable-variables form variables)
  (if (pair? variables)
      (if (Local-Variable-mutable? (car variables))
          (make-Box-Creation 
           (car variables)
           (boxify-mutable-variables form (cdr variables)) )
          (boxify-mutable-variables form (cdr variables)) )
      form ) )  

;;; end of box.scm
