;;; $Id: inian.scm,v 1.2 1995/06/24 14:15:14 queinnec Exp $

;;; Simple-minded initialization analysis. It tries to safely determine
;;; a set of variables that can be safely accessed ie they are written
;;; before being read so it is not necessary to check whether they are
;;; initialized or not. The analysis may also find some variables that
;;; will raise an error since they are surely read before being written.

;;; Exportation: 
;;;                (initialization-analyze! Program)
;;; End of exportations.
;;; Importations:
;;;    from walk.scm     update-walk!
;;; End of importations.

;;; Just use the regular code-walker that has the advantage to walk
;;; programs in the order of fields which generally is the evaluation
;;; order. Stops at the first expression which may not finish (looping or
;;; raising errors).

(define (initialization-analyze! e)
  (call/cc (lambda (exit) (inian! e (lambda () (exit 'finished)))))
  e )

;;;==================================================================
;;; This is the class of Exceptions raised by this analysis.
;;; Uninitialized-Variable is raised when a variable (global or local
;;; if introduced by a uninitialized-let form) is evaluated before
;;; being defined. For this continuable exception, if the handler returns
;;; #t then the analysis will assume the variable is in fact really
;;; initialized and will continue. If #f, the analysis stops.

(define-class Uninitialized-Variable Exception
  ( variable   ; a Variable
    ) )

(define-class Initialization-Warning Compilation-Warning
  () )

;;; These are the functions to raise exceptions or warnings.
;;; In both cases, these exceptions are continuable and so can be 
;;; ignored by exception handler if required.

(define (inian-warning msg . hints)
  (diagnose (make-Initialization-Warning 
             "Initialization analysis" msg hints )
            #t ) )

(define (inian-error variable)
  (diagnose (make-Uninitialized-Variable variable)
            #t ) )

;;;==================================================================
;;; This is the generic function that approximates which variables are
;;; (A) surely initialized and which one are (B) surely not
;;; initialized. What to do in these situations depends on the caller
;;; of the compiler: for instance, a compilation required by eval might not
;;; be interested in (B).

;;; NOTE: we use update-walk! for that analysis so don't forget that
;;; methods should behave as the identity if they return. They may
;;; escape thus finishing prematurely the analysis.

(define-generic (inian! (e) exit)
  (update-walk! inian! e exit) )

(define-method (inian! (e Global-Assignment) exit)
  (call-next-method)
  (let ((gv (Global-Assignment-variable e)))
    (unless (Global-Variable-initialized? gv)
      (set-Global-Variable-initialized?! gv #t)
      (inian-warning "Surely initialized variable" gv) )
    e ) )

(define-method (inian! (e Global-Reference) exit)
  (let ((gv (Global-Reference-variable e)))
    (cond ((Predefined-Variable? gv) e)
          ((Global-Variable-initialized? gv) e)
          (else 
           ;; if the handler returns #t, this means that the variable
           ;; was found to be initialized by other means.
           (if (inian-error gv)
               (set-Global-Variable-initialized?! gv #t)
               (exit) ) ) )
    e ) )

(define-method (inian! (e Alternative) exit)
  (inian! (Alternative-condition e) exit)
  (exit) )

(define-method (inian! (e Application) exit)
  (call-next-method)
  (exit) )

(define-method (inian! (e Function) exit)
  e )

;;; end of inian.scm
