;;; $Id: lift.scm,v 1.5 1994/10/15 17:32:43 queinnec Exp $

;;; This file computes closed variables and (virtually) lift functions.

;;; Exportation: 
;;;                (lift! Program)
;;; End of exportations.

;;;==================================================================
;;; These are the new classes needed for that transformation.

;;; Free-Reference refines Local-Reference when accessing a free
;;; variable. 

(define-class Free-Local-Reference Local-Reference ())

;;; A Flat-Function holds the list of the local variables that appear
;;; free in its body. Strictly speaking, this is not a list but an
;;; instance of Free-Environment (or No-Free) ie a Form to allow to
;;; walk them.

(define-class Flat-Function With-Temp-Function 
  ( free             ; an instance of Free-Environment or No-Free
    ) )

(define-class Free-Environment Form (first others))

(define-class No-Free Form ())

;;;==================================================================
;;; lifting transformation. Identifies free variables and
;;; abstractions. Pay attention to the walking order since free
;;; variables are collected via side-effects.

(define (lift! o)
  (lift-procedures! o #f '()) )

;;; The code walker. Walk O, store free variables in FLATFUN (the
;;; closest abstraction), all bound variables appear in VARS.

(define-generic (lift-procedures! (o Program) flatfun vars)
  (update-walk! lift-procedures! o flatfun vars) )

;;; Collect free variables into flatfun. This works for references whether
;;; direct or through a Box-Read or Box-Write instances.

(define-method (lift-procedures! (o Local-Reference) flatfun vars)
  (let ((v (Local-Reference-variable o)))
    (if (memq v vars)
        o (begin (adjoin-free-variable! flatfun o)
                 (make-Free-Local-Reference v) ) ) ) )

;;; Transform Functions into flat-functions. Don't forget that free
;;; variables of the newfun abstraction may be free in flatfun.  This
;;; first method is necessary if temporary variables were not
;;; collected; the second takes directly a With-Temp-Function.

(define-method (lift-procedures! (o Function) flatfun vars)
  (let* ((localvars (Function-variables o))
         (body (Function-body o))
         (newfun (make-Flat-Function 
                  localvars
                  body
                  '()
                  (make-No-Free) )) )
    (set-Flat-Function-body! newfun (lift-procedures! body newfun localvars))
    (let ((free* (Flat-Function-free newfun)))
      (set-Flat-Function-free! newfun (lift-procedures! free* flatfun vars)) )
    ;; sort free references and order them along vars.  FUTURE
    newfun ) )

(define-method (lift-procedures! (o With-Temp-Function) flatfun vars)
  (let* ((localvars (Function-variables o))
         (body (Function-body o))
         (newfun (make-Flat-Function 
                  localvars 
                  body 
                  (With-Temp-Function-temporaries o)
                  (make-No-Free) )) )
    (set-Flat-Function-body! newfun (lift-procedures! body newfun localvars))
    (let ((free* (Flat-Function-free newfun)))
      (set-Flat-Function-free! newfun (lift-procedures! free* flatfun vars)) )
    ;; sort free references and order them along vars.  FUTURE
    newfun ) )

;;; Fix-Let is also a binding form that may shadow free variables.

(define-method (lift-procedures! (o Fix-Let) flatfun vars)
  (set-Fix-Let-arguments!
   o (lift-procedures! (Fix-Let-arguments o) flatfun vars) )
  (let ((newvars (append (Fix-Let-variables o) vars)))
    (set-Fix-Let-body! o (lift-procedures! (Fix-Let-body o) flatfun newvars))
    o ) )

;;; Adjoin a reference to a free variable if not already present in
;;; FLATFUN.  The list of such free references is kept in a subclass
;;; of Program (Free-Environment) since it will be evaluated to build
;;; closures, it might be code-walked etc.

(define (adjoin-free-variable! flatfun ref)
  (let check ((free* (Flat-Function-free flatfun)))
    (if (No-Free? free*)
        (set-Flat-Function-free! 
         flatfun (make-Free-Environment ref (Flat-Function-free flatfun)) )
        (unless (eq? (Reference-variable ref)
                     (Reference-variable (Free-Environment-first free*)) )
          (check (Free-Environment-others free*)) ) ) ) )

;;;==================================================================
;;; Count terms

(define-method (number-of (o Free-Environment))
  (+ 1 (number-of (Free-Environment-others o))) )

(define-method (number-of (o No-Free)) 0)

;;; end of lift.scm
