;;; $Id: quotes.scm,v 1.3 1994/10/12 19:07:58 queinnec Exp $

;;; This transformation linearizes quotations. Quotation extraction
;;; must be performed before. To linearize a quotation is to explode
;;; it in a myriad of simple operations to rebuild the data. It may
;;; also try to share subparts of quotations. 

;;; Exportations:
;;;        (linearize-quotation! Flattened-Program) : Flattened-Program
;;;        some classes.
;;; end of exportations

;;;==================================================================
;;; New classes 

;;; An enriched class for Flattened-Program with an additional field
;;; holding the linarized quotations.

(define-class Linearized-Flattened-Program Flattened-Program
  ( linearizations  ; a list of Quotation-Variables
    initialization  ; a Form initializing Quotation-Variables
    ) )

;;; New Quotation-Variables are gathered as subclasses of
;;; Linearized-Quotation-Variable.

(define-class Linearized-Quotation-Variable Quotation-Variable
  ( ) :virtual )

;;; Immediate-Quotation-Variable refines Quotation-Variable. Their
;;; value can ony be a boolean, the empty list, a character, a string
;;; or a small number. They are numbered independently of
;;; Quotation-Variables.

(define-class Immediate-Quotation-Variable Linearized-Quotation-Variable
  ( ) )

(define-class Symbol-Quotation-Variable Linearized-Quotation-Variable
  ( string      ; a Immediate-Quotation-Variable holding the string upon
                ; which a symbol must be made.
    ) )

(define-class Cons-Quotation-Variable Linearized-Quotation-Variable
  ( car         ; a Quotation-Variable holding the car
    cdr         ; a Quotation-Variable holding the cdr
    ) )

(define-class Vector-Quotation-Variable Linearized-Quotation-Variable
  ( terms       ; a list of Quotation-Variable for all terms.
    ) )

;;;==================================================================
;;; Classify quotations and try to share them. 
;;; Note: the initialization field may be left as #f if there are no 
;;; quotations at all. Otherwise it is a Sequence of Global-Assignments
;;; assigning the Quotation-Variables to their value.

(define (linearize-quotations! prg)
  (let* ((quotations (Flattened-Program-quotations prg))
         (newprg (make-Linearized-Flattened-Program
                  (Flattened-Program-body prg)
                  (Flattened-Program-quotations prg)
                  (Flattened-Program-definitions prg)
                  '()
                  '() )) )
    (set-Linearized-Flattened-Program-initialization!
     newprg      
     (let gen ((result #f)
               (qv* quotations) )
       (if (pair? qv*)
           (let* ((qvar  (car qv*))
                  (value (Quotation-Variable-value qvar))
                  (lqv   (analyze-quotation! value newprg))
                  (init  (make-Global-Assignment 
                          (make-Global-Reference lqv)
                          qvar )) )
             (gen (if result (make-Sequence result init) init)
                  (cdr qv*) ) )
           result ) ) )
    ;; order the linearized variables.
    (set-Linearized-Flattened-Program-linearizations!
     newprg (reverse! (Linearized-Flattened-Program-linearizations newprg)) )
    newprg ) )

;;; Adjoin a new quotation variable to the list of quotations of a
;;; program.  Variables are named after their rank in the list of
;;; quotations. Pay attention that the same naming scheme was adopted
;;; for Quotation-Variable.

(define (adjoin-quotation! qvar prg)
  (let ((quotations (Linearized-Flattened-Program-linearizations prg))
        (qv (already-seen-value? (Quotation-Variable-value qvar) prg)) )
    (or qv
        (begin
          (set-Linearized-Quotation-Variable-name! qvar (length quotations))
          (set-Linearized-Flattened-Program-linearizations! 
           prg (cons qvar quotations) )
          qvar ) ) ) )

;;; Decompose (linearize) a value and return a new Quotation-Variable
;;; to describe it. The result is inserted in the list of the
;;; quotations of the program by the caller of
;;; analyze-quotation!. When Quotation-Variables are created, they are
;;; given the temporary number -1. This number will be changed by
;;; adjoin-quotation!.

(define (analyze-quotation! value prg)
  (or (already-seen-value? value prg)
      (adjoin-quotation!
       (cond 
        ((or (string? value)
             (integer? value)
             (null? value)
             (char? value)
             (boolean? value) )
         (make-Immediate-Quotation-Variable -1 value) )
        ((symbol? value)
         (make-Symbol-Quotation-Variable
          -1
          value
          (analyze-quotation! (symbol->string value) prg) ) )
        ((pair? value)
         (make-Cons-Quotation-Variable
          -1
          value
          (analyze-quotation! (car value) prg)
          (analyze-quotation! (cdr value) prg) ) )
        ((vector? value)
         (make-Vector-Quotation-Variable
          -1
          value
          (map (lambda (v) (analyze-quotation! v prg))
               (vector->list value) ) ) )
        (else (linearize-quotation-error
               "Unknown type of value" value )) )
       prg ) ) )
  
;;; Look for a value in the current list of quoted values.

(define (already-seen-value? value prg)
  (let scan ((qvars (Linearized-Flattened-Program-linearizations prg)))
    (and (pair? qvars)
         (let ((qvar (car qvars)))
           (if (equal? (Quotation-Variable-value qvar) value)
               qvar
               (scan (cdr qvars)) ) ) ) ) )

;;; end of quotes.scm
