;;; $Id: qdeval.scm,v 1.12 1996/08/27 18:39:00 queinnec Exp $

;;; This file defines a quick and dirty interpreter to test objectified
;;; programs. It directly interprets the Abstract Syntax Tree. It
;;; allows to test the various transformations by direct
;;; interpretation of the resulting AST.

(define evaluate-error
  (create-compilation-exception-raiser "Evaluation") )

;;;==================================================================
;;; Basic machinery to evaluate abstract syntax tree without any
;;; transformations ie right after objectification.

;;; E 	: Program instance
;;; SR	: Alist mapping Variable to values.

;;; These are the classes for primitives and user-defined closures.

(define-class RunTime-Procedure Object (body variables environment))
(define-class RunTime-Primitive Object (address comparator arity))

;;; find a binding for a variable in the current environment.
;;; A binding is represented by a pair.

(define (find-binding var sr)
  (let ((v (assq var sr)))
    (unless (pair? v) (evaluate-error "Lost variable" var))
    v ) )

;;; The generic function for evaluation of the abstract syntax tree.

(define-generic (evaluate (e Program) sr))

(define-method (evaluate (e Local-Reference) sr)
  (let ((v (find-binding (Local-Reference-variable e) sr)))
    (cdr v) ) )

(define undefined-value (cons 'undefined 'value))

(define-method (evaluate (e Global-Reference) sr)
  (let ((v (find-binding (Global-Reference-variable e) sr)))
    (if (eq? (cdr v) undefined-value)
        (evaluate-error "Uninitialized variable" e)
        (cdr v) ) ) )

(define-method (evaluate (e Global-Assignment) sr)
  (let ((v (find-binding (Global-Assignment-variable e) sr)))
    (let* ((old (cdr v))
           (value (evaluate (Global-Assignment-form e) sr)) )
      (set-cdr! v value)
      old ) ) )

(define-method (evaluate (e Predefined-Reference) sr)
  (let ((v (find-binding (Predefined-Reference-variable e) sr)))
    (cdr v) ) )

(define-method (evaluate (e Local-Assignment) sr)
  (let* ((ref (Local-Assignment-reference e))
         (v (find-binding (Local-Reference-variable ref) sr)) )
    (unless (pair? v) (evaluate-error "Lost reference" e))
    (let ((old (cdr v))
          (value (evaluate (Local-Assignment-form e) sr)) )
      (set-cdr! v value)
      old ) ) )

(define-method (evaluate (e Function) sr)
  (make-RunTime-Procedure (Function-body e)
                          (Function-variables e)
                          sr ) )

(define-method (evaluate (e Alternative) sr)
  (if (evaluate (Alternative-condition e) sr)
      (evaluate (Alternative-consequent e) sr)
      (evaluate (Alternative-alternant e) sr) ) )

(define-method (evaluate (e Sequence) sr)
  (begin (evaluate (Sequence-first e) sr)
         (evaluate (Sequence-last e) sr) ) )

(define-method (evaluate (e Constant) sr)
  (Constant-value e) )

(define-method (evaluate (e Arguments) sr)
  (cons (evaluate (Arguments-first e) sr)
        (evaluate (Arguments-others e) sr) ) )

(define-method (evaluate (e No-Argument) sr)
  '() )

(define-method (evaluate (e Regular-Application) sr)
  (let ((f (evaluate (Regular-Application-function e) sr))
        (args (evaluate (Regular-Application-arguments e) sr)) )
    (invoke f args) ) )

(define-generic (invoke (f) args))

(define-method (invoke (f RunTime-Procedure) args)
  (unless (let check ((variables (RunTime-Procedure-variables f))
                      (args args) )
            (if (pair? variables)
                (or (Local-Variable-dotted? (car variables))
                    ;; (car variables) is a regular variable
                    (and (pair? args)
                         (check (cdr variables) (cdr args)) ) )
                (not (pair? args)) ) )
    (evaluate-error "Wrong arity" f args) )
  (evaluate (RunTime-Procedure-body f)
            (sr-extend (RunTime-Procedure-environment f)
                       (RunTime-Procedure-variables f) 
                       args ) ) )

(define-method (invoke (f RunTime-Primitive) args)
  (unless ((RunTime-Primitive-comparator f) 
           (length args) (RunTime-Primitive-arity f) )
    (evaluate-error "Wrong arity" f args) )
  (apply (RunTime-Primitive-address f) args) )

(define (sr-extend sr variables values)
  (if (pair? variables)
      (if (Local-Variable-dotted? (car variables))
          ;; don't recurse, it's a Dotted-Variable (and necesserarily
          ;; the last one).
          (cons (cons (car variables) values)
                sr )
          ;; a regular variable
          (if (pair? values)
              (cons (cons (car variables) (car values))
                    (sr-extend sr (cdr variables) (cdr values)) )
              (evaluate-error "Not enough values" variables) ) )
      (if (null? values) 
          sr
          (evaluate-error "Too much values" values) ) ) )

(define-method (evaluate (e Predefined-Application) sr)
  (let ((v (find-binding (Predefined-Application-variable e) sr.current)))
    (let ((f (cdr v))
          (args (evaluate (Predefined-Application-arguments e) sr)) )
      (invoke f args) ) ) )

(define-method (evaluate (e Fix-Let) sr)
  (let ((args (evaluate (Fix-Let-arguments e) sr)))
    (evaluate (Fix-Let-body e) 
              (sr-regular-extend sr (Fix-Let-variables e) args) ) ) )

;;; (assume (every? Regular-Variable? variables))

(define (sr-regular-extend sr variables values)
  (if (pair? variables)
      (if (pair? values)
          (cons (cons (car variables) (car values))
                (sr-regular-extend sr (cdr variables) (cdr values)) )
          (evaluate-error "Not enough values" variables) )
      (if (null? values) sr
          (evaluate-error "Too much values" values) ) ) )

;;;===================================================================
;;; Boxes are coded in the cdr of pairs, why not !

(define-method (evaluate (e Box-Read) sr)
  (cdr (evaluate (Box-read-reference e) sr)) )

(define-method (evaluate (e Box-Write) sr)
  (set-cdr! (evaluate (Box-Write-reference e) sr)
            (evaluate (Box-Write-form e) sr) ) )

(define-method (evaluate (e Box-Creation) sr)
  (let ((v (find-binding (Box-Creation-variable e) sr)))
    (set-cdr! v (cons 'box (cdr v)))
    (evaluate (Box-Creation-body e) sr) ) )

;;;===================================================================
;;; New methods introduced by colltemp.scm

(define-class With-Temp-RunTime-Procedure RunTime-Procedure
  ( temporaries 
    ) )

(define-method (evaluate (e With-Temp-Function) sr)
  (make-With-Temp-RunTime-Procedure
   (With-Temp-Function-body e)
   (With-Temp-Function-variables e)
   sr
   (With-Temp-Function-temporaries e) ) )

(define-method (evaluate (e Preallocated-Fix-Let) sr)
  (let ((args (evaluate (Preallocated-Fix-Let-arguments e) sr)))
    (for-each (lambda (var val)
                (let ((v (find-binding var sr)))
                  (set-cdr! v val) ) )
              (Preallocated-Fix-Let-variables e)
              args )
    (evaluate (Preallocated-Fix-Let-body e) sr) ) )

(define-method (invoke (f With-Temp-RunTime-Procedure) args)
  (unless (let check ((variables (RunTime-Procedure-variables f))
                      (args args) )
            (if (pair? variables)
                (or (Local-Variable-dotted? (car variables))
                    ;; (car variables) is a regular variable
                    (and (pair? args)
                         (check (cdr variables) (cdr args)) ) )
                (not (pair? args)) ) )
    (evaluate-error "Wrong arity" f args) )
  (evaluate (RunTime-Procedure-body f)
            (sr-extend 
             (let ((temp* (With-Temp-RunTime-Procedure-temporaries f)))
               (sr-extend (RunTime-Procedure-environment f)
                          temp*
                          (map (lambda (v) undefined-value) temp*) ) )
             (RunTime-Procedure-variables f)
             args ) ) )

;;;===================================================================
;;; These are the new methods introduced by lift.scm

(define-method (evaluate (e Flat-Function) sr)
  (make-With-Temp-RunTime-Procedure
   (With-Temp-Function-body e)
   (With-Temp-Function-variables e)
   (evaluate (Flat-Function-free e) sr)
   (With-Temp-Function-temporaries e) ) )

(define-method (evaluate (e No-Free) sr)
  sr.current )

(define-method (evaluate (e Free-Environment) sr)
  (cons (cons (Reference-variable (Free-Environment-first e))
              (evaluate (Free-Environment-first e) sr) )
        (evaluate (Free-Environment-others e) sr) ) )

(define-method (evaluate (e Free-Local-Reference) sr)
  (let ((v (find-binding (Free-Local-Reference-variable e) sr)))
    (cdr v) ) )

;;;===================================================================
;;; These are the new methods introduced by extract.scm

;;; This vector will contain the definitions of all lifted functions.

(define *runtime-functions* '())

(define-method (evaluate (o Flattened-Program) sr)
  (set! *runtime-functions* 
        (apply vector (reverse (Flattened-Program-definitions o))) )
  ;; update the global environment for No-Free method.
  (set! sr.current 
        (sr-regular-extend 
         sr
         (Flattened-Program-quotations o)
         (map Quotation-Variable-value 
              (Flattened-Program-quotations o) ) ) )
  (evaluate (Flattened-Program-body o) sr.current) )

(define-method (evaluate (e Closure-Creation) sr)
  (let* ((index (Function-Definition-index (Closure-Creation-definition e)))
         (def (vector-ref *runtime-functions* index)) )
    (make-With-Temp-RunTime-Procedure
     (Function-Definition-body def)
     (Function-Definition-variables def)
     (evaluate (Closure-Creation-free e) sr)
     (Function-Definition-temporaries def) ) ) )

;;;===================================================================
;;; show some objects

(define-method (show (o Quotation-Variable) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "<Quoted" stream)
    (display (Quotation-Variable-name o) stream)
    (display "= " stream)
    (display (Quotation-Variable-value o) stream) ) )

(define-method (show (o Variable) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "<" stream)
    (display (Class-name (object->class o)) stream)
    (display " " stream)
    (display (Variable-name o) stream)
    (display ">" stream) ) )

;;;===================================================================
;;; New methods for quotes.scm (linearization of quotations).

(define-method (evaluate (o Linearized-Flattened-Program) sr)
  (set! *runtime-functions* 
        (apply vector (reverse (Flattened-Program-definitions o))) )
  ;; update the global environment for No-Free method.
  ;; First, allocate global quotation variables.n
  (set! sr.current
        (enrich sr.current
                (Flattened-Program-quotations o) ) )
  ;; Second, allocate and initialize linearized quotation variables
  (set! sr.current 
        (enrich sr.current
                (Linearized-Flattened-Program-linearizations o) ) )
  ;; Third, initialize quotation variables.
  (let ((init (Linearized-Flattened-Program-initialization o)))
    (when init (evaluate init sr.current)) )
  ;; Everything is ready now for the evaluation of the body of the program
  (evaluate (Flattened-Program-body o) sr.current) )

(define (enrich sr qvars)
  (if (pair? qvars)
      (let ((sr (elaborate (car qvars) sr)))
        (enrich sr (cdr qvars)) )
      sr ) )

;;; This generic function sets up the initial global environment for
;;; quotations. Use a special generic function since evaluate is
;;; restricted to subclasses of Program.

(define-generic (elaborate (o Quotation-Variable) sr)))

(define-method (elaborate (o Quotation-Variable) sr)
  (cons (cons o undefined-value) sr) )

(define-method (elaborate (o Immediate-Quotation-Variable) sr)
  (cons (cons o (Immediate-Quotation-Variable-value o)) 
        sr ) )

(define-method (elaborate (o Symbol-Quotation-Variable) sr)
  (let* ((b (find-binding (Symbol-Quotation-Variable-string o) sr))
         (str (cdr b)) )
    (cons (cons o (string->symbol str))
          sr ) ) )

(define-method (elaborate (o Cons-Quotation-Variable) sr)
  (let* ((ba (find-binding (Cons-Quotation-Variable-car o) sr))
         (bd (find-binding (Cons-Quotation-Variable-cdr o) sr)) )
    (cons (cons o (cons (cdr ba) (cdr bd)))
          sr ) ) )

(define-method (elaborate (o Vector-Quotation-Variable) sr)
  (let ((b* (map (lambda (qv) (find-binding qv sr))
                 (Vector-Quotation-Variable-terms o) )))
    (cons (cons o (apply vector (map cdr b*)))
          sr ) ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; Initial global compile-time environment

(define g.current
  (make-Global-Environment '() '()) )

;;; Initial runtime environment

(define sr.current '())

;;; Define values or functions in the predefined global environment.

(define-icslas-macro (definitial name value)
  `(let ((v (make-Predefined-Variable ',name #f)))
     (set-Global-Environment-predefined! 
      g.current (cons v (Global-Environment-predefined g.current)) )
     (set! sr.current (cons (cons v ,value) sr.current))
     ',name ) )

(define-icslas-macro (defprimitive name value desc)
  (let ((comparator (if (memq desc '(0 1 2 3)) '= '>=))
        (arity (case desc ((0 1 2 3) desc) ((>=0) 0) ((>=2) 2)
                 (else (eval-error 'incorrect-arity arity)) )) )
    `(let ((v (make-Predefined-Variable
               ',name (make-Functional-Description ,comparator ,arity) ))
           (f (make-RunTime-Primitive ,value ,comparator ,arity)) )
       (set-Global-Environment-predefined! 
        g.current (cons v (Global-Environment-predefined g.current)) )
       (set! sr.current (cons (cons v f) sr.current))
       ',name ) ) )

;;; Define an uninitialized location in the user global environment.

(define-icslas-macro (defvariable name)
  `(let ((v (make-Global-Variable ',name #f)))
     (set-Global-Environment-global! 
      g.current (cons v (Global-Environment-global g.current)) )
     (set! sr.current (cons (cons v undefined-value) sr.current))
     ',name ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Initialization of the predefined global environment.

(definitial t #t)
(definitial f #f)
(definitial nil '())

;;; This variable is needed for (begin) and (if #f p)
(definitial unspecified 813)

(defprimitive cons cons 2)
(defprimitive car car 1)
(defprimitive cdr cdr 1)
(defprimitive pair? pair? 1)
(defprimitive symbol? symbol? 1)
(defprimitive eq? eq? 2)
(defprimitive set-car! set-car! 2)
(defprimitive set-cdr! set-cdr! 2)
(defprimitive + + 2)
(defprimitive - - 2)
(defprimitive = = 2)
(defprimitive < < 2)
(defprimitive > > 2)
(defprimitive * * 2)
(defprimitive <= <= 2)
(defprimitive >= >= 2)
(defprimitive remainder remainder 2)
(defprimitive display display 1)

(defprimitive apply
  (lambda (f . args)
    (if (pair? args)
        (invoke f (let flat ((args args))
                    (if (null? (cdr args)) (car args)
                        (cons (car args) (flat (cdr args))) ) ))
        (evaluate-error "incorrect arity" 'apply) ) )
  >=2 )

(defprimitive call/cc
  (lambda (f)
    (call/cc 
     (lambda (k)
       (invoke f (list (make-RunTime-Primitive
                        k = 1 ))) ) ) )
  1 )

(defprimitive list
  (lambda args args)
  >=0 )    

;;; Define the global variables needed to test the evaluator on scheme.tst.

(defvariable a)
(defvariable foo)
(defvariable fact)
(defvariable primes)

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Testing the various transformations with qdeval. The
;;; compile-expression function is written so that you can stop
;;; transforming O at any point and test the result with evaluate.

;;; This function defines how to compile a program and mainly the
;;; various transformations that are to be applied on the program.
;;; Here we have:
;;;    box transformation,
;;;    initialization analysis,
;;;    collect temporary variables,
;;;    lifting in place,
;;;    extraction of quotations and closed lambdas,
;;;  ...                                         FUTURE

(define (compile-expression e)
  (let ((o (objectify e (make-Local-Environment '()) g.current)))
    ;; Alphatize variable names (to make it easier for C)
    (set! o (alphatize! o))
    ;; Anormalization
    (set! o (a-normalize o))
    ;; Add the Box transformation to remove mutable local variables:
    (set! o (insert-box! o))
    ;; Figure which global variables are surely initialized
    (set! o (initialization-analyze! o))
    ;; Closurize the whole program
    (set! o (make-Regular-Application
             (make-No-Argument)
             (make-Function '() o) ))
    ;; Collect temporary variables (turn Function into With-Temp-Function). 
    ;; DON'T perform this transformation if you want a SCHEME behavior!
    ;(set! o (gather-temporaries! o))
    ;; Lambda-lift in place (turn With-Temp-Function into Flat-Function)
    (set! o (lift! o))
    ;; Migrate quotations and functions towards the highest level.
    (set! o (extract-things! o))
    ;; linearize quotations
    (set! o (linearize-quotations! o))
    ;;  Return the result of all these transformations
    o ) )

(define-generic (qdeval-handler (exc) continuable?))

(define-method (qdeval-handler (exc Compilation-Exception) c?)
  (display "   *** *** *** " stderr)
  (display (Compilation-Exception-phase exc) stderr)
  (display " Exception *** *** ***" stderr)
  (newline stderr)
  (display (Compilation-Exception-message exc) stderr)
  (newline stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (newline stderr) )
            (Compilation-Exception-culprits exc) )
  (call-next-method) )

(define-method (qdeval-handler (exc Compilation-Warning) c?)
  (display (Compilation-Warning-phase exc) stderr)
  (display " Warning: " stderr)
  (display (Compilation-Warning-message exc) stderr)
  (newline stderr)
  (for-each (lambda (hint) 
              (show hint stderr)
              (newline stderr) )
            (Compilation-Warning-culprits exc) ) )

(define *create-global-variables-on-the-fly* #f)

(define-method (qdeval-handler (exc Unknown-Variable) c?)
  (if *create-global-variables-on-the-fly*
      (let ((v (make-Global-Variable (Unknown-Variable-name exc) #f))
            (g (Unknown-Variable-g exc)) )
        (set-Global-Environment-global!
         g (cons v (Global-Environment-Global g)) )
        (set! sr.current (cons (cons v undefined-value) sr.current))
        (make-Global-Reference v) )
      ;;; abort!
      (call-next-method) ) )

(define (icslas)
  (set! *create-global-variables-on-the-fly* #t)
  (interpreter
   "Icslas? "
   "Icslas= "
   #t
   (lambda (read print err)
     (define-method (qdeval-handler (exc Exception) c?) (err exc))
     (lambda ()
       (monitor qdeval-handler 
         (let* ((e (icslas-expand (read)))
                (prg (compile-expression e)) )
           (unveil prg)(newline) ; DEBUG
           (print (evaluate prg sr.current)) ) ) ) ) ) )

(define (test-icslas file)
  (set! *create-global-variables-on-the-fly* #f)
  (suite-test
   file
   "Icslas? "
   "Icslas= "
   #t
   (lambda (read check err)
     (define-method (qdeval-handler (exc Exception) c?) (err exc))
     (lambda ()
       (monitor qdeval-handler 
         (let ((e (icslas-expand (read))))
           (check (evaluate (compile-expression e) sr.current)) ) ) ) )
   equal? ) )

;;; Test: (test-icslas "../DEA/scheme.tst")

;;; end of qdeval.scm
