;;; $Id: check.scm,v 1.7 1995/06/26 12:16:23 queinnec Exp $

;;; This file defines how syntax is checked when an Sexp is converted
;;; into a Program instance. These functions return true if the syntax
;;; is correct otherwise they call the syntax-check-error function to provoke
;;; an anomaly. These functions only check the current level of the
;;; Sexp, deeper levels will be triggered by the objectification of
;;; subSexps. 

;;; Exportation: 
;;;                (check-* Sexp)
;;; End of exportations.
;;; Importations:
;;;    from utils.scm          (syntax-check-error string Sexp)
;;; End of importations.

;;; Don't check (pair? e), it is true by definition of objectify !

;;; (<form> <form> ...)

(define (check-application e)
  (if (null? (cdr (last-pair e)))
      #t
      (syntax-check-error "Dot in an application" e) ) )

;;; (lambda <variable-list> <body>)
;;; <variable-list> := ()  | ( <variable> . <variable-list> ) | <variable>
;;; Check if variables appear duplicated: this is illegal as in R4RS.

(define (check-abstraction e)
  (if (pair? (cdr e))
      (and (check-variable-list? (cadr e))
           (if (pair? (cddr e))
               #t
               (syntax-check-error "Body is missing" e) ) )
      (syntax-check-error "Variables are missing" e) ) )

(define (non-duplicated-variable? v seen)
  (if (memq v seen)
      (syntax-check-error "Duplicated variable" v)
      #t ) )

(define (check-variable-list? e)
  (define (variable-list? e seen)
    (or (and (pair? e)
             (symbol? (car e))
             (non-duplicated-variable? (car e) seen)
             (variable-list? (cdr e) (cons (car e) seen)) )
        (null? e)
        (and (symbol? e)
             (non-duplicated-variable? e seen) ) ) )
  (variable-list? e '()) )

;;; (set! <reference> <form>)

(define (check-assignment e)
  (if (and (pair? (cdr e))
           (check-reference (cadr e))
           (pair? (cddr e))
           (null? (cdddr e)) )
      #t
      (syntax-check-error "Incorrect assignment" e) ) )

;;; <reference> 
;;; This is left as it is for further extensions.

(define (check-reference e)
  (if (symbol? e)
      #t
      (syntax-check-error "Not a variable" e) ) )

;;; (begin <form> ...)
;;; The form (begin) is legal but returns an unspecified value.

(define (check-sequence e)
  (if (null? (cdr (last-pair e)))
      #t
      (syntax-check-error "Dot in a begin form" e) ) )

;;; (if <form> <form> [ <form> ] )
;;; Alternatives may be binary or ternary as in R4RS.

(define (check-alternative e)
  (if (and (pair? (cdr e))
           (pair? (cddr e)) )
      (if (or (and (pair? (cdddr e))
                   (null? (cddddr e)) )
              (null? (cdddr e)) )
          #t
          (syntax-check-error "Too long alternative" e) )
      (syntax-check-error "Incomplete alternative" e) ) )

;;; (quote <data>)

(define (check-quotation e)
  (if (pair? (cdr e))
      (if (null? (cddr e))
          #t
          (syntax-check-error "Too much quoted value" e) )
      (syntax-check-error "No quoted value" e) ) )

;;; (monitor <handler> <form>...)

(define (check-monitoring e)
  (if (pair? (cdr e))
      (if (null? (cdr (last-pair e)))
          #t
          (syntax-check-error "Dot in a monitor form" e) ) ) )

;;; (bind-exit (<variable>) <form> ...)

(define (check-escaper e)
  (if (pair? (cdr e))
      (if (null? (cdr (last-pair e)))
          (if (and (pair? (cadr e))
                   (symbol? (car (cadr e)))
                   (null? (cdr (cadr e))) )
              #t
              (syntax-check-error "Incorrect bind-exit variable" e) )
          (syntax-check-error "Dot in a bind-exit form" e) ) ) )

;;; end of check.scm
