;;; $Id: objify.scm,v 1.11 1997/02/28 19:50:26 queinnec Exp $

;;; This file converts a Sexp into a Program object. The functions to
;;; check the syntax of the Sexp are in the check.scm file. This
;;; objectification is the first step applied to any program before
;;; interpretation or compilation.

;;; Exportation: 
;;;                (objectify Sexp r g)
;;;                (number-of o)
;;; End of exportations.
;;; Importations:
;;;    from progr.scm    needs Meroon and Program subclasses
;;;         check.scm    (check-* Sexp) functions
;;;         ???          (extend-environment r vars)
;;;                      (find-variable? r|g var)
;;;                      (find-predefined-variable? g var)
;;; Requires a handler for Unknown-Variable.
;;; End of importations.

;;;=================================================================
;;; Objectification engine

;;; Use: first read (for example with the usual read function) an
;;; Sexpression, then converts it into a Program object with
;;; (objectify (read) r g). The R argument is a local environment
;;; holding the instances representing the local variables while G
;;; represents the global environment (predefined or not). R and G are
;;; instances of some classes describing environments. Some operations
;;; exist on them:
;;; 	(find-variable? symbol r) : Reference
;;; 	(find-variable? symbol g) : Reference
;;; 	(find-predefined-variable? symbol g) : Predefined-Variable
;;; 	(extend-environment r variables)

(define (objectify e r g)
  (if (atom? e)
      (if (symbol? e)
          (objectify-reference e r g)
          (objectify-quotation e r g) )
      (if (and (symbol? (car e))
               (find-variable? (car e) r) )
          ;; Local variables have precedence upon special forms.
          ;; This is not R4RS compliant!
          (begin  (check-application e)
                  (objectify-application (car e) (cdr e) r g) )
          (let ((p (assq (car e) *special-forms*)))
            (if (pair? p)
                ((cdr p) e r g)
                (begin
                  (check-application e)
                  (objectify-application (car e) (cdr e) r g)) ) ) ) ) )

;;; The Alist of special forms to objectify.

(define *special-forms* '())

(define-icslas-macro (define-special-form name objifier)
  `(begin 
     (set! *special-forms*
           (cons (cons ',name ,objifier)
                 *special-forms* ) )
     ',name ) )

;;; Special forms.

(define-special-form quote
  (lambda (e r g)
    (check-quotation e)
    (objectify-quotation (cadr e) r g) ) )

(define-special-form if
  (lambda (e r g)
    (check-alternative e)
    (if (pair? (cdddr e))
        (objectify-alternative3 (cadr e) (caddr e) (cadddr e)
                                r g )
        (objectify-alternative2 (cadr e) (caddr e)
                                r g ) ) ) )

(define-special-form begin
  (lambda (e r g)
    (check-sequence e)
    (objectify-sequence (cdr e) r g) ) )

(define-special-form set!
  (lambda (e r g)
    (check-assignment e)
    (objectify-assignment (cadr e) (caddr e) r g) ) )

(define-special-form lambda
  (lambda (e r g)
    (check-abstraction e)
    (objectify-function (cadr e) (cddr e) r g) ) )

;;; In Icslas the monitor special form controls exception handling.
;;; There are also the dynamic-let and dynamic-ref special forms.

(if-icslas-feature (or Icslas)
  (begin 
    (define-special-form monitor
      (lambda (e r g)
        (check-monitoring e)
        (objectify-monitoring (cadr e) (cddr e) r g) ) )
    ;; call/ep is the equivalent predefined function.
    ;(define-special-form bind-exit
    ;  (lambda (e r g)
    ;    (check-escaper e)
    ;    (objectify-escaper (caadr e) (cddr e) r g) ) )
    )
  )

;;;=================================================================
;;; Objectification methods.

(define (objectify-quotation value r g)
  (make-Constant value) )

;;; Alternatives can be binary or ternary. (if #f a) returns the value
;;; of the `unspecified' predefined global variable.

(define (objectify-alternative2 ec et r g)
  (make-Alternative (objectify ec r g)
                    (objectify et r g)
                    (make-Predefined-Reference
                     (find-predefined-variable? 'unspecified g) ) ) )

(define (objectify-alternative3 ec et ef r g)
  (make-Alternative (objectify ec r g)
                    (objectify et r g)
                    (objectify ef r g) ) )

;;; (begin) is not an error but returns the value of the `unspecified'
;;; predefined global variable.

(define (objectify-sequence e* r g)
  (if (pair? e*)
      (if (pair? (cdr e*))
          (let* ((a (objectify (car e*) r g))
                 (d (objectify-sequence (cdr e*) r g)) )
            (make-Sequence a d) )
          (objectify (car e*) r g) )
      (make-Predefined-Reference 
       (find-predefined-variable? 'unspecified g) ) ) )

;;; A monitor form has an implicit-begin body.

(define (objectify-monitoring h e* r g)
  (make-Monitoring (objectify h r g)
                   (objectify-sequence e* r g) ) )

;;; an escaper form has an implicit-begin body.

(define (objectify-escaper var e* r g)
  (let* ((ov   (make-Local-Variable var #f #f))
         (newr (extend-environment r (list ov))) )
    (make-Escaper ov (objectify-sequence e* newr g)) ) )

;;; Process an application. Identifies three subcases: a call to a
;;; predefined variable, a closed application where the first term is
;;; an explicit abstraction, and regular applications.

(define (objectify-application f e* r g)
  (let ((ff  (objectify f r g))
        (ee* (convert2arguments (map (lambda (e) (objectify e r g)) e*))) )
    (cond ((Function? ff)                             ;; REMOVE
           (process-closed-application ff ee* r g) )  ;; REMOVE
          ((Predefined-Reference? ff) 
           (let* ((fvf  (Predefined-Reference-variable ff))
                  (desc (Predefined-Variable-description fvf)) )
             (if (Functional-Description? desc)
                 (if ((Functional-Description-comparator desc)
                      (length e*) (Functional-Description-arity desc) )
                     (make-Predefined-Application ee* fvf)
                     (objectify-error "Incorrect predefined arity" f e*) )
                 (make-Regular-Application ee* ff) ) ) )
          (else (make-Regular-Application ee* ff)) ) ) )

;;; Process closed applications, handle the case of dotted variables.
;;; This could have been done by a post transformation but it is now
;;; written there and I do not know of any transformations that introduce
;;; such terms.

(define (process-closed-application f e* r g)
  (let ((v* (Function-variables f))
        (b  (Function-body f)) )
    (if (and (pair? v*) (Local-Variable-dotted? (car (last-pair v*))))
        (process-nary-closed-application f e* r g) 
        (if (= (number-of e*) (length (Function-variables f)))
            (make-Fix-Let (Function-variables f) e* (Function-body f))
            (objectify-error "Incorrect regular arity" f e*) ) ) ) )

;;; Process a closed application where the function is nary. Convert 
;;; ((lambda (x ... . last) ...) a ... x y z) into
;;; ((lambda (x ... last) ...) a ... (cons x (cons y (cons z '()))))
;;; So the dotted variable now appears as a regular variable.

(define (process-nary-closed-application f e* r g)
   (let* ((v* (Function-variables f))
          (b  (Function-body f))
          (o (make-Fix-Let 
              v*
              (let gather ((e* e*) (v* v*))
                (if (Local-Variable-dotted? (car v*))
                    (make-Arguments 
                     (let pack ((e* e*))
                       (if (Arguments? e*)
                           (make-Predefined-Application
                            (make-Arguments 
                             (Arguments-first e*) 
                             (make-Arguments 
                              (pack (Arguments-others e*)) 
                              (make-No-Argument) ) )
                            (find-predefined-variable? 'cons g) )
                           (make-Constant '()) ) )
                     (make-No-Argument) )
                    (if (Arguments? e*)
                        (make-Arguments (Arguments-first e*) 
                                        (gather (Arguments-others e*) 
                                                (cdr v*) ) )
                        (objectify-error "Incorrect dotted arity" f e*) ) ) )
              b )) )
     ;; This variable should no longer be considered as dotted.
     (set-Local-Variable-dotted?! (car (last-pair v*)) #f)
     o ) )
 
;;; Convert a list of Program instances into a Program instance, more
;;; precisely an Arguments instance terminated by a No-Argument instance.

(define (convert2arguments e*)
  (if (pair? e*)
      (make-Arguments (car e*) (convert2arguments (cdr e*)))
      (make-No-Argument) ) )

;;; compute the length of a sequence of things.

(define-generic (number-of (o)))

(define-method (number-of (o Arguments))
  (+ 1 (number-of (Arguments-others o))) )

(define-method (number-of (o No-Argument)) 0)

;;; There is an asymetry between Local- and Global- Assignment, the
;;; first field may be a reference or a variable.

(define (objectify-assignment variable e r g)
  (let ((ov (objectify variable r g))
        (of (objectify e r g)) )
    (cond ((Local-Reference? ov)
           ;; NOTE: mark the local variable as mutable (for future use).
           (set-Local-Variable-mutable?! (Local-Reference-variable ov) #t)
           (make-Local-Assignment of ov) )
          ((Global-Reference? ov)
           (make-Global-Assignment of (Global-Reference-variable ov)) )
          ((Predefined-Reference? ov)
           (objectify-error "Immutable predefined" variable)) ) ) )

(define (objectify-function names body r g)
  (let* ((vars (objectify-variables-list names))
         (b    (objectify-sequence body (extend-environment r vars) g)) )
    (make-Function vars b) ) )

(define (objectify-variables-list names)
  (if (pair? names)
      (cons (make-Local-Variable (car names) #f #f)
            (objectify-variables-list (cdr names)) )
      (if (symbol? names)
          (list (make-Local-Variable names #f #t))
          '() ) ) )

;;; Objectify a reference. Look for the variable in the local then in
;;; the global environment. Return the appropriate reference.

(define (objectify-reference variable r g)
  (or (find-variable? variable r)
      (find-variable? variable g)
      (objectify-free-global-reference variable r g) ) )

;;; end of objify.scm
