/* $Id: dmbgl.c,v 1.70 1998/06/01 20:18:33 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmbgl.c
 * This is the Bigloo runtime interface with DMeroon.
 * This file defines the routines that are specific to Bigloo and required
 * by DMeroon. It also defines the SchemeScript protocol.

 * This file exports an additional function which is written in Scheme
 * to allow multiple sizes to be conveniently given to the allocator.

	SIGNATURE:
 (Object (DMbgl_allocate Class . anything))

 */

#include "dmeroon.h"

DM_DefineVersionString(dmbgl,
  "$Id: dmbgl.c,v 1.70 1998/06/01 20:18:33 queinnec Exp $");

DM_Context* DMbgl_context = DM_NULL;

/* RESOURCE DM_CheckReturnedStatus
 * This macro may be used to check the result of a call to the API.
 * Rather than longjmp-ing to the most recent context, it raises an
 * exception for the underlying language.
 */

#define DM_CheckReturnedStatus(context, errcode) \
  if ( (errcode) != DME_SUCCESS ) { \
    DM_RaiseException(context, errcode); \
  }

#define DM_RaiseException(context, errcode) \
  DM_raise_exception(context, errcode, __FILE__, __LINE__)

/* {{{ Allocation

 * RESOURCE DM_allocate_bare_instance
 * This function allows DMeroon to allocate a DMeroon object using the
 * allocator of the support language. 
 * 
 * This is an unsafe allocator since the higher levels ensures that
 * <class> is a Class and <size> is a natural number. The size of the 
 * DMeroon and Bigloo prefixes is automatically added to the previous
 * size.

 NOTE: On Sony, Boehm's GC_malloc return addresses that are multiple
 of 4 whereas, alignment of the most stringent type (as handled by
 gcc) is 8.  For C compatibility since the kernel classes are defined
 in C and must be readable by DMeroon, DM_allocate_bare_instance must
 provide addresses that are multiple of 8. This would be a waste of
 room if performed everywhere, so I optimize this as follows. Such a
 stringent alignment is only needed when there are fields of size 8 or
 at least two indexed fields. So we correct the address returned by
 malloc for these cases. This information is coded in the
 DM_CLASS_IS_STRINGENT bit. 
 */

DM_Object*
DM_allocate_bare_instance (DM_Context *context,
                           DM_Class   *class, 
                           DM_nat3    size ) {
  DM_SupportPointer wrapped_o;

  if ( DM_HasOption(class, DM_CLASS_IS_STRINGENT) ) {
    size += DM_ROOM_FOR_REALIGNMENT;
    wrapped_o = DM_MALLOC(DM_PrefixSize() + size);
    if ( wrapped_o ) {
      wrapped_o = (void*)DM_OffsetRealign((DM_SupportAddress)wrapped_o);
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  } else {
    wrapped_o = DM_MALLOC(DM_PrefixSize() + size);
    if ( wrapped_o ) {
      /* NOTHING */
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  }

  /* Structure the object for Bigloo. */
  wrapped_o->header = MAKE_HEADER(DM_HEADER, DM_PrefixSize() + size);

  return DM_FromSupport(wrapped_o);
}

/* RESOURCE DM_allocate
 * This is the higher level function used directly from Bigloo to
 * allocate an instance. It is very specific since it takes a variable
 * number of arguments in Bigloo and a fix one in C. Its name in C
 * is DMbgl_allocate, its name in Scheme will be DM-allocate (this
 * mapping is defined in Bigloo/sig2bgl.scm).

 * FUTURE: This binding is costly in memory, an argc++argv style
 * interface would be preferrable.
 */

DM_SupportPointer 
DMbgl_allocate (DM_SupportPointer wrapped_class,
                DM_SupportPointer wrapped_list_of_args) {
  DM_ErrorCode errcode;
  DM_nat3      argv[DM_MAX_INDEXED_FIELDS];
  DM_nat3      argc = 0;
  DM_Object    *o;

  if (   !(DM_IsDMeroonInstance(wrapped_class)) ) {
    DM_RaiseException(DMbgl_context, DME_NOT_DMEROON_INSTANCE);
  }

  while ( PAIRP(wrapped_list_of_args) ) {
    DM_SupportPointer i_wrapped = CAR( wrapped_list_of_args );
    if ( ! DM_IsSupportFixnum(i_wrapped) )
      DM_RaiseException(DMbgl_context, DME_NOT_NUMBER);
    if ( argc >= DM_MAX_INDEXED_FIELDS )
      DM_RaiseException(DMbgl_context, DME_INCORRECT_NUMBER_OF_SIZES);
    {
      DM_SupportFixnum n = DM_FromSupportFixnum(i_wrapped);
      
      if ( n < 0 ) 
        DM_RaiseException(DMbgl_context, DME_NEGATIVE_SIZE);
      argv[argc++] = n;
    }
    wrapped_list_of_args = CDR( wrapped_list_of_args );
  }
  errcode = DMeroon_allocate(DMbgl_context,
                             &o,
                             DM_FromSupport(wrapped_class), 
                             argc, 
                             argv );
  DM_CheckReturnedStatus(DMbgl_context, errcode);
  return DM_ToSupport(o);
}

/* }}}
 * {{{ GC related stuff
 *
 * We also need to include this header so we can use the
 * GC_register_displacement from Boehm GC. This is copied from 
 * the gc.h file of the 4.3 release.
 */

typedef unsigned long GC_word;
extern void GC_register_displacement (GC_word size);
typedef void (*GC_finalization_proc)(void * obj, void * client_data);
extern void GC_register_finalizer(void * obj,
                                  GC_finalization_proc fn, 
                                  void * cd,
                                  GC_finalization_proc *ofn, 
                                  void ** ocd);
extern void GC_gcollect (void);

/* 
 * Force a GC.
 */

void
GC_gc () 
{

  DM_DebugAtLevel(3,
    fprintf(DM_debug_stream, "GC collecting\n");
    /* Clear the BoundedStack that holds the most recent Requests/Answers.
     * The simplest thing is to reallocate a fresh one. */
    { 
      DM_BoundedStack *bs = DM_Allocate1(DMbgl_context, BoundedStack, 2);
  
      bs->index     = 0;
      DM_lastreqans = bs;
    }
);
  
  /* Invoke GC. */
  (void) GC_gcollect();
}

/* 
 * When Exit instances are to be recycled, a message must be sent to the
 * remotely referenced Entry. This requires cooperation with the GC.
 * Fortunately, Boehm's GC provides this ability.
 */

void
DM_exit_finalizer (void *wrapped_exit, void *client_data ) {
  DM_Object *exit = DM_FromC(wrapped_exit);

  DM_DebugAtLevel(2,
    if ( ! DM_IsA(DMbgl_context, exit, Exit) ) {
      DM_Abort(DMbgl_context, DME_NOT_AN_EXIT);
    }
    fprintf(DM_debug_stream, "GC: %p is useless\n", exit);
);

  /* Remove the Exit instance from the Exits hashtable */
  DM_remove_exit(DMbgl_context, (DM_Exit*)exit);
}

void
DM_set_exit_finalization (DM_Exit *exit) {
  DMeroon_object *wrapped_exit = DM_ToC(exit);

  GC_register_finalizer(wrapped_exit, DM_exit_finalizer, NULL, NULL, NULL);
}

/* }}}
 * {{{ Printing

 * RESOURCE DM_print_char_star
 * These functions belong to Bigloo and know how to print a string or
 * a fixnum on a variety of output ports: file-port, string-port, or
 * Tcp of DMeroon.  When writing on a Tcp, the len argument if non zero
 * is taken for granted: len characters will be output even if some null
 * chars appear in string. There is an additional case when debugging
 * where the function accepts the port to be stdout or stderr.
 */

void
DM_print_char_star (DM_Context *context,
                    char       *string,
                    DM_nat3    len,
                    void       *output_port) {
  DM_nat3 n = len ? len : strlen(string);

  if ( output_port == NULL ) return;

  DM_DebugAtLevel(0,
    /* Recognize standard outgoing C streams to print on them. */
    if (   (output_port == (void*)stdout)
        || (output_port == (void*)stderr) ) {
      FILE *stream = (FILE*) output_port;

      fprintf(stream, "%*s", (int)n, string);
      fflush(stream);
      return;
    }
);

  if ( DM_IsObject(output_port) ) {
    if ( DM_IsA(context, output_port, OutBuffer) ) {
      DM_DeclareNarrowedObject(outbuffer, OutBuffer, output_port);
      
      DM_print_char_star_in_outbuffer(context, string, n, outbuffer);
      return;
    }
  } else {
    /* UGLY: Revert to the SupportPointer. */
    DM_SupportPointer wrapped_port = DM_ToSupport(output_port);

    if ( OUTPUT_STRING_PORTP( wrapped_port ) ) {
      strputs(string, wrapped_port);
      return;
    }
    
    if ( OUTPUT_PORTP( wrapped_port ) ) {
      FILE *fout = OUTPUT_PORT( wrapped_port ).file;
      fputs(string, fout);
      return;
    }
  }

  DM_Abort(context, DME_PRINT_CHAR_STAR_PROBLEM);
}

/* }}}
 * {{{ Errors
 * This function raises an error that may be caught by Bigloo.
 */

static obj_t DMbgl_DMeroon_string;

void
DM_raise_exception (DM_Context    *context,
                    DM_ErrorCode  code,
                    char          *file,
                    unsigned long line) 
{

  DM_DebugAtLevel(10,
    fflush(stdout);
    fprintf(stderr, "\n\n???!!! Error %lu in %s:%lu\n",
            (unsigned long)code, file, line );
);

  FAILURE(DMbgl_DMeroon_string,
          BINT( code ),
          MAKE_PAIR( string_to_bstring( file ),
                    MAKE_PAIR( BINT( line ),
                              BNIL ) ) );
  /* NOT REACHED */
  return;
}

/* }}}
 * {{{ Bigloo Scheme protocol

 * This protocol allows to send Scheme programs to be evaluated by
 * Bigloo.  These programs are transferred as strings and thus need
 * the non-standard open-input-string function to be read.  The
 * program is accumulated until complete then it is interpreted. The
 * program is determined to be complete when the last characters read
 * are \0\0\r\n. See DMeroon/Commands/urlscript.prl that sends such
 * scripts.
 */

/* RESOURCE DM_handle_scheme_script
 * This function is Bigloo dependent. It receives a string containing
 * some Scheme code and an OutBuffer in which to print the results.
 */

static void
DM_handle_scheme_script (DM_Context   *context,
                         char         *cstring,
                         DM_nat3      length,
                         DM_OutBuffer *outbuffer ) {
  obj_t bstring, bresult;

  /* First: Convert the cstring into a Bigloo string.
   * Code snarfed from runtime/Clib/string.c).  */
  bstring = GC_MALLOC_ATOMIC( STRING_SIZE + length );

#if( !defined( TAG_STRING ) || defined( BUMPY_GC ) )
  bstring->string_t.header = MAKE_HEADER( STRING_TYPE, 0 );
#endif	
  bstring->string_t.length = length;
  memcpy( &(bstring->string_t.char0), cstring, length );

  /* Second: Evaluate the string. This function is defined in dm.bgl
   * and should always return a string (errors occuring during the
   * evaluation of the string should be trapped, displayed in the
   * resulting string). 
   * NOTE: continuations may alter the evaluation... */
  bresult = DM_special_evals(bstring);
  if ( ! STRINGP(bresult) ) {
    DM_Abort(context, DME_NOT_BIGLOO_STRING);
  }

  /* Copy the result to the OutBuffer. */
  DM_print_char_star(context, 
                     BSTRING_TO_STRING( bresult ), 
                     STRING_LENGTH(bresult), 
                     outbuffer);
}

/* RESOURCE DM_scheme_protocol_handler
 * This function (not exported) implements the SchemeScript protocol.
 * A SchemeScript request is complete when the four last bytes are 
 * NUL NUL CR LF. The string is then evaluated by Bigloo.
 */

static DM_c_boolean
DM_scheme_protocol_handler (DM_Context  *context,
                            DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;
  DM_nat3 index = inbuffer->index;
  DM_nat3 lf_index = index;

  /* Look for the next LF. DM_handle_protocol ensures it exists. */
  while ( inbuffer->byte[lf_index] != '\n' ) 
    lf_index++;

  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    char         *q         = &(inbuffer->byte[inbuffer->maximum]);
    DM_OutBuffer *outbuffer = connection->emission;

    if (   (inbuffer->maximum > 4 )
        && ( *(--q) == '\n' )
        && ( *(--q) == '\r' )
        && ( *(--q) == '\0' )
        && ( *(--q) == '\0' ) ) {
      /* The SchemeScript request is complete. */
      DM_nat3 length = inbuffer->maximum - 4 - lf_index;

      DM_handle_scheme_script(context,
                              &(inbuffer->byte[lf_index+1]), 
                              length,
                              outbuffer );
      /* In all cases, clean up the Connection after evaluating the script. */
      inbuffer->index = inbuffer->maximum;
      /* close the port so the client knows it is finished. */
      DM_SetOption(connection, DM_PORT_IS_DRIED);
      return DM_TRUE;
    } else {
      /* The SchemeScript request is not yet complete. */
      return DM_FALSE;
    }
  } else {
    /* NOTHING: not a connection!  */
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* }}}
 * {{{ Initialization

 * Bigloo has its own allocator and DMeroon uses it so nothing needs to
 * be initialized since all predefined objects are statically built.
 * In fact, we allocate some Bigloo objects to ease error reporting.
 * NOTE: This routine is the main and single initialization routine, the
 * user must be aware of. Therefore, DM_initialize must call every other
 * internal initialization routines.
 */

void
DM_bigloo_initialize (DM_Context* context) 
{
  DM_LinkVersionStringIntoSite(dmbgl);
 
  /* Allocate a string to qualify DMeroon exceptions for Bigloo.  */
  DMbgl_DMeroon_string = string_to_bstring( "DMeroon" );

  /* Check some facts.
   * Is DM_SUPPORT_FIXNUM_MAXIMUM the greatest Bigloo fixnum ? */
  { DM_SupportFixnum m = 
      ( ((DM_SupportFixnum)1)
       << ( (CHAR_BIT*sizeof(DM_SupportFixnum)) - TAG_SHIFT - 1) ) - 1;
    DM_SupportFixnum v = DM_SUPPORT_FIXNUM_MAXIMUM;
    DM_SupportFixnum vplus1 = DM_FromSupportFixnum(DM_ToSupportFixnum(v+1));
    if ( m != v )
      DM_Abort(DMbgl_context, DM_SUPPORT_FIXNUM_MAXIMUM_PROBLEM);
    /* We should wrap-around here and if we don't then it is not
     * not the greatest Bigloo fixnum. */
    if ( vplus1 >= v )
      DM_Abort(DMbgl_context, DM_SUPPORT_FIXNUM_MAXIMUM_PROBLEM);
  }

  /* Test a little the macros of dmbgl.h 
   * ONE: test changing from/to Support. */
  { DM_Object         *o        = (DM_Object*)(DM_ReferenceToClass(Class));
    DM_SupportPointer wrapped_o = DM_ToSupport(o);
    DM_Object         *oo       = DM_FromSupport(wrapped_o);

    if ( oo != o ) {
      DM_Abort(DMbgl_context, DME_TO_FROM_PROBLEM);
    }
  }

  /* TWO: Test IsObject. */
  { DM_Object *o = DM_TheConstant(nil);

    if ( ! DM_IsObject(o) ) {
      long type;

      type = TYPE(DM_ToSupport(o));
      DM_Abort(DMbgl_context, DME_IS_OBJECT_PROBLEM);
    }
  }

  /* THREE: Test if DM_allocate_bare_instance satisfies IsObject.. 
   * This is normally the first allocation to be ever performed. */
  { DM_Object *o = DM_Allocate0(context, Object);

    if ( ! DM_IsObject(o) )
      DM_Abort(DMbgl_context, DME_ALLOCATE_PROBLEM);
  }

  /* Advise Boehm's GC that DMeroon uses interior pointers.  Also advise
   * that we may correct misalignment (see allocate_bare_instance).  */
  GC_register_displacement(DM_PrefixSize());
  GC_register_displacement((1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);
  GC_register_displacement(DM_PrefixSize()
                           + (1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);

  /* Initialize the various packages (in order).  */
  DM_scm_initialize(context);
  DM_newapi_initialize(context);

  /* Then initialize the default context and the DMbgl_context
   * variable which is used whenever a context is missing. */
  { 
    DM_Context* _default = DM_TheCurrentSite()->context;
    
    _default->handler = DM_CodeOf(DM_abort);
#ifdef DM_DEBUG
    _default->errstream = stderr; 
#else
    _default->errstream = NULL;
#endif
    DM_ClearOption(_default, DM_CONTEXT_CHECK_DMEROON);
    DM_SetOption(_default, DM_CONTEXT_CHECK_CLASS);
    DM_ClearOption(_default, DM_CONTEXT_ABORT_IMMEDIATELY);
    DMbgl_context = DM_AllocateContext(context);
  }

  /* Finally, register the Bigloo Scheme protocol. 
   * This protocol is protected by the current DM_password.  */
  DM_add_restricted_protocol(context, "SchemeScript", 
                             DM_scheme_protocol_handler);

}

/* }}}
 * end of dmbgl.c */
