/* $Id: dmbgl.h,v 1.34 1998/05/31 14:51:24 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 
 * DOCUMENTATION dmbgl.h
 * This is the Bigloo header file for DMeroon.
 * It defines how DMeroon objects are represented, accessed, managed, etc.
 * It mainly contains macros that will be used to instantiate the 
 * dmeroon.h and dm.h header files (which of course must be included after)
 * as well as the dmbgl.c file (which contains routines specific to Bigloo).
 *
 * DMeroon objects have the prefix Bigloo expects: they are Bigloo
 * values.  These objects will have a unique tag Bigloo actually
 * ignores.  This is the precise layout for DMeroon instances as seen
 * from Bigloo:
 *
 * typedef struct DMeroon_object {
 *   header_t     header;           <-- bigloo prefix      <== Bigloo address
 *   long         _padding;
 *   DM_reference _class;           <-- DMeroon prefix
 *   proxy information;
 *   blah blah blah;                <-- DMeroon content    <== DMeroon reference
 * } DMeroon_object;
 *
 * Pay attention that pointers in Bigloo designates the word containing the
 * Bigloo header while DMeroon references designate the first word of the
 * content of a Dmeroon object. The macros ToSupport and FromSupport make
 * the appropriate conversions.
 */

#ifndef DMeroon_Bigloo_H
#define DMeroon_Bigloo_H

#define DM_SUPPORT_LANGUAGE "bigloo"

/* Since this is the binding for Bigloo, include Bigloo header. 
 * This was "bigloo.h" before 1.9, this is now "bigloo1.9c.h" */
#include DM_BIGLOO_HEADER_FILE

extern obj_t strputs (char* s, obj_t port);
extern obj_t c_string_to_symbol();
extern obj_t make_real(double);

/* This is the name of the basic allocator of Bigloo. */

#define DM_MALLOC GC_MALLOC

/*
 * This is the value used in DMeroon objects header for Bigloo.
 */

#define DM_HEADER 56423

/* 
 * These are the macros that recognize a Bigloo object as a DMeroon instance.
 * DM_IsDMeroonInstance tests if a Bigloo pointer corresponds to a DMeroon
 * value while DM_IsObject tests if DMeroon value is a DMeroon object.
 */ 

#define DM_IsDMeroonInstance(p) \
  ( POINTERP(p) && (TYPE(p) == DM_HEADER) )

#define DM_IsObject(o) \
   ( ( o ) && (DM_HEADER == TYPE(DM_ToSupport(o))) )

/*
 * Here are the types of the various kinds of values handled by Bigloo.
 * DM_SupportBoolean is only for advertizing predicates.
 * DM_SupportVoid means that the returned value is unspecified.
 */

#define DM_SupportPointer  obj_t
#define DM_SupportFixnum   int_t
#define DM_SupportFloatnum double
#define DM_SupportBoolean  DM_SupportPointer
#define DM_SupportVoid     DM_SupportPointer

/* 
 * This is how Bigloo values are recognized according to their types.
 */

#define DM_IsSupportFixnum(wrapped_o) \
  INTEGERP(wrapped_o)
#define DM_IsSupportCharacter(wrapped_o) \
  CHARP(wrapped_o)
#define DM_IsSupportFloat(wrapped_o) \
  DM_IsSupportDouble(wrapped_o)
#define DM_IsSupportDouble(wrapped_o) \
  REALP(wrapped_o)
#define DM_IsSupportLongDouble(wrapped_o) \
  DM_IsSupportDouble(wrapped_o)
#define DM_IsSupportString(wrapped_o) \
  STRINGP(wrapped_o)

/* 
 * This is how Bigloo values are built.
 */

#define DM_ToSupport(dmeroon) \
  ( BREF(((((DMeroon_object*) (dmeroon))) - 1)) )
#define DM_ToSupportFixnum(int) \
  BINT(int)
#define DM_ToSupportCharacter(char) \
  BCHAR(char)
#define DM_ToSupportFloat(float) \
  DM_ToSupportDouble((double)float)
#define DM_ToSupportDouble(double) \
  DOUBLE_TO_REAL(double)
#define DM_ToSupportLongDouble(longdouble) \
  DM_ToSupportDouble((double)longdouble)

/*
 * This is how Bigloo values are decoded.
 * PROBLEM there: 
 *   Why the cascade (CREF then DMeroon) is "not computable at load time"
 *   as gcc says ???
 #define DM_FromSupport(bigloo) \
  ((void*)(((DMeroon_object*) CREF(bigloo)) + 1))
 * Since actually CREF was the identity, remove it! 
 */

#define DM_FromSupport(bigloo) \
  ((void*)((((DMeroon_object*) (bigloo))) + 1))
#define DM_FromSupportFixnum(wrapped_int) \
  CINT(wrapped_int)
#define DM_FromSupportCharacter(wrapped_char) \
  CCHAR(wrapped_char)
#define DM_FromSupportFloat(wrapped_float) \
  DM_FromSupportDouble(wrapped_float)
#define DM_FromSupportDouble(wrapped_double) \
  REAL_TO_DOUBLE(wrapped_double)
#define DM_FromSupportLongDouble(wrapped_longdouble) \
  DM_FromSupportDouble(wrapped_longdouble)
#define DM_FromSupportString(wrapped_string) \
  BSTRING_TO_STRING(wrapped_string)

/*
 * These are Bigloo representations of #f and #t.
 */

#define DM_SupportFalse BFALSE
#define DM_SupportTrue  BTRUE

/* 
 * When a C procedure has nothing better to return, it returns
 * DM_SupportDone.  Generally such a procedure has type DM_SupportVoid
 * corresponding to C void. The C void type should be avoided since it
 * makes invocation of procedures harder: it is better to always
 * return a value.
 */

#define DM_SupportDone DM_SupportTrue

/*
 * The Bigloo header occupies one word (32 or 64bits), DMeroon adds
 * two words, the alignment constraint forces us to add one word of
 * padding. To test the parameters of the DM_SupportPrefixInitialization
 * macro I choose to encode the size of the wrapped_object as the padding.
 */

#define DM_SupportPrefix() \
  header_t        _header; \
  long            _padding;

#define DM_SupportPrefixInitialization(type, varname) \
  MAKE_HEADER(DM_HEADER, sizeof(type)), \
  sizeof(type),

/*
 * Limits set by Bigloo.
 * DM_SUPPORT_FIXNUM_MAXIMUM is the most positive small integer Bigloo 
 * can represent. DM_SUPPORT_FIXNUM_MINIMUM is the least negative one
 * and must be defined as its opposite.

 * On a 32bits cpu with a 2bits tag, this yields:
 *   DM_SUPPORT_FIXNUM_MAXIMUM = 536870911
 * Observe that we are insensitive to integer representation such as
 * complement to 1 or 2 (but we loose one integer!).
 * On a 64bits cpu with a 3bits tag, this yields:
 *   DM_SUPPORT_FIXNUM_MAXIMUM = 1152921504606846976

 * This definition is correct but cannot be used in #if statement since
 * it contains casts and sizeof operators. I therefore rewrite it a little
 * to be more static: DM_SupportFixnum is int_t which is currently long.
 * This assumption is checked in DM_initialize (dmbgl.c).
 *#define DM_SUPPORT_FIXNUM_MAXIMUM \
 * ((DM_SupportFixnum) \
 * ((((DM_SupportFixnum)1)<<( (CHAR_BIT*sizeof(DM_SupportFixnum)) \
 *                           - TAG_SHIFT -1 ))-1))

 * NOTE: on alpha (a 64bit/word cpu) long takes 8 bytes so is equivalent
 * to long long on a 32bit/word cpu. Therefore this definition holds:
 */

#define DM_SUPPORT_FIXNUM_MAXIMUM \
  ((LONG_MAX)>>TAG_SHIFT)

#define DM_SUPPORT_FIXNUM_MINIMUM (-DM_SUPPORT_FIXNUM_MAXIMUM)

/*
 * Unfortunately long long are not handled by cpp except on machines with
 * 64bit/word, even Gnu's cpp does not appreciate the official definition
 * of long_long_max entities with their LL suffix. If you redefine them
 * without the LL suffix then, on 32bit/word cpus, these definitions lead
 * to too big unsigned numbers that's why I remove them.
 */

#undef LONG_LONG_MAX
#undef ULONG_LONG_MAX
/*	#define LONG_LONG_MAX  9223372036854775807	*/
/*	#define ULONG_LONG_MAX 18446744073709551615	*/

/* 
 * This is the precise function that initializes both DMeroon and the
 * binding language (here Bigloo).
 */

#define DM_INITIALIZE DM_bigloo_initialize

#endif /* DMeroon_Bigloo_H */

/* DOCUMENTATION
 * This is the second part of this header file. 
 * It defines the functions that are required by DMeroon to run.
 *
 * DM_initialize should be called before you use DMeroon (somewhere in
 *   your main).
 *
 * DM_allocate_bare_instance allocates an instance of <class> with a given
 *   size (expressed in bytes).
 *
 * DM_is_object tests if a pointer is the address of a DMeroon object.
 *
 * NOTE: These prototypes use DMeroon definitions so they must be processed
 * after all DMeroon header files.
 */

#ifdef DMeroon_SUPPORT_INCLUDED_AGAIN

extern DM_Object*
DM_allocate_bare_instance (DM_Context *context,
                           DM_Class   *class, 
                           DM_nat3    size );

extern void
DM_finalize_exit (void *wrapped_exit, void* client_data);

extern void
DM_set_exit_finalization (DM_Exit *exit);

extern void
DM_raise_exception (DM_Context    *context,
                    DM_ErrorCode  code, 
                    char          *file, 
                    unsigned long line);

extern void
DM_print_char_star (DM_Context *context,
                    char       *string,
                    DM_nat3    len,
                    void       *output_port);

extern void
DM_bigloo_initialize (DM_Context* context);

/* Extra functions for Bigloo. */

extern DM_SupportPointer 
DM_allocate (DM_SupportPointer wrapped_class, 
             DM_SupportPointer wrapped_list_of_args);

extern void GC_gc ();

/* Add as well all the functions of the generic stub. */

#include "dmscm.h"

/* This function is defined in dm.bgl */

extern obj_t DM_special_evals (obj_t bigloostring);

extern DM_Context* DMbgl_context;

#endif /*  DMeroon_SUPPORT_INCLUDED_AGAIN */

/* end of dmbgl.h */
