;;; $Id: sig2bgl.scm,v 1.33 1997/08/17 19:30:45 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; This file converts signatures into appropriate interfaces for
;;; foreign languages such as module for Bigloo.

;;;{{{ Mapping types to Bigloo

;;; Convert a type into a Bigloo type.
;;; Here we choose to let the dmscm.c generic layer to check the types and
;;; classes as well as to coerce values back and from.

(set! types-mapping
      (map (lambda (type)
             (cons (string->symbol
                    (list->string
                     (map char-downcase
                          (string->list 
                           (symbol->string type) ) ) ) )
                   (string->symbol "obj") ) )
           '(void string char fixnum floatnum anything boolean 
                  nat1 nat2 nat3 int1 int2 int3 
                  offset index options translation ) ) )

;;;}}}
;;;{{{ Main function

;;; Read signatures, produce the header of the module then its body.
;;; This is the generation required for Bigloo to interface a .c or .h
;;; file.

(define (prepare-c-4bigloo sigfile hfile bglfile)
  (complete-types-mapping)
  (let ((signatures (read-signatures-file sigfile)))
    ;; generate the .bgh file with the appropriate Bigloo headers
    (call-with-output-file hfile
      (lambda (out)
        (produce-scheme-comment "Automatically generated (Don't edit!)" out)
        (display "(directives" out)(newline out)(newline out)
        (for-each (lambda (sig)
                    (produce-scheme-comment sig out)
                    (let ((headers (produce-header sig)))
                      (for-each (lambda (h) 
                                  (write h out)
                                  (newline out) )
                                headers ) ) )
                  signatures )
        (newline out)(display ") ; end of directives" out)(newline out)
        (produce-scheme-eof out) ) )
    ;; generate a scheme file 
    (call-with-output-file bglfile
      (lambda (out)
        (produce-scheme-comment "Automatically generated (Don't edit!)" out)
        (for-each (lambda (sig)
                    (produce-scheme-comment sig out)
                    (let ((e (produce-body sig)))
                      (if e (begin 
                              (write e out)
                              (newline out) )) ) )
                  signatures )
        (produce-scheme-eof out) ) )
    'done ) )

;;;}}}
;;;{{{ Build Bigloo ex/im-port directives.

;;; In fact they generate a list of directives, one to import a C
;;; resource and another one to export it. That is one foreign
;;; directive followed by one export directive.

(define (produce-header sig)
  (if (pair? (cadr sig))
      (if (eq? (string->symbol "macro") (car (cadr sig)))
          (produce-macro-header sig)
          (produce-function-header sig) )
      (produce-constant-header sig) ) )

(define (produce-constant-header sig)
  (let ((type (car sig))
        (name (cadr sig)) )
    `((foreign 
       (macro obj ,name 
              ,(case-cs type
                 ((nat1 nat2 nat3 netnat2 netnat3
                        int1 int2 int3 
                        translation displacement index )
                  (string-append "BINT(" 
                                 (symbol->string name)
                                 ")" ) )
                 ((void string char anything boolean)
                  (sigtools-error "Does not know this type of constant" sig) )
                 (else
                  (if (memq type (map string->symbol
                                      '("Object" "Class" "Type"
                                        "Site" "Clock" "Field" ) ))
                      (string-append "DM_ToSupport(DM_FromC(&"
                                     (symbol->string name)
                                     "_object))" )
                      (sigtools-error "Unknown type constant" type) ) ) ) ) )
      (export ,(convert-name name)) ) ) )

;;; Due to problems in Bigloo with the reexportation of inline functions 
;;; imported as foreign functions, I remove all inlines.

(define (produce-function-header sig)
  (let* ((type  (car sig))
         (call  (cadr sig))
         (name  (car call))
         (types (cdr call)) )
    `((foreign (,(convert-type type) ,name
                                     ,(flat (convert-list-of-types types))
                                     ,(symbol->string name) ))
      (export (,(convert-name name) . ,types)) ) ) )

(define (produce-macro-header sig)
  (let* ((type  (car sig))
         (call  (cdr (cadr sig)))
         (name  (car call))
         (types (cdr call)) )
    `((foreign
       (macro ,(convert-type type) ,name
                                    ,(convert-list-of-types types) 
                                    ,(symbol->string name) )
       ;; This would be better but leads to a redefinition in Bigloo1.7b
       ;(define ,(convert-type type) ,(convert-macro-name name)
       ;                             ,(convert-list-of-types types) 
       ;                             ,(symbol->string name) )
       )
      ;; a foreign directive is transitive for the compiler but not
      ;; for the interpreter (not visible from repl of main.bgl)
      (export (,(convert-macro-name name) . ,types))
      ) ) )

;;;}}}
;;;{{{ Build a Scheme function invoking the C function.

;;; Convert a signature into a function that checks the type of its
;;; arguments. Return #f if there is nothing to emit.

(define (produce-body sig)
  (if (pair? (cadr sig))
      (if (eq? (string->symbol "macro") (car (cadr sig)))
          (produce-macro-body sig)
          (produce-function-body sig) )
      (produce-constant-body sig) ) )

(define (produce-constant-body sig)
  (let* ((type (car sig))
         (name (cadr sig))
         (scheme-name (convert-name name)) )
    `(define ,scheme-name ,name) ) )

;;; Generate a Scheme function calling directly the C function name.

(define (produce-function-body sig)
  (let* ((type      (car sig))
         (call      (cadr sig))
         (name      (car call))
         (types     (cdr call))
         (variables (convert-sigtypes-to-names types)) )
    `(define (,(convert-name name) . ,variables)
       (,name . ,(flat variables)) ) ) )

;;; Macros are for internal use, they do not check the type of their 
;;; arguments (to avoid infinite regression).

(define (produce-macro-body sig)
  (let* ((type      (car sig))
         (call      (cdr (cadr sig)))
         (name      (car call))
         (types     (cdr call))
         (variables (convert-sigtypes-to-names types)) )
    `(define (,(convert-macro-name name) . ,variables)
       (,name . ,variables) ) ) )

;;; }}}
;;; {{{ Obsolete parts

;;; Transform the result of a foreign function.
;;; In fact accept it as it is.

(define (coerce-from from exp)
  exp )

;;; Due to the possible dotted argument list, define a special mapper
;;; and take care of the possibly dotted variable. In fact this only
;;; occurs for the signature of DM_allocate in Bigloo.

(define (map-coerce-to types variables)
  (if (pair? types)
      (cons (coerce-to (car types) (car variables))
            (map-coerce-to (cdr types) (cdr variables)) )
      (if (null? types)
          '() 
          (list variables) ) ) )
  
;;; Check an argument held in variable VAR to be of DMeroon type TYPE.

(define (coerce-to type var)
  (if (DMeroon-class-type? type)
      (if (eq? type (string->symbol "Object"))
          `(if (DM_is_object ,var)
               ,var
               (DM-error "Wrong type: should be" ',type ,var) )
          `(if (and (DM_is_object ,var)
                    (DM_is_a ,var ,(symbol-concatenate "DM-" type "-class")) )
               ,var
               (DM-error "Wrong type: should be" ',type ,var) ) )
      (case-cs type
        ((anything stream void) var)
        ((string) `(if (string? ,var) 
                       ,var 
                       (DM-error "Not a string" ,var) ))
        ((char) `(if (char? ,var) 
                     ,var
                     (DM-error "Not a character" ,var) ))
        ((boolean) `(if (boolean? ,var)
                        ,var
                        (DM-error "Not a boolean" ,var) ))
        ((vector) `(if (vector? ,var)
                       ,var
                       (DM-error "Not a vector" ,var) ))
        ((fixnum) `(if (integer? ,var) 
                       ,var
                       (DM-error "Not a fixnum" ,var) ))
        ((floatnum) `(if (real? ,var) 
                         ,var
                         (DM-error "Not a floatnum" ,var) ))
        ((nat1 index)
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   (<= ,var 255) )
              ,var
              (DM-error "Not a nat1" ,var) ))
        ((nat2 netnat2 translation)
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   (<= ,var 65535) )
              ,var
              (DM-error "Not a nat2" ,var) ))
        ((nat3 netnat3 options offset displacement)
         ;; ATTENTION, bigloo on 32bit word machines, read 4294967295
         ;; as -1. In that case, avoid the check.
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   (if (< 0 4294967295)
                       (<= ,var 4294967295)
                       #t ) )
              ,var
              (DM-error "Not a nat3" ,var) ))
        ((int1)
         `(if (and (integer? ,var)
                   (>= ,var -127)
                   (<= ,var +127) )
              ,var
              (DM-error "Not a int1" ,var) ))
        ((int2)
         `(if (and (integer? ,var)
                   (>= ,var -32767)
                   (<= ,var +32767) )
              ,var
              (DM-error "Not a int2" ,var) ))
        ((int3)
         ;; ATTENTION, bigloo on 32bit word machines, read +/-2147483648
         ;; as 0. In that case avoid the check.
         `(if (and (integer? ,var)
                   (if (< -2147483648 0) (>= ,var -2147483648) #t)
                   (if (< 0 +2147483648) (<= ,var +2147483648) #t) )
              ,var
              (DM-error "Not a int3" ,var) ))
        (else (sigtools-error "Cannot coerce to " type)) ) ) )

;;; }}}
;;; {{{ Special names

(set! special-names 
      ;; from DMeroon/c/dmscm.h
      (list (cons (string->symbol "DMscm_is_object")
                  (string->symbol "DM-is-object?" ) )
            (cons (string->symbol "DMscm_initialize")
                  (string->symbol "DM-initialize!" ) )
            (cons (string->symbol "DMscm_is_a")
                  (string->symbol "DM-is-a?" ) )
            (cons (string->symbol "DMscm_is_a_subclass")
                  (string->symbol "DM-is-a-subclass?" ) )
            (cons (string->symbol "DMscm_class_of")
                  (string->symbol "DM-class-of" ) )
            (cons (string->symbol "DMscm_site_of")
                  (string->symbol "DM-site-of" ) )
            (cons (string->symbol "DMscm_clock_of")
                  (string->symbol "DM-clock-of" ) )
            (cons (string->symbol "DMscm_receive")
                  (string->symbol "DM-receive" ) )
            (cons (string->symbol "DMscm_send")
                  (string->symbol "DM-send" ) )
            (cons (string->symbol "DMscm_connect")
                  (string->symbol "DM-connect" ) )
            (cons (string->symbol "DMscm_create_subclass")
                  (string->symbol "DM-create-subclass" ) )
            (cons (string->symbol "DMscm_smash_class")
                  (string->symbol "DM-smash-class" ) )
            (cons (string->symbol "DMscm_set_site_of")
                  (string->symbol "DM-set-site-of!" ) )
            (cons (string->symbol "DMscm_set_clock_of")
                  (string->symbol "DM-set-clock-of!" ) )
            (cons (string->symbol "DMscm_set_sharable")
                  (string->symbol "DM-set-sharable!" ) )
            (cons (string->symbol "DMscm_set_copyable")
                  (string->symbol "DM-set-copyable!" ) )
            (cons (string->symbol "DMscm_set_initialized")
                  (string->symbol "DM-set-initialized!" ) )
            (cons (string->symbol "DMscm_is_initialized")
                  (string->symbol "DM-is-initialized?" ) )
            (cons (string->symbol "DMscm_eq")
                  (string->symbol "DM-eq?" ) )
            (cons (string->symbol "DMscm_field_value")
                  (string->symbol "DM-field-value" ) )
            (cons (string->symbol "DMscm_indexed_field_value")
                  (string->symbol "DM-indexed-field-value" ) )
            (cons (string->symbol "DMscm_set_field_value")
                  (string->symbol "DM-set-field-value!" ) )
            (cons (string->symbol "DMscm_set_indexed_field_value")
                  (string->symbol "DM-set-indexed-field-value!" ) )
            (cons (string->symbol "DMscm_indexed_field_length")
                  (string->symbol "DM-indexed-field-length" ) )
            (cons (string->symbol "DMscm_field_item_type")
                  (string->symbol "DM-field-item-type" ) )
            (cons (string->symbol "DMscm_field_repetition_type")
                  (string->symbol "DM-field-repetition-type" ) )
            (cons (string->symbol "DMscm_serve")
                  (string->symbol "DM-serve" ) )
            ;; from Bigloo/dmbgl.c
            (cons (string->symbol "DMbgl_allocate")
                  (string->symbol "DM-allocate" ) )
            ) )

;;; }}}
;;; end of sig2bgl.scm
