;;; $Id: DMeroonet.scm,v 1.8 1997/08/17 19:30:45 queinnec Exp $

;;; This files defines a small object system compatible with Meroon but
;;; using DMeroon objects. It mainly offers generic functions over DMeroon.

;;;==================================================================
;;; DMeroonet does not restrict the number of classes that can be created.
;;; They are held in a global variable.

(define *classes* '())

;;; Converts a symbol to a Class with that name.

(define (->Class name)
  (let lookup ((l *classes*))
    (if (pair? l)
        (if (DM-str-of-cmp name (car l) <Class.name>)
            (car l)
            (lookup (cdr l)) )
        #f ) ) )

;;; DMeroonet does not restrict the number of generic functions that can
;;; be created. They are held in a global variable.

(define *generics* '())

;;; Converts a symbol to a generic instance with such a name. Returns
;;; the first one ie the newest one.

(define (->Generic name)
  (let lookup ((l *generics*))
    (if (pair? l)
        (if (eq? name (Generic-name (car l)))
            (car l)
            (lookup (cdr l)) )
        #f ) ) )

;;; This variable holds the last class defined at macroexpand-time. It
;;; avoids to rebuild the class at evaluation-time if immediately
;;; after as under a toplevel loop. It is useful for separate
;;; compilation (see the define-class macro).

(define *last-defined-class* #f)

;;;==================================================================
;;; Take a list of symbols and return a symbol with the concatenated
;;; name.  Useful to create derived names.

(define (symbol-concatenate . names)
  (string->symbol (apply string-append (map symbol->string names))) )

;;;==================================================== Class definition
;;; Syntax: (define-class <name-of-the-class> <name-of-the-super-class>
;;;             ( <name-of-a-Mono-Field>          |
;;;               (= <name-of-a-Mono-Field> ...)  |
;;;               (* <name-of-a-Poly-Field> ...)
;;;               ... ) )

;;; The macro call is supposed to be correct: it does not check that
;;; name is a correct name, that super-name is a correct name that
;;; names a currently existing class. The strategy is still to create
;;; the Class instance and to ask it to generate its equivalent code
;;; as the result of the macro call. Were Class-generate-related-names be
;;; a generic function as in Meroon V3, this would allow to define new
;;; metaclass with alternate expansion.

(define-icslas-macro (define-class name super-name
                       own-field-descriptions . ignore-class-options )
  (set! *last-defined-class* #f)
  (let ((class (register-class name super-name own-field-descriptions)))
    (set! *last-defined-class* class)
    `(begin
       (if (not *last-defined-class*)
         (register-class ',name ',super-name ',own-field-descriptions) )
      ,(Class-generate-related-names class) ) ) )

;;; Generate from a class all the definitions of the related functions.
;;; As in Meroon V3 the Class instance is stored in a global variable
;;; but this time, its name is surrounded with < and > as in Dylan.

(define (Class-generate-related-names class)
  (let* ((name (Class-name class))
         (class-variable-name (symbol-concatenate '< name '>))
         (predicate-name (symbol-concatenate name '?))
         (maker-name (symbol-concatenate 'make- name))
         (allocator-name (symbol-concatenate 'allocate- name)) )
    `(begin
       (define ,class-variable-name (->Class ',name))
       (define ,predicate-name (make-predicate ,class-variable-name))
       (define ,maker-name (make-maker ,class-variable-name))
       (define ,allocator-name (make-allocator ,class-variable-name))
       ,@(map (lambda (field) (Field-generate-related-names field class))
              (Class-fields class) )
       ',(Class-name class) ) ) )

;;; Generate from a field all the definitions of the related accessors.
;;; An instance of Field refers to the class that introduced it.

;;; Names are constructed with the current class rather than the
;;; defining class, that is why the class is also given to this
;;; function. The lengther is also generated in case of a Poly-Field.

(define (Field-generate-related-names field class)
  (let* ((fname (Field-name field))
         (cname (Class-name class))
         (cname-variable (symbol-concatenate '< cname '>))
         (reader-name (symbol-concatenate cname '- fname))
         (writer-name (symbol-concatenate 'set- cname '- fname '!)) )
    `(begin
       (define ,reader-name
         (make-reader
          (retrieve-named-field ,cname-variable ',fname) ) )
       (define ,writer-name
         (make-writer
          (retrieve-named-field ,cname-variable ',fname) ) )
       ,@(if (DM-is-field-indexed field)
             `((define ,(symbol-concatenate cname '- fname '-length)
                 (make-lengther
                  (retrieve-named-field ,cname-variable ',fname) ) ))
             '() ) ) ) )

;;; Convert a name and a class into a field.

(define (retrieve-named-field class name)
  (->Field class name) )

;;;==================================================== Predicate
;;; Create from a class its appropriate recognizer.

(define (make-predicate class)
  (lambda (o) (and (DM-is-object o)
                   (DM-is-a o class) )) )

;;; Check if an object is an instance of a class. If you appreciate a
;;; dangerous life just nullify it!

(define (check-class-membership o class)
  (if (not (and (DM-is-object o) (DM-is-a o class)))
    (DMeroonet-error "Wrong class" o class) ) )

;;;==================================================== Allocation
;;; Create from a class its appropriate allocator. Sizes must be a
;;; list of positive integers, fields must only be Mono- or Poly-Fields.

(define (make-allocator class)
  (lambda sizes
    (apply DM-allocate class sizes) ) )

(define (make-maker class)
  (lambda arguments
    (apply DM-make class arguments) ) )

;;;==================================================== Access

(define (make-reader field)
  (if (DM-is-field-indexed field)
      (lambda (o) (DM-field-value o field))
      (lambda (o i) (DM-indexed-field-value o field i)) ) )

(define (make-writer field)
  (if (DM-is-field-indexed field)
      (lambda (o v) (DM-set-field-value! o field v))
      (lambda (o i v) (DM-set-indexed-field-value! o v field i)) ) )

(define (make-lengther field)
  (lambda (o) (DM-indexed-field-length o field)) )

;;;==================================================== Class creation
;;; Create a class instance and graft it into the inheritance hierarchy.

(define (register-class name super-name own-field-descriptions)
  (apply DM-create-class name 
         (->Class super-name)
         (length own-field-descriptions)
         (parse-field-descriptions own-field-descriptions) ) )

;;; Parse field descriptions and convert them into something suitable to
;;; DM-create-class.

(define (parse-field-descriptions descs)
  (if (pair? descs)
      (if (pair? (car descs))
          (let* ((desc (car descs))
                 (type (car desc))
                 (name (cadr desc)) )
            (case type
              ((=) `("-field" ,(symbol->string name) "-mutable" 
                              . ,(parse-field-descriptions (cdr descs)) ) )
              ((*) `("-field" ,(symbol->string name) "-mutable" 
                              "-indexed" DM-nat3-type
                              . ,(parse-field-descriptions (cdr descs)) ) ) ) )
          `("-field" ,(symbol->string (car descs)) "-mutable"
            . ,(parse-field-descriptions (cdr descs)) ) )
      '() ) )

;;;==================================================== Generic functions
;;; A generic function is represented by a closure that encloses an
;;; instance of Generic which itself contains a dispatch table. This
;;; table is a vector containing methods that are retrieved using the
;;; number associated to class. 

(define-icslas-macro (define-generic call . body)
  (parse-variable-specifications
   (cdr call)
   (lambda (discriminant variables)
     (let ((generic (gensym)))          ; make generic hygienic
       `(define ,(car call)
          (let ((,generic
                 (register-generic 
                  ',(car call)
                  (lambda ,(flat-variables variables)
                    ,(if (pair? body)
                         `(begin . ,body)
                         `(DMeroonet-error 
                           "No method" ',(car call)
                           . ,(flat-variables variables) ) ) )
                  ',(cdr call) ) ))
            (lambda ,(flat-variables variables)
              ((determine-method (Generic-dispatch-table ,generic)
                                 (DM-is-object->class ,(car discriminant)) )
               . ,(flat-variables variables) ) ) ) ) ) ) ) )

;;; Create and register a generic function.

(define (register-generic generic-name default signature)
  (let ((generic (make-Generic generic-name
                               default
                               signature )) )
    (set! *generics* (cons generic *generics*))
    generic ) )

;;; Parse a list of specifications, extract all the variables and the
;;; discriminating variable among them. 
;;; Restriction: Silently ignore if there is more than one
;;; discriminant variable ie multi-methods are not implemented.

(define (parse-variable-specifications specifications k)
  (if (pair? specifications)
      (parse-variable-specifications
       (cdr specifications)
       (lambda (discriminant variables)
         (if (pair? (car specifications))
             (k (car specifications) 
                (cons (caar specifications) variables) )
             (k discriminant (cons (car specifications) variables)) ) ) )
      (k #f specifications) ) )

;;; Flatten a list of variables to avoid a final dotted variable. This
;;; is necessary since all methods (even the default method) are
;;; represented by closures with a fix arity.

(define (flat-variables variables)
  (if (pair? variables)
      (cons (car variables) (flat-variables (cdr variables)))
      (if (null? variables) variables (list variables)) ) )

;;;==================================================== Method creation
;;; Syntax: (define-method ( <name-of-the-generic-function
;;;                          . <description-of-the-variables> )
;;;              <body-of-the-method> )
;;; <description-of-the-variables> 
;;;     ::= ( <variable-name> . <description-of-the-variables> )
;;;      |  ( ( <variable-name> <class-name> ) 
;;;           . <description-of-the-variables> )
;;;      |  <variable-name>
;;; The discriminating variable appears with a class.

;;; Define a method on a generic function. The dispatch table to
;;; update is found via the name of the generic function in the list
;;; of all generic instances. 

(define-icslas-macro (define-method call . body)
  (parse-variable-specifications
   (cdr call)
   (lambda (discriminant variables)
     (let ((g (gensym))(c (gensym)))    ; make g and c hygienic
       `(set! ,(car call)
              (register-method
               ',(car call)
               (lambda (,g ,c)
                 (let ((supermethod           ;; PB sur Object
                        (determine-method
                         (Generic-dispatch-table ,g)
                         (DM-ClassToSuper ,c (- (DM-ClassToSuperLength ,c) 2)) ) ))
                   (lambda ,(flat-variables variables)
                     (define (call-next-method)
                       (supermethod . ,(flat-variables variables)) )
                     . ,body ) ) )
               ',(cadr discriminant)
               ',(cdr call) ) ) ) ) ) )

(define (register-method generic-name pre-method class-name signature)
  (let* ((generic        (->Generic generic-name))
         (class          (->Class class-name))
         (new-method     (pre-method generic class))
         (dispatch-table (Generic-dispatch-table generic)) )
    (check-signature-compatibility generic signature)
    (set-Generic-dispatch-table! 
     generic (cons (cons class new-method) dispatch-table) )
    #t ) )

;;; Check signature compatibility between a generic function and a method.
;;; The two signatures can have both a dotted variable.

(define (check-signature-compatibility generic signature)
  (define (coherent-signatures? la lb)
    (if (pair? la)
        (if (pair? lb)
            (and (or ;; similar discriminating variable
                     (and (pair? (car la)) (pair? (car lb)))
                     ;; similar regular variable
                     (and (symbol? (car la)) (symbol? (car lb))) )
                 (coherent-signatures? (cdr la) (cdr lb)) )
            #f )
        (or (and (null? la) (null? lb))
            ;; similar dotted variable
            (and (symbol? la) (symbol? lb)) ) ) )
  (if (not (coherent-signatures? (Generic-signature generic) signature))
    (meroonet-error "Incompatible signatures" generic signature) ) )

;;; Determine the method to apply in a generic function. 
;;; The dispatch-table is a list of ( (class . method) ... . default )

(define (determine-method dispatchers class)
  (let lookup ((l dispatchers))
    (if (pair? l)
        (if (DM-is-a-subclass class (caar l))
            (cdar l)
            (lookup (cdr l)) )
        l ) ) )
;;;   FUTURE ordonner les methodes par profondeur decroissante    FUTURE

;;; Generic functions.

(define-generic (show (o) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display o stream) ) )

(define-generic (clone (o))
  (DM-clone o) )

;;; end of DMeroonet.scm
