;;; $Id: except.scm,v 1.3 1995/06/24 14:15:17 queinnec Exp $

;;; This file defines the possible exceptions that might be raised
;;; during the normal course of the compilers. They serve to customize 
;;; these compilers wrt generic functions.

;;;=================================================================
;;; Exceptions

(define-class Exception Object () :virtual)

;;; This is the exception raised by the objectify function when
;;; encountering an unknown global variable. The handler if returning
;;; should return a Reference. More precisely it can abort or adjoin
;;; this new global variable to the global mutable environment.

(define-class Unknown-Variable Exception
  ( name           ; the name of the unknown variable (a symbol)
    r              ; the current lexical environment
    g              ; the global environment
    ) )

;;; Two other exceptions exist for errors or warnings. They may be 
;;; signalled during compilation or evaluation. 

;;; This exception is used whenever an exceptional situation is found
;;; during the compilation but the situation is not exceptional enough
;;; to deserve a dedicated exception.

(define-class Regular-Exception Exception
  ( phase          ; the name of the phase of the compilation that raised the
                   ; the exception (a string).
    message        ; a string that explains the problem.
    culprits       ; a list of additional values detailing the exception.
    ) :virtual )

;;; This is the class of the exceptions that are raised by the underlying
;;; runtime.

(define-class Underlying-Exception Regular-Exception ())

(define-class Compilation-Exception Regular-Exception ())

;;; Sometimes it is useful to report some hints to the user. Warnings
;;; are modelized by continuable exceptions.

(define-class Compilation-Warning Regular-Exception ())

;;; This is an even lighter warning just to make the user sure that
;;; something is happening.

(define-class Compilation-Report Compilation-Warning ())

;;;=================================================================
;;; Exception raisers.

;;; New unknown global variables are signalled, the handler is to be
;;; provided by the caller of objectify. If the handler returns it
;;; must return a reference. The handler may also refuse to create the
;;; global variable or create it on the fly.

(define (objectify-free-global-reference variable r g)
  (diagnose (make-Unknown-Variable variable r g) 
            #t ) )

;;; This function builds functions to signal non continuable exceptions.

(define (create-compilation-exception-raiser phase)
  (lambda (msg . culprits)
    (diagnose (make-Compilation-Exception phase msg culprits)
              #f ) ) )

;;; This function creates functions to signal warnings ie continuable
;;; exceptions.

(define (create-compilation-warning-raiser phase)
  (lambda (msg . hints)
    (diagnose (make-Compilation-Warning phase msg hints)
              #t ) ) )

;;; Report progress to the user.

(define (report-progress msg . hints)
  (diagnose (make-Compilation-Report "Progress report" msg hints) #t) )

;;;==================================================================
;;; Many handlers are similar, they print the error and exit or print
;;; the warning and continue. They differ generally on how to cope
;;; with unknown global variables and that's all. So it is useful to
;;; provide a generic handler that does just that so you can link your
;;; own handler onto it.

(define-generic (basic-handler (exc) c?))

(define-method (basic-handler (exc Compilation-Warning) c?)
  (display (Compilation-Warning-phase exc) stderr)
  (display " Warning: " stderr)
  (display (Regular-Exception-message exc) stderr)
  (newline stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (newline stderr) )
            (Regular-Exception-culprits exc) )
  ;; return to the diagnoser
  #t )

(define-method (basic-handler (exc Compilation-Report) c?)
  (display (Regular-Exception-message exc) stderr)
  (display " " stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (display " " stderr) )
            (Regular-Exception-culprits exc) )
  (newline stderr)
  ;; return to the diagnoser
  #t )

(define-method (basic-handler (exc Regular-Exception) c?)
  (display (Regular-Exception-message exc) stderr)
  (newline stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (newline stderr) )
            (Regular-Exception-culprits exc) ) )

(define-method (basic-handler (exc Underlying-Exception) c?)
  (display "   *** *** *** Underlying Exception *** *** ***" stderr)
  (newline stderr)
  ;; display the other fields of the exception with the method on
  ;; Regular-Exception (just above).
  (call-next-method) )

(define-method (basic-handler (exc Compilation-Exception) c?)
  (display "   *** *** *** " stderr)
  (display (Compilation-Exception-phase exc) stderr)
  (display " Exception *** *** ***" stderr)
  (newline stderr)
  ;; display the other fields of the exception with the method on
  ;; Regular-Exception (just above).
  (call-next-method) )

;;; end of excpt.scm
