<;;; $Id: dm2ch.scm,v 1.39 1998/06/10 16:31:25 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; DOCUMENTATION dm2ch.scm
;;; This script must be run by a case-sensitive Scheme implementation.
;;; It takes a .dm file defining classes, instances and function signatures 
;;; and converts it into a .c and a .h files. 

;;; Restriction: after an indexed field, only indexed fields can appear.
;;; Fortunately, this is checked by this processor.

;;; NOTE: This file uses filutl.scm so load it before. It also uses
;;; the case-sensitive reader from TOP/Library/read.scm.   

;;;==================================================================
;;;            Utilities

;;; maps a predicate on alist and tells if at least one element answers true. 

(define (any? p . args)
  (let ormap ((args args) (value #f))
    (if (let any-at-end? ((ls args))
          (and (pair? ls)
               (or (not (pair? (car ls)))
                   (any-at-end? (cdr ls)))))
        value
        (let ((value (apply p (map car args))))
          (or value (ormap (map cdr args) value))))))

;;; concatenate all sort of names.

(define (symbol-concatenate . names)
  (string->symbol
   (apply string-append
          (map (lambda (thing)
                 (cond ((symbol? thing)
                        (symbol->string thing) )
                       ((string? thing) thing)
                       ((number? thing)
                        (number->string thing) )
                       (else (process-error
                              "non coercible to String"
                              thing )) ) )
               names ) ) ) )

;;; Display a list of terms with a prefix, some separator and a final suffix. 

(define (display-list out list beg mid end)
  (display beg out)
  (if (pair? list)
      (let pr ((list list))
        (display (car list) out)
        (if (pair? (cdr list))
            (begin (display mid out)
                   (pr (cdr list)) ) ) ) )
  (display end out) )

(define (caddddr x) (caddr (cddr x)))

;;; Returns l1 less the terms from l2.

(define (list-difference l1 l2)
  (if (pair? l1)
      (if (member (car l1) l2)
          (list-difference (cdr l1) l2)
          (cons (car l1) (list-difference (cdr l1) l2)) )
      '() ) )

;;;==================================================================

;;; DOCUMENTATION process-file
;;; This function converts a .dm file into a .c and a .h file. 
;;; First it reads all the definitions of classes that are contained in the
;;; context files then it converts the basefile.

(define (process-file basefile . context-files)
  (let* ((exps    (file->list (string-append basefile ".dm") read-cs))
         (cfile   (string-append basefile ".c"))
         (hfile   (string-append basefile ".h"))
         (shortname (extract-filename-sans-extension
                     (extract-basefilename basefile) ))
         (macname (string-append "DM_" shortname "_H")) )
    ;; Load context files.
    (for-each 
     (lambda (file)
       (process #f #f (file->list (string-append file ".dm") read-cs)) )
     context-files )
    (call-with-output-file cfile
      (lambda (cout)
        (display "/* Automatically generated (Don't touch) */" cout)
        (newline cout)
        (newline cout)
        (display "#include \"dmeroon.h\"" cout)
        (newline cout)
        (display "#include \"" cout)
        (display shortname cout)
        (display ".h\"" cout)
        (newline cout)
        (call-with-output-file hfile
          (lambda (hout)
            (display "/* Automatically generated (Don't touch) */" hout)
            (newline hout)
            (newline hout)
            (display "#ifndef " hout)
            (display macname hout)
            (newline hout)
            (display "#define " hout)
            (display macname hout)
            (newline hout)
            ;; The real work
            (process cout hout exps)
            ;; end of real work
            (newline hout)
            (newline hout)
            (display "#endif /* " hout)
            (display macname hout)
            (display " */" hout)
            (newline hout)
            (display "/* eof */" hout)
            (newline hout) ) )
        (newline cout)
        (display "/* eof */" cout)
        (newline cout) ) ) )
  basefile )

;;; Process every expression and output their translation onto cout and hout.
;;; Expressions may define classes, instances or signatures.

(define (process cout hout exps)
  (for-each 
   (lambda (exp)
     (if (pair? exp)
         (begin
           ;;(display `(processing ,exp))(newline)
           (case-cs (car exp)
             ((define-class)
              (apply process-define-class cout hout (cdr exp)) )
             (else (process-error "Unknown form" exp)) ) )
         (process-error "Unknown form" exp)) )
   exps ) )

;;;==================================================================
;;; Process a class definition.
;;; The *classes* variable accumulates all classes definitions to manage
;;; inheritance. Every entry defines a class like that.
;;;     (name superclassname properfieldastheyare allfields)

(define (make-class name superclass-name proper-fields all-fields options mcn)
  (vector '--class-- name superclass-name proper-fields 
          all-fields options mcn ) )
(define (check-class c)
  (if (and (vector? c)
           (= (vector-length c) 7)
           (eq? (vector-ref c 0) '--class--) )
      #t
      (process-error "Internal error, not a class" c) ) )
(define (class-name c)
  (check-class c)
  (vector-ref c 1) )
(define (class-superclass-name c)
  (check-class c)
  (vector-ref c 2) )
(define (class-proper-field-descriptions c)
  (check-class c)
  (vector-ref c 3) )
(define (class-fields c)
  (check-class c)
  (vector-ref c 4) )
(define (class-options c)
  (check-class c)
  (vector-ref c 5) )
(define (class-metaclass-name c)
  (check-class c)
  (vector-ref c 6) )
(define (class-proper-fields class)
  (let ((all-fields (class-fields class)))
    (list-tail all-fields
               (- (length all-fields)
                  (length (class-proper-field-descriptions class)) ) ) ) )

;;; Look for a class in a list of vectors.

(define (find-class name lv)
  (if (pair? lv)
      (if (eq? name (class-name (car lv)))
          (car lv)
          (find-class name (cdr lv)) )
      #f ) )

(define *classes* 
  (list (make-class (string->symbol "Object") 
                    #f '() '() '() 
                    (string->symbol "Class") )) )
        
;;; For a class, compute the list of its superclasses including itself.

(define (compute-super-classes classname)
  (let ((super (find-class classname *classes*)))
    (if (not super) (process-error "Missing class" classname))
    (if (eq? classname (string->symbol "Object"))
        (list super)
        (cons super 
              (compute-super-classes (class-superclass-name super)) ) ) ) )

;;; Retrieve the class that introduced a field.

(define (class-introducing-field class field)
  (let scan ((supers (reverse (compute-super-classes (class-name class)))))
    (if (pair? supers)
        (if (member field (class-fields (car supers)))
            (car supers)
            (scan (cdr supers)) )
        (process-error "Cannot find introducing class" class field) ) ) )

;;; returns the sublist of indexed fields.

(define (filter-indexed-fields fields)
  (if (pair? fields)
      (if (field-reptyp-name (car fields))
          (cons (car fields) (filter-indexed-fields (cdr fields)))
          (filter-indexed-fields (cdr fields)) )
      '() ) )

(define (make-field name itemtyp-name reptyp-name options referring)
  (vector '--field-- name itemtyp-name reptyp-name options referring) )
(define (check-field f)
  (if (and (vector? f)
           (= (vector-length f) 6)
           (eq? (vector-ref f 0) '--field--) )
      #t
      (process-error "Internal error: not a field" f) ) )
(define (field-name f)
  (check-field f)
  (vector-ref f 1) )
(define (field-itemtyp-name f)
  (check-field f)
  (vector-ref f 2) )
(define (field-reptyp-name f)
  (check-field f)
  (vector-ref f 3) )
(define (field-options f)
  (check-field f)
  (vector-ref f 4) )
(define (field-referring f)
  (check-field f)
  (vector-ref f 5) )

(define (find-metaclass-name options)
  (define metaclass-kw (string->symbol ":metaclass"))
  (let search ((options options))
    (if (pair? options)
        (if (eq? metaclass-kw (car options))
            (if (pair? (cdr options))
                (cadr options)
                (process-error "Metaclass name missing") )
            (search (cdr options)) )
        (string->symbol "Class") ) ) )

(define (remove-metaclass-options options)
  (define metaclass-kw (string->symbol ":metaclass"))
  (let search ((options options))
    (if (pair? options)
        (if (eq? metaclass-kw (car options))
            (if (pair? (cdr options))
                (remove-metaclass-options (cddr options))
                (process-error "Metaclass name missing") )
            (cons (car options) (search (cdr options))) )
        options ) ) )

(define (process-define-class cout hout name supername fields . options)
  (if *verbose* (begin (display "[")(display name)(display "]")(newline)))
  (if (find-class name *classes*)
      (process-error "Class redefinition" name) )
  (let ((super (find-class supername *classes*)))
    (if (not super) (process-error "Unnown superclass" supername))
    (let* ((all-fields (append (class-fields super) 
                               (map parse-field fields) ))
           (mcn (find-metaclass-name options))
           (class (make-class
                   name supername fields all-fields 
                   (compute-deduced-options
                    (make-class name supername fields all-fields options mcn) )
                   mcn ))
           (complete? #t) )
      (if (and (not (find-class mcn *classes*))
               (not (equal? mcn (string->symbol "Class"))) )
          (process-error "Unknown metaclass" mcn) )
      (set! *classes* (cons class *classes*))
      (set! complete? (check-class-restriction! class))
      (cond 
       ((and cout hout)
        ;; generate structures, extern declaration and macro in the H file
        (if (not complete?) (generate-struct-warning hout class))
        (generate-struct hout class)
        (generate-extern hout class)
        (generate-struct-generator hout class)
        ;; generate objects in the C file
        (generate-signatures cout class)
        (generate-proper-field-objects cout class)
        (generate-documentation cout class)
        (generate-proper-class-object cout class) )
       (else 
        ;; No generation for context files.
        'nothing ) ) ) )
  name )

(define (compute-deduced-options class)
  (let* ((fields (class-fields class))
         (options (class-options class))
         (indexed-fields (filter-indexed-fields (class-fields class)))
         (other-options '()) )
    (if (> (length indexed-fields) 1)
        (set! other-options (cons (string->symbol ":stringent")
                                  other-options )) )
    ;; a mutable field implies that the class is mutable.
    (for-each 
     (lambda (field)
       (if (memq (string->symbol ":mutable") (field-options field))
           (if (not (memq (string->symbol ":mutable") other-options))
               (set! other-options (cons (string->symbol ":mutable")
                                         other-options )) ) ) )
     fields )
    ;; remove metaclass option if present
    (set! options (remove-metaclass-options options))
    ;; merge deduced and other options
    (append other-options options) ) )

;;; Currently, after an indexed field, the generated C structure will
;;; not show the additional fields so it is necessary to emit a warning.

(define (check-class-restriction! class)
  (define result #t)
  (let check ((fields (class-fields class)))
    (if (pair? fields)
        (if (field-reptyp-name (car fields))
            (or (let all-indexed ((fields (cdr fields)))
                  (if (pair? fields)
                      (begin 
                        (set! result #f)
                        (if (field-reptyp-name (car fields))
                            (all-indexed (cdr fields))
                            #f ) )
                      #t ) )
                (let all-regular ((fields (cdr fields)))
                  (if (pair? fields)
                      (if (not (field-reptyp-name (car fields)))
                          (all-regular (cdr fields))
                          (begin
                            ;(warn "Mix of regular and indexed fields after indexed field")
                            (set! result #f) ) )
                      #t ) ) )
            (check (cdr fields)) ) ) )
  result )

;;; Parse and check the definition of a field.
;;;  field ::= ( fieldname :type typename 
;;;                        [:indexing-type typename ]
;;;                    [ :mutable ]
;;;                    [ :local ]
;;;                    [ :volatile ]
;;;		       [ :presend ] )

(define (parse-field field)
  (let* ((name    #f)
         (itemtyp #f)
         (reptyp  #f)
         (options '())
         (referring #f) )
    (if (not (pair? field)) (process-error "Missing field name" field))
    (if (symbol? (car field))
        (set! name (car field))
        (process-error "Invalid field name" (car field)) )
    (let scan ((field (cdr field)))
      (if (pair? field)
          (case-cs (car field)
            ((:type)
             (if itemtyp (process-error "Multiple :type" field))
             (if (not (pair? (cdr field))) 
                 (process-error "Missing type" field) )
             (set! itemtyp (cadr field))
             (if (and (not (eq? itemtyp (string->symbol "reference")))
                      referring )
                 (process-error "Not a reference field" name) )
             (scan (cddr field)) )

            ((:indexing-type)
             (if reptyp (process-error "Multiple :indexing-type" field))
             (if (not (pair? (cdr field)))
                 (process-error "Missing indexing type" field) )
             (set! reptyp (cadr field))
             (scan (cddr field)) )

            ((:refer)
             (if referring (process-error "Multiple :refer" field))
             (if (not (pair? (cdr field)))
                 (process-error "Missing referring class" field) )
             (if (and itemtyp
                      (not (eq? itemtyp (string->symbol "reference"))) )
                 (process-error "Not a reference field" name) )
             (set! referring (cadr field))
             (scan (cddr field)) )

            ((:volatile :local :mutable :presend :secret)
             (set! options (append options (list (car field))))
             (scan (cdr field)) )

            (else (process-error "Unknown field option" field)) ) ) )
    (if itemtyp
        (begin
          ;; By default the referring type is Object.
          (if (eq? itemtyp (string->symbol "reference"))
              (if (not referring)
                  (set! referring (string->symbol "Object")) ) )
          (make-field name itemtyp reptyp options referring) )
        (process-error "Type not specified" field) ) ) )

;;;==================================================================
;;; Generate the signatures of these objects in the .sig file.
;;; something like:
;;;     (Class DM_Point_class)
;;;     (Field DM_Point_x_field)

(define (generate-signatures sout class) 
  (newline sout)
  (display "/* SIGNATURES:" sout)
  (newline sout)
  (let ((sc `(,(string->symbol "Class")
              ,(symbol-concatenate "DM_" (class-name class) "_class") )))
    (display "   " sout)
    (display sc sout)
    (newline sout)
    (for-each (lambda (field)
                (let ((sf `(,(string->symbol "Field")
                            ,(symbol-concatenate 
                              "DM_" (class-name class) "_"
                              (field-name field) "_field" ))))
                  (display "   " sout)
                  (display sf sout)
                  (newline sout) ) )
              (class-proper-fields class) ) )
  (newline sout)
  (display " */" sout)
  (newline sout)
  (newline sout) )

;;;==================================================================
;;; Generate a macro to allocate instances with precise indexed field
;;; lengths.  ATTENTION: it is not possible to use that macro with a
;;; null size, C does not allow something like item[0]. I finally
;;; understood why: item[1] starts an array with some padding to force
;;; the alignment; item[0] is similar to nothing that does not force
;;; realignment!

;;; Generates something like:
;;;    #define DM_GeneratePointStruct(sizes) \
;;;    ... item[sizes] ...

(define (generate-struct-generator hout class)
  (let* ((name (class-name class))
         (indexed-fields 
          (let filter ((r (map field-reptyp-name (class-fields class))))
            (if (pair? r)
                (if (car r)
                    (cons (car r) (filter (cdr r)))
                    (filter (cdr r)) )
                '() ) ) ) )
    (if (pair? indexed-fields)
        (let ((indexes #f))
          (do ((i 0 (+ i 1))
               (result '())
               (indexed-fields indexed-fields (cdr indexed-fields)) )
              ((null? indexed-fields)
               (set! indexes (reverse result)) )
            (set! result
                  (cons (string->symbol
                         (string-append 
                          "_" (number->string i) ) )
                        result ) ) )
          (newline hout)
          (display "#define DM_Generate" hout)
          (display name hout)
          (display-list hout indexes "Struct(" "," ") \\")
          (newline hout)
          (display "struct DM_" hout)
          (display name hout)
          (display-list hout indexes "_##" "##_##" " { \\")
          (let gen ((indexes               indexes)
                    (indexed-field-before? #f)
                    (fields                (class-fields class)) )
            (if (pair? fields)
                (let ((field (car fields)))
                  (newline hout)
                  (if (field-reptyp-name field)
                      (cond 
                       (indexed-field-before?
                        (display " DM_RealignToMostStringent( \\" hout)
                        (newline hout)
                        (generate-special-indexed-struct-field
                         hout field (car indexes) )
                        (display " \\" hout)
                        (gen (cdr indexes) #t (cdr fields))
                        (newline hout)
                        (display " ); \\" hout) )
                       (else
                        (generate-special-indexed-struct-field 
                         hout field (car indexes) )
                        (display " \\" hout)
                        (gen (cdr indexes) #t (cdr fields)) ) )
                      (cond 
                       (indexed-field-before?
                        (display " DM_RealignToMostStringent( \\" hout)
                        (newline hout)
                        (generate-regular-struct-field hout field)
                        (display " \\" hout)
                        (gen indexes #t (cdr fields))
                        (newline hout)
                        (display " ); \\" hout) )
                       (else
                        (generate-regular-struct-field hout field)
                        (display " \\" hout)
                        (gen indexes indexed-field-before? (cdr fields))
                        ) ) ) ) ) )
          (newline hout)
          (display "}" hout)
          (newline hout) ) ) ) )

(define (generate-special-indexed-struct-field hout field index)
  (let ((name    (field-name field))
        (itemtyp (field-itemtyp-name field))
        (reptyp  (field-reptyp-name field)) )
    (case-cs reptyp
      ((nat1 nat2 nat3)
       (display "  DM_" hout)
       (display reptyp hout)
       (display " " hout)
       (display name hout)
       (display "_length; \\" hout)
       (newline hout)
       (display "  DM_" hout)
       (display itemtyp hout)
       (display " " hout)
       (display name hout)
       (display "[" hout)
       (display index hout)
       (display "];" hout) )
      (else (process-error "Unknown indexing type" reptyp)) ) ) )  

;;;==================================================================
;;; Generate the extern directives exporting the objects representing
;;; the class and its fields. Something like:
;;;     DM_DefineWrappedType(DM_Point);
;;;     extern DM_Class_3_2_6_wrapped DM_Point_class_object;
;;;     extern DM_Field_1_3_wrapped DM_Point_xy_field_object;

(define (generate-extern hout class)
  (let ((name   (class-name class))
        (fields (class-fields class)) )
    (newline hout)
    (begin
      (display "DM_DefineWrappedType(DM_" hout)
      (display name hout)
      (display ");" hout)
      (newline hout)
      (newline hout) )
    (let ((supers (compute-super-classes name)))
      (generate-specialized-class hout class)
      ;; export the class object
      (display "extern DM_" hout)
      (display (class-metaclass-name class) hout)
      (display "_" hout)
      (display (length supers) hout)
      (display "_" hout)
      (display (length (class-fields (car supers))) hout)
      (display "_" hout)
      (display (+ 1 (string-length (symbol->string name))) hout)
      (display "_wrapped DM_" hout)
      (display name hout)
      (display "_class_object;" hout)
      (newline hout)
      (newline hout)
      ;; export its proper fields only
      (let* ((all-fields    (class-fields (car supers)))
             (proper-fields (class-proper-fields (car supers)))
             (inherited-fields (list-difference all-fields proper-fields)) )
        (do ((offsets-number (+ 1 (length (filter-indexed-fields 
                                           inherited-fields ))))
             (i              0             (+ i 1))
             (fields         proper-fields (cdr fields)) )
            ((null? fields))
          (let ((field (car fields)))
            (generate-specialized-field hout field offsets-number)
            (display "extern DM_Field_" hout)
            (display offsets-number hout)
            (display "_" hout)
            (display (+ 1 (string-length (symbol->string (field-name field))))
                     hout )
            (display "_wrapped DM_" hout)
            (display name hout)
            (display "_" hout)
            (display (field-name field) hout)
            (display "_field_object;" hout)
            (newline hout)
            (case-cs (field-reptyp-name field)
               ((nat1 nat2 nat3 nat4)
                (set! offsets-number (+ offsets-number 1)) )
               (else 'nothing) ) ) ) ) ) ) )

;;; Generate automatically the correct specialized metaclasses for C.
;;; Protect these definitions from double definition with a macro.

(define *already-generated-specialized-classes* '())

(define (generate-specialized-class hout class)
  (let* ((supers (compute-super-classes (class-name class)))
         (mcn (class-metaclass-name class))
         (n1 (length supers))
         (n2 (length (class-fields (car supers))))
         (n3 (+ 1 (string-length (symbol->string (class-name class))))) )
    (if (member (list n1 n2 n3 mcn) *already-generated-specialized-classes*)
        #t
        (let ((macname (symbol-concatenate 
                        "DM_Specialized" mcn "_" n1 "_" n2 "_" n3)))
          (display "#ifndef " hout)
          (display macname hout)
          (newline hout)
          (display "#define " hout)
          (display macname hout)
          (newline hout)          
          ;; Define the appropriate class structure
          (if (= 0 n2)
              (display "DM_DefineVerySpecialized" hout)
              (display "DM_DefineSpecialized" hout) )
          (display mcn hout)
          (display "(" hout)
          (display n1 hout)
          (display "," hout)
          (display n2 hout)
          (display "," hout)
          (display n3 hout)
          (display ");" hout)
          (newline hout)
          (display "#endif" hout)
          (newline hout)
          (newline hout)
          (set! *already-generated-specialized-classes*
                (cons (list n1 n2 n3 mcn)
                      *already-generated-specialized-classes* ) ) ) ) ) )

(define *already-generated-specialized-fields* '())

(define (generate-specialized-field hout field offsets-number)
  (let ((sl (+ 1 (string-length (symbol->string (field-name field))))))
    (if (member (cons offsets-number sl)
                *already-generated-specialized-fields* )
        #t
        (let* ((macname (symbol-concatenate 
                         "DM_SpecializedField_" offsets-number
                         "_" sl )) )
          (newline hout)
          (display "#ifndef " hout)
          (display macname hout)
          (newline hout)
          (display "#define " hout)
          (display macname hout)
          (newline hout) 
          (display "DM_DefineSpecializedField(" hout)
          (display offsets-number hout)
          (display "," hout)
          (display sl hout)
          (display ");" hout)
          (newline hout)
          (display "#endif" hout)
          (newline hout)
          (newline hout)
          (set! *already-generated-specialized-fields*
                (cons (cons offsets-number sl)
                      *already-generated-specialized-fields* ) ) ) ) ) )

;;;==================================================================
;;; Generate the typedef'ed struct corresponding to the class.
;;; Something like:
;;;     typedef struct DM_Point {
;;;       DM_reference x;
;;;     } DM_Point;

(define (generate-struct hout class)
  (let ((name   (class-name class))
        (fields (class-fields class)) )
    (if (pair? fields)
        (begin
          (newline hout)
          (display "typedef struct DM_" hout)
          (display name hout)
          (display " {" hout)
          (newline hout)
          (for-each 
           (let ((indexed-field-before? #f))
             (lambda (field)
               (if indexed-field-before?
                   'dont-output-anything
                   (begin
                     (if (field-reptyp-name field)
                         (set! indexed-field-before? #t) )
                     (generate-struct-field hout field) ) ) ) )
           fields )
          (display "} DM_" hout)
          (display name hout)
          (display ";" hout)
          (newline hout) )
        ;; The class has no field, it is the equivalent to Object and 
        ;; has the same definition Object has in dmeroon.h [This is 
        ;; necessary to allow "struct DM_Value*" to be legal in C.]
        (begin
          (newline hout)
          (display "typedef struct DM_" hout)
          (display name hout)
          (display " {" hout)
          (newline hout)
          (display "  char _void;" hout)
          (newline hout)
          (display "} DM_" hout)
          (display name hout)
          (display ";" hout)
          (newline hout) ) ) ) )

;;; Generate a warning for incomplete C structures. That is classes
;;; with fields following an indexed field. They cannot be translated
;;; into static C.

(define (generate-struct-warning hout class)
  (newline hout)
  (display "/* WARNING: the following C structure is incomplete since" hout)
  (newline hout)
  (display " * all fields after the first indexed field are missing." hout)
  (newline hout)
  (display " */" hout)
  (newline hout) )

;;; Generate code to define a C struct.

(define (generate-struct-field hout field)
  (if (field-reptyp-name field)
      (generate-indexed-struct-field hout field)
      (generate-regular-struct-field hout field) )
  (newline hout) )

(define (generate-regular-struct-field hout field)
  (let ((name    (field-name field))
        (itemtyp (field-itemtyp-name field)) )
    (case-cs itemtyp
      ((data code handler_code char1 offset 
        date key class_options field_options port_options site_options
        external_options context_options file_options request_options
        nat1 nat2 nat3 nat4 int1 int2 int3 int4
        netnat2 netnat3 float1 float2 float3 )
       (display "  DM_" hout)
       (display itemtyp hout)
       (display " " hout)
       (display name hout) )
      ((reference)
       (display "  struct DM_" hout)
       (display (field-referring field) hout)
       (display "* " hout)
       (display name hout) )
      ((code handler_code)
       (display " void* (*" hout)
       (display name hout)
       (display ")()" hout) )
      (else (process-error "Unknown type" itemtyp)) )
    (display ";" hout) ) )

(define (generate-indexed-struct-field hout field)
  (let ((name    (field-name field))
        (itemtyp (field-itemtyp-name field))
        (reptyp  (field-reptyp-name field)) )
    (case-cs reptyp
      ((nat1 nat2 nat3)
       (display "  DM_" hout)
       (display reptyp hout)
       (display " " hout)
       (display name hout)
       (display "_length;" hout)
       (newline hout)
       (case-cs itemtyp
         ((data code handler_code char1 offset 
           date key class_options field_options port_options site_options
           external_options context_options file_options request_options
           nat1 nat2 nat3 nat4 int1 int2 int3 int4
           netnat2 netnat3 float1 float2 float3 )
          (display "  DM_" hout)
          (display itemtyp hout)
          (display " " hout)
          (display name hout) )
         ((reference)
          (display "  struct DM_" hout)
          (display (field-referring field) hout)
          (display "* " hout)
          (display name hout) )
         ((code)
          (display " void* (*" hout)
          (display name hout)
          (display ")()" hout) )
         (else (process-error "Unknown type" itemtyp)) )
       (display "[1];" hout) )
      (else (process-error "Unknown indexing type" reptyp)) ) ) )

;;;==================================================================
;;; Generates some documentations in the C file that may be extracted.

(define (generate-documentation cout class)
  (newline cout)
  (display "/* RESOURCE " cout)
  (display (class-name class) cout)
  (newline cout)
  (let ((supers (compute-super-classes (class-name class))))
    (if (cdr supers)
        (begin (display "This class inherits from " cout)
               (display (class-name (cadr supers)) cout)
               (display "." cout)
               (newline cout) ) ) )
  (display " */" cout)
  (newline cout) )

;;;==================================================================
;;; Generate the fields objects in the C file.
;;; Something like:
;;;    DM_Field_1_4_wrapped DM_Point_xyz_field_object = {
;;;      ...
;;;    };

(define (generate-proper-field-objects cout class)
  (let ((indexed-before 0))
    (let pr ((fields (class-fields class)))
      (if (pair? fields)
          (begin
            (if (eq? class (class-introducing-field class (car fields)))
                (generate-proper-field-object 
                 cout class (car fields) indexed-before ) )
            (if (field-reptyp-name (car fields))
                (set! indexed-before (+ indexed-before 1)) )
            (pr (cdr fields)) ) ) ) ) )

(define (generate-proper-field-object cout class field indexed-before)
  (newline cout)
  (let ((metafield (symbol-concatenate
                    "DM_Field_" 
                    (+ 1 indexed-before) 
                    "_"
                    (+ 1 (string-length (symbol->string (field-name field))))
                    "_wrapped" ))
        (varname (symbol-concatenate
                  "DM_" (class-name class) "_"
                  (field-name field) "_field_object" )) )
    (display metafield cout)
    (display " " cout)
    (display varname cout)
    (display " = {" cout)
    (newline cout)
    (display "  DM_SupportPrefixInitialization(" cout)
    (display metafield cout)
    (display "," cout)
    (display varname cout)
    (display ")" cout)
    (newline cout)
    (display "  DM_ReferenceToClass(Field)," cout)
    (newline cout)
    (display "  DM_ProxyInitialization()," cout)
    (newline cout)
    (display "  { { {" cout)
    (newline cout)
    (display-field-options cout field)
    (newline cout)
    (display "    DM_ReferenceToClass(" cout)
    (display (class-name class) cout)
    (display ")," cout)
    (newline cout)
    ;; referring type is meaningful for references only
    (display "    " cout)
    (if (field-referring field)
        (begin (display "DM_ReferenceToClass(" cout)
               (display (field-referring field) cout)
               (display ")" cout) )
        (display "DM_NULL" cout) )
    (display "," cout)
    (newline cout)
    ;; number of offsets
    (display "    " cout)
    (display (+ 1 indexed-before) cout)
    (display "," cout)
    (newline cout)
    (display "    {" cout)
    (newline cout)
    (display-field-offsets cout class field)
    (display "    }," cout)
    (newline cout)
    ;; Realign for the name
    (display "    { {" cout)
    (newline cout)
    (display "    " cout)
    (display (+ (string-length (symbol->string (field-name field))) 1) cout)
    (display "," cout)
    (newline cout)
    (display "    " cout)
    (write (symbol->string (field-name field)) cout)
    (newline cout)
    (display "    } }" cout)
    (newline cout)
    (display "  } } }" cout)
    (newline cout)
    (display "};" cout)
    (newline cout) ) )

;;; This function outputs the offsets. It is this function that
;;; forbids classes to have regular fields after an indexed field.

(define (display-field-offsets cout class field)
  (let ((first? #t))
    (let pr ((fields (class-fields class)))
      (if (pair? fields)
          (if (equal? field (car fields))
              (display-field-offset cout class field first?)
              (begin 
                (if (field-reptyp-name (car fields))
                    (begin
                      (display-field-offset cout class (car fields) first?)
                      (set! first? #f) ) )
                (pr (cdr fields)) ) ) ) ) ) )

(define (display-field-offset cout class field first?)
  (cond
   (first?
    (display "      DM_PackOffset(DM_OffsetOf(DM_" cout)
    (display (class-name class) cout)
    (display "," cout)
    (display (field-name field) cout)
    (if (field-reptyp-name field)
        (display "_length" cout) )
    (display ")," cout) )
   (else (display "      DM_PackOffset(0," cout)) )
  (newline cout)
  (display "                    DM_" cout)
  (display (or (field-reptyp-name field) "nought") cout)
  (display "_index," cout)
  (newline cout)
  (display "                    DM_" cout)
  (display (field-itemtyp-name field) cout)
  (display "_index )," cout)
  (newline cout) )

(define (display-field-options cout field)
  (display "    " cout)
  (let* ((options (field-options field))
         (coptions (map (lambda (option)
                          (string->symbol
                           (case-cs option
                              ((:mutable)  "DM_FIELD_IS_MUTABLE")
                              ((:local)    "DM_FIELD_IS_LOCAL")
                              ((:volatile) "DM_FIELD_IS_VOLATILE")
                              ((:presend)  "DM_FIELD_IS_PRE_SENDABLE")
                              ((:secret)   "DM_FIELD_IS_SECRET")
                              (else (process-error "Unknown field option"
                                                   option )) ) ) )
                        options )) )
    (if (pair? coptions)
        (display-list cout coptions "" "|" "")
        (display "DM_FIELD_DEFAULT_OPTIONS" cout) )
    (if (field-reptyp-name field)
        (display "|DM_FIELD_IS_INDEXED" cout) )
    (display "," cout) ) )

;;;==================================================================
;;; Generate the class object in the C file.

(define (generate-proper-class-object cout class)
  (newline cout)
  (let* ((name      (class-name class))
         (supers    (compute-super-classes name))
         (classvar  (symbol-concatenate
                     "DM_" name "_class_object" ))
         (mcn       (class-metaclass-name class))
         (metaclass-c-type (symbol-concatenate
                            "DM_" mcn "_"
                            (length supers) "_"
                            (length (class-fields (car supers))) "_"
                            (+ 1 (string-length (symbol->string name)))
                            "_wrapped" )) )
    (display metaclass-c-type cout)
    (display " " cout)
    (display classvar cout)
    (display " = {" cout)
    (newline cout)
    (display "  DM_SupportPrefixInitialization(" cout)
    (display metaclass-c-type cout)
    (display "," cout)
    (display classvar cout)
    (display ")" cout)
    (newline cout)
    (display "  DM_ReferenceToClass(" cout)
    (display mcn cout)
    (display ")," cout)
    (newline cout)
    (display "  DM_ProxyInitialization()," cout)
    (newline cout)
    (display "  { { {" cout)
    (newline cout)
    (display "    " cout)
    (display-class-option cout class)
    (newline cout)
    (display "    " cout)
    (display (length supers) cout)
    (display "," cout)
    (newline cout)
    (display "    {" cout)
    (for-each (lambda (class)
                (newline cout)
                (display "      DM_ReferenceToClass(" cout)
                (display (class-name class) cout)
                (display ")," cout) )
              (reverse supers) )
    (newline cout)
    (display "    }," cout)
    (newline cout)
    (display "    { {" cout)
    (newline cout)
    (display "      " cout)
    (display (length (class-fields (car supers))) cout)
    (display "," cout)
    (display-field-references cout class)
    (newline cout)
    ;; Realign for the name
    (display "      { {" cout)
    (newline cout)
    (display "        " cout)
    (display (+ (string-length (symbol->string (class-name class))) 1) cout)
    (display "," cout)
    (newline cout)
    (display "        " cout)
    (write (symbol->string (class-name class)) cout)
    (newline cout)
    (display "      } }" cout)
    (newline cout)
    (display "    } }" cout)
    (newline cout)
    (display "  } } }" cout)
    (newline cout)
    (display "};" cout)
    (newline cout) ) )    

(define (display-class-option cout class)
  (let ((options (class-options class)))
    (if (pair? options)
        (let ((coptions 
               (map (lambda (option)
                      (string->symbol
                       (case-cs option
                          ((:immotile)  "DM_CLASS_IS_IMMOTILE")
                          ((:stringent) "DM_CLASS_IS_STRINGENT")
                          ((:mutable)   "DM_CLASS_IS_MUTABLE")
                          ((:volatile)  "DM_CLASS_IS_VOLATILE")
                          ((:copyable)  "DM_CLASS_IS_COPYABLE")
                          ((:uninstantiable) "DM_CLASS_IS_NOT_INSTANTIABLE")
                          ((:unsubclassable) "DM_CLASS_IS_NOT_SUBCLASSABLE")
                          (else (process-error "Unknown class option"
                                               option )) ) ) )
                    options )))
          (display-list cout coptions 
                        "" "|" "|DM_PREDEFINED_CLASS_DEFAULT_OPTIONS" ) )
        (display "DM_PREDEFINED_CLASS_DEFAULT_OPTIONS" cout) )
    (if (pair? (filter-indexed-fields (class-fields class)))
        (display "|DM_CLASS_IS_INDEXED" cout) )
    (display "," cout) ) )

(define (display-field-references cout class)
  (let ((fields (class-fields class)))
    (if (pair? fields)
        (begin 
          (newline cout)
          (display "      {" cout)
          (newline cout)
          (let pr ((fields fields))
            (if (pair? fields)
                (begin 
                  (display "        DM_ReferenceTo(DM_" cout)
                  (display (class-name (class-introducing-field 
                                        class (car fields) ))
                           cout )
                  (display "_" cout)
                  (display (field-name (car fields)) cout)
                  (display "_field)," cout) 
                  (newline cout)
                  (pr (cdr fields)) ) ) )
          (display "      }," cout) ) ) ) )          

;;;==================================================================
;;; Report error

(define (process-error msg . hints)
  (newline)
  (display "*** *** *** *** *** ERROR *** *** *** *** ***")
  (newline)
  (display msg)
  (newline)
  (display hints)
  (newline)
  (error) )


;;; Test with elk: (load 'trace.scm)
;;; Test: (begin (load "Bootstrap/dm2ch.scm") (process-file "DMeroon/c/dmstruct"))


;;; end of dm2ch.scm
