;;; $Id: expfil.scm,v 1.7 1995/07/29 18:02:28 queinnec Exp $

;;; This file when interpreted by (nearly) any Scheme system processes
;;; the source files of Icslas and DMeroon to remove nearly all Icslas
;;; and DMeroon macros. In order not to restrain compilers, Scheme
;;; syntax is respected (let, let*, do ... are left in place). This is
;;; an example of a neutral expansion.

;;; The exact work is to expand all files so that the resulting files
;;; may appear as regular Icslas files. All macros are removed and
;;; particularly the macros that conditionalize the sources
;;; (if-icslas-feature) except the `monitor' macro that corresponds to
;;; a regular special form in Icslas. Concerning functions, the
;;; conditionalization is done with respect to Icslas ie
;;; strict-r4rs. This is why functions such as atom? are kept.

;;; Caution, this file is evaluated in the TARGET directory!
;;; NOTE: this file must not use macros at all to be portable.

;;; Load the configuration parameters. This configuration file is 
;;; generated in the TARGET directory.

(load "config.scm")

;;;==================================================================
;;; You can use any Scheme implementation you want but it should offer
;;; an eval function. 

(define (neutral-expand-eval e)
  ;; (display `(about to eval ,e)) ; DEBUG
  (eval e) )

;;; Since it is not the case that any Scheme implementation have an
;;; eval feature, here is a portable simulation:

;(define *transfer-variable* #f)

;(define (neutral-expand-eval e)
;  (define *temp-file* "/tmp/Eval.scm")
;  (call-with-output-file *temp-file*
;    (lambda (out)
;      (write `(set! *transfer-variable* ,e) out)
;      (newline out) ) )
;  (load *temp-file*)
;  *transfer-variable* )

;;;==================================================================
;;; Load the neutral expansion engine.

(load (string-append TOP "/Macros/clexpand.scm"))

;;; Adjoin the define-icslas-macro to this engine.
;;; NOTE: Same code as in the beginning of Macros/clmacros.scm

(define (special-walk-define-icslas-macro e w)
  (let* ((call      (cadr e))
         (name      (car call))
         (variables (cdr call))
         (w         (extend-with-variables w variables))
         (body      (walk* (cddr e) w)) )
    ;; report that this macro will be removed.
    (display `(removing ,name))
    ;; Build and evaluate the expander on the fly.    
    (let* ((expander `(lambda (macroexpand . ,variables) . ,body))
           (expanded-expander (walk expander w)) )
      (set! remove-icslas-and-dmeroon-macros-environment
            (extend-with-repetitive-macro 
             remove-icslas-and-dmeroon-macros-environment
             name 
             (neutral-expand-eval expanded-expander) ) )
      ;; Nothing is left after this definition.
      `(quote ,name) ) ) )

(define (remove-icslas-and-dmeroon-macros-environment e)
  (if (pair? e)
      (case (car e)
        ((define-icslas-macro) special-walk-define-icslas-macro)
        (else (neutral-walking-env e)) )
      (neutral-walking-env e) ) )

;;; This function is to be called from process-icslas-sources.
;;; It takes a filename, read then expand its content, write it in 
;;; a new file. The file is expanded as in:
;;;     TOP/d1/d2/f.scm ---> (TARGET/)d1/d2/f.scm
;;; NOTE: the target directories must exist before.
;;; NOTE: This file is evaluated in the TARGET directory.
;;; Since Meroonet is written in a (bad) style (ie in one file), it is 
;;; necessary to load Meroonet in order to expand it. So first we expand
;;; Meroonet then we load it then we process the other files.

(define (neutral-expand-file basefile eval?)
  ;; Read a whole file into a list of expressions.
  ;; This definition is local to avoid redefining it when expanding
  ;; the Bootstrap/filutl.scm file.
  (define (file->list filename)
    (call-with-input-file filename
      (lambda (in)
        (let loop ((e (read in)))
          (if (eof-object? e)
              '()
              (cons e (loop (read in))) ) ) ) ) )
  (let* ((e*      (file->list (string-append TOP "/" basefile ".scm")))
         (newfile (string-append basefile ".SCM")) )
    (call-with-output-file newfile
      (lambda (out)
        (display ";;; automatically expanded (Don't touch!)" out)
        (newline out)
        (newline out)
        (for-each
         (lambda (e) 
           (let ((ne (walk e remove-icslas-and-dmeroon-macros-environment)))
             (write-flattening-begin-forms ne out)
             ;; evaluate on the fly and prints a dot to show some progress.
             (if eval? (neutral-expand-eval ne))
             (display ".") ) )
         e* )
        (display ";;; eof." out)
        (newline out) ) ) ) )

;;; Many macros generate (begin . forms), flatten them while printed.
;;; Suppress also useless quotations.

(define (write-flattening-begin-forms e out)
  (cond ((and (pair? e) (eq? (car e) 'begin))
         (for-each (lambda (ne)
                     (write-flattening-begin-forms ne out) )
                   (cdr e) ) )
        ((or (and (pair? e) (eq? (car e) 'quote))
             (boolean? e) (number? e) )
         ;; don't print ie remove that form.
         #t )
        ((and (pair? e) (eq? (car e) 'define))
         (write e out)
         (newline out)
         (newline out) )
        ;; Don't want to see the value of these forms (while interpreting):
        (else (write `(begin ,e #t) out)
              (newline out)
              (newline out) ) ) )

;;;==================================================================
;;; Load the processor of Icslas sources.

(load (string-append TOP "/icslas.scm"))

;;; Load the precise features describing what must be generated. This
;;; file should have already be generated in the TARGET/Bootstrap
;;; directory.

(load "Bootstrap/features.scm")

;;;==================================================================
;;; Everything is ready: process every source files (due to
;;; process-icslas-sources, they are processed in order which is very
;;; convenient for state-macros). Before that expand and eval on the fly
;;; the Meroonet system.

(define (work)
  (neutral-expand-file "Bind/meroonet" #t)
  (process-icslas-sources
   icslas-features
   (lambda (basefile)
     (newline)
     (display "expanding ")
     (display basefile)
     (neutral-expand-file basefile #f) ) )
  (display "all files expanded.")
  (newline) )

;;; Test: (work)

;;; end of expfil.scm
