;;; $Id: filutl.scm,v 1.4 1998/06/10 16:31:25 queinnec Exp $

;;; This file defines some utilities to handle filenames, read or
;;; write them.

;;;==================================================================
;;; Manipulate file components.
;;; directory component is everything before the last /

(define (extract-directory file)
  (let ((len (string-length file))
        (lastslash 0) )
    (do ((i 0 (+ i 1)))
        ((>= i len) (substring file 0 lastslash)) 
      (if (char=? #\/ (string-ref file i))
          (set! lastslash i) ) ) ) )
;;; Test: (extract-directory "foo.scm")
;;;       (extract-directory "bar/foo.scm")
;;;       (extract-directory "/bar/hux/foo.scm")
;;;       (extract-directory "~/../foo/bar/foo.scm")
;;;       (extract-directory "~/")
;;;       (extract-directory "")

;;; filename-sans-extension is everything before the last dot.

(define (extract-filename-sans-extension file)
  (let* ((len (string-length file))
         (bfile (extract-basefilename file))
         (blen (string-length bfile))
         (found #f)
         (lastdot 0) )
    (do ((i 0 (+ i 1)))
        ((>= i blen) 
         (substring file 0 (if found (+ (- len blen) lastdot) len)) )
      (if (char=? #\. (string-ref bfile i))
          (begin 
            (set! found #t)
            (set! lastdot i) ) ) ) ) )
;;; Test: (extract-filename-sans-extension "foo.scm")
;;;       (extract-filename-sans-extension "bar/foo.scm")
;;;       (extract-filename-sans-extension "/bar/hux/foo.scm")
;;;       (extract-filename-sans-extension "/bar/hux/foo.scm.gz")
;;;       (extract-filename-sans-extension "~/../foo/bar/foo.scm")
;;;       (extract-filename-sans-extension "~/../foo/bar.foo/scm")
;;;       (extract-filename-sans-extension "~/")
;;;       (extract-filename-sans-extension "")

;;; basefilename is everything after the last /

(define (extract-basefilename file)
  (let ((len (string-length file))
        (lastslash 0) )
    (do ((i 0 (+ i 1)))
        ((>= i len) (substring file lastslash len)) 
      (if (char=? #\/ (string-ref file i))
          (set! lastslash (+ i 1)) ) ) ) )
;;; Test: (extract-basefilename "foo.scm")
;;;       (extract-basefilename "bar/foo.scm")
;;;       (extract-basefilename "/bar/hux/foo.scm")
;;;       (extract-basefilename "~/../foo/bar/foo.scm")
;;;       (extract-basefilename "~/")
;;;       (extract-basefilename "")

;;; Read a whole file into a list of expressions.
;;; The read function to use may be given, by default, this is read
;;; which is case-insensitive. 

(define (file->list filename . reader)
  (let ((reader (if (pair? reader) (car reader) read)))
    (call-with-input-file filename
      (lambda (in)
        (let loop ()
          (let ((e (reader in)))
            (if (eof-object? e)
                '()
                (cons e (loop)) ) ) ) ) ) ) )

;;; Print a series of expressions into a file

(define (list->file expressions filename)
  (call-with-output-file filename
    (lambda (out)
      (for-each (lambda (e)
                  (write e out)
                  (newline out) )
                expressions )
      (newline out) ) ) )

;;; This case-cs macro introduces case-sensitivity for quoted symbols.
;;; This is useful when using a case-sensitive reader.
;;; CAUTION: this macro may not be understood by some implementation
;;; due to the use of define-macro. Gambit does not like it.

(define-macro (case-cs exp . clauses)
  `(case ,exp
     . ,(map (lambda (clause)
               (if (pair? (car clause))
                   `((,@(map (lambda (s)
                               (string->symbol
                                (list->string
                                 (map char-upcase
                                      (string->list 
                                       (symbol->string s) ) ) ) ) )
                             (car clause) )
                      ,@(map (lambda (s)
                               (string->symbol
                                (list->string
                                 (map char-downcase
                                      (string->list 
                                       (symbol->string s) ) ) ) ) )
                             (car clause) ) ) . ,(cdr clause) )
                   clause ) )
             clauses ) ) )

;;; end of filutl.scm
