;;; $Id: BCappl.scm,v 1.11 1995/07/24 18:35:15 queinnec Exp $

;;; This file implements bytecoded object files, how they are stored,
;;; written, read, loaded and run.

;;;==================================================================
;;; This set of functions depends on the format used to record
;;; bytecoded files and on the structure of BCRT-Modules.

;;; Store the result of a Byte-Compilation into a file so it can be
;;; read with file->BCRT-Module.

(define (Byte-Compilation->file bc target)
  (report-progress "Recording bytecoded file" target)
  (truncate-bytecode-vector! bc)
  (call-with-output-file target
    (lambda (out)
      (display ";;; Name" out)
      (newline out)
      (write (Byte-Compilation-name bc) out)
      (newline out)
      (newline out)
      (display ";;; Global variable names" out)
      (newline out)
      (write (reverse! (map (lambda (gv)
                              (symbol->string (Variable-name gv)) )
                            (Global-Environment-global 
                             (Byte-Compilation-global bc) ) ))
             out )
      (newline out)
      (newline out)
      (display ";;; Quotations" out)
      (newline out)
      (write (reverse! (map (lambda (qv)
                              (BC-Quotation-variable-value qv) )
                            (Byte-Compilation-quotations bc) ))
             out )
      (newline out)
      (newline out)
      (display ";;; Bytecode " out)
      (display (list (vector-length (Byte-Compilation-bytes bc))) out)
      (newline out)
      (write (Byte-Compilation-bytes bc) out)
      (newline out) ) ) )

;;; Read a compiled module from the file system.

(define (file->BCRT-Module filename)
  (call-with-input-file filename
    (lambda (in)
      (let* ((name              (read in))
             (varnames          (read in))
             (quotations        (read in))
             (bytes             (read in)) )
        (create-BCRT-Module name varnames quotations bytes) ) ) ) )

(define (create-BCRT-Module name varnames quotations bytes)
  (let* ((varnames-length   (length varnames))
         (quotations-length (length quotations))
         (bytes-length      (vector-length bytes))
         (bc                (allocate-BCRT-Module 
                             bytes-length
                             quotations-length
                             varnames-length )) )
    (set-BCRT-Module-name! bc (convert-to-quotation name))
    (do ((i 0 (+ i 1)))
        ((>= i bytes-length))
      (set-BCRT-Module-byte! bc i (vector-ref bytes i)) )
    (do ((i 0 (+ i 1))
         (quotations quotations (cdr quotations)) )
        ((>= i quotations-length))
      (set-BCRT-Module-quotation! 
       bc i (convert-to-quotation (car quotations)) ) )
    (do ((i 0 (+ i 1))
         (varnames varnames (cdr varnames)) )
        ((>= i varnames-length))
      (set-BCRT-Module-varname! bc i (convert-to-quotation (car varnames))) )
    bc ) )             

;;; Convert a Byte-Compilation object to a BCRT-Module.  This
;;; conversion is done directly in memory and do not require to write
;;; in a file and to reread it. Otherwise this function should be
;;; equivalent to Byte-Compilation->file composed with file->BCRT-Module.

(define (Byte-Compilation->BCRT-Module bc)
  (truncate-bytecode-vector! bc)
  (create-BCRT-Module 
   (Byte-Compilation-name bc)
   (reverse! (map (lambda (gv)
                    (symbol->string (Variable-name gv)) )
                  (Global-Environment-global 
                   (Byte-Compilation-global bc) ) ))
   (reverse! (map (lambda (qv)
                    (BC-Quotation-variable-value qv) )
                  (Byte-Compilation-quotations bc) ))
   (Byte-Compilation-bytes bc) ) )

;;;==================================================================
;;; Install a BCRT-module in the toplevel of the user requesting it.
;;; This mainly consists to instantiate the module ie to allocate a
;;; global environment for its global mutable variables, to link them
;;; with the user toplevel so boxes are shared. 
;;; The varnames in Module only serve to link the module with the user
;;; toplevel.

;;;                                     a Module
;;;   bc -----------------------------> +----------+ <----*module*
;;;                         +------->   | bytecode |
;;; *genv*----+             |           |  ...     | <---*pc*
;;;           V             |           | quote0   |
;;;         ModuleEnv       |           | quote1   |
;;;  me --> +-------+       |           |   ...    |
;;;         |   ------------+           | varname0 |
;;;         | Box0  |                   | varname1 |
;;;         | Box1  |                   |   ...    |
;;;         |  ...  |                   +----------+
;;;         +-------+                    a User
;;;                        *user*----> +----------+
;;;                                    | name     |
;;;                                    | toplevel-----+
;;;                                    +----------+   |
;;;                                                   |
;;;                                                   V  a Named-Env
;;;                                        +-----------------+
;;;                                        |varname0 --> Box0|
;;;                                        |     ...         |
;;;                                        +-----------------+

(define (install-module! named-env bm)
  (report-progress "Installing module"
                   (BCRT-Module-name bm)
                   "..." )
  (let* ((varname-length (BCRT-Module-varname-length bm))
         (me (allocate-BCRT-Module-Environment varname-length)) )
    (set-BCRT-Module-Environment-module! me bm)
    (do ((i 0 (+ i 1)))
        ((>= i varname-length))
      (set-BCRT-Module-Environment-element! 
       me i (find-associated-box named-env (BCRT-Module-varname bm i)) ) )
    me ) )

;;; The linking engine.
;;; It maps global variable names to shared boxes.
;;; A Named-Env represents a flat name space of all global variables 
;;; a user may have.

(define (find-associated-box named-env varname)
  (let ((v (assoc varname (BCRT-Named-Env-mapping named-env))))
    (if (pair? v) (cdr v)
        (let ((box (make-BCRT-Named-Box BCRT-uninitialized
                                        (convert-to-quotation varname) )))
          (set-BCRT-Named-Env-mapping!
           named-env (cons (cons varname box) 
                           (BCRT-Named-Env-mapping named-env)) )
          box ) ) ) )

;;;==================================================================
;;; Run an installed module.

(define (BCRT-run-module me)
  (report-progress "Evaluating module" 
                   (BCRT-Module-name
                    (BCRT-Module-Environment-module me) )                     
                   "..." )
  (if *debug*
      (begin (display "---------Code----------" stderr)
             (newline stderr)
             (for-each (lambda (inst) 
                         (display inst stderr) (newline stderr) )
                       (BCRT-disassemble 
                        (BCRT-Module-Environment-module me) ) ) ) )
  (let ((results '()))
    (call/cc
     (lambda (exit)
       (define (thread-terminate!)
         (let ((result *val*))
           ;;(display "Gotten result: ")(show result)
           (set! results (cons result results))
           (if (scheduler-call)
               (BCRT-run-byte-machine)
               (exit results) ) ) )
       (set! *val* BCRT-uninitialized)
       ;;(set! *type* '???)
       (set! *genv* me)
       (set! *env* (make-BCRT-Null-Environment))
       ;; Deduced register:
       (set! *module* (BCRT-Module-Environment-module me))
       (set! *pc* 0)
       (set! *cont* (make-BCRT-Internal-Continuation "" thread-terminate!))
       (set! *all-threads*
             (list (make-BCRT-Thread
                    *user* *genv* *env* *pc* *val* *cont* )) )
       (scheduler-call)
       (BCRT-run-byte-machine) ) ) ) )

;;; end of BCappl.scm
