;;; $Id: BCenv.scm,v 1.20 1997/08/17 19:30:46 queinnec Exp $

;;; This file defines the predefined environments for the bytecode
;;; compiler.  The following forms define predefined variables in the
;;; compilation environment and in the runtime environment.

;;;==================================================================
;;; This is the global predefined environment as defined in BCpred.scm
;;; To simplify things the primitives that are defined here keep the
;;; same indexes they have in the original g.icslas.

(define g.scm.icslas '())

;;; This class refines Functional-Description to describe a predefined
;;; binding bound to a primitive. The Scheme value is stored in it to
;;; ease the construction of the predefined global immutable environment.

(define-class BC-Functional-Description Functional-Description
  ( value   ; 
    ) )

;;; This macro allows to define an association in g.scm.icslas for a
;;; variable holding a constant as value.

(define-icslas-macro (definitial name . value)
  (define (find-index name g)
    (let search ((g g))
      (if (pair? g)
          (if (eq? name (Predefined-Variable-name (car g)))
              (BC-Predefined-Variable-rank (car g))
              (search (cdr g)) )
          '(length g.scm.icslas) ) ) )
  `(begin
     (set! g.scm.icslas 
           (cons (make-BC-Predefined-Variable 
                  ',name 
                  (make-Constant-Description 
                   ,(if (pair? value) 
                        (car value)
                        'BCRT-uninitialized) )
                  ,(find-index name g.icslas) )
                 g.scm.icslas ) )
     ',name ) )

;;; Define the usual constants needed for the test suite.
;;; unspecified is needed by objectify (for instance in (begin)).

(definitial f BCRT-false)
(definitial t BCRT-true) 
(definitial nil BCRT-nil)
(definitial unspecified BCRT-unspecified)
(definitial end-of-file BCRT-end-of-file)

;;; This macro allows to define primitives.
;;; A primitive always terminates, just return a result in the *val*
;;; register and resumes the bytecode machine.  

;;; This macro defines a regular primitive.

(define-icslas-macro (defprimitive name address arity)
  (define (arity->variables arity)
    (case arity
      ((0) '())
      ((1) '(a))
      ((2) '(a b))
      ((3) '(a b c))
      ((4) '(a b c d))
      (else (Byte-Code-error "No such arity" arity)) ) )
  (if (>= arity 0)
      `(definternalprimitive ,name
         ,(let ((variables (arity->variables arity)))
            `(lambda ,variables
               (set! *val* (,address . ,variables))
               (BCRT-resume *cont*) ) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a primitive" name) ) )

;;; This macro defines a predicate, the truth value must be converted
;;; into a BCRT truth value.

(define-icslas-macro (defpredicate name address arity)
  (define (arity->variables arity)
    (case arity
      ((0) '())
      ((1) '(a))
      ((2) '(a b))
      ((3) '(a b c))
      ((4) '(a b c d))
      (else (Byte-Code-error "No such arity" arity)) ) )
  (if (>= arity 0)
      `(definternalprimitive ,name 
         ,(let ((variables (arity->variables arity)))
            `(lambda ,variables
               (set! *val* (if (,address . ,variables)
                               BCRT-true BCRT-false ))
               (BCRT-resume *cont*) ) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a predicate" name) ) )

;;; This macro defines an arithmetic function.

(define-icslas-macro (defarithmeticprimitive name address arity)
  (define (arity->variables arity)
    (case arity
      ((0) '())
      ((1) '(a))
      ((2) '(a b))
      ((3) '(a b c))
      ((4) '(a b c d))
      (else (Byte-Code-error "No such arity" arity)) ) )
  (if (>= arity 0)
      `(definternalprimitive ,name
         ,(let ((variables (arity->variables arity)))
            `(lambda ,variables
               (set! *val* (create-BCRT-number 
                            (,address . ,(map (lambda (v)
                                                `(BCRT-number-value ,v) )
                                              variables )) ))
               (BCRT-resume *cont*) ) )
         ,arity )
      (Byte-Code-error "Incorrect arity for an arithmetic primitive" name) ) )

;;; This macro defines special functions with arity n or n+1. This is
;;; necessary for a lot of R4RS functions such as display, make-vector.

(define-icslas-macro (defspecialprimitive name address arity)
  (define (find-index name g)
    (let search ((g g))
      (if (pair? g)
          (if (eq? name (Predefined-Variable-name (car g)))
              (BC-Predefined-Variable-rank (car g))
              (search (cdr g)) )
          '(length g.scm.icslas) ) ) )
  `(begin
     (set! g.scm.icslas
           (cons (make-BC-Predefined-Variable
                  ',name 
                  (make-BC-Functional-Description
                   (lambda (a arity) (or (= a arity) (= a (+ arity 1))))
                   ,arity
                   (make-BCRT-Primitive
                    -1
                    (lambda ()
                      (let ((n (BCRT-Environment-value-length *val*)))
                        (if (not (or (= n ,(+ arity 1)) (= n ,(+ arity 2))))
                            (BCRT-error "Wrong arity" ',name) )
                        (listify-frame! *val* 0 '() cons)
                        (set! *val* (apply ,address
                                           (BCRT-Environment-value *val* 0) ))
                        (BCRT-resume *cont*) ) )
                    ',name ) )
                  ,(find-index name g.icslas) )
                 g.scm.icslas ) )
     ',name ) )

;;; This macro defines nary function. The underlying function is applied
;;; on all its arguments.

(define-icslas-macro (defnaryprimitive name address arity)
  (define (arity->variables arity)
    (case arity
      ((0) '())
      ((1) '(a))
      ((2) '(a b))
      ((3) '(a b c))
      ((4) '(a b c d))
      (else (Byte-Code-error "No such arity" arity)) ) )
  (if (< arity 0)
      `(definternalprimitive ,name 
         (lambda ,(arity->variables (- (+ arity 1)))
           (listify-frame! *val* ,(- (+ arity 1)) '() cons)
           (set! *val* 
                 (apply ,address 
                        ,@(map (lambda (i)
                                 `(BCRT-Environment-value *val* ,i) )
                               (case (- (+ arity 1))
                                 ((0) '())
                                 ((1) '(0))
                                 ((2) '(0 1))
                                 (else (Byte-Code-error "Weird narity" 
                                                        arity )) ) )
                        (BCRT-Environment-value *val* ,(- (+ arity 1))) ) )
           (BCRT-resume *cont*) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a nary primitive" name) ) )

(define-icslas-macro (defnarypredicate name address arity)
  (if (< arity 0)
      `(defnaryprimitive ,name 
         (lambda args 
           (if (apply ,address args) BCRT-true BCRT-false) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a nary primitive" name) ) )

(define-icslas-macro (defnaryarithmeticprimitive name address arity)
  (if (< arity 0)
      `(defnaryprimitive ,name 
         (lambda args 
           (create-BCRT-number (apply ,address (map BCRT-number-value args))) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a nary primitive" name) ) )

(define-icslas-macro (defnaryarithmeticpredicate name address arity)
  (if (< arity 0)
      `(defnaryprimitive ,name 
         (lambda args 
           (if (apply ,address (map BCRT-number-value args))
               BCRT-true BCRT-false) )
         ,arity )
      (Byte-Code-error "Incorrect arity for a nary primitive" name) ) )

;;; This is the internal macro to define primitives.

(define-icslas-macro (definternalprimitive name address arity)
  (define (find-index name g)
    (let search ((g g))
      (if (pair? g)
          (if (eq? name (Predefined-Variable-name (car g)))
              (BC-Predefined-Variable-rank (car g))
              (search (cdr g)) )
          '(length g.scm.icslas) ) ) )
  `(begin
     (set! g.scm.icslas 
           (cons (make-BC-Predefined-Variable
                  ',name
                  (make-BC-Functional-Description 
                   ,(if (>= arity 0) '= '>=)
                   ,(if (>= arity 0) arity (- (+ arity 1)))
                   (make-BCRT-Primitive
                    ,arity
                    ,address
                    ',name ) )
                  ,(find-index name g.icslas) )
                 g.scm.icslas ) )
     ',name ) )

;;; The generic creator of numbers.

(define (create-BCRT-number num)
  (cond ((integer? num) (make-BCRT-fixnum num))
        ((real? num)   (make-BCRT-float num)) 
        (else (BCRT-error "Unknown type of number" num)) ) )

;;;==================================================================

;(defpredicate boolean? 
;  (lambda (o) (or (eq? o BCRT-true)
;                  (eq? o BCRT-false) ))
;  1 )

(defpredicate eqv? eqv? 2)
(defpredicate eq? eq? 2)

;(define (BCRT-equal? o1 o2)
;  (or (eq? o1 o2)
;      (egal o1 o2) ) )
;(defpredicate equal? BCRT-equal? 2)

(defprimitive cons (lambda (hd tl) (make-BCRT-Pair tl hd)) 2)
;(defprimitive car BCRT-Pair-head 1)
;(defprimitive cdr BCRT-Pair-tail 1)
;(defpredicate pair? BCRT-Pair? 1)
;(defprimitive set-car! set-BCRT-Pair-head! 2)
;(defprimitive set-cdr! set-BCRT-Pair-tail! 2)

(define (listify-frame! v* arity nil cons)
  (let loop ((index (- (BCRT-Environment-value-length v*) 1))
             (result nil) )
    (if (= arity index)
        (set-BCRT-Environment-value! v* arity result)
        (loop (- index 1)
              (cons (BCRT-Environment-value v* (- index 1))
                    result ) ) ) ) )

(definternalprimitive list
  (lambda ()
    (listify-frame! *val* 0 BCRT-nil (lambda (hd tl) (make-BCRT-Pair tl hd)))
    (set! *val* (BCRT-Environment-value *val* 0))
    (BCRT-resume *cont*) )
   -1 ) ; at least zero arguments

;(defpredicate symbol? BCRT-Symbol? 1)
;(defprimitive symbol->string BCRT-Symbol-name 1)

;;; Pay attention that BCRT-symbols-alist maps Scheme strings to BCRT-Symbols.

(defprimitive string->symbol 
  (lambda (bcrt-str)
    (let ((str (convert-to-scheme bcrt-str)))
      (let search ((alist BCRT-symbols-alist))
        (if (pair? alist)
            (let ((symb (cdar BCRT-symbols-alist)))
              (if (equal? bcrt-str (BCRT-Symbol-name symb))
                  symb
                  (search (cdr alist)) ) )
            (let ((symb (make-BCRT-Symbol str)))
              (set! BCRT-symbols-alist
                    (cons (cons str symb)
                          BCRT-symbols-alist ) )
              symb ) ) ) ) )
  1 )

;;; Numbers are now coded as indirect instances of BCRT-number.

;;(defpredicate integer? integer? 1)
;(defpredicate rational? (lambda (o) #f) 1)
;;(defpredicate real? real? 1)
;(defpredicate complex? (lambda (o) #f) 1)
;;(defpredicate number? number? 1)

;;(defpredicate exact? exact? 1)
;;(defpredicate inexact? inexact? 1)

(defnaryarithmeticpredicate = = -3)
(defnaryarithmeticpredicate < < -3)
(defnaryarithmeticpredicate > > -3)
(defnaryarithmeticpredicate <= <= -3)
(defnaryarithmeticpredicate >= >= -3)

;;(defpredicate zero? zero? 1)
;;(defpredicate positive? positive? 1)
;;(defpredicate negative? negative? 1)
;;(defpredicate odd? odd? 1)
;;(defpredicate even? even? 1)

;;(defnaryprimitive max max -2)
;;(defnaryprimitive min min -2)
(defnaryarithmeticprimitive + + -1)
(defnaryarithmeticprimitive * * -1)
(defnaryarithmeticprimitive - - -2)
(defnaryarithmeticprimitive / / -2)
;;(defprimitive abs abs 1)
(defarithmeticprimitive quotient quotient 2)
(defarithmeticprimitive remainder remainder 2)
(defarithmeticprimitive modulo modulo 2)
;;(defnaryprimitive gcd gcd -1)
;;(defnaryprimitive lcm lcm -1)

;(defprimitive numerator numerator 1)
;(defprimitive denominator denominator 1)

(defarithmeticprimitive floor floor 1)
(defarithmeticprimitive ceiling ceiling 1)
(defarithmeticprimitive truncate truncate 1)
(defarithmeticprimitive round round 1)

;(defprimitive rationalize rationalize 2)

;;; no trigonometric functions

(defarithmeticprimitive sqrt sqrt 1)
(defarithmeticprimitive expt expt 2)

;(defprimitive make-rectangular make-rectangular 2)
;(defprimitive make-polar make-polar 2)
;(defprimitive real-part real-part 1)
;(defprimitive imag-part imag-part 1)
;(defprimitive angle angle 1)

;;(defprimitive exact->inexact exact->inexact 1)
;;(defprimitive inexact->exact inexact->exact 1)

;(defprimitive number->string
;  (lambda (n . radix)
;    (let ((s (apply number->string n radix)))
;      (apply make-BCRT-String (string->list s)) ) )
;  1 )

;(defspecialprimitive string->number 
;  (lambda (string . radix)
;    (let ((str (convert-to-scheme string)))
;      (apply string->number str radix) ) )
;  1 )

;(defpredicate char? char? 1)
;(defpredicate char=? char=? 2)
;(defpredicate char<? char<? 2)
;(defpredicate char>? char>? 2)
;(defpredicate char<=? char<=? 2)
;(defpredicate char>=? char>=? 2)
;(defpredicate char-ci=? char-ci=? 2)
;(defpredicate char-ci<? char-ci<? 2)
;(defpredicate char-ci>? char-ci>? 2)
;(defpredicate char-ci<=? char-ci<=? 2)
;(defpredicate char-ci>=? char-ci>=? 2)
;(defpredicate char-alphabetic? char-alphabetic? 1)
;(defpredicate char-numeric? char-numeric? 1)
;(defpredicate char-whitespace? char-whitespace? 1)
;(defpredicate char-upper-case? char-upper-case? 1)
;(defpredicate char-lower-case? char-lower-case? 1)

;(defprimitive char->integer char->integer 1)
;(defprimitive integer->char integer->char 1)
;(defprimitive char-upcase char-upcase 1)
;(defprimitive char-downcase char-downcase 1)

;(defpredicate string? BCRT-String? 1)
(defspecialprimitive make-string 
  (lambda (n . item)
    (let ((n (BCRT-Number-value n))
          (char (if (pair? item) (car item)
                    (BCRT-String-character BCRT-all-characters 32) )))
      (if (BCRT-Character? char)
          (let ((result (allocate-BCRT-String n)))
            (do ((i 0 (+ i 1)))
                ((>= i n))
              (set-BCRT-String-character! result i char) )
            result )
          (BCRT-error "Not a char" item) ) ) )
  1 )
(defnaryprimitive string 
  (lambda chars
    (apply make-BCRT-String (length chars) chars) )
  -1 )
;(defprimitive string-length BCRT-String-character-length 1)
;(defprimitive string-ref BCRT-String-character 2)
;(defprimitive string-set! set-BCRT-String-character! 3)

(defprimitive string-copy clone 1)

;(defpredicate vector? BCRT-Vector? 1)
(defspecialprimitive make-vector 
  (lambda (n . item)
    (let ((n (BCRT-Number-value n))
          (result (allocate-BCRT-Vector n)) )
      (if (pair? item)
          (do ((i 0 (+ i 1)))
              ((>= i n))
            (set-BCRT-Vector-item! result i (car item)) ) )
      result ) )
  1 )
(defnaryprimitive vector 
  (lambda items
    (apply make-BCRT-Vector (length items) items) )
  -1 )
;(defprimitive vector-length BCRT-Vector-item-length 1)
;(defprimitive vector-ref BCRT-Vector-item 2)
;(defprimitive vector-set! set-BCRT-Vector-item! 3)

(defpredicate procedure?
  (lambda (o) (or (BCRT-Closure? o)
                  (BCRT-Primitive? o)
                  (BCRT-Continuation? o) ))
  1 )

;;; The arg1 variable is ignored, its value will be fetched in *val*.

(definternalprimitive apply
  (lambda (f arg1)
    (define (BCRT-length l)
      (if (BCRT-Pair? l)
          (+ 1 (BCRT-length (BCRT-Pair-tail l)))
          0 ) )
    (let ((n (BCRT-Environment-value-length *val*)))
      (if (>= n 3)
          (let* ((last-arg-index (- n 2))
                 (last-arg (BCRT-Environment-value *val* last-arg-index))
                 (size (+ last-arg-index (BCRT-length last-arg)))
                 (frame (allocate-BCRT-Environment size)) )
            (do ((i 1 (+ i 1)))
                ((>= i last-arg-index))
              (set-BCRT-Environment-value! 
               frame (- i 1) (BCRT-Environment-value *val* i) ) )
            (do ((i (- last-arg-index 1) (+ i 1))
                 (last-arg last-arg (BCRT-Pair-tail last-arg)))
                ((not (BCRT-Pair? last-arg)))
              (set-BCRT-Environment-value! frame i (BCRT-Pair-head last-arg)) )
            (set! *val* frame)
            (BCRT-invoke f) )
          (BCRT-error "Incorrect arity" 'apply) ) ) )
  -3 ) ; at least two argument (given to the f and arg1 variables)

(definternalprimitive call/cc 
  (lambda (f)
    (set! *val* (make-BCRT-Environment
                 BCRT-uninitialized 2 *cont* BCRT-nil ))
    (BCRT-invoke f) )
  1 )

;;; Ports are encoded with the underlying objects for now.

;(defpredicate input-port? input-port? 1)
;(defpredicate output-port? output-port? 1)
(define BCRT-stdin
  (make-BCRT-Input-Port
   (current-input-port)
   BCRT-false                           ; closed?
   (allocate-BCRT-String 0)             ; buffer
   (convert-to-quotation 0)             ; index
   BCRT-false                           ; poken
   (allocate-BCRT-String 80)            ; zone
   (convert-to-quotation 0)
   ) )
(define BCRT-stdout
  (make-BCRT-Output-Port
   (current-output-port)
   BCRT-false                           ; closed?
   (allocate-BCRT-String 100)           ; buffer
   (convert-to-quotation 0)             ; index
   ) )
(definitial stdin BCRT-stdin)
(definitial stdout BCRT-stdout)

;;; Wait for real dynamic binding              FUTURE
(defprimitive current-input-port 
  (lambda () BCRT-stdin)
  0 )
(defprimitive current-output-port 
  (lambda () BCRT-stdout)
  0 )
(defprimitive open-input-file 
  (lambda (file)
    (let* ((filename (convert-to-scheme file))
           (fd       (open-input-file filename)) )
      (make-BCRT-Input-Port
       fd 
       BCRT-false                       ; closed
       (allocate-BCRT-String 0)         ; buffer
       (convert-to-quotation 0)         ; index
       BCRT-false                       ; poken
       (allocate-BCRT-String 80)        ; zone
       (convert-to-quotation 0)         ; zoneindex
       ) ) )
  1 )
(defprimitive open-output-file 
  (lambda (file)
    (let* ((filename (convert-to-scheme file))
           (fd       (open-output-file filename)) )
      (make-BCRT-Output-Port
       fd 
       BCRT-false                       ; closed
       (allocate-BCRT-String 100)       ; buffer
       (convert-to-quotation 0)         ; index
       ) ) )
  1 )
(defprimitive close-input-port 
  (lambda (bcrt-port)
    (set-BCRT-Input-Port-poken! bcrt-port BCRT-false)
    (set-BCRT-Port-closed?! bcrt-port BCRT-true) )
  1 )
(defprimitive close-output-port 
  (lambda (bcrt-port)
    (BCRT-flush-port bcrt-port)
    (set-BCRT-Port-closed?! bcrt-port BCRT-true) )
  1 )
(define (BCRT-flush-port bcrt-port)
  (let* ((buffer (BCRT-Port-buffer bcrt-port))
         (fd     (BCRT-Port-fd bcrt-port))
         (index  (convert-to-scheme (BCRT-Port-index bcrt-port))) )
    (do ((i 0 (+ i 1)))
        ((>= i index))
      (display (integer->char
                (BCRT-Character-code
                 (BCRT-String-character buffer i) ) )
               fd ) )
    (flush-output-port fd)                  ;; BIGLOO dependence
    (set-BCRT-Port-index! bcrt-port (convert-to-quotation 0)) ) )

;;; Still to write with dynamic binding            FUTURE
;;; (def? with-input-file ? 2)
;;; (def? with-output-file ? 2)  

;;;(defspecialprimitive read read 0)
;;;(defspecialprimitive read-char read-char 0)
;;;(defspecialprimitive peek-char peek-char 0)
(defprimitive do-read-char
  (lambda (bcrt-port)
    (if (not (eq? (BCRT-Port-closed? bcrt-port) BCRT-false))
        (BCRT-error "Closed port" bcrt-port 'do-read-char)
        (let ((char (read-char (BCRT-Input-Port-fd bcrt-port))))
          (if (eof-object? char)
              BCRT-end-of-file
              (BCRT-String-character BCRT-all-characters
                                     (char->integer char) ) ) ) ) )
  1 )
;(defpredicate eof-object? eof-object? 1)
;;; This one is not portable
;;;(defprimitive char-ready? char-ready? 0)

(defprimitive display-char 
  (lambda (bcrt-char bcrt-port)
    (if (not (eq? (BCRT-Port-closed? bcrt-port) BCRT-false))
        (BCRT-error "Closed port" 'display-char)
        (BCRT-insert-char! bcrt-port bcrt-char) ) )
  2 )

(defprimitive display-string 
  (lambda (bcrt-string bcrt-port)
    (if (not (eq? (BCRT-Port-closed? bcrt-port) BCRT-false))
        (BCRT-error "Closed port" 'display-string)
        (let ((n (BCRT-String-character-length bcrt-string)))
          (do ((i 0 (+ i 1)))
              ((>= i n))
            (BCRT-insert-char! 
             bcrt-port (BCRT-String-character bcrt-string i) ) ) ) ) )
  2 )
(defprimitive display-newline 
  (lambda (bcrt-port)
    (if (not (eq? (BCRT-Port-closed? bcrt-port) BCRT-false))
        (BCRT-error "Closed port" 'display-newline)
        (begin (BCRT-insert-char! 
                bcrt-port (BCRT-String-character bcrt-all-characters
                                                 (char->integer #\newline) ) )
               (BCRT-flush-port bcrt-port) ) ) )
  1 )
(define (BCRT-insert-char! bcrt-port bcrt-char)
  (let* ((buffer (BCRT-Port-buffer bcrt-port))
         (index  (convert-to-scheme (BCRT-Port-index bcrt-port)))
         (n      (BCRT-String-character-length buffer)) )
    (if (> n index)
        (begin
          (set-BCRT-String-character! buffer index bcrt-char)
          (set-BCRT-Port-index! bcrt-port (convert-to-quotation (+ 1 index))) )
        (begin 
          (BCRT-flush-port bcrt-port)
          (BCRT-insert-char! bcrt-port bcrt-char) ) ) ) )         

;;(defspecialprimitive write show 1)
;;(defspecialprimitive display show 1)
;;(defspecialprimitive newline newline 0)
;;(defspecialprimitive write-char write-char 1)

;;; transcript-on filename                 FUTURE
;;; transcript-off                 FUTURE

;;;==================================================================
;;; Object related primitives.
;;; They make Meroon apparent from Icslas.
;;; NOTE: actually allocation is not yet visible.

(defprimitive field-value field-value 2)
(defprimitive indexed-field-value 
  (lambda (o field i)
    (field-value o field (convert-to-scheme i)) )
  3 )
(defprimitive indexed-field-length 
  (lambda (o field)
    (create-BCRT-number (indexed-field-length o field)) )
  2 )
(defprimitive set-field-value! set-field-value! 3)
(defprimitive set-indexed-field-value!
  (lambda (o v field i)
    (set-field-value! o v field (convert-to-scheme i)) )
  4 )
(defpredicate is-a? is-a? 2)
(defprimitive object->class object->class 1)

(definitial <Pair> BCRT-Pair-class)
(definitial <Pair.tail> (list-ref (Class-fields BCRT-Pair-class) 0))
(definitial <Pair.head> (list-ref (Class-fields BCRT-Pair-class) 1))
(definitial <Symbol> BCRT-Symbol-class)
(definitial <Symbol.name> (list-ref (Class-fields BCRT-Symbol-class) 0))
(definitial <String> BCRT-String-class)
(definitial <String.character> (list-ref (Class-fields BCRT-String-class) 0))
(definitial <Vector> BCRT-Vector-class)
(definitial <Vector.item> (list-ref (Class-fields BCRT-Vector-class) 0))
(definitial <Character> BCRT-Character-class)
(definitial <Character.code> (list-ref (Class-fields BCRT-Character-class) 0))
(definitial <Number> BCRT-Number-class)
(definitial <Fixnum> BCRT-fixnum-class)
(definitial <Float> BCRT-float-class)
(definitial <OutputPort> BCRT-Output-Port-class)
(definitial <InputPort> BCRT-Input-Port-class)
(let ((field (list-ref (Class-fields BCRT-Input-Port-class) 4)))
  (if (equal? 'poken (Field-name field))
      (definitial <InputPort.poken> field) 
      (Byte-Code-error "Wrong <InputPort.poken> field") ) )
(let ((field (list-ref (Class-fields BCRT-Input-Port-class) 5)))
  (if (equal? 'zone (Field-name field))
      (definitial <InputPort.zone> field) 
      (Byte-Code-error "Wrong <InputPort.poken> field") ) )
(let ((field (list-ref (Class-fields BCRT-Input-Port-class) 6)))
  (if (equal? 'zoneindex (Field-name field))
      (definitial <InputPort.zoneindex> field) 
      (Byte-Code-error "Wrong <InputPort.poken> field") ) )

(define BCRT-all-characters
  (let ((result (allocate-BCRT-String 256)))
    (do ((i 0 (+ i 1)))
        ((>= i 256))
      (set-BCRT-String-character! 
       result i (make-BCRT-Character i) ) )
    result ) )

(definitial all-characters BCRT-all-characters)

(definitial to-upper-case-characters
  (let ((result (allocate-BCRT-String 256)))
    (do ((i 0 (+ i 1)))
        ((>= i 256))
      (set-BCRT-String-character! 
       result i (make-BCRT-Character
                 (char->integer (char-upcase (integer->char i))) ) ) )
    result ) )

(definitial to-lower-case-characters
  (let ((result (allocate-BCRT-String 256)))
    (do ((i 0 (+ i 1)))
        ((>= i 256))
      (set-BCRT-String-character! 
       result i (make-BCRT-Character 
                 (char->integer (char-downcase (integer->char i))) ) ) )
    result ) )

;;;==================================================================
;;; Special primitives for debug

(defpredicate toggle-trace! 
  (lambda ()
    (set! *debug* (not *debug*))
    *debug* )
  0 )

;;;==================================================================
;;; Special primitives.

;;; This eval corresponds to an eval at toplevel. It ignores the
;;; current lexical environment as well as the current module
;;; environment. The expression is compiled in g.scm.icslas which only
;;; knows the predefined immutable global environment. The resulting
;;; bytecode segment will be linked in the current User's toplevel.

(define (BCRT-eval bcrt-exp)
  (let* ((exp (convert-to-scheme bcrt-exp))
         (e   (icslas-expand exp))
         (bc  (compile->Byte-Code "dynamic evaluation" e g.scm.icslas))
         (bm  (Byte-Compilation->BCRT-Module bc))
         (me  (install-module! (BCRT-User-toplevel *user*)
                               bm )) )
    (if *debug*
        (begin (display "---------Code----------" stderr)
               (newline stderr)
               (for-each (lambda (inst) 
                           (display inst stderr) (newline stderr) )
                         (BCRT-disassemble bm) ) ) )
    (set! *cont* 
          (make-BCRT-Context-Continuation
           *cont*
           0                             ; *pc*
           (make-BCRT-Null-Environment)  ; *env*
           me ) )                        ; *genv*
    (BCRT-resume *cont*) ) )

(definternalprimitive eval BCRT-eval 1)

(define (BCRT-load bcrt-filename)
  (let* ((bm  (file->BCRT-Module (convert-to-scheme bcrt-filename)))
         (me  (install-module! (BCRT-User-toplevel *user*)
                               bm )) )
    (set! *cont* 
          (make-BCRT-Context-Continuation
           *cont*
           0                             ; *pc*
           (make-BCRT-Null-Environment)  ; *env*
           me ) )                        ; *genv*
    (BCRT-resume *cont*) ) )

(definternalprimitive load BCRT-load 1)

;;; (byte) compiling a file.

(define (BCRT-compile-file filename)
  (byte-compile-file icslas-expand g.scm.icslas (convert-to-scheme filename)) )

(defprimitive compile-file BCRT-compile-file 1)

;;;==================================================================
;;; Exception Handlers.
;;; The invokation frame for diagnose will be used to invoke the handler.

(define (BCRT-diagnose exception continuable?)
  (set! *env* *val*)   ; save the invokation frame
  (set! *val* *cont*)  ; register where the continuation will be scanned.
  (BCRT-lookup-handler exception continuable? *val*) )

(define-generic (BCRT-lookup-handler exception continuable? (k))
  ;; Invoke the system default handler.
  (set! *val* *env*)  ;  restaure the invocation frame.
  (diagnose exception #f) )

;;; The continuation frame that intercepts values coming from the
;;; handler invocation must be run under the sponsorship of the
;;; previous handler.

(define-method (BCRT-lookup-handler exception c? (k BCRT-Continuation))
  (set! *val* (BCRT-Continuation-next k))
  (if (eq? c? BCRT-false)
      (set! *cont*
            (make-BCRT-Internal-Continuation
             *cont*
             (lambda ()
               (BCRT-error "Non continuable anomaly" exception) ) ) ) )
  (BCRT-lookup-handler exception c? *val*) )

(define-method 
  (BCRT-lookup-handler exception c? (k BCRT-Skip-Handler-Continuation))
  (set! *val* (BCRT-Skip-Handler-Continuation-alternate k))
  (BCRT-lookup-handler exception c? *val*) )

(define-method (BCRT-lookup-handler exception c? (k BCRT-Handler-Continuation))
  (set! *val* *env*)  ;  restaure the invocation frame.
  (set! *cont* (make-BCRT-Skip-Handler-Continuation 
                *cont* (BCRT-Continuation-next k) ))
  (if (eq? c? BCRT-false)
      (set! *cont*
            (make-BCRT-Internal-Continuation
             *cont*
             (lambda ()
               (BCRT-error "Non continuable anomaly" exception) ) ) ) )
  (BCRT-invoke (BCRT-Handler-Continuation-handler k)) )
  
(definternalprimitive diagnose (lambda (e c)(BCRT-diagnose e c)) 2)

;;; BCRT-error signals non continuable anomalies.
;;; NOTE: it has to call BCRT-diagnose with the correct frame in *val*.

(define (BCRT-error msg . hints)
  (let* ((anomaly (make-BCRT-Underlying-Anomaly
                   (convert-to-quotation msg)
                   hints ))
         (frame (make-BCRT-Environment 'wait 3 anomaly BCRT-false BCRT-nil)) )
    (set! *val* frame)
    (BCRT-diagnose anomaly BCRT-false) ) )

;;; end of BCenv.scm
