;;; $Id: BCgen.scm,v 1.20 1995/09/19 11:53:37 queinnec Exp $

;;; This file defines the bytecode compiler.
;;; The codes assigned to the bytecode instructions appear near the end
;;; of this file. 

;;; Exportations:
;;;        (compiler->Byte-Code Sexp predefined.env) : Byte-Compilation
;;;        BC-<instruction>-code 
;;; end of exportations.

;;;==================================================================
;;;          Additional classes.

;;; This class defines a bytecode compilation. This object is enriched
;;; during the compilation and will contain at last the result of the
;;; compilation.

(define-class Byte-Compilation Object
  ( original             ;; the Sexpression to be compiled
    ast                  ;; the corresponding abstract syntax tree
    global               ;; a Global-Environment
    quotations           ;; a list of Quotation-Variables
    index                ;; where to store a byte in the next field
    bytes                ;; a vector of bytes
    name                 ;; a string.
    ) )

;;; Global and Predefined variables have an index recorded in them.
;;; It is therefore easy to generate their address in the global frames.

(define-class BC-Predefined-Variable Predefined-Variable
  ( rank ; rank in the global predefined environment.
    ) )

(define-class BC-Global-Variable Global-Variable
  ( rank ; rank in the global mutable environment
    ) )

;;; Quotations are collected then referred to through global variables.
;;; The 'name' of a Quotation-Variable is a natural number, the value field
;;; contains the associated quoted value.

(define-class BC-Quotation-Variable Variable 
  ( value              ; a Scheme value.
    ) )

;;;==================================================================
;;; Compile an expression into a byte-program.
;;; This compilation is performed under the sponsorship of a special 
;;; exception handler. G.init is the common immutable predefined global 
;;; environment. Note that, currently, no expansion is made.

(define (compile->Byte-Code name sexp g.init)
  (monitor compile->Byte-Code-handler
   (report-progress "Parsing...")
    (let* ((g  (make-Global-Environment g.init '()))
           (o  (objectify sexp (make-Local-Environment '()) g))
           (bc (make-Byte-Compilation sexp o g '() 0 
                                      (make-vector 10 BC-nop-code)
                                      name )) )
      ;; add the box transformation to remove mutable variables
      (report-progress "Boxing mutable variables...")
      (set! o (insert-box! o))
      (set-Byte-Compilation-ast! bc o)
      ;; figure out which variables are surely initialized
      (report-progress "Initialization analysis...")
      (set! o (initialization-analyze! o))
      (set-Byte-Compilation-ast! bc o)
      ;; compile towards bytecodes
      (report-progress "Emitting bytes...")
      (->Byte-Code o '() g #t bc)
      (emit BC-module-border-code bc)
      (report-progress "Emitted" (Byte-Compilation-index bc) "bytes")
      ;; return the  Byte-Compilation object
      bc ) ) )

;;;==================================================================
;;; This is the error handler used when compiling expressions to
;;; bytecodes. It catches dynamic creation of unreferenced global
;;; variables.

(define-generic (compile->Byte-Code-handler (exc) c?)
  ;; resignal the same exception to the dynamically enclosing handler.
  (diagnose exc c?) )

;;; This handler just add the newly found global variable to the
;;; global environment automatically. This allows to make a global
;;; variable exist as soon as it is mentioned. 

(define-method (compile->Byte-Code-handler (exc Unknown-Variable) c?)
  (let* ((g   (Unknown-Variable-g exc))
         (gvs (Global-Environment-global g))
         (v   (make-BC-Global-Variable (Unknown-Variable-name exc) 
                                       #f
                                       (length gvs) )) )
    (set-Global-Environment-global! g (cons v gvs))
    (make-Global-Reference v) ) )

;;; Ignore uninitialized variables but return #f to leave them non initialized.

(define-method (compile->Byte-Code-handler (exc Uninitialized-Variable) c?)
  #f )

;;;==================================================================
;;; Utilities to manage the vector of bytes.
;;; They're actually quite slow:  To be improved          FUTURE

;;; This function adds a byte to the bytevector stored in a
;;; Byte-Compilation.  When the bytevectors is exhausted, we just
;;; double it. Returns the address of the stored byte.

(define (emit byte bc)
  (let* ((index (Byte-Compilation-index bc))
         (bytes (Byte-Compilation-bytes bc))
         (vlen  (vector-length bytes))
         (step  1024) )
    (if (and (= 0 (modulo index step)) (>= index step))
        (report-progress "Current ByteCode vector size:" 
                         (quotient index step) "K." ) )
    (if (>= index vlen)
        (let ((newbytes (make-vector (* 2 vlen) BC-nop-code)))
          (do ((i 0 (+ i 1)))
              ((>= i vlen))
            (vector-set! newbytes i (vector-ref bytes i)) )
          (set! bytes newbytes)
          (set-Byte-Compilation-bytes! bc newbytes) ) )
    (vector-set! bytes index byte)
    (set-Byte-Compilation-index! bc (+ index 1))
    index ) )

;;; extend the bytevector if it cannot hold an additional number of
;;; bytes to be emitted in some future.

(define (can-emit! bc n)
  (let* ((index (+ n (Byte-Compilation-index bc)))
         (bytes (Byte-Compilation-bytes bc))
         (vlen  (vector-length bytes)) )
    (cond 
     ((< index vlen) bytes)
     (else 
      (let ((newbytes (make-vector (* 2 vlen) BC-nop-code)))
        (do ((i 0 (+ i 1)))
            ((>= i vlen))
          (vector-set! newbytes i (vector-ref bytes i)) )
        (set-Byte-Compilation-bytes! bc newbytes)
        newbytes ) ) ) ) )
        
;;; To double the bytevector anytimes it is exhausted leads to
;;; normalize the bytevector when the compilation is over.

(define (truncate-bytecode-vector! bc)
  (let* ((index (Byte-Compilation-index bc))
         (bytes (Byte-Compilation-bytes bc))
         (newbytes (make-vector index BC-nop-code)) )
    (do ((i 0 (+ i 1)))
        ((>= i index))
      (vector-set! newbytes i (vector-ref bytes i)) )
    (set-Byte-Compilation-bytes! bc newbytes)
    bc ) )

;;; Patch the byte at address1 to hold the offset necessary to jump at
;;; address2. This jump is pc relative. If the jump is greater than 255,
;;; translate the bytes to give more room for the jump-offset.
;;; More visually:
;;;          +---------------------------------------------------+
;;;          |    goto ?      ........        instruction ... end
;;;                    ^                      ^
;;;                   address1               address2
;;; The instruction to be executed after (goto ?) should be instruction.
;;;
;;; This may become if delta =  address2 - address1 and delta < 256
;;;          +---------------------------------------------------+
;;;          |    goto delta  ........        instruction ... end
;;;                    ^                      ^
;;;                   address1               address2
;;; This is so since the meaning of goto is to add delta to the PC.
;;; If delta >= 256 then patch the goto code to become a long-goto and
;;; translate all the instructions from address1+1 until infinite.
;;;          +---------------------------------------------------+-----+
;;;          |LONGgoto delta delta ........        instruction ... end+1
;;;                    ^                           ^               
;;;                   address1                 address2+1      
;;; This function returns the size of the performed translation.

(define (patch-jump! bc address1 address2)
  (let ((jump  (- (- address2 address1) 1))
        (index (Byte-Compilation-index bc))
        (bytes (Byte-Compilation-bytes bc)) )
    (cond
     ((<= jump 0)
      (Byte-Code-error "Negative jump offset" bc address1 address2) )
     ((<= jump 255)
      (vector-set! bytes address1 jump)
      0 )
     ((<= jump 65535)
      ;; report a warning                                     DEBUG
      (Byte-Code-warning "Translated code" jump)
      (set! bytes (can-emit! bc 1))
      (do ((i index (- i 1)))
          ((<= i address1))
        (vector-set! bytes i (vector-ref bytes (- i 1))) )
      (set-Byte-Compilation-index! bc (+ index 1))
      (let ((jump1 (modulo jump 256))
            (jump2 (quotient jump 256)) )
        ;; patch the previous instruction (just add one to its code) ie
        ;;   short-* instructions are turned into long-* instructions.
        ;;   long-* instructions are turned into long-long-* instructions.
        ;; NOTE: This requires instruction code to respect some law.
        (vector-set! bytes (- address1 1) 
                     (+ 1 (vector-ref bytes (- address1 1))) )
        (vector-set! bytes address1 jump2)
        (vector-set! bytes (+ address1 1) jump1) )
      1 )
     (else
      (Byte-Code-error "Too big jump offset" bc address1 address2) ) ) ) )

;;; Codes waiting for a numeric argument can expect it under different
;;; forms whether it fits in one, two or four bytes. The various
;;; opcodes can be computed from the shortest one. Small numbers are
;;; encoded with appropriate opcodes.

(define (special-emit-number number code bc)
  (cond ((<= number 255)
         (case number
           ((0) (emit (+ code 3) bc))
           ((1) (emit (+ code 4) bc))
           ((2) (emit (+ code 5) bc))
           ((3) (emit (+ code 6) bc))
           ((4) (emit (+ code 7) bc))
           (else (emit code bc)
                 (emit number bc) ) ) )
        ((<= number 65535)
         (emit (+ code 1) bc)
         (emit (quotient number 256) bc)
         (emit (modulo number 256) bc) )
        (else
         (emit (+ code 2) bc)
         (let loop ((i 4)(n number))
           (if (> i 0)
               (let ((p (quotient n 256))
                     (q (modulo n 256)) )
                 (loop (- i 1) p)
                 (emit q bc) ) ) ) ) ) )

(define (emit-number number code bc)
  (cond ((<= number 255)
         (emit code bc)
         (emit number bc) )
        ((<= number 65535)
         (emit (+ code 1) bc)
         (emit (quotient number 256) bc)
         (emit (modulo number 256) bc) )
        (else
         (emit (+ code 2) bc)
         (let loop ((i 4)(n number))
           (if (> i 0)
               (let ((p (quotient n 256))
                     (q (modulo n 256)) )
                 (loop (- i 1) p)
                 (emit q bc) ) ) ) ) ) )

;;;==================================================================
;;; Compile the Ast into byte-codes.

(define-generic (->Byte-Code (o) r g tail? bc))

;;;==================================================================
;;; Variables (code snarfed from TreeCode/treefy.scm)
;;; Compute lexical indexes to access variables in the environment.

;;; Return the offset of variable among variables or #f if absent.

(define (compute-lexical-offset variable variables)
  (let scan ((offset 0)
             (variables variables) )
    (if (pair? variables)
        (if (eq? (car variables) variable)
            offset
            (scan (+ 1 offset) (cdr variables)) )
        #f ) ) )

;;; Return where is variable among the list of list of variables R.

(define (compute-depth-and-offset variable r)
  (let scan ((depth  0)
             (offset 0)
             (r      r) )
    (if (pair? r)
        (let ((n (compute-lexical-offset variable (car r))))
          (if n (cons depth n)
              (scan (+ 1 depth) 0 (cdr r)) ) )
        #f ) ) )

(define-method (->Byte-Code (o Local-Reference) r g tail? bc)
  (let* ((variable     (Local-Reference-variable o))
         (depth+offset (compute-depth-and-offset variable r)) )
    (if depth+offset
        (let ((depth  (car depth+offset))
              (offset (cdr depth+offset)) )
          (case depth
            ((0) (special-emit-number 
                  offset BC-short-shallow-reference-code bc ))
            ((1) (special-emit-number
                  offset BC-short-depth1-reference-code bc ))
            (else        ;; OTHER CASES TO WRITE                  FUTURE
             (emit BC-short+short-deep-reference-code bc)
             (emit depth bc)
             (emit offset bc) ) )
          (if tail? (emit BC-return-code bc)) )
        (Byte-Code-error "Lost variable" o) ) ) )

;;; This instruction depends on the order along which predefined
;;; variables are defined in the predefined environment.

(define-method (->Byte-Code (o Predefined-Reference) r g tail? bc)
  (let ((variable (Reference-variable o)))
    (case (Predefined-Variable-name variable)
      ;; If constants are wanted.
      ;;((t) (emit BC-load-t-code bc))
      ;;((f) (emit BC-load-f-code bc))
      ;;((nil) (emit BC-load-nil-code bc))
      (else (->Byte-Code variable r g tail? bc)) )
    (if tail? (emit BC-return-code bc)) ) )

;;; Methods on variables leave their value or associated Box if
;;; mutable in the *val* register. Then the reference/assignment can
;;; further work on this box.

(define-method (->Byte-Code (o Global-Reference) r g tail? bc)
  (let ((variable (Reference-variable o)))
    (->Byte-Code variable r g #f bc)
    (cond 
     ((BC-Quotation-Variable? variable) #f)
     ((Global-Variable? variable)
      (if (Global-Variable-initialized? variable)
          (emit BC-box-ref-code bc)
          (emit BC-box-ref-and-check-code bc) ) )
     (else (Byte-Code-error "unknown type of global" variable)) )
    (if tail? (emit BC-return-code bc)) ) )

(define-method (->Byte-Code (o Global-Assignment) r g tail? bc)
  (->Byte-Code (Global-Assignment-variable o) r g #f bc)
  (emit BC-push-value-code bc)
  (->Byte-Code (Global-Assignment-form o) r g #f bc)
  (emit BC-pop-box-and-write-code bc)
  (if tail? (emit BC-return-code bc)) )

;;; Quotations are handled on the fly.
;;; share quotations ???     FUTURE

(define-method (->Byte-Code (o Constant) r g tail? bc)
  (let ((value (Constant-value o)))
    (cond 
     ((eq? value #f)
      (emit BC-load-false-code bc) )
     ((eq? value #t)
      (emit BC-load-true-code bc) )
     ((eq? value '())
      (emit BC-load-nil-code bc) )
     ((integer? value)
      (cond ((= value 0) (emit BC-load-zero-code bc))
            ((= value 1) (emit BC-load-one-code bc))
            ((<= 0 value)
             (emit-number value BC-load-nat1-code bc) )
            (else 
             (emit-number (- value) BC-load-negative-nat1-code bc) ) ) )
     (else
      (let* ((qv* (Byte-Compilation-quotations bc))
             (qv  (make-BC-Quotation-Variable (length qv*)
                                              (Constant-value o) )) )
        (set-Byte-Compilation-quotations! bc (cons qv qv*))
        (->Byte-Code qv r g #f bc) ) ) )
    (if tail? (emit BC-return-code bc)) ) )

;;;==================================================================
;;; Code generation for variables.

(define-method (->Byte-Code (o BC-Global-Variable) r g tail? bc)
  (let ((n (BC-Global-Variable-rank o)))
    (emit-number n BC-short-global-ref-code bc) ) )

(define-method (->Byte-Code (o BC-Quotation-Variable) r g tail? bc)
  (let ((n (BC-Quotation-Variable-name o)))
    (emit-number n BC-short-quotation-ref-code bc) ) )

(define-method (->Byte-Code (o BC-Predefined-Variable) r g tail? bc)
  (let ((n (BC-Predefined-Variable-rank o)))
    (emit-number n BC-short-predefined-ref-code bc) ) )

;;;==================================================================
;;; Control forms.

(define-method (->Byte-Code (o Alternative) r g tail? bc)
  (->Byte-Code (Alternative-condition o) r g #f bc)
  (emit BC-if-false-short-goto+-code bc)
  (let ((patch1 (emit 0 bc)))    ;  +pc-offset to be patched
    (->Byte-Code (Alternative-consequent o) r g tail? bc)
    (cond 
     (tail? (patch-jump! bc patch1 (Byte-Compilation-index bc))
            (->Byte-Code (Alternative-alternant o) r g tail? bc) )
     (else
      (emit BC-short-goto+-code bc)
      (let ((patch2 (emit 0 bc)))  ;  +pc-offset to be patched
        (->Byte-Code (Alternative-alternant o) r g tail? bc)
        (let ((delta (patch-jump! bc patch2 (Byte-Compilation-index bc))))
          (patch-jump! bc patch1 (+ patch2 1 delta)) ) ) ) ) ) )

(define-method (->Byte-Code (o Sequence) r g tail? bc)
  (->Byte-Code (Sequence-first o) r g #f bc)
  (->Byte-Code (Sequence-last o) r g tail? bc) )

;(define-method (->Byte-Code (o Monitoring) r g tail? bc)
;  (->Byte-Code (Monitoring-handler o) r g #f bc)
;  (emit BC-push-handler-code bc)
;  (->Byte-Code (Monitoring-body o) r g #f bc)
;  (emit BC-pop-handler-code bc)
;  (if tail? (emit BC-return-code bc)) )

;;; NOTE: we can compile body with tail? and omit pop-handler which can
;;; handled with a resume method if in tail position.
(define-method (->Byte-Code (o Monitoring) r g tail? bc)
  (->Byte-Code (Monitoring-handler o) r g #f bc)
  (emit BC-push-handler-code bc)
  (->Byte-Code (Monitoring-body o) r g tail? bc)
  (if (not tail?) (emit BC-pop-handler-code bc)) )

(define-method (->Byte-Code (o Escaper) r g tail? bc)
  (emit BC-short-create-escape-frame-code bc)
  (let ((patch1 (emit 0 bc))) ; +pc-offset to be patched
    (->Byte-Code (Escaper-body o)
                 (cons (list (Escaper-variable o)) r)
                 g tail? bc )
    (patch-jump! bc patch1 (Byte-Compilation-index bc))
    (emit BC-unlink-env-code bc)
    (if tail? (emit BC-return-code bc)) ) )

;;;==================================================================
;;; function and applications

(define-method (->Byte-Code (o Function) r g tail? bc)
  (emit BC-create-closure-and-short-goto+-code bc)
  (let* ((patch1    (emit 0 bc))
         (variables (Function-variables o))
         (n         (length variables)) )
    (cond ((and (pair? variables)
                (Local-Variable-dotted? (car (last-pair variables))) )
           (special-emit-number (- n 1) BC-short-check->=arity-code bc) )
          (else 
           (special-emit-number (+ n 1) BC-short-check-=arity-code bc) ) )
    (->Byte-Code (Function-body o) (cons variables r) g #t bc)
    (patch-jump! bc patch1 (Byte-Compilation-index bc)) )
  (if tail? (emit BC-return-code bc)) )

(define-method (->Byte-Code (o Regular-Application) r g tail? bc)
  (->Byte-Code (Regular-Application-function o) r g #f bc)
  (emit BC-push-value-code bc)
  (->Byte-Code-argument (Application-arguments o) r g 0 1 bc)
  (if tail?
      (emit BC-pop-function-and-goto-code bc)
      (emit BC-pop-function-and-invoke-code bc) ) )

(define-generic (Arguments-length (o Form)) 0)

(define-method (Arguments-length (o Arguments))
  (+ 1 (Arguments-length (Arguments-others o))) )

(define-method (->Byte-Code (o Predefined-Application) r g tail? bc)
  (let ((var  (Predefined-Application-variable o))
        (args (Application-arguments o)) )
    (cond 
     ;((= 1 (Arguments-length args))                       FUTURE
     ; (special protocol for monadic function) )
     (else
      (->Byte-Code-argument args r g 0 1 bc)
      (if tail? 
          (emit BC-primitive-goto-code bc)
          (emit BC-primitive-invoke-code bc) )
      (emit (BC-Predefined-Variable-rank var) bc) ) ) ) )

(define-method (->Byte-Code (o Fix-Let) r g tail? bc)
  (->Byte-Code-argument (Fix-Let-arguments o) r g 0 1 bc)
  (emit BC-link-env-code bc)
  (->Byte-Code (Fix-Let-body o) (cons (Fix-Let-variables o) r) g tail? bc)
  (if (not tail?) (emit BC-unlink-env-code bc)) )

;;;==================================================================
;;; Compile the arguments of an application.
;;; At the end, allocate a frame to contain them all. Add some additional
;;; room in the frame if required. This is not needed for let forms.

(define-generic (->Byte-Code-argument (o) r g rank delta bc))

(define-method (->Byte-Code-argument (o No-Argument) r g rank delta bc)
  (special-emit-number (+ delta rank) 
                       BC-short-create-activation-frame-code 
                       bc ) )

(define-method (->Byte-Code-argument (o Arguments) r g rank delta bc)
  (->Byte-Code (Arguments-first o) r g #f bc)
  (cond 
   ;; There is no remaining arguments.
   ((No-Argument? (Arguments-others o))
    (special-emit-number (+ delta rank 1) 
                         BC-short-create-activation-frame-and-push-code 
                         bc ) )
   (else
    (emit BC-push-value-code bc)
    (->Byte-Code-argument (Arguments-others o) r g (+ rank 1) delta bc)
    (special-emit-number rank BC-short-pop-in-frame-code bc) ) ) )

;;;==================================================================
;;; Box handling.

(define-method (->Byte-Code (o Box-Creation) r g tail? bc)
  (let ((index (compute-lexical-offset (Box-Creation-variable o) (car r))))
    (emit BC-boxify-variable-code bc)
    (emit index bc)
    (->Byte-Code (Box-Creation-body o) r g tail? bc) ) )

(define-method (->Byte-Code (o Box-Read) r g tail? bc)
  (->Byte-Code (Box-Read-reference o) r g #f bc)
  (emit BC-box-ref-code bc)
  (if tail? (emit BC-return-code bc)) )

(define-method (->Byte-Code (o Box-Write) r g tail? bc)
  (->Byte-Code (Box-Write-reference o) r g #f bc)
  (emit BC-push-value-code bc)
  (->Byte-Code (Box-Write-form o) r g #f bc)
  (emit BC-pop-box-and-write-code bc)
  (if tail? (emit BC-return-code bc)) )

;;; end of BCgen.scm
