;;; $Id: BCinst.scm,v 1.21 1995/07/24 18:35:18 queinnec Exp $

;;; The bytecode machine simulated in Scheme.

;;;==================================================================
;;; The bytecode machine.
;;; It has some registers.

;;; *val*, *type* holds the current value and its type.
;;; *module*, *pc* refers to the next instruction to execute.
;;; *env* is the local environment, a linked list of BCRT-Environment
;;;        *env* --> +--------+    +--------+
;;;                  | next: ----> |        |
;;;                  | value0 |    |        |
;;;                  | value1 |    |        |
;;;                  | reserve|    +--------+
;;;                  +--------+
;;; *penv* is the predefined immutable global environment
;;;       *penv* --> +--------+
;;;                  | unused |
;;;                  | value0 |
;;;                  |   ...  |
;;;                  | valuen |
;;;                  +--------+
;;; *genv* is the mutable global environment
;;;       *genv* --> +--------+      +--------+
;;;                  | next: ------> | unused |
;;;                  |  box0  |      | quote0 |
;;;                  |   ...  |      |   ...  |
;;;                  |  boxn ---+    +--------+
;;;                  +--------+ |
;;;                             +----> +-------+
;;;                                    | value |
;;;                                    +-------+
;;; 
;;; *cont* is the stack (the continuation), a linked list of values
;;; and return contexts.
;;;        *cont* :=  ( value ... (Context pc,code,env) ... )
;;; *thread* is the current thread.
;;; *user* is the current user.

(define *val* 'wait)
(define *type* 'wait)
(define *pc* 'wait)
(define *module* 'wait)
(define *env* 'wait)
(define *genv* 'wait)
(define *penv* 'wait)
(define *cont* 'wait)
(define *thread* 'wait)
(define *user* 'wait)

;;; Some of these registers are there only to speed up some accesses.
;;; *module* can be deduced from *genv* but is used in this simulation
;;; since *pc* is relative to the bytecode-vector of *module*. 

;;;==================================================================
;;; How to run the bytecode machine.
;;; Trace mode is controlled with *debug*.
;;; Statistics is controlled with *statistics*.

(define *debug* #f)
(define *statistics* #t)

(define (BCRT-run-byte-machine)
  (let* ((code (BCRT-Module-byte *module* *pc*))
         (size (vector-ref BCRT-instruction-size code)) )
    (if *debug* (BCRT-show-machine-state))
    (if *statistics* (BCRT-count-instruction code))
    (set! *pc* (+ *pc* size))
    (case size
      ((1) ((vector-ref BCRT-instruction-behavior code)))
      ((2) ((vector-ref BCRT-instruction-behavior code)
            (BCRT-Module-byte *module* (- *pc* 1)) ))
      ((3) ((vector-ref BCRT-instruction-behavior code)
            (BCRT-Module-byte *module* (- *pc* 2))
            (BCRT-Module-byte *module* (- *pc* 1)) )) )
    (BCRT-run-byte-machine) ) )

;;; Register which instructions are used.

(define BCRT-statistics (make-vector 256 0))

(define (BCRT-count-instruction code)
  (vector-set! BCRT-statistics code
               (+ 1 (vector-ref BCRT-statistics code)) ) )

(define (BCRT-show-statistics)
  (define total 0)
  (display "Executed instructions statistics" stderr)(newline stderr)
  (do ((i 0 (+ i 1)))
      ((>= i 256) 
    (let ((name (vector-ref BCRT-instruction-name i)))
      (if (not (eq? name 'Illegal))
          (begin
            (display name stderr)
            (display ": " stderr)
            (display (vector-ref BCRT-statistics i) stderr)
            (set! total (+ total (vector-ref BCRT-statistics i)))
            (newline stderr) ) ) ) ) )
  (display "Total number of executed instructions: " stderr)
  (display total stderr)(newline stderr) )

;;; Show the current state of the machine.

(define (BCRT-show-machine-state)
  (display "----Machine state---" stderr)(newline stderr)
  (display "VAL=" stderr)(show *val* stderr)(newline stderr)
  (show *env* stderr)
  (display "STACK=" stderr)(show *cont* stderr)(newline stderr)
  (display "PC=" stderr)(display *pc* stderr)
  (display " instruction to execute=" stderr)
  (display (BCRT-disassemble-instruction *module* *pc*) stderr)(newline stderr)
  #f )

;;;==================================================================
;;; Invoke functions.
;;; When invoked, the fresh activation frame to use is in *val*.
;;; It has one more field than strictly necessary for the sake of nary 
;;; functions. 

(define-generic (BCRT-invoke (fun BCRT)))

(define-method (BCRT-invoke (fun BCRT-Closure))
  (set! *pc* (BCRT-Closure-pc fun))
  (set-BCRT-Environment-next! *val* (BCRT-Closure-environment fun))
  (set! *genv* (BCRT-Closure-globalenv fun))
  ;; deduced register:
  (set! *module* (BCRT-Module-Environment-module *genv*))
  (set! *env* *val*) )

(define-method (BCRT-invoke (fun BCRT-Primitive))
  (let ((nargs (BCRT-Environment-value-length *val*)))
    (case (BCRT-Primitive-arity fun)
      ((0) (if (= nargs 1)
               ((BCRT-Primitive-address fun))
               (BCRT-error "Too much arguments" fun) ))
      ((1) (if (= nargs 2)
               ((BCRT-Primitive-address fun)
                (BCRT-Environment-value *val* 0) )
               (BCRT-error "Wrong arity" fun) ))
      ((2) (if (= nargs 3)
               ((BCRT-Primitive-address fun)
                (BCRT-Environment-value *val* 0)
                (BCRT-Environment-value *val* 1) )
               (BCRT-error "Wrong arity" fun) ))
      ((3) (if (= nargs 4)
               ((BCRT-Primitive-address fun)
                (BCRT-Environment-value *val* 0)
                (BCRT-Environment-value *val* 1)
                (BCRT-Environment-value *val* 2) )
               (BCRT-error "Wrong arity" fun) ))
      ((4) (if (= nargs 5)
               ((BCRT-Primitive-address fun)
                (BCRT-Environment-value *val* 0)
                (BCRT-Environment-value *val* 1)
                (BCRT-Environment-value *val* 2)
                (BCRT-Environment-value *val* 3) )
               (BCRT-error "Wrong arity" fun) ))
      ;; just give the mandatory argument. Others are in *val*.
      ((-1) (if (>= nargs 0)
                ((BCRT-Primitive-address fun))
               (BCRT-error "Wrong arity" fun) ))
      ((-2) (if (>= nargs 1)
                ((BCRT-Primitive-address fun)
                 (BCRT-Environment-value *val* 0) )
               (BCRT-error "Wrong arity" fun) ))
      ((-3) (if (>= nargs 2)
                ((BCRT-Primitive-address fun)
                 (BCRT-Environment-value *val* 0)
                 (BCRT-Environment-value *val* 1) )
               (BCRT-error "Wrong arity" fun) )) 
      (else (BCRT-error "Weird arity" fun)) ) ) )

(define-method (BCRT-invoke (k BCRT-Continuation))
  (if (not (= 2 (BCRT-Environment-value-length *val*)))
      (BCRT-error "Incorrect arity for continuation" k)
      (begin 
        (set! *cont* k)
        (set! *val* (BCRT-Environment-value *val* 0))
        (BCRT-resume *cont*) ) ) )

(define-method (BCRT-invoke (k BCRT-Escaper-Context-Continuation))
  (if (not (= 2 (BCRT-Environment-value-length *val*)))
      (BCRT-error "Incorrect arity for exit continuation") )
  (if (BCRT-verify-extent (BCRT-Escaper-Context-Continuation-next k)
                          *cont* )                                   ; HACK
      (begin
        (set! *cont* k)
        (set! *val* (BCRT-Environment-value *val* 0))
        (BCRT-resume *cont*) )
      (BCRT-error "obsolete continuation" k) ) )

(define-generic (BCRT-verify-extent k (cont))
  #f )

(define-method (BCRT-verify-extent k (cont BCRT-Continuation))
  (or (eq? k cont)
      (BCRT-verify-extent k (BCRT-Continuation-next cont)) ) )

;;; boxify following variables. This is done whenever an Environment
;;; is created to ensure the atomicity of its construction.

(define (BCRT-boxify-variables)
  (let ((code (BCRT-Module-byte *module* *pc*)))
    (if (= BC-boxify-variable-code code)
        (let ((index (BCRT-Module-byte *module* (+ *pc* 1))))
          (set-BCRT-Environment-value! 
           *env* index (make-BCRT-Box 
                        (BCRT-Environment-value *env* index) ) )
          (set! *pc* (+ *pc* 2))
          (BCRT-boxify-variables) ) ) ) )

;;; When an internal continuation is invoked, the corresponding
;;; continuation frame is still on the stack. For a Context continuation
;;; the frame is popped into the registers and popped either.

(define-generic (BCRT-resume (k)))

(define-method (BCRT-resume (k BCRT-Internal-Continuation))
  ((BCRT-Internal-Continuation-address k)) )

(define-method (BCRT-resume (k BCRT-Context-Continuation))
  (set! *genv* (BCRT-Context-Continuation-genv k))
  (set! *env* (BCRT-Context-Continuation-env k))
  ;; deduced register:
  (set! *module* (BCRT-Module-Environment-module *genv*))
  (set! *pc* (BCRT-Context-Continuation-pc k))
  (set! *cont* (BCRT-Continuation-next k)) )

;;; This method is due to the monitor compiling method.
(define-method (BCRT-resume (k BCRT-Handler-Continuation))
  (set! *cont* (BCRT-Continuation-next k))
  (BCRT-resume *cont*) )

(define-method (BCRT-resume (k BCRT-Skip-Handler-Continuation))
  (set! *cont* (BCRT-Continuation-next k))
  (BCRT-resume *cont*) )

;;;==================================================================
;;; Create the runtime global predefined immutable environment:
;;;
;;;   penv  -------> +---------+
;;;                  | unused  |
;;;                  | predef0 |
;;;                  |  ...    |
;;;                  +---------+

;;; This function converts the list of Predefined variables into a
;;; BCRT-Environment.

(define (create-BCRT-predefined-environment gvs)
  (let ((sg (allocate-BCRT-Environment (length gvs))))
    (set-BCRT-Environment-next! sg (make-BCRT-Null-Environment))
    (for-each (lambda (gv)
                (set-BCRT-Environment-value!
                 sg 
                 (BC-Predefined-Variable-rank gv)
                 (let ((desc (Predefined-Variable-description gv)) )
                   (cond
                    ((Constant-Description? desc)
                     (Constant-Description-value desc) )
                    ((BC-Functional-Description? desc)
                     (BC-Functional-Description-value desc) )
                    (else (BCRT-error "No associated value" desc)) ) ) ) )
              gvs )
    sg ) )

;;;==================================================================
;;; Convert quotations into values.

(define (convert-to-quotation o)
  (cond 
   ((eq? o #f) BCRT-false)
   ((eq? o #t) BCRT-true)
   ((eq? o '()) BCRT-nil)
   ((integer? o) (make-BCRT-fixnum o))
   ((real? o) (make-BCRT-float o))
   ((char? o) (BCRT-String-character BCRT-all-characters
                                     (char->integer o) ))
   ((pair? o)
    (make-BCRT-Pair (convert-to-quotation (cdr o))
                    (convert-to-quotation (car o)) ) )
   ((vector? o)
    (apply make-BCRT-Vector (vector-length o) 
           (map convert-to-quotation (vector->list o)) ) )
   ((string? o)
    (apply make-BCRT-String (string-length o) 
           (map convert-to-quotation (string->list o)) ) )
   ((symbol? o)
    (BCRT-string->symbol (symbol->string o)) )
   (else (Byte-Code-error "Cannot quote this value" o)) ) )

;;; A ultra-naive implementation of symbols.
;;; This alist contains (scheme-string . quoted-symbol) tuples.

(define BCRT-symbols-alist '())

(define (BCRT-string->symbol string)
  (let ((r (let search ((alist BCRT-symbols-alist))
             (if (pair? alist)
                 (if (string-ci=? string (caar alist))
                     (car alist)
                     (search (cdr alist)) )
                 #f ) )))
    (if (pair? r)
        (cdr r)
        (let* ((str (apply make-BCRT-String 
                           (string-length string)
                           (map convert-to-quotation
                                (map preferred-case (string->list string)) ) ))
               (s (make-BCRT-Symbol str)) )
          (set! BCRT-symbols-alist (cons (cons string s) BCRT-symbols-alist))
          s ) ) ) )

;;;==================================================================
;;; These are the instructions of the simulated bytecode machine.
;;; Three vectors resp. define the names, the sizes and the behavior 
;;; of the instructions. 

(define BCRT-instruction-name (make-vector 256 'Illegal))

(define BCRT-instruction-size (make-vector 256 255))

(define BCRT-instruction-behavior 
  (make-vector 256 (lambda () 
                     (BCRT-error "No such instruction") )) )

;;; Disassemble bytecode.

(define (BCRT-disassemble module)
  (let ((vlen (BCRT-Module-byte-length module)))
    (let disasm ((index 0))
      (if (< index vlen)
          (let* ((code (BCRT-Module-byte module index))
                 (size (vector-ref BCRT-instruction-size code)) )
            (cons (BCRT-disassemble-instruction module index)
                  (disasm (+ index size)) ) )
          '() ) ) ) )

(define (BCRT-disassemble-instruction module index)
  (let* ((code (BCRT-Module-byte module index))
         (size (vector-ref BCRT-instruction-size code))
         (name (vector-ref BCRT-instruction-name code)) )
    `(,name . ,(case size
                 ((1) '())
                 ((2) (list (BCRT-Module-byte module (+ index 1))))
                 ((3) (list (BCRT-Module-byte module (+ index 1))
                            (BCRT-Module-byte module (+ index 2)) ))
                 (else (BCRT-error "Weird bytecode" 
                                   module index )) ) ) ) )

;;;==================================================================
;;; This macro defines an instruction, its name, number of arguments,
;;; its associated encoding byte and its simulated "semantics".
;;; It also defines the BC-<instruction>-code variable.

(define-icslas-macro (definstruction call . body)
  (let ((code (string->symbol
               (apply string-append
                      (map symbol->string 
                           (list 'BC- (car call) '-code) ) ) )))
    `(begin
       (vector-set! BCRT-instruction-name ,code ',(car call))
       (vector-set! BCRT-instruction-size ,code ,(length call))
       (vector-set! BCRT-instruction-behavior 
                    ,code (lambda ,(cdr call) . ,body) )
       ',(car call) ) ) )

;;;==================================================================
;;; The instructions themselves.

(definstruction (nop)
  #t )

(definstruction (link-env) 
  (set-BCRT-Environment-next! *val* *env*)
  (set! *env* *val*)
  (BCRT-boxify-variables) )

(definstruction (unlink-env) 
  (set! *env* (BCRT-Environment-next *env*)) )

(definstruction (return) 
  (BCRT-resume *cont*) )

(definstruction (pop-function-and-invoke) 
  (let ((fun (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (make-BCRT-Context-Continuation
                  (BCRT-Continuation-next *cont*)
                  *pc*
                  *env*
                  *genv*
                  ))
    (BCRT-invoke fun) ) )

(definstruction (pop-function-and-goto) 
  (let ((fun (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (BCRT-invoke fun) ) )

;;; Primitives return with BCRT-resume so they expect the return address
;;; on the stack when invoked.

(definstruction (primitive-invoke rank)
  (let ((fun (BCRT-Environment-value *penv* rank)))
    (set! *cont* (make-BCRT-Context-Continuation
                  *cont*
                  *pc*
                  *env*
                  *genv*
                  ))
    (BCRT-invoke fun) ) )

(definstruction (primitive-goto rank)
  (let ((fun (BCRT-Environment-value *penv* rank)))
    (BCRT-invoke fun) ) )

(definstruction (load-nat1 i) 
  (set! *val* (convert-to-quotation i)) )

(definstruction (load-nat2 i j) 
  (set! *val* (convert-to-quotation (+ (* 256 i) j))) )

(definstruction (load-nat3 i j k l) 
  (let* ((n1 (+ (* 256 i) j))
         (n2 (+ (* 256 n1) k))
         (n  (+ (* 256 n2) l)) )
    (set! *val* n) ) )

(definstruction (load-negative-nat1 i) 
  (set! *val* (convert-to-quotation (- i))) )

(definstruction (load-negative-nat2 i j) 
  (let ((n (+ (* 256 i) j)))
    (set! *val* (convert-to-quotation (- n))) ) )

(definstruction (load-negative-nat3 i j k l) 
  (let* ((n1 (+ (* 256 i) j))
         (n2 (+ (* 256 n1) k))
         (n  (+ (* 256 n2) l)) )
    (set! *val* (convert-to-quotation (- n))) ) )

(definstruction (load-char1 i)
  (set! *val* (BCRT-String-character BCRT-all-characters i)) )

;;; Leave room for more specialized offsets ?

(definstruction (short-global-ref offset) 
  (set! *val* (BCRT-Module-Environment-element *genv* offset)) )

(definstruction (long-global-ref offset1 offset2) 
  (let ((n (+ (* 256 offset1) offset2)))
    (set! *val* (BCRT-Module-Environment-element *genv* n)) ) )

(definstruction (short-quotation-ref i) 
  (set! *val* (BCRT-Module-quotation
               (BCRT-Module-Environment-module *genv*)
               i )) )

(definstruction (long-quotation-ref i j) 
  (let ((n (+ (* 256 i) j))) 
    (set! *val* (BCRT-Module-quotation
                 (BCRT-Module-Environment-module *genv*)
                 n )) ) )

(definstruction (short-predefined-ref i) 
  (set! *val* (BCRT-Environment-value *penv* i)) )

(definstruction (if-false-short-goto+ offset) 
  (if (eq? *val* BCRT-false)
      (set! *pc* (+ *pc* offset)) ) )

(definstruction (if-false-long-goto+ offset1 offset2) 
  (let ((offset (+ (* 256 offset1) offset2)))
    (if (eq? *val* BCRT-false)
        (set! *pc* (+ *pc* offset)) ) ) )

(definstruction (if-false-long-long-goto+ offset1 offset2 offset3 offset4) 
  (let* ((n1     (+ (* 256 offset1) offset2))
         (n2     (+ (* 256 n1) offset3))
         (offset (+ (* 256 n2) offset4)) )
    (if (eq? *val* BCRT-false)
        (set! *pc* (+ *pc* offset)) ) ) )

(definstruction (short-goto+ offset) 
  (set! *pc* (+ *pc* offset)) )

(definstruction (long-goto+ offset1 offset2) 
  (let ((offset (+ (* 256 offset1) offset2)))
    (set! *pc* (+ *pc* offset)) ) )

(definstruction (long-long-goto+ offset1 offset2 offset3 offset4) 
  (let* ((n1     (+ (* 256 offset1) offset2))
         (n2     (+ (* 256 n1) offset3))
         (offset (+ (* 256 n2) offset4)) )
    (set! *pc* (+ *pc* offset)) ) )

(definstruction (box-ref) 
  (set! *val* (BCRT-Box-value *val*)) )

;;; Leave the empty box in the *val* register.

(definstruction (box-ref-and-check) 
  (let ((v (BCRT-Box-value *val*)))
    (if (eq? v BCRT-uninitialized)
        (BCRT-error "Uninitialized box" *val*)
        (set! *val* v) ) ) )

(definstruction (pop-box-and-write) 
  (let ((box (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (let ((old (BCRT-Box-value box)))
      (set-BCRT-Box-value! box *val*)
      (set! *val* old) ) ) )

(definstruction (boxify-variable index) 
  ;; This is done by BCRT-invoke directly so it cannot be interrupted.
  (BCRT-error "Cannot execute pseudo-instruction" 'boxify-variable) )

(definstruction (create-closure-and-short-goto+ offset) 
  (set! *val* (make-BCRT-closure *pc* *env* *genv*))
  (set! *pc* (+ *pc* offset)) )

(definstruction (create-closure-and-long-goto+ offset1 offset2) 
  (let ((offset (+ (* 256 offset1) offset2)))
    (set! *val* (make-BCRT-closure *pc* *env* *genv*))
    (set! *pc* (+ *pc* offset)) ) )

(definstruction (create-closure-and-long-long-goto+
                 offset1 offset2 offset3 offset4 ) 47
  (let* ((n1     (+ (* 256 offset1) offset2))
         (n2     (+ (* 256 n1) offset3))
         (offset (+ (* 256 n2) offset4)) )
    (set! *val* (make-BCRT-closure *pc* *env* *genv*))
    (set! *pc* (+ *pc* offset)) ) )

(definstruction (short-shallow-reference index) 
  (set! *val* (BCRT-Environment-value *env* index)) )

(definstruction (shallow-reference0) 
  (set! *val* (BCRT-Environment-value *env* 0)) )

(definstruction (shallow-reference1) 
  (set! *val* (BCRT-Environment-value *env* 1)) )

(definstruction (shallow-reference2) 
  (set! *val* (BCRT-Environment-value *env* 2)) )

(definstruction (shallow-reference3) 
  (set! *val* (BCRT-Environment-value *env* 3)) )

(definstruction (shallow-reference4) 
  (set! *val* (BCRT-Environment-value *env* 4)) )

;;; FUTURE internal recursion on objects that may not be present:
;;; Generate additional bytecodes ?
(definstruction (short+short-deep-reference depth index) 
  (let search ((depth depth)
               (r     *env*) )
    (if (= depth 0)
        (set! *val* (BCRT-Environment-value r index))
        (search (- depth 1) (BCRT-Environment-next r)) ) ) )

(definstruction (push-value) 
  (set! *cont* (make-BCRT-Value-Continuation *cont* *val*)) )

(definstruction (short-create-activation-frame size) 
  (set! *val* (allocate-BCRT-Environment size)) )

(definstruction (create-activation-frame0) 
  (set! *val* (allocate-BCRT-Environment 0)) )

(definstruction (create-activation-frame1) 
  (set! *val* (allocate-BCRT-Environment 1)) )

(definstruction (create-activation-frame2) 
  (set! *val* (allocate-BCRT-Environment 2)) )

(definstruction (create-activation-frame3) 
  (set! *val* (allocate-BCRT-Environment 3)) )

(definstruction (create-activation-frame4) 
  (set! *val* (allocate-BCRT-Environment 4)) )

(definstruction (short-create-activation-frame-and-push size) 
  (let ((frame (allocate-BCRT-Environment size)))
    (set-BCRT-Environment-value! frame (- size 2) *val*)
    (set! *val* frame) ) )

(definstruction (create-activation-frame1-and-push) 
  (let ((frame (allocate-BCRT-Environment 1)))
    (set! *val* frame) ) )

(definstruction (create-activation-frame2-and-push) 
  (let ((frame (allocate-BCRT-Environment 2)))
    (set-BCRT-Environment-value! frame 0 *val*)
    (set! *val* frame) ) )

(definstruction (create-activation-frame3-and-push) 
  (let ((frame (allocate-BCRT-Environment 3)))
    (set-BCRT-Environment-value! frame 1 *val*)
    (set! *val* frame) ) )

(definstruction (create-activation-frame4-and-push) 
  (let ((frame (allocate-BCRT-Environment 4)))
    (set-BCRT-Environment-value! frame 2 *val*)
    (set! *val* frame) ) )

(definstruction (short-pop-in-frame rank) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* rank val) ) )    

(definstruction (pop0-in-frame) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* 0 val) ) )    

(definstruction (pop1-in-frame) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* 1 val) ) )    

(definstruction (pop2-in-frame) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* 2 val) ) )    

(definstruction (pop3-in-frame) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* 3 val) ) )    

(definstruction (pop4-in-frame) 
  (let ((val (BCRT-Value-Continuation-value *cont*)))
    (set! *cont* (BCRT-Continuation-next *cont*))
    (set-BCRT-Environment-value! *val* 4 val) ) )

(definstruction (short-depth1-reference index) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) index)) )

(definstruction (depth1-reference0) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) 0)) )

(definstruction (depth1-reference1) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) 1)) )

(definstruction (depth1-reference2) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) 2)) )

(definstruction (depth1-reference3) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) 3)) )

(definstruction (depth1-reference4) 
  (set! *val* (BCRT-Environment-value (BCRT-Environment-next *env*) 4)) )

;;; Activation frame may hold one value more than the number of actual
;;; arguments (for the sake of n-ary functions).

(definstruction (short-check->=arity arity) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (check->=arity0) 
  (let ((got (BCRT-Environment-value-length *env*))
        (arity 0) )
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (check->=arity1) 
  (let ((got (BCRT-Environment-value-length *env*))
        (arity 1) )
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (check->=arity2) 
  (let ((got (BCRT-Environment-value-length *env*))
        (arity 2) )
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (check->=arity3) 
  (let ((got (BCRT-Environment-value-length *env*))
        (arity 3) )
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (check->=arity4) 
  (let ((got (BCRT-Environment-value-length *env*))
        (arity 4) )
    (if (not (> got arity))
        (BCRT-error "Incorrect arity (got/expected)" got arity) )
    (listify-frame! *env* arity BCRT-nil 
                    (lambda (hd tl) (make-BCRT-Pair tl hd)) )
    (BCRT-boxify-variables) ) )

(definstruction (short-check-=arity arity+1) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (= got arity+1))
        (BCRT-error "Incorrect arity (got/expected)" got arity+1) )
    (BCRT-boxify-variables) ) )

(definstruction (check-=arity1) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (= got 1))
        (BCRT-error "Incorrect arity (got/expected)" got 1) )
    (BCRT-boxify-variables) ) )

(definstruction (check-=arity2) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (= got 2))
        (BCRT-error "Incorrect arity (got/expected)" got 2) )
    (BCRT-boxify-variables) ) )

(definstruction (check-=arity3) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (= got 3))
        (BCRT-error "Incorrect arity (got/expected)" got 2) )
    (BCRT-boxify-variables) ) )

(definstruction (check-=arity4) 
  (let ((got (BCRT-Environment-value-length *env*)))
    (if (not (= got 4))
        (BCRT-error "Incorrect arity (got/expected)" got 3) )
    (BCRT-boxify-variables) ) )

;;; Select some bytecodes for important constants.

(definstruction (load-false) 
  (set! *val* BCRT-false) )

(definstruction (load-true) 
  (set! *val* BCRT-true) )

(definstruction (load-nil) 
  (set! *val* BCRT-nil) )

(definstruction (load-zero) 
  (set! *val* (convert-to-quotation 0)) )

(definstruction (load-one) 
  (set! *val* (convert-to-quotation 1)) )

;;;==================================================================

(define *all-threads* '())

(define (scheduler-call)
  (if (pair? *all-threads*)
      (let ((thread (scheduler-dequeue!)))
        (set! *thread* thread)
        (set! *user* (BCRT-Thread-user thread))
        (set! *genv* (BCRT-Thread-genv thread))
        (set! *env* (BCRT-Thread-env thread))
        ;; deduced register:
        (set! *module* (BCRT-Module-Environment-module *genv*))
        (set! *pc* (BCRT-Thread-pc thread))
        (set! *val* (BCRT-Thread-val thread))
        ;;(set! *type* (BCRT-Thread-type thread))
        (set! *cont* (BCRT-Thread-cont thread))
        thread )
      ;; if there is no more thread, the machine will stop.
      #f ) )

(define (scheduler-suspend!)
  (set-BCRT-Thread-user! *thread* *user*)
  (set-BCRT-Thread-genv! *thread* *genv*)
  (set-BCRT-Thread-env!  *thread* *env*)
  (set-BCRT-Thread-pc!   *thread* *pc*)
  (set-BCRT-Thread-val!  *thread* *val*)
  (set-BCRT-Thread-type! *thread* *type*)
  (set-BCRT-Thread-cont! *thread* *cont*)
  (scheduler-enqueue! *thread*) )

(define (scheduler-enqueue! thread)
  (set! *all-threads* (append! *all-threads*
                               (list thread) )) )

(define (scheduler-dequeue!)
  (if (pair? *all-threads*)
      (let ((new-thread (car *all-threads*)))
        (set! *all-threads* (cdr *all-threads*))
        new-thread )
      (BCRT-error "No more threads") ) )

(definstruction (suicide)
  (scheduler-call) )

(definstruction (suspend)
  (scheduler-suspend!)
  (scheduler-call) )

;;; These instructions create a new thread that will be run later.

(definstruction (fork-and-short-goto+ offset)
  (scheduler-suspend!)
  (set! *thread* (clone *thread*))
  (set! *pc* (+ *pc* offset)) )

(definstruction (fork-and-long-goto+ offset1 offset2)
  (let* ((offset (+ (* 256 offset1) offset2)))
    (scheduler-suspend!)
    (set! *thread* (clone *thread*))
    (set! *pc* (+ *pc* offset)) ) )

(definstruction (fork-and-long-long-goto+ offset1 offset2 offset3 offset4)
  (let* ((n1     (+ (* 256 offset1) offset2))
         (n2     (+ (* 256 n1) offset3))
         (offset (+ (* 256 n2) offset4)) )
    (scheduler-suspend!)
    (set! *thread* (clone *thread*))
    (set! *pc* (+ *pc* offset)) ) )

;;; Running an installed module (in *val*)

(definstruction (run-module)
  (let ((me *val*))
    (set! *cont* (make-BCRT-Context-Continuation
                  *cont* *pc* *env* *genv* ))
    (set! *genv* me)
    (set! *env* (make-BCRT-Null-Environment))
    (set! *module* (BCRT-Module-Environment-module me))
    (set! *pc* 0) ) )    

;;;  Exception handlers

(definstruction (push-handler)
  (set! *cont* (make-BCRT-Handler-Continuation *cont* *val*)) )

(definstruction (pop-handler)
  (set! *cont* (BCRT-Continuation-next *cont*)) )

;;; Bind-exit forms
;;; Since this is a binding form, it must call boxify-variables.

(definstruction (short-create-escape-frame offset)
  (let ((k (make-BCRT-Escaper-Context-Continuation
            *cont* (+ *pc* offset) 'wait *genv* )))
    (set! *env* (make-BCRT-Environment *env* 1 k))
    (set-BCRT-Escaper-Context-Continuation-env! k *env*)
    (BCRT-boxify-variables) ) )

(definstruction (long-create-escape-frame offset1 offset2)
  (let* ((offset (+ (* 256 offset1) offset2))
         (k (make-BCRT-Escaper-Context-Continuation
             *cont* (+ *pc* offset) 'wait *genv* )))
    (set! *env* (make-BCRT-Environment *env* 1 k))
    (set-BCRT-Escaper-Context-Continuation-env! k *env*)
    (BCRT-boxify-variables) ) )

(definstruction (long-long-create-escape-frame i j k l)
  (let* ((n1 (+ (* 256 i) j))
         (n2 (+ (* 256 n1) k))
         (offset (+ (* 256 n2) l))
         (k (make-BCRT-Escaper-Context-Continuation
             *cont* (+ *pc* offset) 'wait *genv* )))
    (set! *env* (make-BCRT-Environment *env* 1 k))
    (set-BCRT-Escaper-Context-Continuation-env! k *env*)
    (BCRT-boxify-variables) ) )

(definstruction (module-border)
  (BCRT-error "Cannot execute pseudo instruction" 'module-border) )

;;; end of BCinst.scm
