;;; $Id: BCtest.scm,v 1.20 1995/07/26 19:00:47 queinnec Exp $

;;; This file defines how to test the bytecode compiler.

;;;==================================================================
;;; This is the toplevel where all expressions will be evaluated.

(set! *user* (make-BCRT-User "me" (make-BCRT-Named-Env '())))

;;; Build the predefined environment, compile and load the modules of the
;;; predefined libraries. 

(define (prepare-bytecode-evaluation! g)
  (set! *penv* (create-BCRT-predefined-environment g))
  (for-each (lambda (file)
              (define xlofile (string-append file ".xlo"))
              (if (not (file-exists? xlofile))             ;; Bigloo dependent
                  (byte-compile-file icslas-expand g 
                                     (string-append file ".scm") ) )
              (BCRT-run-module
               (install-module! (BCRT-User-toplevel *user*) 
                                (file->BCRT-Module xlofile) ) ) )
            '( "Library/libscm0"
               "Library/libscm1"
               "Library/char"
               "Library/num"
               "Library/print" 
               "Others/tester"
               "Library/read"
               ) ) )

;;; Start an interpreter.
;;; This interpreter shares the same memory between toplevel interactions
;;; since the User-toplevel environment is not reinitialized.

(define (icslas)
  (prepare-bytecode-evaluation! g.scm.icslas)
  (interpreter
   "Icslas? "
   ""
   #t
   (lambda (read print err)
     (define-method (icslas-handler (exc Exception) c?) (err exc))
     (lambda ()
       (monitor icslas-handler 
         (let* ((e  (icslas-expand (read)))
                (bc (compile->Byte-code "Test" e g.scm.icslas))
                (bm (Byte-Compilation->BCRT-Module bc))
                (me (install-module! (BCRT-User-toplevel *user*) bm)) )
           (let ((results (BCRT-run-module me)))
             (BCRT-flush-port BCRT-stdout)
             (for-each (lambda (result)
                         (display "Icslas= ")
                         (show result)
                         (newline) )
                       results )
             (print "") ) ) ) ) ) ) )

;;; A driver for tests suites. It uses the tester.scm file.

(define (test-icslas file)
  (prepare-bytecode-evaluation! g.scm.icslas)
  (suite-test
   file
   "Icslas? "
   ""
   #t
   (lambda (read check err)
     (define-method (icslas-handler (exc Exception) c?) (err exc))
     (lambda ()
       (monitor icslas-handler 
         (let ((e (read)))
           ;; skip that tests
           (if (member e '( XYZZY (SET! XYZZY 3)
                            ((LAMBDA (X Y) XYZZY) 1 2) ))
               (err '***)
               (let* ((ee  `(let ((e (bind-exit (k)
                                       (monitor (lambda (exc c?)
                                                  (k (list "FAIL" exc c?)) )
                                         ,e ) )))
                              (display "Icslas= ")
                              (display e)
                              (newline)
                              e ))
                      (exp (icslas-expand ee))
                      (bc  (compile->Byte-code "test" exp g.scm.icslas))
                      (bm  (Byte-Compilation->BCRT-Module bc))
                      (me  (install-module! (BCRT-User-toplevel *user*) bm)) )
                 ;(set! *debug* #t)
                 (let ((results (BCRT-run-module me)))
                   (BCRT-flush-port BCRT-stdout)
                   (check (convert-list-of-results-to-scheme results))
                   ) ) ) ) ) ) )
   (lambda (expected obtained)
     (or (equal? expected obtained)
         (if (pair? obtained) 
             (if (null? (cdr obtained))
                 (equal? expected (car obtained))
                 (set-equal? expected obtained) )
             #f ) ) ) ) )
;;; Test: (test-icslas "Test/scheme.tst")

;;; Convert a list of results to Scheme. If one of them is a list starting
;;; with the "FAIL" string, then an error was caught by Icslas.

(define (convert-list-of-results-to-scheme values)
  (define (convert results values)
    (if (pair? values)
        (let ((r (convert-to-scheme (car values))))
          (if (and (pair? r) (equal? (car r) "FAIL"))
              '***
              (convert (cons r results) (cdr values)) ) )
        results ) )
  (convert '() values) )

;;; This new tester uses tester.scm in Icslas not in the underlying Scheme.

(define (new-test-icslas file)
  (prepare-bytecode-evaluation! g.scm.icslas)
  (let* ((e `(begin 
               (suite-test ,file
                "Icslas? " "Icslas= " #t
                (lambda (read print error)
                  (lambda () 
                    (monitor (lambda (e c?) (err e))
                      (print (eval (read))) ) ) )
                (lambda (expected obtained)
                  (or (equal? expected obtained)
                      (if (pair? obtained) 
                          (if (null? (cdr obtained))
                              (equal? expected (car obtained))
                              (set-equal? expected obtained) )
                          #f ) ) ) ) ))
         (exp (icslas-expand e))
         (bc  (compile->Byte-code "fulltest" exp g.scm.icslas))
         (bm  (Byte-Compilation->BCRT-Module bc))
         (me  (install-module! (BCRT-User-toplevel *user*) bm)) )
    ;(set! *debug* #t)
    (let ((results (BCRT-run-module me)))
      (map convert-to-scheme results) ) )) )
;;; Test: (new-test-icslas "Test/scheme.tst")


;;;==================================================================
;;; The error handler.
;;; A method is dynamically added to it when (icslas) or (test-icslas..) is 
;;; invoked. An uncaught exception just exits these functions. Normally
;;; the default method of icslas-handler should never be used since the
;;; dynamically added method should catch any raised exception.

(define-generic (icslas-handler (exc) continuable?)
  (basic-handler exc continuable?) )

;;; Use the basic handler which knows how to display warnings and reports.

(define-method (icslas-handler (exc Compilation-Warning) continuable?)
  (basic-handler exc continuable?) )

;;; Consider all non-initialized variables as non initialized.

(define-method (icslas-handler (exc Uninitialized-Variable) c?)
  (if c? #f (call-next-method)) )

;;; Catch anomalies raised by the Bytecode machine.

(define-method (icslas-handler (exc BCRT-Anomaly) c?)
  (display "-------- ByteCode Anomaly -------")(newline)
  (show (BCRT-Anomaly-msg exc))(newline)
  (show (BCRT-Anomaly-hints exc))(newline)
  (diagnose exc c?) )

;;;==================================================================
;;; Convert an Icslas value into a Scheme value (needed for tests).

(define-generic (convert-to-scheme (o))
  (cond ((integer? o) o)
        (else (diagnose (list "Cannot convert back to Scheme" o) #f)) ) )

(define-method (convert-to-scheme (o BCRT-Value))
  (BCRT-Value-scheme o) )

(define-method (convert-to-scheme (o BCRT))
  o )

(define-method (convert-to-scheme (o BCRT-Symbol))
  (string->symbol (convert-to-scheme (BCRT-Symbol-name o))) )

(define-method (convert-to-scheme (o BCRT-String))
  (let* ((size   (BCRT-String-character-length o))
         (result (make-string size)) )
    (do ((i 0 (+ i 1)))
        ((>= i size))
      (string-set! result i (convert-to-scheme (BCRT-String-character o i))) )
    result ) )

(define-method (convert-to-scheme (o BCRT-Pair))
  (cons (convert-to-scheme (BCRT-Pair-head o))
        (convert-to-scheme (BCRT-Pair-tail o)) ) )

(define-method (convert-to-scheme (o BCRT-Vector))
  (let* ((size   (BCRT-Vector-item-length o))
         (result (make-vector size)) )
    (do ((i 0 (+ i 1)))
        ((>= i size))
      (vector-set! result i (convert-to-scheme (BCRT-Vector-item o i))) )
    result ) )

(define-method (convert-to-scheme (o BCRT-Character))
  (integer->char (BCRT-Character-code o)) )

(define-method (convert-to-scheme (o BCRT-number))
  (BCRT-number-value o) )

;;;==================================================================
;;; Some methods to see objects.

(define-method (show (o Global-Variable) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "#<GlobalVariable " stream)
    (display (Global-Variable-name o) stream)
    (display ">" stream) ) )

(define-method (show (o BCRT-Environment) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (let ((n (BCRT-Environment-value-length o)))
      (do ((i 0 (+ i 1)))
          ((>= i n))
        (display "ENV[" stream)
        (display i stream)
        (display "]=" stream)
        (show (BCRT-Environment-value o i) stream)
        (newline) ) ) ) )

(define-method (show (o BCRT-Null-Environment) . stream)
  #f ) 

(define-method (show (o BCRT-Continuation) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (call-next-method)
    (display " ")
    (show (BCRT-Continuation-next o) stream) ) )

(define-method (show (o BCRT-Primitive) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "#<Primitive:" stream)
    (display (BCRT-Primitive-name o) stream)
    (display ">" stream) ) )

(define-method (show (o BCRT-Pair) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "{" stream)
    (show (BCRT-Pair-head o) stream)
    (display " . " stream)
    (show (BCRT-Pair-tail o) stream)
    (display "}" stream) ) )

(define-method (show (o BCRT-Vector) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port)))
        (len (BCRT-Vector-item-length o)) )
    (display "#{" stream)
    (do ((i 0 (+ i  1)))
        ((>= i len))
      (display " " stream)
      (show (BCRT-Vector-item o i) stream) )
    (display "}" stream) ) )

(define-method (show (o BCRT-Value) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (cond
     ((eq? o BCRT-false) (display "#!false" stream))
     ((eq? o BCRT-true) (display "#!true" stream))
     ((eq? o BCRT-nil) (display "{}" stream))
     ((eq? o BCRT-end-of-file) (display "#<EOF>" stream))
     ((eq? o BCRT-uninitialized) (display "#{UFO}" stream))
     (else (display o stream)) ) ) )

(define-method (show (o BCRT-Symbol) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "|" stream)
    (display (convert-to-scheme (BCRT-Symbol-name o)) stream)
    (display "|" stream) ) )

(define-method (show (o BCRT-String) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port)))
        (n (BCRT-String-character-length o)) )
    (display "\"" stream)
    (do ((i 0 (+ i 1)))
        ((>= i n))
      (show (convert-to-scheme (BCRT-String-character o i)) stream) )
    (display "\"" stream) ) )

(define-method (show (o BCRT-Closure) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "#<Closure(pc=" stream)
    (show (BCRT-Closure-pc o) stream)
    (display "@" stream)
    (show (BCRT-Module-name 
           (BCRT-Module-Environment-module
            (BCRT-Closure-globalenv o) ) ) stream )
    (display ")>" stream) ) )

(define-method (show (o BCRT-Number) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display (BCRT-Number-value o) stream) ) )

(define-method (show (o BCRT-Box) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "#<a" stream)
    (if (eq? BCRT-uninitialized (BCRT-Box-value o))
        (display "n empty" stream) )
    (display " Box>" stream) ) )

(define-method (show (o BCRT-Named-Box) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "#<a" stream)
    (if (eq? BCRT-uninitialized (BCRT-Box-value o))
        (display "n empty" stream) )
    (display " Box(" stream)
    (show (BCRT-Named-Box-name o) stream)
    (display ")>" stream) ) )

(define-method (show (o BCRT-Named-Env) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (for-each (lambda (b) (show b stream))
              (BCRT-Named-Env-mapping o) )
    (newline stream) ) )

(define-method (show (o BCRT-User) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display "--------------- User: " stream)
    (show (BCRT-User-name o) stream)
    (display " ---------------" stream)
    (newline stream)
    (show (BCRT-User-toplevel o) stream) ) )

(define-method (show (o BCRT-Module-Environment) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port)))
        (n (BCRT-Module-Environment-element-length o)) )
    (display "Shared boxes: " stream)
    (newline stream)
    (do ((i 0 (+ i 1)))
        ((>= i n))
      (show `(,i . ,(BCRT-Module-Environment-element o i)) stream) )
    (newline stream)
    (show (BCRT-Module-Environment-module o) stream) ) )

(define-method (show (o BCRT-Module) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port)))
        (qn (BCRT-Module-quotation-length o))
        (bn (BCRT-Module-byte-length o))
        (vn (BCRT-Module-varname-length o)) )
    (display "--------- Module " stream)
    (show (BCRT-Module-name o) stream)
    (display " ---------------" stream)
    (newline stream)
    (display "Quotations: " stream)
    (do ((i 0 (+ i 1)))
        ((>= i qn))
      (show `(,i . ,(BCRT-Module-quotation o i)) stream) )
    (newline stream)
    (display "Variables: " stream)
    (do ((i 0 (+ i 1)))
        ((>= i vn))
      (show `(,i . ,(BCRT-Module-varname o i)) stream) )
    (newline stream)
    (display "Bytecode: " stream)
    (newline)
    (for-each (lambda (i) (display i stream) (newline stream))
              (BCRT-disassemble o) ) ) )

;;; end of BCtest.scm
