/* $Id: dmxla.c,v 1.30 1998/06/03 17:55:51 queinnec Exp $ */

#include "xla.h"

DM_DefineStringWithCstring(dmxla_version,
  "$Id: dmxla.c,v 1.30 1998/06/03 17:55:51 queinnec Exp $");

/* {{{ Global variables
 * RESOURCE DMxla_context
 * This is a context object to be used when no other context is available.
 * It is only used in (sort of) signal handlers.

 * A similar variable exists in dmscm.c
 * It is initialized in DM_icslas_initialize.
 */

DM_Context* DMxla_context = DM_NULL;

/* }}}
 * Convert a DM_SupportFixnum into an allocated Fixnum.
 * Convert a DM_SupportFloatnum into an allocated Floatnum.
 */

DM_SupportPointer
DM_support_fixnum_to_fixnum (DM_Context* context,
                             DM_SupportFixnum i) {
  DM_Int4 *tmp = DM_Allocate0(context, Int4);

  tmp->value = i;
  return DM_ToSupport(tmp);
}

DM_SupportPointer
DM_support_floatnum_to_floatnum (DM_Context* context,
                                 DM_SupportFloatnum flo) {
  DM_Float2 *tmp = DM_Allocate0(context, Float2);

  tmp->value = flo;
  return DM_ToSupport(tmp);
}

/******************************************************************
 * Allocation.
 * This is an unsafe allocator since the higher levels ensures that
 * <class> is a Class and <size> is a natural number. The size of the 
 * DMeroon prefix is automatically added to the previous size.

 NOTE: On Sony, Boehm's GC_malloc return addresses that are multiple
 of 4 whereas, alignment of the most stringent type (as handled by
 gcc) is 8.  For C compatibility since the kernel classes are defined
 in C and must be readable by DMeroon, DM_allocate_bare_instance must
 provide addresses that are multiple of 8. This would be a waste of
 room if performed everywhere, so I optimize this as follows. Such a
 stringent alignment is only needed when there are fields of size 8 or
 at least two indexed fields. So we correct the address returned by
 malloc for these cases. This information is coded in the
 DM_CLASS_IS_STRINGENT bit. 

 */

DM_Object*
DM_allocate_bare_instance (DM_Context *context,
                           DM_Class   *class, 
                           DM_nat3    size ) {
  DMeroon_object *p;

  if ( DM_HasOption(class, DM_CLASS_IS_STRINGENT) ) {
    size += DM_ROOM_FOR_REALIGNMENT;
    p = DM_MALLOC(DM_PrefixSize() + size);
    if ( p ) {
      p = (void*)DM_OffsetRealign((DM_SupportAddress)p);
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  } else {
    p = DM_MALLOC(DM_PrefixSize() + size);
    if ( p ) {
      /* NOTHING */
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  }

#if defined(DM_NON_ZEROING_MALLOC)
  /* DM_MALLOC returns instances that should be zeroed since DMeroon
   * expects fresh instances to be filled with zeroes. */
  memset(p, 0, DM_PrefixSize() + size);
#endif

  return DM_FromC(p);
}

#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC
/******************************************************************
 *                    Cooperation with GC
 *
 * When Exit instances are to be recycled, a message must be sent to the
 * remotely referenced Entry. This requires cooperation with the GC.
 * Fortunately, Boehm GC provides this ability.
 */

void
DM_exit_finalizer (void *exit, void *client_data) {
  /* Remove the Exit instance from the Exits hashtable */
  DM_remove_exit(DMxla_context, exit);
}

#endif

void
DM_set_exit_finalization (DM_Exit *exit) {
#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC
  GC_register_finalizer(exit, DM_exit_finalizer, NULL, NULL, NULL);
#endif
}



/******************************************************************
 * These are global mutable variables for the internal use of 
 * exception handling.
 */

DM_Anomaly *DM_exception = NULL;
jmp_buf    *DM_jmp_buf   = NULL;

void
DM_raise_exception (DM_Context *context,
                    DM_nat3    code,
                    char       *file,
                    DM_nat3    line)
{

  DM_DebugAtLevel(10,
    fflush(stdout);
    fprintf(stderr, "Error %lu in %s:%lu\n",
            (unsigned long)code,
            file,
            (unsigned long)line );
);

  { DM_Anomaly *anomaly = DM_Allocate0(DMxla_context, Anomaly);
    
    anomaly->code  = code;
    DM_exception   = anomaly;

    if ( DM_jmp_buf != NULL ) {
      longjmp(*DM_jmp_buf, DM_JUMPBUF_EXCEPTION);
    } else {
      fprintf(stderr,"No exception handler\n");
      exit(1);
    }
  }
}

/* RESOURCE DM_print_char_star
 * The generic printer on which all DMeroon and Icslas print routines
 * are based. It knows how to write a C string into a Port or a C stream.
 * It does not flush the Port (but it does on a C stream).

 * NOTE: for long strings, would be better to have `start' and `len'
 * to allow to chop them in small pieces. There is a problem on strings
 * that are larger than the buffer ???                 FUTURE.
 */

void
DM_print_char_star (DM_Context        *context,
                    char              *string,
                    DM_nat3           len,
                    void              *wrapped_port) {
  DM_nat3       n = len ? len : strlen(string);

  if ( wrapped_port == NULL ) return;

  DM_DebugAtLevel(0,
    if (   (wrapped_port == (DM_SupportPointer)stdout)
        || (wrapped_port == (DM_SupportPointer)stderr) ) {
      FILE *stream = (FILE*) wrapped_port;

      fprintf(stream, "%*s", (int)n, string);
      fflush(stream);
      return;
    }
);

  { DM_OutBuffer *outbuffer = DM_FromSupport(wrapped_port);
    if ( DM_IsA(context, outbuffer, OutBuffer) ) {
      DM_print_char_star_in_outbuffer(context, string, n, outbuffer);
      return;
    }
  }

  DM_Abort(context, DME_PRINT_CHAR_STAR_PROBLEM);
}

/******************************************************************
 * Iinitialization of DMeroon for Icslas.
 */

void
DM_icslas_initialize (DM_Context* context) {

  /* Avoid a warning telling that this variable is useless. */
  DM_dmxla_version_string_object = DM_dmxla_version_string_object;
  
#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC
  /* Advise Boehm GC that DMeroon uses interior pointers.
   * Also advise that we may correct misalignment (see allocate_bare_instance).
   */
  GC_register_displacement(DM_PrefixSize());
  GC_register_displacement((1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);
  GC_register_displacement(DM_PrefixSize()
                           + (1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);
#endif

  /* Initialize the rest of the API (in order). */
  DM_scm_initialize(context);
  DM_newapi_initialize(context);

  /* Then initialize the DMxla_context variable. */
  DMxla_context = DM_AllocateContext(context);

  /* Initialize the rest of the binding. */
  DM_xlaenv_initialize(context);
  DM_xlaobj_initialize(context);

}

/* end of dmxla.c */

