/* $Id: xlabci.c,v 1.29 1998/03/19 19:03:41 queinnec Exp $ 
 *
 * This is the ByteCode interpreter written in C.
*/

#include "xla.h"

DM_DefineStringWithCstring(xlabci_version,
  "$Id: xlabci.c,v 1.29 1998/03/19 19:03:41 queinnec Exp $");

/* RESOURCE DM_listify_frame
 * An internal function that takes an activation frame and gathers
 * in a list all arguments past the arity-th one. This is for n-ary
 * Scheme functions.
 * All activation frames are allocated with an extra slot ie f(x,y,z) 
 * is invoked with a 4-slot activation frame. If the arity of f is 
 * "at least one" then y and z must be gathered in a list. The activation
 * frame will be converted from <next,4,x,y,z,void> to 
 * <next,4,x,(y z),void,void>. The frame has an extra slot for the case
 * where f is called with a single argument ie f(x). The activation frame
 * looks like <next,2,x,void> and is converted into <next,2,x,()>. Hence
 * the requirement to hold an extra slot.
 
 * NOTE: Pay attention, in the argc,argv convention argv is a pointer
 * to a vector of DM_Object*. Note also that argc is the exact
 * number of arguments (and not the size of the Environment frame
 * which is one more).

 * arity should be the rank in the frame where must be stored the list
 * gathering extra arguments. Rank is zero-based!
 */

static void
DM_listify_frame (DM_Context*    context,
                  DM_nat3        argc,
                  DM_Environment *env,
                  DM_nat2        arity ) {
  DM_Object *result = DM_TheConstant(nil);
  
  while ( argc > arity ) {

#ifdef DM_ICSLAS_PREFERS_MUTABILITY
    DM_MutablePair   *r = DM_Allocate0(context, MutablePair);
    r->value = env->value[--argc];
    r->rest  = result;
#else
    DM_ImmutablePair *r = DM_Allocate0(context, ImmutablePair);
    r->value = (DM_ImmutableLink*)(env->value[--argc]);
    r->rest  = result;
#endif

    result   = (DM_Object*)r;
  }
  env->value[arity] = result;
}

/* RESOURCE DM_byte_eval
 * The bytecode interpreter.
 * Trace of the Bytecode machine is controlled with DM_DEBUG and DM_trace.
 */

#ifdef DM_DEBUG
DM_nat1 DM_trace = 0;

DM_SupportVoid
DM_toggle_trace (void) {
  DM_trace = 1 - DM_trace;
  return DM_SupportDone;
}
#endif /* !DM_DEBUG */

DM_Object*
DM_byte_eval (DM_Context* context,
              DM_Thread*  thread ) {
  /* Don't care if they are clobbered. */
  DM_nat3              pc      = thread->pc;
  DM_Object            *val    = thread->val;
  DM_Environment       *env    = thread->env;
  DM_ModuleEnvironment *genv   = thread->genv;
  DM_Continuation      *cont   = thread->cont;
  DM_Module            *module = genv->module;
  DM_nat3              instruction_count  = 0;
  DM_nat1              bytecode;
  DM_Object             *fun;
  DM_nat1              i, j, k, l, arity;
  DM_int3              gotten;
  jmp_buf              jb;
  int                  status;

  DM_DebugAtLevel(0,
    DM_ImmutableVector *instr = 
      (DM_ImmutableVector*)(DM_ReferenceTo(DM_instruction_names));
    DM_ImmutableByteVector *sizes = 
      (DM_ImmutableByteVector*)(DM_ReferenceTo(DM_instruction_sizes));
    int m;
);

  /* Prepare exception handling */
  /* reset de ces variables quand on sort de bci           FUTURE */
  DM_jmp_buf = &jb;
  DM_exception = NULL;
  status = setjmp(jb);
  switch ( status ) {
  case 0: {
    thread->jumpbuf = &jb;
    break;
  }
  case DM_JUMPBUF_EXCEPTION: {
    val = (DM_Object*)DM_exception;
    DM_exception = NULL;
    goto handle_exception;
  }
  case DM_JUMPBUF_ESCAPE: {
    /* Normally wrapped_thread should not be clobbered by longjmp. */
    return val;
  }
  }
  
 loop:
  instruction_count++;
  bytecode = module->byte[pc++];

  DM_DebugAtLevel(0,
    if ( DM_trace ) {
      fprintf(stderr, "\t-------Machine state-------\n");
      fprintf(stderr, "val = ");
      DM_show_object(context, val, (DM_SupportPointer)stderr);
      fprintf(stderr, "\nenv =  ");
      DM_show_object(context, (DM_Object*)env, (DM_SupportPointer)stderr);
      fprintf(stderr, "\n");
      fprintf(stderr, "\ncont =  ");
      DM_show_object(context, (DM_Object*)cont, (DM_SupportPointer)stderr);
      fprintf(stderr, "\npc = %u, instruction to run = ", pc);    
      DM_show_object(context, instr->item[bytecode], (DM_SupportPointer)stderr);
      fprintf(stderr, " ie. (");
      for ( m=0 ; m<sizes->item[bytecode] ; m++ ) {
        fprintf(stderr, " %d", module->byte[pc-1+m]);
      }
      fprintf(stderr, " )\n");
      fflush(stderr);
    }
);
  
 loop_on_special_bytecode :
  switch (bytecode) {
  /*                                  look at generated asm    FUTURE */

  case DM_SUSPEND_CODE : {
    thread->pc   = pc;
    thread->val  = val;
    thread->env  = env;
    thread->genv = genv;
    thread->cont = cont;
    return (DM_Object*)thread;
  }

  case DM_SUICIDE_CODE : {
    return DM_FALSE;
  }

  case DM_NOP_CODE : {
    goto loop;
  }

  case DM_HANDLE_EXCEPTION_CODE : {
    handle_exception :                     /* TEMPORARY */

      DM_DebugAtLevel(0,
        exit(2);
);
    return DM_Abort(context, DME_EXCEPTION_CAUGHT);
  }

  case DM_LINK_ENV_CODE : {
   ((DM_Environment*)val)->next = env;
   env = (DM_Environment*)val;
   goto boxify_variables;
 }

  case DM_UNLINK_ENV_CODE : {
    env = env->next;
    goto loop;
  }

  case DM_RETURN_CODE : {
    resume :
    if ( DM_IsA(context, cont, ContextContinuation) ) {
      DM_DeclareNarrowedObject(k, ContextContinuation, cont);
      
      env    = k->env;
      genv   = k->genv;
      module = genv->module;
      pc     = k->pc;
      cont   = k->next;
      goto loop;
    } 
    if ( DM_IsA(context, cont, InternalContinuation) ) {
      thread->pc   = pc;
      thread->val  = val;
      thread->env  = env;
      thread->genv = genv;
      thread->cont = cont;
      ((DM_InternalContinuation*)cont)->address(thread);
      goto loop;
    }
    /* Otherwise skip the continuation */
    cont = cont->next;
    goto resume;
  }

  case DM_POP_FUNCTION_AND_INVOKE_CODE : {
    DM_ContextContinuation *k = DM_Allocate0(context, ContextContinuation);

    fun     = ((DM_Object*)((DM_ValueContinuation*)cont)->value);
    k->next = cont->next;
    k->pc   = pc;
    k->env  = env;
    k->genv = genv;
    cont    = (DM_Continuation*)k;

    invoke :
    if ( DM_IsA(context,  fun, Closure) ) {
      DM_Closure *f = (DM_Closure*)fun;

      pc                           = f->pc;
      ((DM_Environment*)val)->next = f->environment;
      genv                         = f->globalenv;
      module                       = ((DM_ModuleEnvironment*)genv)->module;
      env                          = (DM_Environment*)val;
      goto loop;
    } else 
    
    if ( DM_IsA(context, fun, Primitive) ) {
      DM_int2 nargs   = ((DM_Environment*)val)->value_length;
      DM_int2 arity   = ((DM_Primitive*)fun)->arity;
      DM_code address = ((DM_Primitive*)fun)->address;

      if ( arity>=0 ) {
        if ( arity != (nargs-1) ) {
          DM_Abort(context, DME_WRONG_ARITY);
        } else {
          switch (arity) {
          case 0 : {
            val = ((DM (*)(void)) *address)();
            goto resume;
          } 
          case 1 : {
            DM_Object* a = ((DM_Environment*)val)->value[0];
            val = ((DM (*)(DM)) *address)(a);
            goto resume;
          }
          case 2 : {
            DM_Object* a = ((DM_Environment*)val)->value[0];
            DM_Object* b = ((DM_Environment*)val)->value[1];
            val = ((DM (*)(DM,DM)) *address)(a,b);
            goto resume;
          }
          case 3 : {
            DM_Object* a = ((DM_Environment*)val)->value[0];
            DM_Object* b = ((DM_Environment*)val)->value[1];
            DM_Object* c = ((DM_Environment*)val)->value[2];
            val = ((DM (*)(DM,DM,DM)) *address)(a,b,c);
            goto resume;
          }
          case 4 : {
            DM_Object* a = ((DM_Environment*)val)->value[0];
            DM_Object* b = ((DM_Environment*)val)->value[1];
            DM_Object* c = ((DM_Environment*)val)->value[2];
            DM_Object* d = ((DM_Environment*)val)->value[3];
            val = ((DM (*)(DM,DM,DM,DM)) *address)(a,b,c,d);
            goto resume;
          }
          default: 
            DM_Abort(context, DME_WEIRD_POSITIVE_ARITY);
          }
        }
      } else {
        if ( nargs >= (-(arity+1)) ) {
          DM_Object* argv[DM_MAXIMUM_ARGUMENTS_NUMBER];
          DM_nat3    argc = nargs-1;
          DM_nat3    i;

          for ( i=0 ; i<argc ; i++ ) 
            argv[i] = ((DM_Environment*)val)->value[i];
          val = ((DM (*)(DM_nat3,DM[1])) *address)(argc, argv);
          goto resume;
        } else {
          DM_Abort(context, DME_WRONG_ARITY);
        }
      }
    } else

    /* This is how call/cc, apply are invoked. There is a special bytecode
     * so their definition lies in the current function so they can access
     * all its internal variables. */
    if ( DM_IsA(context, fun, SpecialPrimitive) ) {
      DM_int2 nargs   = ((DM_Environment*)val)->value_length;
      DM_int2 arity   = ((DM_SpecialPrimitive*)fun)->arity;
      
      if ( ( arity >= 0 ) && (arity == (nargs-1) ) ) {
        bytecode = ((DM_SpecialPrimitive*)fun)->bytecode;
        goto loop_on_special_bytecode;
      } 
      if ( ( arity < 0 ) && ( nargs > (- (arity+1) ) ) ) {
        bytecode = ((DM_SpecialPrimitive*)fun)->bytecode;
        goto loop_on_special_bytecode;
      } 
      DM_Abort(context, DME_WRONG_ARITY);
    }

    if ( DM_IsA(context, fun, Continuation) ) {
      DM_int2 nargs   = ((DM_Environment*)val)->value_length;
      
      if ( nargs == 2 ) {
        cont = (DM_Continuation*)fun;
        val = ((DM_Environment*)val)->value[0];
        goto resume;
      } else {
        DM_Abort(context, DME_WRONG_CONTINUATION_ARITY);
      }
    } else {
      DM_Abort(context, DME_CANNOT_APPLY);
    }
  }

  case DM_POP_FUNCTION_AND_GOTO_CODE : {
    fun  = ((DM_ValueContinuation*)cont)->value;
    cont = cont->next;
    goto invoke;
  }

  case DM_PRIMITIVE_INVOKE_CODE : {
    DM_ContextContinuation *k = DM_Allocate0(context, ContextContinuation);

    k->next = cont;
    k->env  = env;
    k->genv = genv;
    cont    = (DM_Continuation*)k;
    i       = module->byte[pc++];
    k->pc   = pc;
    fun     = DM_Primitive(i);

    DM_DebugAtLevel(0,
      if ( fun == DM_TheConstant(uninitialized) )
        DM_Abort(context, DME_MISSING_PRIMITIVE);
);
    goto invoke;
  }

  case DM_PRIMITIVE_GOTO_CODE : {
    i   = module->byte[pc++];
    fun = DM_Primitive(i);

    DM_DebugAtLevel(0,
      if ( fun == DM_TheConstant(uninitialized) )
        DM_Abort(context, DME_MISSING_PRIMITIVE);
);
    goto invoke;
  }

  case DM_LOAD_CHAR1_CODE : {
    i   = module->byte[pc++];
    val = (DM_Object*)(((DM_ImmutableVector*)
                       (DM_ReferenceTo(DM_characters)))->item[i]);
    goto loop;
  }

  case DM_LOAD_NAT1_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    result->value = i;
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_LOAD_NAT2_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    j             = module->byte[pc++];
    result->value = (256*i) + j;
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_LOAD_NAT3_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    j             = module->byte[pc++];
    k             = module->byte[pc++];
    l             = module->byte[pc++];
    result->value = 256*(256*((256*i) + j) + k) + l;
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_LOAD_NEGATIVE_NAT1_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    result->value = -i;
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_LOAD_NEGATIVE_NAT2_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    j             = module->byte[pc++];
    result->value = -((256*i) + j);
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_LOAD_NEGATIVE_NAT3_CODE : {
    DM_Int4 *result = DM_Allocate0(context, Int4);
    
    i             = module->byte[pc++];
    j             = module->byte[pc++];
    k             = module->byte[pc++];
    l             = module->byte[pc++];
    result->value = -(256*(256*((256*i) + j) + k) + l);
    val           = (DM_Object*)result;
    goto loop;
  }

  case DM_SHORT_GLOBAL_REF_CODE : {
    i   = module->byte[pc++];
    val = (DM_Object*)(genv->element[i]);
    goto loop;
  }

  case DM_LONG_GLOBAL_REF_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    val = (DM_Object*)(genv->element[256*i + j]);
    goto loop;
  }

  case DM_SHORT_QUOTATION_REF_CODE : {
    i   = module->byte[pc++];
    /* Normally this access should be error freee. */
    DMeroon_indexed_field_value(context,
                                &val,
                                NULL, 
                                (DM_Object*)module,
                                DM_ReferenceToFieldOfClass(quotation, Module),
                                i );
    goto loop;
  }

  case DM_SHORT_PREDEFINED_REF_CODE : {
    i   = module->byte[pc++];
    val = DM_Primitive(i);
    goto loop;
  }
    
  case DM_IF_FALSE_SHORT_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    if ( val == DM_SupportFalse ) pc = pc + i;
    goto loop;
  }

  case DM_IF_FALSE_LONG_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    if ( val == DM_SupportFalse ) pc = pc + 256*i + j;
    goto loop;
  }

  case DM_IF_FALSE_LONG_LONG_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    k   = module->byte[pc++];
    l   = module->byte[pc++];
    if ( val == DM_SupportFalse ) 
      pc = pc + (256*(256*((256*i) + j) + k) + l);
    goto loop;
  }

  case DM_SHORT_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    pc = pc + i;
    goto loop;
  }

  case DM_LONG_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    pc = pc + 256*i + j;
    goto loop;
  }

  case DM_LONG_LONG_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    k   = module->byte[pc++];
    l   = module->byte[pc++];
    pc = pc + (256*(256*((256*i) + j) + k) + l);
    goto loop;
  }

  case DM_BOX_REF_CODE : {
    val = ((DM_Box*)val)->value;
    goto loop;
  }

  case DM_BOX_REF_AND_CHECK_CODE : {
    DM_Object* result = ((DM_Box*)val)->value;
    if ( result == DM_TheConstant(uninitialized) ) {

      DM_DebugAtLevel(5,
        fprintf(stderr,"Uninitialized box " DM_0X "%p",val);
        if ( DM_IsA(context, val, NamedBox) ) 
          fprintf(stderr," (%s)",((DM_String*)
                                  ((DM_NamedBox*)val)->name)->character );
        fprintf(stderr,"\n");
);

      DM_Abort(context, DME_UNINITIALIZED_BOX);
    } else {
      val = result;
      goto loop;
    }
  }

  case DM_POP_BOX_AND_WRITE_CODE: {
    DM_Box*   box = (DM_Box*)((DM_ValueContinuation*)cont)->value;
    DM_Object* old = box->value;

    cont = cont->next;
    box->value = val;
    val = old;
    goto loop;
  }

  case DM_BOXIFY_VARIABLE_CODE : {
    DM_Abort(context, DME_PSEUDO_INSTRUCTION);
  }
    
  case DM_CREATE_CLOSURE_AND_SHORT_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    val = (DM_Object*)(DM_Allocate0(context, Closure));
    ((DM_Closure*)val)->pc = pc;
    ((DM_Closure*)val)->environment = env;
    ((DM_Closure*)val)->globalenv = genv;
    pc = pc + i;
    goto loop;
  }

  case DM_CREATE_CLOSURE_AND_LONG_GOTO_FORWARD_CODE : {
    i   = module->byte[pc++];
    j   = module->byte[pc++];
    val = (DM_Object*)(DM_Allocate0(context, Closure));
    ((DM_Closure*)val)->pc = pc;
    ((DM_Closure*)val)->environment = env;
    ((DM_Closure*)val)->globalenv = genv;
    pc = pc + 256*i + j;
    goto loop;
  }

  case DM_SHORT_SHALLOW_REFERENCE_CODE : {
    i   = module->byte[pc++];
    val = env->value[i];
    goto loop;
  }

  case DM_SHALLOW_REFERENCE0_CODE : {
    val = env->value[0];
    goto loop;
  }

  case DM_SHALLOW_REFERENCE1_CODE : {
    val = env->value[1];
    goto loop;
  }

  case DM_SHALLOW_REFERENCE2_CODE : {
    val = env->value[2];
    goto loop;
  }

  case DM_SHALLOW_REFERENCE3_CODE : {
    val = env->value[3];
    goto loop;
  }

  case DM_SHALLOW_REFERENCE4_CODE : {
    val = env->value[4];
    goto loop;
  }

  case DM_SHORT_WITH_SHORT_DEEP_REFERENCE_CODE : {
    DM_nat1 depth = module->byte[pc++];
    DM_nat1 i     = module->byte[pc++];
    /*                                              TEMP */
    val = (DM_Object*)env;
    while ( depth-- > 0 ) {
      val = (DM_Object*)(((DM_Environment*)val)->next);
    }
    val = ((DM_Environment*)val)->value[i];
    goto loop;
  }

  case DM_PUSH_VALUE_CODE : {
    DM_ValueContinuation *k = DM_Allocate0(context, ValueContinuation);

    k->next = (DM_Continuation*)cont;
    k->value = val;
    cont = (DM_Continuation*)k;
    goto loop;
  }

  case DM_SHORT_CREATE_ACTIVATION_FRAME_CODE : {
    i   = module->byte[pc++];
    { val = (DM_Object*) DM_Allocate1(context, Environment, i);

    DM_DebugAtLevel(8,
        { int j;
          /* Need to initialize this frame to print while debugging. */
          for ( j=0 ; j<i ; j++ )
            ((DM_Environment*)val)->value[j] = DM_TheConstant(uninitialized);
        }
);

      goto loop;
    }
  }

  case DM_CREATE_ACTIVATION_FRAME0_CODE : {
    val = (DM_Object*)(DM_Allocate1(context, Environment,0));
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME1_CODE : {
    val = (DM_Object*)(DM_Allocate1(context, Environment,1));

    DM_DebugAtLevel(8,
      ((DM_Environment*)val)->value[0] = DM_TheConstant(uninitialized);
);
     
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME2_CODE : {
    val = (DM_Object*)(DM_Allocate1(context, Environment,2));

    DM_DebugAtLevel(8,
      ((DM_Environment*)val)->value[0] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[1] = DM_TheConstant(uninitialized);
);
 
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME3_CODE : {
    val = (DM_Object*)(DM_Allocate1(context, Environment,3));

    DM_DebugAtLevel(8,
      ((DM_Environment*)val)->value[0] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[1] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[2] = DM_TheConstant(uninitialized);
);
 
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME4_CODE : {
    val = (DM_Object*)(DM_Allocate1(context, Environment,4));

    DM_DebugAtLevel(8,
      ((DM_Environment*)val)->value[0] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[1] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[2] = DM_TheConstant(uninitialized);
      ((DM_Environment*)val)->value[3] = DM_TheConstant(uninitialized);
);
 
    goto loop;
  }

  case DM_SHORT_CREATE_ACTIVATION_FRAME_AND_PUSH_CODE : {
    i   = module->byte[pc++];
    { DM_Environment *env = DM_Allocate1(context, Environment,i);

    DM_DebugAtLevel(8,
        { int j;
          /* Need to initialize this frame to print while debugging. */
          for ( j=0 ; j<i ; j++ )
            ((DM_Environment*)env)->value[j] = DM_TheConstant(uninitialized);
        }
);
 
      env->value[i-2] = val;
      val = (DM_Object*)env;
    }
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME1_AND_PUSH_CODE : {
    { DM_Environment* env = DM_Allocate1(context, Environment,1);

    DM_DebugAtLevel(8,
      ((DM_Environment*)env)->value[0] = DM_TheConstant(uninitialized);
);
 
      val = (DM_Object*)env;
    }
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME2_AND_PUSH_CODE : {
    { DM_Environment* env = DM_Allocate1(context, Environment,2);

    DM_DebugAtLevel(8,
      ((DM_Environment*)env)->value[1] = DM_TheConstant(uninitialized);
);
 
      env->value[0] = val;
      val = (DM_Object*)env;
    }
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME3_AND_PUSH_CODE : {
    { DM_Environment* env = DM_Allocate1(context, Environment,3);

    DM_DebugAtLevel(8,
      ((DM_Environment*)env)->value[0] = DM_TheConstant(uninitialized);
      ((DM_Environment*)env)->value[2] = DM_TheConstant(uninitialized);
);
 
      env->value[1] = val;
      val = (DM_Object*)env;
    }
    goto loop;
  }

  case DM_CREATE_ACTIVATION_FRAME4_AND_PUSH_CODE : {
    { DM_Environment* env = DM_Allocate1(context, Environment,4);

    DM_DebugAtLevel(8,
      ((DM_Environment*)env)->value[0] = DM_TheConstant(uninitialized);
      ((DM_Environment*)env)->value[1] = DM_TheConstant(uninitialized);
      ((DM_Environment*)env)->value[3] = DM_TheConstant(uninitialized);
);
 
      env->value[2] = val;
      val = (DM_Object*)env;
    }
    goto loop;
  }

  case DM_SHORT_POP_IN_FRAME_CODE : {
    i   = module->byte[pc++];
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[i] = v;
    }
    goto loop;
  }

 case DM_POP0_IN_FRAME_CODE : {
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[0] = v;
    }
    goto loop;
  }

 case DM_POP1_IN_FRAME_CODE : {
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[1] = v;
    }
    goto loop;
  }

 case DM_POP2_IN_FRAME_CODE : {
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[2] = v;
    }
    goto loop;
  }

 case DM_POP3_IN_FRAME_CODE : {
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[3] = v;
    }
    goto loop;
  }

 case DM_POP4_IN_FRAME_CODE : {
    { DM_Object* v = ((DM_ValueContinuation*)cont)->value;
      cont = cont->next;
      ((DM_Environment*)val)->value[4] = v;
    }
    goto loop;
  }

  case DM_SHORT_DEPTH1_REFERENCE_CODE : {
    i   = module->byte[pc++];
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[i];
    goto loop;
  }

  case DM_DEPTH1_REFERENCE0_CODE : {
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[0];
    goto loop;
  }

  case DM_DEPTH1_REFERENCE1_CODE : {
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[1];
    goto loop;
  }

  case DM_DEPTH1_REFERENCE2_CODE : {
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[2];
    goto loop;
  }

  case DM_DEPTH1_REFERENCE3_CODE : {
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[3];
    goto loop;
  }

  case DM_DEPTH1_REFERENCE4_CODE : {
    /*                                             locality ??? */
    val = ((DM_Environment*)(env->next))->value[4];
    goto loop;
  }

  case DM_SHORT_CHECK_GREATER_EQUAL_ARITY_CODE : {
    arity  = module->byte[pc++]; 
    gotten = env->value_length;
    if ( gotten <= arity ) 
      DM_Abort(context, DME_WRONG_ARITY);
    DM_listify_frame(context, gotten-1, env, arity);

    boxify_variables : {
      bytecode = module->byte[pc];
      if ( bytecode == DM_BOXIFY_VARIABLE_CODE ) {
        DM_Box *box = DM_Allocate0(context, Box);

        instruction_count++; 
        pc++;
        i = module->byte[pc++];
        box->value = env->value[i];
        env->value[i] = (DM_Object*)box;
        goto boxify_variables;
      } else goto loop;
    }
  }

  case DM_CHECK_GREATER_EQUAL_ARITY0_CODE : {
    arity  = 0;
    gotten = env->value_length;
    if ( gotten > arity ) {
      DM_listify_frame(context, gotten-1, env, arity);
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_GREATER_EQUAL_ARITY1_CODE : {
    arity  = 1;
    gotten = env->value_length;
    if ( gotten > arity ) {
      DM_listify_frame(context, gotten-1, env, arity);
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_GREATER_EQUAL_ARITY2_CODE : {
    arity  = 2;
    gotten = env->value_length;
    if ( gotten > arity ) {
      DM_listify_frame(context, gotten-1, env, arity);
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_GREATER_EQUAL_ARITY3_CODE : {
    arity  = 3;
    gotten = env->value_length;
    if ( gotten > arity ) {
      DM_listify_frame(context, gotten-1, env, arity);
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_GREATER_EQUAL_ARITY4_CODE : {
    arity  = 4;
    gotten = env->value_length;
    if ( gotten > arity ) {
      DM_listify_frame(context, gotten-1, env, arity);
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_SHORT_CHECK_EQUAL_ARITY_CODE : {
    arity = module->byte[pc++];
    gotten = env->value_length;
    /* Here, arity is already shifted by one. */
    if ( gotten == arity ) {
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_EQUAL_ARITY1_CODE : {
    arity = 0;
    gotten = env->value_length;
    if ( gotten == (arity+1) ) {
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_EQUAL_ARITY2_CODE : {
    arity = 1;
    gotten = env->value_length;
    if ( gotten == (arity+1) ) {
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_EQUAL_ARITY3_CODE : {
    arity = 2;
    gotten = env->value_length;
    if ( gotten == (arity+1) ) {
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_CHECK_EQUAL_ARITY4_CODE : {
    arity = 3;
    gotten = env->value_length;
    if ( gotten == (arity+1) ) {
      goto boxify_variables;
    } else {
      DM_Abort(context, DME_WRONG_ARITY);
    }
  }

  case DM_LOAD_FALSE_CODE : {
    val = DM_SupportFalse;
    goto loop;
  }

  case DM_LOAD_TRUE_CODE : {
    val = DM_SupportTrue;
    goto loop;
  }

  case DM_LOAD_NIL_CODE : {
    val = DM_TheConstant(nil);
    goto loop;
  }

  case DM_LOAD_ZERO_CODE : {
    val = DM_ReferenceTo(DM_fixnum_0);
    goto loop;
  }

  case DM_LOAD_ONE_CODE : {
    val = DM_ReferenceTo(DM_fixnum_1);
    goto loop;
  }

    /* fork and others                    TEMP */

  case DM_PUSH_HANDLER_CODE : {
    DM_HandlerContinuation *hc = DM_Allocate0(context, HandlerContinuation);
    
    hc->next = cont;
    hc->handler = val;
    cont = (DM_Continuation*)hc;
    goto loop;
  }

  case DM_POP_HANDLER_CODE : {
    cont = cont->next;
    goto loop;
  }

  case DM_SHORT_CREATE_ESCAPE_FRAME_CODE : {
    DM_EscaperContextContinuation *ecc =
      DM_Allocate0(context, EscaperContextContinuation);
    DM_Environment* frame = DM_Allocate1(context, Environment,1);

    i                   = module->byte[pc++];
    frame->next         = env;
    frame->value[0]     = (DM_Object*)ecc;
    ecc->next           = cont;
    ecc->pc             = pc + i;
    ecc->env            = frame;
    ecc->genv           = genv;
    goto boxify_variables;
  }

  case DM_CALLCC_CODE : {
    DM_Environment *frame = DM_Allocate1(context, Environment,2);

    frame->value[0] = (DM_Object*)cont;
    fun = ((DM_Environment*)val)->value[0];
    val = (DM_Object*)frame;
    goto invoke;
  }

  case DM_APPLY_CODE : {
    DM_nat3           nargs     = ((DM_Environment*)val)->value_length;
    DM_Object          *last_arg = ((DM_Environment*)val)->value[nargs-2];
    DM_Object          *list     = last_arg;
    DM_nat3           len       = 0;
    DM_Environment   *frame;
    DM_nat3          i;
    
    fun = ((DM_Environment*)val)->value[0];
    while (    DM_IsA(context, list, ImmutablePair) 
            || DM_IsA(context, list, MutablePair) ) {
      list = ((DM_MutablePair*)list)->rest;
      len++;
    }
    frame = DM_Allocate1(context, Environment, nargs + len - 2);
    for ( i=0 ; i<(nargs-3) ; i++ ) 
      frame->value[i] = ((DM_Environment*)val)->value[i+1];
    list = last_arg;
    while (    DM_IsA(context, list, ImmutablePair )
            || DM_IsA(context, list, MutablePair ) ) {
      frame->value[i++] = ((DM_MutablePair*)list)->value;
      list = ((DM_MutablePair*)list)->rest;
    }
    val = (DM_Object*)frame;
    goto invoke;
  }

  default: {
    DM_Abort(context, DME_ILLEGAL_BYTECODE);
  }
  }
  /* Normally not reached. */
  return DM_SupportDone;
}

void
DM_xlabci_initialize (DM_Context *context) {

  /* Just to avoid a warning. */
  DM_xlabci_version_string_object = DM_xlabci_version_string_object;

}  

/* end of xlabci.c */
