/* $Id: xlaboot.c,v 1.32 1998/06/01 19:30:41 queinnec Exp $ 
 
 * This file defines how to start Icslas as a stand-alone program.
 * It provides a lot of options to indicate where to get the initial
 * module to run.
 */

#include "xla.h"
#include "xlaboot.h"

DM_DefineStringWithCstring(xlaboot_version,
  "$Id: xlaboot.c,v 1.32 1998/06/01 19:30:41 queinnec Exp $");

/* RESOURCE DM_open_boot_port
 * Opens a file and return a Port instance (with its associated Buffer).
 * If there is a problem, return DM_NULL. This function is very similar
 * to DM_open_input_port and share the same method for initialization.
 */

DM_Object*
DM_open_boot_port (DM_Context* context, 
                   char *filename ) {
  int fd;
    
  DM_ResetErrno();
  if ( (fd = open(filename, O_RDONLY, DM_OPEN_MODMASK)) < 0 ) {
    return DM_NULL;
  } else {
    DM_BootPort *bootport = DM_Allocate0(context, BootPort);
    
    bootport->number  = fd;
    (void) DM_establish_input_port(context, (DM_InputPort*)bootport);
    return (DM_Object*)bootport;
  }
}

/* 
 * A function to interact with the ByteCode interpreter.
 * DM_return_continuation returns the value the thread returns.
 */

static void
DM_return (DM_Context* context, DM_Thread* thread) {

  DM_DebugAtLevel(0,
    DM_show_object(context, thread->val, (DM_SupportPointer)stderr);
    fprintf(stderr, "\n");
);

  /* Return thread->val to the caller of DM_byte_eval() which will
   * then extract the returned value from the thread object. */
  longjmp(thread->jumpbuf, DM_JUMPBUF_ESCAPE);
}

/* RESOURCE DM_eval_module
 * Evaluate a Module object in the context of the current User.
 * This function returns the result of the evaluation.
 */

DM_Object*
DM_eval_module (DM_Context* context, 
                DM_Object*  module ) {
  DM_User *user = (DM_User*)(DM_ReferenceTo(DM_user));
  volatile DM_Thread *thread = DM_Allocate0(context, Thread);
  volatile DM_Object* result;
  DM_ModuleEnvironment *me;
  DM_InternalContinuation* k = DM_Allocate0(context, InternalContinuation);
  
  me = DM_link_module(context, user, (DM_Module*)module);
  thread->user   = user;
  thread->env    = (void*)(DM_ReferenceTo(DM_null_environment));
  thread->pc     = 0;
  thread->val    = DM_ReferenceTo(DM_unspecified);
  k->address     = DM_CodeOf(DM_return);
  k->next        = DM_NULL;               /* Normally useless */
  thread->cont   = (DM_Continuation*)k;
  thread->genv   = me;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, k);
    DM_SetSharable(context, thread);
  }
  DM_ProtectWithFreshContext(context, newcontext,
             result = DM_byte_eval(newcontext, (DM_Thread*)thread),
       /* Something went wrong. */
       result = DM_SupportFalse;
                           );
  return (DM_Object*)result;
}

/* RESOURCE DM_work_until_calm
 * Enter the DMeroon event loop and try to evaluate the incoming Modules.
 * Exits when no objects arrive for a given duration. This function returns
 * the value of the last evaluated module.
 */

DM_Object*
DM_work_until_calm (DM_Context* context, 
                    DM_nat3     duration ) {
  DM_Object* m;
  DM_Object* result = DM_SupportFalse;
  DM_ErrorCode errcode;

  do {
    errcode = DMeroon_receive(context, &m, duration);
    /* Reserved for dmscm.c: DM_CheckReturnedStatus(context, errcode); */
    if ( DM_IsA(context, m, Module) ) {
      result = DM_eval_module(context, m);
    } else { 

      DM_DebugAtLevel(2,
        fprintf(DM_FindErrorStream(context), "Ignored a non-Module.\n");
);
      
    }
  } while ( m );
  return result;
}

/* This is the specialized function whose goal is to create a port
 * through which Icslas waits for a Module to evaluate. This is the
 * protocol for booting Icslas from a file.
 */

void
DM_run_file (DM_Context* context, 
             char *file) {
  DM_BootPort* bootport;
  DM_Object*   result;

  bootport = (DM_BootPort*)(DM_open_boot_port(context, file));
  if ( ! bootport ) {
    fprintf(stderr, "No such file: %s\n", file);
    exit(7);
  }
  while ( DM_HasNotOption(bootport, DM_PORT_IS_DRIED) ) {
    result = DM_work_until_calm(context, 1);
  }
}

/* This is the string that answers the -help option. */

#define DM_HELP_OPTIONS_STRING \
 "This is Icslas and DMeroon.\n" \
 "Options may be repeated. They are:\n" \
 "  -help                         displays this text.\n" \
 "  -exit                         exits immediately.\n" \
 "  -banner                       displays a banner.\n" \
 "  -file <filename>              evaluates the module present in that\n" \
 "                                file (there may be more than one).\n" \
 "  -inspect \"inspector\"          forks an external inspector to examine\n" \
 "                                DMeroon data.\n" \
 "  -linger <duration_in_seconds> serves DMeroon requests until a calm\n" \
 "                                of that duration.\n" \
 "  -trace                        traces the progress of the bytecode\n" \
 "                                interpreter.\n"

/* RESOURCE main
 * The main entry point of the Icslas interpreter.
 * It parses the options to decide how to boot. It is possible to boot
 * from a file or a tcp/ip connection. These options may be repeated.
 */

int
main (int argc, char *argv[]) {
  DM_Context* context = DM_NULL;
  DM_Site*    site;
  DM_ErrorCode errcode;
  int i;

  /* To avoid warning */
  DM_xlaboot_version_string_object =   DM_xlaboot_version_string_object;

  /* First initialize DMeroon and its Icslas binding. */
  errcode = DMeroon_initialize(context, &site);
  if ( errcode != DME_SUCCESS ) {
    fprintf(DM_debug_stream, "\tCannot initialize DMeroon (error=%lu).\n",
            (unsigned long)errcode );
    exit(5);
  }
  context = DM_AllocateContext(context);
  DM_xlaenv_initialize(context);
  DM_xlaobj_initialize(context);

  /* then analyze options. */
  for ( i=1 ; i<argc ; i++ ) {

    if ( strcmp(argv[i], "-help") == 0 ) {
      fprintf(stderr, DM_HELP_OPTIONS_STRING);
      exit(0);
      continue;
    }

    if ( strcmp(argv[i], "-banner") == 0 ) {
      fprintf(stderr, "Icslas $Revision: 1.32 $\n");
      continue;
    }

    if ( strcmp(argv[i], "-exit") == 0 ) {
      /* Be sure to stop the interpreter now! */
      exit(0);
    }

    if ( strcmp(argv[i], "-trace") == 0 ) {
      /* Show the evaluation trace of the ByteCode interpreter. */

      DM_DebugAtLevel(0,
        DM_trace = 1;
);

      continue;
    }

    if ( strcmp(argv[i], "-file") == 0 ) {
      /* Evaluate the Module instance marshaled in a file. 
       * The ``-file <file>'' option may be repeated ad libitum.  */
      if ( i<argc-1 ) {
        DM_run_file(context, argv[++i]);
        continue;
      } else {
        fprintf(stderr, "Missing filename\n");
        exit(3);
      }
    }

    if ( strcmp(argv[i], "-linger") == 0 ) {
      /* Ask Icslas to die after this duration if no computation to
       * perform occurs meanwhile. If -linger is the last option, then
       * Icslas will last a long long time. */
      DM_nat3 duration = (i==argc-1) 
                        ? DM_INFINITE_DURATION 
                        : DM_string_to_nat3(context, argv[++i], NULL);

      DM_work_until_calm(context, duration);
      continue;
    }

    DM_DebugAtLevel(0,
      if ( strcmp(argv[i], "-inspect") == 0 ) {
        /* Ask Icslas to fork an inspector. If the string naming the inspector
         * is empty, then use stdin to read debug commands. If the -inspect
         * option is the last one, then consider it as equivalent to "". */
        char *str = (i==argc-1) ? "" : argv[++i];

        if ( strlen(str) == 0 ) {
          fprintf(DM_FindErrorStream(context), 
                  " Entering debug monitor...\n");
          DM_debug_monitor(context, 0);
        } else {
          fprintf(DM_FindErrorStream(context), 
                  " Forking external inspector %s...\n", str);
          DM_fork_inspector(context, str);
        }
        continue;
      }
);

    if ( strcmp(argv[i], "-connect") == 0 ) {
      /* Ask Icslas to connect another site. */
      if ( i<argc-2 ) {
        DM_Site *site;
        char    *sitename = argv[++i];
        char    *portnum  = argv[++i];
        DM_nat2 port      = DM_string_to_nat3(context, portnum, NULL);
        DM_ErrorCode errcode;

        fprintf(DM_FindErrorStream(context), 
                " Connecting to server %s:%d...\n",
                sitename, port);
        errcode = DMeroon_connect(context, &site, sitename, port);
        /* Reserved for dmscm.c: DM_CheckReturnedStatus(context, errcode); */
        continue;
      } else {
        fprintf(DM_FindErrorStream(context), 
                " Missing host/port parameters");
        exit(3);
      }
    }

    fprintf(stderr, "Unrecognized option %s\n", argv[i]);
    exit(1);
  }
  fprintf(stderr, "Ite Icslas est.\n");
  exit(0);
}

/* end of xlaboot.c */
