;;; $Id: Ccomp.scm,v 1.7 1996/08/28 17:47:49 queinnec Exp $

;;; This file defines functions to compile a whole file into a .o file.

;;; Exportations:
;;; end of exportations.

;;;==================================================================
;;; A class for compilation commands.

(define-class Compilation Object
  ( source      ; a filename containing Scheme code
    verbose?    ; a boolean (tells if the compiler is verbose or not)
    ) )

(define-class C-Compilation Compilation
  ( cfile       ; the file where to put C code
    ofile       ; the associated .o file
    ) )

;;; This function may act as an application entry point with options
;;; being the list of options given on the Un*x command.  Compile a
;;; file to C and possibly to a stand-alone.
;;;    cxlas file.scm [-o a.out] [-C file.c] 

(define (compiler-main options)
  (monitor compiler-main-handler
    (let ((cmd (make-C-Compilation #f #f #f #f)))
      (scan-options! cmd options)
      (let ((e `(begin . ,(file->list (Compilation-source cmd)))))
        (when (Compilation-verbose? cmd)
          (display "Expanding...")(newline) )
        (set! e (walk e enriched-icslas-expansion-environment))
        (when (Compilation-verbose? cmd)
          (display "Generate C code...")(newline) )
        (call-with-output-file (C-Compilation-cfile cmd)
          (lambda (out) (compile->C e (predefined-global-env) out)) ) )
      (when (Compilation-verbose? cmd)
        (display "Calling C compiler...")(newline) )
      (call-external-compiler cmd)
      #t ) ) )

;;; The handler to monitor exceptions while compiling.
;;; By default, call the wrapping handler.

(define-generic (compiler-main-handler (exc) c?)
  (diagnose exc c?) )

(define-method (compiler-main-handler (exc Compilation-Exception) c?)
  (flush-output-port stdout)
  (display "   *** *** *** " stderr)
  (display (Compilation-Exception-phase exc) stderr)
  (display " Exception *** *** ***" stderr)
  (newline stderr)
  (display (Compilation-Exception-message exc) stderr)
  (newline stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (newline stderr) )
            (Compilation-Exception-culprits exc) )
  (call-next-method) )

;;; Just prints the warning and resume the diagnoser.

(define-method (compiler-main-handler (exc Compilation-Warning) c?)
  (flush-output-port stdout)
  (display (Compilation-Warning-phase exc) stderr)
  (display " Warning: " stderr)
  (display (Compilation-Warning-message exc) stderr)
  (newline stderr)
  (for-each (lambda (hint) 
              (show hint stderr)
              (newline stderr) )
            (Compilation-Warning-culprits exc) ) )

;;; Forward the exception.

(define-method (compiler-main-handler (exc Exception) c?)
  (flush-output-port stdout)
  (display "!!!??? an exception occurred ???!!!" stderr)
  (newline stderr)
  (display exc stderr)
  (call-next-method) )

;;;==================================================================
;;; Show some specific objects.

(define-method (show (o Variable) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (display (Variable-name o) stream) ) )

;;;==================================================================
;;; How to invoke the C compiler.
;;; This supposes that the TARGET variable is set and tells where is the
;;; directory containing the compiled runtime library. This is the case
;;; if the compiler is started with cxla.

(define-generic (call-external-compiler (cmd)))

;;; The C compiler has to be parameterized with these variables.  If
;;; generating with the syntactic extensions offered by gcc, the
;;; output C is no longer ISO-compliant.

(define C-compiler 
  (if generate-for-gcc
      "gcc "
      "gcc -ansi -pedantic " ) )

;;; Just check that this variable is defined.

TARGET

(define C-flags 
  (string-append " -Wall -O -gg -I" TARGET "/SchemeRT ") )

(define C-libraries 
  (string-append " " 
                 TARGET "/SchemeRT/scheme.o "
                 TARGET (if generate-cps-code
                            "/SchemeRT/schemeklib.o "
                            "/SchemeRT/schemelib.o " ) ) )

(define-method (call-external-compiler (cmd C-Compilation))
  (let* ((status 0)
         (cd-cmd (string-append 
                  "cd " 
                  (extract-directory (C-Compilation-cfile cmd)) ) )
         (cc-cmd (string-append C-compiler
                                C-flags
                                (extract-basefilename
                                 (C-Compilation-cfile cmd) )
                                C-libraries ))
         (complete-shell-command (string-append cd-cmd " ; " cc-cmd)) )
    (set! status (system complete-shell-command))
    (unless (= (modulo status 256) 0)
      (compiler-main-error "C compilation problem" status) )
    cmd ) )

(define compiler-main-error
  (create-compilation-exception-raiser 
   "Code generation" ) )

;;; Process the options.

(define (scan-options! cmd options)
  (let scan ((options options))
    (if (pair? options)
        (cond ((equal? (car options) "-o") 
               (if (pair? (cdr options))
                   (begin (set-C-Compilation-ofile! cmd (cadr options))
                          (scan cmd (cddr options)) )
                   (compiler-main-error "Missing file.o" options) ))
              ((equal? (car options) "-C") 
               (if (pair? (cdr options))
                   (begin (set-C-Compilation-cfile! cmd (cadr options))
                          (scan (cddr options)) )
                   (compiler-main-error "Missing file.c" options) ))
              ((equal? (car options) "-v")
               (set-Compilation-verbose?! cmd #t)
               (scan (cdr options)) )
              (else (if (Compilation-source cmd)
                        (compiler-main-error "Too much file.scm" options)
                        (begin (set-Compilation-source! 
                                cmd (car options) )
                               (scan (cdr options)) ) )) ) ) )
  (unless (Compilation-source cmd)
    (compiler-main-error "No file to compile" cmd) )
  (unless (C-Compilation-cfile cmd)
    (set-C-Compilation-cfile! 
     cmd (compute-default-generated-file cmd) ) )
  cmd )

;;; Compute by default the name of the file to be produced.

(define (compute-default-generated-file cmd)
  (let ((file (Compilation-source cmd)))
    (string-append (extract-filename-sans-extension file)
                   ".c" ) ) )

;;; end of Ccomp.scm
