;;; $Id: Cgen.scm,v 1.11 1996/08/29 15:33:54 queinnec Exp $

;;; Compilation towards C

;;; Exportations:
;;;         (compile->C Sexp Global-Environment output-port)
;;; end of exportations.

;;; Importations:
;;; end of importations.

;;; Enrich the class of Linearized-Flattened-Program to hold the
;;; original code to compile.

(define-class C-Program Linearized-Flattened-Program
  ( original       ; original expression
    globalenv      ; the global environment that was used.
    ) )

;;; More than one generation is possible.

(define generate-for-gcc 
  (if-icslas-feature gcc #t #f) )

(define generate-cps-code
  (if-icslas-feature cps #t #f) )

;;;==================================================================
;;; This function performs the C code generation. It takes a Sexp to
;;; compile, the current global environment and the output-port to write.
;;; Note that the expression to compile is supposed to already be expanded.

(define (compile->C e g out)
  (monitor compile->C-handler
      (let ((prg (prepare-expression e g)))
        (->C prg out) ) ) )

;;; This is the general generic function to convert a program onto a
;;; port towards a C file.

(define-generic (->C (e) out))

(define-method (->C (prg C-Program) out)
    (generate-header             prg out)
    (generate-global-environment prg out)
    (generate-quotations         prg out)
    (generate-functions          prg out)
    (generate-main               prg out)
    (generate-trailer            prg out)
    prg )

;;; Transform an Sexpression to a C-Program.

(define (prepare-expression e g)
  (let ((o (objectify e (make-Local-Environment '()) g)))
    ;; Alphatize variable names (to make it easier for C)
    (set! o (alphatize! o))
    ;; A-normalize
    (set! o (a-normalize o))
    ;; Add the Box transformation to remove mutable local variables:
    (set! o (insert-box! o))
    ;; Figure which global variables are surely initialized
    (set! o (initialization-analyze! o))
    ;; Closurize the whole program
    (set! o (make-Regular-Application
             (make-No-Argument)
             (make-Function '() o) ))
    ;; Collect temporary variables (turn Function into With-Temp-Function). 
    ;; DON'T perform this transformation if you want a SCHEME behavior!
    ;; You don't need it with gcc.
    (if (not generate-for-gcc)
        (set! o (gather-temporaries! o)) )
    ;; Lambda-lift in place (turn With-Temp-Function into Flat-Function)
    (set! o (lift! o))
    ;; Migrate quotations and functions towards the highest level.
    (set! o (extract-things! o))
    ;; linearize quotations
    (set! o (linearize-quotations! o))
    ;;  Return the result of all these transformations after recording
    ;;  the initial parameters of this compilation.
    (make-C-Program
     (Flattened-Program-body o)
     (Flattened-Program-quotations o)
     (Flattened-Program-definitions o)
     (Linearized-Flattened-Program-linearizations o)
     (Linearized-Flattened-Program-initialization o)
     e
     g ) ) )

;;;==================================================================
;;; This is the error handler used while compiling. It catches the
;;; Unknown-Variable exception to define it on the fly but all other
;;; exceptions are re-raised.

(define-generic (compile->C-handler (exc) c?)
  (diagnose exc c?) )

;;; Define new global variables on the fly.

(define-method (compile->C-handler (exc Unknown-Variable) c?)
  (let ((v (make-Global-Variable (Unknown-Variable-name exc) #f))
        (g (Unknown-Variable-g exc)) )
    (set-Global-Environment-global!
     g (cons v (Global-Environment-Global g)) )
    (make-Global-Reference v) ) )

;;; Takes care of the exceptions raised by the initialization analysis.
;;; This means that the variable is by default not initialized.

(define-method (compile->C-handler (exc Uninitialized-Variable) c?)
  #f )

;;;==================================================================
;;;              Generation

;;; File header and trailer are simple.

(define (generate-header prg out)
  (let ((e (C-Program-original prg)))
    (format out "/* Compiler to C $Revision: 1.11 $ ~%")
    (format out " * generate-for-gcc = ~A, generate-cps-code = ~A. ~%~%"
            generate-for-gcc generate-cps-code )
    (pp e out) ; DEBUG
    (format out "~% */~%~%#include \"scheme.h\"~%") ) )

(define (generate-trailer prg out)
  (format out "~%/* End of generated code. */~%") )

;;;==================================================================
;;; Generate global environment. For all Global-Variables allocate a
;;; location to hold it. We also give the real name to the macro so
;;; that some reflective capabilities can be added to C.

;;; gv* is a list of instances of Global-Variable.

(define (generate-global-environment prg out)
  (let* ((globalenv (C-Program-globalenv prg))
         (gv*       (Global-Environment-global globalenv)) )
    (when (pair? gv*)
      (format out "~%/* Global environment: */~%")
      (for-each (lambda (gv) (generate-global-variable out gv))
                gv* )) ) )

(define (generate-global-variable out gv)
  (let ((name (Global-Variable-name gv)))
    (format out "SCM_DefineGlobalVariable(~A,\"~A\");~%"
            (IdScheme->IdC name) name ) ) )

;;; Convert a Scheme name to a C name. All predefined names will be
;;; inserted in the following map so they can be uniformly named. The
;;; map is an Alist ( (symbol . "Cname") ... ).

(define Scheme->C-names-mapping 
  '( (*        . "TIMES")
     (<        . "LESSP")
     (pair?    . "CONSP")
     (set-cdr! . "RPLACD")
     ) )

(set! Scheme->C-names-mapping 
  '( (*        . "TIMES")
     (+        . "PLUS")
     (-        . "DIFFERENCE")
     (/        . "QUOTIENT")
     (>        . "GREATERP")
     (>=       . "NOT_LESSP")
     (<        . "LESSP")
     (<=       . "NOT_GREATERP")
     (=        . "EQN")
     (eq?      . "EQ")
     (pair?    . "CONSP")
     (symbol?  . "SYMBOLP")
     (set-car! . "RPLACA")
     (set-cdr! . "RPLACD")
     ) )

(define (IdScheme->IdC name)
  (let ((v (assq name Scheme->C-names-mapping)))
    (if (pair? v) (cdr v)
        (let ((str (symbol->string name)))
          (let retry ((Cname (compute-Cname str)))
            (if (Cname-clash? Cname Scheme->C-names-mapping)
                (retry (compute-another-Cname str))
                (begin (set! Scheme->C-names-mapping
                             (cons (cons name Cname) 
                                   Scheme->C-names-mapping ) )
                       Cname ) ) ) ) ) ) )

(define (Cname-clash? Cname mapping)
  (let check ((mapping mapping))
    (and (pair? mapping)
         (or (string=? Cname (cdr (car mapping)))
             (check (cdr mapping)) ) ) ) )

;;; These functions compute a C name for a symbol. Scheme symbols
;;; cannot be transleted into a name containing an isolated underscore
;;; so all these names will be used for C generation purposes.

(define compute-another-Cname 
  (let ((counter 1))
    (lambda (str)
      (set! counter (+ 1 counter))
      (compute-Cname (format #f "~A_~A" str counter)) ) ) )

(define (compute-Cname str)
  (define (mapcan f l)
    (if (pair? l)
        (append (f (car l)) (mapcan f (cdr l)))
        '() ) )
  (define (convert-char char)
    (case char
      ((#\_)             '(#\_ #\_))
      ((#\?)             '(#\p))
      ((#\!)             '(#\i))
      ((#\<)             '(#\l))
      ((#\>)             '(#\g))
      ((#\=)             '(#\e))
      ((#\- #\/ #\* #\:) '())
      (else              (list char)) ) )
  (let ((cname (mapcan convert-char (string->list str))))
    (if (pair? cname) (list->string cname) "weird") ) )

;;;==================================================================
;;; Generate static shared quotations. Linearize quotations so they
;;; can be expressed by C constants. Equal? quotations are shared.
;;; And quotations here are completely static.

;;; Linearized quotations are named thing<i> while quotations are named
;;; quotation<i>.

(define (generate-quotations prg out)
  (let ((lqv*  (Linearized-Flattened-Program-linearizations prg)))
    (when (pair? lqv*)
      (format out "~%/* Quotations: */~%")
      (for-each (lambda (lqv) (->C lqv out)) lqv*)
      ;; since lqv* is a pair, the initialization is not #f
      (generate-quotations-aliases 
       (Linearized-Flattened-Program-initialization prg)
       out ) ) ) )

(define-generic (generate-quotations-aliases (form) out))

(define-method (generate-quotations-aliases (form Sequence) out)
  (generate-quotations-aliases (Sequence-first form) out)
  (generate-quotations-aliases (Sequence-last form) out) )

(define-method (generate-quotations-aliases (form Global-Assignment) out)
  (let ((qv (Global-Assignment-variable form))
        (lqv (Reference-variable (Global-Assignment-form form))) )
    (format out "#define quotation~A thing~A~%"
            (Quotation-Variable-name qv)
            (Linearized-Quotation-Variable-name lqv) ) ) )
          
(define-method (->C (qv Symbol-Quotation-Variable) out)
  (let ((strqv (Symbol-Quotation-Variable-string qv)))
    (format out "SCM_DefineSymbol(thing~A_object,thing~A);       /* ~S */~%"
            (Quotation-Variable-name qv)
            (Quotation-Variable-name strqv)
            (Quotation-Variable-value qv) )
    (format out "#define thing~A SCM_Wrap(&thing~A_object)~%"
            (Quotation-Variable-name qv) (Quotation-Variable-name qv) ) ) )

(define-method (->C (qv Cons-Quotation-Variable) out)
  (let ((aqv (Cons-Quotation-Variable-car qv))
        (dqv (Cons-Quotation-Variable-cdr qv)) )
    (format out "SCM_DefinePair(thing~A_object,thing~A,thing~A);  /* ~S */~%"
            (Quotation-Variable-name qv)
            (Quotation-Variable-name aqv)
            (Quotation-Variable-name dqv)
            (Quotation-Variable-value qv) )
    (format out "#define thing~A SCM_Wrap(&thing~A_object)~%"
            (Quotation-Variable-name qv) (Quotation-Variable-name qv) ) ) )

(define-method (->C (qv Vector-Quotation-Variable) out)
  (compile-error "Not yet implemented" o) )

(define-method (->C (qv Immediate-Quotation-Variable) out)
  (let ((value (Quotation-Variable-value qv))
        (index (Quotation-Variable-name qv)) )
    (cond ((null? value)
           (format out "#define thing~A SCM_nil	/* () */~%"
                   index ) )
          ((boolean? value)
           (format out "#define thing~A ~A	/* ~S */~%"
                   index (if value "SCM_true" "SCM_false") value ) )
          ((integer? value)
           (format out "#define thing~A SCM_Int2fixnum(~A)~%"
                   index value ) )
          ((string? value)
           (format out "SCM_DefineString(thing~A_object,\"~A\");~%"
                   index value )
           (format out "#define thing~A SCM_Wrap(&thing~A_object)~%"
                   index index ) )
          (else (compiler-error "Unknown quotation value" value)) ) ) )

;;;==================================================================
;;; To ease writing printing expressions.

(define-icslas-macro (between-parentheses out . body)
  (let ((g (gensym)))
    `(let ((,g ,out))
       (format ,g "(")
       (begin . ,body)
       (format ,g ")") ) ) )

;;;==================================================================
;;; Convert an expression to C

;;; Generate the name used in C for a variable.

(define-generic (variable->C (variable) out))

(define-method (variable->C (variable Variable) out)
  (format out (IdScheme->IdC (Variable-name variable))) )

(define-method (variable->C (variable Renamed-Local-Variable) out)
  (format out "~A_~A" 
          (IdScheme->IdC (Variable-name variable))
          (Renamed-Local-Variable-index variable) ) )

(define-method (variable->C (variable Quotation-Variable) out)
  (format out "quotation~A" (Quotation-Variable-name variable)) )

;;; Works for Local-, Global-, Predefined- References but for
;;; Free-Local-References.

(define-method (->C (e Reference) out)
  (reference->C (Reference-variable e) out) )

(define-generic (reference->C (v Variable) out))

(define-method (reference->C (v Variable) out)
  (variable->C v out) )

;;; Exercice: change the C macro SCM_CheckedGlobal to produce a more
;;; user-friendly error message.

;(define-method (reference->C (v Global-Variable) out)
;  (format out "SCM_CheckedGlobal")
;  (between-parentheses out
;    (variable->C v out) ) )

;;; Use the results of the initialization analysis ie do not check
;;; that initialized variables are initialized.

(define-method (reference->C (v Global-Variable) out)
  (if (Global-Variable-initialized? v)
      (variable->C v out)
      (begin 
        (format out "SCM_CheckedGlobal")
        (between-parentheses out
          (variable->C v out) ) ) ) )

(define-method (->C (e Free-Local-Reference) out)
  (format out "SCM_Free")
  (between-parentheses out
    (variable->C (Free-Local-Reference-variable e) out) ) )

(define-method (->C (e Global-Assignment) out)
  (between-parentheses out
    (variable->C (Global-Assignment-variable e) out)
    (format out "=")
    (->C (Global-Assignment-form e) out) ) ) 

(define-method (->C (e Box-Read) out)
  (format out "SCM_Content")
  (between-parentheses out
    (->C (Box-Read-reference e) out) ) )

(define-method (->C (e Box-Write) out)
  (between-parentheses out
    (format out "SCM_Content")
    (between-parentheses out
      (->C (Box-Write-reference e) out) )
    (format out "=")
    (->C (Box-Write-form e) out) ) )

(define-method (->C (e Box-Creation) out)
  (between-parentheses out
    (variable->C (Box-Creation-variable e) out)
    (format out "= SCM_allocate_box")
    (between-parentheses out
      (variable->C (Box-Creation-variable e) out) )
    (format out ",~%")
    (->C (Box-Creation-body e) out) ) )

;;;  distinguer qqs  cas comme  eq?, ou pair?

(define-method (->C (e Alternative) out)
  (between-parentheses out
    (boolean->C (Alternative-condition e) out)
    (format out "~%? ")
    (->C (Alternative-consequent e) out)
    (format out "~%: ")
    (->C (Alternative-alternant e) out) ) )

(define-generic (boolean->C (e) out)
  (between-parentheses out
    (->C e out)
    (format out " != SCM_false") ) )
    
(define-method (->C (e Sequence) out)
  (between-parentheses out
    (->C (Sequence-first e) out)
    (format out ",~%")
    (->C (Sequence-last e) out) ) )

;;; (define-method (->C (e Constant) ...) does not exist any longer.

(define-method (->C (e Regular-Application) out)
  (let ((n (number-of (Regular-Application-arguments e))))
    (cond ((< n 4)
           (format out "SCM_invoke~A" n)
           (between-parentheses out
             (->C (Regular-Application-function e) out)
             (->C (Regular-Application-arguments e) out) ) )
          (else (format out "SCM_invoke")
                (between-parentheses out
                  (->C (Regular-Application-function e) out)
                  (format out ",~A" n)
                  (->C (Regular-Application-arguments e) out) ) ) ) ) )

(define-method (->C (e Arguments) out)
  (format out ",~%")
  (->C (Arguments-first e) out)
  (->C (Arguments-others e) out) )

(define-method (->C (e No-Argument) out)
  #t )

;;; This is the method for ISO C where it is not possible to have an
;;; expression with local variables. Fortunately gcc offers such a
;;; construct so we can use it.  But if we use that facility of gcc,
;;; we do not need any longer to collect temporary variables nor to
;;; emit their declaration at the beginning of functions.

(define-method (->C (e Fix-Let) out)
  (cond 
   (generate-for-gcc
    (format out "({ ")
    (local-bindings->C (Fix-Let-variables e) (Fix-Let-arguments e) out)
    (->C (Fix-Let-body e) out)
    (format  out " ; })") )
   ;; if not using gcc
   (else
    (between-parentheses out
      (bindings->C (Fix-Let-variables e) (Fix-Let-arguments e) out)
      (->C (Fix-Let-body e) out) ) ) ) )



(define-generic (local-bindings->C variables (arguments) out))

(define-method (local-bindings->C variables (e Arguments) out)
  (format out "SCM ")
  (variable->C (car variables) out)
  (format out " = ")
  (->C (Arguments-first e) out)
  (format out ";~%")
  (local-bindings->C (cdr variables) (Arguments-others e) out) )

(define-method (local-bindings->C variables (e No-Argument) out)
  (format out "") )

(define-generic (bindings->C variables (arguments) out))

(define-method (bindings->C variables (e Arguments) out)
  (variable->C (car variables) out)
  (format out "=")
  (->C (Arguments-first e) out)
  (format out ",~%")
  (bindings->C (cdr variables) (Arguments-others e) out) )

(define-method (bindings->C variables (e No-Argument) out)
  (format out "") )

(define-method (->C (e Predefined-Application) out)
  ((Functional-Description-for-C-generator
    (Predefined-Variable-description
     (Predefined-Application-variable e) ) ) e out ) )

(define (make-predefined-application-generator Cname)
  (lambda (e out)
    (format out "~A" Cname)
    (between-parentheses out
      (arguments->C (Predefined-Application-arguments e) out) ) ) )

(define-generic (arguments->C (e) out))

(define-method (arguments->C (e Arguments) out)
  (->C (Arguments-first e) out)
  (->C (Arguments-others e) out) )

(define-method (arguments->C (e No-Argument) out)
  #t )

;;; Exercice: Close statically if there are no free variables.

(define-method (->C (e Closure-Creation) out)
  (format out "SCM_close")
  (between-parentheses out
    (format out "SCM_CfunctionAddress(function_~A),~A,~A" 
            (Function-Definition-index
             (Closure-Creation-definition e) )
            (generate-arity (Function-Definition-variables
                             (Closure-Creation-definition e) ))
            (number-of (Closure-Creation-free e)) )
    (->C (Closure-Creation-free e) out) ) )

(define (generate-arity variables)
  (let count ((variables variables)(arity 0))
    (if (pair? variables)
        (if (Local-Variable-dotted? (car variables))
            (- (+ arity 1))
            (count (cdr variables) (+ 1 arity)) )
        arity ) ) )

(define-method (->C (e No-Free) out)
  #t )

(define-method (->C (e Free-Environment) out)
  (format out ",")
  (->C (Free-Environment-first e) out)
  (->C (Free-Environment-others e) out) )

;;;==================================================================
;;; Function definitions

(define (generate-functions prg out)
  (let ((definitions (Flattened-Program-definitions prg)))
    (format out "~%/* Functions: */~%")
    (for-each (lambda (def)
                (generate-closure-structure out def)
                (generate-possibly-dotted-definition out def) )
              (reverse definitions) ) ) )

(define (generate-closure-structure out definition)
  (format out "SCM_DefineClosure(function_~A, "
          (Function-Definition-index definition) )
  (generate-local-temporaries (Function-Definition-free definition) out)
  (format out ");~%") )

(define (generate-possibly-dotted-definition out definition)
  (format out "~%SCM_DeclareFunction(function_~A) {~%"
          (Function-Definition-index definition) )
  (let ((vars (Function-Definition-variables definition))
        (rank -1) )
    (for-each (lambda (v)
                (set! rank (+ rank 1))
                (cond ((Local-Variable-dotted? v)
                       (format out "SCM_DeclareLocalDottedVariable(") )
                      ((Variable? v)
                       (format out "SCM_DeclareLocalVariable(") ) )
                (variable->C v out)
                (format out ",~A);~%" rank) )
              vars )
    (if (not generate-for-gcc)
        (let ((temps (Function-Definition-temporaries definition)))
          (when (pair? temps)
            (generate-local-temporaries temps out)
            (format out "~%") ) ) )
    (format out "return ")
    (->C (Function-Definition-body definition) out)
    (format out ";~%}~%~%") ) )

(define (generate-local-temporaries temps out)
  (when (pair? temps)
    (format out "SCM ")
    (variable->C (car temps) out)
    (format out "; ")
    (generate-local-temporaries (cdr temps) out) ) )

;;;==================================================================
;;; Main procedure. We choose to print the result.

(define (generate-main prg out)
  (let ((form (Flattened-Program-body prg)))
    (format out "~%/* Expression: */~%void main(void) {~%")
    (format out "  SCM_print")
    (between-parentheses out
      (->C form out) )
    (format out ";~%  exit(0);~%}~%") ) )

;;; end of Cgen.scm
