/* $Id: dmc.c,v 1.14 1998/06/04 15:16:04 queinnec Exp $  
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 
 * DOCUMENTATION dmc.c
 * This file defines the C interface for DMeroon. It defines the small
 * set of functions that must be provided to DMeroon by the support 
 * language. A cpp variable rules the presence of a simple-minded 
 * memory allocator (no GC then) which is used when no GC is required.
 */

#include "dmeroon.h"

DM_DefineStringWithCstring(dmc_version,
  "$Id: dmc.c,v 1.14 1998/06/04 15:16:04 queinnec Exp $");

/* {{{ Predicate

 * RESOURCE DM_is_object
 * This predicate returns true if given a pointer onto a DMeroon instance.
 * This is one of the rare primitives that takes a SupportPointer.
 */

DM_c_boolean
DM_is_object (DM_SupportPointer wrapped_p) {
  if ( DM_IsDMeroonInstance(wrapped_p) ) {
    return DM_TRUE;
  } else {
    return DM_FALSE;
  }
}

/* }}}
 * {{{ Allocation
 */

#if defined(DM_USE_NO_GC)

/* RESOURCE DMsm_malloc
 * Allocate size bytes. This is a simple-minded memory allocator. It simply
 * slice big malloc'ed pages. This may be used when the regular GC is 
 * suspected to be erroneous (provided the test is short enough to run 
 * without GC).
 */

static char*         DMsm_next_byte       = NULL;
static unsigned long DMsm_remaining_bytes = 0;

void*
DMsm_malloc (size_t size) 
{
  void* result;

  DM_DebugAtLevel(21,
    fprintf(DM_debug_stream, "Allocate %lu\n", (long unsigned)size);
);

  /* Pad size for alignment compatibility. */
  size = DM_OffsetRealign(size);
  if ( size <= DMsm_remaining_bytes ) {
    /* Chop a little the current chunk. */
    DMsm_remaining_bytes -= size;
    result = DMsm_next_byte;
    DMsm_next_byte += size;
    return result;
  } else {
    /* Chunk is no longer big enough, allocate a new chunk. This chunk
     * must be big enough to satisfy the required size. Pad to obtain a
     * multiple of the page size. */
    long unsigned page_size = ((size/DMsm_PAGE_SIZE) + 1) * DMsm_PAGE_SIZE;

    DMsm_next_byte = malloc(page_size);

  DM_DebugAtLevel(21,
    fprintf(DM_debug_stream, "New chunk (size=%lu) at " DM_0X "%p\n", 
            page_size,
            DMsm_next_byte );
);

    if ( DMsm_next_byte == NULL ) {
      /* Allocation failed. */
      DM_Abort(DM_NULL, DME_MEMORY_EXHAUSTED);
      return DM_NULL;
    } else {
      /* Allocation succeeded, reset to zero the chunk, chop a little
       * the new chunk. The old chunk is wasted: too bad but malloc is
       * not compelled to return contiguous chunks. */
      memset(DMsm_next_byte, 0, page_size);
      DMsm_remaining_bytes = page_size - size;
      result = DMsm_next_byte;
      DMsm_next_byte += size;
      return result;      
    }
  }
}

#endif /* DM_USE_NO_GC */

/* RESOURCE DM_allocate_bare_instance 
 * This function allows DMeroon to allocate a DMeroon object using the
 * allocator of the support language. 
 *
 * This is an unsafe allocator since the higher levels ensures that
 * <class> is a Class and <size> is a natural number. The size of the 
 * DMeroon prefix is automatically added to the previous size.

 NOTE: On Sony, Boehm's GC_malloc return addresses that are multiple
 of 4 whereas, alignment of the most stringent type (as handled by
 gcc) is 8.  For C compatibility since the kernel classes are defined
 in C and must be readable by DMeroon, DM_allocate_bare_instance must
 provide addresses that are multiple of 8. This would be a waste of
 room if performed everywhere, so I optimize this as follows. Such a
 stringent alignment is only needed when there are fields of size 8 or
 at least two indexed fields. So we correct the address returned by
 malloc for these cases. This information is coded in the
 DM_CLASS_IS_STRINGENT bit. 
 */

DM_Object*
DM_allocate_bare_instance (DM_Context *context,
                           DM_Class   *class, 
                           DM_nat3    size ) {
  DMeroon_object *p;

  if ( DM_HasOption(class, DM_CLASS_IS_STRINGENT) ) {
    size += DM_ROOM_FOR_REALIGNMENT;
    p = DM_MALLOC(DM_PrefixSize() + size);
    if ( p ) {
      p = (void*)DM_OffsetRealign((DM_SupportAddress)p);
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  } else {
    p = DM_MALLOC(DM_PrefixSize() + size);
    if ( p ) {
      /* NOTHING */
    } else {
      DM_Abort(context, DME_ALLOCATION_PROBLEM);
      return DM_NULL;
    }
  }

#if defined(DM_NON_ZEROING_MALLOC)
  /* DM_MALLOC returns instances that should be zeroed since DMeroon
   * expects fresh instances to be filled with zeroes. */
  memset(p, 0, DM_PrefixSize() + size);
#endif

  /* This is the initialization for the C support language. */
  p->_tag = DMEROON_TAG;

  DM_DebugAtLevel(0,
    /* One word is used for padding, record the size in it for debug. */
    p->_padding = size;   
);

  return DM_FromC(p);
}

/* }}}
 * {{{ Boehm's GC related stuff
 *
 * When Exit instances are to be recycled, a message must be sent to the
 * remotely referenced Entry. This requires cooperation with the GC.
 * Fortunately, Boehm GC provides this ability.
 */

#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC

void
DM_exit_finalizer (void *exit, void *client_data ) {
  /* Remove the Exit instance from the Exits hashtable */
  DM_DebugAtLevel(0,
    fprintf(DM_debug_stream, "Boehm's GC: freeing " DM_0X "%p\n", exit);
);
  DM_remove_exit(NULL, exit);
}

#endif

void
DM_set_exit_finalization (DM_Exit *exit) {

#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC
  GC_register_finalizer(exit, DM_exit_finalizer, NULL, NULL, NULL);
#endif

}

/* }}}
 * {{{ Printing

 * RESOURCE DM_print_char_star 
 * The generic printer on which any DMeroon (and Icslas) print routines
 * are based. It knows how to write a C string into a Port or a C stream.
 * It does not flush the Port (but it does on a C stream).

 * NOTE: for long strings, would be better to have `start' and `len'
 * to allow to chop them in small pieces. There is a problem on strings
 * that are larger than the buffer ???                 FUTURE.
 */

void
DM_print_char_star (DM_Context *context,
                    char       *string,
                    DM_nat3    len,
                    void       *output_port) {
  DM_nat3 n = len ? len : strlen(string);

  if ( output_port == NULL ) return;

  DM_DebugAtLevel(0,
    /* Recognize standard outgoing C streams to print on them. */
    if (   (output_port == (void*)stdout)
        || (output_port == (void*)stderr) ) {
      FILE *stream = (FILE*) output_port;

      fprintf(stream, "%*s", (int)n, string);
      fflush(stream);
      return;
    }
);

  if ( DM_IsA(context, output_port, OutBuffer) ) {
    DM_DeclareNarrowedObject(outbuffer, OutBuffer, output_port);
    
    DM_print_char_star_in_outbuffer(context, string, n, outbuffer);
    return;
  }

  DM_Abort(context, DME_PRINT_CHAR_STAR_PROBLEM);
}

/* }}}
 * {{{ Initialization

 * RESOURCE DM_initialize
 * Initialize DMeroon before use.
 */

void
DM_c_initialize (DM_Context *context) {

  /* Just to avoid a warning. */
  DM_dmc_version_string_object =   DM_dmc_version_string_object;

#if defined(DM_USE_BOEHM_GC) && DM_USE_BOEHM_GC

  /* Advise Boehm's GC that DMeroon uses interior pointers.  Also
   * advise that we may correct misalignment (see
   * allocate_bare_instance).
   */
  GC_register_displacement(DM_PrefixSize());
  GC_register_displacement((1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);
  GC_register_displacement(DM_PrefixSize()
                           + (1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);

#endif

  /* Test a little the macros of dmc.h 
   * ONE: test changing from/to Support. */
  { DM_Object *o = (DM_Object*)(DM_ReferenceToClass(Class));

    if ( DM_FromSupport(DM_ToSupport(o)) != (o) ) {
      DM_Abort(context, DME_TO_FROM_PROBLEM);
    }
  }

  /* TWO: Test IsObject. */
  { DM_Object *o = DM_TheConstant(nil);

    if ( ! DM_IsObject(o) ) {
      DM_Abort(context, DME_IS_OBJECT_PROBLEM);
    }
  }

  /* THREE: Test if DM_allocate_bare_instance satisfies IsObject.. 
   * This is normally the first allocation to be performed. */
  { DM_Object *o = DM_Allocate0(context, Object);

    if ( ! DM_IsObject(o) )
      DM_Abort(context, DME_ALLOCATE_PROBLEM);
  }

  /* Initialize all the API */
  DM_newapi_initialize(context);
}

/* }}}
 * end of dmc.c */
