/* $Id: server.c,v 1.24 1998/05/31 14:51:24 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION server.c
 * This file defines the standard DMeroon server. 
 */

#include "server.h"

/* RESOURCE site
 * The current site.
 */

static DM_Site* site;

/* prints the identification of the server. */

static void
print_banner (DM_Context* context) {
  fprintf(context->errstream, 
          "DMeroon " DM_RELEASE_LEVEL " [" DM_RELEASE_DATE 
          "]. C Server $Revision: 1.24 $\n"
          "(c) 1995-1998 Christian.Queinnec@lip6.fr.\n"
          );
}
     
/* RESOURCE analyze_options
 * Interpret options.
 */

static void
analyze_options (DM_Context* context, 
                 int         argc,
                 char*       argv[] ) {
  DM_ErrorCode errcode;
  int          i;

  for ( i=1 ; i<argc ; i++ ) {

    if ( strcmp(argv[i], "-help") == 0 ) {
      print_banner(context);
      fprintf(context->errstream, "\n"
"-banner\n\t prints the above banner.\n"
"-verbose\n\t displays some information.\n"
"-connect hostname portnumber\n\t "
  "connects to the DMeroon server located on host:port.\n\t "
  "This option may be repeated.\n"
"-linger [duration]\n\t "
  "serves requests until a calm of at least duration occurs.\n\t "
  "A final -linger without duration means quasi-infinity.\n"
              );
      exit(0);
    }

    if ( strcmp(argv[i], "-banner") == 0 ) {
      print_banner(context);
      continue;
    }

    if ( strcmp(argv[i], "-careful1") == 0 ) {
      DM_SetOption(DM_TheCurrentSite()->context, DM_CONTEXT_CHECK_DMEROON);
      DM_SetOption(context, DM_CONTEXT_CHECK_DMEROON);
      continue;
    }

    if ( strcmp(argv[i], "-careful2") == 0 ) {
      DM_SetOption(DM_TheCurrentSite()->context, DM_CONTEXT_CHECK_CLASS);
      DM_SetOption(context, DM_CONTEXT_CHECK_CLASS);
      continue;
    }

    if ( strcmp(argv[i], "-paranoid") == 0 ) {
      DM_SetOption(DM_TheCurrentSite()->context, DM_CONTEXT_ABORT_IMMEDIATELY);
      DM_SetOption(context, DM_CONTEXT_ABORT_IMMEDIATELY);
      continue;
    }

    if ( strcmp(argv[i], "-verbose") == 0 ) {
      DM_SetOption(DM_TheCurrentSite()->context, DM_CONTEXT_VERBOSE);
      DM_SetOption(context, DM_CONTEXT_VERBOSE);
      continue;
    }

    if ( strcmp(argv[i], "-linger") == 0 ) {
      /* Ask the server to die after this duration if nothing occurs
       * meanwhile. If -linger is the last option, then the server
       * will last a very long time. */
      DM_nat3   duration = DM_INFINITE_DURATION;
      DM_Object *o;

      if ( i < argc-1 ) {
        /* -linger is not the last option. */
        duration = DM_string_to_nat3(context, argv[++i], NULL);
      }
      do {
        errcode = DMeroon_receive(context, &o, duration);
        if ( errcode == DME_TIMEOUT ) {
          break;
        } else {
          /* Ignore the received object. Remove the pointer on it to
           * let the GC recycle it. */
          o = DM_NULL;
          continue;
        }
      } while ( 1 );
      continue;
    }

    DM_DebugAtLevel(0,

      if ( strcmp(argv[i], "-inspect") == 0 ) {
        /* Ask the server to fork an inspector. If the string naming the
         * inspector is empty, then use stdin to read debug commands. If
         * the -inspect option is the last one, then consider it as
         * equivalent to "". Inspection is used for debugging. */
        char *str = (i==argc-1) ? "" : argv[++i];

        if ( strlen(str) == 0 ) {
          if ( DM_HasOption(context, DM_CONTEXT_VERBOSE) ) {
            fprintf(context->errstream, "\tEntering debug monitor...\n");
          }
          DM_debug_monitor(context, 0);
        } else {
          if ( DM_HasOption(context, DM_CONTEXT_VERBOSE) ) {
            fprintf(context->errstream, 
                    "\tForking external inspector %s...\n", str);
          }
          DM_fork_inspector(context, str);
        }
        continue;
      }

);

    if ( strcmp(argv[i], "-connect") == 0 ) {
      /* Ask Icslas to connect another site. */
      if ( i<argc-2 ) {
        char    *sitename = argv[++i];
        char    *portnum  = argv[++i];
        DM_nat2 port      = DM_string_to_nat3(context, portnum, NULL);
        DM_Site *othersite;

        if ( DM_HasOption(context, DM_CONTEXT_VERBOSE) ) {
          fprintf(context->errstream, "\tConnecting to server %s:%d...\n",
                  sitename, port);
        }
        errcode = DMeroon_connect(context, &othersite, 
                                  sitename, 
                                  port,
                                  NULL );
        if ( errcode != DME_SUCCESS ) {
          if ( DM_HasOption(context, DM_CONTEXT_VERBOSE) ) {
            fprintf(context->errstream, "\tFailed connection (error=%lu)\n", 
                    (unsigned long)errcode );
          }
          exit(3);
        } else {
          continue;
        }
      }
    }

    fprintf(context->errstream, "Unrecognized option %s\n", argv[i]);
    exit(1);
  }

  if ( argc == 1 ) {
    /* If no option is present, just output the banner. */
    print_banner(context);
  } else {
    /* No more options, exit DMeroon server. */
    if ( DM_HasOption(context, DM_CONTEXT_VERBOSE) ) {
      fprintf(context->errstream, "Ite DMeroon est.\n");
    }
  }
}

/* RESOURCE main
 * Initialize DMeroon, analyze options and enter server.
 */

int
main (int argc, char *argv[]) 
{
  DM_Context   *context = DM_NULL;
  DM_ErrorCode errcode;

  /* Initialize DMeroon but first, tells the context to report errors
   * on stderr. */
  DM_TheCurrentSite()->context->errstream = stderr;
  errcode = DMeroon_initialize(context, &site);
  if ( errcode != DME_SUCCESS ) {
    fprintf(stderr, "\tCannot initialize DMeroon (error=%lu).\n",
            (unsigned long)errcode );
    exit(5);
  }

  /* Since now DMeroon is working, analyze the options under the
   * protection of a fresh Context allocated by
   * DM_ProtectWithContext. */
  DM_ProtectWithContext(context,
          analyze_options(context, argc, argv),
     /* If something went wrong */
     exit(7);
                           );
  exit(0);
}

/* end of server.c */
