/* $Id: dm.h,v 1.102 1998/07/01 10:04:11 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

#ifndef DM_H
#define DM_H

/* 
 * This file contains internal definitions necessary for the
 * compilation of DMeroon. These things are not primarily destined to
 * be used out of DMeroon.  
 *
 * This file depends on some macros provided by the support language.
 */

/* 
 * C booleans are represented by these values. 
 * The type is DM_c_boolean.
 */

#define DM_FALSE 0
#define DM_TRUE  1

/******************************************************************
 *                    BASIC STRUCTURES
 * 
 * These are the types of predefined DMeroon objects.
 *
 * All instances of DMeroon contains at least two fields: one is the
 * instantiation link (and holds a reference to the class instance),
 * the other is the proxy and records migration information. 
 */

#define DM_DMeroonPrefix() \
  struct DM_Class*  _class; \
  struct DM_Object* _proxy

/*
 * This is the size of the DMeroon prefix. It is often two or four words.
 * Recall that this prefix is the concatenation of the header required
 * by the language runtime (malloc- and type- header) followed by
 * DMeroon prefix and some possible padding (that is why this is not
 * equal to sizeof(DMeroon_object)!).
 * This constant may be accessed through the Site object.
 */

#define DM_PrefixSize() \
  DM_OffsetOf(DM_Object_wrapped, _ignored._realigned._content._void)

/* 
 * This is the general type of a DMeroon instance viewed from C.  All
 * DMeroon instances (again viewed from C) will extend (inherit) this
 * structure. Remind that a DMeroon reference is the address of the
 * first field that will follow this structure.
 */

typedef struct DMeroon_object {
  DM_SupportPrefix()
  DM_DMeroonPrefix();
} DMeroon_object;

/******************************************************************
 *                              REFERENCES
 *
 * DM_ReferenceTo returns the address of the object that represents
 * the DMeroon entity named <name>. The entity itself has a name
 * ending with _object. Remember that a reference to a DMeroon
 * instance designates the first word of its content.
 */

#define DM_ReferenceTo(name) \
  (DM_FromC( &( name##_object )))

/* 
 * This macro allows to reference a string defined with DM_DefineString.
 * Strings (for Fields, Classes etc) are defined with a name suffixed
 * with _string. They are declared static.
 */

#define DM_ReferenceToString(name) \
  ((DM_String*)DM_ReferenceTo(DM_##name##_string))

/*
 * This one allows to easily reference a class.
 */

#define DM_ReferenceToClass(name) \
  ((DM_Class*)DM_ReferenceTo(DM_##name##_class))

/*
 * or to a given field within a class.
 */

#define DM_ReferenceToFieldOfClass(fieldname, classname) \
  ((DM_Field*)DM_ReferenceTo(DM_##classname##_##fieldname##_field))

/*
 * This one allows to easily reference a DMeroon or C type.
 * Only used in the dmtypes.c file.
 */

#define DM_ReferenceToDMeroonType(name) \
  ((DM_DMeroonType*)DM_ReferenceTo(DM_##name##_type))

#define DM_ReferenceToCType(name) \
  ((DM_CType*)DM_ReferenceTo(DM_C_##name##_type))

/* 
 * or to a RequestMetaClass
 */

#define DM_ReferenceToRequestMetaClass(name) \
  ((DM_RequestMetaClass*)DM_ReferenceTo(DM_##name##_class))

/* 
 * These macros allow references to implementation constants.
 */
 
#define DM_TheCurrentSite() \
  ((DM_Site*)DM_ReferenceTo(DM_site))

#define DM_TheImplementedTypes() \
  ((DM_ImplementedTypesVector*)DM_ReferenceTo(DM_implemented_types))

#define DM_ThePredefinedClasses() \
  ((DM_PredefinedClassesVector*)DM_ReferenceTo(DM_predefined_classes))

#define DM_TheConstant(name) \
  ((DM_Object*)DM_ReferenceTo(DM_##name))

/* 
 * These two macros take a DMeroon referenced object and return the 
 * components of the DMeroon specific header ie a DMeroon reference onto
 * its class and proxy. 
 */

#define DM_ObjectToClass(context, dmeroon) \
  ((DM_Class*)((DM_ToC(dmeroon))->_class))

#define DM_ObjectToProxy(context, dmeroon) \
  ((DM_Object*)((DM_ToC(dmeroon))->_proxy))

/* 
 * These two macros allow to modify the class and proxy of a DMeroon
 * value.
 */

#define DM_SetClassOf(context, dmeroon, value) \
  ( ((DMeroon_object*)(DM_ToC(dmeroon)))->_class = (value) )

#define DM_SetProxyOf(context, dmeroon, value) \
  ( ((DMeroon_object*)(DM_ToC(dmeroon)))->_proxy = (DM_Object*)(value) )

/* 
 * Declare that a freshly allocated object is now initialized.
 * The proxy is overwritten without verification so this macro is
 * to be used carefully. 
 */

#define DM_SetSharable(context, dmeroon) \
  (DM_SetProxyOf(context, (dmeroon), DM_no_proxy_value))

#define DM_IsSharable(context, dmeroon) \
  (    ( DM_ObjectToProxy(context, dmeroon) == DM_no_proxy_value ) \
    || (     DM_IsA(context, DM_ObjectToProxy(context, dmeroon), Entry) \
          && DM_HasNotOption((DM_Entry*)DM_ObjectToProxy(context, dmeroon), \
                             DM_ENTRY_IS_COPYABLE ) ) )

/* 
 * Some DMeroon objects may be copied rather than shared.  As for
 * DM_SetSharable, the proxy is overwritten without verification so
 * this macro is to be used carefully.
 */

#define DM_SetCopyable(context, dmeroon) \
  DM_SetProxyOf(context, dmeroon, DM_TheConstant(true))

#define DM_IsCopyable(context, dmeroon) \
  (    ( DM_ObjectToProxy(context, dmeroon) == DM_TheConstant(true) ) \
    || (     DM_IsA(context, DM_ObjectToProxy(context, dmeroon), Entry) \
          && DM_HasOption((DM_Entry*)DM_ObjectToProxy(context, dmeroon), \
                          DM_ENTRY_IS_COPYABLE ) ) )

/*
 * DM_CodeOf returns the address of the C function named <name>.
 * Pay attention that in C, pointers to functions are not pointers 
 * on data. The DMeroon type named DM_code takes care of these latter
 * while DM_data takes care of the former.
 */

#define DM_CodeOf(name) \
  ((DM_code)(name))

/*
 * Some classes are predefined. 
 */

#define DM_IsPredefinedClass(class) \
  ( class == \
   (DM_Class*)((DM_ImmutableVector*) DM_ReferenceTo(DM_predefined_classes)) \
              ->item[DM_PredefinedClassNumber(class)] )

/******************************************************************
 * Keys are big structures (16 bytes now).
 */

#define DM_EqualKey(k1, k2) \
   (    ( (k1)._idx  == (k2)._idx )     \
     && ( (k1)._ip   == (k2)._ip )	\
     && ( (k1)._port == (k2)._port )	\
     && ( (k1)._date == (k2)._date )	\
     && ( (k1)._rand == (k2)._rand ) )
     

/******************************************************************
 *                          Variables
 *
 * When an object is known to be of a given class, we may want to 
 * narrow it into its right class. The difference with the previous
 * macro is that there is no runtime check here.
 */

#define DM_DeclareNarrowedObject(name, classname, other) \
  DM_##classname * name = ((DM_##classname*)(other))

/*
 * Test if an instance belongs to a specific class.
 * This is easier to write (and read) than the direct call that uses
 * the DMeroon instance of Class. Caution: it returns a C boolean.
 * Pay attention that o must be a DMeroon reference (this is marked
 * by the DM_ prefix of the macro.
 */

#ifdef __GNUC__

#define DM_IsA(context, o, classname) \
  ( (o) && \
    ({ DM_Class *_oc    = DM_ObjectToClass(context, o); \
       DM_Class *_class = DM_ReferenceToClass(classname); \
       (   ( _oc == _class )  \
        || ({ DM_nat3 _depth = _class->super_length; \
              (   ( _depth <= _oc->super_length ) \
               && ( _class == _oc->super[_depth-1] ) ); }) ); \
     }) )

#else

#define DM_IsA(context, o, classname) \
  ( (o) && DM_is_a(context, o, DM_ReferenceToClass(classname)) )

#endif /* __GNUC__ */

/*
 * Test if an object is a direct instance of a class.
 */

#define DM_IsADirect(context, o, classname) \
  ( DM_ObjectToClass(context, o) == DM_ReferenceToClass(classname) )

/******************************************************************
 *                          CLASSES
 *
 * Define some specialized classes for classes having a given number of
 * fields and ancestors. They are needed in C to statically allocate the
 * class objects.
 *
 * The first macro defines regular classes having some fields (but at 
 * least one).
 */

#define DM_DefineSpecializedClass(supers_number,fields_number,namelen) \
typedef struct DM_Class_##supers_number##_##fields_number##_##namelen { \
  DM_nat3      options; \
  DM_nat3      super_length; \
  DM_reference super[supers_number]; \
  DM_RealignToMostStringent( \
      DM_nat3      field_length; \
      DM_reference field[fields_number]; \
      DM_RealignToMostStringent( \
          DM_nat1      name_length; \
          DM_char1     name[namelen]; \
       ); \
    ); \
} DM_Class_##supers_number##_##fields_number##_##namelen; \
DM_DefineWrappedType(DM_Class_##supers_number##_##fields_number##_##namelen)

/* 
 * This is the class needed by Object or other classes without fields.
 * It has to be specially defined since it is not possible to 
 * use the preceding macro with a field-length of zero.
 */

#define DM_DefineVerySpecializedClass(supers_number,fields_number,namelen) \
typedef struct DM_Class_##supers_number##_##fields_number##_##namelen { \
  DM_nat3      options; \
  DM_nat3      super_length; \
  DM_reference super[supers_number]; \
  DM_RealignToMostStringent( \
      DM_nat3      field_length; \
      /* DM_reference field[fields_number]; */ \
      DM_RealignToMostStringent( \
          DM_nat1      name_length; \
          DM_char1     name[namelen]; \
        ); \
     ); \
} DM_Class_##supers_number##_##fields_number##_##namelen; \
DM_DefineWrappedType(DM_Class_##supers_number##_##fields_number##_##namelen)

/* Request subclasses have a special metaclass and a special macro to
 * be allocated. */

#define DM_DefineSpecializedRequestMetaClass(supers_number,fields_number,namelen) \
typedef struct DM_RequestMetaClass_##supers_number##_##fields_number##_##namelen { \
  DM_nat3      options; \
  DM_nat3      super_length; \
  DM_reference super[supers_number]; \
  DM_RealignToMostStringent( \
      DM_nat3      field_length; \
      DM_reference field[fields_number]; \
      DM_RealignToMostStringent( \
          DM_nat1      name_length; \
          DM_char1     name[namelen]; \
          DM_RealignToMostStringent( \
              DM_handler_code  handler; \
          ); \
       ); \
    ); \
} DM_RequestMetaClass_##supers_number##_##fields_number##_##namelen; \
DM_DefineWrappedType(DM_RequestMetaClass_##supers_number##_##fields_number##_##namelen)

/******************************************************************
 *                          WRAPPED TYPES
 *
 * Remember that a DMeroon instance has, first, a header imposed by the
 * support language, then some additional fields required by DMeroon.
 * For any type, this macro defines the associated wrapped type needed
 * for DMeroon.

 * NOTE: One has to carefully distinguish objects seen from DMeroon from
 * the same objects seen from the support language. These latter objects
 * are called wrapped objects and the variables that hold such wrapped
 * objects have a name starting with the wrapped_ prefix. Nearly every
 * type defined in DMeroon for DMeroon objects has its corresponding
 * wrapped type associated to the corresponding wrapped objects.
 */

#define DM_DefineWrappedType(type) \
typedef struct type##_wrapped { \
  DM_SupportPrefix() \
  DM_DMeroonPrefix(); \
  DM_RealignToMostStringent( \
    type _content; \
    ); \
} type##_wrapped

/* Allocate statically a DMeroon object of a given type. 
 * Pay attention to initialize it with zeroes.  
 * This macro cannot handle objects with indexed fields. */

#define DM_AllocateStatically(name, typename) \
  DM_##typename##_wrapped wrapped_##name##_object = { \
    DM_SupportPrefixInitialization(DM_##typename##_wrapped,\
                                   wrapped_##name##_object) \
    DM_ReferenceToClass(typename), \
    DM_ProxyInitialization() \
  }; \
  DM_##typename *name = ( \
     (DM_HasOption(DM_ReferenceToClass(typename), DM_CLASS_IS_INDEXED) \
      ? DM_Abort(DM_NULL, DME_STATIC_ALLOCATION_PROBLEM) \
      : DM_NULL ), \
     DM_zero(DM_NULL, DM_FromC(&wrapped_##name##_object)), \
     (DM_##typename*)(DM_FromC(&wrapped_##name##_object)))

/******************************************************************
 *                         ALIGNMENTS
 *
 * Alignment constraints are computed on a small structure . (I take
 * the idea from Joel Bartlett's Scheme->C 15mar93/cdecl/sizeof.c
 * program).
 *
 * This macro defines a small structure which is only used by the
 * DM_AlignConstraintFor macro.  */

#define DM_DefineConstrainedStructure(name,type) \
struct DM_alignment_##name { \
  char before_##name; \
  type name; \
  type repeated_##name; \
}

/* 
 * This macro assumes ( sizeof(DM_SupportAddress) >= sizeof(void*) )
 * since we cast address towards the longest possible integer.  This
 * is probably always the case. This assertion is checked in
 * DM_kernel_initialize().
 * 
 * DM_AlignConstraintFor returns a small integer. 2 means that addresses
 * should be even, 4 means that addresses should be multiple of 4 etc.  */

#define DM_AlignConstraintFor(type) \
  DM_OffsetOf(struct DM_alignment_##type,type)

/* 
 * This macro returns the size of a type when repeated in an array.
 * For alignment reason there might be some padding between occurrences
 * of the type. Usually it is the same as the size of the type but this
 * macro allows to introduce a DMeroon type of size 3 for instance and
 * to compute its repeated size ie 4 if the type is 2bytes aligned.
 */

#define DM_RepeatedSizeFor(type) \
  ( DM_OffsetOf(struct DM_alignment_##type,repeated_##type) \
   - DM_OffsetOf(struct DM_alignment_##type,type) )

/*
 * After an indexed field, it is necessary to realign the next stored data
 * with the most stringent alignment (usually 8 bytes for DM_float2).
 * This macro takes care of that.
 */

#define DM_RealignToMostStringent(members) \
  union { /* most stringent alignment */ \
    struct { \
      members \
    } _realigned; \
    DM_MOST_STRINGENT_TYPE _align; \
  } _ignored

/* 
 * These macros compute the offsets for various fields in various
 * types of instances. 
 * 
 * DM_OffsetOf computes the offset of field within class.
 * This offset is measured from the beginning of the content of the DMeroon
 * object.
 */

#ifdef _WINDOWS
#define DM_OffsetOf(aclass,afield) \
  ((DM_SupportAddress) &((aclass*) 0)->afield))
#else
#define DM_OffsetOf(class,field) \
  (  ((DM_SupportAddress) &((*((class*) 0)).field)) \
   - ((DM_SupportAddress)&((*((class*) 0)))) )
#endif /* _WINDOWS */

/* 
 * Realign an address to the most stringent alignment.  Takes a
 * DM_SupportAddress and returns a DM_SupportAddress. Pay attention
 * that address is more than once evaluated so compute it before
 * giving it to DM_OffsetRealign.
 */

#define DM_OffsetRealign(address) \
  DM_OffsetRealignToSomeAlignment(address,DM_MOST_STRINGENT_ALIGNMENT_LOG)

#define DM_OffsetRealignToSomeAlignment(address,alignment_log) \
  ( ( (address) & ((1<<alignment_log)-1) ) \
   ? ((((address)>>alignment_log)+1)<<alignment_log) \
   : (address) )

/*
 * increments a displacement with a given type size.
 */

#define DM_OffsetExtend(t, type) \
  ( (t) + (  ((DM_Type*)(type))->size ) )

/*
 * Pad t until it fits (ie is correctly aligned) for type.
 * Caution: t and type may be multiply evaluated.
 */

#define DM_OffsetPad(t, type) \
  ( (((DM_nat3)(t))& \
     ((((DM_Type*)(type))->alignment)-1)) \
   ? ( (((DM_nat3)(t))| \
        ((((DM_Type*)(type))->alignment)-1)) + 1 ) \
   : (t) )

/*
 * This macro packs the three components of an offset into a single item.
 * Actually this item is a DM_nat3. This is a sort of byte-code encoding
 * telling how to access fields.
 */

#define DM_PackOffset(translation,repetition,item) \
  ( (translation<<16) | (repetition<<8) | item )

/*
 * Access to the various components of the offset.
 */

#define DM_OffsetToTranslation(offset) ((offset)>>16)
#define DM_OffsetToRepetitionTypeIndex(offset) ((offset>>8)&0xff)
#define DM_OffsetToItemTypeIndex(offset) ((offset)&0xff)

/******************************************************************
 *                       STRINGS
 *
 * The DM_DefineString macro allocates a DMeroon string object containing
 * the <name> C string. The C string can be easily retrieved from the 
 * DMeroon object. As all indexed fields, the character field is prefixed
 * by its length. The length is the total number of characters including the
 * final null character.
 * 
 * The DM_DefineStringWithLength allocates a DMeroon string with a precise
 * length but no content.
 */

#define DM_DefineFullyParameterizedString(name,size,content) \
typedef struct struct_##name##_string { \
  DM_nat3      character_length; \
  DM_char1     character[size+1]; \
} struct_##name##_string; \
DM_DefineWrappedType(struct_##name##_string); \
static struct_##name##_string_wrapped \
 DM_##name##_string_object = { \
    DM_SupportPrefixInitialization(struct_##name##_string_wrapped,\
                                   DM_##name##_string_object) \
    DM_ReferenceToClass(String), \
    DM_ProxyInitialization(), \
    { { { \
      size, \
      content \
    } } } \
  }

#define DM_DefineStringWithCstring(name,string) \
  DM_DefineFullyParameterizedString(name,sizeof(string),string)

#define DM_DefineString(name) \
  DM_DefineFullyParameterizedString(name,sizeof(#name),#name)

#define DM_DefineStringWithLength(name,size) \
  DM_DefineFullyParameterizedString(name,size,"unknown")

/******************************************************************
 *                      VERSIONSTRING
 * Define a VersionString to identify a file. 
 * Then link such a VersionString to the others accumulated in site.
 */

#define DM_DefineVersionString(name,string) \
typedef struct DM_VersionString_##name { \
  struct DM_VersionString* next; \
  DM_nat1 name_length; \
  DM_char1 name[5+sizeof(string)]; \
} struct_##name##_version_string; \
DM_DefineWrappedType(struct_##name##_version_string); \
static struct_##name##_version_string_wrapped \
 DM_##name##_version_string_object = { \
    DM_SupportPrefixInitialization(struct_##name##_version_string_wrapped,\
                                   DM_##name##_version_string_object) \
    DM_ReferenceToClass(VersionString), \
    DM_ProxyInitialization(), \
    { { { \
      DM_NULL, /* To be filled with DM_LinkVersionStringIntoSite */ \
      5+sizeof(string), \
      "@(#)" string \
    } } } \
  }

#define DM_ReferenceToVersionString(name) \
  ((DM_VersionString*)DM_ReferenceTo(DM_##name##_version_string))

#define DM_LinkVersionStringIntoSite(name) \
{ DM_Site *site = DM_TheCurrentSite(); \
  DM_VersionString *vs = DM_ReferenceToVersionString(name); \
  vs->next = site->versions; \
  site->versions = vs; \
}

/******************************************************************
 *                      FIELDS
 *
 * The DM_GenerateFieldStruct is automatically generated in dmstruct.h
 * but must be used in that same file before its defining place. I
 * therefore copy it here so DM_DefineSpecializedField may use it. I
 * also change its name to avoid a redefinition warning.
 */

#define DM_GenerateFieldStruct_(_0,_1) \
struct DM_Field_##_0##_##_1 { \
  DM_field_options options; \
  struct DM_Class* introducing; \
  struct DM_Class* referring; \
  DM_nat3 offset_length; \
  DM_offset offset[_0]; \
  DM_RealignToMostStringent( \
      DM_nat1 name_length; \
      DM_char1 name[_1]; \
  ); \
}

#define DM_DefineSpecializedField(offsets_number,namelength) \
typedef DM_GenerateFieldStruct_(offsets_number,namelength) \
  DM_Field_##offsets_number##_##namelength; \
DM_DefineWrappedType(DM_Field_##offsets_number##_##namelength)

/******************************************************************
 *                           OPTIONS
 * Classes, fields and other things have an options field that represent
 * additional boolean values. By default, default options are encoded 
 * as zero. All options field are represented by DM_nat3.
 */

#define DM_HasOption(o, constant) \
   ( (o)->options & (constant) )

#define DM_HasNotOption(o, constant) \
   ( ! DM_HasOption(o, constant) )

#define DM_SetOption(o, constant) \
   ( (o)->options |= constant )

#define DM_ClearOption(o, constant) \
   ( (o)->options &= ~((DM_nat3)(constant)) )

/* Classes that are predefined are recognized with a non-null
 * index coded in the DM_CLASS_PREDEFINED bit field (256 are possible).
 */

#define DM_CLASS_IS_MUTABLE             1<<0
#define DM_CLASS_IS_IMMOTILE            1<<1
#define DM_CLASS_IS_STRINGENT		1<<2
#define DM_CLASS_IS_VOLATILE		1<<3
#define DM_CLASS_HAS_NO_REFERENCE	1<<4
#define DM_CLASS_IS_USABLE		1<<5
#define DM_CLASS_IS_NOT_SUBCLASSABLE	1<<6
#define DM_CLASS_IS_COPYABLE		1<<7
#define DM_CLASS_IS_NOT_INSTANTIABLE	1<<8
#define DM_CLASS_IS_INDEXED		1<<9

#define DM_CLASS_PREDEFINED		(0xff<<24)

#define DM_CLASS_DEFAULT_OPTIONS		0

#define DM_PREDEFINED_CLASS_DEFAULT_OPTIONS	\
     DM_CLASS_IS_USABLE|DM_CLASS_DEFAULT_OPTIONS

#define DM_FIELD_IS_MUTABLE		1<<0	/* 0 or 1 */
#define DM_FIELD_IS_VOLATILE            1<<1	/* 0 or 2 */
#define DM_FIELD_IS_LOCAL               1<<2	/* 0 or 4 */
#define DM_FIELD_IS_PRE_SENDABLE	1<<3	/* 0 or 8 */
#define DM_FIELD_IS_SECRET		1<<4	/* 0 or 16 */
#define DM_FIELD_IS_INDEXED		1<<5    

#define DM_FIELD_DEFAULT_OPTIONS	0

#define DM_PredefinedClassNumber(class) \
  ((DM_nat1)( (class->options)>>24 ))

#define DM_ENTRY_IS_COPYABLE 		(1<<1)
#define DM_ENTRY_IS_UNINITIALIZED	(1<<3)
#define DM_EXIT_DECREMENT_SENT		(1<<4)
#define DM_EXIT_IS_PREFETCHED		(1<<5)
#define DM_EXTERNAL_IS_EXTERNALIZED	(1<<6)
#define DM_EXIT_IS_CHAINED		(1<<7)
#define DM_WAITER_DONE			(1<<0)

/******************************************************************
 *                     LIMITS OF TYPES
 *
 * Constants used to set the range of DMeroon types.
 * Due to overflow, maximum/minimum are the last number that can be 
 * represented so tests must be written as:
 *     ( DM_*_MINIMUM <= value ) && ( value <= DM_*_MAXIMUM )
 *
 * NOTE: To be used at cpp-time, these constants must not be cast therefore
 * they are considered as unsigned long ints and DM_NAT4_MAXIMUM is not
 * representable. Therefore I chose to have a run-time expression for 
 * DM_NAT4_MAXIMUM but it must not be used at cpp-time. Same for DM_INT4.
 */

#define DM_NAT1_MAXIMUM	0xff
#define DM_NAT2_MAXIMUM	0xffff
#define DM_NAT3_MAXIMUM	0xffffffff
#define DM_NAT4_MAXIMUM	\
   ((((DM_nat4)DM_NAT3_MAXIMUM)+1)*DM_NAT3_MAXIMUM + DM_NAT3_MAXIMUM)

#define DM_NETNAT2_MAXIMUM DM_NAT2_MAXIMUM
#define DM_NETNAT3_MAXIMUM DM_NAT3_MAXIMUM

#define DM_OFFSET_MAXIMUM DM_NAT3_MAXIMUM

#define DM_INT1_MAXIMUM	0x7f
#define DM_INT2_MAXIMUM	0x7fff
#define DM_INT3_MAXIMUM	0x7fffffff
#define DM_INT4_MAXIMUM	(DM_NAT4_MAXIMUM/2)

#define DM_INT1_MINIMUM	(- DM_INT1_MAXIMUM)
#define DM_INT2_MINIMUM	(- DM_INT2_MAXIMUM)
#define DM_INT3_MINIMUM	(- DM_INT3_MAXIMUM)
#define DM_INT4_MINIMUM	(- DM_INT4_MAXIMUM)

/******************************************************************
 *                      OTHER MACROS
 *
 * This constant means that the object is local and has not been
 * migrated.  DM_ProxyInitialization() is used when statically
 * initializing a _proxy field while DM_no_proxy_value is used in
 * executable C code. 
 */

#define DM_no_proxy_value (DM_TheConstant(nil))

#define DM_ProxyInitialization() \
   DM_no_proxy_value

/*
 * DM_Displace returns the address of the Tth byte in object O.
 * DM_Displace can be used as an L-value or R-value.
 * 
 * DM_Content reads the content of address P as TYPE.
 */

#define DM_Displace(o,t) \
  ((void*) (((char*) o)+t))

#define DM_Content(type,p) \
   (*((type*) p))


/******************************************************************
 *                       ERRORS
 *
 * Errors detected by DMeroon are reported with DM_Abort.  DM_MayAbort
 * reports warnings, the context argument may decide that warnings
 * should be considered as fatal errors. For errors and warnings, the
 * DM_abort function (to factorize code) will longjmp to a safe
 * place. These safe places may be specified with
 * DM_ProtectWithContext or with DM_ProtectWithFreshContext (similar
 * to Lisp unwind-protect).
 *
 * NOTE: Pay attention on sparc, setjmp seems to forget the value of 
 * local variables. Qualifying them as volatile may be not sufficient.

 * Gcc may know that DM_abort does not return.    FUTURE

 * RESOURCE DM_Abort
 * Raise an exception.
 */

#define DM_Abort(context,code) \
  { DM_Context* _ctx = (context); \
    void (*_handler)(DM_Context*, DM_ErrorCode, char*, unsigned long) = \
       (_ctx != NULL) \
       ? (void (*)(DM_Context*, DM_ErrorCode, char*, unsigned long)) \
                (_ctx->handler) \
       : DM_abort; \
    _handler(_ctx, code, __FILE__, __LINE__); \
    /* Dead code, should never be executed. */ \
    DM_BarfIfReached(_ctx); \
  }

/* RESOURCE DM_MayAbort
 * This one is used for warnings. An option may
 * turn warnings into fatal errors. */

#define DM_MayAbort(context, code) \
  if ( (context) && DM_HasOption(context, DM_CONTEXT_ABORT_IMMEDIATELY) ) { \
    DM_Abort(context, code); \
  } else { \
    /* Nothing */ \
  }

/* RESOURCE DM_BarfIfReached
 * Exit immediately if run. This is to detect, at debug-time, dead code 
 * which should never be executed.
 */

#if defined(DM_DEBUG)
#define DM_BarfIfReached(context) \
  DM_barf_if_reached(context)
#else
#define DM_BarfIfReached(context) \
   /* Dead code: should never be executed. */
#endif

/* RESOURCE DM_AllocateContext
 * A context holds a number of informations and mainly the continuation
 * to escape to in case of problems. Contexts are created by cloning the
 * default context held in the current site.
 */

#define DM_AllocateContext(cntxt) \
  (DM_Context*)(DM_clone((cntxt), (DM_Object*)(DM_TheCurrentSite()->context)))

/* RESOURCE DM_DefineStaticContext
 * Stack-allocate a Context and initialize from the default context. This
 * is somewhat similar to the previous clone operation (except for the 
 * lifetime of the created Context.
 */

#define DM_DefineStaticContext(name) \
  DM_Context_wrapped _wrapped_context_object = { \
    DM_SupportPrefixInitialization(DM_Context_wrapped, \
                                   _wrapped_context_object ) \
    DM_ReferenceToClass(Context), \
    DM_ProxyInitialization() \
  }; \
  DM_Context* name     = (DM_Context*)DM_FromC(&_wrapped_context_object); \
  DM_Context* _default = DM_TheCurrentSite()->context; \
  void*       _junk_   = (memmove(name, _default, sizeof(DM_Context)), \
                          _junk_ = &_junk_ /* To avoid a warning */ )

/* There's a problem with local variables and setjmp. It seems that
 * only volatile (sic) variables are preserved by setjmp but I could
 * not succeed using that feature on sparc. For them I added the
 * trick to flush windows registers but it still does not work.
 */

#if defined(sparc)
/* trick taken out from Aubrey Jaffer's scm. There, it was 
   prefixed with the following comment:
   James Clark came up with this neat one instruction fix for
   continuations on the SPARC.  It flushes the register windows so
   that all the state of the process is contained in the stack. */

#define FLUSH_REGISTER_WINDOWS asm("ta 3")

#else  /* !sparc */

#define FLUSH_REGISTER_WINDOWS /* Nothing */

#endif /* sparc */

/* RESOURCE DM_ProtectWithFreshContext
 * Protect a computation (body) with a context (newcontext). If
 * something wrong occurs then longjmp here and run the errorcleanup
 * code: the errorcleanup must always be run and the exception must
 * always be confined to this block.  The fresh context is cloned from
 * the default one (recorded in the current site). For debug reason,
 * the old context (context) may be obtained from it (in its
 * 'previous' slot).
 */

#define DM_ProtectWithFreshContext(context, newcontext, body, errorcleanup) \
  { jmp_buf     _exit; \
    DM_DefineStaticContext(newcontext); \
    \
    newcontext->previous     = (context); \
    newcontext->continuation = &(_exit); \
    newcontext->errcode      = DME_FAILED_SETJMP; \
    /* Grab current continuation. */ \
    FLUSH_REGISTER_WINDOWS; \
                              /* Should not be setjmp viz call/cc? FUTURE */ \
    if ( setjmp(_exit) ) { \
      /* Something went wrong: someone called the continuation.  \
       * Erase the continuation and perform the cleanup. */ \
      newcontext->continuation = DM_NULL; \
      { errorcleanup; } \
    } else { \
      /* Reinitialize the error code. */ \
      newcontext->errcode = DME_SUCCESS; \
      /* Perform the protected expression. */ \
      { body; } \
      /* No error intervene, reset the continuation and continue. */ \
      newcontext->continuation = DM_NULL; \
    } \
  }

/* RESOURCE DM_ProtectWithContext
 * This macro is somewhat similar to DM_ProtectWithFreshContext except
 * that it protects the evaluation of some code (body) with the
 * current context. However if the current context is null, a local
 * one is allocated (in the heap since its lifetime may exceed the
 * block). If an error is raised during the evaluation of the body,
 * then the errorcleanup code is run and the former Context is
 * restaured. This semantics is reminiscent of unwind-protect or
 * try-finally. 
 */

#define DM_ProtectWithContext(context, body, errorcleanup) \
  { jmp_buf _exit; \
    void*   _old_continuation = NULL; \
    \
    /* context maybe NULL. */ \
    if ( context == DM_NULL ) { \
      context = DM_AllocateContext(context); \
    } else { \
      _old_continuation = context->continuation; \
    } \
    context->continuation = &(_exit); \
    context->errcode      = DME_FAILED_SETJMP; \
    /* Grab current continuation. */ \
    FLUSH_REGISTER_WINDOWS; \
    if ( setjmp(_exit) ) { \
      /* Something went wrong: someone called the continuation.  \
       * Restaure the continuation and perform the cleanup. */ \
      context->continuation = _old_continuation; \
      { errorcleanup; } \
    } else { \
      /* Reinitialize the error code. */ \
      context->errcode = DME_SUCCESS; \
      /* Perform the protected expression. */ \
      { body; } \
      /* No error intervene, restaure the continuation and continue. */ \
      context->continuation = _old_continuation; \
    } \
  }

/******************************************************************
 *                      External objects
 */

#define DM_EXTERNAL_DEFAULT_OPTIONS	0

/******************************************************************
 *                      Requests
 */

#define DM_REQUEST_DEFAULT_OPTIONS	0
#define DM_REQUEST_WANTS_ANSWER		(1<<2)

/******************************************************************
 *                      ALLOCATION
 *
 * These macros allocate an instance of a class (just named with its
 * short name) then return a DMeroon reference.
 */

#define DM_Allocate0(context,class) \
  ((void*)DM_allocate0(context, DM_ReferenceToClass(class)))

#define DM_Allocate1(context,class,i0) \
  ((void*)DM_allocate1(context, DM_ReferenceToClass(class),i0))

#define DM_Allocate2(context,class,i0,i1) \
  ((void*)DM_allocate2(context, DM_ReferenceToClass(class),i0,i1))

#define DM_Allocate3(context,class,i0,i1,i2) \
  ((void*)DM_allocate3(context, DM_ReferenceToClass(class),i0,i1,i2))


/******************************************************************
 *                      PREDEFINED OBJECTS
 *
 * This is the declaration of all predefined objects (but strings).
 *
 * There are some precise naming conventions:
 * -- They are all prefixed by DM_ to avoid name clashes with other libraries.
 * -- All objects are held by a C variable whose name ends with _object.
 * -- The DMeroon type of these objects is mentoned in the penultimate 
 *    component of the name. Here it can be _type_, _class_ or _field_.
 *    For strings, it is _string_.
 * -- Type names are in lower case ie nat1, float2, nought etc.
 *    Therefore their structure is DM_<name>_type_object.
 * -- Class names are capitalized ie Class, Type, String etc.
 *    Therefore their structure is DM_<Name>_class_object.
 * -- Fields mentions the class name and the field name. 
 *    Therefore their structure is DM_<Class>_<field>_field_object.
 */

/* 
 * This macro converts a C statically allocated object into a DMeroon
 * reference.  C allocated objects have a name ending with _object.
 * Caution: This is unrelated to DM_{From/To}Support macros which
 * represents the conversion between DMeroon and some language, here
 * the DM_{From/To}C macros deals with the conversion between static C and
 * DMeroon. The general type for a DMeroon value is DM_Object*.
 */

#define DM_FromC(c_object) \
  ((DM_Object*)(((DMeroon_object*)(c_object))+1))

#define DM_ToC(dmeroon) \
  (((DMeroon_object*)(dmeroon))-1)

DM_DefineWrappedType(DM_Object);

/* 
 * These macros are used to define and to export the structure of
 * constants such as true, false, nil, etc.
 */

#define DM_ExportConstantObject(nm) \
typedef struct DM_##nm##_struct { \
  DM_nat1      name_length; \
  DM_char1     name[sizeof(#nm)+1]; \
} DM_##nm##_struct; \
DM_DefineWrappedType(DM_##nm##_struct); \
extern DM_##nm##_struct_wrapped DM_##nm##_object

#define DM_DefineConstantObject(name) \
DM_##name##_struct_wrapped DM_##name##_object = { \
    DM_SupportPrefixInitialization(DM_##name##_struct_wrapped,\
                                   DM_##name##_object) \
    DM_ReferenceToClass(Constant), \
    DM_ProxyInitialization(), \
    { { { \
      sizeof(#name), \
      #name \
    } } } \
  }

/* DEBUG: Once upon a time, there was a cpp variable named DM_DEBUG,
 * but it was not convenient to debug when DMeroon was compiled
 * without that variable set. I therefore decide to entirely change
 * the inclusion of debug code which now depends on a macro.  The
 * debug level is stored in the site object and may be changed
 * dynamically.
 */

#ifdef DM_DEBUG
#define DM_DebugAtLevel(level,code) \
  { if ( DM_TheCurrentSite()->debug_level > (level) ) { \
      code; \
    } }
#else
#define DM_DebugAtLevel(level,code) /* Nothing at all */
#endif

/* DEBUG: errors are printed on context->errstream whenever possible.
 * They are printed on DM_debug_stream otherwise.
 */

#define DM_FindErrorStream(context) \
  ((context) ? (context->errstream) : DM_debug_stream)

/****************************************************************************
 * errno is a quasi-reserved keyword. It should not be used as a field
 * name inside a struct. But it is still used to access error numbers.
 * The following macros were deemed to be useful, they're probably not
 * and therefore should be removed.
 */

#if defined(DM_USE_OLD_ERRNO)
#define DM_ErrnoValue() errno
#else
#define DM_ErrnoValue() errnb
#endif /* DM_USE_OLD_ERRNO */

#define DM_ResetErrno() \
   ( DM_ErrnoValue() = 0 )

#define DM_IsErrnoSet() \
   ( DM_ErrnoValue() != 0 ) 

#define DM_IsErrnoEqualTo(e) \
   ( DM_ErrnoValue() == (e) )

/* 
 * RequestMetaClass-es must be associated to a handler. Their
 * structure is too complicated so they are not statically initialized
 * (though that may be done).
 */

#define DM_AssociateHandlerToRequestMetaClass(classname,function) \
  { \
    DM_RequestMetaClass *c = DM_ReferenceToRequestMetaClass(classname); \
    DM_Field *f = DM_ReferenceToFieldOfClass(handler, RequestMetaClass); \
    DM_displacement t = DM_compute_offset(context, (DM_Object*)c, f); \
    void *p = DM_Displace(c, t); \
    DM_Content(DM_handler_code, p) = (DM_handler_code)function; \
  }

#endif /* DM_H */

/* end of dm.h */
