/* $Id: dmact.c,v 1.16 1998/08/23 08:54:17 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmact.c
 * This file contains the commands associated to the urlscript language.
 * Actions have a precise signature:
 *      DM_UrlState* (*)(DM_Context*, DM_UrlAction*, DM_UrlState*, void*, ...)
 * and a precise name DM_<command>_command.

 * FUTURE: add a 'go' primitive to migrate computation.
 *             support for [] notation (as before)
 *             remove extract and modify in favor of explode
 *                explode a POST request into many strings 
 *    Update documentation.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmact,
  "$Id: dmact.c,v 1.16 1998/08/23 08:54:17 queinnec Exp $");

/* {{{ Common utilities
 * RESOURCE DM_register_action
 * Insert an UrlAction into the UrlEnv of the current site. The command
 * may be used for all future URL scripts.
 */

void
DM_register_action (DM_Context   *context,
                    DM_UrlAction *action )
{
  DM_hashtable_insert(context, 
                      (DM_Hashtable**)(&(DM_TheCurrentSite()->urlenv)),
                      (DM_Object*)action );
}

/* RESOURCE DM_compute_hash_command
 * Compute the hashcode of an UrlAction.
 */

static DM_nat3
DM_compute_hash_command (DM_Context   *context,
                         DM_UrlAction *action )
{
  return DM_hashpjw(action->name_length - 1, action->name);
}

/* }}} {{{ UrlScript commands.

 * COMMAND entry 
 * Return the Entry with a given key. Don't dereference the Entry.
 * ... nat3 nat3 nat3 nat3            
 *                         entry
 * ... Entry
 */     

static DM_UrlState*
DM_entry_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Nat3      *arg0,
                  DM_Nat3      *arg1,
                  DM_Nat3      *arg2, 
                  DM_Nat3      *arg3 )
{
  DM_key   key;
  DM_Entry *entry;

  key._rand = arg3->value;
  key._date = arg2->value;
  key._ip   = arg1->value;
  key._idx  = arg0->value >>16;
  key._port = arg0->value & 0xffff;
  entry = DM_extract_entry(context, key);
  if ( entry ) {
    DM_push_object_in_bounded_stack(context, 
                                    (DM_Object*)entry, 
                                    &(state->stack));
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_LOST_ENTRY;
  }
  return state;
}

DM_DefineUrlAction4("entry", entry, Nat3, Nat3, Nat3, Nat3);

/* COMMAND fetch
 * Fetch a remote object.
 * ... Entry
 *           fetch
 * ... Object
 */

static DM_UrlState*
DM_fetch_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Object    *arg0 )
{
  DM_EnsurePresent(context, &arg0);
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("fetch", fetch, Object);

/* COMMAND tick
 * Tick a local clock
 * ... Clock
 *           tick
 * ... 
 */

static DM_UrlState*
DM_tick_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Clock     *arg0 )
{
  if ( !(DM_IsA(context, DM_ObjectToProxy(context, arg0), Exit)) ) {
    /* The clock is local */
    DM_increment_clock(context, arg0);
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_REMOTE_CLOCK;
  }
  return state;
}

DM_DefineUrlAction1("tick", tick, Clock);

/* COMMAND modify
 * Modify an object. The string corresponds to the body of a POST request
 * ie a series of name=value&name=value... with some encoding as in:
 * rest=http%3A%2F%2Flispou.lip6.fr%3A56423%2Fnull&value=....
 *
 * ... String Object
 *                    modify
 * ... Object
 */

static DM_UrlState*
DM_modify_command (DM_Context   *context,
                   DM_UrlAction *self,
                   DM_UrlState  *state,
                   void         *output_port,
                   DM_String    *arg0,
                   DM_Object    *arg1 )
{
  DM_ErrorCode errcode = DME_SUCCESS;
  DM_String *namevaluestring = (DM_String*)DM_clone(context, (DM_Object*)arg0);
  char *name;     /* start of a name field. */
  char *value;    /* start of a value representation. */
  char *p = namevaluestring->character;
  DM_Field       *field;
  DM_DMeroonType *itemtyp;
  DM_DMeroonType *reptyp;
  
  if ( arg1 ) {
    /* Push the object so it is easier to print it after modification. */
    DM_push_object_in_bounded_stack(context, arg1, &(state->stack));
    /* Handle all name=value pairs. First, identify where they are. */
    while ( 1 ) {
      /* look for the first equal sign and change it into a \0. */
      name = p;
      while ( *p != '=' ) {
        if ( *p ) {
          p++;
        } else {
          /* The script is empty. The POST problem ensures that the
           * string has a final \0 and no intervening \0 before. */
          goto return_point;
        }
      }
      *p = '\0';
      DM_decode_percent(context, name, p);
      value = ++p;
      /* Look for the first ampersand sign or the end of string. */
      while ( *p != '&' ) {
        if ( *p ) {
          p++;
        } else {
          goto modify_field;
        }
      }
      *p++ = '\0';
      DM_decode_percent(context, value, p);

    modify_field:
      /* Modify a field, name is its name, value the new value. */
      field = DM_cstring_to_field(context, 
                                  DM_ObjectToClass(context, arg1), 
                                  name );
      if ( field ) {
        (void) DMeroon_field_types(context, &itemtyp, &reptyp, field);
        switch ( itemtyp->index ) {

#define Case(type) \
        case DM_##type##_index: { \
          char *q; \
          DM_##type new = DM_string_to_nat3(context, value, &q); \
          \
          if ( *q != '\0' ) { \
            /* Value is not entirely consumed. */ \
            errcode = DME_URLSCRIPT_BAD_MODIFY_VALUE; \
            goto return_point; \
          } \
          errcode = DMeroon_set_field_value(context, NULL, arg1, field, &new); \
          if ( errcode == DME_SUCCESS ) { \
            continue; \
          } else { \
            goto return_point; \
          } \
        break; \
        }

        Case(nat1)
        case DM_netnat2_index:
        Case(nat2)
        case DM_offset_index:
        case DM_date_index:
        case DM_class_options_index:
        case DM_field_options_index:
        case DM_port_options_index:
        case DM_site_options_index:
        case DM_external_options_index:
        case DM_context_options_index: 
        case DM_file_options_index:
        case DM_request_options_index:
        case DM_netnat3_index:
        Case(nat3)
        Case(nat4)
        Case(int1)
        Case(int2)
        Case(int3)
        Case(int4)
        case DM_char1_index: {
          if ( reptyp->index == DM_nought_index ) {
            /* This is a non indexed char1 field. 
            * Don't verify if there is only one character. */
            DM_char1 new = *value;
          
            errcode = DMeroon_set_field_value(context, NULL, arg1, field, &new);
            if ( errcode == DME_SUCCESS ) {
              continue;
            } else {
              goto return_point;
            }
          } else {
            /* This is an indexed field. */
            DM_nat3 i = 0;
            DM_char1 new = *value;
            
            while ( new ) {
              errcode = DMeroon_set_indexed_field_value(context, NULL, arg1,
                                                        field, i++, &new);
              if ( errcode == DME_SUCCESS ) {
                new = *(++value);
              } else {
                goto return_point;
              }
            }
            continue;
          }
        }
        default: {
          errcode = DME_NOT_YET_IMPLEMENTED;
          goto return_point;
        }
        }

#undef Case

      } else {
        errcode = DME_NO_SUCH_FIELD;
        goto return_point;
      }
    }
  } else {
    errcode = DME_URLSCRIPT_NOT_AN_OBJECT;
    goto return_point;
  }

return_point:
  state->errcode = errcode;
  return state;
}

DM_DefineUrlAction2("modify", modify, String, Object);

/* COMMAND extract
 * Extract from a stringified POST request the string associated to a 
 * given name. For instance, it the string is "a=b&c=def" then extracting
 * c from it will yield the "def" string. 
 * ... String String 
 *                   extract
 * ... String
 */

static DM_UrlState*
DM_extract_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_String    *arg0,
                    DM_String    *arg1 )
{
  DM_String*   str;
  char *name = arg1->character;
  size_t namelength = strlen(name);
  char *p = arg0->character;
  char *start, *result;
  DM_nat3 valuelength;

  while ( 1 ) {
    /* Search for the name. Or it is the first name followed by an
     * equal sign or skip until an ampersand or the end of the string.
     * We detect the end with a \0 (the POST protocol ensures that no
     * intervening \0 may occur before the end). */
    start = p;
    while ( *p != '=' ) {
      if ( *p ) {
        p++;
      } else {
        /* This is the end of the string. Return null. */
        DM_push_object_in_bounded_stack(context, DM_NULL, &(state->stack));
        state->errcode = DME_SUCCESS;
        return state;
      }
    }
    /* Here p designates '=', check if it is the name we are looking for. */
    if ( strncmp(start, name, namelength) == 0 ) {
      /* Found!, isolate the value starting from the '=' until '&' or
       * end of string (both ends excluded) */
      result = ++p;
      while ( *p != '&' ) {
        if ( *p ) {
          p++;
        } else {
          /* This is the end of the value. String-ify it and return it. */
          p++;
          valuelength = p-result;
          goto stringify;
        }
      }
      /* Here p designates '&'. Leave room for the final \0. */
      p++;
      valuelength = p - result;
    stringify:
      str = DM_cstring_to_string(context, result, valuelength);
      str->character[valuelength-1] = '\0';
      DM_push_object_in_bounded_stack(context, 
                                      (DM_Object*)str, 
                                      &(state->stack) );
      state->errcode = DME_SUCCESS;
      return state;
    } else {
      /* This is not the name we're looking for. Skip until '&' then
       * continue to search. */
      while ( *p != '&' ) {
        if ( *p ) {
          p++;
        } else {
          /* This is the end of the string. Return null. */
          DM_push_object_in_bounded_stack(context, DM_NULL, &(state->stack));
          state->errcode = DME_SUCCESS;
          return state;
        }
      }
      p++;
      continue;
    }
  }
}

DM_DefineUrlAction2("extract", extract, String, String);

/* COMMAND print byteprint fancyprint
 * pop and print the object on top of the stack. There is more than one
 * way to print an object.
 * ... Object
 *            print
 * ... 
 */

static DM_UrlState*
DM_print_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Object    *arg0 )
{
  if ( DM_IsA(context, arg0, String) ) {
    DM_DeclareNarrowedObject(string, String, arg0);

    DM_print_char_star(context, 
                       string->character,
                       string->character_length,
                       output_port );
  } else if ( DM_IsA(context, arg0, Value) ) {
    /* This is an instance of the abstract Value class. Print the raw
     * content of its first field. */
    DM_Class       *class = DM_ObjectToClass(context, arg0);
    DM_Field       *field = DM_class_field(context, class, 0);
    DM_DMeroonType *itemtyp;

    (void) DMeroon_field_types(context, &itemtyp, DM_NULL, field);
    DM_local_print(context, itemtyp, arg0, 0, output_port);
  } else {
    /* By default, displays the content of the object. */
    DM_html_print_object(context, arg0, output_port);
  }
  DM_print_char_star(context, "\r\n", 2, output_port);
  state->errcode = DME_SUCCESS;
  return state;
}

static DM_UrlState*
DM_byteprint_command (DM_Context   *context,
                      DM_UrlAction *self,
                      DM_UrlState  *state,
                      void         *output_port,
                      DM_Object    *arg0 )
{
  DM_html_byte_print(context, arg0, output_port);
  state->errcode = DME_SUCCESS;
  return state;
}

static DM_UrlState*
DM_fancyprint_command (DM_Context   *context,
                       DM_UrlAction *self,
                       DM_UrlState  *state,
                       void         *output_port,
                       DM_Object    *arg0 )
{
  DM_html_fancy_print(context, arg0, output_port);
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("print", print, Object);
DM_DefineUrlAction1("byteprint", byteprint, Object);
DM_DefineUrlAction1("fancyprint", fancyprint, Object);

/* COMMAND thesite false true nil null
 * Push the current site, the boolean false, true or the constant nil or
 * the constant null.
 * ... 
 *            thesite
 * ... Site
 */

static DM_UrlState*
DM_thesite_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port )
{
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(DM_TheCurrentSite()),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("thesite", thesite);

#if defined(DM_DEBUG)
/* This command sets the debug level of the current site and returns the 
 * previous debug level.
 * This may be dangerous if done from any http client since it may generate 
 * lots of details on the standard output of the DMeroon server.
 * ... Nat3
 *             setdebuglevel  
 * ... Nat3
 */

static DM_UrlState*
DM_setdebuglevel_command (DM_Context   *context,
                          DM_UrlAction *self,
                          DM_UrlState  *state,
                          void         *output_port,
                          DM_Nat3      *arg0 )
{
  DM_Nat3 *result = DM_Allocate0(context, Nat3);

  result->value = DM_TheCurrentSite()->debug_level;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, result);
  }
  DM_TheCurrentSite()->debug_level = arg0->value;
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(result),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("setdebuglevel", setdebuglevel, Nat3);

#endif /* defined(DM_DEBUG) */

/* RESOURCE
 * Prints the stack "en passant".
 */

static DM_UrlState*
DM_dumpstack_command (DM_Context   *context,
                      DM_UrlAction *self,
                      DM_UrlState  *state,
                      void         *output_port )
{
  DM_dump_urlscript_stack(context, state, output_port);
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("dumpstack", dumpstack);

static DM_UrlState*
DM_false_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port )
{
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(DM_TheConstant(false)),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("false", false);

static DM_UrlState*
DM_true_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port )
{
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(DM_TheConstant(true)),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("true", true);

static DM_UrlState*
DM_nil_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port )
{
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(DM_TheConstant(nil)),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("nil", nil);

static DM_UrlState*
DM_null_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port )
{
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)(DM_NULL),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("null", null);

/* COMMAND dup
 * Duplicates the top of the stack.
 * ... Object 
 *                      dup
 * ... Object Object 
 */

static DM_UrlState*
DM_dup_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_Object    *arg0 )
{
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("dup", dup, Object);

/* COMMAND swap
 * Swap the top and the subtop of the stack.
 * ... Object1 Object2 
 *                      swap
 * ... Object1 Object1 
 */

static DM_UrlState*
DM_swap_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0,
                 DM_Object    *arg1 )
{
  DM_push_object_in_bounded_stack(context, arg1, &(state->stack));
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction2("swap", swap, Object, Object);

/* COMMAND pop
 * Pops top of the stack.
 * ... Object
 *              pop
 * ... 
 */

static DM_UrlState*
DM_pop_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_Object    *arg0 )
{
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("pop", pop, Object);

/* COMMAND ==
 * Compare the top and subtop of the stack. Nat3 and Strings are compared 
 * for their content, other objects are just compared on their address.
 *
 * ... Object1 Object2
 *                      ==
 * ... Boolean
 */

static DM_UrlState*
DM_equal_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Object    *arg0,
                  DM_Object    *arg1 )
{
  DM_Object *result;

  if ( arg0 == arg1 ) {
    result = DM_TheConstant(true);
  } else if (    DM_IsA(context, arg0, Nat3) 
              && DM_IsA(context, arg1, Nat3) ) {
    DM_DeclareNarrowedObject(n0, Nat3, arg0);
    DM_DeclareNarrowedObject(n1, Nat3, arg1);

    if ( n0->value == n1->value ) {
      result = DM_TheConstant(true);
    } else {
      result = DM_TheConstant(false);
    }
  } else if (    DM_IsA(context, arg0, String) 
              && DM_IsA(context, arg1, String) ) {
    DM_DeclareNarrowedObject(string0, String, arg0);
    DM_DeclareNarrowedObject(string1, String, arg1);
    
    if (   ( string0->character_length == string1->character_length )
        && ( strncmp(string0->character, 
                     string1->character,
                     string0->character_length ) == 0 ) ) {
      result = DM_TheConstant(true);
    } else {
      result = DM_TheConstant(false);
    }
  } else {
    result = DM_TheConstant(false);
  }
  DM_push_object_in_bounded_stack(context, result, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction2("==", equal, Object, Object);

/* COMMAND not
 * Negate the top of the stack
 * ... Boolean
 *               not
 * ... Boolean
 */

static DM_UrlState*
DM_not_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_Object    *arg0 )
{
  DM_Object *result;

  if ( arg0 == DM_TheConstant(false) ) {
    result = DM_TheConstant(true);
  } else {
    result = DM_TheConstant(false);
  }
  DM_push_object_in_bounded_stack(context, result, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("not", not, Object);

/* COMMAND if
 * Compare the top and subtop of the stack. If Object is true then return 
 * Object2 otherwise return Object1. Therefore a condition is written as
 *                      else then condition if.
 * ... Object1 Object2 Object 
 *                              if
 * ... Object?
 */

static DM_UrlState*
DM_choose_command (DM_Context   *context,
                   DM_UrlAction *self,
                   DM_UrlState  *state,
                   void         *output_port,
                   DM_Object    *arg0,
                   DM_Object    *arg1,
                   DM_Object    *arg2 )
{
  DM_Object *result;

  if ( arg2 == DM_TheConstant(false) ) {
    result = arg0;
  } else {
    result = arg1;
  }
  DM_push_object_in_bounded_stack(context, result, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction3("if", choose, Object, Object, Object);

/* }}} {{{ Arithmetic commands.
 */

#define DM_DefineArithmeticOperatorCommand(name,op) \
static DM_UrlState* \
DM_##name##_command (DM_Context   *context, \
                     DM_UrlAction *self, \
                     DM_UrlState  *state, \
                     void         *output_port, \
                     DM_Nat3      *arg0, \
                     DM_Nat3      *arg1 ) /* Top of stack */ \
{ \
  DM_Nat3 *result = DM_Allocate0(context, Nat3); \
 \
  result->value = arg1->value op arg0->value; \
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) { \
    DM_SetCopyable(context, result); \
  } \
  DM_DebugAtLevel(5, \
    fprintf(DM_FindErrorStream(context), "@@@ %lu %lu " #name "-> %lu\n", \
            (long unsigned)arg0->value, \
            (long unsigned)arg1->value, \
            (long unsigned)result->value ) \
);  \
  DM_push_object_in_bounded_stack(context, (DM_Object*)result, \
                                  &(state->stack)); \
  state->errcode = DME_SUCCESS; \
  return state; \
} \
DM_DefineUrlAction2(#name, name, Nat3, Nat3)

DM_DefineArithmeticOperatorCommand(plus,+);
DM_DefineArithmeticOperatorCommand(minus,-); 
DM_DefineArithmeticOperatorCommand(times,*);
DM_DefineArithmeticOperatorCommand(quotient,/);
DM_DefineArithmeticOperatorCommand(modulo,%);

#if 0
static DM_UrlState* 
DM_times_command (DM_Context   *context, 
                     DM_UrlAction *self, 
                     DM_UrlState  *state, 
                     void         *output_port, 
                     DM_Nat3      *arg0, 
                     DM_Nat3      *arg1 ) 
{ 
  DM_Nat3 *result = DM_Allocate0(context, Nat3); 
 
  result->value = arg0->value * arg1->value; 
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) { 
    DM_SetCopyable(context, result); 
  } 
  DM_push_object_in_bounded_stack(context, (DM_Object*)result, &(state->stack));
  state->errcode = DME_SUCCESS; 
  return state; 
} 
DM_DefineUrlAction2("times", times, Nat3, Nat3);
#endif

/* }}} {{{ Arithmetic predicates
 */

#define DM_DefineArithmeticComparatorCommand(name,op) \
static DM_UrlState* \
DM_##name##_command (DM_Context   *context, \
                     DM_UrlAction *self, \
                     DM_UrlState  *state, \
                     void         *output_port, \
                     DM_Nat3      *arg0, \
                     DM_Nat3      *arg1 ) \
{ \
  DM_Object* result; \
\
  if ( arg1->value op arg0->value ) { \
    result = DM_TheConstant(true); \
  } else { \
    result = DM_TheConstant(false); \
  } \
  DM_push_object_in_bounded_stack(context, (DM_Object*)result, &(state->stack)); \
  state->errcode = DME_SUCCESS; \
  return state; \
} \
DM_DefineUrlAction2(#name, name, Nat3, Nat3)

DM_DefineArithmeticComparatorCommand(lt,<);
DM_DefineArithmeticComparatorCommand(le,<=);
DM_DefineArithmeticComparatorCommand(gt,>=);
DM_DefineArithmeticComparatorCommand(ge,>);
DM_DefineArithmeticComparatorCommand(ne,!=);

/* }}} {{{ Definitions
 * COMMAND define
 * Define a new command word to be some value. The command is defined with
 * a behavior that simply returns itself as value.
 * ... Object String 
 *                   define
 * ...
 */

static DM_UrlState*
DM_return_internal (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port )
{
  DM_push_object_in_bounded_stack(context, self->internal, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

static DM_UrlAction*
DM_search_action_in_vocabulary (DM_Context *context, 
                                DM_UrlEnv  *env,
                                DM_nat3    hashcode,
                                DM_nat3    length,
                                char       *command_name );

static DM_UrlState*
DM_define_command (DM_Context   *context,
                   DM_UrlAction *self,
                   DM_UrlState  *state,
                   void         *output_port,
                   DM_Object    *arg0,
                   DM_String    *arg1 )
{ 
  /* Don't count the final \0. */
  DM_nat3 namelength = arg1->character_length - 1;
  char    *name = &(arg1->character[0]);
  DM_nat3 hashcode = DM_hashpjw(namelength, name);
  DM_UrlAction *oldaction = 
    DM_search_action_in_vocabulary(context, 
                                   state->environment,
                                   hashcode,
                                   namelength,
                                   name );

  if ( oldaction ) {
    /* There's a already such a command in the current vocabulary. 
     * Note that it is possible to shadow a command. */
    state->errcode = DME_URLSCRIPT_COMMAND_REDEFINITION;
    return state;
  } else {
    DM_UrlAction *action =
      DM_Allocate2(context, UrlAction, namelength+1, 0);
  
    action->behavior = DM_CodeOf(DM_return_internal);
    action->internal = arg0;
    memcpy(&(action->name[0]), name, namelength+1);
    DM_hashtable_insert(context, 
                        (DM_Hashtable**)(&(state->environment)), 
                        (DM_Object*)action );
    state->errcode = DME_SUCCESS;
    return state;
  }
}

DM_DefineUrlAction2("define", define, Object, String);

/* COMMAND defcommand
 * Define a new command that evaluates a given string. This is a little
 * like { String eval } "name" define but far easier to understand.
 * ... String String 
 *                   defcommand
 * ...
 */

static DM_UrlState*
DM_eval_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0 );

static DM_UrlState*
DM_eval_internal (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port )
{
  return DM_eval_command(context, 
                         self, 
                         state, 
                         output_port,
                         self->internal );
}

static DM_UrlState*
DM_define_command_command (DM_Context   *context,
                           DM_UrlAction *self,
                           DM_UrlState  *state,
                           void         *output_port,
                           DM_String    *arg0,
                           DM_String    *arg1 )
{
  /* Don't count the final \0. */
  DM_nat3 namelength = arg1->character_length - 1;
  char   *name = &(arg1->character[0]);
  DM_nat3 hashcode = DM_hashpjw(namelength, name);
  DM_UrlAction *oldaction = 
    DM_search_action_in_vocabulary(context, 
                                   state->environment,
                                   hashcode,
                                   namelength,
                                   name );

  if ( oldaction ) {
    /* There's a already such a command in the current vocabulary. 
     * Note that it is possible to shadow a command. */
    state->errcode = DME_URLSCRIPT_COMMAND_REDEFINITION;
    return state;
  } else {
    DM_UrlAction *action =
      DM_Allocate2(context, UrlAction, namelength+1, 0);
  
    action->behavior = DM_CodeOf(DM_eval_internal);
    action->internal = (DM_Object*)arg0;
    memcpy(&(action->name[0]), name, namelength+1);
    DM_hashtable_insert(context, 
                        (DM_Hashtable**)(&(state->environment)), 
                        (DM_Object*)action );
    state->errcode = DME_SUCCESS;
    return state;
  }
}

DM_DefineUrlAction2("defcommand", define_command, String, String);

/* COMMAND eval
 * Pop and evaluate the top object in the current vocabulary if it is a
 * String. Otherwise just leaves it in the stack (in Lisp terms, this is
 * an autoquote value).
 * ... Object
 *                  eval
 * ... ??? ???? ???
 */

static DM_UrlState*
DM_eval_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0 )
{
  if ( DM_IsA(context, arg0, String) ) {
    DM_DeclareNarrowedObject(string0, String, arg0);
    DM_UrlState *newstate = DM_Allocate0(context, UrlState);
  
    newstate->previous    = state;
    newstate->script      = string0;
    newstate->pc          = 0;
    newstate->stack       = state->stack;
    newstate->environment = state->environment;
    return newstate;
  } else {
    /* Don't eval this thing, just push it back. */
    DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
    state->errcode = DME_SUCCESS;
    return state;
  }
}

DM_DefineUrlAction1("eval", eval, Object);

/* COMMAND go
 * Migrate an evaluation onto a remote machine. The remote computation
 * will find the emitting site on its stack.
 *   ... Site 
 *             go
 *   ... Site
 */

static DM_UrlState*
DM_go_command (DM_Context   *context,
               DM_UrlAction *self,
               DM_UrlState  *state,
               void         *output_port,
               DM_Site*     site )
{
  DM_UrlStateMigrationRequest *request =
    DM_Allocate0(context, UrlStateMigrationRequest);
  DM_UrlStateMigrationAnswer *answer;

  request->options     = DM_REQUEST_DEFAULT_OPTIONS;
  request->requesting  = DM_TheCurrentSite();
  request->requested   = site;
  request->urlstate    = state;
  /* TEMP: This prevents the migrated script to use 'print' and other
   * related commands. To fix that, one has to introduce a remote
   * printing facility.                             FUTURE            */
  request->output      = DM_NULL;  
  DM_SendRequest(context, request, site);
  DM_wait_until_answer(context, 
                       DM_TheCurrentSite()->abnormal_duration,
                       (DM_Answer**)&answer,
                       (DM_Request*)request);
  if ( DM_IsA(context, answer, UrlStateMigrationAnswer) ) {
    return answer->urlstate;
  } else {
    DM_Abort(context, DME_INTERNAL_ERROR);
    /* Not reached. */
    return DM_NULL;
  }
}

DM_DefineUrlAction1("go", go, Site);

/* DOCUMENTATION
 * These are the handlers that manage the reception of an UrlStateMigration
 * Request or Answer. 
 */

void
DM_handle_script_evaluation_request (DM_Context                  *context,
                                     DM_UrlStateMigrationRequest *request ) 
{
  DM_UrlStateMigrationAnswer *answer =
    DM_Allocate0(context, UrlStateMigrationAnswer);

  DM_EnsurePresent(context, &request->urlstate);
  DM_SetCopyable(context, answer);
  answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
  answer->requesting = DM_TheCurrentSite();
  answer->requested  = request->requesting;
  answer->request    = (DM_Request*)request;
  answer->urlstate   = DM_interpret_url_script(context,
                                               request->urlstate,
                                               NULL );
  DM_SendAnswer(context, answer, request->requesting);
  
  DM_DebugAtLevel(0,
      DM_memorize_sent_object(context, (DM_Object*)answer);
);

}

void
DM_handle_script_evaluation_answer (DM_Context                 *context,
                                    DM_UrlStateMigrationAnswer *answer ) 
{
  /* Nothing to do. The answer will be dissecated after the
   * DM_wait_until_answer right before. */
}

/* COMMAND begin
 * Continue evaluation with a new local environment. This is to segregate
 * local commands for instance. The 'end' command allows to pop such
 * a local environment.
 */

static DM_UrlState*
DM_begin_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port )
{
  DM_UrlState *newstate = (DM_UrlState*)DM_clone(context, (DM_Object*)state);
  DM_UrlEnv   *newenv   = DM_Allocate1(context, UrlEnv, 2);

  newenv->hasher        = state->environment->hasher;
  newenv->extractor     = state->environment->extractor;
  newenv->next          = (DM_Hashtable*)(state->environment);
  newenv->size          = 0;
  newstate->environment = newenv;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, newenv);
    DM_SetCopyable(context, newstate);
  }
  return newstate;
}

static DM_UrlState*
DM_end_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port )
{
  DM_UrlEnv *env = state->environment;

  if ( env ) {
    state->environment = (DM_UrlEnv*)(env->next);
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_URLSCRIPT_URLENV_UNDERFLOW;
  }
  return state;
}

DM_DefineUrlAction0("begin", begin);
DM_DefineUrlAction0("end", end);

static DM_UrlState*
DM_use_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_UrlEnv    *urlenv )
{
  DM_UrlEnv *env = state->environment;

  DM_push_object_in_bounded_stack(context, (DM_Object*)env, &(state->stack));
  state->environment = urlenv;
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("use", use, UrlEnv);

/* DOCUMENTATION Reification primitive.
 * At the beginning was the 'vocabulary' and the 'stack' primitive that
 * allows to get access to the current UrlEnv and the current stack (for
 * instance to implement other Forth like primitives). Now there is only
 * the 'state' command that allows to retrieve the previous values as
 * follows: 
 * 
 * { state.stack } "stack" defcommand
 * { state.environment } "vocabulary" defcommand
 *
 * ... 
 *            state
 * ... State
 */

static DM_UrlState*
DM_state_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port )
{
  DM_push_object_in_bounded_stack(context, (DM_Object*)state, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction0("state", state);

/* COMMAND address
 * Pop an address and return the DMeroon object at that address. The address
 * here means the DMeroon reference.
 * ... Nat3
 *             address
 * ... Object
 */

static DM_UrlState*
DM_address_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_Nat3      *arg0 )
{
  DM_Object *o = (DM_Object*)((DM_SupportAddress)(arg0->value));
  if ( DM_IsObject(o) ) {
    DM_push_object_in_bounded_stack(context, o, &(state->stack));
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_NOT_REGULAR_OBJECT;
  }
  return state;
}

DM_DefineUrlAction1("address", address, Nat3);

/* }}} {{{ DMeroon API
 * COMMAND classof
 * Return the class of a DMeroon object.
 * ... Object
 *             classof
 * ... Class
 */

static DM_UrlState*
DM_classof_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_Object    *arg0 )
{
  DM_Class *class = DM_ObjectToClass(context, arg0);

  DM_push_object_in_bounded_stack(context, (DM_Object*)class, &(state->stack));
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("classof", classof, Object);

/* COMMAND siteof
 * Return the class of a DMeroon object.
 * ... Object
 *             siteof
 * ... Site
 */

static DM_UrlState*
DM_siteof_command (DM_Context   *context,
                   DM_UrlAction *self,
                   DM_UrlState  *state,
                   void         *output_port,
                   DM_Object    *arg0 )
{
  DM_Site *site;

  state->errcode = DMeroon_site_of(context, &site, arg0);
  DM_push_object_in_bounded_stack(context, (DM_Object*)site, &(state->stack));
  return state;
}

DM_DefineUrlAction1("siteof", siteof, Object);

/* COMMAND clockof
 * Return the class of a DMeroon object.
 * ... Object
 *             clockof
 * ... Clock
 */

static DM_UrlState*
DM_clockof_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_Object    *arg0 )
{
  DM_Clock *clock;

  state->errcode = DMeroon_clock_of(context, &clock, arg0);
  DM_push_object_in_bounded_stack(context, (DM_Object*)clock, &(state->stack));
  return state;
}

DM_DefineUrlAction1("clockof", clockof, Object);

/* COMMAND proxyof
 * Return the proxy of a DMeroon object.
 * ... Object
 *             proxyof
 * ... External
 */

static DM_UrlState*
DM_proxyof_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_Object    *arg0 )
{
  DM_Object *external = DM_ObjectToProxy(context, arg0);

  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)external, 
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("proxyof", proxyof, Object);

/* COMMAND class
 * Return the predefined class with a given name.
 * ... String
 *             class
 * ... Class
 */

static DM_UrlState*
DM_class_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_String    *arg0 )
{
  DM_Class *class = 
    DM_cstring_to_predefined_class(context, arg0->character);
  
  if ( class ) {
    DM_push_object_in_bounded_stack(context, 
                                    (DM_Object*)class,
                                    &(state->stack) );
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_NO_SUCH_CLASS;
  }
  return state;
}

DM_DefineUrlAction1("class", class, String);

/* COMMAND type
 * Return the predefined type with a given name
 * ... String
 *             type
 * ... Type
 */

static DM_UrlState*
DM_type_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_String    *arg0 )
{
  DM_DMeroonType *type = DM_cstring_to_type(context, arg0->character);
  
  if ( type ) {
    DM_push_object_in_bounded_stack(context, 
                                    (DM_Object*)type,
                                    &(state->stack) );
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_NO_SUCH_TYPE;
  }
  return state;
}

DM_DefineUrlAction1("type", type, String);

/* COMMAND field
 * Return the predefined field of a predefined class with given names.
 * ... Class String
 *                    field
 * ... Field
 */

static DM_UrlState*
DM_field_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Class     *arg0,
                  DM_String    *arg1 )
{
  DM_Field *field = DM_cstring_to_field(context, arg0, arg1->character);
  
  if ( field ) {
    DM_push_object_in_bounded_stack(context, 
                                    (DM_Object*)field,
                                    &(state->stack) );
    state->errcode = DME_SUCCESS;
  } else {
    state->errcode = DME_NO_SUCH_FIELD;
  }
  return state;
}

DM_DefineUrlAction2("field", field, Class, String);

/* COMMAND site
 * Return the site with a given hostname and portnumber.
 * ... String Nat3
 *                    site
 * ... Site
 */

static DM_UrlState*
DM_site_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_String    *arg0,
                 DM_Nat3      *arg1 )
{
  DM_nat3 ip = DM_get_ip_address(context, arg0->character, 0);
  if ( ip != INADDR_NONE ) {
    DM_Site *site = DM_extract_site(context, ip, arg1->value);

    if ( site ) {
      DM_push_object_in_bounded_stack(context, 
                                      (DM_Object*)site,
                                      &(state->stack) );
      state->errcode = DME_SUCCESS;
      return state;
    }
  } 
  state->errcode = DME_NO_SUCH_SITE;
  return state;
}

DM_DefineUrlAction2("site", site, String, Nat3);

/* COMMAND allocate
 * Return the predefined field of a predefined class with given names.
 * ... Class Nat3 ...
 *                     allocate
 * ... Site
 */

static DM_UrlState*
DM_allocate_command (DM_Context   *context,
                     DM_UrlAction *self,
                     DM_UrlState  *state,
                     void         *output_port,
                     DM_Object    *arg0 )
{
  DM_nat3 sizes[DM_MAX_INDEXED_FIELDS];
  DM_nat3 i = 0;
  
  while ( 1 ) {
    if ( DM_IsA(context, arg0, Class) ) {
      DM_DeclareNarrowedObject(class, Class, arg0);
      DM_Object *o = DM_allocate_instance(context, class, i, sizes);

      DM_push_object_in_bounded_stack(context, o, &(state->stack));
      state->errcode = DME_SUCCESS;
      return state;
    } else if ( DM_IsA(context, arg0, Nat3) ) {
      DM_DeclareNarrowedObject(n, Nat3, arg0);

      sizes[i++] = n->value;
      arg0 = DM_pop_from_bounded_stack(context, state->stack);
      continue;
    } else {
      state->errcode = DME_URLSCRIPT_WRONG_ARGUMENT;
      return state;
    }
  }
}

DM_DefineUrlAction1("allocate", allocate, Object);

/* COMMAND clone
 * Clone an object.
 * ... Object
 *            clone
 * ... Object
 */

static DM_UrlState*
DM_clone_command (DM_Context   *context,
                  DM_UrlAction *self,
                  DM_UrlState  *state,
                  void         *output_port,
                  DM_Object    *arg0 )
{
  DM_push_object_in_bounded_stack(context, 
                                  DM_clone(context, arg0),
                                  &(state->stack) );
  state->errcode = DME_SUCCESS;
  return state;
}

DM_DefineUrlAction1("clone", clone, Object);

/* COMMAND setsharable
 * Initialize an object to be sharable.
 * ... Object
 *            setsharable
 * ... Object
 */

static DM_UrlState*
DM_setsharable_command (DM_Context   *context,
                        DM_UrlAction *self,
                        DM_UrlState  *state,
                        void         *output_port,
                        DM_Object    *arg0 )
{
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  state->errcode = DMeroon_set_sharable(context, arg0);
  return state;
}

DM_DefineUrlAction1("setsharable", setsharable, Object);

/* COMMAND setcopyable
 * Initialize an object to be copyable.
 * ... Object
 *            setcopyable
 * ... Object
 */

static DM_UrlState*
DM_setcopyable_command (DM_Context   *context,
                        DM_UrlAction *self,
                        DM_UrlState  *state,
                        void         *output_port,
                        DM_Object    *arg0 )
{
  DM_push_object_in_bounded_stack(context, arg0, &(state->stack));
  state->errcode = DMeroon_set_copyable(context, arg0);
  return state;
}

DM_DefineUrlAction1("setcopyable", setcopyable, Object);

/* COMMAND setsiteof
 * Change the site that owns an object.
 * ... Object Site
 *                 setsiteof
 * ... Site
 */

static DM_UrlState*
DM_setsiteof_command (DM_Context   *context,
                      DM_UrlAction *self,
                      DM_UrlState  *state,
                      void         *output_port,
                      DM_Object    *arg0,
                      DM_Site      *arg1 )
{
  DM_Site *oldsite;

  state->errcode = DMeroon_set_site_of(context, &oldsite, arg0, arg1);
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)oldsite,
                                  &(state->stack) );
  return state;
}

DM_DefineUrlAction2("setsiteof", setsiteof, Object, Site);

/* COMMAND setclockof
 * Change the site that owns an object.
 * ... Object Clock
 *                 setclockof
 * ... Clock
 */

static DM_UrlState*
DM_setclockof_command (DM_Context  *context,
                      DM_UrlAction *self,
                      DM_UrlState  *state,
                      void         *output_port,
                      DM_Object    *arg0,
                      DM_Clock     *arg1 )
{
  DM_Clock *oldclock;

  state->errcode = DMeroon_set_clock_of(context, &oldclock, arg0, arg1);
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)oldclock,
                                  &(state->stack) );
  return state;
}

DM_DefineUrlAction2("setclockof", setclockof, Object, Clock);

/* COMMAND connect
 * Connect to a DMeroon server at a given hostname and port.
 * ... String Nat3 String
 *                        connect
 * ... Site
 */

static DM_UrlState*
DM_connect_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_String    *arg0,
                    DM_Nat3      *arg1,
                    DM_String    *arg2 )
{
  DM_Site *site;

  state->errcode = DMeroon_connect(context, &site, arg0->character,
                            arg1->value, arg2->character );
  DM_push_object_in_bounded_stack(context, (DM_Object*)site, &(state->stack));
  return state;
}

DM_DefineUrlAction3("connect", connect, String, Nat3, String);

/* COMMAND send
 * Send an object towards a site.
 * ... Object Site
 *                  send
 * ... 
 */

static DM_UrlState*
DM_send_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0,
                 DM_Site      *arg1 )
{
  state->errcode = DMeroon_send(context, arg0, arg1);
  return state;
}

DM_DefineUrlAction2("send", send, Object, Site);

/* COMMAND receive
 * Receive an object. Pushes that object and a boolean to tell whether 
 * an object was received.
 * ... 
 *                       receive
 * ... Object Boolean
 */

static DM_UrlState*
DM_receive_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port )
{
  DM_Object *o;
  DM_Object *result;
  
  state->errcode = DMeroon_receive(context, &o, 0);
  DM_push_object_in_bounded_stack(context, o, &(state->stack));
  if ( state->errcode == DME_SUCCESS ) {
    result = DM_TheConstant(true);
  } else {
    result = DM_TheConstant(false);
  }
  DM_push_object_in_bounded_stack(context, result, &(state->stack));
  return state;
}

DM_DefineUrlAction0("receive", receive);

/* COMMAND get
 * Get the content of a field 
 * ... Object Field
 *                  get
 * ... Object
 */

static DM_UrlState*
DM_get_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_Object    *arg0,
                DM_Field     *arg1 )
{
  DM_DMeroonType *itemtyp;
  DM_Object *o;
  
  (void) DMeroon_field_types(context, &itemtyp, DM_NULL, arg1);
  switch ( itemtyp->index ) {
  case DM_reference_index: {
    state->errcode = DMeroon_field_value(context, &o, DM_NULL, arg0, arg1);
    DM_push_object_in_bounded_stack(context, o, &(state->stack));
    return state;
  }
  default: {
    state->errcode = DME_URLSCRIPT_NOT_AN_OBJECT;
    return state;
  }
  }
}

DM_DefineUrlAction2("get", get, Object, Field);

/* COMMAND set
 * Set the content of a field, return the former content.
 * ... Object Field Object
 *                          set
 * ... Object
 */

static DM_UrlState*
DM_set_command (DM_Context   *context,
                DM_UrlAction *self,
                DM_UrlState  *state,
                void         *output_port,
                DM_Object    *arg0,
                DM_Field     *arg1,
                DM_Object    *arg2 )
{
  DM_Object *old;
  DM_DMeroonType *itemtyp;
  
  (void) DMeroon_field_types(context, &itemtyp, DM_NULL, arg1);
  switch ( itemtyp->index ) {
  case DM_reference_index: {
    state->errcode = DMeroon_set_field_value(context, &old, arg0, arg1, &arg2);
    DM_push_object_in_bounded_stack(context, old, &(state->stack));
    return state;
  }
  default: {
    state->errcode = DME_URLSCRIPT_INCOMPATIBLE_ARGUMENT;
    return state;
  }
  }
}

DM_DefineUrlAction3("set", set, Object, Field, Object);

/* COMMAND xget
 * Get the content of a field 
 * ... Object Field Nat3
 *                       xget
 * ... Object
 */

static DM_UrlState*
DM_xget_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0,
                 DM_Field     *arg1,
                 DM_Nat3      *arg2 )
{
  DM_Object *o;
  DM_DMeroonType *itemtyp;
  
  (void) DMeroon_field_types(context, &itemtyp, DM_NULL, arg1);
  switch ( itemtyp->index ) {
  case DM_reference_index: {
    state->errcode = DMeroon_indexed_field_value(context, &o, DM_NULL,
                                          arg0, arg1, arg2->value );
    DM_push_object_in_bounded_stack(context, o, &(state->stack));
    return state;
  }
  default: {
    state->errcode = DME_URLSCRIPT_INCOMPATIBLE_ARGUMENT;
    return state;
  }
  }
}

DM_DefineUrlAction3("xget", xget, Object, Field, Nat3);

/* COMMAND xlength
 * Get the length of an indexed field 
 * ... Object Field 
 *                   xlength
 * ... Nat3
 */

static DM_UrlState*
DM_xlength_command (DM_Context   *context,
                    DM_UrlAction *self,
                    DM_UrlState  *state,
                    void         *output_port,
                    DM_Object    *arg0,
                    DM_Field     *arg1 )
{
  DM_nat4 length;
  
  state->errcode = 
    DMeroon_indexed_field_length(context, &length, DM_NULL, arg0, arg1);
  if ( state->errcode == DME_SUCCESS ) {
    DM_Nat3 *result = DM_Allocate0(context, Nat3);
    
    /* Truncates nat4 into nat3. */
#ifdef DM_HAS_LONG_LONG_ENTITIES
    result->value = length;
#else
    result->value = length._low;
#endif
    DM_push_object_in_bounded_stack(context, 
                                    (DM_Object*)result,
                                    &(state->stack) );
    state->errcode = DME_SUCCESS;
  }
  return state;
}

DM_DefineUrlAction2("xlength", xlength, Object, Field);

/* COMMAND xset
 * Set the content of an indexed field, return the former content.
 * ... Object Field Nat3 Object
 *                               xset
 * ... Object
 */

static DM_UrlState*
DM_xset_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Object    *arg0,
                 DM_Field     *arg1,
                 DM_Nat3      *arg2,
                 DM_Object    *arg3 )
{
  DM_Object *o;
  DM_DMeroonType *itemtyp;
  
  (void) DMeroon_field_types(context, &itemtyp, DM_NULL, arg1);
  switch ( itemtyp->index ) {
  case DM_reference_index: {
    state->errcode = DMeroon_set_indexed_field_value(context, &o, 
                                              arg0, arg1, arg2->value, &arg3 );
    DM_push_object_in_bounded_stack(context, o, &(state->stack));
    return state;
  }
  default: {
    state->errcode = DME_URLSCRIPT_INCOMPATIBLE_ARGUMENT;
    return state;
  }
  }
}

DM_DefineUrlAction4("xset", xset, Object, Field, Nat3, Object);

/* COMMAND createsubclass
 * Create a subclass.
 * ... Class String Nat3 String Nat3 Type Type Class
 *                                                    createsubclass
 * ... Class
 */

static DM_UrlState*
DM_createsubclass_command (DM_Context   *context,
                           DM_UrlAction *self,
                           DM_UrlState  *state,
                           void         *output_port,
                           DM_Class       *arg0,
                           DM_String      *arg1,
                           DM_Nat3        *arg2,
                           DM_String      *arg3,
                           DM_Nat3        *arg4,
                           DM_DMeroonType *arg5,
                           DM_DMeroonType *arg6,
                           DM_Class       *arg7 )
{
  DM_Class *class;
  
  state->errcode = DMeroon_create_subclass(context, &class,
                                    arg0, arg1->character,
                                    arg2->value, arg3->character,
                                    arg4->value, arg5, arg6, arg7 );
  DM_push_object_in_bounded_stack(context, (DM_Object*)class, &(state->stack));
  return state;
}

DM_DefineUrlAction8("createsubclass", createsubclass, 
                    Class, String, Nat3, String, Nat3, Type, Type, Class );

/* COMMAND smashclass
 * ... Class String Nat3 Nat3
 *                              smashclass
 * ... Class
 */

static DM_UrlState*
DM_smashclass_command (DM_Context   *context,
                       DM_UrlAction *self,
                       DM_UrlState  *state,
                       void         *output_port,
                       DM_Class     *arg0,
                       DM_String    *arg1,
                       DM_Nat3      *arg2 )
{
  DM_Class *newclass;

  state->errcode = DMeroon_smash_class(context, &newclass,
                                arg0, arg1->character, 
                                arg2->value, arg2->value );
  DM_push_object_in_bounded_stack(context, 
                                  (DM_Object*)newclass,
                                  &(state->stack) );
  return state;
}

DM_DefineUrlAction4("smashclass", smashclass, Class, String, Nat3, Nat3);

/* COMMAND exit
 * Exits immediately the URLscript interpreter and return the top of the
 * stack as error code (this may be DME_SUCCESS). 
 *   ... Nat3
 *              exit
 *
 * FUTURE: The current implementation prevents error catching. It
 * might also be interesting to leave the current state on the stack
 * to allow inspection. 
 */

static DM_UrlState*
DM_exit_command (DM_Context   *context,
                 DM_UrlAction *self,
                 DM_UrlState  *state,
                 void         *output_port,
                 DM_Nat3      *arg0 )
{
  /* Remove the continuation and the script to interpret. */
  state->previous = DM_NULL;
  state->pc = state->script->character_length;
  state->errcode = arg0->value;
  return state;
}

DM_DefineUrlAction1("exit", exit, Nat3);

/* }}} {{{ Interpreter

 * This function is used as the 'extractor' of UrlEnv. It ensures that 
 * UrlActions extracted from the table are present. 
 */

static DM_Object*
DM_ensure_urlaction_present (DM_Context *context,
                             DM_Object* *o )
{
  if ( *o ) {
    DM_EnsurePresent(context, o);
  }
  return *o;
}


/* RESOURCE DM_search_action
 * Search a command in an environment. Return NULL if no command of that 
 * name is found. UrlEnv have as extractor the previous function so 
 * entries in UrlEnv are always present when compared.
 */

static DM_UrlAction*
DM_search_action_in_vocabulary (DM_Context *context, 
                                DM_UrlEnv  *env,
                                DM_nat3    hashcode,
                                DM_nat3    length,
                                char       *command_name )
{
  DM_HashtableExtract(context,
                      (DM_Hashtable*)env, 
                      DM_UrlAction*,
                      hashcode,
                      (    (length == object->name_length-1)
                        && (strncmp(command_name,
                                    object->name,
                                    length ) == 0 ) )
                      );
}

static DM_UrlAction*
DM_search_action (DM_Context *context, 
                  DM_UrlEnv  *env,
                  DM_nat3    length,
                  char       *name )
{
  DM_nat3 hashcode = DM_hashpjw(length, name);
  DM_UrlAction* result = DM_NULL;

  do {
    if ( env ) {
      DM_EnsurePresent(context, &env);
      /* Search in current vocabulary. */
      result = DM_search_action_in_vocabulary(context, env, hashcode, 
                                              length, name);
      env = (DM_UrlEnv*)(env->next);
    } else {
      return DM_NULL;
    }
  } while ( result == DM_NULL );
  return result;
}

/* RESOURCE DM_search_field
 * Given a Class, return the field that has a particular name.
 */

DM_Field*
DM_search_field (DM_Context *context, 
                 DM_Class   *class,
                 DM_nat3    length, 
                 char       *field_name )
{
  DM_nat3 fn;
  DM_nat3 i;

  DM_ensure_class_usable(context, class);
  fn = DM_class_indexed_field_length(context, class);
  for ( i=0 ; i<fn ; i++ ) {
    DM_Field *field = DM_class_field(context, class, i);

    if ( strncmp(field_name,
                 DM_predefined_name(context, (DM_Object*)field),
                 length ) == 0 ) {
      return field;
    }
  }
  return DM_NULL;
}

/* RESOURCE DM_interpret_url_script
 * Process a script and output its result on the output stream. The url
 * script language is a tiny, stack-based language. The final NUL character
 * of the script marks the end of the program. The program may extend on
 * several lines. When called from a 'GET' request, the script is truncated
 * to its first line. 
 * 
 * This is the basic interpreter, it only recognizes (and pushes)
 * constants such as characters (surrounded with ''), strings
 * (surrounded with "" or by matching {}), positive numbers (prefix 0
 * or 0x are allowed). Other words are considered as commands that
 * must be run. Comments are introduced with #. Fields within objects
 * may be read with the usual dot notation provided they lead to a 
 * DMeroon object. Commands may be separated with whitespaces or /.
 */

DM_UrlState*
DM_interpret_url_script (DM_Context  *context,
                         DM_UrlState *state,
                         void        *output_port ) 
{
  char *script;
  char c;

  /* Don't pre-fetch the previous state yet. */
  DM_EnsurePresent(context, &(state->script));
  DM_EnsurePresent(context, &(state->stack));

  while ( 1 ) {

    if ( state->pc >= state->script->character_length ) {
      /* The script is finished but does not end with \0. */
      c = '\0';
    } else {
      script = state->script->character;
      c = script[state->pc];
    }

    switch ( c ) {

    case '\0': {
      /* This is the end of the script. */
      if ( state->previous ) {
        /* pop the control stack back to the previous script but keep
         * the current stack and environment. */
        DM_EnsurePresent(context, &(state->previous));
        state->previous->stack = state->stack;
        state->previous->environment = state->environment;
        state = state->previous;
        DM_EnsurePresent(context, &(state->script));
        continue;
      } else {
        state->errcode = DME_SUCCESS;
        return state;
      }
    }

    case '#': {
      /* Skip comments until the end of the line ie an LF character. This 
       * allows for LF ended lines as in Unix or CR-LF lines as in DOS. */
      do {
        state->pc += 1;
        c = script[state->pc];
        if ( c == '\0' ) {
          state->errcode = DME_SUCCESS;
          return state;
        }
      } while ( c != '\n' );
      state->pc += 1;
      continue;
    }

    case ' ':
    case '\t':
    case '\n':
    case '\r':
    case '/': {
      /* Skip over spaces and equivalent characters. */
      state->pc += 1;
      continue;
    }

    case '.': {
      /* Dot means an access to a field of the current object. */
      char     *q     = &(script[(state->pc)+1]);
      DM_nat3  length = 0;
      DM_Field *field;

      /* Compute the length of the name of the field. */
      while ( 1 ) {
        switch ( *q ) {
        case '\0':
        case '#':
        case '.':
        case '\'':
        case '\"':
        case ' ': 
        case '\t':
        case '\n':
        case '\r':
        case '/': {
          goto lookup_field;
        }
        default: {
          length++;
          q++;
        }
        }
      }
      
    lookup_field: 
      if ( state->stack->index > 0 ) {
        DM_Object *o = DM_pop_from_bounded_stack(context, state->stack);

        DM_EnsurePresent(context, &o);
        q = &(script[(state->pc)+1]);
        field = DM_search_field(context, 
                                DM_ObjectToClass(context, o),
                                length, 
                                q );
        if ( field ) {
          DM_DMeroonType *dmtype;
          DM_Object      *oo;

          state->pc += 1;          /* Skip the dot. */
          (void) DMeroon_field_types(context, &dmtype, DM_NULL, field);
          /* FUTURE: check for index ???????????????????????? */
          if ( dmtype->index != DM_reference_index ) {
            /* FUTURE: create a Value ????????????????????????? */
            state->errcode = DME_URLSCRIPT_NOT_AN_OBJECT;
            return state;
          }
          state->errcode = DMeroon_field_value(context, &oo, &dmtype, o, field);
          if ( DME_SUCCESS == state->errcode ) {
            DM_push_object_in_bounded_stack(context,
                                            (DM_Object*)oo,
                                            &(state->stack) );
            state->pc += length;  /* Skip the name of the field. */
            continue;
          } else {
            /* Something wrong happened in the API call. */
            return state;
          }
        } else {
        /* No field-name was recognized. */
          state->errcode = DME_URLSCRIPT_UNKNOWN_FIELD;
          return state;
        }
      } else {
        /* There should be at least one object in the stack. */
        state->errcode = DME_URLSCRIPT_STACK_UNDERFLOW;
        return state;
      }
    }

    case '\'': {
      /* Seems to be a character to be pushed. */
      if ( script[(state->pc) + 2] == '\'' ) {
        DM_Char1 *o = DM_Allocate0(context, Char1);
        
        o->value = script[(state->pc) + 1];
        if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
          DM_SetCopyable(context, o);
        }
        DM_push_object_in_bounded_stack(context, 
                                        (DM_Object*)o, 
                                        &(state->stack) );
        state->pc += 3;    /* Skip the syntax of the character: 'x' */
        continue;
      } else {
        state->errcode = DME_URLSCRIPT_BAD_CHARACTER_SYNTAX;
        return state;
      }
    }

    case '"': {
      /* Read and push a String. */
      DM_String* o;
      char *q = &(script[(state->pc)+1]);
      unsigned long length = 0;
      
      /* First, compute the length of the string. */
      while ( 1 ) {
        switch ( *q ) {
        case '\0': {
          state->errcode = DME_URLSCRIPT_BAD_STRING_SYNTAX;
          return state;
        case '"': {
          goto make_string;
        }
        case '\\': {
          length++;
          q += 2;
          continue;
        }
        default: {
          length++;
          q++;
          continue;
        }
        }
      }
      }

    make_string:
      o =DM_Allocate1(context, String, length+1);
      length = 0;
      state->pc += 1;          /* Skip the opening ". */
      while ( 1 ) {
        char c = script[(state->pc)++];

        switch ( c ) {
        case '"': {
          /* state->pc is already on the character that follows 
           * the closing ". */
          goto string_made;
        }
        case '\\': {
          o->character[length++] = script[(state->pc)++];
          continue;
        }
        default: {
          o->character[length++] = c;
          continue;
        }
        }
      }
      
    string_made:
      if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
        DM_SetCopyable(context, o);
      }
      DM_push_object_in_bounded_stack(context, 
                                      (DM_Object*)o,
                                      &(state->stack) );
      continue;
    }

    case '{': {
      /* This is also a String ending with a matching }. The eval
       * command may evaluate it. 
       * FUTURE why not make it an executable string ? */
      DM_String *string;
      char *q = &(script[(state->pc) + 1]);
      DM_nat3 level = 1;   /* Count inner matching { and }. */
      DM_nat3 length = 0;
      DM_c_boolean within_string = DM_FALSE;

      /* Compute the length of the string until the matching }. */
      while ( 1 ) {
        char c = *q++;

        switch ( c ) {
        case '{': {
          if ( ! within_string ) {
            level++;
          }
          length++;
          continue;
        }
        case '}': {
          /* End of script unless within string. */
          if ( ! within_string ) {
            level--;
            if ( level == 0 ) {
              goto allocate_script;
            }
          }
          length++;
          continue;
        }
        case '"': {
          /* Toggle in_string. */
          if ( within_string ) {
            within_string = DM_FALSE;
          } else {
            within_string = DM_TRUE;
          }
          length++;
          continue;
        }
        case '\\': {
          /* Ignore next character. */
          q++;
          length += 2;
          continue;
        }
        default: {
          length++;
          continue;
        }
        }
      }

      /* Allocate and push a String (don't forget the final NUL char). */
    allocate_script:
      string = DM_Allocate1(context, String, length+1);
      memcpy(&(string->character[0]), &(script[state->pc + 1]), length);
      string->character[length] = '\0';
      if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
        DM_SetCopyable(context, string);
      }
      DM_push_object_in_bounded_stack(context, 
                                      (DM_Object*)string, 
                                      &(state->stack) );
      state->pc += 1+length+1;
      continue;
    }

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9': {
      /* Seems to be a number. */
      char *q = &(script[state->pc]);
      char *p;
      DM_nat3 n = DM_string_to_nat3(context, q, &p);

      if ( p == q ) {
        /* No number was recognized. */
        state->errcode = DME_URLSCRIPT_BAD_NUMBER_SYNTAX;
        return state;
      } else {
        DM_Nat3 *o = DM_Allocate0(context, Nat3);

        o->value = n;
        if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
          DM_SetCopyable(context, o);
        }
        DM_push_object_in_bounded_stack(context, 
                                        (DM_Object*)o, 
                                        &(state->stack) );
        state->pc += p - &(script[state->pc]);
        continue;
      }
    }
    
    default: {
      char          *q     = &(script[state->pc]);
      DM_nat3       length = 0;
      DM_UrlAction* action;

      /* Compute the length (the number of characters) of the word. */
      while ( 1 ) {
        switch ( *q ) {
        case '\0':
        case '#':
        case '.':
        case '\'':
        case '\"':
        case ' ': 
        case '\t':
        case '\n':
        case '\r':
        case '/': {
          goto lookup_command;
        }
        default: {
          length++;
          q++;
        }
        }
      }
      
    lookup_command: 
      q = &(script[state->pc]);
      action = DM_search_action(context, 
                                state->environment, 
                                length, 
                                q );
      if ( DM_IsA(context, action, UrlAction) ) {
        DM_DeclareActionBehavior(function, action->behavior);

        /* Verify arity and classes of arguments. */
        { 
          DM_nat1 arity;
          DM_BoundedStack *stack = state->stack;
          DM_nat3 height = stack->index;

          DMeroon_indexed_field_length(context, &arity, DM_NULL, 
                  (DM_Object*)action,
                  DM_ReferenceToFieldOfClass(argument, UrlAction) );
          if ( arity <= height ) {
            DM_Class *argclass;
            DM_nat1  i;

            for ( i=0 ; i<arity ; i++ ) {
              DM_Object *arg;

              /* Make sure the argument is present. */
              DM_EnsurePresent(context, &(stack->item[height-arity+i]));
              arg = stack->item[height-arity+i];
              /* Make sure the class of the expected argument is present. */
              DMeroon_indexed_field_value(context, &argclass, DM_NULL,  
                    (DM_Object*)action,
                    DM_ReferenceToFieldOfClass(argument, UrlAction), i );
              if (    ( arg == DM_NULL ) 
                   || DM_is_a(context, arg, argclass) ) {
                continue;
              } else {
                state->errcode = DME_URLSCRIPT_WRONG_ARGUMENT;
                return state;
              }
            }
            /* Prepare return address (ie update pc). */
            state->pc += length;
            switch ( arity ) {
            case 0: {
              state = function(context, action, state, output_port);
              break;
            }
            case 1: {
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1 );
              break;
            }
            case 2: {
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2 );
              break;
            }
            case 3: {
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3 );
              break;
            }
            case 4: {
              DM_Object *arg4 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3, arg4 );
              break;
            }
            case 5: {
              DM_Object *arg5 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg4 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3, arg4, arg5 );
              break;
            }
            case 6: {
              DM_Object *arg6 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg5 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg4 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3, arg4, arg5, arg6 );
              break;
            }
            case 7: {
              DM_Object *arg7 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg6 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg5 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg4 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3, arg4, arg5, arg6, arg7 );
              break;
            }
            case 8: {
              DM_Object *arg8 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg7 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg6 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg5 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg4 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg3 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg2 = DM_pop_from_bounded_stack(context, stack);
              DM_Object *arg1 = DM_pop_from_bounded_stack(context, stack);

              state = function(context, action, state, output_port, 
                               arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8 );
              break;
            }
            default : {
              state->errcode = DME_URLSCRIPT_TOO_MUCH_ARGUMENTS;
              return state;
            }
          }
          } else {
            state->errcode = DME_URLSCRIPT_BAD_ARITY;
            return state;
          }
        }
        /* Check resulting state */
        if ( state->errcode != DME_SUCCESS ) {
          return state;
        } else {
          continue;
        }
      } else {
        /* No word was recognized. */
        state->errcode = DME_URLSCRIPT_UNKNOWN_COMMAND;
        return state;
      }
    }
    }
  }
}

/* }}}
 * Initialization
 */

void
DM_act_initialize (DM_Context *context)
{
  DM_LinkVersionStringIntoSite(dmact);

  /* Register all commands but, first initialize the urlenv field of
   * the current site with the predefined standard environment of the
   * UrlScript language. */
  {
    DM_UrlEnv *env = DM_Allocate1(context, UrlEnv, 253);

    env->hasher    = DM_CodeOf(DM_compute_hash_command);
    env->extractor = DM_CodeOf(DM_ensure_urlaction_present);
    env->next      = DM_NULL;
    env->size      = 0;
    DM_TheCurrentSite()->urlenv = env;
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      DM_SetCopyable(context, env);
    }
  }
        
  DM_RegisterUrlAction(context, entry);
  DM_RegisterUrlAction(context, fetch);
  DM_RegisterUrlAction(context, print);
  DM_RegisterUrlAction(context, byteprint);
  DM_RegisterUrlAction(context, fancyprint);
  DM_RegisterUrlAction(context, thesite);
  DM_RegisterUrlAction(context, dup);
  DM_RegisterUrlAction(context, swap);
  DM_RegisterUrlAction(context, pop);
  DM_RegisterUrlAction(context, equal);
  DM_RegisterUrlAction(context, not);
  DM_RegisterUrlAction(context, false);
  DM_RegisterUrlAction(context, true);
  DM_RegisterUrlAction(context, nil);
  DM_RegisterUrlAction(context, null);
  DM_RegisterUrlAction(context, eval);
  DM_RegisterUrlAction(context, go);
  DM_RegisterUrlAction(context, begin);
  DM_RegisterUrlAction(context, end);
  DM_RegisterUrlAction(context, use);
  DM_RegisterUrlAction(context, state);
  DM_RegisterUrlAction(context, dumpstack);
  DM_RegisterUrlAction(context, exit);
  DM_RegisterUrlAction(context, choose);
  DM_RegisterUrlAction(context, modify);
  DM_RegisterUrlAction(context, extract);
  DM_RegisterUrlAction(context, define);
  DM_RegisterUrlAction(context, define_command);
  DM_RegisterUrlAction(context, tick);
  DM_RegisterUrlAction(context, classof);
  DM_RegisterUrlAction(context, siteof);
  DM_RegisterUrlAction(context, clockof);
  DM_RegisterUrlAction(context, class);
  DM_RegisterUrlAction(context, type);
  DM_RegisterUrlAction(context, field);
  DM_RegisterUrlAction(context, site);
  DM_RegisterUrlAction(context, allocate);
  DM_RegisterUrlAction(context, clone);
  DM_RegisterUrlAction(context, setsharable);
  DM_RegisterUrlAction(context, setcopyable);
  DM_RegisterUrlAction(context, setsiteof);
  DM_RegisterUrlAction(context, setclockof);
  DM_RegisterUrlAction(context, connect);
  DM_RegisterUrlAction(context, send);
  DM_RegisterUrlAction(context, receive);
  DM_RegisterUrlAction(context, get);
  DM_RegisterUrlAction(context, xget);
  DM_RegisterUrlAction(context, set);
  DM_RegisterUrlAction(context, xset);
  DM_RegisterUrlAction(context, xlength);
  DM_RegisterUrlAction(context, createsubclass);
  DM_RegisterUrlAction(context, smashclass);

  /* Arithmetic */
  DM_RegisterUrlAction(context, plus);
  DM_RegisterUrlAction(context, minus);
  DM_RegisterUrlAction(context, times);
  DM_RegisterUrlAction(context, quotient);
  DM_RegisterUrlAction(context, modulo);
  DM_RegisterUrlAction(context, lt);
  DM_RegisterUrlAction(context, le);
  DM_RegisterUrlAction(context, gt);
  DM_RegisterUrlAction(context, ge);
  DM_RegisterUrlAction(context, ne);

#if defined(DM_DEBUG)
  /* These are dangerous commands that are there for testing purposes. */
  DM_RegisterUrlAction(context, proxyof);
  DM_RegisterUrlAction(context, setdebuglevel);
  DM_RegisterUrlAction(context, address);
#endif /* defined(DM_DEBUG) */

  /* Record handlers for request/answer. */

DM_AssociateHandlerToRequestMetaClass(UrlStateMigrationRequest,DM_handle_script_evaluation_request);
DM_AssociateHandlerToRequestMetaClass(UrlStateMigrationAnswer,DM_handle_script_evaluation_answer);

}

/* end of dmact.c */
