/* $Id: dmcache.c,v 1.9 1998/04/13 20:31:45 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 
 * DOCUMENTATION dmcache.c
 * This file contains the functions to handle the communication caches.
 * These caches are also used as regular stacks (with push and pop).
 */

#include "dmeroon.h"

DM_DefineVersionString(dmcache,
  "$Id: dmcache.c,v 1.9 1998/04/13 20:31:45 queinnec Exp $");

/* {{{ Cache utilities
 *
 * Report on stack usage above that level of debug. 
 */

#define DM_DEBUG_CACHE 7

/* RESOURCE DM_push_object_in_bounded_stack
 * Push an object into a stack. Double the size of that stack if needed.
 * Return the index of the pushed object.
 */

DM_nat3
DM_push_object_in_bounded_stack (DM_Context       *context,
                                 DM_Object        *o,
                                 DM_BoundedStack* *location ) {
  DM_DeclareNarrowedObject(stack, BoundedStack, *location);
  DM_nat3 i, used_index;
  DM_nat3 init_status;

  DM_DebugAtLevel(DM_DEBUG_CACHE,
    fprintf(DM_FindErrorStream(context), 
            "Pushed " DM_0X "%p in stack " DM_0X "%p[%lu]\n",
            o, stack, (unsigned long)stack->index );
);

  used_index = stack->index;
  if ( stack->index >= stack->item_length ) {
    /* Increase the size of the stack and replace the old one. */
    DM_BoundedStack *newstack = DM_Allocate1(context,
                                              BoundedStack,
                                              2*stack->item_length );
    *location = newstack;
    /* Give to the new stack the same status as the old one. */
    (void) DM_do_is_initialized(context, &init_status, (DM_Object*)stack);
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      switch ( init_status ) {
      case DM_OBJECT_IS_COPYABLE: {
        DM_SetCopyable(context, newstack);
        break;
      }
      case DM_OBJECT_IS_SHARABLE: {
        DM_SetSharable(context, newstack);
        break;
      }
      default: {
        /* The previous stack was uninitialized. Leave the new one
         * uninitialized too. */
      }
      }
    }
    for ( i=0 ; i<stack->item_length ; i++ ) {
      newstack->item[i] = stack->item[i];
    }
    newstack->item[i++] = o;
    newstack->index     = i;
  } else {
    /* The stack is sufficiently big to hold this new object. */
    stack->item[stack->index++] = o;
  }
  return used_index;
}

/* RESOURCE DM_pop_from_bounded_stack
 * Clear popped positions to avoid keeping objects too long. This will 
 * ease GC.
 */

DM_Object*
DM_pop_from_bounded_stack (DM_Context      *context,
                           DM_BoundedStack *stack ) {
  if ( stack->index > 0 ) {
    DM_Object* o;

    o = stack->item[--(stack->index)];
    stack->item[stack->index] = DM_NULL;
    return o;
  } else {
    DM_Abort(context, DME_STACK_UNDERFLOW);
    return DM_NULL;
  }
}

/* RESOURCE DM_top_of_bounded_stack
 * Return the object on top of the stack but don't pop it.
 */

DM_Object*
DM_top_of_bounded_stack (DM_Context      *context,
                         DM_BoundedStack *stack ) {
  if ( stack->index > 0 ) {
    return stack->item[stack->index];
  } else {
    DM_Abort(context, DME_STACK_UNDERFLOW);
    return DM_NULL;
  }
}

/* RESOURCE DM_adjust_stack
 * Reset stack height to a given position.
 */

void
DM_adjust_stack (DM_Context      *context,
                 DM_BoundedStack *stack,
                 DM_nat3         height ) 
{
  if ( stack->index >= height ) {
    while ( stack->index > height ) {
      (void) DM_pop_from_bounded_stack(context, stack);
    }
  } else {
    DM_Abort(context, DME_STACK_UNDERFLOW);
  }
}

/* RESOURCE DM_swap_top_of_bounded_stack
 */

void
DM_swap_top_of_bounded_stack (DM_Context      *context,
                              DM_BoundedStack *stack ) {
  if ( stack->index > 1 ) {
    /* Don't pay attention to extractor. */
    DM_Object *o1 = stack->item[stack->index-1];
    DM_Object *o2 = stack->item[stack->index-2];
    stack->item[stack->index-1] = o2;
    stack->item[stack->index-2] = o1;
  } else {
    DM_Abort(context, DME_STACK_UNDERFLOW);
  }
}

/* RESOURCE DM_dup_top_of_bounded_stack
 */

void 
DM_dup_top_of_bounded_stack (DM_Context       *context,
                             DM_BoundedStack* *stack_location ) {
  DM_BoundedStack *stack = *stack_location;

  if ( stack->index > 0 ) {
    DM_push_object_in_bounded_stack(context,
                                    stack->item[stack->index-1],
                                    stack_location );
  } else {
    DM_Abort(context, DME_STACK_UNDERFLOW);
  }
}

/* RESOURCE DM_reset_stack
 * Empties a stack.
 */

void
DM_reset_stack (DM_Context      *context,
                DM_BoundedStack *stack) 
{
  DM_nat3 i;
  
  DM_DebugAtLevel(DM_DEBUG_CACHE,
    fprintf(DM_FindErrorStream(context), 
            "Reset stack " DM_0X "%p\n", stack);
);

  for ( i=0 ; i<stack->index ; i++ ) {
    stack->item[i] = DM_NULL;
  }
  stack->index = 0;
}

/* RESOURCE DM_reset_cache
 * Empties a cache.
 */

void
DM_reset_cache (DM_Context      *context,
                DM_MutableVector *cache )
{
  DM_nat3 i;
  
  DM_DebugAtLevel(DM_DEBUG_CACHE,
    fprintf(DM_FindErrorStream(context), 
            "Reset cache " DM_0X "%p\n", cache);
);

  for ( i=0 ; i<cache->item_length ; i++ ) {
    cache->item[i] = DM_NULL;
  }
}

/* RESOURCE DM_extract_from_cache
 * Extracts an Object from a Cache.
 */

DM_Object*
DM_extract_from_cache (DM_Context       *context,
                       DM_MutableVector *cache,
                       DM_nat3          index ) {

  DM_DebugAtLevel(DM_DEBUG_CACHE,
    fprintf(DM_FindErrorStream(context),
            "Referred to cache " DM_0X "%p[%lu]\n",
            cache, (unsigned long)index );
);

  if ( index < cache->item_length ) {
    return cache->item[index];
  } else {
    DM_Abort(context, DME_OUT_OF_CACHE);
    return DM_NULL;
  }
}

/* RESOURCE DM_is_in_cache
 * Look whether an object is in the cache and returns its index.
 * If not in cache return an index out of the cache.
 */

DM_nat3
DM_is_in_cache (DM_Context       *context,
                DM_Object        *o, 
                DM_MutableVector *cache ) {
  DM_nat3 i;

  /* Look in the cache. 
   * Should use a hashtable but what if objects move in memory ? */
  for ( i=0 ; i<cache->item_length ; i++ ) {
    DM_Object *seen = cache->item[i];

    if ( o == seen ) {

      DM_DebugAtLevel(DM_DEBUG_CACHE,
        fprintf(DM_FindErrorStream(context), 
                "Found " DM_0X "%p in cache " DM_0X "%p[%lu]\n", 
                o, cache, (unsigned long)i );
);
    
      return i;
    }
  }
  return i;
}

/* RESOURCE DM_is_in_stack
 * Look whether an object is in the stack and returns its index.
 * If not in stack return an index out of the stack.
 */

DM_nat3
DM_is_in_stack (DM_Context      *context,
                DM_Object       *o, 
                DM_BoundedStack *stack ) {
  DM_nat3 i;

  /* Look in the stack. 
   * Should use a hashtable but what if objects move in memory ? */
  for ( i=0 ; i<stack->index ; i++ ) {
    DM_Object *seen;
      
    seen = stack->item[i];
    if ( o == seen ) {

      DM_DebugAtLevel(DM_DEBUG_CACHE,
        fprintf(DM_FindErrorStream(context), 
                "Found " DM_0X "%p in stack " DM_0X "%p[%lu]\n", 
                o, stack, (unsigned long)i );
);
    
      return i;
    }
  }
  return i;
}

/* }}}
 * {{{ Initialize the module.
 */

void
DM_cache_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmcache);
}

/* }}}
 *  end of dmcache.c */
