/* $Id: dmcomm.c,v 1.123 1998/07/01 10:04:11 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* This file defines the basic layer for migrating DMeroon objects.
 * It defines the Entry and Exit objects, Clocks and HashTables to
 * manage them.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmcomm,
  "$Id: dmcomm.c,v 1.123 1998/07/01 10:04:11 queinnec Exp $");

/* {{{ Static data
 *
 * Statically allocate the main clock of the site.  The hashtables
 * (for entries and exits) are not statically allocated to let them
 * grow, they are dynamically allocated in DM_comm_initialize() since
 * big static tables cause troubles to some linkers.
 */

DM_Clock_wrapped DM_general_clock_object = {
  DM_SupportPrefixInitialization(DM_Clock_wrapped,
                                 DM_general_clock_object)
  DM_ReferenceToClass(Clock),
  DM_ProxyInitialization(),
  { { {
    0,   /* time    */
    0,   /* clocker */
  } } }
};


/* }}}
 * {{{ Reference counter management
 * 
 * Reference counters are held in volatile fields ie fields which are
 * not mutable by the user but whose content may vary. We therefore
 * provide C functions to modify them since it is not possible to
 * mutate them with regular high-level DMeroon primitives. Moreover,
 * when the reference-count reaches zero, the entry is removed from
 * the Entries table.
 */

void
DM_increment_counter (DM_Context  *context,
                      DM_External *external ) {
  external->counter++;
}

void
DM_decrement_counter (DM_Context *context,
                      DM_Entry   *entry ) {
  entry->counter--;

  DM_DebugAtLevel(0,
      fprintf(DM_FindErrorStream(context), "New refcount is %lu for "
              DM_0X "%p\n", (unsigned long)(entry->counter), entry );
);

  if ( entry->counter <= 0 ) {
    DM_remove_entry(context, (DM_External*)entry);
    DM_ClearOption(entry, DM_EXTERNAL_IS_EXTERNALIZED);
  }
}

/* }}}
 * {{{ Time management

 * RESOURCE DM_propagate_information
 * Propagate through a Connection the clocks that evolved. The information
 * to propagate are those that have a 'clocker' field greater than the
 * 'clocker' field of the Connection. Also propagate the Sites that become
 * known (this will ease future references).
 */

void
DM_propagate_information (DM_Context   *context,
                          DM_OutBuffer *outbuffer) {
  DM_Site      *site      = DM_TheCurrentSite();
  DM_Site      *othersite = DM_NULL;
  DM_Hashtable *entries   = site->entries;
  DM_Hashtable *exits     = site->exits;
  DM_nat3      i;

  if ( DM_IsA(context, outbuffer->port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, outbuffer->port);

    /* Send decrement messages for the useless Exit items coming from
     * othersite and mark them useless not to resend decrement
     * messages. These useless Exit items are linked in the 'freed'
     * field of the othersite. */
    if ( DM_IsA(context, connection->site, Site) ) {
      othersite = connection->site;
      while ( othersite->freed ) {
        DM_Exit *exit = othersite->freed;

        DM_DebugAtLevel(5,
          fprintf(DM_FindErrorStream(context), 
                  "Sending decrement message for exit " 
                  DM_0X "%p\n", exit );
);

        othersite->freed = exit->next;
        /* Help GC, remove a pointer. If recollected, this exit will have
         * the DM_EXIT_DECREMENT_SENT and DM_EXIT_IS_CHAINED options set.
         * Moreover there is no longer a finalizer waiting for it. */
        exit->next = DM_NULL;
        DM_print_byte(context, DM_DECREMENT_COMMAND, outbuffer);
        DM_send_key(context, outbuffer, &(exit->number));
        DM_SetOption(exit, DM_EXIT_DECREMENT_SENT);
      }
    }

    /* Examine all present externalized clocks and propagate the ones
     * that changed since the last time a message was sent to
     * othersite. For that we scan the Entries hash-table looking for
     * Clocks. */
    for ( i=0 ; i<entries->item_length ; i++ ) {
      /* The Entry may be NULL but DM_IsA takes care of that. */ 
      DM_Entry *entry = (void*)entries->item[i];

      if ( DM_IsA(context, entry, Exit) ) {
        /* This is a reexported Exit item. They are scanned below. */
        continue;
      } else if ( DM_IsA(context, entry, Entry) ) {
        if ( DM_IsA(context, entry->object, Clock) ) {
          DM_DeclareNarrowedObject(clock, Clock, entry->object);
          
          if ( clock->clocker > connection->clocker ) {
            /* Send this clock, it evolved. */

            DM_DebugAtLevel(5,
              fprintf(DM_FindErrorStream(context), 
                      "Sending clock " 
                      DM_0X "%p\n", clock );
);

            DM_print_byte(context, DM_UPDATE_CLOCK_COMMAND, outbuffer);
            DM_write_object_reference(context, (DM_Object*)clock, outbuffer);
            DM_send_nat4(context, outbuffer, &(clock->time));
          } else {
            /* The clock has not changed since previous message. */
          }
        } else {
          /* Not a Clock: ignore it. */
        }
      } else {
        /* This is not an Entry ie it's an empty position of the Hashtable. */
      }
    }

#if 0                     /* USELESS */
    /* Examine all known remote Sites. Sites must be propagated before
     * remote clocks to avoid sending a Clock without its owning
     * Site. */
    {
      DM_Hashtable *sites     = site->sites;

      for ( i=0 ; i<sites->item_length ; i++ ) {
        /* The Site may be NULL. `*/
        DM_Site *a_site = (DM_Site*)(sites->item[i]);

        if (    DM_IsA(context, a_site, Site) 
                && ( a_site != DM_TheCurrentSite() )
                && ( a_site != othersite ) ) {
          /* The current site is already transmitted, skip it. Also note
           * that its 'clocker' fields is regularly incremented. */
          if ( a_site->clocker > connection->clocker ) {
            DM_Entry *e = DM_externalize(context, (DM_Object*)a_site);

            if ( DM_IsA(context, e, Exit) ) {
              DM_DeclareNarrowedObject(exit, Exit, e);
            
              DM_DebugAtLevel(5,
                fprintf(DM_FindErrorStream(context), 
                        "Propagated Site " DM_0X "%p\n", 
                        a_site);
);

              /* FUTURE: should only send usable class ?????????? */
              DM_print_byte(context, DM_PROG2_COMMAND, outbuffer);
              DM_print_byte(context, DM_DEFINE_REMOTE_SITE_COMMAND, outbuffer);
              DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
              /* the following sends a SITE ip port date length command. */
              DM_write_object(context, (DM_Object*)a_site, outbuffer);
              DM_write_object_content(context, (DM_Object*)a_site, outbuffer);
              DM_send_key(context, outbuffer, &(exit->number));
            } else {
              DM_Abort(context, DME_NOT_AN_EXIT);
            }
          } else {
            /* The Site was already propagated. */
          }
        } else {
          /* Not a Site: ignore it. */
        }
      }
    }
#endif /* 0                           USELESS */

    /* Examine now remote clocks (they may be reexported or not).
    * Propagate the ones that evolved since last message to
    * othersite. */
    { 
      DM_DeclareExtractorFunction(extractor, exits->extractor);

      /* Examine all known remote Clocks. */
      for ( i=0 ; i<exits->item_length ; i++ ) {
        /* The Exit may be NULL. */
        DM_Exit *exit;

        if ( extractor ) {
          exit = (*extractor)(context, &(exits->item[i]));
        } else {
          exit = (void*)(exits->item[i]);
        }
        if ( DM_IsA(context, exit, Exit) ) {
          DM_ShortCutReference(context,  &(exit->object));
          if ( DM_IsA(context, exit->object, Clock) ) {
            DM_DeclareNarrowedObject(clock, Clock, exit->object);

            if ( clock->clocker > connection->clocker ) {
              /* Send this clock. */
              DM_print_byte(context, DM_UPDATE_CLOCK_COMMAND, outbuffer);
              DM_write_object_reference(context, (DM_Object*)clock, outbuffer);
              DM_send_nat4(context, outbuffer, &(clock->time));

              DM_DebugAtLevel(0,
                fprintf(DM_FindErrorStream(context), 
                        "Propagated remote Clock " 
                        DM_0X "%p\n", clock );
);

            } else {
              /* The clock does not change since previous message. */
            }
          } else {
            /* Not a Clock: ignore it. */
          }
        } else {
          /* This is not an Exit ie it's an empty position of the Hashtable. */
        }
      }
    }

    /* Since I sent all Clocks with a clocker greater than the clocker
     * of the Connection, I can update it to the one of the site. */
    connection->clocker = site->clocker;
  } else {
    /* This is not a Connection, don't propagate any information. */
  }
}

/* }}}
 * {{{ Managing Entry items
 *
 * DOCUMENTATION Entry items
 * Entry items are kept in an Hashtable. They are hashed with the number
 * under which they are externally known (and this number cannot be 0).
 */

/* RESOURCE DM_semi_externalize
 * This function associates an object to an Entry item (if not yet
 * present).  The object may be uninitialized (the Entry is marked
 * DM_ENTRY_IS_UNINITIALIZED). If an Entry is created, it is not
 * registered so its key is zero as well as its reference counter.
 */

DM_Entry*
DM_semi_externalize (DM_Context *context,
                     DM_Object  *o) 
{
  DM_Object *proxy = DM_ObjectToProxy(context, o);

  if ( DM_IsA(context, proxy, Entry) ) {
    /* The object has already an Entry (maybe an Exit) associated to it. */
    DM_DeclareNarrowedObject(entry, Entry, proxy);

    return entry;
  } else {
    /* Here the object is local, not semi-externalized but may have
     * been specified as copyable or sharable or may even be still
     * uninitialized. 

     * NOTE: The Entry item is left in the uninitialized state.
     */
    DM_Entry *entry = DM_Allocate0(context, Entry);

    DM_SetProxyOf(context, o, entry);
    entry->counter = 0;
    entry->options = DM_EXTERNAL_DEFAULT_OPTIONS;
    entry->object  = o;
    entry->clock   = DM_NULL;

    if ( proxy == DM_TheConstant(uninitialized) ) {
      /* The object is not initialized. */
      DM_SetOption(entry, DM_ENTRY_IS_UNINITIALIZED);
    } else {   
      /* The object is initialized. */
      DM_ClearOption(entry, DM_ENTRY_IS_UNINITIALIZED);
      if ( proxy == DM_TheConstant(true) ) {
        DM_SetOption(entry, DM_ENTRY_IS_COPYABLE);
      } else {
        /* Non copyable means sharable. */
        DM_ClearOption(entry, DM_ENTRY_IS_COPYABLE);
      }
    }
    /* Always (at least) semi-externalize the class. */
    { DM_Class *class = DM_ObjectToClass(context, o);

      (void) DM_semi_externalize(context, (DM_Object*)class);
    }
    return entry;
  }
}

/* RESOURCE DM_externalize
 * This function makes an object reachable from the network. It
 * associates it with a registered Entry item. If the object is
 * remote, it may also be externalized that is, the cache may be
 * reexported. Even an uninitialized object may be exported.
 */

DM_Entry*
DM_externalize (DM_Context *context,
                DM_Object  *o) 
{
  DM_Object *proxy = DM_ObjectToProxy(context, o);
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Entry  *entry;

  if ( DM_IsA(context, proxy, Entry) ) {
    /* The object has already an Entry (maybe an Exit) associated to it. */
    entry = (DM_Entry*) proxy;
  } else {
    entry = DM_semi_externalize(context, o);
    if (    DM_HasNotOption(entry, DM_ENTRY_IS_COPYABLE)
            && DM_HasOption(class, DM_CLASS_IS_MUTABLE) ) {
      /* Since the object is mutable and sharable (even if
       * uninitialized), there must be a monitoring clock. If there is
       * no monitoring clock then give it the general clock of the site.
       */
      if ( entry->clock ) {
        /* NOTHING: There is already a monitoring clock.  */
      } else {
        entry->clock = DM_TheCurrentSite()->clock;
      }
    } else {
      /* no need for a clock (the object is immutable or copyable).  */
      entry->clock = DM_NULL;
    }
  }

  /* Register the Entry. */
  if ( DM_HasNotOption(entry, DM_EXTERNAL_IS_EXTERNALIZED) ) {
    /* The Entry was not yet registered, register it! */
    DM_register_entry(context, (DM_External*)entry);
    /* Always externalize the class. */
    (void) DM_externalize(context, (DM_Object*)class);
  }
  return entry;
}

/* Hash Entries only on the low part of their external key. */

#define DM_ComputeHashEntry(number) \
  (number._idx)

static DM_nat3
DM_compute_hash_entry (DM_Context *context,
                       DM_Entry   *entry) {
  return DM_ComputeHashEntry(entry->number);
}

/* RESOURCE DM_extract_entry
 * This function extracts an Entry instance from the Entries hashtable.
 * An Entry is found after its external number (a nat4). If no Entry is
 * found, return DM_FALSE. 

 FUTURE:  Use DM_key* instead of DM_key for speed.                 FUTURE
 */

DM_Entry*
DM_extract_entry (DM_Context *context,
                  DM_key     num) {
  DM_HashtableExtract(context, 
                      DM_TheCurrentSite()->entries,
                      DM_Entry*,
                      DM_ComputeHashEntry(num),
                      DM_EqualKey(object->number, num)
                      );
}

/* RESOURCE DM_remove_entry
 * Remove an entry from the entries hashtable.
 */

void
DM_remove_entry (DM_Context *context,
                 DM_External *external) {
  DM_Site *site = DM_TheCurrentSite();
  
  DM_hashtable_remove(context, &(site->entries), (DM_Object*)external);  
  DM_ClearOption(external, DM_EXTERNAL_IS_EXTERNALIZED);
}

/* RESOURCE DM_compute_external_key

 * Compute a DM_key number suitable for externalization. This number
 * is computed with some randomness for security reason. The key is
 * made of the ip and port number of the host, the current date and a
 * random counter (this prevents forgeing legal keys). There is a 
 * slight possibility of collision if allocating more than 2^16 objects
 * per second and two of them get the same 4bytes random number. This
 * is normally avoided by the periodicity of the random generator. 
 */

static DM_nat2 DM_key_counter = 0;

static void
DM_compute_key (DM_Context *context,
                DM_key     *key ) {
  key->_date = DM_get_date(context);
  key->_ip   = DM_TheCurrentSite()->address;
  key->_port = DM_TheCurrentSite()->port;   /* in network order */
  key->_rand = random();
  key->_idx  = DM_key_counter++;
}

/* RESOURCE DM_register_entry
 * Register an External object in the Entries hashtable. An external
 * number is computed at random under which it will be externally 
 * known. This external number cannot be null so an object is known
 * to be externalized if the number of its Entry is not null. The
 * function also checks that the external number is unique on the
 * current site.
 */

void
DM_register_entry (DM_Context  *context,
                   DM_External *external) {
  DM_Site *site = DM_TheCurrentSite();

  DM_DebugAtLevel(5,
    if ( ! DM_IsA(context, external, External) ) {
      DM_Abort(context, DME_INTERNAL_ERROR);
    }
);

  if ( DM_HasNotOption(external, DM_EXTERNAL_IS_EXTERNALIZED) ) {
    /* The External object is not registered. */
    DM_SetOption(external, DM_EXTERNAL_IS_EXTERNALIZED);
    external->counter = 0;
    if ( DM_IsA(context, external, Exit) ) {
      /* NOTHING: An Exit item is reexported under the key it already
       * has.  Therefore an object keeps a constant key everywhere
       * even if it is migrated. */
    } else {
      DM_compute_key(context, &external->number);
    }

    DM_DebugAtLevel(13,
      fprintf(DM_FindErrorStream(context), "Registered External ");
      DM_key_print(context, &(external->number), DM_FindErrorStream(context));
      fprintf(DM_FindErrorStream(context), "\n");
);

    DM_hashtable_insert(context, &(site->entries), (DM_Object*)external);
  } else {
    /* The object is already externalized. */
    
    DM_DebugAtLevel(13,
      fprintf(DM_FindErrorStream(context), 
              "Already registered Entry " DM_0X "%p\n",
              external );
);

  }
}

/* }}}
 * {{{ Managing Exit items.
 *
 * Exit items are kept in the Exits Hashtable. They are simply hashed
 * for now to ease debug.
 */

#define DM_ComputeHashExit(key) \
  ((DM_nat3)(key._idx))

static DM_nat3
DM_compute_hash_exit (DM_Context *context,
                      DM_Exit    *exit) {
  return DM_ComputeHashExit(exit->number);
}

/* RESOURCE DM_remove_exit
 * Remove an exit from the exits hashtable.
 * NOTE: to remove an Exit from the entries hashtable, use DM_remove_entry
 * but at that time, the Exit cannot be reclaimed since it is useful.
 * Any time an Exit is recycled, send a decrement message to the remotely
 * referenced Entry item.
 */

void
DM_remove_exit (DM_Context *context,
                DM_Exit    *exit) {
  DM_Site *site = DM_TheCurrentSite();
  
  DM_DebugAtLevel(5,
    fprintf(DM_FindErrorStream(context), 
            "Useless Exit item " DM_0X "%p, registered as ", exit);
    DM_key_print(context, &(exit->number), DM_FindErrorStream(context));
    fprintf(DM_FindErrorStream(context), "\n");
);

  if ( DM_HasOption(exit, DM_EXIT_IS_CHAINED) ) {
    /* NOTHING: The Exit item has already been taken care of. */
  } else if ( DM_HasNotOption(exit, DM_EXIT_DECREMENT_SENT) ) {
    /* Chain the Exit in the list of Exit items of the from site (not
     * the owner) to be freed.  Since this is detected by the GC, don't
     * allocate (that is use the 'freed' field)! */

    DM_DebugAtLevel(2,
      fprintf(DM_FindErrorStream(context), 
              "GC: freeing " DM_0X "%p\n", exit);
);

    DM_SetOption(exit, DM_EXIT_IS_CHAINED);
    exit->next = exit->from->freed;
    exit->from->freed = exit;
    DM_hashtable_remove(context, &(site->exits), (DM_Object*)exit); 
  } else {
    /* NOTHING: The Exit has already been taken care of that is, a
     * decrement message is in transit (or the exit was not registered
     * as in DM_DEFINE_EMITTER_COMMAND). */
  }
}

/* RESOURCE DM_extract_exit
 * This function extracts an Exit from an hashtable. The Exit item is 
 * found with its external number (host-ordered) and the IP address of
 * its owner (host-ordered). If no such Exit is found, return DM_FALSE.
 */

DM_Exit*
DM_extract_exit (DM_Context *context,
                 DM_key     num ) {
  DM_HashtableExtract(context, 
                      DM_TheCurrentSite()->exits,
                      DM_Exit*,
                      DM_ComputeHashExit(num),
                      DM_EqualKey(object->number, num) 
                      );
}

/* RESOURCE DM_register_exit
 * Create and register an Exit item. If a similar one already exists,
 * return it. The similarity is checked by comparing the owner and the
 * external number on that owner. This function always return an Exit 
 * item (or an Entry item if the object is local).
 *
 * Any creation of an Exit item is followed by a call to this function
 * to register this Exit item in the Exits table (and to associate a
 * finalizer with it).
 */

DM_Object*
DM_register_exit (DM_Context *context,
                  DM_Exit    *exit) {
  DM_Site *site      = DM_TheCurrentSite();
  DM_Site *othersite = exit->owner;

  if ( othersite == site ) {
    /* The object is in fact local so we return an Entry instead and
       we do not register the Exit at all. The Entry should exist of
       course.  */
    DM_Entry *entry = DM_extract_entry(context, exit->number);
    
    if ( entry ) {
      /* Attention: this is a local object so we return an Entry and 
       * not an Exit.  */
      return (DM_Object*)entry;
    } else {
      DM_Abort(context, DME_LOST_ENTRY);
      /* NOT REACHED */
      return DM_NULL;
    }
  } else {
    /* The object is remote. */
    DM_Exit *e = DM_extract_exit(context, exit->number);

    if ( e ) {

      DM_DebugAtLevel(7,
        fprintf(DM_FindErrorStream(context), 
                "Received Exit " DM_0X "%p similar to Exit " DM_0X "%p\n",
                exit, e );
);
      
      return (DM_Object*)e;
    } else {
      /* Notify the GC we are interested in finalizing that object. */
      DM_set_exit_finalization(exit);
      /* Register the Exit item. */
      DM_hashtable_insert(context, &(site->exits), (DM_Object*)exit);

    DM_DebugAtLevel(13,
      fprintf(DM_FindErrorStream(context), "Registered Exit ");
      DM_key_print(context, &(exit->number), DM_FindErrorStream(context));
      fprintf(DM_FindErrorStream(context), "\n");
);

      return (DM_Object*)exit;
    }
  }
}

/* }}}
 * {{{ Managing Sites
 
 * DOCUMENTATION Sites hashtable
 * Sites are hashed with their IP number and port numbers. A precise
 * instance of Site is also qualified by a date (not participating to
 * the hashing) indicating when it was started. This allows to
 * differentiate two different servers acting sequentially on the same
 * machine and same port.  
 *
 * In fact, this hashtable contains Sites. When we hear of some
 * site just by ip,port,date we create a Site to represent
 * it. 
 */

#define DM_ComputeHashSite(ip, port) \
   ((DM_nat3)((ip)+(port)))

static DM_nat3
DM_compute_hash_site (DM_Context *context,
                      DM_Site *site) {
  return DM_ComputeHashSite(ntohl(site->address), 
                            ntohs(site->port));
}

/* RESOURCE DM_register_site
 * Register a Site in the global hashtable of sites. 
 * Assume the site is not already present in this hashtable.
 */

void
DM_register_site (DM_Context *context,
                  DM_Site    *site) 
{
  DM_Site *thesite = DM_TheCurrentSite();

  DM_DebugAtLevel(7,
    fprintf(DM_FindErrorStream(context), "Ah! a new site " DM_0X "%p\n", site);
);

  thesite->clocker++;
  site->clocker = thesite->clocker;
  DM_hashtable_insert(context, &(thesite->sites), (DM_Object*)site);
}

/* RESOURCE DM_remove_site
 * Remove a site from the global hashtable of sites. This is done when
 * the site is obsolete ie we know of a younger site with the same 
 * address and port.
 */

void
DM_remove_site (DM_Context *context,
                DM_Site *site) {
  DM_Site *thesite = DM_TheCurrentSite();
  
  DM_hashtable_remove(context, &(thesite->sites), (DM_Object*)site);
}  

/* RESOURCE DM_extract_site
 * Extract a Site with its ip and port number (both in host-order).
 * The comparison with the starting time should be done by the caller.
 * 
 * NOTE: Normally no error is possible (even if the hashtable is full).
 */

DM_Site*
DM_extract_site (DM_Context *context,
                 DM_nat3 ip,
                 DM_nat2 port ) {
  DM_Site *site = DM_TheCurrentSite();

  DM_HashtableExtract(
      context, 
      site->sites,
      DM_Site*,
      DM_ComputeHashSite(ip, port),
      (    ( (ip) == htonl(object->address) )
        && ( (port) == htons(object->port) ) )
                      );
}

/* }}} 
 * {{{ Reifying values

 * RESOURCE DM_reify_value
 * convert a value of a DMeroon type into a DMeroon object (with a single
 * field with that value). This is the basic function above which, several
 * specialized utilites exist.
 *
 * NOTE: It's very expensive but I use it for the mutation protocol to
 * lessen the number of specialized classes (subclasses of
 * ObjectMutationRequest) to create.
 */

DM_Value*
DM_reify_value (DM_Context *context,
                DM_DMeroonType *type,
                void           *p ) {
  DM_Value *result = (DM_Value*) DM_allocate0(context, type->dmeroon_class);

  memcpy(result, p, type->size);
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, result);
  }
  return result;
}

/* RESOURCE DM_reify_value_from_field
 * Convert a value into a DMeroon object. The value is specified with an
 * object and a regular field.
 */

DM_Value*
DM_reify_value_from_field (DM_Context *context,
                           DM_Object *o,
                           DM_Field  *field ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3 offset = field->offset[field->offset_length-1];
  DM_DMeroonType    *type = 
    implemented_types->item[DM_OffsetToItemTypeIndex(offset)];
  DM_displacement   t;

  if ( DM_check_access(context, o, field, DM_DEFAULT_ACCESS) ) {
    DM_Abort(context, DME_BAD_ACCESS);
    return DM_NULL;
  } else {
    t = DM_compute_offset(context, o, field);
    return DM_reify_value(context, type, DM_Displace(o, t));
  } 
}

/* RESOURCE DM_reify_value_from_indexed_field
 * Convert a value into a DMeroon object. The value is specified with an
 * object, an indexed field and an index.
 */

DM_Value*
DM_reify_value_from_indexed_field (DM_Context *context,
                                   DM_Object *o,
                                   DM_Field  *field,
                                   DM_nat3   index ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3 offset = field->offset[field->offset_length-1];
  DM_DMeroonType    *type = 
    implemented_types->item[DM_OffsetToItemTypeIndex(offset)];
  DM_displacement    t;

  if ( DM_check_access(context, o, field, DM_INDEXED_ACCESS) ) {
    DM_Abort(context, DME_BAD_ACCESS);
    return DM_NULL;
  } else {
    t = DM_indexed_compute_offset(context, o, field, index);
    return DM_reify_value(context, type, DM_Displace(o, t));
  }
}

/* RESOURCE DM_patch_from_reified_value
 * Patch o.f with the reified value. Return the reified previous value.
 */

DM_Value*
DM_patch_from_reified_value (DM_Context *context,
                             DM_Object *o,
                             DM_Field  *field,
                             DM_Value  *value ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3 offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *type = 
    implemented_types->item[DM_OffsetToItemTypeIndex(offset)];
  DM_displacement    t;
  DM_Value           *result;

  if ( DM_check_access(context, o, field, DM_DEFAULT_ACCESS) ) {
    DM_Abort(context, DME_BAD_PATCH);
    return DM_NULL;
  } else {
    t = DM_compute_offset(context, o, field);
    result = DM_reify_value(context, type, DM_Displace(o, t));
    memcpy(DM_Displace(o, t), value, type->size);
    DM_increment_clock_of(context, o);
    return result;
  }
}

/* RESOURCE DM_patch_from_reified_value
 * Patch o.f[i] with the reified value. Return the reified previous value.
 */

DM_Value*
DM_indexed_patch_from_reified_value (DM_Context *context,
                                     DM_Object *o,
                                     DM_Field  *field,
                                     DM_nat3   index,
                                     DM_Value  *value ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3 offset = field->offset[field->offset_length-1];
  DM_DMeroonType    *type = 
    implemented_types->item[DM_OffsetToItemTypeIndex(offset)];
  DM_displacement   t;
  DM_Value          *result;

  if ( DM_check_access(context, o, field, DM_INDEXED_ACCESS) ) {
    DM_Abort(context, DME_BAD_PATCH);
    return DM_NULL;
  } else {
    t = DM_indexed_compute_offset(context, o, field, index);
    result = DM_reify_value(context, type, DM_Displace(o, t));
    memcpy(DM_Displace(o, t), value, type->size);
    DM_increment_clock_of(context, o);
    return result;
  }
}

/* }}}
 */

/* {{{ Basic encoders//decoders
 * Now a big bunch of methods for encoding and decoding. 
 *
 * NOTE: To ensure that the encoding is always done in the last
 * OutBuffer, all encoders should start with the SendThing macro. This
 * macro begins by setting the emission OutBuffer to its right value
 * before encoding.  The other way to print is to use
 * DM_print_char_in_outbuffer which performs the same check.
 */

void
DM_send_nought (DM_Context *context,
                DM_OutBuffer *emission, 
                void *p) {
  DM_Abort(context, DME_UNSENDABLE_TYPE);
}

void
DM_receive_nought (DM_Context *context,
                   DM_InBuffer *reception,
                   void *p) {
  DM_Abort(context, DME_UNRECEIVABLE_TYPE);
}

void
DM_send_nat1 (DM_Context *context,
              DM_OutBuffer *emission,
              void *p) {
  DM_nat1       byte        = DM_Content(DM_nat1,p);
  
  SendThing(context, emission, 1,
    DM_SendOneByte(context, emission, byte)
  )
}

void
DM_receive_nat1 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  ReceiveThing(context, reception, 1,
    DM_Content(DM_nat1,p) = DM_ReceiveOneByte(context, reception)
  )
}

void
DM_send_nat2 (DM_Context *context,
              DM_OutBuffer *emission,
              void *p) {
  DM_nat2       natural     = DM_Content(DM_nat2,p);
  DM_nat1       byte;

  SendThing(context, emission, 2,
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  natural = (natural - byte)/256;             /* FUTURE - is useless  */
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  )
}

void
DM_receive_nat2 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_nat2       result      = 0;
  DM_nat2       scale       = 256;

  ReceiveThing(context, reception, 2,
  result = DM_ReceiveOneByte(context, reception);
  result += scale*DM_ReceiveOneByte(context, reception);
  DM_Content(DM_nat2,p) = result;
  )
}

void
DM_send_nat3 (DM_Context *context,
              DM_OutBuffer *emission,
              void *p) {
  DM_nat3       natural     = DM_Content(DM_nat3,p);
  DM_nat1       byte;
  unsigned int  i;

  SendThing(context, emission, 4,
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  for ( i=3 ; i>0 ; i-- ) {
    natural = (natural - byte)/256;
    byte = (DM_nat1)natural;
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_nat3 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_nat3       result      = 0;
  DM_nat3       scale       = 256;
  unsigned int  i;

  ReceiveThing(context, reception, 4,
  result = DM_ReceiveOneByte(context, reception);
  for ( i=3 ; i>0 ; i-- ) {
    result += scale*DM_ReceiveOneByte(context, reception);
    scale *= 256;
  }
  DM_Content(DM_nat3,p) = result;
  )
}

void
DM_send_nat4 (DM_Context *context,
              DM_OutBuffer *emission, 
              void *p) {
  DM_nat4       natural     = DM_Content(DM_nat4,p);
  DM_nat1       byte;
  unsigned int  i;

  SendThing(context, emission, 8,
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  for ( i=7 ; i>0 ; i-- ) {
    natural = (natural - byte)/256;
    byte = (DM_nat1)natural;
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_nat4 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_nat4       result      = 0;
  DM_nat4       scale       = 256;
  unsigned int  i;
  
  ReceiveThing(context, reception, 8,
  result = DM_ReceiveOneByte(context, reception);
  for ( i=7 ; i>0 ; i-- ) {
    result += scale*DM_ReceiveOneByte(context, reception);
    scale *= 256;
  }
  DM_Content(DM_nat4,p) = result;
  )
}

void
DM_send_key (DM_Context   *context,
             DM_OutBuffer *emission, 
             void         *p) {
  DM_key key = DM_Content(DM_key,p);

  SendThing(context, emission, 16,
    DM_send_nat2(context, emission, &(key._idx));
    DM_send_nat2(context, emission, &(key._port));
    DM_send_nat3(context, emission, &(key._ip));
    DM_send_nat3(context, emission, &(key._date));
    DM_send_nat3(context, emission, &(key._rand));
  )
}

void
DM_receive_key (DM_Context  *context,
                DM_InBuffer *reception,
                void        *p) {
  DM_key key;
  
  ReceiveThing(context, reception, 16,
    DM_receive_nat2(context, reception, &(key._idx));
    DM_receive_nat2(context, reception, &(key._port));
    DM_receive_nat3(context, reception, &(key._ip));
    DM_receive_nat3(context, reception, &(key._date));
    DM_receive_nat3(context, reception, &(key._rand));
    DM_Content(DM_key,p) = key;
  )
}

/*
 * Integers are encoded by their absolute value prefixed by one bit
 * encoding the sign (0 if positive).
 */

#define EncodeNegativeByte(byte) ( (byte) | (1<<7) )
#define IsNegativeByte(byte) ( (byte)&(1<<7) )
#define DecodeNegativeByte(byte) ( (byte)&((1<<7)-1) )

void
DM_send_int1 (DM_Context *context,
              DM_OutBuffer *emission,
              void *p) {
  DM_int1       integer     = DM_Content(DM_int1,p);

  SendThing(context, emission, 1,
  if ( integer>=0 ) {
    DM_SendOneByte(context, emission,integer);
  } else {
    DM_SendOneByte(context, emission,EncodeNegativeByte(-integer));
  }
  )
}

void
DM_receive_int1 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  long int      tmp;

  ReceiveThing(context, reception, 1,
  tmp = DM_ReceiveOneByte(context, reception);
  if ( IsNegativeByte(tmp) ) {
    DM_Content(DM_int1,p) = - DecodeNegativeByte(tmp);
  } else {
    DM_Content(DM_int1,p) = tmp;
  }
  )
}

void
DM_send_int2 (DM_Context *context,
              DM_OutBuffer *emission, 
              void *p) {
  DM_int2       integer     = DM_Content(DM_int2,p);
  DM_nat2       natural     = integer;
  DM_nat1       negative    = 0;
  DM_nat1       byte;
  
  SendThing(context, emission, 2,
  if ( integer<0 ) { negative = 1; natural = -integer; }
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  natural = (natural - byte)/256;
  byte = (DM_nat1)natural;
  if ( negative ) {
    DM_SendOneByte(context, emission,EncodeNegativeByte(byte));
  } else {
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_int2 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_int2       result      = 0;
  DM_int2       scale       = 256;
  DM_nat1       byte;
  
  ReceiveThing(context, reception, 2,
  result = DM_ReceiveOneByte(context, reception);
  byte = DM_ReceiveOneByte(context, reception);
  if ( IsNegativeByte(byte) ) {
    result += scale*DecodeNegativeByte(byte);
    result = -result;
  } else {
    result += scale*byte;
  }
  DM_Content(DM_nat2,p) = result;
  )
}

void
DM_send_int3 (DM_Context *context,
              DM_OutBuffer *emission, 
              void *p) {
  DM_int3       integer     = DM_Content(DM_int3,p);
  DM_nat3       natural     = integer;
  DM_nat1       negative    = 0;
  DM_nat1       byte;
  unsigned int  i;

  SendThing(context, emission, 4,
  if ( integer<0 ) { negative = 1; natural = -integer; }
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  for ( i=2 ; i>0 ; i-- ) {
    natural = (natural - byte)/256;
    byte = (DM_nat1)natural;
    DM_SendOneByte(context, emission,byte);
  }
  natural = (natural - byte)/256;
  byte = (DM_nat1)natural;
  if ( negative ) {
    DM_SendOneByte(context, emission,EncodeNegativeByte(byte));
  } else {
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_int3 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_int3       result      = 0;
  DM_int3       scale       = 256;
  DM_nat1       byte;
  unsigned int  i;
  
  ReceiveThing(context, reception, 4,
  result = DM_ReceiveOneByte(context, reception);
  for ( i=2 ; i>0 ; i-- ) {
    result += scale*DM_ReceiveOneByte(context, reception);
    scale *= 256;
  }
  byte = DM_ReceiveOneByte(context, reception);
  if ( IsNegativeByte(byte) ) {
    result += scale*DecodeNegativeByte(byte);
    result = -result;
  } else {
    result += scale*byte;
  }
  DM_Content(DM_int3,p) = result;
  )
}

void
DM_send_int4 (DM_Context *context,
              DM_OutBuffer *emission, 
              void *p) {
  DM_int4       integer     = DM_Content(DM_int4,p);
  DM_nat4       natural     = integer;
  DM_nat1       negative    = 0;
  DM_nat1       byte;
  unsigned int  i;

  SendThing(context, emission, 8,
  if ( integer<0 ) { negative = 1; natural = -integer; }
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  for ( i=6 ; i>0 ; i-- ) {
    natural = (natural - byte)/256;
    byte = (DM_nat1)natural;
    DM_SendOneByte(context, emission,byte);
  }
  natural = (natural - byte)/256;
  byte = (DM_nat1)natural;
  if ( negative ) {
    DM_SendOneByte(context, emission,EncodeNegativeByte(byte));
  } else {
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_int4 (DM_Context *context,
                 DM_InBuffer *reception,
                 void *p) {
  DM_int4       result      = 0;
  DM_int4       scale       = 256;
  DM_nat1       byte;
  unsigned int  i;
  
  ReceiveThing(context, reception, 8,
  result = DM_ReceiveOneByte(context, reception);
  for ( i=6 ; i>0 ; i-- ) {
    result += scale*DM_ReceiveOneByte(context, reception);
    scale *= 256;
  }
  byte = DM_ReceiveOneByte(context, reception);
  if ( IsNegativeByte(byte) ) {
    result += scale*DecodeNegativeByte(byte);
    result = -result;
  } else {
    result += scale*byte;
  }
  DM_Content(DM_int4,p) = result;
  )
}

void
DM_send_netnat2 (DM_Context *context,
                 DM_OutBuffer *emission,
                 void *p) {
  DM_nat2       natural     = ntohs(DM_Content(DM_netnat2,p));
  DM_nat1       byte;
  
  SendThing(context, emission, 2,
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  natural = (natural - byte)/256;
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  )
}

void
DM_receive_netnat2 (DM_Context *context,
                    DM_InBuffer *reception,
                    void *p) {
  DM_nat2       result      = 0;
  DM_nat2       scale       = 256;
  
  ReceiveThing(context, reception, 2,
  result = DM_ReceiveOneByte(context, reception);
  result += scale*DM_ReceiveOneByte(context, reception);
  DM_Content(DM_netnat2,p) = htons(result);
  )
}

void
DM_send_netnat3 (DM_Context *context,
                 DM_OutBuffer *emission, 
                 void *p) {
  DM_nat3       natural     = ntohl(DM_Content(DM_netnat3,p));
  DM_nat1       byte;
  unsigned int  i;
  
  SendThing(context, emission, 4,
  byte = (DM_nat1)natural;
  DM_SendOneByte(context, emission,byte);
  for ( i=3 ; i>0 ; i-- ) {
    natural = (natural - byte)/256;
    byte = (DM_nat1)natural;
    DM_SendOneByte(context, emission,byte);
  }
  )
}

void
DM_receive_netnat3 (DM_Context *context,
                DM_InBuffer *reception,
                    void *p) {
  DM_nat3       result      = 0;
  DM_nat3       scale       = 256;
  unsigned int  i;
  
  ReceiveThing(context, reception, 4,
  result = DM_ReceiveOneByte(context, reception);
  for ( i=3 ; i>0 ; i-- ) {
    result += scale*DM_ReceiveOneByte(context, reception);
    scale *= 256;
  }
  DM_Content(DM_netnat3,p) = htonl(result);
  )
}

/*
 * I must read the sources of XDR! There is surely some other way
 * that is mathematically more precise.
 */

void
DM_send_float1 (DM_Context *context,
                DM_OutBuffer *emission, 
                void *p) {
  DM_float1     value       = DM_Content(DM_float1,p);
  int           exponent;
  float         mntss       = ((float)(frexp(((double)value),&exponent)));
  DM_int3       mantissa    = mntss/FLT_EPSILON;
  DM_int2       expnt       = exponent;

  SendThing(context, emission, 6,
  DM_send_int3(context, emission,&mantissa);
  DM_send_int2(context, emission,&expnt);
  )
}

void
DM_receive_float1 (DM_Context *context,
                   DM_InBuffer *reception, 
                   void *p) {
  DM_int3 mantissa;
  DM_int2 expnt;
  double  mntss;

  ReceiveThing(context, reception, 6,
  DM_receive_int3(context, reception,&mantissa);
  DM_receive_int2(context, reception,&expnt);
  mntss = mantissa*FLT_EPSILON;
  DM_ResetErrno();
  DM_Content(DM_float1,p) = ((DM_float1)(ldexp(mntss, (int)expnt)));
  if ( DM_IsErrnoSet() ) 
    DM_Abort(context, DME_FLOAT_DOMAIN_ERROR);
  )
}

void
DM_send_float2 (DM_Context *context,
                DM_OutBuffer *emission,
                void *p) {
  DM_float2     value       = DM_Content(DM_float2,p);
  int           exponent;
  double        mntss       = frexp(value,&exponent);
  DM_int4       mantissa    = mntss/DBL_EPSILON;
  DM_int2       expnt       = exponent;

  SendThing(context, emission, 10,
  DM_send_int4(context, emission,&mantissa);
  DM_send_int2(context, emission,&expnt);
  )
}

void
DM_receive_float2 (DM_Context *context,
                DM_InBuffer *reception, 
                   void *p) {
  DM_int4       mantissa;
  DM_int2       expnt;
  DM_float2     value;
  double        mntss;

  ReceiveThing(context, reception, 10,
  DM_receive_int4(context, reception,&mantissa);
  DM_receive_int2(context, reception,&expnt);
  mntss = mantissa*DBL_EPSILON;
  DM_ResetErrno();
  value = ldexp(mntss, (int)expnt);
  DM_Content(DM_float2,p) = (DM_float2)value;
  if ( DM_IsErrnoSet() ) 
    DM_Abort(context, DME_FLOAT_DOMAIN_ERROR);
  )
}

void
DM_send_float3 (DM_Context *context,
                DM_OutBuffer *emission, 
                void *p) {
  DM_Abort(context, DME_NOT_YET_IMPLEMENTED);
}

void
DM_receive_float3 (DM_Context *context,
                   DM_InBuffer *reception, 
                   void *p) {
  DM_Abort(context, DME_NOT_YET_IMPLEMENTED);
}

/*
 * No need to encode the translation which is site-dependent.
 * But don't forget to recompute the layout after reception.
 */

void
DM_send_offset (DM_Context *context,
                DM_OutBuffer *emission, 
                void *p) {
  DM_offset     offset      = DM_Content(DM_offset,p);

  SendThing(context, emission, 2,
  DM_SendOneByte(context, emission,DM_OffsetToRepetitionTypeIndex(offset));
  DM_SendOneByte(context, emission,DM_OffsetToItemTypeIndex(offset));
  )
}

void
DM_receive_offset (DM_Context *context,
                   DM_InBuffer *reception,
                   void *p) {
  DM_nat1       itemtypindex;
  DM_nat1       reptypindex;

  ReceiveThing(context, reception, 2,
  reptypindex = DM_ReceiveOneByte(context, reception);
  itemtypindex = DM_ReceiveOneByte(context, reception);
  DM_Content(DM_offset,p) = DM_PackOffset(0,reptypindex,itemtypindex);
  )
}

/*
 * Encode a field with type data or code. 
 * There is no guarantee on the success of such a transmission.
 * Bytes are copied as they stand may be truncated or not between
 * sites with different sizes. 
 * The byte sequence is prefixed by its size to verify some things.
 */

void
DM_send_raw_field (DM_Context *context,
                   DM_DMeroonType *type,
                   void           *p,
                   DM_OutBuffer   *emission) {
  DM_nat1 size = type->size;
  DM_nat1 i, byte;
  DM_nat1 *q = p;

  SendThing(context, emission, 1+size,
    DM_SendOneByte(context, emission, size);
    for ( i=0 ; i<size ; i++, q++ ) {
      byte = DM_Content(DM_nat1,q);
      DM_SendOneByte(context, emission, byte);
    }
  )
}

void
DM_receive_raw_field (DM_Context *context,
                      DM_DMeroonType *type,
                      void           *p,
                      DM_InBuffer    *reception ) {
  DM_nat1 size = type->size;
  DM_nat1 i, orgsize, sink;
  DM_nat1 *q = p;

  /* Get how many bytes were sent. */
  if ( DM_CanReceiveBytes(context, reception, 1) ) { 
    orgsize = DM_ReceiveOneByte(context, reception);
    /* copy sent bytes to their destination. */
    if ( DM_CanReceiveBytes(context, reception, orgsize) ) { 
      for ( i=0 ; (i<size)&&(i<orgsize) ; i++, q++ ) {
        DM_Content(DM_nat1,q) = DM_ReceiveOneByte(context, reception);
      }
    }
    /* consume extra sent bytes if any. */
    if ( DM_CanReceiveBytes(context, reception, size-orgsize) ) { 
      for ( i=(size<orgsize)?size:orgsize ; i<orgsize ; i++ ) {
        sink = DM_ReceiveOneByte(context, reception);
      }
    } else {
      DM_Abort(context, DME_MISSING_BYTES);
    }
  } else {
    DM_Abort(context, DME_MISSING_BYTES);
  }
}

/* DOCUMENTATION Convert handlers into numbers and back. 
 * This table accumulates the possible handler_code functions, their code,
 * and their name.  

 * FUTURE: should rank these methods by frequency. 
 */

typedef DM_ErrorCode (*DM_HC)(void*, void*);

DM_handler_code_descriptor DM_handler_code_table[20] = {
  {91, (DM_HC)DM_handle_site_connection_request, 
   "DM_handle_site_connection_request"},
  {92, (DM_HC)DM_handle_object_send_request, 
   "DM_handle_object_send_request"},
  {93, (DM_HC)DM_handle_object_fetch_request,
   "DM_handle_object_fetch_request"},
  {94, (DM_HC)DM_handle_object_indexed_mutation_request,
   "DM_handle_object_indexed_mutation_request"},
  {95, (DM_HC)DM_handle_object_mutation_request,
   "DM_handle_object_mutation_request"},
  {96, (DM_HC)DM_handle_script_evaluation_request,
   "DM_handle_script_evaluation_request"},
  {97, (DM_HC)DM_handle_owner_change_request,
   "DM_handle_owner_change_request"},
  {98, (DM_HC)DM_handle_remote_owner_change_request,
   "DM_handle_remote_owner_change_request"},
  {191, (DM_HC)DM_handle_site_connection_answer,
   "DM_handle_site_connection_answer"},
  {192, (DM_HC)DM_handle_object_fetch_answer,
   "DM_handle_object_fetch_answer"},
  {193, (DM_HC)DM_handle_object_mutation_answer,
   "DM_handle_object_mutation_answer"},
  {194, (DM_HC)DM_handle_owner_change_answer,
   "DM_handle_owner_change_answer"},
  {195, (DM_HC)DM_handle_remote_owner_change_answer,
   "DM_handle_remote_owner_change_answer"},
  {196, (DM_HC)DM_handle_failure_answer,
   "DM_handle_failure_answer"},
  {197, (DM_HC)DM_handle_script_evaluation_answer,
   "DM_handle_script_evaluation_answer"},
  /* This marks the end of this table. */
  {0, NULL, NULL}
};

/* RESOURCE DM_send_handler_code
 * Marshal a handler_code value (a pointer to one of the above functions)
 * into an index.
 */

void
DM_send_handler_code (DM_Context     *context,
                      DM_OutBuffer   *emission,
                      void           *p ) 
{
  DM_handler_code fn = DM_Content(DM_handler_code, p);
  DM_nat1 i = 0;

  for ( i=0 ; DM_handler_code_table[i].code != 0 ; i++ ) {
    if ( fn == DM_handler_code_table[i].handler ) {
      SendThing(context, emission, 1,
         DM_SendOneByte(context, emission, i);
);
        return;
    } else {
      continue;
    }
  }
  DM_Abort(context, DME_UNKNOWN_HANDLER_CODE);
}

/* RESOURCE DM_receive_handler_code
 * Unmarshal an index into a handler_code value (a pointer to one of 
 * the above functions).
 */

void
DM_receive_handler_code (DM_Context     *context,
                         DM_InBuffer    *reception,
                         void           *p ) 
{
  DM_nat1 i;

  ReceiveThing(context, reception, 1,
  i = DM_ReceiveOneByte(context, reception);
);
  DM_Content(DM_handler_code, p) = DM_handler_code_table[i].handler;
}

/* }}}
 * {{{ Initialization
 */

void
DM_comm_initialize (DM_Context *context) 
{
  DM_Site *site = DM_TheCurrentSite();
  DM_Hashtable *entries = 
    DM_Allocate1(context, Hashtable, DM_HASHTABLE_INITIAL_SIZE);
  DM_Hashtable *exits   = 
    DM_Allocate1(context, Hashtable, DM_HASHTABLE_INITIAL_SIZE);
  DM_Hashtable *sites   = 
    DM_Allocate1(context, Hashtable, DM_HASHTABLE_INITIAL_SIZE);
  DM_Hashtable *oblist  = 
    DM_Allocate1(context, Hashtable, DM_HASHTABLE_INITIAL_SIZE);

  DM_LinkVersionStringIntoSite(dmcomm);

  /* initialize the random generator. */
  srandom((int)(DM_get_date(context)));

  /* Initialize the entries hashtable.  */
  entries->hasher = DM_CodeOf(DM_compute_hash_entry);
  entries->extractor = DM_NULL;
  entries->next = DM_NULL;
  entries->size = 0;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, entries);
  }

  /* Initialize the exits hashtable. */
  exits->hasher = DM_CodeOf(DM_compute_hash_exit);
  exits->extractor = DM_CodeOf(DM_fake_pointer);
  exits->next = DM_NULL;
  exits->size = 0;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, exits);
  }

  /* Initialize the sites hashtable. */
  sites->hasher = DM_CodeOf(DM_compute_hash_site);
  sites->extractor = DM_NULL;
  sites->next = DM_NULL;
  sites->size = 0;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, sites);
  }

  /* Initialize the oblist hashtable. */
  oblist->hasher = DM_CodeOf(DM_compute_hash_symbol);
  oblist->extractor = DM_NULL;
  oblist->next = DM_NULL;
  oblist->size = 0;
  site->oblist  = oblist;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, oblist);
  }

#if !defined(DM_SAFETY) || (DM_SAFETY==0)
  /* insert the entries table in the site object.   */
  site->entries = entries;
  site->exits   = exits;
  site->sites   = sites;
  /* Now that site->sites is initialized, register the current site. */
  (void)DM_register_site(context, site);
#else 
  /* To be done in some global variables.             FUTURE */
#endif

  /* Externalize the Site: create and register an Entry for it. */
  DM_externalize(context, (DM_Object*)site);
}

/* }}}
 * end of dmcomm.c */
