/* $Id: dmdbg.c,v 1.56 1998/06/05 18:38:25 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmdbg.c
 * This file contains various functions for the low-level debugging of
 * DMeroon. These functions generally accepts as last arguments the stream
 * onto which they must print. If NULL then, by default, this is the value
 * of DM_debug_stream which is initially stderr. These functions try to
 * guess missing parameters (ie parameters specified as NULL or 0).

 * NOTE: Errors are reported with printf on DM_debug_stream, debug
 * printing is done with DM_print_char_star and others so they may be
 * redirected to other streams and thus allow remote enquiries. To avoid
 * possible problems with the automatic protocol decoding of streams,
 * the first character of any line is a whitespace.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmdbg,
  "$Id: dmdbg.c,v 1.56 1998/06/05 18:38:25 queinnec Exp $");

/* RESOURCE DM_debug_stream
 * This is the variable that holds the stream onto which the debug
 * functions print their messages. This variable is global and
 * mutable; this is not the usual rule in DMeroon but this is useful
 * for debug. Initially this is stderr.
 */

FILE *DM_debug_stream = stderr;

/* {{{ Debug-dependent utilities
 * No need to have these functions in a working DMeroon, so don't
 * compile them in non-debug mode. However the DM_debug_initialize
 * function is always present.
 */

#ifdef DM_DEBUG

/*     {{{ Basic printing utilities

 * RESOURCE DM_print_memory_slice
 * Displays a memory slice in hexadecimal.
 */

void
DM_print_memory_slice (DM_Context   *context,
                       void         *q, 
                       DM_nat3      length, 
                       char         *ascii_prefix,
                       char         *ascii_postfix,
                       void         *output_port ) {
  void           *real_output_port;
  unsigned char  *p = q;
  DM_nat3        i, j;
  unsigned short byte1, byte2, byte3, byte4;
  char           line[80];
  char           *c = q, c1, c2, c3, c4;

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*) context->errstream;
  }

  DM_print_char_star(context, "\r\n", 2, real_output_port);

  /* Do the real work 4 bytes by 4 bytes, with lines of 16 bytes. */
  for ( i=0 ; i<length ; i+=16 ) {
    sprintf(line, " " DM_0X "%p:", p);
    DM_print_char_star(context, line, 0, real_output_port);
    for ( j=0 ; j<16 ; j+=4 ) {
      byte1 = *p++; byte2 = *p++; byte3 = *p++; byte4 = *p++;
      sprintf(line, " %02hx%02hx%02hx%02hx", byte1, byte2, byte3, byte4);
      DM_print_char_star(context, line, 0, real_output_port);
    }
    DM_print_char_star(context, "   ", 3, real_output_port);
    DM_print_char_star(context, ascii_prefix, 0, real_output_port);
    for ( j=0 ; j<16 ; j+=4 ) {
      c1 = *c++; c2 = *c++; c3 = *c++; c4 = *c++; 
      if ( !isprint(c1) ) c1 = '.';
      if ( !isprint(c2) ) c2 = '.';
      if ( !isprint(c3) ) c3 = '.';
      if ( !isprint(c4) ) c4 = '.';
      sprintf(line, " %c%c%c%c", c1, c2, c3, c4);
      DM_print_char_star(context, line, 0, real_output_port);
    }
    DM_print_char_star(context, ascii_postfix, 0, real_output_port);
    DM_print_char_star(context, "\r\n", 2, real_output_port);
  }
}

/* RESOURCE DM_print_types
 * Print the description of all DMeroon types and C types.
 * This routine is called whenever something wrong is discovered 
 * about types when checking them (see dmtypes.c). This will ease 
 * redefining correctly the dmsizes.h file.
 */

void
DM_print_types (DM_Context *context,
                void *output_port) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  void *real_output_port;
  int  i;
  char line[80];

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*) context->errstream;
  }

  DM_print_char_star(context, " \tType descriptions\r\n", 0, real_output_port);

  for ( i=0 ; i<implemented_types->item_length ; i++ ) {
    DM_DMeroonType* the_type = implemented_types->item[i];

    sprintf(line,
            " DMeroon type(%u) " DM_0X "%p %s:\t size=%u alignment=%u \t%s\r\n",
            the_type->index,
            the_type,
            DM_predefined_name(context, (DM_Object*)the_type),
            the_type->size,
            the_type->alignment,
            (the_type->alignment==0)?"NOT IMPLEMENTED":"");
    DM_print_char_star(context, line, 0, real_output_port);
  }
  DM_print_char_star(context, "\r\n", 2, real_output_port);
  
  /* also print the C types. */
  while ( (i++)<256 ) {
    /* In fact, this should be a CType but they are sufficiently similar
     * that they may be handled similarly. */
    DM_DMeroonType* the_type = implemented_types->item[i];

    if ( ( the_type ) && (DM_IsA(context, the_type, Type)) ) {
      sprintf(line,
              " C type(%u) " DM_0X "%p %s:\t size=%u alignment=%u \t%s\r\n",
              the_type->index,
              the_type,
              DM_predefined_name(context, (DM_Object*)the_type),
              the_type->size,
              the_type->alignment,
              (the_type->alignment==0)?"NOT IMPLEMENTED":"");
      DM_print_char_star(context, line, 0, real_output_port);
    }
  }
}

/* RESOURCE DM_print_object
 * Shallow prints in a readable form the content of a DMeroon
 * instance, field by field, indexed or regular fields.  Also
 * indicates its class, proxy and overall options.
 */

void
DM_print_object (DM_Context *context,
                 DM_Object *o,
                 void      *output_port) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  void           *real_output_port;
  DM_Object      *proxy;
  DM_Class       *class;
  char           line[80];
  DM_nat3        fn;
  DM_translation t       = 0;
  DM_Field       *field;
  DM_nat3        on;
  DM_offset      last_offset;
  DM_nat1        itemtypindex, reptypindex;
  DM_DMeroonType *itemtyp, *reptyp;
  DM_nat3        i, j, rep;
  void           *p;
  
  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*) context->errstream;
  }

  if ( ! DM_IsObject(o) ) {
    DM_print_char_star(context, " Not a DMeroon object.\r\n", 
                       0, real_output_port );
    return;
  }

  proxy         = DM_ObjectToProxy(context, o);
  class         = DM_ObjectToClass(context, o);
  fn            = DM_class_indexed_field_length(context, class);
  
  if ( DM_IsA(context, proxy, Entry) ) {
    DM_DeclareNarrowedObject(entry, Entry, proxy);
    
    if ( DM_IsA(context, entry, Exit) ) {
      DM_print_char_star(context, " Remote ", 0, real_output_port);
      if ( DM_is_up_to_date(context, o) ) {
        DM_print_char_star(context, "Uptodate", 0, real_output_port);
      } else {
        DM_print_char_star(context, "Obsolete", 0, real_output_port);
      }
    }
    if ( DM_HasOption(entry, DM_ENTRY_IS_UNINITIALIZED) ) {
      DM_print_char_star(context, " Uninitialized", 0, real_output_port);
    } else if ( DM_HasOption(entry, DM_ENTRY_IS_COPYABLE) ) {
      DM_print_char_star(context, " Copyable", 0, real_output_port);
    } else {
      DM_print_char_star(context, " Sharable", 0, real_output_port);
    }
  } else {
    /* The object is local and not externalized. */
    DM_print_char_star(context, " Local", 0, real_output_port);
    if ( proxy == DM_TheConstant(uninitialized) ) {
      DM_print_char_star(context, " Uninitialized", 0, real_output_port);
    } else if ( proxy == DM_TheConstant(true) ) {
      DM_print_char_star(context, " Copyable", 0, real_output_port);
    } else {
      /* Here proxy == DM_no_proxy_value */
      DM_print_char_star(context, " Sharable", 0, real_output_port);
    }
  }
  sprintf(line, " DMeroon instance at " DM_0X "%p of ", o);
  DM_print_char_star(context, line, 0, real_output_port);
  if ( DM_IsPredefinedClass(class) ) {
    DM_print_char_star(context, "predefined ", 0, real_output_port);
  }
  DM_print_char_star(context, "class ", 0, real_output_port);
  DM_print_char_star(context, 
                     DM_predefined_name(context, (DM_Object*)field),
                     0,
                     real_output_port );
  DM_print_char_star(context, "\r\n", 2, real_output_port);
  sprintf(line, " _class= " DM_0X "%p, _proxy= " DM_0X "%p\r\n",
          class, proxy );
  DM_print_char_star(context, line, 0, real_output_port);

  for ( i=0 ; i<fn ; i++ ) {
    field        = DM_class_field(context, class, i);
    on           = field->offset_length;
    last_offset  = field->offset[on-1];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp      = implemented_types->item[itemtypindex];
      
    if ( reptypindex == DM_nought_index ) {
      /* This field is not indexed. */
      t = DM_OffsetPad(t, itemtyp);
      DM_print_char_star(context, " ", 1, real_output_port);
      DM_print_char_star(context, 
                         DM_predefined_name(context, (DM_Object*)field),
                         0,
                         real_output_port );
      DM_print_char_star(context, "(", 1, real_output_port);
      DM_print_char_star(context, 
                         DM_predefined_name(context, (DM_Object*)itemtyp),
                         0,
                         real_output_port );
      DM_print_char_star(context, "): ", 3, real_output_port);
      DM_local_print(context, itemtyp, o, t, real_output_port);
      DM_print_char_star(context, "\r\n", 2, real_output_port);
      t = DM_OffsetExtend(t, itemtyp);
    } else {
      t   = DM_OffsetPad(t, reptyp);
      p   = DM_Displace(o, t);
      rep = DM_read_repetition_factor(context, p, reptypindex);
      t   = DM_OffsetExtend(t, reptyp);
      t   = DM_OffsetPad(t, itemtyp);
      if ( rep == 0 ) {
        /* empty indexed field */
        DM_print_char_star(context, " ", 1, real_output_port);
        DM_print_char_star(context, 
                           DM_predefined_name(context, (DM_Object*)field),
                           0,
                           real_output_port );
        DM_print_char_star(context, "(", 1, real_output_port);
        DM_print_char_star(context, 
                           DM_predefined_name(context, (DM_Object*)itemtyp),
                           0,
                           real_output_port );
        DM_print_char_star(context, ")[none]\r\n", 0, real_output_port);        
      } else {
        for ( j=0 ; j<rep ; j++ ) {
          DM_print_char_star(context, " ", 1, real_output_port);
          DM_print_char_star(context, 
                             DM_predefined_name(context, (DM_Object*)field),
                             0,
                             real_output_port );
          DM_print_char_star(context, "(", 1, real_output_port);
          DM_print_char_star(context, 
                             DM_predefined_name(context, (DM_Object*)itemtyp),
                             0,
                             real_output_port );
          DM_print_char_star(context, ")[", 2, real_output_port);
          DM_C_unsigned_long_int_print(context, &j, real_output_port);
          DM_print_char_star(context, "]: ", 3, real_output_port);
          DM_local_print(context, itemtyp, o, t, real_output_port);
          DM_print_char_star(context, "\r\n", 2, real_output_port);
          t += itemtyp->repsize;
        }
      }
      /* Realign after any non final indexed field. */
      t = DM_OffsetRealign(t);
    }
  }
}

/* RESOURCE DM_print_bytes
 * Displays the first bytes composing a DMeroon instance.
 * If length is null then displays the whole DMeroon instance.
 *  The header of the object (support and DMeroon header) is also printed.
 */

void
DM_print_bytes (DM_Context *context,
                DM_Object *o, 
                DM_nat3   length,
                void      *output_port ) {
  void* real_output_port;
  void* p;

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*) context->errstream;
  }

  /* Compute a sensible default size and address to print. 
   * Also print the header for the support language. */
  if ( length == 0 ) {
    p = DM_ToC(o);
    length = DM_PrefixSize();
    if ( DM_IsDMeroonInstance(DM_ToSupport(o)) ) {
      length += DM_size(context, o);
    } else {
      /* Just print 16 bytes */
      length += 16;
    }
  }
  DM_print_memory_slice(context, p, length, "", "", real_output_port);
}

/* RESOURCE DM_print_ports
 * Shows the status of all file-descriptors below a given number.
 * It simply displays the result of a call to select. If n is zero
 * then tries to show the status of all file-descriptors.
 */

void
DM_print_ports (DM_Context *context,
                DM_nat3 n,
                void    *output_port) {
  DM_ports_vector *active_ports = 
    (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));
  void              *real_output_port;
  DM_Port           *port;
  unsigned int      i;
  int               nfound;
  struct timeval    tv;
  char              line[100];
  /* Initialize fd_sets with all active connections. */
  fd_set            fd_current_ports_to_read;
  fd_set            fd_current_ports_to_write;
  fd_set            fd_current_ports_to_check;

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*) context->errstream;
  }

  /* Compute a sensible default for n if zero. */
  n = n ? n+1 : getdtablesize();

  /* Mark that we are interested by the status of all sockets. */
  FD_ZERO(&fd_current_ports_to_read);
  FD_ZERO(&fd_current_ports_to_write);
  FD_ZERO(&fd_current_ports_to_check);
  for ( i=0 ; i<n ; i++ ) {
    FD_SET(i, &fd_current_ports_to_read);
    FD_SET(i, &fd_current_ports_to_write);
    FD_SET(i, &fd_current_ports_to_check);
  }

  tv.tv_sec = 0;
  tv.tv_usec = 0;
  DM_ResetErrno();

retry:
  /* Try to enquire all ports. Start from the highest possible n
   * and try select until EBADF does not occur. */
  if ( (nfound = select(n,
                        &fd_current_ports_to_read,
                        &fd_current_ports_to_write,
                        &fd_current_ports_to_check,
                        &tv )) < 0 ) {
    if ( DM_IsErrnoEqualTo(DM_EBADF) ) {

      DM_DebugAtLevel(11,
        fprintf(DM_FindErrorStream(context), "Selecting only %d channels\n", n);
);

      if ( n-- > 0 ) goto retry;
    }
    DM_print_char_star(context, "Cannot select", 0, real_output_port);
    return;
  }

  sprintf(line, " DM_max_port_number is %u.\r\n",  DM_max_port_number);
  DM_print_char_star(context, line, 0, real_output_port);
  sprintf(line, " select returns %d.\r\n",  nfound);
  DM_print_char_star(context, line, 0, real_output_port);
  if ( nfound > 0 ) {
    for ( i=0 ; i<n ; i++ ) {
      port = active_ports->port[i];
      if (   (FD_ISSET(i, &fd_current_ports_to_read))
          || (FD_ISSET(i, &fd_current_ports_to_write))
          || (FD_ISSET(i, &fd_current_ports_to_check)) ) {
        sprintf(line, " File descriptor %d:%s%s%s",
                i,
                (FD_ISSET(i, &fd_current_ports_to_read))?" readable":"",
                (FD_ISSET(i, &fd_current_ports_to_write))?" writable":"",
                (FD_ISSET(i, &fd_current_ports_to_check))?" exception":"" );
       DM_print_char_star(context, line, 0, real_output_port);
        if ( port ) {
          sprintf(line, " (" DM_0X "%p)%s%s%s",
                  port,
                  DM_HasOption(port, DM_PORT_IS_DRIED)?" DRIED":"",
                  DM_HasOption(port, DM_PORT_IS_PEERLESS)?" PEERLESS":"",
                  DM_HasOption(port, DM_PORT_IS_CLOSED)?" CLOSED":"" );
          DM_print_char_star(context, line, 0, real_output_port);
        }
        DM_print_char_star(context, "\r\n", 2, real_output_port);
      }
    }
  }
}

/*     }}}
 *     {{{ Interactive debug monitoring
 */

/* RESOURCE DM_handle_debug_request
 * Handles a debug request, returns DM_TRUE if interaction should continue
 * DM_FALSE otherwise. 
 *
 * The external debugger may issue some commands to inspect the state
 * of DMeroon:
 *    q|quit    to quit
 *    ?|help    to know the commands
 *    types     to display the DMeroon and C types characteristics
 *    ports     to display the state of Port connections
 *    site      to display the site object
 *    constants to display the usual constants
 *    p|print   to print a DMeroon object in plain text
 *    b|bytes   to display the bytes representing a DMeroon object
 *    d|dump    to display the bytes of a memory slice (2 numbers as arguments)
 *    en|entry  to display an Entry item
 *    ex|exit   to display an Exit item
 */

DM_c_boolean
DM_handle_debug_request (DM_Context *context,
                         char       *request,
                         void       *output_port) {
  char         *p, *q;
  void         *real_output_port;
  DM_nat3      address;
  DM_nat3      sz;
  DM_c_boolean printend = DM_TRUE;

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    if ( context ) {
      real_output_port = (void*) context->errstream;
    } else {
      real_output_port = DM_debug_stream;
    }
  }

#define EqualTo(string) \
  ( ! strncmp(request, string, sizeof(string)-1) )
       
  /* Commands without arguments. */
  if ( EqualTo("inspector") || EqualTo("?") || EqualTo("help") ) {
    /* The 'inspector' word is used to signal that the external inspector
     * is ready to send other commands. */
    DM_print_char_star(context, 
           " \tDMeroon Inspector (Protocol $Revision: 1.56 $)\r\n"
           " Possible commands are: ?, quit, \r\n"
           "   types, ports, site, constants, \r\n"
           "   bytes 0xaddress, print 0xaddress, dump 0xaddress size, \r\n"
           "   entry 0xnumber, exit 0xnumber portnumber\r\n",
           0,
           real_output_port );
    goto epilogue;
  }
  if ( EqualTo("q") || EqualTo("quit") ) {
    return DM_FALSE;
  }
  if ( EqualTo("types") || EqualTo("ty") ) {
    DM_print_types(context, real_output_port);
    goto epilogue;
  }
  if ( EqualTo("site") || EqualTo("si") ) {
    DM_print_object(context, (DM_Object*)DM_TheCurrentSite(), real_output_port);
    goto epilogue;
  }
  if ( EqualTo("ports") || EqualTo("po") ) {
    DM_print_ports(context, 0, real_output_port);
    goto epilogue;
  }
  if ( EqualTo("constants") || EqualTo("co") ) {
    char line[80];

    sprintf(line, " Nil  \t\t is " DM_0X "%p\r\n", DM_TheConstant(nil));
    DM_print_char_star(context, line, 0, real_output_port);
    sprintf(line, " True \t\t is " DM_0X "%p\r\n", DM_TheConstant(true));
    DM_print_char_star(context, line, 0, real_output_port);
    sprintf(line, " False\t\t is " DM_0X "%p\r\n", DM_TheConstant(false));
    DM_print_char_star(context, line, 0, real_output_port);
    sprintf(line, " Uninitialized\t is " DM_0X "%p\r\n",
            DM_TheConstant(uninitialized) );
    DM_print_char_star(context, line, 0, real_output_port);
    sprintf(line, " Missing\t is " DM_0X "%p\r\n",
            DM_TheConstant(missing) );
    DM_print_char_star(context, line, 0, real_output_port);
    goto epilogue;
  }
  
  /* Commands now have a first argument which is an address (or an
   * external number), decode it and store it into the 'address'
   * (DM_nat3) variable. First skip the letters that end the command
   * name then additional slashes and whitespaces. Determine from the
   * prefix, the numerical base to decode the number: 0 for octal, 0x
   * for hexadecimal.  */

  p = request;
  while ( (*p) && isalnum(*p) ) p++;
  while ( (*p) && ( ( !(isalnum(*p)) ) || (*p=='/') ) ) p++;
  address = DM_string_to_nat3(context, p, &q);
  if ( p == q ) {
    DM_print_char_star(context, 
      " Unrecognized command or incorrect numeric first argument.\r\n", 
                       0, real_output_port );
    goto epilogue;
  } else { 
    p = q; 
  }
  
  if ( EqualTo("b") || EqualTo("bytes") || EqualTo("by") ) {
    DM_print_bytes(context, 
                   (DM_Object*)(DM_SupportAddress)address, 
                   (DM_nat3) 0,
                   real_output_port );
    goto epilogue;
  }
  if ( EqualTo("p") || EqualTo("print") || EqualTo("pr") ) {
    DM_print_object(context, 
                    (void*)(DM_SupportAddress)address, 
                    real_output_port );
    goto epilogue;
  }
  /* No longer correct since now DM_key is different from DM_nat4.
  if ( EqualTo("entry") || EqualTo("en") ) {
    DM_Entry* e = DM_extract_entry(context, address);

    if ( e ) {
      DM_print_object(context, (DM_Object*)e, real_output_port);
    } else {
      DM_print_char_star(context, " No such Entry\r\n", 0, real_output_port);
    }
    goto epilogue;
  }                                           */

  /* Commands have a second argument which is a size,
   * decode it and store it into the 'sz' variable. */
  while ( (*p) && ( ( !(isalnum(*p)) ) || (*p=='/') ) ) p++;
  sz = DM_string_to_nat3(context, p, &q);
  if ( p == q ) {
    DM_print_char_star(context, 
      " Unrecognized command or incorrect numeric second argument.\r\n", 
                       0, real_output_port);
    goto epilogue;
  } else { 
    p = q; 
  }
  
  if ( EqualTo("d") || EqualTo("dump") || EqualTo("du") ) {
    DM_print_memory_slice(context, 
                          (void*)(DM_SupportAddress)address, 
                          (DM_nat3) sz,
                          "", "",
                          real_output_port );
    goto epilogue;
  }
  /* No longer correct since DM_key is different from DM_nat4.
  if ( EqualTo("exit") || EqualTo("ex") ) {
    DM_Exit* e = DM_extract_exit(context, address);

    if ( e ) {
      DM_print_object(context, (DM_Object*)e, real_output_port);
    } else {
      DM_print_char_star(context, " No such Exit\r\n", 0, real_output_port);
    }      
    goto epilogue;
  }                                                              */
  
  DM_print_char_star(context, " Unrecognized command\r\n", 0, real_output_port);

  /* This code is there to be shared since all commands output end
   * with the word "end." */
 epilogue:
  if ( printend ) {
    DM_print_char_star(context, " end.\r\n", 0, real_output_port);
  }
  return DM_TRUE;
}

/* RESOURCE DM_debug_monitor
 * Reads an input stream, and interprets the incoming requests.
 * You can run it with (DM-debug-interpret 0); this command will
 * take its command from stdin. ATTENTION, reading is done very 
 * naively and without any parsing.  TO BE IMPROVED.
 */

void
DM_debug_monitor (DM_Context *context,
                  DM_nat1    in) {
  char request[80];
  int  size;

  while ( (size = read(in, request, 80)) > 0 ) {
    /* Call DM_handle_debug_request again and again until it returns
     * DM_FALSE to mean that the interaction is finished. */
    if ( ! DM_handle_debug_request(context, request, NULL) )
      return;
  }
}

/* RESOURCE DM_fork_inspector
 * This function forks an external inspector to allow an easier inspection
 * of DMeroon instances. The external inspector may be explicitly specified
 * or found in the DMEROON_EXTERNAL_INSPECTOR environment variable. This
 * external inspector will have its stdout piped towards DM_debug_monitor
 * and must respect this protocol. DMeroon waits ten seconds (by help
 * default) for the external inspector to issue its first command, this
 * command may be the "inspector" word for instance.
 */

void
DM_fork_inspector (DM_Context *context,
                   char       *command) {
  int pid, i;
  int tmp;
  int fd[2];
  
  /* Compute a sensible command. */
  if ( (command == NULL) || (strlen(command)==0) ) {
    command = getenv("DMEROON_EXTERNAL_INSPECTOR");
    if ( command == NULL ) {
      fprintf(DM_FindErrorStream(context), "No specified external debugger\n");
      return;
    } 
  }

  DM_ResetErrno();
  if ( pipe(fd) != 0 ) {
    fprintf(DM_FindErrorStream(context), "Cannot create pipe\n");
    return;
  }

  DM_ResetErrno();
  if ( (pid = fork()) < 0 ) {
    fprintf(DM_FindErrorStream(context), "Cannot fork\n");
    goto end;
  }
  if ( pid == 0 ) { 
    /* This is the forked process.
     * Its stdout is mapped onto fd[1] to allow communications. 
     * This simple protocol allow for tcl clients... 
     */
    int  nfds    = getdtablesize();
    char *argv[2];
    
    argv[0] = command;
    argv[1] = NULL;
    DM_ResetErrno();
    tmp = close(0);
    tmp = close(1);
    tmp = close(fd[0]);
    if ( (tmp = dup2(fd[1], 1)) < 0 ) {
      fprintf(DM_FindErrorStream(context), "Cannot replace stdout\n");
      exit(1);
    }
    /* close all other descriptors but keep stderr. */
    DM_ResetErrno();
    for ( i=3 ; i<=nfds ; i++ ) 
      tmp = close(i);
    DM_ResetErrno();
    if ( (execv(command, argv)) < 0 ) {
      fprintf(DM_FindErrorStream(context), "Cannot exec %s\n", command);
      exit(1);
    }
  } else {
    /* This is still the current process.
     * Interpret debug requests as they come from fd[0] but it will
     * wait at most ten seconds for the inspector to come up.
     */
    fd_set         fds;
    int            nfound;
    struct timeval timeout;

    FD_ZERO(&fds);
    FD_SET(fd[0], &fds);
    DM_ResetErrno();
    tmp = close(fd[1]);
    timeout.tv_sec  = DM_EXTERNAL_INSPECTOR_TIMEOUT;
    timeout.tv_usec = 0;
    DM_ResetErrno();
    if ( (nfound = select(fd[0]+1,
                          &fds,
                          (fd_set*)NULL,
                          (fd_set*)NULL,
                          &timeout )) < 0 ) {
      fprintf(DM_FindErrorStream(context), "Cannot select\n");
      goto end;
    }
    if ( (nfound>0) && FD_ISSET(fd[0], &fds) ) {
      DM_debug_monitor(context, fd[0]);
    } else {
      fprintf(DM_FindErrorStream(context), "No living inspector\n");
      /* You have to kill the external inspector by yourself. */
    }
  }

end:
  close(fd[0]);
  close(fd[1]);
}

#endif /* DM_DEBUG */

/*     }}}
 * }}}
 * {{{ Initialization
 */

void
DM_debug_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmdbg);
}

/* }}}
 * end of dmdbg.c */
