/* $Id: dmextr.c,v 1.67 1998/05/06 20:03:12 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/*
 * This file contains additional extra libray routines that are not
 * necessary for the kernel but still useful.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmextr,
  "$Id: dmextr.c,v 1.67 1998/05/06 20:03:12 queinnec Exp $");

static char* nibbles2characters = "0123456789abcdef";

/* RESOURCE DM_cstring_to_string
 * Convert a C string into an instance of String. The resulting String is
 * not initialized.
 */

DM_String*
DM_cstring_to_string (DM_Context *context,
                      char       *s,
                      DM_nat3    length ) 
{
  DM_nat3   size = (length) ? length : (strlen(s)+1);
  DM_String *o   = DM_Allocate1(context, String, size);

  memcpy(o->character, s, size);
  return o;
}

/* RESOURCE DM_cstring_to_type 
 * Converts a C string into a DMeroon Type. Return DM_NULL if not found.
 * NOTE What if we want to retrieve a CType ???
 */

DM_DMeroonType*
DM_cstring_to_type (DM_Context *context,
                    char *s) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat1 i;

  for ( i=0 ; i<256 ; i++ ) {
    DM_DMeroonType *type = implemented_types->item[i];

    if ( type ) {
      if ( strcmp(s, DM_predefined_name(context, (DM_Object*)type)) ) {
        continue;
      } else {
        return type;
      }
    } else {
      /* DMeroon types are contiguous, the first one being DM_NULL is
       * the last DMeroon type. */
      return DM_NULL;
    }
  }
  return DM_NULL;
}

/* RESOURCE DM_cstring_to_predefined_class
 * Converts a C string into a predefined class with that name. Return
 * DM_NULL if not found.
 */

DM_Class*
DM_cstring_to_predefined_class (DM_Context *context,
                                char *s) {
  DM_ImmutableVector *predefined_classes =
    (DM_ImmutableVector*) DM_ReferenceTo(DM_predefined_classes);
  DM_nat3 i;

  for ( i=1 ; i<predefined_classes->item_length ; i++ ) {
    DM_Class *class = (DM_Class*)(predefined_classes->item[i]);
    
    if ( DM_IsA(context, class, Class) ) {
      char *cname = DM_predefined_name(context, (DM_Object*)class);

      if ( strcmp(s, cname) ) {
        continue;
      } else {
        return class;
      }
    }
  }
  return DM_NULL;
}

/* DOCUMENTATION nat4 numbers
 * Provide some library functions to manipulate nat4 numbers (fortunately
 * DMeroon does not require a full-blown set of nat4 primitives).
 * Nat4 numbers are defined as as structure of two nat3 numbers.
 * The second argument is not a nat4 number.
 */

DM_nat4
DM_nat4_add (DM_nat4 a, DM_nat3 b) {

#ifdef DM_HAS_LONG_LONG_ENTITIES
  return a+b;
#else
  DM_nat4 result = a;
  DM_nat3 r      = a._low + b;

  result._low = r;
  if ( r < b ) {
    result._high++;
  }
  return result;
#endif

}

DM_nat4
DM_nat4_subtract (DM_nat4 a, DM_nat3 b) {

#ifdef DM_HAS_LONG_LONG_ENTITIES
  return a-b;
#else
  DM_nat4 result = a;
  DM_nat3 r      = a._low - b;

  result._low = r;
  if ( r > b ) {
    result._high--;
  }
  return result;
#endif

}

DM_nat4
DM_nat4_multiply (DM_nat4 a, DM_nat3 b) {

#ifdef DM_HAS_LONG_LONG_ENTITIES
  return a*b;
#else
  DM_nat4 result = a;
  DM_nat4 result2;
  DM_nat3 r;

  if ( b == 0 ) {
    result._high = 0;
    result._low  = 0;
    return result;
  } else {
  restart:
    if ( b%2 == 0 ) {
      r = 2 * result._low;
      result._high = 2 * result._high;
      if ( r < result._low ) {
        result._high++;
      }
      result._low = r;
      b = b/2;
      goto restart;
    } else {
      if ( b == 1 ) {
        return result;
      } else {
        result2 = DM_nat4_multiply(result, b-1);
      } 
      result2._high += result._high;
      r = result2._low + result._low;
      if ( r < result2._low ) {
        result2._high++;
      }
      result2._low = r;
      return result2;
    }
  }
#endif

}

/* RESOURCE DM_nat4_last_nibble
 * Returns the last nibble of a nat4 number. In the same time divide that
 * nat4 number by 16. CAUTION: the nat4 number is modified in place.
 */

DM_nat1
DM_nat4_last_nibble (DM_nat4 *a) {

#ifdef DM_HAS_LONG_LONG_ENTITIES
  DM_nat4 temp = *a;

  *a = temp/((DM_nat4) 16);
  return temp % ((DM_nat4) 16);
#else
  DM_nat3 high = a->_high;
  DM_nat3 low  = a->_low;

  a->_high = high>>4;
  a->_low  = (low>>4) | ((high & 0xff)<<28);
  return low & 0xff;
#endif

}


/* RESOURCE DM_string_to_nat3
 * Convert a string into a nat3 number. Numbers are decimal unless prefixed
 * by 0 or 0x or 0X which means octal, hexa or hexa resp. The third parameter
 * is left on the first non recognized character.
 */

DM_nat3
DM_string_to_nat3 (DM_Context *context,
                   char       *p,
                   char       **q) {
  unsigned int base = 10;
  DM_nat3      address = 0;
  char         c;

  if ( *p == '0' ) {
    p++;
    if ( ( *p == 'x' ) || ( *p == 'X' ) ) {
      p++; base = 16;
    } else {
      base = 8;
    }
  }

  while ( (c = *p) ) {
    if ( ( '0' <= c ) && ( c <= '7' ) ) {
      p++; 
      address = address*base + c - '0';
    } else if ( (base!=8) && ( '8' <= c ) && ( c <= '9' ) ) {
      p++; 
      address = address*base + c - '0';
    } else if ( (base==16) && ( 'a' <= c ) && ( c <= 'f' ) ) {
      p++; 
      address = address*base + c - 'a' + 10;
    } else if ( (base==16) && ( 'A' <= c ) && ( c <= 'F' ) ) {
      p++; 
      address = address*base + c - 'A' + 10;
    } else {
      break;
    }
  }
  if ( q ) {
    *q = p;
  }
  return address;
}

void
DM_nat4_to_string (DM_Context *context,
                   DM_nat4    number,
                   void       *output_port) {
  char    line[17];
  char    *p     = line + 16;
  DM_nat1 nibble;
  DM_nat2 i;

  *p = '\0';
  for ( i=0; i<16; i++ ) {
    nibble = DM_nat4_last_nibble(&number);
    *(--p) = nibbles2characters[nibble];
  }
  DM_print_char_star(context, line, 16, output_port);
}

/******************************************************************
 *                Managing Symbols for Scheme-like languages.
 *
 * These functions only manage a hashtable mapping char[] into Symbols. 
 * It is up to the language to use one such table or more not excluding
 * other variants.
 *
 * The hash function used to hash strings is that of P J Weinberger as
 * cited in the Dragoon book from Aho, Sethi and Ullman [page 436].
 * I modified it so it can be used on strings not terminated by NUL.
 */

DM_nat3 
DM_hashpjw (DM_nat3 length,
            char*   string )
{
  char *p;
  unsigned h = 0, g;
  
  for ( p=string ; length-->0 ; p++ ) {
    h = (h<<4) + (*p);
    g = h & 0xf0000000;
    if ( g ) {
      h = h^(g>>24);
      h = h^g;
    }
  } 
  return h;
}

DM_nat3
DM_compute_hash_symbol (DM_Context *context,
                        DM_Symbol  *symbol) {
  return DM_hashpjw(symbol->character_length, symbol->character);
}

static DM_Symbol*
DM_extract_symbol (DM_Context *context,
                   DM_nat3    length,
                   char       *cstring ) 
{
  DM_Site *site = DM_TheCurrentSite();

  DM_HashtableExtract(
    context,  
    site->oblist,
    DM_Symbol*,
    DM_hashpjw(length, cstring),
    (    (length == object->character_length)
      && (strncmp(cstring, object->character, length) == 0) )
    );
}

/* RESOURCE DM_intern_symbol
 * Intern a symbol in the local oblist. The pname may contain NUL
 * characters, it may be not finished by a NUL (a la C) as well.
 */

DM_Symbol*
DM_intern_symbol (DM_Context *context,
                  DM_Symbol  *symbol )
{
  DM_Site   *site = DM_TheCurrentSite();
  DM_Symbol *s;
  
  s = DM_extract_symbol(context, 
                        symbol->character_length,
                        symbol->character );
  if ( s ) {
    return s;
  } else {
    DM_hashtable_insert(context, &(site->oblist), (DM_Object*)symbol);
    return symbol;
  }
}

/******************************************************************
 * Various conversion routines.
 */

/* 
 * This routine converts an integer into a Field with that rank.
 * It returns DM_NULL if no such field is found.
 */

DM_Field*
DM_number_to_field (DM_Context *context,
                    DM_Class *class,
                    DM_nat3  number ) {
  DM_nat3 fn = DM_class_indexed_field_length(context, class);

  if ( number < fn )
    return DM_class_field(context, class, number);
  return DM_NULL;
}

/*
 * This one converts a C string into a Field.
 */

DM_Field*
DM_cstring_to_field (DM_Context *context,
                     DM_Class *class,
                     char     *string ) {
  DM_nat3 fn = DM_class_indexed_field_length(context, class);
  DM_nat3 i;

  for ( i=0 ; i<fn ; i++ ) {
    DM_Field *field = DM_class_field(context, class, i);

    if ( !( strcmp(string, DM_predefined_name(context, (DM_Object*)field)) ) ) {
      return field;
    }
  }
  return DM_NULL;
}

/*
 * This one converts a DMeroon String into a Field.
 */

DM_Field*
DM_string_to_field (DM_Context *context,
                    DM_Class  *class,
                    DM_String *string ) {
  return DM_cstring_to_field(context, class, string->character);
}

/* RESOURCE DM_index_to_type
 * Converts an index (a nat1) into a type with that number.  Pay
 * attention that the Type instance may exist unimplemented.  This can
 * be checked by the zeroness of the alignment field.  Underlying C
 * types can also be accessed (with indexes close to 255).
 */

DM_Type*
DM_index_to_type (DM_Context *context,
                  DM_nat1 index) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();

  /* It is possible that the C representation of a DM_nat1 be wider than
   * [0..255] so check it. */
  if ( index >= 256 ) {
      return DM_NULL;
  } else {
    return (DM_Type*)implemented_types->item[index];
  }
}

/* RESOURCE dm_index_to_predefined_class
 * Converts an index into a predefined class with that number. Return
 * DM_NULL otherwise.
 */

DM_Class*
DM_index_to_predefined_class (DM_Context *context,
                              DM_nat1 index) {
  DM_PredefinedClassesVector *predefined_classes = DM_ThePredefinedClasses();

  /* It is possible that the C representation of a DM_nat1 be wider
   * than [0..255] so check it. The first predefined class (Object)
   * has number 1. */
  if ( (index < 1) || (index >= predefined_classes->item_length) ) {
      return DM_NULL;
  } else {
    return predefined_classes->item[index];
  }
}

/* RESOURCE DM_remove_last_queuelink
 * Returns and removes the last object of a queue. Returns DM_NULL if
 * no such object exist. The argument is the address of a location 
 * referring to a QueueLink. This location may be modified if containing 
 * a single object.
 */

DM_Object*
DM_remove_last_queuelink (DM_Context*   context,
                          DM_QueueLink* *location ) {
  if ( *location ) {
    void *o = *location;

    if ( DM_IsA(context, o, QueueLink) ) {
      DM_DeclareNarrowedObject(queuelink, QueueLink, o);

      if ( DM_IsA(context, queuelink->rest, QueueLink) ) {
        DM_QueueLink *previous = queuelink;

        queuelink = queuelink->rest;
        while ( DM_IsA(context, queuelink->rest, QueueLink) ) {
          previous  = queuelink;
          queuelink = queuelink->rest;
        }
        previous->rest = DM_NULL;
        return queuelink->value;
      } else {
        /* Patch the location directly since it contains a list of
         * exactly one QueueLink. */
        *location = DM_NULL;
        return queuelink->value;
      }
    } else {
      /* The content of location was not a QueueLink. */
      DM_MayAbort(context, DME_NOT_QUEUELINK);
      return DM_NULL;
    }
  } else {
    /* The queue was empty. */
    return DM_NULL;
  }
}

/* RESOURCE DM_enqueue_received_object
 * Enqueue an object onto the local site. Don't resume immediately 
 * waiting threads.
 */

void
DM_enqueue_received_object (DM_Context *context,
                            DM_Object  *o ) {
  DM_Site      *site      = DM_TheCurrentSite();
  DM_QueueLink *queuelink = DM_Allocate0(context, QueueLink);

  /* No need to check whether o is an Exit or a real object. It
   * may be safely written in a field provided QueueLinks are
   * read with the API. */
  queuelink->value = o;
  queuelink->rest  = site->objects;
  site->objects    = queuelink;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, queuelink);
  }
}

/* RESOURCE DM_enqueue_received_request_or_answer
 * Enqueue a request or an answer onto the local site. They will be
 * handled later on (after decoding the message). 
 */

void
DM_enqueue_received_request_or_answer (DM_Context *context,
                                       DM_Object  *o )
{
  DM_Site      *site      = DM_TheCurrentSite();
  DM_QueueLink *queuelink = DM_Allocate0(context, QueueLink);

  /* No need to check whether o is an Exit or a real object. It
   * may be safely written in a field provided QueueLinks are
   * read with the API. This is not currently not true    FUTURE */
  queuelink->value = o;
  queuelink->rest  = site->arrived;
  site->arrived    = queuelink;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, queuelink);
  }
}

/******************************************************************
 * Handle Files.
 */

/* RESOURCE DM_load_file
 * Read a local file and build a File instance representing it.
 */

DM_File*
DM_load_file (DM_Context *context,
              char       *filename )
{
  /* FUTURE                      TO BE DONE */
  return DM_NULL;
}

/* RESOURCE DM_store_file
 * write a File instance into a local file.
 */

DM_c_boolean
DM_store_file (DM_Context *context,
               DM_File    *file,
               char       *dirpath )
{
  /* FUTURE                      TO BE DONE */
  return DM_TRUE;
}

/******************************************************************
 * As any other modules there is an initialization routine.
 */

void
DM_extr_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmextr);
}

/* end of dmextr.c */
