/* $Id: dmhash.c,v 1.21 1998/05/25 19:01:26 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 *
 * DOCUMENTATION dmhash.c
 * This file defines a set of simplistic functions to manage hash-tables.
 * It uses Knuth Vol3, page 519 and followings. (linear probing). 
 *
 * Objects may be inserted (doubling the size of the hashtable if needed),
 * objects may also be removed. Extraction of object is not offered since
 * it depends too much on how to compare objects; nevertheless a macro 
 * allows to simplify its writing (see dmhash.h file).
 *
 * The interface is quite uniform, all hashtables are given as the
 * address of a location holding an Hashtable. These functions are low
 * level; they assume hashtable to be reachable by a single location so
 * they may be doubled without indirection.
 *
 * The Hashtable is initially filled with DM_NULL. This is nice since
 * the allocator guarantees freshly allocated objects to be zero-filled.
 *
 *.FIGURE Hashtable
 *          +---------+
 *          |         |               a Hashtable
 * address: |     ----------X-------> +----------+
 *          |         |     |         |          |
 *          +---------+     |             .... 
 *                          |         |          |
 *                          |         +----------+
 *                          V
 *                          Doubled Hashtable
 *                          +--------+
 *                          |        |
 *
 *.ENDFIGURE
 *
 * A Hashtable contains a hashing function that takes an Object and
 * returns a DM_nat3 number. Empty entries hold DM_NULL.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmhash,
  "$Id: dmhash.c,v 1.21 1998/05/25 19:01:26 queinnec Exp $");

/* {{{ Inserting
 *
 * RESOURCE DM_hashtable_insert
 * This function adds a DMeroon object to a hashtable. 
 * It assumes that the hashtable is not full. After insertion, 
 * the size of the Hashtable is doubled if sufficiently dense.
 */

void
DM_hashtable_insert (DM_Context    *context,
                     DM_Hashtable* *hashtable_address,
                     DM_Object     *object ) {
  DM_Hashtable *hashtable = *hashtable_address;
  DM_nat3       hsize     = hashtable->item_length;
  DM_DeclareHasherFunction(hasher, hashtable->hasher);
  DM_DeclareExtractorFunction(extractor, hashtable->extractor);
  DM_nat3       h         = hsize;
  DM_nat3       hk        = ((*hasher)(context, object)) % hsize;
  DM_Object     *o;

  while ( h-- > 0 ) {
    if ( extractor ) {
      o = (*extractor)(context, &(hashtable->item[hk]));
    } else {
      o = hashtable->item[hk];
    }
    if ( o ) {

      DM_DebugAtLevel(DM_DEBUG_HASHTABLE,
        fprintf(stderr, "Skip non-empty " DM_0X "%p[%lu]\n", 
                hashtable, (unsigned long)hk);
);

      hk = ( hk > 0 ) ? hk-1 : hsize-1;
    } else {
      if ( extractor ) {
        hashtable->item[hk] = (*extractor)(context, &object);
      } else {
        hashtable->item[hk] = object;
      }
      hashtable->size += 1;

      DM_DebugAtLevel(DM_DEBUG_HASHTABLE,
        fprintf(stderr, "Inserted Object " DM_0X "%p in " DM_0X "%p[%lu]\n", 
                object, hashtable, (unsigned long)hk);
);

      /* Double the hashtable if its density is over 75%. */
      if ( 4*hashtable->size >= 3*hashtable->item_length )
        DM_hashtable_double(context, hashtable_address);
      return;
    }
  }
  /* Arriving here means that the hashtable is completely full. 
   * Do not raise an error but just return false. */
  DM_Abort(context, DME_HASHTABLE_OVERFLOW);
  return;
}

/* }}}
 * {{{ Doubling
 *
 * RESOURCE DM_hashtable_double
 * Replace a Hashtable by a new one with a doubled size. The given
 * location is patched with the new Hashtable.
 */

void
DM_hashtable_double (DM_Context    *context,
                     DM_Hashtable* *hashtable_address) {
  DM_Hashtable *old = *hashtable_address;
  DM_DeclareExtractorFunction(extractor, old->extractor);
  DM_Hashtable *new = DM_Allocate1(context, Hashtable, 2 * old->item_length);
  DM_Object    *object;
  DM_nat3      i;

  DM_DebugAtLevel(DM_DEBUG_HASHTABLE,
    fprintf(DM_FindErrorStream(context),
            "Doubling hashtable " DM_0X "%p into "
            DM_0X "%p ...\n", old, new);
);

  new->hasher = old->hasher;
  new->extractor = old->extractor;
  new->next   = old->next;
  new->size   = 0;

  /* Copy the old table in the new one. */
  for ( i=0 ; i<old->item_length ; i++ ) {
    if ( extractor ) {
      object = (*extractor)(context, &(old->item[i]));
    } else {
      object = old->item[i];
    }
    if ( object ) {
      DM_hashtable_insert(context, &new, object);
    } else {
      /* Nothing: this position is empty. */
    }
  }

  DM_DebugAtLevel(DM_DEBUG_HASHTABLE,
    fprintf(DM_FindErrorStream(context), 
            "Doubled hashtable " DM_0X "%p into "
            DM_0X "%p\n", old, new);
);

  /* replace the old hashtable. */
  *hashtable_address = new;
  /* All internal hashtables are copyable. This avoids errors when
   * externalizing them for debug. */
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, new);
  }
}

/* }}}
 * {{{ Removing
 *
 * RESOURCE DM_hashtable_remove
 * Remove an object from an Hashtable.
 * Currently the size of the Hashtable is not reduced even if it becomes
 * sparse (but the interface tolerates that feature).
 */

void
DM_hashtable_remove (DM_Context    *context,
                     DM_Hashtable* *hashtable_address,
                     DM_Object     *object ) {
  DM_Hashtable *hashtable = *hashtable_address;
  DM_nat3      hsize      = hashtable->item_length;
  DM_DeclareHasherFunction(hasher, hashtable->hasher);
  DM_DeclareExtractorFunction(extractor, hashtable->extractor);
  DM_nat3      h          = hsize;
  DM_nat3      h1         = hsize;
  DM_nat3      hk         = ((*hasher)(context, object)) % hsize;
  DM_nat3      last_hk;
  DM_nat3      he;
  DM_Object    *o;
  
  while ( h-- > 0 ) {
    if ( extractor ) {
      o = (*extractor)(context, &(hashtable->item[hk]));
    } else {
      o = hashtable->item[hk];
    }
    if ( o ) {
      if ( o != object ) {   
        hk = ( hk > 0 ) ? hk-1 : hsize-1;
      } else {
        hashtable->size -= 1;

        DM_DebugAtLevel(DM_DEBUG_HASHTABLE,
          fprintf(DM_FindErrorStream(context), 
                  "Removing Object " DM_0X "%p from " DM_0X "%p[%lu]\n",
                  object, hashtable, (unsigned long)hk );
);

        while ( h1-- > 0 ) {
          /* erase position hk, memorize it as last_hk */
          hashtable->item[hk] = DM_NULL;
          last_hk = hk;
        R2:
          hk = ( hk > 0 ) ? hk-1 : hsize-1;
          if ( extractor ) {
            o = (*extractor)(context, &(hashtable->item[hk]));
          } else {
            o = hashtable->item[hk];
          }
          if ( o ) {
            he = ((*hasher)(context, o)) % hsize;
            if (   ( (hk <= he) && (he < last_hk) )
                || ( (he < last_hk ) && (last_hk < hk) )
                || ( (last_hk < hk ) && (hk <= he) ) ) {
              goto R2;
            } else {
              if ( extractor ) {
                hashtable->item[last_hk] = (*extractor)(context, &o);
              } else {
                hashtable->item[last_hk] = o;
              }
            }
          } else {
            return;
          }
        }
        return;
      }
    } else {
      /* Cannot remove an absent object. */
      DM_Abort(context, DME_NOT_IN_HASHTABLE);
      return;
    }
  }
  /* Cannot remove an absent object. */
  DM_Abort(context, DME_NOT_IN_HASHTABLE);
  return;
}
  
/* }}}
 * DOCUMENTATION 
 * Extracting objects (see dmhash.h file).

 * Initializing.
 */

void
DM_hash_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmhash);
}

/* end of dmhash.c */
