/* $Id: dmhtml.c,v 1.97 1998/06/01 19:30:43 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmhtml.c
 * This file contains the html server that allows to inspect DMeroon
 * object using html clients. The main function answers http requests
 * (the 'GET' protocol). There exists a tiny language (the url
 * scripting language) that allows to perform actions on DMeroon
 * objects. Http requests are in fact one-liner url scripting expressions.
 * The post protocol, the way mutable objects are displayed was first studied
 * by Jing Wang (Summer 1997).

 * Printing objects in html requires the presence of some basic UrlActions.
 * These are: site, entry, fancyprint, entry, fetch

 * TODO: verify that nowhere is printed a null char.
 * These functions are too long and duplicate many code.   FUTURE
 */

#include "dmeroon.h"

DM_DefineVersionString(dmhtml,
  "$Id: dmhtml.c,v 1.97 1998/06/01 19:30:43 queinnec Exp $");

/*  {{{ HTML utilities
 
 * These are the main functions to print objects according to three
 * possible modes: 
 *   DM_html_byte_print 
 *   DM_html_fancy_print
 *   DM_html_object_print
 * 
 * Utilities are:
 *       DM_html_print_object_description
 *       DM_html_print_anchor
 *       DM_html_print_mutable_field
 */

static void
DM_html_print_anchor (DM_Context *context,
                      DM_Object  *o,
                      char       *label,
                      void       *output_port);

static void
DM_html_print_address (DM_Context *context,
                       DM_Object  *o,
                       void       *output_port);

static void
DM_html_do_print_mutable_field (DM_Context     *context,
                                DM_DMeroonType *itemtyp,
                                DM_Object      *o,
                                DM_translation t,
                                DM_Field       *field,
                                unsigned int   size,
                                unsigned int   length,
                                void           *output_port) 
{
  char *fieldname = DM_predefined_name(context, (DM_Object*)field);
  DM_Site *site = DM_TheCurrentSite();
  char s[128];
  
  DM_print_char_star(context, " <INPUT NAME=\"", 0, output_port);
  DM_print_char_star(context, fieldname, 0, output_port);
  DM_print_char_star(context, "\" VALUE=\"", 0, output_port);
  if ( itemtyp->index == DM_reference_index ) {
    sprintf(s, "http://%s:%lu/", 
            site->name, (unsigned long)(ntohs(site->port)) );
    DM_print_char_star(context, s, 0, output_port);
    DM_html_print_address(context, o, output_port);
  } else {
    DM_local_print(context, itemtyp, o, t, output_port);
  }
  sprintf(s, "\" SIZE=%u MAXLENGTH=%u >\r\n", size, length);
  DM_print_char_star(context, s, 0, output_port);
}

static void
DM_html_print_mutable_field (DM_Context     *context,
                             DM_DMeroonType *itemtyp,
                             DM_Object      *o,
                             DM_translation t,
                             DM_Field       *field,
                             void           *output_port )
{ unsigned int size   = 0;
  unsigned int length = 0;

  /* Give the possibility to follow the link simply.  Choose a visual
   * symbol to click on. (suggestion for a better one ?) */
  if ( itemtyp->index == DM_reference_index ) {
    DM_html_print_anchor(context, o, 

#if defined(DM_DMEROON_PREFERS_TEXT)
       "<FONT color=green>&nbsp;&Ouml;&nbsp;</FONT>", 
#else
       "<img src=\"icons/object.gif\" width=18 height=18 alt=\"[an object]\">",
#endif

                         output_port);
  }

  switch ( itemtyp->index ) {
  case DM_nought_index: DM_Abort(context, DME_CANNOT_PRINT_NOUGHT);
    /* FUTURE: this may be coded as a field in itemtyp ???    FUTURE */
  case DM_nat1_index:    size = 4; break;
  case DM_nat2_index:    size = 6; break;
  case DM_nat3_index:    size = 10; break;
  case DM_nat4_index:    size = 20; break;
  case DM_int1_index:    size = 5; break;
  case DM_int2_index:    size = 7; break;
  case DM_int3_index:    size = 11; break;
  case DM_int4_index:    size = 21; break;
  case DM_date_index:    size = 24; break;
  case DM_reference_index: size=92; break;
  case DM_data_index:    size = 10; break;
  case DM_code_index:    size = 10; break;
  case DM_char1_index:   size = 1; break;
  case DM_float1_index:  size = 40; break;
  case DM_float2_index:  size = 40; break;
  case DM_float3_index:  size = 60; break;
  case DM_netnat2_index: size = 6; break;
  case DM_netnat3_index: size = 10; break;
  default:               size = 60; break;
  }
  length = size;
  DM_html_do_print_mutable_field(context, itemtyp, o, t, field,
                                 size, length, output_port );
}

static void
DM_html_key_print (DM_Context *context,
                   DM_key     key,
                   void       *output_port ) {
  char line[80];

  sprintf(line, "0x%04lx%04lx/0x%08lx/0x%08lx/0x%08lx",
          (unsigned long) key._idx,
          (unsigned long) key._port,
          (unsigned long) key._ip,
          (unsigned long) key._date,
          (unsigned long) key._rand );
  DM_print_char_star(context, line, 0, output_port);
}

/* RESOURCE DM_html_print_address
 * Displays an url script to find an object. If the object is
 * externalized, it will be reached with an <external-number>/entry.object
 * otherwise with 0x<address>/address. It is not guaranteed that the
 * object will stay at the same address (because of the GC) so this
 * URL may become dangling later.

 * NOTE: The url script produced does not print in any way the object.
 */

static void
DM_html_print_address (DM_Context *context,
                       DM_Object  *o,
                       void       *output_port) {
  if ( o ) {
    DM_DeclareNarrowedObject(entry, Entry, DM_externalize(context, o));

    if ( o == (DM_Object*)DM_TheCurrentSite() ) {
      /* If this is the current site, refer to it with the 'site'
       * command directly. This is more readable. */
      DM_print_char_star(context, "thesite", 7, output_port);
    } else {
      /* Refer to the object by its Entry. */
      DM_html_key_print(context, entry->number, output_port);
      DM_print_char_star(context, "/entry.object", 0, output_port);
    }
  } else {
    DM_print_char_star(context, "null", 4, output_port);
  }
}

/* RESOURCE DM_print_class_mention
 * Mention the class name if possible.
 */

static void
DM_print_class_mention (DM_Context *context, 
                        DM_Class   *class,
                        void       *output_port )
{
  if ( class ) {
    DM_print_char_star(context, "(a <B>", 0, output_port);
    DM_print_char_star(context, 
                       DM_predefined_name(context, (DM_Object*)class),
                       0,
                       output_port );
    DM_print_char_star(context, "</B>) ", 0, output_port);
  }
}

/* RESOURCE DM_html_print_exit_address 
 * Print anchors for Exit items. To follow these anchors will lead to the
 * remote machines. This function only exists to factor code within
 * DM_html_print_anchor.
 */

static void
DM_html_print_exit_address (DM_Context *context,
                            DM_Site    *site,
                            DM_key     number,
                            DM_Class   *class,
                            char       *label,
                            void       *output_port ) {
  char line[100];

  DM_print_char_star(context, "<A href=\"http://", 0, output_port);
  /* Use IP number instead of the name of the site to mark the debug mode. */
  DM_netnat3_print(context, &(site->address), output_port);
  sprintf(line, ":%lu/", (unsigned long)(ntohs(site->port)));
  DM_print_char_star(context, line, 0, output_port); 
  DM_html_key_print(context, number, output_port);
  sprintf(line, "/entry.object/fancyprint\"> ");
  DM_print_char_star(context, line, 0, output_port);
  DM_print_class_mention(context, class, output_port);
  if ( label ) {
    sprintf(line, "[%s] ", label);
    DM_print_char_star(context, line, 0, output_port);
  }
  DM_print_char_star(context, "</A>\r\n", 6, output_port);
}
  
/* RESOURCE DM_html_print_anchor
 * Prints an html anchor and a label towards an object. Try to devise sensible
 * anchors and labels for most objects. This function is called to display
 * the reference-fields of objects.
 *
 * Predefined constants are named, classes are referenced with their names,
 * externalized objects with their external number and other uninitialized
 * objects with their (potentially dangling) address. Small strings are
 * also handled specially.
 *
 * Exit items are not forced to be local to be displayed, they are rather
 * displayed with a double URL to their from and owner counterparts. A third
 * URL allows to see their inner details (with possible actions on them).
 *
 * The 'label' argument may indicate a name rather than letting this
 * function chooses one.
 */

static void
DM_html_print_anchor (DM_Context *context,
                      DM_Object  *o,
                      char       *label,
                      void       *output_port) {
  DM_Site *site = DM_TheCurrentSite();
  char    line[100];

  if ( o ) {
    DM_Class *class = DM_ObjectToClass(context, o);

    if ( DM_IsA(context, o, Exit) ) {
      /* The object is an Exit item, prints some anchors to access it. 
       * Anchors allow to access it on the from or owner site as well 
       * as its current state on the current site. */
      DM_DeclareNarrowedObject(exit, Exit, o);
    
      DM_html_print_exit_address(context, 
                                 exit->from, 
                                 exit->number,
                                 exit->class,
                                 "from",
                                 output_port );
      DM_print_char_star(context, "\r\n", 2, output_port);
      DM_html_print_exit_address(context, 
                                 exit->owner, 
                                 exit->number,
                                 exit->class,
                                 "owner",
                                 output_port );
      sprintf(line, "<A href=\"http://%s:%lu/",
              site->name, (unsigned long)(ntohs(site->port)) );
      DM_print_char_star(context, line, 0, output_port);
      if ( DM_HasNotOption(exit, DM_EXTERNAL_IS_EXTERNALIZED) ) {
        DM_register_entry(context, (DM_External*)exit);
      }
      DM_html_key_print(context, exit->number, output_port);
      DM_print_char_star(context, 
                         "/entry/fancyprint\"> "
                         "<EM>(Exit item)</EM></A>\r\n", 
                         0, output_port );
    } else {
      /* The object is not an Exit item, print the anchor. */
      sprintf(line, "<A href=\"http://%s:%lu/",
              site->name, (unsigned long)(ntohs(site->port)) );
      DM_print_char_star(context, line, 0, output_port);
      DM_html_print_address(context, o, output_port);
      DM_print_char_star(context, "/fancyprint\">\r\n", 0, output_port);

      /* Use the suggested name if present. */
      if ( label ) {
        DM_print_char_star(context, label, 0, output_port);
      } else {
        /* Now, find a reasonable label for that anchor. */
        if ( o == DM_TheConstant(true) ) {
          DM_print_char_star(context, "<B>true</B>", 0, output_port);
        } else if ( o == DM_TheConstant(false) ) {
          DM_print_char_star(context, "<B>false</B>", 0, output_port);
        } else if ( o == DM_TheConstant(nil) ) {
          DM_print_char_star(context, "<B>nil</B>", 0, output_port);
        } else if ( o == DM_TheConstant(uninitialized) ) {
          DM_print_char_star(context, "<B>uninitialized</B>", 0, output_port);
        } else if ( o == DM_TheConstant(missing) ) {
          DM_print_char_star(context, "<B>missing</B>", 0, output_port);
        } else if ( o == (DM_Object*)DM_TheCurrentSite() ) {
          DM_print_char_star(context, "<B>site</B>", 0, output_port);
        } else {
          DM_Entry *entry = DM_externalize(context, o);
    
          if ( DM_IsADirect(context, o, Class) ) {
            DM_DeclareNarrowedObject(class, Class, o);

            DM_print_char_star(context, "<B>", 3, output_port);
            DM_print_char_star(context, 
                               DM_predefined_name(context, (DM_Object*)class),
                               0,
                               output_port );
            DM_print_char_star(context, "</B>", 4, output_port);
          } else if (   DM_IsADirect(context, o, DMeroonType) 
                        || DM_IsADirect(context, o, CType)  ) {
            DM_DeclareNarrowedObject(type, Type, o);
            
            DM_print_char_star(context, "<B>", 3, output_port);
            DM_print_char_star(context, 
                               DM_predefined_name(context, (DM_Object*)type), 
                               0,
                               output_port );
            DM_print_char_star(context, "</B>", 4, output_port);
          } else if ( DM_IsADirect(context, o, String) ) {
            DM_String *string = (DM_String*)o;
            
            if ( string->character_length > DM_MAX_CHARS_FOR_IMMEDIATE_DISPLAY )
              goto regular_reference_anchor;
            /* Display the content of the String. */
            DM_print_char_star(context, "(a <B>String</B>) \"<TT>", 
                               0, output_port);
            /* NOTE: attention mosaic is touchy wrt to \0 characters. */
            DM_print_string(context, string, output_port);
            DM_print_char_star(context, "</TT>\"\r\n", 0, output_port);
          } else {
            /* The object is not predefined, has no sensible short nickname,
             * then we refer to it via its external number. */
          regular_reference_anchor:
            DM_print_class_mention(context, class, output_port);
            DM_html_key_print(context, entry->number, output_port);
          }
        }
      }

      /* Close the label. */
      DM_print_char_star(context, "</A>", 4, output_port);
    }
  } else {
    DM_print_char_star(context, "<B>null</B>", 0, output_port);
  }
}

/* RESOURCE DM_html_print_object_description
 * Displays a short description of an object with its main characteristics
 * ie mutable/immutable, remote/local uptodate/obsolete.
 */

static void
DM_html_print_object_description (DM_Context *context,
                                  DM_Object *o,
                                  void      *output_port) {
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Object *proxy = DM_ObjectToProxy(context, o);

  DM_print_char_star(context, "<P>", 3, output_port);

  if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) )  {
    DM_print_char_star(context, "Mutable ", 0, output_port);
  } else {
    DM_print_char_star(context, "Immutable ", 0, output_port);
  }

  if ( DM_IsA(context, proxy, Entry) ) {
    DM_DeclareNarrowedObject(entry, Entry, proxy);

    if ( DM_IsA(context, proxy, Exit) ) {
      DM_print_char_star(context, "remote ", 0, output_port);
      if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) )  {
        if ( DM_is_up_to_date(context, o) ) {
          DM_print_char_star(context, "uptodate ", 0, output_port);
        } else {
          DM_print_char_star(context, "obsolete ", 0, output_port);
        }
      } else {
        /* The object is immutable do always up to date. */
      }
    } else {
      DM_print_char_star(context, "local ", 0, output_port);
    }
    if ( DM_HasOption(entry, DM_ENTRY_IS_UNINITIALIZED) ) {
      DM_print_char_star(context, "uninitialized ", 0, output_port);
    } else if ( DM_HasOption(entry, DM_ENTRY_IS_COPYABLE) ) {
      DM_print_char_star(context, "copyable ", 0, output_port);
    } else {
      DM_print_char_star(context, "sharable ", 0, output_port);
    }
  } else {
    if ( proxy == DM_TheConstant(uninitialized) ) {
      DM_print_char_star(context, "uninitialized ", 0, output_port);
    } else if ( proxy == DM_TheConstant(true) ) {
      DM_print_char_star(context, "copyable ", 0, output_port);
    } else {
      DM_print_char_star(context, "sharable ", 0, output_port);
    }
  }

  DM_print_char_star(context, "instance of ", 0, output_port);
  if ( DM_IsPredefinedClass(class) ) {
    DM_print_char_star(context, "predefined ", 0, output_port);
  }
  DM_print_char_star(context, "class\r\n", 0, output_port);
  DM_html_print_anchor(context, (DM_Object*)class, NULL, output_port);
  DM_print_char_star(context, ".\r\n", 3, output_port);
}

/* RESOURCE DM_html_other_view_print
 * This functions displays an anchor allowing alternate view on an object.
 * Alternate views are 'byte', 'object' and 'fancy'.
 */

static void
DM_html_other_view_print (DM_Context *context,
                          DM_Object *o,
                          char      *label,
                          void      *output_port ) {
  DM_Site *site = DM_TheCurrentSite();
  char    line[100];
    
  sprintf(line, " <A href=\"http://%s:%lu/",
          site->name, (unsigned long)(ntohs(site->port)) );
  DM_print_char_star(context, line, 0, output_port);
  DM_html_print_address(context, o, output_port);
  sprintf(line, "/%s\"><EM>(%s)</EM></A>\r\n", label, label);
  DM_print_char_star(context, line, 0, output_port);
}

/* RESOURCE DM_html_print_object
 * Displays the content of an object in HTML. This is the standard
 * method that displays object's fields one after one another. However 
 * it tries to handle repetition of characters in a fancier way.
 */

void
DM_html_print_object (DM_Context *context,
                      DM_Object  *o,
                      void       *output_port) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();

  if ( o ) {
    DM_Class          *class        = DM_ObjectToClass(context, o);
    DM_Object         *oo;
    DM_nat3           fn     = DM_class_indexed_field_length(context, class);
    DM_translation    t      = 0;
    DM_Field          *field;
    DM_nat3           on;
    DM_offset         last_offset;
    DM_nat1           itemtypindex, reptypindex;
    DM_DMeroonType    *itemtyp, *reptyp;
    DM_nat3           i, j, rep;
    void              *p;
    char              line[255];

    DM_html_print_object_description(context, o, output_port);

#if defined(DM_DEBUG)
    DM_html_other_view_print(context, o, "byteprint", output_port);
#endif /* defined(DM_DEBUG) */

    if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) ) {
      /* If uninitialized then fields may be modified.      FUTURE
                              || ! DM_is_initialized(context, o) */
      DM_Site *site = DM_TheCurrentSite();

      sprintf(line, "<FORM METHOD=POST ACTION=\"http://%s:%lu/",
              site->name, (long unsigned)(ntohs(site->port)) );
      DM_print_char_star(context, line, 0, output_port);
      DM_html_print_address(context, o, output_port);
      DM_print_char_star(context, "/modify\">\r\n", 0, output_port);
    }

    for ( i=0 ; i<fn ; i++ ) {
      field           = DM_class_field(context, class, i);
      on              = field->offset_length;
      last_offset     = field->offset[on-1];
      reptypindex     = DM_OffsetToRepetitionTypeIndex(last_offset);
      reptyp          = implemented_types->item[reptypindex];
      itemtypindex    = DM_OffsetToItemTypeIndex(last_offset);
      itemtyp         = implemented_types->item[itemtypindex];
      
      if ( reptypindex == DM_nought_index ) {
        /* The current field is a regular (not indexed) field. */
        t = DM_OffsetPad(t, itemtyp);
        if ( DM_HasOption(field, DM_FIELD_IS_SECRET) ) {
          /* Don't display secret fields. */
        } else {
          DM_print_char_star(context, "<P><B>", 6, output_port);
          DM_print_char_star(context, 
                             DM_predefined_name(context, (DM_Object*)field),
                             0,
                             output_port );
          DM_print_char_star(context, "</B>(", 5, output_port);
          DM_print_char_star(context, 
                             DM_predefined_name(context, (DM_Object*)itemtyp),
                             0,
                             output_port );
          DM_print_char_star(context, "):\n", 3, output_port);
          /* Display the value (handle specially references). */
          if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
            /* The field is mutable, display it in a box with the post
             * protocol except if a reference. */
            if ( itemtypindex == DM_reference_index ) {
              p  = DM_Displace(o, t);
              oo = DM_Content(DM_reference, p);
              if ( oo ) {
                /* shortcut references, this is more useful for the user (if
                 * not for debug). */
                DM_ShortCutReference(context, p);
              }
              DM_html_print_mutable_field(context, itemtyp, oo, t, 
                                          field, output_port );
            } else {
              /* Case of all other non reference fields. */
              DM_html_print_mutable_field(context, itemtyp, o, t, 
                                          field, output_port );
            }
            DM_print_char_star(context, "\r\n", 2, output_port);
          } else {
            /* The field is immutable. */
            if ( itemtypindex == DM_reference_index ) {
              p  = DM_Displace(o, t);
              oo = DM_Content(DM_reference, p);
              if ( oo ) {
                /* shortcut references, this is more useful for the user (if
                 * not for debug). */
                DM_ShortCutReference(context, p);
              }
              DM_html_print_anchor(context, oo, NULL, output_port);
            } else {
              /* Case of all other non reference fields. */
              DM_local_print(context, itemtyp, o, t, output_port);
            }
            DM_print_char_star(context, "\r\n", 2, output_port);
          }
        }
        t = DM_OffsetExtend(t, itemtyp);
      } else {
        /* The field is indexed. */
        t   = DM_OffsetPad(t, reptyp);
        p   = DM_Displace(o, t);
        rep = DM_read_repetition_factor(context, p, reptypindex);
        t   = DM_OffsetExtend(t, reptyp);
        t   = DM_OffsetPad(t, itemtyp);
        if ( DM_HasOption(field, DM_FIELD_IS_SECRET) ) {
          /* Don't display secret fields. */
          t += rep * (itemtyp->repsize);
        } else {
          if ( rep == 0 ) {
            /* empty indexed field */
            DM_print_char_star(context, "<P><B>", 6, output_port);
            DM_print_char_star(context, 
                               DM_predefined_name(context, (DM_Object*)field),
                               0,
                               output_port );
            DM_print_char_star(context, "</B>(", 5, output_port);
            DM_print_char_star(context, 
                               DM_predefined_name(context, (DM_Object*)itemtyp),
                               0, 
                               output_port );
            DM_print_char_star(context, ")[none]\r\n", 9, output_port);
          } else {
            /* The field is indexed with a non null number of
             * items. There is a special case for characters since it
             * is easier to view them as a string. */
            if ( itemtypindex == DM_char1_index ) {
              /* Display a repeated sequence of characters. */
              j = rep-1; /* remove the final NUL character.   HACK, FUTURE  */
              p = DM_Displace(o, t);
              DM_print_char_star(context, "<P><B>", 6, output_port);
              DM_print_char_star(context, 
                                 DM_predefined_name(context, (DM_Object*)field),
                                 0,
                                 output_port );
              DM_print_char_star(context, "</B>(", 5, output_port);
              DM_print_char_star(context, 
                                 DM_predefined_name(context, 
                                                    (DM_Object*)itemtyp),
                                 0, 
                                 output_port );
              DM_print_char_star(context, ")[0..", 5, output_port);
              DM_C_unsigned_long_int_print(context, &j, output_port);
              DM_print_char_star(context, "]: <TT>", 0, output_port);
              if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
                char *fieldname = DM_predefined_name(context, 
                                                     (DM_Object*)field );

                DM_print_char_star(context, 
                                   " <INPUT TYPE=TEXT NAME=\"", 
                                   0, output_port );
                DM_print_char_star(context, fieldname, 0, output_port);
                DM_print_char_star(context, "\" VALUE=\"", 0, output_port);
              }
              for ( j=0 ; j<rep-1 ; j++ ) {
                /* HTTP clients are often touchy with respect to \0, so
                 * make sure no NUL char is emitted. Also don't print the
                 * final NUL at the end of the String.
                 FUTURE: convert these chars into %xx http encoding.     */
                char c = *((char*)p)++;
                DM_print_char(context, (isprint(c))?(c):'?', output_port);
              }
              if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
                char s[64];

                DM_print_char_star(context, "\" ", 0, output_port);
                sprintf(s, "SIZE=%u MAXLENGTH=%u >\r\n", rep, rep);
                DM_print_char_star(context, s, 0, output_port);
              }
              DM_print_char_star(context, "</TT>\r\n", 0, output_port);
              t += rep * (itemtyp->repsize);
            } else {
              /* Display a repetition of things (which are not characters).
               * FUTURE: would be good to display compactly contiguous
               * repeated values. */
              for ( j=0 ; j<rep ; j++ ) {
                DM_print_char_star(context, "<P><B>", 6, output_port);
                DM_print_char_star(context, 
                                   DM_predefined_name(context, 
                                                      (DM_Object*)field ),
                                   0,
                                   output_port );
                DM_print_char_star(context, "</B>(", 5, output_port);
                DM_print_char_star(context, 
                                   DM_predefined_name(context, 
                                                      (DM_Object*)itemtyp ), 
                                   0,
                                   output_port );
                DM_print_char_star(context, ")[", 2, output_port);
                DM_C_unsigned_long_int_print(context, &j, output_port);
                DM_print_char_star(context, "]:\n", 3, output_port);
                /* Display the value (handle specially references). */
                if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
                  /* The field is mutable, display it in a box with the post
                   * protocol except if a reference. */
                  if ( itemtypindex == DM_reference_index ) {
                    p  = DM_Displace(o, t);
                    oo = DM_Content(DM_reference, p);
                    if ( oo ) {
                      /* shortcut references, this is more useful for the
                       * user (if not for debug) */
                      DM_ShortCutReference(context, p);
                    }
                    DM_html_print_mutable_field(context, itemtyp, oo, t, 
                                                field, output_port );
                  } else {
                    DM_html_print_mutable_field(context, itemtyp, o, t, 
                                                field, output_port );
                  }
                  DM_print_char_star(context, "\r\n", 2, output_port);
                  t += itemtyp->repsize;
                } else {
                  /* The field is immutable. */
                  if ( itemtypindex == DM_reference_index ) {
                    p  = DM_Displace(o, t);
                    oo = DM_Content(DM_reference, p);
                    if ( oo ) {
                      /* shortcut references, this is more useful for the
                       * user (if not for debug) */
                      DM_ShortCutReference(context, p);
                    }
                    DM_html_print_anchor(context, oo, NULL, output_port);
                  } else {
                    DM_local_print(context, itemtyp, o, t, output_port);
                  }
                  DM_print_char_star(context, "\r\n", 2, output_port);
                  t += itemtyp->repsize;
                }
              }
            }
          }
        }
        /* Realign after any non final indexed field. */
        t = DM_OffsetRealign(t);
      }
    }

    if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) ) {
      DM_print_char_star(context, 
                         "<HR><CENTER>\r\n"
                         "<INPUT TYPE=SUBMIT VALUE=\"Modify\">\r\n"
                         "</CENTER><HR></FORM>\r\n",
                         0, output_port );
    }

  } else {
    DM_print_char_star(context, "The <B>null</B> value.\r\n", 0, output_port);
  }
}

/* RESOURCE DM_html_print_actions
 * Prints a number of anchors to ease debug and perform some actions 
 * on objects. Offer to fetch remote objects for an Exit. Offer to tick
 * a Clock.
 */

static void
DM_html_print_actions (DM_Context *context,
                       DM_Object *o,
                       void      *output_port ) {
  DM_Site *site = DM_TheCurrentSite();
  char line[100];

  if ( DM_IsA(context, o, Exit ) ) {
    DM_DeclareNarrowedObject(exit, Exit, o);

    DM_print_char_star(context, "<P> Actions: <A href=\"http://", 
                       0, output_port);
    DM_print_char_star(context, site->name, 0, output_port);
    sprintf(line, ":%lu/", (unsigned long)(ntohs(site->port)));
    DM_print_char_star(context, line, 0, output_port);
    if ( DM_HasNotOption(exit, DM_EXTERNAL_IS_EXTERNALIZED) ) {
      DM_register_entry(context, (DM_External*)exit);
    }
    DM_html_key_print(context, exit->number, output_port);
    DM_print_char_star(context, 
                       "/entry/fetch/fancyprint\"> "
                       "Fetch</A> \r\n", 
                       0, output_port);

  } else if ( DM_IsA(context, o, Clock) ) {
    DM_Entry *entry = DM_externalize(context, o);

    DM_print_char_star(context, "<P> Actions: <A href=\"http://", 
                       0, output_port);
    DM_print_char_star(context, site->name, 0, output_port);
    sprintf(line, ":%lu/", (unsigned long)(ntohs(site->port)));
    DM_print_char_star(context, line, 0, output_port);
    DM_html_key_print(context, entry->number, output_port);
    DM_print_char_star(context, 
                       "/entry.object/dup/tick/fancyprint\"> "
                       "Tick</A>\r\n",
                       0, output_port);
  }
}

/* RESOURCE DM_html_fancy_print
 * This function is a generic function to display a DMeroon instance in
 * html. It may use different printers according to the class of the 
 * object to print.
 */

void
DM_html_fancy_print (DM_Context *context,
                     DM_Object *o,
                     void      *output_port ) {
  if ( DM_IsA(context, o, Dictionary) ) {
    /* To be done                             FUTURE */
  } 
  if ( DM_IsA(context, o, Class) ) {
    /* To be done                             FUTURE */
  } 
  
  /* Default method. Propose some actions on the object, then
   * print the object in the standard way. */
  DM_html_print_actions(context, o, output_port);
  DM_html_print_object(context, o, output_port);
}

/* RESOURCE DM_html_byte_print 
 * This function displays an object with bytes. Since this function is
 * most of the time used for debug, it allows some actions to be 
 * performed on objects.
 */

void
DM_html_byte_print (DM_Context *context,
                    DM_Object *o,
                    void      *output_port ) {
  DM_Site   *site  = DM_TheCurrentSite();
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Object *proxy = DM_ObjectToProxy(context, o);
  char line[100];

#ifdef DM_DEBUG
  DM_html_print_object_description(context, o, output_port);

  DM_html_other_view_print(context, o, "fancyprint", output_port);

  DM_html_print_actions(context, o, output_port);

  DM_print_char_star(context, "<P> Class: <A href=\"http://", 0, output_port);
  DM_print_char_star(context, site->name, 0, output_port);
  sprintf(line, ":%lu/" DM_0X "%p/address/byteprint\"> " 
          DM_0X "%p </A> aka\r\n", 
          (unsigned long)(ntohs(site->port)), 
          class,
          class );
  DM_print_char_star(context, line, 0, output_port);
  DM_html_print_anchor(context, (DM_Object*)class, NULL, output_port);
  DM_print_char_star(context, "\r\n", 2, output_port);
  
  DM_print_char_star(context, "<P> Proxy: <A href=\"http://", 0, output_port);
  DM_print_char_star(context, site->name, 0, output_port);
  sprintf(line, ":%lu/" DM_0X "%p/address/byteprint\"> " 
          DM_0X "%p </A>\r\n", 
          (unsigned long)(ntohs(site->port)), 
          proxy,
          proxy );
  DM_print_char_star(context, line, 0, output_port);
  DM_print_char_star(context, "\r\n", 2, output_port);
  
  DM_print_char_star(context, "<P> Prefix:", 0, output_port);
  DM_print_char_star(context, "<PRE>", 5, output_port);
  { void *p = DM_ToC(o);
  DM_print_memory_slice(context, p, DM_PrefixSize(),  
                        "<CODE>", "</CODE>", output_port);
  }
  DM_print_char_star(context, "</PRE>", 6, output_port);
  
  /* There's a problem here since DM_print_memory_slice prints hex numbers
   * followed by their ascii printable equivalent. These characters may
   * contain forbidden characters for HTML such as < or > etc. How to
   * prevent their interpretation ? */
  DM_print_char_star(context, "<P> Content:", 0, output_port);
  DM_print_char_star(context, "<PRE>", 5, output_port);
  DM_print_memory_slice(context, o, DM_size(context, o), 
                        "<CODE>", "</CODE>", output_port);
  DM_print_char_star(context, "</PRE>\r\n", 8, output_port);
#else
  DM_print_char_star(context, "Sorry, this is not possible on this site!", 
                     0, output_port);
#endif /* defined(DM_DEBUG) */

}

/* }}}
 * {{{ decoding URL requests

 * DOCUMENTATION decoding URL requests
 * A request is a series of words. A word may represent an address, a
 * number, a string, a character or a command. There is a series of
 * functions to decode them, they usually return NULL if they fail and
 * the rest of the request if something appropriate is
 * recognized. They may take additional arguments stipulating where to
 * store the decoded result. 
 */

static char weekday[7][4] = {
  "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"
};
static char month[12][4] = {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun", 
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" 
};

/* RESOURCE DM_ctime
 * return a correctly formated string representing the current date plus
 * some delta (expressed in seconds). It's a little like ctime except for
 * the signature and the format (RFC 1123 recommended by HTTP).
 */

static char*
DM_ctime (DM_Context *context,
          char       *line,
          char       *prefix,
          DM_nat3    delta,
          char       *suffix )
{
  DM_TIME_T t = DM_get_date(context) + delta;
  struct tm *tv;   

  /* The signature of gmtime does not seem to be thread-safe since it
   * allocates a struct tm somewhere and returns it. */
  tv = gmtime(&t);
  sprintf(line, "%s%3s, %02d %3s %4d %02d:%02d:%02d GMT%s",
          prefix,
          weekday[tv->tm_wday],
          tv->tm_mday,
          month[tv->tm_mon],
          1900 + tv->tm_year,
          tv->tm_hour,
          tv->tm_min,
          tv->tm_sec,
          suffix );
  return line;
}

/* RESOURCE DM_dump_urlscript_stack
 * Prints the URLscript stack in html. 
 */

void
DM_dump_urlscript_stack (DM_Context  *context,
                         DM_UrlState *state,
                         void        *output_port ) 
{
  unsigned long height = state->stack->index;
  char line[200];
  long int i;
  DM_Object *oo;

  if ( output_port ) {
    /* Display the content of the stack with links. */
    if ( height == 0 ) {
      DM_print_char_star(context, "<HR> Empty stack.\r\n", 0, output_port);
    } else {
      sprintf(line, "<HR> Stack (height=%lu) content:\r\n<OL start=0>",
              height );
      DM_print_char_star(context, line, 0, output_port);
      for ( i=height-1 ; i>=0 ; i-- ) {
        oo = state->stack->item[i];
        DM_print_char_star(context, "<LI>", 0, output_port);
        DM_html_print_anchor(context, oo, NULL, output_port);
        DM_print_char_star(context, "\r\n", 0, output_port);
      }
      DM_print_char_star(context, 
                         " (bottom of the stack)</OL>\r\n<HR>\r\n", 
                         0, output_port);
    }
  } else {
    /* Nothing since there is no output_port */
  }
}

/* RESOURCE DM_handle_url_script
 * Process a script and output its result on the output stream. The url
 * script language is a tiny, stack-based language. The final NUL character
 * of the script marks the end of the program. The program may extend on
 * several lines. When called from a 'GET' request, the script is truncated
 * to its first line. This function is only invoked from DM_answer_html_request.
 */

static DM_ErrorCode
DM_handle_url_script (DM_Context  *context,
                      DM_UrlState *state,
                      void        *output_port ) 
{
  /* Confine errors while interpreting the script in order to 
   * produce an error message. */
  DM_ProtectWithContext(context,
     state = DM_interpret_url_script(context, state, output_port),
           /* An error occurred. */
           state->errcode = context->errcode;
           goto produce_error_message;
);
  if ( state->pc == 0 ) {
    /* The script was empty, output the help message. */
    goto produce_help_message;
  } else {
    if ( state->errcode ) {
      goto produce_error_message;
    } else {
      /* How many objects are in the stack ? */
      switch ( state->stack->index ) {
      case 0: {
        /* Just return the errcode, print nothing. */
        return state->errcode;
      }
      case 1: {
        /* Display automagically the single object left in the stack. */
        DM_Object *arg0 = DM_pop_from_bounded_stack(context, state->stack);
        
        DM_html_print_object(context, arg0, output_port);
        DM_print_char_star(context, "\r\n", 2, output_port);
        return state->errcode;
      } 
      default: {
        /* There are unconsumed arguments left on the stack. */
        state->errcode = DME_URLSCRIPT_UNCONSUMED_STACK;
        goto produce_left_argument_message;
      }
      }
    }
  }

  /* From now, these are all error handlers. */

produce_left_argument_message:
  DM_print_char_star(context, 
                     "<P> Script ended with at least two objects in stack\r\n",
                     0, output_port);
  goto produce_error_message;
    
produce_error_message: {
    char *script = &(state->script->character[state->pc]);
    DM_nat3 script_length = strlen(script);
    DM_nat3 length = (40<script_length) ? 40 : script_length;
    char line[200];

    /* This error code is read by the urlscript Perl program. 
     * Don't change the string that wraps it! 
     * NOTE: I often have problems with directives such %40s which, in
     * fact, don't limite themselves to 40 characters. I did it by hand. */
    sprintf(line, "\r\n<!-- URLscript error %lu -->\r\n"
            "<P><HR> Error %lu occurred in script near:\r\n"
            "<PRE>&gt;&gt;&gt;&gt;&gt;", 
            (long unsigned)state->errcode,
            (long unsigned)state->errcode );
    DM_print_char_star(context, line, 0, output_port);
    DM_print_char_star(context, script, length, output_port);
    DM_print_char_star(context, 
                       "&lt;&lt;&lt;&lt;&lt;</PRE><P>\r\n"
                       "<P> Erroneous state: ", 0, output_port);
    DM_html_print_anchor(context, (DM_Object*)state, NULL, output_port);
    DM_dump_urlscript_stack(context, state, output_port);
    goto produce_help_message;
  }

produce_help_message:
  DM_print_char_star(
                     context, 
                     "\r\n<HR width=\"50%\"><HR> "
                     "<P><img src=\"icons/dmeroon.gif\" alt=\"DMeroon\"> "
                     "Server $Revision: 1.97 $ \r\n"
                     "<P> This server allows you to inspect all DMeroon \r\n"
                     "objects published by this "
                     "<A href=\"thesite/fancyprint\">site</A>. \r\n" 
                     "<P> You may even know more about <A href=\""
                     DMEROON_URL 
                     "\"> DMeroon</A> and get some help\r\n"
                     " or, simply look at <A href=\""
                     DMEROON_QUEINNEC_URL
                     "\"> me</A>. \r\n",
                     0,
                     output_port );
  return state->errcode;
}

/* }}}
 * {{{ HTTP protocol

 * RESOURCE DM_answer_gif_request
 * Returns a gif correctly wrapped with MIME conventions.
 */

static void
DM_answer_gif_request (DM_Context    *context,
                       char          *gif_name,
                       char          *gif,
                       unsigned long gif_size,
                       void          *real_output_port )
{
  char      line[500];

  DM_print_char_star(context, 
    "HTTP/1.0 200 Document follows\r\n"
    "From: " DM_RELEASE_EMAIL "\r\n"
    "Content-type: image/gif\r\n"
    "Server: DMeroon/" DM_RELEASE_LEVEL " " DM_RELEASE_DATE "\r\n"
    "Pragma: cache\r\n"
                     , 0, real_output_port);   
  DM_ctime(context, line, "Date: ", 0, "\r\n");
  DM_print_char_star(context, line, 0, real_output_port); 
  DM_ctime(context, line, "Expires: ", 365*24*60*60, "\r\n"); /* Add a year */
  DM_print_char_star(context, line, 0, real_output_port);
  sprintf(line, "Content-length: %lu\r\n\r\n", gif_size);
  DM_print_char_star(context, line, 0, real_output_port);  
  DM_print_char_star(context, gif, gif_size, real_output_port);
}

/* RESOURCE DM_answer_txt_request
 * Emits a text file correctly wrapped with MIME conventions.
 */

static void
DM_answer_txt_request (DM_Context *context,
                       char       *txt_name,
                       char       *txt,
                       void       *real_output_port )
{
  char      line[500];
  
  DM_print_char_star(context, 
        "HTTP/1.0 200 OK\r\n"
        "From: " DM_RELEASE_EMAIL "\r\n"
        "Content-type: text/plain\r\n"
        "Server: DMeroon/" DM_RELEASE_LEVEL " " DM_RELEASE_DATE "\r\n"
                     , 0, real_output_port);
  DM_ctime(context, line, "Date: ", 0, "\r\n");
  DM_print_char_star(context, line, 0, real_output_port);
  DM_ctime(context, line, "Expires: ", 365*24*60*60, "\r\n\r\n");
  DM_print_char_star(context, line, 0, real_output_port);
  DM_print_char_star(context, txt, 0, real_output_port);
}

/* RESOURCE DM_answer_no_such_file
 * Emits the answer "no such file" correctly wrapped with MIME conventions.
 */

static void
DM_answer_no_such_file (DM_Context *context,
                        char       *request,
                        void       *real_output_port)
{
  char      line[500];

  DM_print_char_star(context, 
      "HTTP/1.0 404 Not found\r\n"
      "From: " DM_RELEASE_EMAIL "\r\n"
      "Content-type: text/html\r\n"
      "Server: DMeroon/" DM_RELEASE_LEVEL " " DM_RELEASE_DATE "\r\n"
      "Pragma: no-cache\r\n"
                     , 0, real_output_port);  
  DM_ctime(context, line, "Date: ", 0, "\r\n\r\n"); 
  DM_print_char_star(context, line, 0, real_output_port);
  DM_print_char_star(context, 
      "<HEAD><TITLE>File not found</TITLE></HEAD>\r\n"
      "<BODY><H1>File not found</H1>\r\n"
      "<P>Sorry!</BODY>\r\n\r\n"
                     , 0, real_output_port);
}

/* RESOURCE DM_answer_html_request
 * Decodes a request and answers it in HTML.
 * The request starts with the first non space character following the 
 * 'get' keyword (the one which is recognized by DM_handle_protocol).
 * Push 'argument' in the URLscript stack if present. This is to cope with 
 * the POST request. The body of the POST request is turned into a String
 * and pushed in the stack where URLscript commands may analyze it.

 * There are many shapes for the get request. 
 * old Mosaic just sends 'get /words/words HTML1.0' while Netscape 
 * suffixes that first line with lot of lines indicating the various
 * types of MIME documents it can handle until a CRLFCRLF end.
 */

void
DM_answer_html_request (DM_Context *context,
                        char       *request,
                        DM_Object  *argument,
                        void       *output_port ) {
  DM_Site *site = DM_TheCurrentSite();
  void    *real_output_port;
  char    line[500];

  /* Compute a sensible default output stream. */
  if ( output_port ) {
    real_output_port = output_port;
  } else {
    real_output_port = (void*)context->errstream;
  }

  DM_DebugAtLevel(5,
      fprintf(DM_FindErrorStream(context), "Got request \"%s\".\n", request);
);

  /* Remove possible leading slash. */
  if ( request[0] == '/' ) {
    request++;
  }

  /* Hack: to allow icons, sounds, texts to be emitted by DMeroon,
   * handle specially the first word if /icons/ or /sounds/ or
   * /robots.txt. Take care when generating CR LF for the MIME header
   * since there must be a clear separation between the MIME header
   * and the body of the answer. 
   */

#define FirstWordIsEqualTo(s) (strncmp(request, s, sizeof(s)-1) == 0)

  /* Prevent compliant robots to enter a DMeroon site. 
   * There's also a meta tag generated below. */
  if ( FirstWordIsEqualTo("robots.txt") ) {
    return DM_answer_txt_request(context, request, 
             "User-agent: *\r\n"
             "Disallow: /\r\n"
                                 , real_output_port);
  }

  /* Serve DMeroon icons (so there is no need of a separate file to
   * hold the icon). The name of the icon follows the icons/
   * prefix. */
  if ( FirstWordIsEqualTo("icons/") ) {
    request += sizeof("icons/") -1;

    if ( FirstWordIsEqualTo("dmeroon.gif") ) {
      /* dmeroon.gif is defined in DM_dmeroon_gif, see dmgif.c file */
      return DM_answer_gif_request(context, request, 
                                   (char*)DM_dmeroon_gif, 
                                   sizeof(DM_dmeroon_gif),
                                   real_output_port );
    } else if ( FirstWordIsEqualTo("object.gif") ) {
      /* Same icon for now.                                     FUTURE */
      return DM_answer_gif_request(context, request, 
                                   (char*)DM_dmeroon_gif, 
                                   sizeof(DM_dmeroon_gif),
                                   real_output_port );
    } else {
      return DM_answer_no_such_file(context, request, real_output_port);
    }
  }

#undef FirstWordIsEqualTo
 
  /* This is a real URLscript request. Generate the MIME header for an
   * answer in html.
   * FUTURE In case of errors, change 200 into 500 ???
   */
  DM_print_char_star(context, 
    "HTTP/1.0 200 Document follows\r\n"
    "From: " DM_RELEASE_EMAIL "\r\n"
    /* "Content-length: 1234567890\r\n", It is possible to do without
     this field, so there is no need to reserve place then to compute
     the length of the emitted answer then to patch this
     field. Closing the answer stream of chars is sufficient for the
     http client to detect the end. */
    "Content-type: text/html\r\n"
    "Server: DMeroon/" DM_RELEASE_LEVEL " " DM_RELEASE_DATE "\r\n"
    "Pragma: no-cache\r\n"
                     , 0, real_output_port);
  DM_ctime(context, line, "Date: ", 0, "\r\n");
  DM_print_char_star(context, line, 0, real_output_port);
  DM_ctime(context, line, "Expires: ", 365*24*60*60, "\r\n\r\n");
  DM_print_char_star(context, line, 0, real_output_port);
  /* Generate the html content */
  DM_print_char_star(context, " <HTML>"
                     "<BASE href=\"http://", 0, real_output_port);
  DM_print_char_star(context, site->name, 0, real_output_port);
  sprintf(line, ":%lu/\">\r\n <HEAD>\r\n", (unsigned long)(ntohs(site->port)));
  DM_print_char_star(context, line, 0, real_output_port);
  /* Ask compliant robots not to follow these links (see robots.txt above).
   * Other meta keys are in order.                   FUTURE */
  DM_print_char_star(context, 
          "<META NAME=\"robots\" CONTENT=\"NOINDEX, NOFOLLOW\">\r\n", 
                     0, real_output_port);
  DM_print_char_star(context, "<TITLE> ", 0, real_output_port);
  DM_print_char_star(context, request, 0, real_output_port);
  DM_print_char_star(context, " </TITLE></HEAD>\r\n"
                              "<BODY bgcolor=\"white\">\r\n"
                              "<H1> From <A href=\""
                              DMEROON_URL "\">DMeroon</A>@", 
                     0, real_output_port );
  DM_print_char_star(context, site->name, 0, real_output_port);
  sprintf(line, ":%lu", (unsigned long)(ntohs(site->port)));
  DM_print_char_star(context, line, 0, real_output_port);
  DM_print_char_star(context, " </H1><P>\r\n", 0, real_output_port);
  
  { 
    DM_UrlState     *state = DM_Allocate0(context, UrlState);
    DM_BoundedStack *stack = DM_Allocate1(context, BoundedStack, 8);
    DM_UrlEnv       *env   = DM_Allocate1(context, UrlEnv, 2);

    state->previous = DM_NULL;
    state->script   = DM_cstring_to_string(context, request, 0);
    state->pc       = 0;
    state->stack    = stack;
    if ( argument ) {
      DM_push_object_in_bounded_stack(context, argument, &(state->stack));
    }
    /* Protect the standard predefined environment (so user definition
     * do not pollute it). */
    env->hasher    = DM_TheCurrentSite()->urlenv->hasher;
    env->extractor = DM_TheCurrentSite()->urlenv->extractor;
    env->next = (DM_Hashtable*)(DM_TheCurrentSite()->urlenv);
    env->size = 0;
    state->environment = env;
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      DM_SetCopyable(context, state);
      DM_SetCopyable(context, stack);
      DM_SetCopyable(context, env);
    }
    DM_handle_url_script(context, state, real_output_port);
  }

  /* RedBaron does not display the last field of a Dictionary if <P>
   * is not present right there !? */
  DM_print_char_star(context, "<P></BODY></HTML>\r\n\r\n", 0, real_output_port);
}

/* }}}
 * {{{ Initialization
 */

void
DM_html_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmhtml);
}

/* }}}
 * end of dmhtml.c */
