/* $Id: dmkern.c,v 1.128 1998/08/23 08:54:18 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* DOCUMENTATION dmkern.c
 * This file defines the basic DMeroon memory model.
 */

#include "dmeroon.h"

/* 
 * This variable controls whether the classes from Icslas are included
 * in the sequence of predefined classes. 
 * This is ugly and will be changed.                       FUTURE
 */

#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
#include "xla.h"
#endif /* defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT */

DM_DefineVersionString(dmkern,
  "$Id: dmkern.c,v 1.128 1998/08/23 08:54:18 queinnec Exp $");


/* {{{ Static resources

 * RESOURCE DM_Object_class
 * The Object class is the root of the inheritance tree. Recall that 
 * DMeroon only offers single inheritance. The Object class is itself
 * represented by a DMeroon object whose C representation is held in 
 * the C variable DM_Object_class_object. The Object class is a predefined
 * class with number 1.
 */

DM_Class_1_0_7_wrapped DM_Object_class_object = {
  DM_SupportPrefixInitialization(DM_Class_1_0_7_wrapped,
                                 DM_Object_class_object)
  DM_ReferenceToClass(Class),
  DM_ProxyInitialization(),
  { { {
    /* Object has number 1. */
    (1<<24)|DM_PREDEFINED_CLASS_DEFAULT_OPTIONS, 
    1,                          	/* One superclass: itself! */
    {
      DM_ReferenceToClass(Object)
    },
    { {
      0,                         	/* no fields */
      { {
        7,
        "Object"
      } }
    } }
  } } }
};

/* RESOURCE DM_nil and other constants
 * DMeroon offers a bunch of constants not to be confused with the 
 * NULL pointer (which means "no referenced object"). They have a
 * special class so they can record their name.
 * 
 * Among constants are nil, true, false etc. They are immediately 
 * useful to encode Scheme objects. 
 */

DM_DefineConstantObject(nil);

DM_DefineConstantObject(uninitialized);

DM_DefineConstantObject(missing);

DM_DefineConstantObject(true);

DM_DefineConstantObject(false);

/* }}}
 * {{{ Predicates
 */

/* RESOURCE DM_is_a
 * This predicate returns true if a DMeroon instance is of a given class.
 * It assumes that class is a DMeroon class instance. This is a 
 * constant-time predicate.

 * NOTE: For C code, use the DM_IsA macro.
 */

DM_c_boolean
DM_is_a (DM_Context *context,
         DM_Object *o, 
         DM_Class  *class) {
  DM_Class  *oc    = DM_ObjectToClass(context, o);

  if ( oc == class ) {
    return DM_TRUE;
  } else {
    DM_nat3 depth = class->super_length;

    if (   ( depth <= oc->super_length )
        && ( class == oc->super[depth-1] ) ) {
      return DM_TRUE;
    } else {
      return DM_FALSE;
    }
  }
}

/* RESOURCE DM_is_subclass
 * This predicate returns true if a class is a subclass of a given class.
 * This is a constant time predicate.
 */

DM_c_boolean
DM_is_subclass (DM_Context *context,
                DM_Class *class1, 
                DM_Class *class2) {
  if ( class1 == class2 ) {
    return DM_TRUE;
  } else {
    DM_nat3 depth = class2->super_length;

    if ( ( depth <= class1->super_length )
        && ( class2 == class1->super[depth-1] ) ) {
      return DM_TRUE;
    } else {
      return DM_FALSE;
    }
  }
}

/* RESOURCE DM_is_up_to_date
 * This predicate returns true if a remote object is validly cached.
 * If validly cached, its mutable fields may be read otherwise a request
 * to its owning site will be issued to refresh the cache. If invalidly 
 * cached, its immutable fields may still be read. If not cached at all,
 * a request will be issued to cache it locally.

 * NOTE: The difficulty is that the clock is not necessarily present.
 */

DM_c_boolean
DM_is_up_to_date (DM_Context *context,
                  DM_Object  *o) {
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Object *proxy = DM_ObjectToProxy(context, o);
  DM_Object *clk;

  if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) ) {
    if ( DM_IsA(context, proxy, Exit) ) {
      /* The object is remote. */
      DM_DeclareNarrowedObject(exit, Exit, proxy);

      /* Try to shortcut the reference to its clock. */
      DM_ShortCutReference(context,  &(exit->clock));
      clk = (DM_Object*)exit->clock;
      if ( DM_IsA(context, clk, Clock)  ) {
        DM_DeclareNarrowedObject(clock, Clock, clk);
      
        if ( clock->time == exit->time ) {
          return DM_TRUE;
        } else {
          return DM_FALSE;
        }
      } else {
        /* The clock is not present (clk should then be an Exit), so
         * it is safe to say that the object is obsolete. No need to
         * ask for the clock, a prefetch request has normally been
         * emitted while this Exit was decoded. */
        return DM_FALSE;
      }
    } else {
      /* The object is mutable but local so necessarily up to date. */
      return DM_TRUE;
    }
  } else {
    /* The Object is immutable so necessarily up to date. */
    return DM_TRUE;
  }
}

/* }}}
 * {{{ Clock utilities

 * RESOURCE DM_increment_clock
 * Increments a local Clock. Increments as well the clocker field of the 
 * current Site to mention that the clock has to be propagated. We use
 * the ticker value so in DM_handle_object_fetch_answer we can compare
 * a clock and ticker without problem.
 */

__inline__ void
DM_increment_clock (DM_Context *context,
                    DM_Clock *clock) {
  DM_Site *site = DM_TheCurrentSite();

  clock->clocker = site->clocker = clock->time = ++(site->ticker);
}

/* RESOURCE DM_increment_clock_of
 * Increment the clock monitoring a local object (if there is one and if 
 * the object is externalized).
 */

void
DM_increment_clock_of (DM_Context *context,
                       DM_Object *o) {
  DM_Object *proxy = DM_ObjectToProxy(context, o);
  
  if ( DM_IsA(context, proxy, Exit) ) {
    DM_Abort(context, DME_REMOTE_CLOCK);
  } else if ( DM_IsA(context, proxy, Entry) ) {
    DM_DeclareNarrowedObject(entry, Entry, proxy);

    if ( DM_IsA(context, entry->clock, Clock) ) {
      DM_increment_clock(context, entry->clock);
    } else {
      /* The object is externalized but has no clock. */
    }
  } else {
    /* The object is local and has no associated clock. */
  }
}

/* }}}
 * {{{ Reading/writing repetition factors.

 * RESOURCE DM_read_repetition_factor/DM_write_repetition_factor
 * Reads the repetition factor within a DMeroon instance.
 * The result is always a nat3 ie nat4 repetition factor cannot be 
 * handled by this function.
 */

DM_nat3
DM_read_repetition_factor (DM_Context *context,
                           void *p,
                           DM_nat1 reptypindex) {
  switch (reptypindex) {
  case DM_nat1_index : {
    return DM_Content(DM_nat1, p);
  }
  case DM_nat2_index : {
    return DM_Content(DM_nat2, p);
  }
  case DM_nat3_index : {
    return DM_Content(DM_nat3, p);
  }
  case DM_nat4_index : {
    return (DM_nat3)DM_Content(DM_nat4, p);
  }
  default : 
    DM_Abort(context, DME_BAD_REPETITION_TYPE);
    return (0); /* Useless */
  }
}

/* }}}
 * {{{ Computing displacements

 * Compute displacement towards a regular field or an indexed field.

 * To speed up this operation some numbers are already computed (by
 * DM_compute_layout). An offset is coded as a nat3 structured as follows:
 *            offset = [  translation, reptypindex, itemtypindex ]
 *                           nat2          nat1        nat1
 * itemtypindex is the index of the type of the field,
 * reptypindex is the type of the index if the field is indexed.
 *             nought means no repetition and has 0 as index.
 * translation is an important number. 

 * For a class without indexed field, translation is the offset of the 
 * field wrt DMeroon reference point. Suppose Point to be a class
 * with two fields x:nat2, y:nat2 then on a 32bits cpu, the offset of x
 * will be [0,nought,nat2] and that of y: [2,nought,nat2].
 * 
 * When there is one indexed field: suppose we add an (nat1-)indexed
 * (nat3-)field z to our Point class: the offset of z will be:
 * [4,nat1,nat3]. 4 is the offset for the size of the field (the nat1)
 * 4+size(nat1)+padfor(nat3)=6 is the address of [0], 10 the address of
 * z[1] etc. That is the address of z[i] is
 *      4+size(nat1)+padfor(nat3)+ i*repeatedsize(nat3)
 * 
 * Things become more complex after an indexed field. Fields now have
 * (at least) two offsets to describe their access. Suppose we add an
 * (char1-)a regular field to our previous class. Then the 'a' field
 * has offsets [4,nat1,nat3]+[0,nought,char1] and the addres of a is:
 *   4+size(nat1)+padfor(nat3)+ i*repeatedsize(nat3) + padforrealignment + 0
 *   where i is the nat1 number read at 4.
 */

DM_displacement
DM_compute_offset (DM_Context *context,
                   DM_Object *o,
                   DM_Field  *field ) {
  DM_nat3            on      = field->offset_length;
  DM_offset          offset;
  DM_displacement    t;
  DM_nat1            reptypindex, itemtypindex;
  DM_DMeroonType     *reptyp, *itemtyp;
  DM_nat3            rep;
  unsigned int       i;
  void               *p;

  /* The important case of classes without indexed fields. */
  if ( on == 1 ) {
    offset = field->offset[0];
    return DM_OffsetToTranslation(offset);
  } else {

    /* Here, field is after at least one indexed field. */
    DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
    t = 0;
    for ( i=0 ; i<on-1 ; i++ ) {
      offset       = field->offset[i];
      reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
      reptyp       = implemented_types->item[reptypindex];
      itemtypindex = DM_OffsetToItemTypeIndex(offset);
      itemtyp      = implemented_types->item[itemtypindex];
      
      t += DM_OffsetToTranslation(offset);
      /* Useless: t = DM_OffsetPad(t, reptyp); */
      p       = DM_Displace(o, t);
      rep     = DM_read_repetition_factor(context, p, reptypindex);
      t       = DM_OffsetExtend(t, reptyp);
      t       = DM_OffsetPad(t, itemtyp);
      t       += rep * itemtyp->repsize;
      /* Realign after any indexed field. */
      t       = DM_OffsetRealign(t);
    }

    /* Exploit now the last offset. */
    offset = field->offset[on-1];
    t += DM_OffsetToTranslation(offset);
    return t;
  }
}

/* RESOURCE DM_indexed_compute_offset
 * Computes a displacement within a DMeroon object where is the
 * index'th value of the indexed 'field'. The range of the index is
 * checked and an error is reported if out of range. This makes the
 * function uneasy to use unless the range is checked before hand but
 * at that time it is often useless to use that function since all the
 * necessary computations were already done.
 */

DM_displacement
DM_indexed_compute_offset (DM_Context *context,
                           DM_Object *o,
                           DM_Field  *field,
                           DM_nat3   index ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3            on      = field->offset_length;
  DM_offset          offset;
  DM_displacement    t;
  DM_nat1            reptypindex, itemtypindex;
  DM_DMeroonType     *reptyp, *itemtyp;
  DM_nat3            rep;
  void               *p;

  t            = DM_compute_offset(context, o, field);
  offset       = field->offset[on-1];
  reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
  reptyp       = implemented_types->item[reptypindex];
  itemtypindex = DM_OffsetToItemTypeIndex(offset);
  itemtyp      = implemented_types->item[itemtypindex];
  p            = DM_Displace(o, t);
  rep          = DM_read_repetition_factor(context, p, reptypindex);

  if ( index < rep ) {
    t = DM_OffsetExtend(t, reptyp);
    t = DM_OffsetPad(t, itemtyp);
    t += index * itemtyp->repsize;
    return t;
  } else {
    DM_Abort(context, DME_TOO_BIG_INDEX);
    return (0); /* Useless */
  }
}

/* 
 *
 * Accessing fields of instances of class Class.
 * This is difficult since fields are gathered in an indexed field which 
 * is in second position after the indexed field gathering superclasses.
 * These accessors are unsafe, they must be applied to classes only.
 *  
 * The following code tries to be portable, it assumes nothing (I think) on
 * representations but the following which seems to be pretty general:
 * pad1 is the padding size allowing a DM_reference to follow a DM_nat3.
 * pad2 is the padding size allowing a DM_reference to follow a DM_reference.
 * pad3 is the padding size allowing any type to follow a DM_reference.
 * Generally for 32bits machine, pad1 = pad2 = 0 !
 *
 *                     +--------------+ +------+
 *      super_length        nat3          pad1
 *                     +--------------+ +-----+
 *      super[0]          reference       pad2
 *                     +--------------+ +------+
 *        ....
 *                     +--------------+ +------+ +------+
 *      super[last]       reference       pad2     pad3
 *                     +--------------+ +------+ +------+
 *     field_length        nat3           pad1
 *                     +--------------+ +------+
 *      field[0]          reference       pad2
 *                     +--------------+ +------+
 *        ....
 * In the following size2 = sizeof(DM_reference)+pad2
 *                  size1 = sizeof(DM_nat3) + pad1
 *
 * Unfortunately pad3 is not constant since it has to align field_length
 * to the most stringent alignment constraint (generally 8 bytes). This
 * most stringent alignment is given by the DM_OffsetRealign macro.
 *
 * Implementation NOTES:
 * Here is the shape of DM-Class-class on Alpha for Bigloo. Observe
 * Bigloo tag on 8 bytes now, references on 8 bytes and swapped halves.
1:=> (DM-print-bytes dm-class-class  100  (current-output-port))

140001458: 0006e339 00000000 40001478 00000001 
140001468: 00000000 00000000 00000000 00000000 
140001478: 400000c0 00000001 00000006 00000002 
140001488: 40001310 00000001 40001478 00000001 
140001498: 00000004 00000000 40001370 00000001 
1400014a8: 400013b0 00000001 400013f0 00000001 
1400014b8: 40001430 00000001 

 * and here is the equivalent one on SUN 4 bsd.
1:=>  (DM-print-bytes dm-class-class  100  (current-output-port))

fb020: 0003719d 000fb030 00000000 00000000 
fb030: 000fa454 00000006 00000002 000faf58 
fb040: 000fb030 00000000 00000004 000faf90 
fb050: 000fafb8 000fafe0 000fb008 00000000 
 */

/* RESOURCE DM_class_indexed_field_length
 * This function returns the number of fields of a class.
 */

DM_nat3
DM_class_indexed_field_length (DM_Context *context,
                               DM_Class   *class) {
  DM_SupportAddress address =
    ((DM_SupportAddress) &class->super[0]) 
    + (class->super_length)
      * DM_ReferenceToDMeroonType(reference)->repsize;
  void *p = ((void*) DM_OffsetRealign(address));
  /* p points to field_length. */

  return DM_Content(DM_nat3, p);
}

/* RESOURCE DM_class_field_address
 * Return the address of the ith field of a class.
 */

DM_Field**
DM_class_field_address (DM_Context *context,
                        DM_Class *class, 
                        DM_nat3  i ) {
  DM_SupportAddress address1 = 
    ((DM_SupportAddress) &class->super[0])
    + (class->super_length)
     * DM_ReferenceToDMeroonType(reference)->repsize;
  DM_SupportAddress address2 = DM_OffsetRealign(address1);
  DM_SupportAddress address3 = 
    DM_OffsetExtend(address2, DM_ReferenceToDMeroonType(nat3));
  DM_SupportAddress address4 =
    DM_OffsetPad(address3, DM_ReferenceToDMeroonType(reference));
  void *p = (void*)( address4
                    + i*(DM_ReferenceToDMeroonType(reference)->repsize) );
  
  return p;
}

/* RESOURCE DM_fake_pointer
 * This function converts a pointer into something that
 * contains the same information while not recognized as a pointer by the
 * GC.  I took the code of Boehm's GC (see REVEAL_POINTER in gc.h). It must
 * be idempotent ie it fakes the pointer into something that may faked again
 * into the initial pointer.

 * NOTE: Another solution would be to declare to Boehm's GC that the
 * exits table does not contain any pointers therefore should not be
 * walked. 

 * In fact, this function takes a location holding a pointer to fake
 * but it returns the faked pointer.
 */

void*
DM_fake_pointer (DM_Context *context,
                 void*      *p) 
{
  DM_SupportAddress address = *((DM_SupportAddress*) p);
  DM_SupportAddress newaddress;

  /* Preserve NULL pointers. */
  if ( address ) {
    newaddress = ~address;
    /* *((DM_SupportAddress*) p) = newaddress; */
    return (void*)newaddress;
  } else {
    return NULL;
  }
}

/* RESOURCE DM_predefined_name
 * returns the name of a predefined object ie a C string of char.
 * Most predefined objects have as last field a repetition of characters
 * naming them. This makes easy to find names of classes and fields
 * when remotely sent.
 */

char*
DM_predefined_name (DM_Context *context,
                    DM_Object  *o ) 
{
  if ( DM_IsA(context, o, Class) ) {
    DM_displacement t = 
      DM_indexed_compute_offset(context, 
                                o, 
                                DM_ReferenceToFieldOfClass(name, Class),
                                0 );
    
    return DM_Displace(o, t);
  } else if ( DM_IsA(context, o, Field) ) {
    DM_displacement t = 
      DM_indexed_compute_offset(context, 
                                o, 
                                DM_ReferenceToFieldOfClass(name, Field),
                                0 );
    
    return DM_Displace(o, t);
  } else if ( DM_IsA(context, o, Site) ) {
    DM_DeclareNarrowedObject(site, Site, o);
    return site->name;
  } else if ( DM_IsA(context, o, Constant) ) {
    DM_DeclareNarrowedObject(constant, Constant, o);
    
    return constant->name;
  } else if ( DM_IsA(context, o, DMeroonType) ) {
    DM_DeclareNarrowedObject(dmeroontype, DMeroonType, o);
    
    return dmeroontype->name;
  } else if ( DM_IsA(context, o, CType) ) {
    DM_DeclareNarrowedObject(ctype, CType, o);

    return ctype->name;
  } else if ( DM_IsA(context, o, UrlAction) ) {
    DM_DeclareNarrowedObject(action, UrlAction, o);

    return action->name;
  } else {
    return DM_NULL;
  }
}

/* RESOURCE DM_class_field
 * This function returns the ith field of a class.
 */

DM_Field*
DM_class_field (DM_Context *context,
                DM_Class *class, 
                DM_nat3  i ) {
  void *p = DM_class_field_address(context, class, i);

  return DM_Content(DM_reference, p);
}

/* RESOURCE DM_set_class_field
 * This is an internal function that allows to overwrite a given field
 * within a class. DON'T USE IT!
 */

void
DM_set_class_field (DM_Context *context,
                    DM_Class   *class, 
                    DM_nat3    i, 
                    DM_Field   *field) {
  void *p = DM_class_field_address(context, class, i);

  DM_Content(DM_reference, p) = (DM_reference)field;
  /* No need to return the former value for this writer which is only 
     used internally and not advertized to users. */
  /* Check if the field is indexed and if it is, set its option as
   * well as the option of the class. */
  {
    DM_nat3   on          = field->offset_length;
    DM_offset last_offset = field->offset[on - 1];
    DM_nat1   reptypindex = DM_OffsetToRepetitionTypeIndex(last_offset);
    
    if ( reptypindex ) {
      DM_SetOption(field, DM_FIELD_IS_INDEXED);
      DM_SetOption(class, DM_CLASS_IS_INDEXED);
    }
  }
}

/* }}}
 * {{{ Sizing, cloning objects

 * RESOURCE DM_size
 * Compute the size in bytes of a DMeroon instance. 
 * This size does not include the size of the prefix/header (nor the 
 * support language one nor the DMeroone one) so DM_size is compatible of 
 * the sizeof pseudo-function of C.
 *
 * This size of course depends on the current site but it is useful to 
 * know it for various purposes (dumping memory, allocation etc.)
 */

DM_nat3
DM_size (DM_Context *context,
         DM_Object *o) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class          *class  = DM_ObjectToClass(context, o);
  DM_nat3           fn      = DM_class_indexed_field_length(context, class);
  DM_Field          *field;
  DM_nat3           on;
  DM_displacement   t = 0;
  DM_offset         offset;
  DM_nat1           reptypindex, itemtypindex;
  DM_DMeroonType    *reptyp, *itemtyp;
  DM_nat3           rep;
  DM_nat3           i;
  void              *p;

  /* Case of classes without fields. */
  if ( fn == 0 ) return 0;

  field = DM_class_field(context, class, fn-1);
  on    = field->offset_length;
  for ( i=0 ; i<on ; i++ ) {
    offset       = field->offset[i];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(offset);
    itemtyp      = implemented_types->item[itemtypindex];
    
    t += DM_OffsetToTranslation(offset);
    if ( reptypindex == DM_nought_index ) {
      /* i == on-1 : only the last offset may be non indexed. */
      /* Useless: t = DM_OffsetPad(t, itemtyp); */
      t   = DM_OffsetExtend(t, itemtyp);
    } else {
      /* Useless: t = DM_OffsetPad(t, reptyp); */
      p   = DM_Displace(o, t);
      rep = DM_read_repetition_factor(context, p, reptypindex);
      t   = DM_OffsetExtend(t, reptyp);
      t   = DM_OffsetPad(t, itemtyp);
      t   += rep * itemtyp->repsize;
      /* Realign after any non final indexed field. */
      if ( i<on-1 ) t = DM_OffsetRealign(t);
    }
  }
  return t;
}

/* RESOURCE DM_clone
 * Clone a DMeroon instance. It returns a new instance whose content is 
 * similar to the original one. This is a shallow copy.
 * The DMeroon header is not copied so the result is uninitialized.
 */

DM_Object*
DM_clone (DM_Context *context,
          DM_Object *o) {
  DM_nat3   size   = DM_size(context, o);
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Object *clone = DM_allocate_bare_instance(context, class, size);
  
  memcpy(clone, o, size);
  DM_SetClassOf(context, clone, class);
  DM_SetProxyOf(context, clone, DM_TheConstant(uninitialized));
  return clone;
}

/* RESOURCE DM_zero
 * Fill an object with zeroes.
 */

DM_Object*
DM_zero (DM_Context *context,
         DM_Object  *o )
{
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class          *class  = DM_ObjectToClass(context, o);
  DM_nat3           fn      = DM_class_indexed_field_length(context, class);
  DM_Field          *field;
  DM_nat3           on;
  DM_displacement   t = 0;
  DM_offset         offset;
  DM_nat1           reptypindex, itemtypindex;
  DM_DMeroonType    *reptyp, *itemtyp;
  DM_nat3           rep;
  DM_nat3           i;
  void              *p;

  /* Case of classes without fields. */
  if ( fn == 0 ) return o;

  field = DM_class_field(context, class, fn-1);
  on    = field->offset_length;
  for ( i=0 ; i<on ; i++ ) {
    offset       = field->offset[i];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(offset);
    itemtyp      = implemented_types->item[itemtypindex];
    
    t += DM_OffsetToTranslation(offset);
    if ( reptypindex == DM_nought_index ) {
      /* i == on-1 : only the last offset may be non indexed. */
      /* Useless: t = DM_OffsetPad(t, itemtyp); */
      p = DM_Displace(o, t);
      memset(p, 0, itemtyp->size);
      t   = DM_OffsetExtend(t, itemtyp);
    } else {
      /* Useless: t = DM_OffsetPad(t, reptyp); */
      p   = DM_Displace(o, t);
      rep = DM_read_repetition_factor(context, p, reptypindex);
      t   = DM_OffsetExtend(t, reptyp);
      t   = DM_OffsetPad(t, itemtyp);
      p = DM_Displace(o, t);
      memset(p, 0, rep * itemtyp->repsize);
      t   += rep * itemtyp->repsize;
      /* Realign after any non final indexed field. */
      if ( i<on-1 ) t = DM_OffsetRealign(t);
    }
  }
  return o;
}

/* }}}
 * {{{ Class utilities

 * RESOURCE DM_compute_layout
 * Computes all the offsets of the fields of a class.
 *
 * A Field instance contains a series of Offsets structured as follows:
 *     ( translation,  repetition-type-index, item-type-index )
 * The DM_compute_layout function recomputes the translation part of
 * all these offsets according to the types mentioned at its right.
 * This computation is site-dependent since and must be redone whenever
 * a Class instance (or a related Field) arrives on a site.
 */

void
DM_compute_layout (DM_Context *context,
                   DM_Class *class) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3           fn = DM_class_indexed_field_length(context, class);
  DM_translation    t  = 0;
  DM_Field          *field, *previous_field;
  DM_nat3           on;
  DM_offset         last_offset, new_offset;
  DM_nat1           itemtypindex, reptypindex;
  DM_DMeroonType    *itemtyp, *reptyp;
  DM_nat3           i, j;

  for ( i=0 ; i<fn ; i++ ) {
    field = DM_class_field(context, class, i);
    on            = field->offset_length;
    last_offset   = field->offset[on-1];
    reptypindex   = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp        = implemented_types->item[reptypindex];
    itemtypindex  = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp       = implemented_types->item[itemtypindex];

    if ( reptypindex == DM_nought_index ) {
      t = DM_OffsetPad(t, itemtyp);
    } else {
      t = DM_OffsetPad(t, reptyp);
    }
    new_offset = DM_PackOffset(t, reptypindex, itemtypindex);

    if ( on > 1 ) {
      /* Copy the offsets from the previous field. If on is greater
       * than 1 that means that there is at least one (indexed)
       * field before; therefore it is safe to assume the previous field to
       * always exist. */
      previous_field = DM_class_field(context, class, i-1);
      for ( j=0 ; j<on ; j++ )
        field->offset[j] = previous_field->offset[j];
    }

    field->offset[on-1] = new_offset;
    if ( reptypindex != DM_nought_index ) {
      /* After an indexed field, translation is reset to 0. */
      t = 0;
    } else {
      t = DM_OffsetExtend(t, itemtyp);
    }
  }
}

/* RESOURCE DM_compute_characteristics
 * This function takes a freshly created class and computes its global
 * characteristics (mutable, fixed size, contains references etc.).
 * The class must not be used to allocate instances before that operation.

 * NOTE: Not to bother complicating the dm2ch compiler, this function
 * is applied on all predefined classes. Could be made statically !!!
 */

void
DM_compute_characteristics (DM_Context *context,
                            DM_Class *class) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Type           *mst   = 
    (DM_Type*)(DM_FromC(&DM_MOST_STRINGENT_TYPE_OBJECT));
  DM_nat3           fn     = DM_class_indexed_field_length(context, class);
  DM_Field          *field;
  DM_nat3           on;
  DM_offset         last_offset;
  DM_nat1           itemtypindex, reptypindex;
  DM_DMeroonType    *itemtyp, *reptyp;
  DM_nat3           i;
  long              class_has_references = 0;

  for ( i=0 ; i<fn ; i++ ) {
    field = DM_class_field(context, class, i);
    on            = field->offset_length;
    last_offset   = field->offset[on-1];
    reptypindex   = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp        = implemented_types->item[reptypindex];
    itemtypindex  = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp       = implemented_types->item[itemtypindex];
    
    /* If an instance has a field that contains C things, then it may
     * be copied but not migrated. */
    if (    (itemtypindex == DM_data_index) 
         || (itemtypindex == DM_code_index)
         || (itemtypindex >= DM_MAXIMUM_TYPES_NUMBER)
         || DM_HasOption(field, DM_FIELD_IS_LOCAL) ) {
      DM_SetOption(class, DM_CLASS_IS_IMMOTILE);
    }

    /* A single mutable field and the class is mutable. */
    if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
      DM_SetOption(class, DM_CLASS_IS_MUTABLE);
    }

    /* A single volatile field and the class is volatile. */
    if ( DM_HasOption(field, DM_FIELD_IS_VOLATILE) ) {
      DM_SetOption(class, DM_CLASS_IS_VOLATILE);
    }

    /* Note if the class contains references (this helps GC). */
    if ( itemtypindex == DM_reference_index ) {
      class_has_references = 1;
    }

    /* Is the class stringent?  It is stringent if a field is
     * stringent (ie has the alignment constraint of the most
     * stringent type) or if the class has fields after an indexed
     * field. */
    if (    (reptyp->alignment == mst->alignment)
         || (itemtyp->alignment == mst->alignment)
         || (on >= 2) ) {
      DM_SetOption(class, DM_CLASS_IS_STRINGENT);
    }
  }
  
  /* This is a global characteristic. */
  if ( !class_has_references ) {
    DM_SetOption(class, DM_CLASS_HAS_NO_REFERENCE);
  }
}

/* RESOURCE DM_structure_inheritance
 * This function stuffs the fields of a fresh class with what it inherits
 * from its superclass.
 */

void
DM_structure_inheritance (DM_Context *context,
                          DM_Class *class,
                          DM_Class *superclass ) {
  DM_nat3  depth       = superclass->super_length;
  DM_nat3  nfields     = DM_class_indexed_field_length(context, superclass);
  DM_nat3  i;

  /* Fill all superclasses. */
  for ( i=0 ; i<depth ; i++ ) {
    class->super[i] = superclass->super[i];
  }
  class->super[depth] = class;

  /* Fill inherited fields from superclass. */
  for ( i=0 ; i<nfields ; i++ ) {
    DM_set_class_field(context, class, i, DM_class_field(context, superclass, i));
  }
}

/* RESOURCE DM_update_object
 * Copy the source object o2 into the target object o1 respecting its local 
 * fields. Check as well classes and sizes.
 */

void
DM_update_object (DM_Context *context,
                  DM_Object *o1, 
                  DM_Object *o2 )
{
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class *class = DM_ObjectToClass(context, o1);

  /* Compare classes. */
  if ( class != DM_ObjectToClass(context, o2) ) {
    DM_Abort(context, DME_DISSIMILAR_CLASSES);
  } else {
    DM_nat3           fn = DM_class_indexed_field_length(context, class);
    DM_Field          *field;
    DM_nat3           on;
    DM_offset         last_offset;
    DM_nat1           itemtypindex, reptypindex;
    DM_DMeroonType    *itemtyp, *reptyp;
    DM_nat3           i, rep1, rep2;
    void              *p1, *p2;
    DM_translation    oldt, t = 0;

    DM_DebugAtLevel(5,
      fprintf(DM_FindErrorStream(context),
              "Updating " DM_0X "%p with " DM_0X "%p\n",
              (void*)o1, (void*)o2 );
);
    
    for ( i=0 ; i<fn ; i++ ) {
      field         = DM_class_field(context, class, i);
      on            = field->offset_length;
      last_offset   = field->offset[on-1];
      reptypindex   = DM_OffsetToRepetitionTypeIndex(last_offset);
      reptyp        = implemented_types->item[reptypindex];
      itemtypindex  = DM_OffsetToItemTypeIndex(last_offset);
      itemtyp       = implemented_types->item[itemtypindex];
    
      if ( DM_HasOption(field, DM_FIELD_IS_LOCAL) ) { 
        /* The field is local don't overwrite it. */
        if ( reptypindex == DM_nought_index ) {
          /* This field is not indexed. */
          t = DM_OffsetPad(t, itemtyp);
          t = DM_OffsetExtend(t, itemtyp);
        } else {
          t    = DM_OffsetPad(t, reptyp);
          p1   = DM_Displace(o1, t);
          rep1 = DM_read_repetition_factor(context, p1, reptypindex);
          p2   = DM_Displace(o2, t);
          rep2 = DM_read_repetition_factor(context, p2, reptypindex);
          if ( rep1 != rep2 ) {
            DM_Abort(context, DME_DISSIMILAR_SKELETONS);
          } else {
            t    = DM_OffsetExtend(t, reptyp);
            t    = DM_OffsetPad(t, itemtyp);
            t += rep1 * itemtyp->repsize;  
            /* Realign after any non final indexed field. */
            t = DM_OffsetRealign(t);
          }
        }
      } else {
        /* The field is not local, overwrite it. */
        if ( reptypindex == DM_nought_index ) {
          /* The field is not indexed. */
          t = DM_OffsetPad(t, itemtyp);
          p1 = DM_Displace(o1, t);
          p2 = DM_Displace(o2, t);
          memcpy(p1, p2, itemtyp->size);
          t = DM_OffsetExtend(t, itemtyp);
        } else {
          /* This is an indexed field. */
          t    = DM_OffsetPad(t, reptyp);
          p1   = DM_Displace(o1, t);
          rep1 = DM_read_repetition_factor(context, p1, reptypindex);
          p2   = DM_Displace(o2, t);
          rep2 = DM_read_repetition_factor(context, p2, reptypindex);
          if ( rep1 != rep2 ) {
            DM_Abort(context, DME_DISSIMILAR_SKELETONS);
          } else {
            /* Copy the whole indexed field including its repetition factor. */
            oldt = t;
            t   = DM_OffsetExtend(t, reptyp);
            t   = DM_OffsetPad(t, itemtyp);
            t += rep1 * itemtyp->repsize;
            memcpy(p1, p2, t - oldt);
            /* Realign after any non final indexed field. */
            t = DM_OffsetRealign(t);
          }
        }
      }
    }
  }
}

/* }}}
 * {{{ Allocation
 *
 * RESOURCE DM_allocate
 * Allocate an instance with any number of indexed fields. Returns DM_NULL
 * if this is not possible to allocate such an instance. [This is 
 * idealistic since probably the memory system is exhausted and nothing
 * more may be done.]
 *
 * Predefined allocators for 0, 1, 2, and 3 indexed fields also exist.
 *
 * RESTRICTION: Repetition factors are limited to DM_nat3 values.
 *
 *.FIGURE allocation hierarchy
 *  DM_allocate<i>
 *     calls DMeroon_do_allocate_instance
 *           calls DM_allocate_bare_instance (provided by the support language)
 *.ENDFIGURE
 */

DM_Object*
DM_allocate_instance (DM_Context *context,
                      DM_Class *class,
                      DM_nat3  number_of_sizes,
                      DM_nat3  sizes[] ) {
  DM_Object    *o;
  DM_ErrorCode status;

  status = DMeroon_do_allocate(context, &o, class, number_of_sizes, sizes);
  if ( status == DME_SUCCESS ) {
    return o;
  } else {
    DM_Abort(context, status);
    return DM_NULL;
  }
}

DM_Object*
DM_allocate0 (DM_Context *context,
              DM_Class *class) {
  DM_Object    *o;
  DM_ErrorCode status;

  status = DMeroon_do_allocate(context, &o, class, 0, DM_NULL);
  if ( status == DME_SUCCESS ) {
    return o;
  } else {
    DM_Abort(context, status);
    return DM_NULL;
  }
}

DM_Object*
DM_allocate1 (DM_Context *context,
              DM_Class *class,
              DM_nat3  i0 ) {
  DM_Object    *o;
  DM_ErrorCode status;

  status = DMeroon_do_allocate(context, &o, class, 1, &i0);
  if ( status == DME_SUCCESS ) {
    return o;
  } else {
    DM_Abort(context, status);
    return DM_NULL;
  }
}

DM_Object*
DM_allocate2 (DM_Context *context,
              DM_Class *class,
              DM_nat3  i0,
              DM_nat3  i1 ) {
  DM_Object    *o;
  DM_ErrorCode status;
  DM_nat3      sizes[2];

  sizes[0] = i0;
  sizes[1] = i1;
  status = DMeroon_do_allocate(context, &o, class, 2, sizes);
  if ( status == DME_SUCCESS ) {
    return o;
  } else {
    DM_Abort(context, status);
    return DM_NULL;
  }
}

DM_Object*
DM_allocate3 (DM_Context *context,
              DM_Class *class,
              DM_nat3  i0,
              DM_nat3  i1,
              DM_nat3  i2 ) {
  DM_Object    *o;
  DM_ErrorCode status;
  DM_nat3      sizes[3];

  sizes[0] = i0;
  sizes[1] = i1;
  sizes[2] = i2;
  status = DMeroon_do_allocate(context, &o, class, 3, sizes);
  if ( status == DME_SUCCESS ) {
    return o;
  } else {
    DM_Abort(context, status);
    return DM_NULL;
  }
}

/* }}}
 * {{{ Initialization and predefined classes
 
 * RESOURCE DM_predefined_classes_object
 * This Immutable vector holds all the predefined classes of DMeroon.
 * The first position (at index 0) should not be used. 
 */

DM_PredefinedClassesVector_256_wrapped DM_predefined_classes_object = {
  DM_SupportPrefixInitialization(DM_PredefinedClassesVector_256_wrapped,
                                 DM_predefined_classes_object)
  DM_ReferenceToClass(ImmutableVector),
  DM_ProxyInitialization(),
  { { {
    256,                                /* Temporary see dm_kernel_initialize */
    {
      DM_NULL,				/* no class at position 0 */
      DM_ReferenceToClass(Object),	/* Object has number 1 */
      DM_ReferenceToClass(Constant),
      DM_ReferenceToClass(Type),
      DM_ReferenceToClass(CType),
      DM_ReferenceToClass(DMeroonType),
      DM_ReferenceToClass(RequestMetaClass),
      DM_ReferenceToClass(Field),
      DM_ReferenceToClass(Class),
      DM_ReferenceToClass(String),
      DM_ReferenceToClass(MutableVector),
      DM_ReferenceToClass(ImmutableVector),
      DM_ReferenceToClass(External),
      DM_ReferenceToClass(Entry),
      DM_ReferenceToClass(EntryWithDelayedRequests),
      DM_ReferenceToClass(EntryWithClass),
      DM_ReferenceToClass(Exit),
      DM_ReferenceToClass(Clock),
      DM_ReferenceToClass(Hashtable),
      DM_ReferenceToClass(BoundedStack),
      DM_ReferenceToClass(ImmutableLink),
      DM_ReferenceToClass(ImmutablePair),
      DM_ReferenceToClass(QueueLink),
      DM_ReferenceToClass(Symbol),
      DM_ReferenceToClass(Boolean),
      DM_ReferenceToClass(True),
      DM_ReferenceToClass(False),
      DM_ReferenceToClass(ImmutableByteVector),
      DM_ReferenceToClass(Null),
      DM_ReferenceToClass(VersionString),
      DM_ReferenceToClass(Port),
      DM_ReferenceToClass(Tcp),
      DM_ReferenceToClass(Connection),
      DM_ReferenceToClass(Server),
      DM_ReferenceToClass(Buffer),
      DM_ReferenceToClass(InBuffer),
      DM_ReferenceToClass(OutBuffer),
      DM_ReferenceToClass(Marker),
      DM_ReferenceToClass(Protocol),
      DM_ReferenceToClass(Site),
      DM_ReferenceToClass(Dictionary),
      DM_ReferenceToClass(Value),
      DM_ReferenceToClass(Number),
      DM_ReferenceToClass(ExactNumber),
      DM_ReferenceToClass(SignedNumber),
      DM_ReferenceToClass(Int1),
      DM_ReferenceToClass(Int2),
      DM_ReferenceToClass(Int3),
      DM_ReferenceToClass(Int4),
      DM_ReferenceToClass(UnsignedNumber),
      DM_ReferenceToClass(Nat1),
      DM_ReferenceToClass(Nat2),
      DM_ReferenceToClass(Nat3),
      DM_ReferenceToClass(Nat4),
      DM_ReferenceToClass(InexactNumber),
      DM_ReferenceToClass(Float1),
      DM_ReferenceToClass(Float2),
      DM_ReferenceToClass(Float3),
      DM_ReferenceToClass(Char1),
      DM_ReferenceToClass(NetNat2),
      DM_ReferenceToClass(NetNat3),
      DM_ReferenceToClass(Data),
      DM_ReferenceToClass(Code),
      DM_ReferenceToClass(Reference),
      DM_ReferenceToClass(Date),
      DM_ReferenceToClass(Key),
      DM_ReferenceToClass(ClassOptions),
      DM_ReferenceToClass(FieldOptions),
      DM_ReferenceToClass(PortOptions),
      DM_ReferenceToClass(SiteOptions),
      DM_ReferenceToClass(ExternalOptions),
      DM_ReferenceToClass(ContextOptions),
      DM_ReferenceToClass(FileOptions),
      DM_ReferenceToClass(RequestOptions),
      DM_ReferenceToClass(HandlerCode),
      DM_ReferenceToClass(Request),
      DM_ReferenceToClass(SiteConnectionRequest),
      DM_ReferenceToClass(ObjectSendRequest),
      DM_ReferenceToClass(ObjectFetchRequest),
      DM_ReferenceToClass(SiteFetchRequest),
      DM_ReferenceToClass(ObjectMutationRequest),
      DM_ReferenceToClass(ObjectIndexedMutationRequest),
      DM_ReferenceToClass(OwnerChangeRequest),
      DM_ReferenceToClass(RemoteOwnerChangeRequest),
      DM_ReferenceToClass(UrlStateMigrationRequest),
      DM_ReferenceToClass(Answer),
      DM_ReferenceToClass(SiteConnectionAnswer),
      DM_ReferenceToClass(ObjectFetchAnswer),
      DM_ReferenceToClass(ObjectMutationAnswer),
      DM_ReferenceToClass(OwnerChangeAnswer),
      DM_ReferenceToClass(RemoteOwnerChangeAnswer),
      DM_ReferenceToClass(FailureAnswer),
      DM_ReferenceToClass(UrlStateMigrationAnswer),
      DM_ReferenceToClass(Context),
      DM_ReferenceToClass(Waiter),
      DM_ReferenceToClass(FileWaiter),
      DM_ReferenceToClass(ReceptionWaiter),
      DM_ReferenceToClass(DurationWaiter),
      DM_ReferenceToClass(AnswerWaiter),
      DM_ReferenceToClass(EvaluationWaiter),
      DM_ReferenceToClass(Fact),
#if defined(DM_SCHEME_IS_PRESENT) && DM_SCHEME_IS_PRESENT
      DM_ReferenceToClass(Generic),
      DM_ReferenceToClass(SchemeMethod),
      DM_ReferenceToClass(SchemeContinuation),
      DM_ReferenceToClass(Invokation),
#endif /* DM_SCHEME_IS_PRESENT */
      DM_ReferenceToClass(File),
      DM_ReferenceToClass(UrlAction),
      DM_ReferenceToClass(UrlState),
      DM_ReferenceToClass(UrlEnv),
#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
      DM_ReferenceToClass(MutablePair),
      DM_ReferenceToClass(MutableString),
      DM_ReferenceToClass(Character),
      DM_ReferenceToClass(StreamPort),
      DM_ReferenceToClass(InputPort),
      DM_ReferenceToClass(BootPort),
      DM_ReferenceToClass(OutputPort),
      DM_ReferenceToClass(Number),
      DM_ReferenceToClass(Closure),
      DM_ReferenceToClass(Primitive),
      DM_ReferenceToClass(SpecialPrimitive),
      DM_ReferenceToClass(Environment),
      DM_ReferenceToClass(NullEnvironment),
      DM_ReferenceToClass(Box),
      DM_ReferenceToClass(NamedBox),
      DM_ReferenceToClass(Continuation),
      DM_ReferenceToClass(InternalContinuation),
      DM_ReferenceToClass(HandlerContinuation),
      DM_ReferenceToClass(SkipHandlerContinuation),
      DM_ReferenceToClass(ValueContinuation),
      DM_ReferenceToClass(ContextContinuation),
      DM_ReferenceToClass(EscaperContextContinuation),
      DM_ReferenceToClass(Thread),
      DM_ReferenceToClass(ToplevelThread),
      DM_ReferenceToClass(Anomaly),
      DM_ReferenceToClass(UnderlyingAnomaly),
      DM_ReferenceToClass(ObsoleteContinuationAnomaly),
      DM_ReferenceToClass(NonContinuableAnomaly),
      DM_ReferenceToClass(Module),
      DM_ReferenceToClass(ModuleEnvironment),
      DM_ReferenceToClass(User),
#endif /* defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT */
      /* sentinel mark the end (used in DM_kernel_initialize) */
      DM_NULL,
    }
  } } }
};

/* Initialization.  The main goal of that initialization is to
 * initialize all predefined classes with sequential number. Object
 * has number 1. A global ImmutableVector holds all these classes and
 * DM_NULL among them.
 */

void
DM_kernel_initialize (DM_Context *context) {
  DM_PredefinedClassesVector *predefined_classes = DM_ThePredefinedClasses();
  DM_nat3 i;

  DM_LinkVersionStringIntoSite(dmkern);

  /* Set the rank of all predefined classes and set the size of the
   * ImmutableVector predefined_classes. Also set up the layout. 
   * Start the iteration from 1 since 0 contains NULL. */

  for ( i=1 ; i<predefined_classes->item_length ; i++ ) {
    if ( predefined_classes->item[i] ) {
      DM_DeclareNarrowedObject(class, Class, predefined_classes->item[i]);

      class->options |= i<<24;
      /* Already done: DM_SetOption(class, DM_CLASS_IS_USABLE); */
      DM_compute_characteristics(context, class);
    } else {
      /* The vector contains DM_NULL at position i, skeletizes the 
       * vector to that size. */
      if ( i < 256 ) {
        predefined_classes->item_length = i;
      } else {
        /* The vector of predefined classes is too short. You cannot
         * extend it since predefined classes are encoded on a byte
         * and this is required in many places. */
        DM_Abort(context, DME_NOT_ENOUGH_PREDEFINED_CLASSES);
      }
    }
  }

DM_AssociateHandlerToRequestMetaClass(SiteConnectionRequest,DM_handle_site_connection_request);
DM_AssociateHandlerToRequestMetaClass(ObjectSendRequest,DM_handle_object_send_request);
DM_AssociateHandlerToRequestMetaClass(ObjectFetchRequest,DM_handle_object_fetch_request);
DM_AssociateHandlerToRequestMetaClass(SiteFetchRequest,NULL); /* Useful ? */
DM_AssociateHandlerToRequestMetaClass(ObjectMutationRequest,DM_handle_object_mutation_request);
DM_AssociateHandlerToRequestMetaClass(ObjectIndexedMutationRequest,DM_handle_object_indexed_mutation_request);
DM_AssociateHandlerToRequestMetaClass(OwnerChangeRequest,DM_handle_owner_change_request);
DM_AssociateHandlerToRequestMetaClass(RemoteOwnerChangeRequest,DM_handle_remote_owner_change_request);
DM_AssociateHandlerToRequestMetaClass(SiteConnectionAnswer,DM_handle_site_connection_answer);
DM_AssociateHandlerToRequestMetaClass(ObjectFetchAnswer,DM_handle_object_fetch_answer);
DM_AssociateHandlerToRequestMetaClass(ObjectMutationAnswer,DM_handle_object_mutation_answer);
DM_AssociateHandlerToRequestMetaClass(OwnerChangeAnswer,DM_handle_owner_change_answer);
DM_AssociateHandlerToRequestMetaClass(FailureAnswer,DM_handle_failure_answer);
}

/* }}}
 * end of dmkern.c */
