/* $Id: dmnet.c,v 1.128 1998/06/05 18:38:25 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* DOCUMENTATION dmnet.c
 * This file handles the low-level network-oriented aspects of DMeroon
 * that is, how to open/close servers, Tcp connections, how to fill/flush
 * buffers, how to read/write buffers or messages. 
 */

#include "dmeroon.h"

DM_DefineVersionString(dmnet,
  "$Id: dmnet.c,v 1.128 1998/06/05 18:38:25 queinnec Exp $");

/* {{{ Managing Ports

 * RESOURCE DM_active_ports
 * This Vector instance holds all active connections watched by
 * DMeroon.  Since it is a global and mutable resource, it must be
 * accessed with great care (that's why DMeroon is supposed to execute
 * in a single thread).

 * NOTE: We are actually not interested by the first three predefined
 * streams.  But even if the following initialization assumes that
 * they have fd numbers 0, 1 and 2, we don't care.

 * NOTE: On most machines FD_SETSIZE is 256. It raises up to 4096 on
 * alpha.  This seems unrelated to getdtablesize() which often returns
 * 32. To avoid such a huge table, maybe we can use an Hashtable ???
 */

DM_ports_vector_wrapped DM_active_ports_object = {
  DM_SupportPrefixInitialization(DM_ports_vector_wrapped,
                                 DM_active_ports_object)
  DM_ReferenceToClass(MutableVector),
  DM_ProxyInitialization(),
  { { {
    FD_SETSIZE,   /* assume FD_SETSIZE >= 3 */
    {
      DM_NULL,     /* stdin    */
      DM_NULL,     /* stdout   */
      DM_NULL,     /* stderr   */
    }
  } } }
};

/* DOCUMENTATION DM_max_port_number
 * This is the highest number associated to a socket used by
 * DMeroon. It is maintained by DM_(in)activate_port procedures.
 * This number is necessary for select(2).
 */

unsigned DM_max_port_number = 0;

/* DOCUMENTATION DM_current_ports_to_read/write
 * These are used by select(2) to determine the state of the
 * connections watched over by DMeroon.  Every active port has the
 * corresponding bits set. These masks are used to produce the fd for
 * select(2). They are initialized in DM_net_initialize(). Any freshly
 * created server or connection has an associated bit in it.
 *
 * The result after select(2) is kept in DM_last_ports_to_read/write.

 * NOTE: It is possible (as in sched.scm) to invoke DM_activate_port
 * on its own objects (InputPort for Icslas for instance). 
 */

fd_set DM_current_ports_to_read;
fd_set DM_current_ports_to_write;

fd_set DM_last_ports_to_read;
fd_set DM_last_ports_to_write;

/* RESOURCE DM_activate_port
 * Register a new Port in the sequence of Ports to be watched by DMeroon.
 * It is possible to register any kind of Object with a file-descriptor.
 * The third argument tells if the Object must be watched for read, or
 * write or both. This third argument is a combination of DM_PORT_TO_READ 
 * and DM_PORT_TO_WRITE constants.
 */

void
DM_activate_port (DM_Context *context,
                  DM_nat2    sockfd, 
                  DM_Port    *port,
                  DM_int2    io ) {
  DM_ports_vector *active_ports = 
    (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));

  DM_DebugAtLevel(21,
    fprintf(DM_FindErrorStream(context), "Activated Port[%d]\n", sockfd);
);

  active_ports->port[sockfd] = port;
  if ( io & DM_PORT_TO_READ )
    FD_SET(sockfd, &DM_current_ports_to_read);
  if ( io & DM_PORT_TO_WRITE )
    FD_SET(sockfd, &DM_current_ports_to_write);
  if ( sockfd > DM_max_port_number ) 
    DM_max_port_number = sockfd;
  /* Record these two options in the Port so it can be watched for. 
   * Ignore if this is not a Port (see Bigloo/dmports.c file). */
  if ( DM_IsA(context, port, Port) ) {
    DM_SetOption(port, io);
  }
}

/* RESOURCE DM_inactivate_port
 * Remove a Port from the sequence of Ports that are watched by DMeroon.
 */

void
DM_inactivate_port (DM_Context *context,
                    DM_nat2    sockfd) {
  DM_ports_vector *active_ports = 
    (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));

  DM_DebugAtLevel(21,
    fprintf(DM_FindErrorStream(context), "Inactivated Port[%d]\n", sockfd);
);

  active_ports->port[sockfd] = DM_NULL;
  FD_CLR(sockfd, &DM_current_ports_to_read);
  FD_CLR(sockfd, &DM_current_ports_to_write);
  /* try to lower DM_max_port_number to improve DM_select. */
  if ( sockfd == DM_max_port_number ) {
    while (    ( DM_max_port_number > 0 )
           && ( ! active_ports->port[DM_max_port_number] ) ) {
      DM_max_port_number--;
    }
  }
}

/* }}}
 * {{{ Server utilities

 * DOCUMENTATION
 * Generally, the allocation and the initialization of DMeroon objects is
 * clearly separated into two diffferent functions. This allows the user
 * to define subclasses of the DMeroon predefined internal classes and
 * still use the initialization with the (call-next-method) feature.

 * RESOURCE DM_establish_server
 * This function binds a Server instance to an open socket and makes it
 * listening to that socket. This function returns DM_NULL if 
 * something wrong happens, otherwise it returns the Server.

 * If you allocate yourself your own Server instance, you can still use
 * that function which will only set up its 'number' field with the
 * appropriate file-descriptor.
 */

DM_Server*
DM_establish_server (DM_Context *context,
                     DM_Server  *server) {
  int     sockfd, tmp;
  void    *tmpaddr = &tmp;
  DM_nat2 backlog  = server->backlog;

  DM_DebugAtLevel(50,
    fprintf(DM_FindErrorStream(context), "Serving port %d with " DM_0X "%p.\n",
            ntohs(server->port), server);
);

  server->options &= DM_PORT_PERMANENT_OPTIONS;
  if ( (sockfd = socket(server->family, 
                        SOCK_STREAM, 
                        IPPROTO_TCP)) < 0 ) {
    DM_SetOption(server, DM_PORT_FAILED_SOCKET);
    DM_MayAbort(context, DME_FAILED_SOCKET);
    return DM_NULL;
  }
  server->number = sockfd;
  if ( (setsockopt(sockfd, 
                   SOL_SOCKET, 
                   SO_REUSEADDR,
                   tmpaddr,
                   sizeof(tmp))) < 0 ) {
    DM_SetOption(server, DM_PORT_FAILED_SETSOCKOPT);
    DM_MayAbort(context, DME_FAILED_SETSOCKOPT);
    return DM_NULL;
  }
  if ( (bind(sockfd, 
             ((struct sockaddr*) &(server->family)),
             sizeof(DMeroon_sockaddr_in) )) < 0 ) {
    DM_SetOption(server, DM_PORT_FAILED_BIND);
    DM_MayAbort(context, DME_FAILED_BIND);
    return DM_NULL;
  }
  if ( (listen(sockfd,backlog)) < 0 ) {
    DM_SetOption(server, DM_PORT_FAILED_LISTEN);
    DM_MayAbort(context, DME_FAILED_LISTEN);
    return DM_NULL;
  }

  /* Record this new active server and return it. 
   * No need to write on a Server so use only one direction. */
  DM_activate_port(context, sockfd, (DM_Port*)server, DM_PORT_TO_READ);
  return server;
}

/* DOCUMENTATION host/network order
 * IP addresses, port numbers may appear in host-order or network-order.
 * As a rule, these numbers are always given by the user in host-order.

 * RESOURCE DM_create_server
 * Create a Server instance, initialize it and establish a server with it
 * that is, try to bind it to a listening socket. Return the Server
 * instance otherwise return DM_NULL. The port parameter is a regular
 * nat2 (host-ordered).
 */

DM_Server*
DM_create_server (DM_Context *context,
                  DM_nat2    port, 
                  DM_nat2    backlog) {
  DM_Server *server = DM_Allocate0(context, Server);
  
  /* Initialize all the fields of this server except its 'number' field
   * which will hold the file descriptor when bound to a socket. 
   */
  server->retries    = DM_PORT_WRITE_RETRY_NUMBER;   /* Useful for a server ? */
  server->options    = DM_PORT_DEFAULT_OPTIONS;
  server->family     = AF_INET;
  server->port       = htons(port);
  server->address    = htonl(INADDR_ANY);
  server->last_read  = 0;
  server->last_write = 0;
  /* This number counts the number of incoming connections. */
  server->connected = 0;
  server->backlog = backlog;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, server);
  }

  return DM_establish_server(context, server);
}

/* RESOURCE DM_establish_incoming_connection
 * Read a connection request on a Server instance and initializes a
 * Connection instance to represent it. The 'number' field of the 
 * Connection will be set up appropriately (and of course the 'options'
 * field and the other fields of the sockaddr part). 

 * You may allocate your own Connection instance and apply this
 * function on it (that is not using DM_accept_connection). 
 * This function returns DM_NULL if something went wrong.
 */

DM_Connection*
DM_establish_incoming_connection (DM_Context    *context,
                                  DM_Server     *server,
                                  DM_Connection *connection) {
  int           sockaddr_len = sizeof(DMeroon_sockaddr_in);
  int           sockfd, tmp;
  void          *tmpaddr = &tmp;
  struct linger linger;

  DM_DebugAtLevel(10,
    fprintf(DM_FindErrorStream(context), 
            "Accepting connection from " DM_0X "%p.\n", server);
);

  server->options &= DM_PORT_PERMANENT_OPTIONS;
  DM_ResetErrno();
  /* this is a possibly blocking system call. */
  if ( (sockfd = accept(server->number,
                        ((struct sockaddr*) &(connection->family)),
                        /* The address, not the value !!! */
                        &sockaddr_len )) < 0 ) {
    if ( DM_IsErrnoEqualTo(DM_EINTR) ) 
      return DM_NULL;
    DM_SetOption(server, DM_PORT_FAILED_ACCEPT);
    DM_MayAbort(context, DME_FAILED_ACCEPT);
    return DM_NULL;
  }
               
  DM_DebugAtLevel(10,
    fprintf(DM_FindErrorStream(context), 
            "Accepted connection: " DM_0X "%p.\n", connection);
);

#if !defined(DMEROON_IS_MUTE)
  { struct in_addr ia;
    
    ia.s_addr = connection->address;
    fprintf(DM_FindErrorStream(context), 
            "**Connection from %s\n", inet_ntoa(ia));
  }
#endif
  
  connection->number  = sockfd;
  connection->options = DM_CONNECTION_DEFAULT_OPTIONS;

  /* Set up this new connection */
  if ( setsockopt(sockfd,
                  SOL_SOCKET,
                  SO_KEEPALIVE,
                  tmpaddr,
                  sizeof(tmp))  < 0 ) {
    DM_SetOption(connection, DM_PORT_FAILED_SETSOCKOPT);
    DM_MayAbort(context, DME_FAILED_SETSOCKOPT);
    return DM_NULL;
  }
  /* SO_DONTLINGER seems to be more graceful but it is not offered on
   * my old SonyNews. */
  linger.l_onoff  = 1;					  /* linger is on */
  linger.l_linger = DM_TheCurrentSite()->normal_duration; /* seconds */
  if ( setsockopt(sockfd,
                  SOL_SOCKET,
                  SO_LINGER,
                  (void*)&linger,
                  sizeof(linger))  < 0 ) {
    DM_SetOption(connection, DM_PORT_FAILED_SETSOCKOPT);
    DM_MayAbort(context, DME_FAILED_SETSOCKOPT);
    return DM_NULL;
  }

  /* Statistics: Increment the number of connections accepted by the
   * Server and record the current date. */
  server->connected++;
  server->last_read = DM_get_date(context);
  
  /* Record this new active connection and return it. */
  DM_activate_port(context, sockfd, (DM_Port*)connection, 
                  DM_PORT_TO_READ | DM_PORT_TO_WRITE );
  return connection;
}

/* RESOURCE DM_accept_connection
 * Create a Connection instance, initialize it and establish an
 * incoming connection with it. A connection is accompanied by
 * two buffers that must be allocated and initialized altogether.
 * Return DM_NULL in case of problems otherwise return the Connection.
 */

DM_Connection*
DM_accept_connection (DM_Context *context,
                      DM_Server  *server) {
  DM_Connection *connection = DM_Allocate0(context, Connection);
  DM_InBuffer   *readbuf    = 
    DM_Allocate1(context, InBuffer, DM_INITIAL_BUFFER_SIZE);
  DM_OutBuffer  *writebuf   = 
    DM_Allocate1(context, OutBuffer, DM_INITIAL_BUFFER_SIZE);
  DM_BoundedStack  *readstack  = DM_Allocate1(context, BoundedStack, 2);
  DM_BoundedStack  *writestack = DM_Allocate1(context, BoundedStack, 2);
  DM_MutableVector *readcache  = DM_Allocate1(context, MutableVector, 2);
  DM_MutableVector *writecache = DM_Allocate1(context, MutableVector, 2);

  readstack->index  = 0;
  writestack->index = 0;

  readbuf->port    = (DM_Port*)connection;
  readbuf->cache   = readcache;
  readbuf->stack   = readstack;
  readbuf->mark    = DM_NULL;
  readbuf->index   = 0;
  readbuf->maximum = 0;
  readbuf->amount  = 0;
  readbuf->next    = DM_NULL;

  writebuf->port    = (DM_Port*)connection;
  writebuf->cache   = writecache;
  writebuf->stack   = writestack;
  writebuf->mark    = DM_NULL;
  writebuf->index   = 0;
  writebuf->maximum = 0;
  writebuf->amount  = 0;
  writebuf->next    = DM_NULL;

  connection->retries    = DM_PORT_WRITE_RETRY_NUMBER;
  connection->options    = DM_CONNECTION_DEFAULT_OPTIONS;
  connection->last_read  = 0;
  connection->last_write = 0;
  connection->reception  = readbuf;
  connection->emission   = writebuf;
  connection->emitqueue  = writebuf;
  connection->clocker    = 0;
  connection->site       = DM_NULL;

  /* Use the default settings from the current Site instance. */
  connection->protocol = DM_TheCurrentSite()->protocol;
  connection->handler  = DM_NULL;

  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, readbuf);
    DM_SetSharable(context, writebuf);
    DM_SetSharable(context, readcache);
    DM_SetSharable(context, writecache);
    DM_SetSharable(context, connection);
  }

  return DM_establish_incoming_connection(context, server, connection);
}

/* }}}
 * {{{ IP addresses management routines.
 *
 * RESOURCE DM_cxxvii_0_0_i
 * This ip address represents 'localhost'. It is stored in network order.
 */

static DM_netnat3 DM_cxxvii_0_0_i;

/* RESOURCE DM_get_ip_address
 * Converts a string into an IP address. The string may be the name of
 * an host, or an IP number written in dot notation. INADDR_NONE is
 * returned if the address cannot be decoded. The additional argument
 * tells in case of multiply possible addresses which one to choose. 0
 * is the first choice. Attention: the returned number is a regular nat3
 * number (host-ordered).

 * NOTE: man pages (on Sony News) say that the referred hostent is in
 * a static area so it does not need to be free-d.
 */

DM_nat3 
DM_get_ip_address (DM_Context *context,
                   char       *hostname, 
                   DM_nat3    choice ) {
  unsigned long  ip;
  struct hostent *hp;
  unsigned int   i;

  ip = inet_addr(hostname);
  /* ip is in network order (and so is DM_cxxvii_0_0_i). */
  if ( ip != INADDR_NONE ) {
    /* Takes care of 'localhost' which is often (always?) 127.0.0.1. */
    if ( ip == DM_cxxvii_0_0_i ) {
      return ntohl(DM_TheCurrentSite()->address);
    } else {
      return ntohl(ip); 
    }
  } else if ( (hp = gethostbyname(hostname)) == NULL ) {

#if defined(DM_HAS_HERROR)
    DM_DebugAtLevel(0,
      herror(NULL);
);
#endif

    DM_MayAbort(context, DME_FAILED_GETHOSTBYNAME);
    return ntohl(INADDR_NONE);
  } else {
    for ( i=0 ; ; i++ ) {
      unsigned long *addr = (unsigned long*)(hp->h_addr_list[i]);

      if ( addr ) {
        if ( i == choice ) {
          /* (*addr)[] are in network order. */
          ip = *addr;
          /* Takes care of 'localhost' which is often (always?) 127.0.0.1. */
          if ( ip == DM_cxxvii_0_0_i ) {
            return ntohl(DM_TheCurrentSite()->address);
          } else {
            return ntohl(ip); 
          }
        } else {
          /* this is not the requires choice-th address. */
          continue;
        }
      } else {
        /* No more ip addresses are available. Report failure. */
        DM_MayAbort(context, DME_HOST_NAME_CHOICE);
        return ntohl(INADDR_NONE);
      }
    }
  }
}

/* }}}
 * {{{ Client utilities

 * RESOURCE DM_establish_outgoing_connection
 * Take a Connection towards some host and initialize the TCP/IP
 * channel.  Return the Connection if successful, otherwise return
 * DM_NULL. You may create the Connection by yourself or use the
 * DM_create_connection function.  DM_establish_outgoing_connection
 * will only set up the 'number' field with the appropriate
 * file-descriptor and the 'family' field.
 */

DM_Connection*
DM_establish_outgoing_connection (DM_Context    *context,
                                  DM_Connection *connection) {
  int sockfd;

  /* initialize connection. */
  connection->options = DM_CONNECTION_DEFAULT_OPTIONS;

  DM_DebugAtLevel(10,
    { struct in_addr ia;

      ia.s_addr = connection->address;
      fprintf(DM_FindErrorStream(context),
              "Remote host Internet address=\"%s\".\n",
              inet_ntoa(ia) );
    }
);

  connection->family = AF_INET;
  if ( (sockfd = socket(connection->family, 
                        SOCK_STREAM, 
                        IPPROTO_TCP)) < 0 ) {
    DM_SetOption(connection, DM_PORT_FAILED_SOCKET);
    DM_MayAbort(context, DME_FAILED_SOCKET);
    return DM_NULL;
  }
  connection->number = sockfd;

  DM_ResetErrno();
  if ( (connect(sockfd, 
                ((struct sockaddr*) &(connection->family)),
                sizeof(DMeroon_sockaddr_in) )) < 0 ) {
    DM_SetOption(connection, DM_PORT_FAILED_CONNECT);
    DM_MayAbort(context, DME_FAILED_CONNECT);
    return DM_NULL;
  }

  /* Record this new active connection and return it. */
  DM_activate_port(context, sockfd, (DM_Port*)connection, 
                  DM_PORT_TO_READ | DM_PORT_TO_WRITE );
  return connection;
}

/* RESOURCE DM_create_connection
 * Allocate buffers, connection and host instances then establish a
 * connection with a remote host. Returns that Connection if successfully
 * established, otherwise returns DM_NULL.
 */

DM_Connection*
DM_create_connection (DM_Context *context,
                      char       *hostname, 
                      DM_nat2    remote_port,
                      char       *password ) {
  DM_Connection *connection = DM_Allocate0(context, Connection);
  DM_InBuffer   *reception  =
    DM_Allocate1(context, InBuffer, DM_INITIAL_BUFFER_SIZE);
  DM_OutBuffer  *emission   =
    DM_Allocate1(context, OutBuffer, DM_INITIAL_BUFFER_SIZE);
  DM_BoundedStack *stacki  = DM_Allocate1(context, BoundedStack, 2);
  DM_BoundedStack *stacko  = DM_Allocate1(context, BoundedStack, 2);
  DM_MutableVector *cachei = DM_Allocate1(context, MutableVector, 2);
  DM_MutableVector *cacheo = DM_Allocate1(context, MutableVector, 2);

  stacki->index = 0;
  stacko->index = 0;

  emission->port    = (DM_Port*)connection;
  emission->cache   = cachei;
  emission->stack   = stacki;
  emission->mark    = DM_NULL;
  emission->index   = 0;
  emission->maximum = 0;
  emission->amount  = 0;
  emission->next    = DM_NULL;

  reception->port    = (DM_Port*)connection;
  reception->cache   = cacheo;
  reception->stack   = stacko;
  reception->mark    = DM_NULL;
  reception->index   = 0;
  reception->maximum = 0;
  reception->amount  = 0;
  reception->next    = DM_NULL;

  connection->retries    = DM_PORT_WRITE_RETRY_NUMBER;
  connection->options    = DM_CONNECTION_DEFAULT_OPTIONS;
  connection->port       = htons(remote_port);
  connection->address    = htonl(DM_get_ip_address(context, hostname, 0));
  connection->last_read  = 0;
  connection->last_write = 0;
  connection->reception  = reception;
  connection->emission   = emission;
  connection->emitqueue  = emission;
  connection->clocker    = 0;
  connection->site       = DM_NULL;
  if ( password ) {
    connection->password = DM_cstring_to_string(context, password, 0);
  } else {
    connection->password = DM_NULL;
  }

  /* Use the default protocol from the Site instance. */
  connection->protocol = DM_TheCurrentSite()->protocol;
  connection->handler  = DM_NULL;

  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, cachei);
    DM_SetSharable(context, cacheo);
    DM_SetSharable(context, emission);
    DM_SetSharable(context, reception);
    DM_SetSharable(context, connection);
    if ( connection->password) {
      DM_SetSharable(context, connection->password);
    }
  }
  
  return DM_establish_outgoing_connection(context, connection);
}

/* }}}
 * {{{ Closing Port

 * RESOURCE DM_finish_port
 * This function inactivates a file descriptor if not yet done. It does
 * not check whether there are still some bytes to read/write. This function
 * returns DM_TRUE if the port is effectively closed. Decode the options
 * field to know more about it. No closed Port still exists among the 
 * active_ports.

 * NOTE: Leaves all the opened connections in a server ?
 * NOTE: and if there are pending connections to be accepted ?
 */

DM_c_boolean
DM_finish_port (DM_Context *context,
                DM_Port    *port) {
  int sockfd = port->number;

  DM_DebugAtLevel(10,
    fprintf(DM_FindErrorStream(context), 
            "Finishing port " DM_0X "%p.\n", port);
);

  if ( DM_HasOption(port, DM_PORT_IS_CLOSED) ) {
    return DM_FALSE;
  } else {
    port->options &= DM_PORT_PERMANENT_OPTIONS;
    /* Don't check whether the readbuf and writebuf still contains
     * bytes unread or unsent. Just close the port. */
    DM_ResetErrno();
    if ( (close(sockfd) != 0) ) {

#ifdef DM_CLOSE_MAY_RETURN_EAGAIN
      /* This is the first OS-dependent patch!
       * On FreeBSD, close sometimes returns EAGAIN.
       * The port can no longer be closed (we get error EBADF if we
       * try to) meaning that the port is closed. Maybe the closure
       * is delayed ?  */
      if ( DM_IsErrnoEqualTo(DM_EAGAIN) ) 
        goto port_is_closed;
#endif

      DM_DebugAtLevel(10,
        fprintf(DM_FindErrorStream(context), 
                "Cannot close port " DM_0X "%p with errno=%d.\n", 
                port, DM_ErrnoValue() );
);

      DM_SetOption(port, DM_PORT_FAILED_CLOSE); 
      DM_MayAbort(context, DME_FAILED_CLOSE);
      return DM_FALSE;
    }

#ifdef DM_CLOSE_MAY_RETURN_EAGAIN
  port_is_closed:
#endif

    /* Remove this port from the active ones */
    DM_inactivate_port(context, sockfd);
    DM_SetOption(port, DM_PORT_IS_CLOSED);
    return DM_TRUE;
  }
}

/* }}}
 * {{{ Message layer
 *     {{{ Reception side
 
 * RESOURCE DM_rearrange_inbuffer
 * Modify an InBuffer to make it more suitable to receive missing parts
 * of message headers or content. The meaningful bytes are shifted
 * to the left and the buffer might grow if sufficiently full. If
 * newsize is 0, the buffer is just doubled. Returns the (possibly new) 
 * rearranged buffer.
 */

DM_InBuffer*
DM_rearrange_inbuffer (DM_Context  *context,
                       DM_InBuffer *inbuffer,
                       DM_nat3     newsize ) {
  DM_nat3 index = inbuffer->index;
  DM_nat3 i;
  
  if ( index != 0 ) {
    /* Shift the meaningful content of the buffer to the left to
     * remove the useless bytes before the beginning of a header. 
     * FUTURE: Should try to only perform a well-aligned bit-blit.      ???
     */
    char *p      = (char*)(inbuffer->byte);
    char *q      = (char*)(&(inbuffer->byte[index]));
    DM_nat3 left = inbuffer->maximum - index;

    DM_DebugAtLevel(5,
      if ( left>0 ) {
        fprintf(DM_FindErrorStream(context), 
                "Keeping %d characters in buffer " DM_0X "%p.\n", 
                left, inbuffer );
      }
);

    /* May probably use memcpy if correct on overlapping areas. */
    for ( i=left ; i>0 ; i-- ) *p++ = *q++;
    inbuffer->index   = 0;
    inbuffer->maximum -= index;
    return inbuffer;
  } else {

    /* Compute a sensible default for newsize if null. */
    newsize = (newsize>0) ? newsize : 2*(inbuffer->byte_length);

    /* Increase the size of the buffer if sufficiently full.  This
     * criterium is probably too rough, should be dependent on the
     * average size of a message.                               FUTURE
     */
    if ( inbuffer->maximum >= ((2*(inbuffer->byte_length))/3) ) {
      DM_InBuffer *new = DM_Allocate1(context, InBuffer, newsize);
        
      DM_DebugAtLevel(5,
        fprintf(DM_FindErrorStream(context),
                "Increased InBuffer " DM_0X "%p to %d\n", 
                new, newsize);
);

      new->port    = inbuffer->port;
      new->cache   = inbuffer->cache;
      new->stack   = inbuffer->stack;
      new->index   = inbuffer->index;
      new->maximum = inbuffer->maximum;
      new->amount  = inbuffer->amount;
      memcpy(new->byte, inbuffer->byte, inbuffer->maximum);
      if ( DM_IsA(context, inbuffer->port, Connection) ) {
        DM_DeclareNarrowedObject(connection, Connection, inbuffer->port);

        connection->reception = new;
      } else {
        DM_MayAbort(context, DME_NOT_CONNECTION);
      }
      if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
        DM_SetSharable(context, new);
      }
      return new;
    } else {
      /* There is still some room available in the current buffer,
       * wait for it to be filled. */
      return inbuffer;
    }
  }
}

/* RESOURCE DM_fill_buffer
 * This function fills an InBuffer associated to a file descriptor. It
 * detects EOF and sets the 'options' field of the associated Port
 * accordingly.  If there are some unread bytes they are shifted to
 * the head of the InBuffer. This function returns DM_TRUE whenever
 * some bytes were read.  It also tries to enlarge the InBuffer if
 * necessary.
 */

DM_c_boolean
DM_fill_buffer (DM_Context  *context,
                DM_InBuffer *inbuffer) {
  DM_Port *port   = inbuffer->port;
  DM_nat2 fd      = port->number;
  DM_nat3 maximum = inbuffer->maximum;
  DM_nat3 size    = inbuffer->byte_length;
  int     read_bytes;

  DM_DebugAtLevel(5,
    fprintf(DM_FindErrorStream(context),
            "Filling InBuffer " DM_0X "%p[%lu..%lu[.\n", 
            inbuffer, (long unsigned)maximum, (long unsigned)size );
);

  port->options &= DM_PORT_PERMANENT_OPTIONS;
  if ( DM_HasOption(port, DM_PORT_IS_DRIED) )
    /* port has reached EOF (or is supposed to), don't try to read now. */
    return DM_FALSE;

  /* Check if the InBuffer needs to be enlarged since it is worthless
   * to read bytes in a quasi-full buffer.  */
  if ( maximum > ((3*size)/4) ) {
    inbuffer = DM_rearrange_inbuffer(context, inbuffer, 0);
    return DM_fill_buffer(context, inbuffer);
  } else {
    /* There is some room for new bytes to be read. */
    DM_ResetErrno();
    read_bytes = read(fd, &(inbuffer->byte[maximum]), size-maximum);
    if ( read_bytes < 0 ) {
      if ( DM_IsErrnoEqualTo(DM_EINTR) ) {
        return DM_FALSE;
      } else if ( DM_IsErrnoEqualTo(DM_ECONNRESET) ) {
        /* ``Connection reset by peer'' is equivalent to EOF. */
        read_bytes = 0;
      } else {
        DM_SetOption(port, DM_PORT_FAILED_READ);
        DM_MayAbort(context, DME_FAILED_READ);
        return DM_FALSE;
      }
    }

    /* EOF reached: record it in the permanent options of the connection
     * and return False to signal it.  */
    if ( read_bytes == 0 ) {
      DM_SetOption(port, DM_PORT_IS_DRIED);

      DM_DebugAtLevel(5,
        fprintf(DM_FindErrorStream(context),
                "EOF reached on " DM_0X "%p\n", inbuffer);
);

      return DM_FALSE;
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context),
              "Filled InBuffer " DM_0X "%p with %u bytes.\n",
              inbuffer, read_bytes);
      DM_print_memory_slice(context, 
                            &(inbuffer->byte[maximum]),
                            read_bytes, 
                            "", "",
                            (DM_SupportPointer)(context->errstream) );
);

    inbuffer->maximum     = maximum + read_bytes;
    inbuffer->amount     += read_bytes;
    if ( DM_IsA(context, port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, port);

      connection->last_read = DM_get_date(context);
    }
    return DM_TRUE;
  }
}

/*     }}}
 *     {{{ Emission side

 * Sometimes a signal can be generated if there is no longer a peer able
 * to read a socket. In that case, this handler will be called.
 * Good authors tell that the simpler/shorter the handler is the better it
 * is.

 * NOTE: it is not clear to me, when this signal is received, how to
 * find which port it refers. Experiments on Sony News show that a TCP
 * "connection reset by peer" is still found readable and writable by
 * select.
 */

static DM_sigpipe_occurred = 0;

static void
DM_sigpipe_signal (int sigcode) {
  DM_sigpipe_occurred += 1;

  DM_DebugAtLevel(0,
    fprintf(DM_debug_stream, "One unknown Port got peerless.\n");
);

}

/* RESOURCE DM_flush_buffer 
 * This function tries to write the content of an OutBuffer onto the
 * associated file-descriptor. If it is not possible to write the
 * whole content then the 'maximum' field records up to where the
 * buffer was flushed. DM_TRUE is returned if <<all>> bytes were
 * successfully output.  The 'options' field of the associated Port is
 * enriched to allow you to decode the various events that occurred
 * while writing. You only need to decode it when the result is
 * DM_FALSE.
 */

DM_c_boolean
DM_flush_buffer (DM_Context   *context,
                 DM_OutBuffer *outbuffer) {
  DM_Port  *port  = outbuffer->port;
  DM_nat2  fd     = port->number;
  DM_char1 *bytes = &(outbuffer->byte[outbuffer->maximum]);
  DM_nat3  size   = outbuffer->index - outbuffer->maximum;
  int      written_bytes;

  DM_DebugAtLevel(5,
    fprintf(DM_FindErrorStream(context), 
            "Flushing OutBuffer " DM_0X "%p[%lu..%lu[.\n", 
            outbuffer, (long unsigned)(outbuffer->maximum), 
            (long unsigned)(outbuffer->maximum + size) );
);

  if ( outbuffer->mark ) {
    /* A message is being composed, don't emit its beginning. */
    size = outbuffer->mark->index - outbuffer->maximum;

    DM_DebugAtLevel(5,
      fprintf(DM_FindErrorStream(context),
              "Flushing while message is being composed.\n");
);

    return DM_FALSE;
  }

  port->options &= DM_PORT_PERMANENT_OPTIONS;
  DM_ResetErrno();

  if ( size == 0 )
    return DM_TRUE;

  /* FUTURE: could use writev to send also the subsequent buffers. */

  if ( (written_bytes = write(fd, bytes, size)) < 0 ) {
    DM_SetOption(port, DM_PORT_FAILED_WRITE);
    if ( DM_IsErrnoEqualTo(DM_EPIPE) )
      DM_SetOption(port, DM_PORT_IS_PEERLESS);
    return DM_FALSE;
  }
  
  DM_DebugAtLevel(5,
    fprintf(DM_FindErrorStream(context), 
            "Flushed %d bytes from OutBuffer " DM_0X "%p.\n",
            written_bytes, outbuffer);
);

  DM_DebugAtLevel(9,
    DM_print_memory_slice(context, 
                          bytes, 
                          written_bytes, 
                          "", "", 
                          (DM_SupportPointer)(context->errstream) );
);

  outbuffer->amount  += written_bytes;
  outbuffer->maximum += written_bytes;
  size               -= written_bytes;
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);

    connection->last_write = DM_get_date(context);
  }
  if ( size > 0 ) {
    DM_SetOption(port, DM_PORT_PARTIALLY_DRAINED);

    DM_DebugAtLevel(5,
      fprintf(DM_FindErrorStream(context), 
              "Still %u remaining bytes in OutBuffer " DM_0X "%p.\n",
              size, outbuffer);
);

      return DM_FALSE;
    }
  return DM_TRUE;
}

/* RESOURCE DM_rearrange_outbuffer
 * Shift non output bytes to the left of the OutBuffer to save room.
 */

static DM_OutBuffer*
DM_rearrange_outbuffer (DM_Context   *context,
                        DM_OutBuffer *outbuffer) {
  char         *p, *q;
  DM_nat3      size, i;

  p    = (char*)(&(outbuffer->byte[0]));
  q    = (char*)(&(outbuffer->byte[outbuffer->maximum]));
  size = outbuffer->index - outbuffer->maximum;
  /* Should take care of word-alignment to speed up bit-blitting. */
  for ( i=size ; i>0 ; i-- ) *p++ = *q++;
  outbuffer->index   -= outbuffer->maximum;
  outbuffer->maximum = 0;
  return outbuffer;
}

/* RESOURCE DM_handle_writable_connection
 * When a Port is found writable, try to flush the queue of waiting
 * OutBuffers. Return DM_TRUE if something was done.

 * This code is here since it is related to the management of OutBuffers.

 * FUTURE: is it possible for a buffer to be flushed with an unfinished 
 * message ??? This would be an error since messages may be aborted.
 */

DM_c_boolean
DM_handle_writable_connection (DM_Context    *context,
                               DM_Connection *connection) {
  DM_OutBuffer *outbuffer     = connection->emitqueue;
  DM_OutBuffer *currentbuffer = connection->emission;

  if (   ( outbuffer != currentbuffer )
      || (outbuffer->maximum < outbuffer->index) ) {

    /* There are some bytes to flush (or in a previous buffer or in
     * the last buffer). */
    if ( DM_flush_buffer(context, outbuffer) ) {
      /* The buffer was entirely flushed. */
      if ( outbuffer != currentbuffer ) {
        /* and not the last one, so abandon it! But don't forget to
         * accumulate statistics. */
        connection->emitqueue = (DM_OutBuffer*)(outbuffer->next);
        currentbuffer->amount += outbuffer->amount;
      } else {
        DM_rearrange_outbuffer(context, outbuffer);
      }
      return DM_TRUE;
    } else {
      /* The port was only partially_drained, just wait for the PORT
       * channel to become writable again. Rearrange OutBuffer if this
       * is the last one. */
      if ( outbuffer == currentbuffer ) {
        /* This is the last OutBuffer ie the one that is currently written in
         * by DM_print_char_star_in_outbuffer. */
        DM_rearrange_outbuffer(context, outbuffer);
      }
      /* There are still some bytes to flush: analyze why? */
      return DM_FALSE;
    }
  } else {
    /* There were no character to flush. Keep the connection alive if
     * inactive for some time but don't perturb any unfinished message. */
    if ( DM_HasOption(connection, DM_PORT_RUNS_DMEROON) ) {
      if ( outbuffer->mark ) {
        /* NOTHING: There's an unfinished message. 
         * FUTURE: But wait, this is not normal since messages should be 
         * written atomically. Raise a warning to elucidate that question. */
        DM_DebugAtLevel(5,
          fprintf(DM_FindErrorStream(context), "=== Message problem\n");
);
        DM_Abort(context, DME_MESSAGE_UNFINISHED);
        return DM_FALSE;
      } else {
        if ( (connection->last_write + DM_TheCurrentSite()->cyclic_duration)
            < DM_get_date(context) ) {
          /* Send the 0000 subcode to keep the connection alive. */
          outbuffer = DM_ensure_large_enough_outbuffer(context, outbuffer, 5);
          outbuffer->byte[(outbuffer->index)++] = 0;
          outbuffer->byte[(outbuffer->index)++] = 0;
          outbuffer->byte[(outbuffer->index)++] = 0;
          outbuffer->byte[(outbuffer->index)++] = 0;
          outbuffer->byte[(outbuffer->index)++] = '\n';

          DM_DebugAtLevel(7,
            fprintf(DM_FindErrorStream(context), "Ping %s:%lu\n",
                    connection->site->name,
                    (unsigned long)ntohs(connection->site->port) );
);

          /* Pinging does not count as a real action so return false.
           * But since the buffer is filled with 5 bytes, it must be
           * drained. */
          (void) DM_flush_buffer(context, outbuffer);
          return DM_FALSE;
        } else {
          /* NOTHING: the connection was recently written. */
          return DM_FALSE;
        }
      }
    } else {
      /* This is not a DMeroon protocol-ed connection. */
      return DM_FALSE;
    }
  }
}

/* RESOURCE DM_print_char_star_in_outbuffer
 * Insert a C string in an OutBuffer.
 * This function is the specialized method of the generic DM_print_char_star 
 * for an OutBuffer. It returns DM_TRUE if it succeeds inserting the string
 * in the OutBuffer, otherwise it returns DM_FALSE if an overflow buffer
 * was needed. Generally, you do not need to check these status.

 * NOTE: to print onto a Connection is better since it is easier to get
 * the last OutBuffer to print in.             FUTURE ?
 */

DM_c_boolean
DM_print_char_star_in_outbuffer (DM_Context   *context,
                                 char         *string, 
                                 DM_nat3      len,
                                 DM_OutBuffer *outbuffer) {
  DM_Port *port = outbuffer->port;
  
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    int           n = len ? len : strlen(string);
    int           firsts;
    DM_OutBuffer *outbuffer = connection->emission;

    /* Get the last buffer of the chain. */
    while ( outbuffer->next ) {
      outbuffer = (DM_OutBuffer*)(outbuffer->next);
    }

    /* Is the buffer large enough or not ? */
    if ( (outbuffer->index + n) <= outbuffer->byte_length ) {
      memcpy(&(outbuffer->byte[outbuffer->index]), string, n);

      DM_DebugAtLevel(10,
        { 
          DM_nat3 i;
          fprintf(DM_FindErrorStream(context), "[");
          for ( i=0 ; i<n ; i++ ) {
            fprintf(DM_FindErrorStream(context), "[%#hx]", 
                    outbuffer->byte[outbuffer->index+i] );
          }
          fprintf(DM_FindErrorStream(context), "]");
        }
);

      outbuffer->index += n;
      return DM_TRUE;
    } else {
      /* Fill completely the current buffer and allocate a new one to
       * hold the rest.  Chain it in the 'emitqueue' field of the
       * associated port. */
      DM_OutBuffer *new = 
        DM_ensure_large_enough_outbuffer(context, outbuffer, n);

      firsts = outbuffer->byte_length - outbuffer->index;
      memcpy(&(outbuffer->byte[outbuffer->index]), string, firsts);

      DM_DebugAtLevel(9,
        { 
          DM_nat3 i;
          fprintf(DM_FindErrorStream(context), "[");
          for ( i=0 ; i<firsts ; i++ ) {
            fprintf(DM_FindErrorStream(context), "[%#hx]", 
                    outbuffer->byte[outbuffer->index+i] );
          }
        }
);

      outbuffer->index += firsts;
      memcpy(&(new->byte[0]), string+firsts, n-firsts);

      DM_DebugAtLevel(9,
        { 
          DM_nat3 i;
          for ( i=0 ; i<n-firsts ; i++ ) {
            fprintf(DM_FindErrorStream(context), "[%#hx]", new->byte[i]);
          }
          fprintf(DM_FindErrorStream(context), "]");
        }
);

      new->index += n-firsts;
      connection->emission = new;
      return DM_FALSE;
    }
  } else {
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* RESOURCE DM_ensure_large_enough_outbuffer
 * Check whether an OutBuffer is large enough to hold a given number of bytes.
 * If this is not the case, then the OutBuffer is extended (with a new
 * larger buffer) and this new buffer is returned.
 */

DM_OutBuffer*
DM_ensure_large_enough_outbuffer (DM_Context   *context,
                                  DM_OutBuffer *outbuffer,
                                  DM_nat3      number) {
  /* Find the last OutBuffer of the linked list. */
  while ( outbuffer->next ) {
    outbuffer = (DM_OutBuffer*)(outbuffer->next);
  }

  if ( (outbuffer->index + number) <= outbuffer->byte_length ) {
    /* The OutBuffer is large enough. */
    return outbuffer;
  } else {
    DM_nat3      newlength   = ( number>2*(outbuffer->byte_length) )
                                ? number : 2*(outbuffer->byte_length);
    DM_OutBuffer *new        = DM_Allocate1(context, OutBuffer, newlength);

    DM_DebugAtLevel(15,
      fprintf(DM_FindErrorStream(context), 
              "Linked new OutBuffer " DM_0X "%p to current " DM_0X "%p\n",
              (void*)new, (void*)outbuffer );
);
    
    new->port       = outbuffer->port;
    if ( DM_IsA(context, outbuffer->port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, outbuffer->port);
      /* The Connection always refer to the OutBuffer into which bytes
       * may be printed. */
      connection->emission = new;
    }
    new->cache      = outbuffer->cache;
    new->stack      = outbuffer->stack;
    new->index      = 0;
    new->maximum    = 0;
    new->amount     = 0;
    new->next       = DM_NULL;
    /* propagate the marking state of the previous buffer. */
    new->mark       = outbuffer->mark;
    outbuffer->next = (DM_Buffer*)new;
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      DM_SetSharable(context, new);
    }
    return new;
  }
}

/* DOCUMENTATION Message layer
 * These functions define the message exchange protocol over tcp/ip connections.
 * It might be strange to have messages over a reliable byte stream but it
 * is necessary and more convenient for the causality of DMeroon that 
 * messages are exchanged atomically ie encoded/decoded atomically.

 * RESOURCE DM_start_message
 * This function means that a message is going to be recorded in an 
 * OutBuffer. It checks that no message is actually being sent and just
 * record where the size of the message will be recorded.
 */

DM_OutBuffer*
DM_start_message (DM_Context   *context,
                  DM_OutBuffer *outbuffer,
                  DM_c_boolean boolean ) 
{
  if ( outbuffer->mark ) {
    /* A message is already started, this is not normal! */
    DM_Abort(context, DME_MESSAGE_UNFINISHED);
    return DM_NULL;    
  } else {
    /* The OutBuffer must contain at least four bytes free to hold the
     * size of the meessage.  Initialize the size with 1 which is not
     * a legal size since the size must contain itself (that is, is at
     * least equal to 4). */
    DM_OutBuffer *last_outbuffer = 
      DM_ensure_large_enough_outbuffer(context, outbuffer, 4);

    (void) DM_mark_buffer(context, (DM_Buffer*)last_outbuffer);
    last_outbuffer->byte[(last_outbuffer->index)++] = 1;
    last_outbuffer->byte[(last_outbuffer->index)++] = 0;
    last_outbuffer->byte[(last_outbuffer->index)++] = 0;
    last_outbuffer->byte[(last_outbuffer->index)++] = 0;
    if ( boolean ) {
      /* While bootstrapping the Connection, don't send clocks. */
      DM_propagate_information(context, last_outbuffer);
    }
    return last_outbuffer;
  }
}

/* RESOURCE DM_mark_buffer
 * Record the current state of the Outbuffer in a Mark. This will be
 * used to compute lengths in linked buffers.
 */

DM_Marker*
DM_mark_buffer (DM_Context *context,
                DM_Buffer  *buffer )
{
  DM_Marker *marker = DM_Allocate0(context, Marker);

  marker->next   = buffer->mark;
  marker->buffer = buffer;
  marker->index  = buffer->index;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, marker);
  }
  buffer->mark = marker;
  return marker;
}

/* RESOURCE DM_size_since_mark
 * Compute the number of bytes present in a (linked list of) Buffers
 * starting from a Marker. 
 *
 * NOTE: Markers may also be linked. The head buffer may not know all
 * markers.  The last buffer knows them all.
 *
 * FIGURE
 *      +-----------------------------------(next)------------+
 *      V                                                     |
 *   Marker       Buffer                                      |
 *   |    |<(mark)|+   |                                      |
 *   | +-(buffer)>|    |                                      |
 *   +----+       | +->|(next)-> Buffer           Marker      |
 *                +----+         | +--|-(mark)--->|    |      |
 *                               |    | <-(buffer)|--+ |      |
 *                               +----+           |  +-|------+
 *                                                +----+
 * ENDFIGURE
 */

DM_nat3
DM_size_since_mark (DM_Context *context,
                    DM_Marker  *marker )
{
  DM_nat3   total = 0;
  DM_nat3   index   = marker->index;
  DM_Buffer *buffer = marker->buffer;

  do {
    total += buffer->index - index;
    index = 0;
    buffer = buffer->next;
  } while ( buffer );
    return total;
}

/* RESOURCE DM_patch_size
 * Patch a size under a mark. We cannot use DM_send_nat3 
 * since it increments the index of the OutBuffer.
 */

void
DM_patch_size (DM_Context *context,
               DM_Marker  *marker,
               DM_nat3    size )
{
  DM_DeclareNarrowedObject(outbuffer, OutBuffer, marker->buffer);
  DM_nat3 index = marker->index;

  outbuffer->byte[index++] = size % 256;
  size /= 256;
  outbuffer->byte[index++] = size % 256;
  size /= 256;
  outbuffer->byte[index++] = size % 256;
  size /= 256;
  outbuffer->byte[index] = size % 256;
}

/* RESOURCE DM_cancel_message
 * Cancel the current message if any. This is used by the remote failure
 * mechanism (see DM_report_failure). Don't forget to reset the cache if 
 * necessary to resynchronize it.
 */

void
DM_cancel_message (DM_Context   *context,
                   DM_OutBuffer *outbuffer) {
  DM_Marker *marker;

  /* Find the last OutBuffer of the linked list. */
  while ( outbuffer->next ) {
    outbuffer = (DM_OutBuffer*)(outbuffer->next);
  }
  if ( outbuffer->mark ) {
    /* A message is already started. */
    marker = outbuffer->mark;

    /* Find the last marker then the first buffer. */
    while ( marker->next ) {
      marker = marker->next;
    }
    outbuffer        = (DM_OutBuffer*)(marker->buffer);
    outbuffer->index = marker->index;
    outbuffer->mark  = DM_NULL;
    /* Don't reuse the linked buffers. */
    outbuffer->next  = DM_NULL;
  } else {
    /* NOTHING to cancel. */
  }
}

/* RESOURCE DM_finish_message
 * Mark that the message is now totally written into the OutBuffers.
 * Computes then stores the size of a message. The message starts at
 * the recorded position until the last char of the last linked buffer. 
 * You may give the buffer where the message started or the one where it
 * stopped.
 */

void
DM_finish_message (DM_Context   *context,
                   DM_OutBuffer *outbuffer,
                   DM_Site      *site )
{
  DM_OutBuffer *first_outbuffer;
  DM_Marker    *marker;
  DM_nat3      index;
  DM_nat3      total = 0;

  /* Find the last OutBuffer of the linked list. */
  while ( outbuffer->next ) {
    outbuffer = (DM_OutBuffer*)(outbuffer->next);
  }
  /* There should be only one mark. */ 
  marker = outbuffer->mark;
  /* Get the Buffer and the position where the message started if any. */
  if ( ! marker ) {
    DM_Abort(context, DME_MESSAGE_UNBEGUN);
  }

  DM_DebugAtLevel(8,
    if ( marker->next ) {
      DM_Abort(context, DME_UNFINISHED_MARK);
    }
);
  
  first_outbuffer = (DM_OutBuffer*)(marker->buffer);
  index           = marker->index;

  /* Output the epilogue. The size of the message contains it.
   * Caution, this may create a new buffer. */
  DM_print_char_star_in_outbuffer(context, "\n", 1, outbuffer);
  
  /* Compute the size of the message and reset the "message started" flag. */
  total = DM_size_since_mark(context, marker);
  DM_patch_size(context, marker, total);
  outbuffer = first_outbuffer;
  while ( outbuffer ) {
    outbuffer->mark = DM_NULL;
    outbuffer = (DM_OutBuffer*)(outbuffer->next);
  }

  /* Record statistics in the target site. Site may be NULL when
   * handshaking with an unknown site (in DM_site_handshake). */
  if ( site ) {
    site->amount += total;
  }

  DM_DebugAtLevel(8,
    fprintf(DM_FindErrorStream(context), " \tMessage of size %lu\n", 
            (unsigned long)total );
    outbuffer = first_outbuffer;
    DM_print_memory_slice(context, 
                          &(outbuffer->byte[index]), 
                          outbuffer->index - index, 
                          "", "", 
                          (DM_SupportPointer)(context->errstream) );
    outbuffer = (DM_OutBuffer*)(outbuffer->next);
    while ( outbuffer ) {
      DM_print_memory_slice(context, 
                            &(outbuffer->byte), 
                            outbuffer->index, 
                            "", "", 
                            (DM_SupportPointer)(context->errstream) );
      outbuffer = (DM_OutBuffer*)(outbuffer->next);
    }
    /* Give a view of the cache */
    if ( DM_IsA(context, first_outbuffer->cache, BoundedStack) ) {
      DM_DeclareNarrowedObject(cache, BoundedStack, first_outbuffer->cache);
      DM_nat3 i;

      fprintf(DM_FindErrorStream(context), 
              " \t Content of the cache " DM_0X "%p[0..%lu]\n", 
              cache, (unsigned long)(cache->index) );
      for ( i=0 ; i<cache->index ; i++ ) {
        DM_Object *o = cache->item[i];  

        if ( DM_IsObject(o) ) {
          DM_Class *class = DM_ObjectToClass(context, o);

          fprintf(DM_FindErrorStream(context), " [%lu]: " DM_0X "%p (a %s)\n", 
                  (unsigned long)i, 
                  o, 
                  DM_predefined_name(context, (DM_Object*)class) );
        }
      }
      fprintf(DM_FindErrorStream(context), " \t End of cache \n");
    }
);

}

/*     }}}
 * }}}
 * {{{ initialization
 */

/*
 * Initializes the module.

 * Verify that DMeroon_sockaddr_in faithfully implements sockaddr_in.
 * Not found a way to use the names of the fields with sizeof.
 * So I compare the DMeroon_sockaddr_in definition (in dmnet.h) with
 * the definition of sockaddr_in as given in socket.
 */

void
DM_net_initialize (DM_Context *context) 
{
  DM_Site *site = DM_TheCurrentSite();

  DM_LinkVersionStringIntoSite(dmnet);

  /* Check the DMeroon_sockaddr_in structure definition. 
     Remember that there are 8 useless bytes at the end of a sockaddr_in. */

#define DM_VerifySockAddr(field1,type1,field2,type2) \
  if ( sizeof(type1) != sizeof(type2) ) \
    DM_Abort(context, DME_INCORRECT_SOCKADDRIN_STRUCT)

  if ( sizeof(DMeroon_sockaddr_in) != sizeof(struct sockaddr_in) )
    DM_Abort(context, DME_INCORRECT_SOCKADDRIN_SIZE);
  DM_VerifySockAddr(family,  DM_nat2, sin_family, short);
  DM_VerifySockAddr(port,    DM_nat2, sin_port,   u_short);
  DM_VerifySockAddr(address, DM_nat3, sin_addr,   struct in_addr);
  /* verify that the DMeroon_sockaddr_in has the correct length in Tcp */
  if ( (DM_OffsetOf(DM_Tcp,zero) 
        + sizeof(DM_nat4) /* zero */
        - DM_OffsetOf(DM_Tcp,family) ) != sizeof(struct sockaddr_in) )
    DM_Abort(context, DME_INCORRECT_SOCKADDRIN_STRUCT);

  /* Check whether inet constants are the same in host-order. */
  if ( INADDR_NONE != ntohl(INADDR_NONE) )
    DM_Abort(context, DME_INADDR_NONE_DEPENDENCY);
  if ( INADDR_ANY != ntohl(INADDR_ANY) )
    DM_Abort(context, DME_INADDR_ANY_DEPENDENCY);
  
  /* Initialize the vector of active connections and insert it in the
   * site object.  */
  { DM_ports_vector *active_ports = 
      (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));
 
    if ( FD_SETSIZE < 3 )
      DM_Abort(context, DME_FD_SETSIZE_DEPENDENCY);
    site->ports = (DM_Object*)active_ports;
  }

  /* Initialize the fdsets for DM_analyze_all_ports. */
  FD_ZERO(&DM_current_ports_to_read);
  FD_ZERO(&DM_current_ports_to_write);
  FD_ZERO(&DM_last_ports_to_read);
  FD_ZERO(&DM_last_ports_to_write);

  /* Trap SIGPIPE occurring for sockets. */
  if ( (signal(SIGPIPE, DM_sigpipe_signal)) == BADSIG ) 
    DM_Abort(context, DME_FAILED_SIGPIPE_HANDLER);

  /* Computes the IP address of 'localhost'. Don't care if this
   * variable is set to INADDR_NONE. The result is in network order.*/
  { struct hostent *hp = gethostbyname("localhost");
    if ( hp ) {
      unsigned long *addr = (unsigned long*)(hp->h_addr_list[0]);
      if ( addr ) {
        DM_cxxvii_0_0_i = *addr;
      } else {
        DM_cxxvii_0_0_i = INADDR_NONE;
      }
    } else {
      DM_cxxvii_0_0_i = INADDR_NONE;
    }
  }

  { /* Create the default server.  It is possible to specify the port
     * to serve via the system environment. In either case the
     * environment is extended to contain a binding for
     * DMEROON_SERVING_PORT. If that number is zero, do not create a
     * server. */
    DM_Server *server; 
    DM_nat2   portnum = DM_DMEROON_FAVORITE_PORT;
    char      *portnumstr;

    if ( (portnumstr = getenv("DMEROON_SERVING_PORT")) ) {
      portnum    = DM_string_to_nat3(context, portnumstr, NULL);
    }
    site->port = htons(portnum);
    if ( portnum ) {
      /* Record the identity of the DMeroon server in that variable.
       * It will be only visible from forked processes if any. */
      char dmeroonid[255];

      sprintf(dmeroonid, "%s:%d", site->name, portnum);
      if ( DM_setenv(context, "DMEROON", dmeroonid) != 0 ) 
        DM_Abort(context, DME_CANNOT_SETENV);
      server = DM_create_server(context, portnum, DM_DMEROON_FAVORITE_BACKLOG);
      if ( server ) {
        fprintf(DM_FindErrorStream(context), "DMeroon's URL http://%s/\n", dmeroonid);
      } else {
        DM_Abort(context, DME_CANNOT_CREATE_SERVER);
      }
    } else {
      /* A null port number means do not create a Server. */
    }
  }
}

/* }}}
 * end of dmnet.c */
