/* $Id: dmnet.h,v 1.69 1998/05/06 20:03:15 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

#ifndef DM_NET_H
#define DM_NET_H

/*
 * This is the header file for the basic network part of DMeroon.
 */

/* DOCUMENTATION
 * The universal DMeroon site is defined by these two macros.  From this
 * DMeroon site, you can update your sources and get other valuable
 * objects.
 */

#define DM_ORIGINAL_REPOSITORY_PORT 	56423

#if defined(SITE_is_blaye)

#define DM_ORIGINAL_REPOSITORY_NAME 	"blaye.inria.fr"

#else

#define DM_ORIGINAL_REPOSITORY_NAME 	"cornas.inria.fr"

#endif

/* RESOURCE DM_DMEROON_FAVORITE_PORT
 * By default, DMeroon uses the not yet well known port: 56423.
 */

#define DM_DMEROON_FAVORITE_PORT 	56423
#define DM_DMEROON_FAVORITE_BACKLOG	5

/* RESOURCE DM_DMEROON_FAVORITE_PERIOD 
 * After that duration, the server will automatically exit.
 */

#define DM_DMEROON_FAVORITE_PERIOD	300

/* RESOURCE DM_CYCLIC_DURATION
 * This is the frequency of the DMeooon event loop. No wait is longer than
 * this one since there are some basic management tasks to run with this
 * frequency.
 */

#define DM_CYCLIC_DURATION 	15 /* seconds */

/* RESOURCE DM_NORMAL_DURATION
 * This is an internal constant used in DM_work when waiting for
 * the answer of a request. Normally 30 seconds should suffice for a
 * request to cross the world (twice) and get answered.
 */

#define DM_NORMAL_DURATION 	30 /* seconds */

/* RESOURCE DM_ABNORMAL_DURATION
 * The abnormal delay means that waiting more than that is really 
 * abnormal and should raise a time-out.
 */

#define DM_ABNORMAL_DURATION 	120 /* seconds */

/* DOCUMENTATION
 * The initial size of buffers is equal to 1K
 * Since I must build the C type of Buffers with such a size, I must
 * explicitly write by hand the definitions that use the value of 
 * BUFSIZ since I am not sure that it is possible to use the value 
 * of a C macro within a call to a C macro.

 * NOTE: BUFSIZ is often 1024 but Alpha defines it to 8192 and SGI to 4096.
 * These are quite a lot, should probably use a Hashtable to record Ports.

 * NOTE: It is unclear for me that BUFSIZ is appropriate when writing on 
 * a socket. So don't bother with BUFSIZ and takes 1024 on all machines.
 */

#define DM_INITIAL_BUFFER_SIZE 1024

/******************************************************************
 *                   Additional types
 *
 * They are defined so they can be understood as DMeroon instances.
 * DM_INITIAL_BUFFER_SIZE is the normal size of port buffers.
 *
 * First, mimic a sockaddr_in with a DMeroon object. The conformity of
 * this fake structure is checked in DM_net_initialize. There should
 * be 8 null bytes at the end (maybe they are used by the kernel).
 */

typedef struct DMeroon_sockaddr_in {
  DM_nat2    family;
  DM_netnat2 port;
  DM_netnat3 address;
  DM_nat4    zero;
} DMeroon_sockaddr_in;

/*
 * Port, Tcp, Connection
 * number is the field-descriptor (fd) number associated to the stream
 * family,port,address form the content of the sockaddr_in structure.
 * NOTE: caution, the sockaddr_in structure must be correctly aligned.
 * This structure will be inherited by Server and Connection classes.
 * There is a constant that says how many time, an unsuccessful 
 * write is retried. This constant is put in a mutable field so it can
 * be changed by the user.
 */

#define DM_PORT_WRITE_RETRY_NUMBER 5

/* DOCUMENTATION Port options
 * To be machine independent, all errors are coded as bits in options.
 * These errors are automatically set by the functions that apply on 
 * Port, Buffers etc.
 *
 * DM_PORT_IS_WRITABLE/READABLE is determined and set by select.
 * DM_PORT_IS_DRIED means that the port cannot deliver bytes any longer 
 *    ie an EOF was found while reading on it.
 * DM_PORT_IS_PEERLESS means that no one is at the other side 
 *   (for a Port or a pipe) ie you cannot write any longer to it.
 * DM_PORT_TO_READ/WRITE records the way the Port is used.
 * DM_PORT_USES_PROTOCOL qualifies an Icslas port that is read using
 * protocols and not by the read function.
 *
 * DM_PORT_FAILED_something is for errors.
 */

#define DM_PORT_DEFAULT_OPTIONS		0

#define DM_PORT_TO_READ			1<<0
#define DM_PORT_TO_WRITE		1<<1

#define DM_PORT_IS_WRITABLE             1<<2 
#define DM_PORT_IS_READABLE             1<<3
#define DM_PORT_IS_DRIED		1<<4
#define DM_PORT_IS_PEERLESS		1<<5
#define DM_PORT_IS_CLOSED		1<<6
#define DM_PORT_RUNS_DMEROON		1<<7

#define DM_PORT_FAILED_SETSOCKOPT	1<<9
#define DM_PORT_FAILED_RETRIED_WRITE	1<<10
#define DM_PORT_FAILED_WRITE		1<<11
#define DM_PORT_FAILED_READ		1<<12
#define DM_PORT_FAILED_BIND		1<<13
#define DM_PORT_FAILED_LISTEN		1<<14
#define DM_PORT_FAILED_ACCEPT		1<<15
#define DM_PORT_FAILED_CLOSE		1<<16
#define DM_PORT_FAILED_CONNECT		1<<17
#define DM_PORT_FAILED_SOCKET		1<<18

#define DM_PORT_PARTIALLY_DRAINED	1<<20

#define DM_IP_FAILED_GETHOSTBYNAME	1<<21

#define DM_PORT_WITH_RECURSIVE_ENCODING	1<<24
#define DM_PORT_WITHOUT_PREFETCHING	1<<25
#define DM_PORT_WITHOUT_CACHING		1<<26

#define DM_CONNECTION_DEFAULT_OPTIONS	\
   ( DM_PORT_DEFAULT_OPTIONS | \
     DM_PORT_WITHOUT_CACHING )                /* TEMP */

/* DOCUMENTATION DM_PORT_PERMANENT_OPTIONS
 * Some of these conditions are permanent: once set, they are permanent.
 * In fact, this is not exactly the case of DM_PORT_IS_WRITABLE which may
 * be more exactly qualified as volatile. The non permanent options are
 * reset before any read/write operations and serve as a means to report
 * problems.
 */

#define DM_PORT_PERMANENT_OPTIONS \
	( DM_PORT_TO_READ \
        | DM_PORT_TO_WRITE \
        | DM_PORT_IS_READABLE \
        | DM_PORT_IS_WRITABLE \
        | DM_PORT_IS_DRIED \
	| DM_PORT_IS_PEERLESS \
        | DM_PORT_IS_CLOSED \
        | DM_PORT_RUNS_DMEROON \
        | DM_PORT_WITHOUT_PREFETCHING \
        | DM_PORT_WITHOUT_CACHING \
         )

/*
 * Active connections are recorded in a DMeroon Vector instance.
 */

typedef struct DM_ports_vector {
  DM_nat3      port_length;
  DM_reference port[FD_SETSIZE];
} DM_ports_vector;

/*
 * Remember that all these types must be prefixed.
 */

DM_DefineWrappedType(DM_ports_vector);

extern fd_set DM_current_ports_to_read;
extern fd_set DM_current_ports_to_write;
extern fd_set DM_last_ports_to_read;
extern fd_set DM_last_ports_to_write;
extern unsigned DM_max_port_number;

/*******************************************************************
 *             Additional externalizations
 */

extern DM_ports_vector_wrapped DM_active_ports_object;

extern void
DM_activate_port (DM_Context *context,
                  DM_nat2    sockfd,
                  DM_Port    *port,
                  DM_int2    io );

extern void
DM_inactivate_port (DM_Context *context,
                    DM_nat2    sockfd);

extern DM_Server*
DM_establish_server (DM_Context *context,
                     DM_Server  *server);

extern DM_Server*
DM_create_server (DM_Context *context,
                  DM_nat2    port, 
                  DM_nat2    backlog);

extern DM_Connection*
DM_establish_incoming_connection (DM_Context    *context,
                                  DM_Server     *server,
                                  DM_Connection *connection);

extern DM_Connection*
DM_accept_connection (DM_Context *context,
                      DM_Server  *server);

extern DM_nat3 
DM_get_ip_address (DM_Context *context,
                   char       *remote_hostname, 
                   DM_nat3    choice );

extern DM_Connection*
DM_establish_outgoing_connection (DM_Context    *context,
                                  DM_Connection *connection);

extern DM_Connection*
DM_create_connection (DM_Context *context,
                      char       *hostname,
                      DM_nat2    remote_port,
                      char       *password );

extern DM_c_boolean
DM_finish_port (DM_Context *context,
                DM_Port    *port);

extern DM_InBuffer*
DM_rearrange_inbuffer (DM_Context  *context,
                       DM_InBuffer *inbuffer,
                       DM_nat3     newsize );

extern DM_c_boolean
DM_fill_buffer (DM_Context  *context,
                DM_InBuffer *inbuffer);

extern DM_c_boolean
DM_flush_buffer (DM_Context   *context,
                 DM_OutBuffer *outbuffer);

extern DM_c_boolean
DM_handle_writable_connection (DM_Context    *context,
                               DM_Connection *connection);

extern DM_c_boolean
DM_print_char_star_in_outbuffer (DM_Context   *context,
                                 char         *string, 
                                 DM_nat3      len,
                                 DM_OutBuffer *outbuffer);

extern DM_OutBuffer*
DM_ensure_large_enough_outbuffer (DM_Context   *context,
                                  DM_OutBuffer *outbuffer,
                                  DM_nat3      number);

extern DM_Marker*
DM_mark_buffer (DM_Context *context,
                DM_Buffer  *buffer );

extern DM_nat3
DM_size_since_mark (DM_Context *context,
                    DM_Marker  *marker );

extern void
DM_patch_size (DM_Context *context,
               DM_Marker  *marker,
               DM_nat3    size );

extern DM_OutBuffer*
DM_start_message (DM_Context   *context,
                  DM_OutBuffer *outbuffer,
                  DM_c_boolean boolean );

extern void 
DM_cancel_message (DM_Context   *context,
                   DM_OutBuffer *outbuffer);

extern void
DM_finish_message (DM_Context   *context,
                   DM_OutBuffer *outbuffer,
                   DM_Site      *site);

extern void 
DM_net_initialize (DM_Context *context);

#endif /* DM_NET_H */

/* end of dmnet.h */
