/* $Id: dmnewapi.c,v 1.62 1998/06/01 19:30:44 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmnewapi.c
 * This file defines the new API of DMeroon.
 * This new API is more configurable and thread-safe. Not to interfere
 * for the moment, API functions are prefixed with DMeroon instead of DM.
 *
 * These functions always return an error code that if non null means that
 * some problem was encountered. These functions also ``return'' values
 * storing them in addresses that are given as arguments (generally the 
 * second argument). In the following, variables immediately prefixed 
 * with a star are return values. In case of errors, the content of these
 * return-variables is undefined and should not be used.
 * 
 * General points:
 *
 * The context argument is assumed to be NULL or an instance of Context.
 * 
 * These primitives are generally implemented with two C functions: a
 * careful one that check the received arguments and a careless
 * one. The checks are themselves divided into two parts: checking the
 * DMeroon-ity of the objects and checking their classes. There are
 * two separate options in the Context to specify this behavior.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmnewapi,
  "$Id: dmnewapi.c,v 1.62 1998/06/01 19:30:44 queinnec Exp $");

/* {{{ Initialization

 * RESOURCE DMeroon_initialize 
 * Initialize DMeroon library and should be called before any other use
 * of the DMeroon library. It is possible to give NULL as second
 * argument to ignore the returned current site. Most often this primitive
 * is also invoked with a NULL context since one cannot create dynamically 
 * such a Context before DMeroon is initialized.
 *
 * This is a non-blocking primitive.

 * The Context is not used by this function. There might be additional
 * arguments (argc,argv) or they may be passed through the Context in
 * some future release.                                      FUTURE
 */

DM_ErrorCode
DMeroon_initialize (DM_Context* context,
                    DM_Site*    *returned_site ) 
{
  /* Allocate statically an instance of Context since most often 
   * context is NULL here. This object will be filled below. */
  DM_DefineStaticContext(newcontext);

  /* Call DMeroon_initialize only once! */
  static DM_nat1 already_done = DM_FALSE;

  if ( !already_done ) {
    already_done = DM_TRUE;
    /* Use a non-null context if none is given. */
    if ( context == NULL ) {
      context = newcontext;
    }
    /* Call the initialization of the support language. 
    * This will directly abort in case of problems. */
    DM_INITIALIZE(context);
  }

  if ( returned_site )
    *returned_site = DM_TheCurrentSite();
  return DME_SUCCESS;
}

/* }}}
 * {{{ Structure accessor

 * RESOURCE DMeroon_is_a
 * Test if a DMeroon object is an instance of a class. The returned boolean
 * is a C boolean (see constants DM_TRUE and DM_FALSE).
 * 
 * This is a non-blocking predicate. The class of a present object is
 * always present.
 */

DM_ErrorCode
DMeroon_is_a (DM_Context*  context,
              DM_c_boolean *returned_boolean,
              DM_Object*   object,
              DM_Class*    class ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(class)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, class, Class)) ) {
      return DME_NOT_CLASS;
    }
  }

  { DM_Class *oc = DM_ObjectToClass(context, object);
  
    if ( oc == class ) {
      *returned_boolean = DM_TRUE;
    } else {
      DM_nat3 depth = class->super_length;
    
      if (   ( depth <= oc->super_length )
          && ( class == oc->super[depth-1] ) ) {
        *returned_boolean = DM_TRUE;
      } else {
        *returned_boolean = DM_FALSE;
      }
    }
  }
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_is_a_subclass
 * Test if a DMeroon class is a subclass of another one. The returned boolean
 * is a C boolean (see constants DM_TRUE and DM_FALSE).
 * 
 * This is a non-blocking predicate. The superclasses of a present class 
 * are always present.
 */

DM_ErrorCode
DMeroon_is_a_subclass (DM_Context*  context,
                       DM_c_boolean *returned_boolean,
                       DM_Class*    subclass,
                       DM_Class*    class ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(subclass))
          || !(DM_IsObject(class)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, subclass, Class)) ) {
      return DME_NOT_CLASS;
    }
    if ( !(DM_IsA(context, class, Class)) ) {
      return DME_NOT_CLASS;
    }
  }

  if ( subclass == class ) {
    *returned_boolean = DM_TRUE;
  } else {
    DM_nat3 depth = class->super_length;
    
    if (   ( depth <= subclass->super_length )
           && ( class == subclass->super[depth-1] ) ) {
      *returned_boolean = DM_TRUE;
    } else {
      *returned_boolean = DM_FALSE;
    }
  }
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_class_of
 * Returns the class of an object.
 *
 * This is a non-blocking function. The class of a present object is 
 * always present.
 */

DM_ErrorCode
DMeroon_class_of (DM_Context* context,
                  DM_Class*   *returned_class,
                  DM_Object*  object ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  *returned_class = DM_ObjectToClass(context, object);
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_site_of
 * Return the Site owning a DMeroon object.
 * 
 * This primitive may block since a present object may have a non-usable site.
 */

DM_ErrorCode
DMeroon_site_of (DM_Context* context,
                 DM_Site*    *returned_site,
                 DM_Object*  object ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  DM_ProtectWithContext(context, { 
       DM_Object *proxy = DM_ObjectToProxy(context, object);
       
       if ( DM_IsA(context, proxy, Exit) ) {
         /* The object is remote. */
         DM_DeclareNarrowedObject(exit, Exit, proxy);
         
         DM_ensure_site_usable(context, exit->owner);
         *returned_site = exit->owner;
       } else {
         /* The object is local. */
         *returned_site = DM_TheCurrentSite();
       }
     },
       /* Nothing even if something went wrong, just continue in sequence. */
);
  return context->errcode;
}

/* RESOURCE DMeroon_clock_of
 * Return the clock (if any) that monitors an object.
 * Returns DM_NULL if not monitored by a clock.
 * 
 * This  primitive does not block.
 */

DM_ErrorCode
DMeroon_clock_of (DM_Context* context,
                  DM_Clock*   *returned_clock,
                  DM_Object*  object ) 
{
  volatile DM_ErrorCode errcode = DME_SUCCESS;

  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  { 
    DM_Object *proxy = DM_ObjectToProxy(context, object);

    if ( DM_IsA(context, proxy, Exit) ) {
      /* The object is remote. */
      DM_DeclareNarrowedObject(exit, Exit, proxy);

      *returned_clock = exit->clock;
    } else if ( DM_IsA(context, proxy, Entry) ) {
      /* The object is local and has an associated Entry item. */
      DM_DeclareNarrowedObject(entry, Entry, proxy);

      *returned_clock = entry->clock;
    } else {
      /* The object is local and not externalized. */
      *returned_clock = DM_NULL;
    }
  }
  return errcode;
}

/* }}}
 * {{{ Communications

 * RESOURCE DMeroon_receive
 * Receive an object. This primitive may wait forever. This is the primitive
 * normally used by an application that implements a server: it waits for
 * objects to come (with DMeroon_receive) and handle them appropriately. 
 */

static DM_ErrorCode
DM_do_receive (DM_Context* context,
               DM_Object*  *returned_object,
               DM_nat3     duration ) 
{
  DM_Site   *site    = DM_TheCurrentSite();
  DM_Object *o;
  
received_object:
  o = DM_remove_last_queuelink(context, &(site->objects));
  if ( o ) {
    DM_ensure_present(context, &o);
    *returned_object = o;
    return DME_SUCCESS;
  } else {
    /* No object is present. */
    *returned_object = DM_NULL;
    if ( site->handler == DM_CodeOf(DM_accumulate_object) ) {
      /* This handler accumulates incoming objects in site->objects. */
      /* The queue of already received objects is empty.  */
      DM_ErrorCode errcode = 
        DM_wait_until_object_reception(context, 
                                       duration,
                                       (DM_Object**)(&(site->objects)) );

      if ( errcode == DME_SUCCESS ) {
        /* An object was received, handle it. */
        goto received_object;
      } else {
        /* Nothing happened during the event loop, return that no object
         * was received while waiting at most duration.  */
        return errcode;
      }
    } else {
      /* There is a receiver function installed by the user so it is
       * hopeless to wait for objects. */
      return DME_CANNOT_RECEIVE;
    }
  }
}

DM_ErrorCode
DMeroon_receive (DM_Context* context,
                 DM_Object*  *returned_object,
                 DM_nat3     duration ) 
{
  DM_ProtectWithContext(context,
       context->errcode = DM_do_receive(context, returned_object, duration),
     /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* RESOURCE DMeroon_install_receiver
 * Any incoming object will be given to the function specified in argument
 * to DMeroon_install_receiver. This function has void (*)(context, object)
 * as signature.
 */

DM_ErrorCode
DMeroon_install_receiver (DM_Context* context,
                          DM_code     function ) 
{
  DM_Site *site = DM_TheCurrentSite();

  if ( function ) {
    site->handler = function;
    /* Already arrived objects will be handled by the server loop. */
  } else {
    /* a NULL function means that we reset the handling of incoming
     * objects to the default ie accumulate them in site->objects. */
    site->handler = DM_CodeOf(DM_accumulate_object);
  }
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_serve
 * Let the DMeroon server manage its space for at least duration.
 * Return true if some requests were served.
 */

DM_ErrorCode
DMeroon_serve (DM_Context*  context,
               DM_nat3      duration ) 
{
  DM_ProtectWithContext(context,
    (void) DM_wait_until_duration(context, duration),
    /* Nothing even if something went wrong. */
);
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_send
 * Send the reference of an object towards a site. 
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_send (DM_Context* context,
            DM_Object*  o,
            DM_Site*    site ) 
{
  if ( site == DM_TheCurrentSite() ) {
    /* Send an object to the current site is easy, we just enqueue it.
     * Don't signal object reception, DM_receive will take that burden. */
    DM_enqueue_received_object(context, o);
  } else {
    /* Send a reference onto the object so the receiver can ask for it. */
    DM_send_object_send_request(context, o, site);
  }
  return DME_SUCCESS;
}

DM_ErrorCode
DMeroon_send (DM_Context* context,
              DM_Object*  object,
              DM_Site*    site ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(site)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, site, Site)) ) {
      return DME_NOT_SITE;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_send(context, object, site),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* RESOURCE DM_connect
 * Merge the given DMeroon Server with the current DMeroon space. This
 * function returns the Site instance describing the remote server if 
 * reachable.
 *
 * This primitive may block.
 *
 * The Site instance holds an Hashtable that contains all the known
 * sites.  This Hashtable maps ip# and port# (in host order) towards
 * instances of Site.

 * FUTURE password to be taken into account               FUTURE
 */

static DM_ErrorCode
DM_do_connect (DM_Context* context,
               DM_Site*    *returned_site,
               char*       hostname,
               DM_nat2     port_number,
               char*       password ) 
{
  DM_nat3 ip = DM_get_ip_address(context, hostname, 0);

  if ( ip == ntohl(INADDR_NONE) ) {
    /* No IP number for that host. 
     * Should try the other choices.    FUTURE */
    return DME_UNKNOWN_SITE;
  } else {
    DM_Site       *site = DM_extract_site(context, ip, port_number);
    DM_Connection *connection;

    if ( site ) {
      *returned_site = site;
      /* The Site was already known. */
      if ( site == DM_TheCurrentSite() ) {
        /* This is the current site, return it immediately. There is no
         * need to check whether there is an open connection towards it. */
        return DME_SUCCESS;
      } else {
        /* The goal of DM_connect is to manage/create a direct
         * connection to the remote site. Check if there is already
         * one but check before if the site instance is usable. */
        DM_ensure_site_usable(context, site);
        if ( site->connection ) {
          /* There is a direct Connection to that site.  It is the
           * duty of the event loop to remove closed Connection. */
          return DME_SUCCESS;
        } else {
          /* No connection exists towards this Site: open one! */
          goto open_new_connection;
        }
      }
    } else {
      /* There is no such already known Site, establish a connection
       * to that host:port and wait for the Site to come. */
    open_new_connection:

      connection = DM_create_connection(context, hostname, port_number, 
                                        password );
      if ( connection ) {
        if ( DM_site_handshake(context, connection) == DME_SUCCESS ) {
          DM_ensure_site_usable(context, connection->site);
          *returned_site = connection->site;
          return DME_SUCCESS;
        } else {
          /* A connection was open, but the handshake of sites went wrong. */
          return DME_UNHANDSHAKABLE_SITE;
        }
      } else {
        /* Not possible to connect to that host:port, abandon! */
        return DME_UNCONNECTABLE_SITE;
      }
    }
  }
}

DM_ErrorCode
DMeroon_connect (DM_Context* context,
                 DM_Site*    *returned_site,
                 char*       hostname,
                 DM_nat2     port_number,
                 char*       password ) 
{
  DM_ProtectWithContext(context,
       context->errcode = DM_do_connect(context, 
                                        returned_site, 
                                        hostname,
                                        port_number,
                                        password ),
     /* Nothing even if something went wrong. */
);
  return context->errcode;
}


/* }}}
 * {{{ Class management

 * RESOURCE DMeroon_create_subclass
 * Create a subclass with one additional field. 
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_create_subclass (DM_Context*      context,
                       DM_Class*        *returned_class,
                       DM_Class*        superclass,
                       char*            class_name,
                       DM_class_options class_options,
                       char*            field_name,
                       DM_field_options field_options,
                       DM_DMeroonType*  reptyp,
                       DM_DMeroonType*  itemtyp,
                       DM_Class*        referred_class ) 
{
  /* This may wait */
  DM_ensure_class_usable(context, superclass);
  {
    DM_nat3  inherited_supers = superclass->super_length;
    DM_nat3  inherited_fields = 
      DM_class_indexed_field_length(context, superclass);
    DM_Class *newclass        = DM_Allocate3(context, Class, 
                                              inherited_supers+1, 
                                              inherited_fields+1,
                                              1+strlen(class_name) );
    char *cname = DM_predefined_name(context, (DM_Object*)newclass);
    DM_Field  *field;
  
    strcpy(cname, class_name);
    newclass->options |= class_options;
    /* Copies inherited superclasses, inherited fields ... */
    DM_structure_inheritance(context, newclass, superclass);
    if ( inherited_fields == 0 ) {
      field = DM_Allocate2(context, Field, 1, 1+strlen(field_name));
    } else {
      DM_Field  *last_field  = 
        DM_class_field(context, superclass, inherited_fields-1);
      DM_nat3   on              = last_field->offset_length;
      DM_offset last_offset     = last_field->offset[on-1];
      DM_nat1   lastreptypindex = DM_OffsetToRepetitionTypeIndex(last_offset);

      field = DM_Allocate2(context, 
                           Field, 
                           ( lastreptypindex == DM_nought_index )
                            ? on : on+1,
                           1+strlen(field_name) );
      /* Don't need to copy inherited offsets, 
       * DM_compute_layout will do that. */
    }
    { char *nm = DM_predefined_name(context, (DM_Object*)field);
      strcpy(nm, field_name);
    }
    field->options    |= field_options;
    field->introducing = newclass;
    if ( itemtyp == DM_ReferenceToDMeroonType(reference) ) {
      if ( DM_IsA(context, referred_class, Class) ) {
        field->referring   = referred_class;
      } else {
        return DME_NOT_A_CLASS;
      }
    }
    field->offset[field->offset_length-1] = 
      DM_PackOffset(0, reptyp->index, itemtyp->index);
    DM_set_class_field(context, newclass, inherited_fields, field);
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      DM_SetSharable(context, field);
    }
    DM_compute_characteristics(context, newclass);
    DM_compute_layout(context, newclass);
    /* Mark that the class is usable now but only if initialized. */
    if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
      DM_SetOption(newclass, DM_CLASS_IS_USABLE);
      DM_SetSharable(context, newclass);
    }
    *returned_class = newclass;
    return DME_SUCCESS;
  }
}

DM_ErrorCode
DMeroon_create_subclass (DM_Context*      context,
                         DM_Class*        *returned_class,
                         DM_Class*        superclass,
                         char*            class_name,
                         DM_class_options class_options,
                         char*            field_name,
                         DM_field_options field_options,
                         DM_DMeroonType*  reptyp,
                         DM_DMeroonType*  itemtyp,
                         DM_Class*        referred_class ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(superclass))
          || !(DM_IsObject(reptyp))
          || !(DM_IsObject(itemtyp)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, superclass, Class)) ) {
      return DME_NOT_CLASS;
    }
    if (   !(DM_IsObject(reptyp)) 
        || !(DM_IsA(context, reptyp, Type)) ) {
      return DME_NOT_TYPE;
      /* Repetition types may only be nought or nat1-4. */
      if ( reptyp->index > DM_nat4_index ) {
        return DME_NOT_REPETITION_TYPE;
      }
    }
    if (   !(DM_IsObject(itemtyp)) 
        || !(DM_IsA(context, itemtyp, Type)) ) {
      return DME_NOT_TYPE;
    }
  }
  if ( DM_HasOption(superclass, DM_CLASS_IS_NOT_SUBCLASSABLE) ) {
    context->errcode = DME_NOT_SUBCLASSABLE;
  } else {
    DM_ProtectWithContext(context,
       context->errcode = DM_do_create_subclass(context,
                                                returned_class,
                                                superclass,
                                                class_name,
                                                class_options,
                                                field_name,
                                                field_options,
                                                reptyp,
                                                itemtyp,
                                                referred_class ),
     /* Nothing even if something went wrong. */
);
  }
  return context->errcode;
}

/* RESOURCE DM_smash_class
 * Clone a class with less superclasses but with the same fields. This
 * function is needed by the previous one that only knows how to subclass
 * with a single field. The resulting class must still be a subclass of
 * Object.
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_smash_class (DM_Context*      context,
                   DM_Class*        *returned_class,
                   DM_Class*        class,
                   char*            class_name,
                   DM_class_options class_options,
                   DM_nat3          level ) 
{
  DM_ensure_class_usable(context, class);
  {
    DM_nat3  nfields = DM_class_indexed_field_length(context, class);
    DM_nat3  i;
    DM_Field *field, *oldfield;
    DM_Class *cl;

    /* Test if it is possible to smash so many levels. Of course,
     * Object must stay the same. */
    if ( level <= class->super_length-1 ) {
      DM_Class *newclass = DM_Allocate3(context, Class, 
                                        class->super_length - level + 1,
                                        nfields,
                                        1+strlen(class_name) );
      char *cname = DM_predefined_name(context, (DM_Object*)newclass);

      strcpy(cname, class_name);
      newclass->options |= class_options;
      /* Fill all superclasses. */
      for ( i=0 ; i<(class->super_length - level) ; i++ ) {
        newclass->super[i] = class->super[i];
      }
      newclass->super[class->super_length - level] = newclass;
      /* Fill fields from cloned class. */
      for ( i=0 ; i<nfields ; i++ ) {
        oldfield = DM_class_field(context, class, i);
        field = (DM_Field*)DM_clone(context, (DM_Object*)oldfield);
        DM_set_class_field(context, newclass, i, field);
        /* Correct 'introducing' field. */
        cl = field->introducing;
        if ( cl->super_length >= newclass->super_length ) {
          field->introducing = newclass;
        }
        if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
          DM_SetSharable(context, field);
        }
      }
      DM_compute_characteristics(context, newclass);
      DM_compute_layout(context, newclass);
      /* Mark that the class is usable now but only if initialized. */
      if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
        DM_SetOption(newclass, DM_CLASS_IS_USABLE);
        DM_SetSharable(context, newclass);
      }
      *returned_class = newclass;   
      return DME_SUCCESS;
    } else {
      /* Cannot smash so many levels. */
      return DME_BAD_LEVEL;
    }
  }
}

DM_ErrorCode
DMeroon_smash_class (DM_Context*      context,
                     DM_Class*        *returned_class,
                     DM_Class*        class,
                     char*            class_name,
                     DM_class_options class_options,
                     DM_nat3          level ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(class)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, class, Class)) ) {
      return DME_NOT_CLASS;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_smash_class (context, 
                                             returned_class,
                                             class,
                                             class_name,
                                             class_options,
                                             level ),
     /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* }}}
 * {{{ Migration

 * RESOURCE DMeroon_set_site_of
 * Make a site be the owner of an object. 
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_set_site_of (DM_Context* context,                   /* BOGUS!!! */
                   DM_Site*    *returned_previous_site,
                   DM_Object*  o,
                   DM_Site*    site ) {

#if 0

  if ( site == DM_TheCurrentSite() ) {
    /* We are on the site that must become the new owner. */
    DM_Object *proxy = DM_ObjectToProxy(context, o);
    
    if ( DM_IsA(context, proxy, Exit) ) {
      /* The object is remote. */
      DM_DeclareNarrowedObject(exit, Exit, proxy);
      DM_OwnerChangeRequest *request = 
        DM_Allocate0(context, OwnerChangeRequest);
      DM_EntryWithDelayedRequests *entry = 
        DM_Allocate0(context, EntryWithDelayedRequests);

      entry->counter    = 0;
      entry->options    = DM_EXTERNAL_DEFAULT_OPTIONS;
      entry->object     = DM_NULL;
      entry->clock      = DM_NULL;
      entry->threads    = NULL;
      DM_register_entry(context, (DM_External*)entry);
      request->options    = DM_REQUEST_DEFAULT_OPTIONS;
      DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
      request->requesting = DM_TheCurrentSite();
      request->requested  = exit->owner;
      /* Create a shortcutable reference. */
      request->object     = (DM_Object*)exit;
      request->entry_num  = entry->number;
      DM_SetSharable(context, request);
      DM_externalize(context, (DM_Object*)request);
      DM_SendRequest(context, request, exit->owner);
      /* If no answer is returned, a timeout will be raised. */
      (void) DM_wait_until_answer(context, (DM_Request*)request);
      return DME_SUCCESS;
    } else {
      /* The object is local, so its current owner is also the new one. */
      return DME_SUCCESS;
    }
  } else {
    /* Send a request to the the owning site. */
    DM_RemoteOwnerChangeRequest *request = 
      DM_Allocate0(context, RemoteOwnerChangeRequest);

    request->options    = DM_REQUEST_DEFAULT_OPTIONS;
    DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
    request->requesting = DM_TheCurrentSite();
    request->requested  = site;
    request->object     = (DM_Object*)DM_externalize(context, o);
    DM_SetCopyable(context, request);
    DM_externalize(context, (DM_Object*)request);
    DM_SendRequest(context, request, site);
    DM_wait_until_answer(context, 
                         DM_TheCurrentSite()->abnormal_duration,
                         &answer,
                         (DM_Request*)request);
    return DME_SUCCESS;
  }
#endif /* 0 */

  return DME_NOT_YET_IMPLEMENTED;
}

DM_ErrorCode
DMeroon_set_site_of (DM_Context* context,                   /* BOGUS!!! */
                     DM_Site*    *returned_previous_site,
                     DM_Object*  object,
                     DM_Site*    site ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(site)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, site, Site)) ) {
      return DME_NOT_SITE;
    }
  }

  return DME_NOT_YET_IMPLEMENTED;

  DM_ProtectWithContext(context,
       context->errcode = DM_do_set_site_of(context, 
                                            returned_previous_site,
                                            object,
                                            site ),
     /* Nothing even if something went wrong. */
);
  return context->errcode;
}

static DM_ErrorCode
DM_do_set_clock_of (DM_Context* context,
                    DM_Clock*   *returned_previous_clock,
                    DM_Object*  object,
                    DM_Clock*   clock ) {
  DM_Class  *class = DM_ObjectToClass(context, object);
  DM_Object *proxy = DM_ObjectToProxy(context, object);

  if ( DM_HasOption(class, DM_CLASS_IS_MUTABLE) ) {
    if ( DM_IsA(context, proxy, Exit) ) {
      /* The object is remote. */
        
      return DME_NOT_YET_IMPLEMENTED;

    } else if ( DM_IsA(context, proxy, Entry) ) {
      /* The object is local and already externalized. */
      DM_DeclareNarrowedObject(entry, Entry, proxy);

      *returned_previous_clock = entry->clock;
      entry->clock = clock;
      return DME_SUCCESS;
    } else {
      /* The object is local without associated Entry item.
       * Semi-externalize it and record the clock in the fresh Entry
       * item. */
      DM_Entry *entry = DM_semi_externalize(context, object);

      entry->clock    = clock;
      *returned_previous_clock = DM_NULL;      /* No previous clock */
      return DME_SUCCESS;
      }
  } else {
    /* An immutable object cannot have a clock. */
    return DME_NOT_MUTABLE;
  }
}

DM_ErrorCode
DMeroon_set_clock_of (DM_Context* context,                   /* BOGUS !!! */
                      DM_Clock*   *returned_previous_clock,
                      DM_Object*  object,
                      DM_Clock*   clock ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(clock)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, clock, Clock)) ) {
      return DME_NOT_CLOCK;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_set_clock_of(context, 
                                             returned_previous_clock,
                                             object,
                                             clock ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* }}}
 * {{{ Field access

 * RESOURCE DM_check_access
 * Checking the legality of accesses. This function exists to share all
 * access check. It is parameterized by three boolean values characterizing
 * the access (indexed or not, modification or not, length query or not).
 * Wrapped_type if not NULL indicates which type is expected from that
 * access.
 *
 * This function is an internal function that should not be there.
 */

DM_ErrorCode
DM_check_access (DM_Context *context,
                 DM_Object* o,
                 DM_Field*  field,
                 DM_nat3    access_status ) {
  /* Check the instance wrt field. */
  if ( DM_is_a(context, o, field->introducing) ) {
    /* The instance has such a field. */
    DM_offset last_offset = field->offset[field->offset_length - 1];
    DM_nat1   reptypindex = DM_OffsetToRepetitionTypeIndex(last_offset);

    /* Cannot access a secrete field. */
    if ( DM_HasOption(field, DM_FIELD_IS_SECRET) ) {
      return DME_SECRET_FIELD;
    }
    
    if ( access_status & (DM_INDEXED_ACCESS|DM_LENGTH_ACCESS) ) {
      /* Means that we are checking a DM-indexed-{get,set} invocation 
       * or a DM_length operation. */
      if ( reptypindex == DM_nought_index )
        return DME_NOT_INDEXED_FIELD;
    } else {
      /* This is a regular access. */
      if ( reptypindex != DM_nought_index )
        return DME_NOT_REGULAR_FIELD;
    }
  
    if ( access_status & DM_WRITE_ACCESS ) {
      /* Means that we are checking a DM-{regular,indexed-}set invocation:
       * Or the field is mutable or the object is uninitialized. */
      if ( DM_HasOption(field, DM_CLASS_IS_MUTABLE) ) {
        return DME_SUCCESS;
      } else {
        DM_Object *proxy = DM_ObjectToProxy(context, o);

        if ( DM_IsA(context, proxy, Entry) ) {
          DM_DeclareNarrowedObject(entry, Entry, proxy);

          if ( DM_HasOption(entry, DM_ENTRY_IS_UNINITIALIZED) ) {
            return DME_SUCCESS;
          } else {
            return DME_NOT_MUTABLE_FIELD;
          }
        } else if ( proxy == DM_TheConstant(uninitialized) ) {
          return DME_SUCCESS;
        } else {
          return DME_NOT_MUTABLE_FIELD;
        }
      }
    }
    return DME_SUCCESS;
  } else {
    /* The instance has no such field. */
    return DME_WRONG_INSTANCE;
  }
}

/* RESOURCE DMeroon_field_value
 * Reads the content of a regular field of an Object. 
 * It is useful to return, at the same time, the DMeroon type of the
 * read value. This helps to decode that value.
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_field_value (DM_Context*     context,
                   void            *returned_value,
                   DM_DMeroonType* *returned_type,
                   DM_Object*      object,
                   DM_Field*       field ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_offset          last_offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *itemtyp  = 
    implemented_types->item[DM_OffsetToItemTypeIndex(last_offset)];
  DM_displacement    t;
  DM_ErrorCode       errcode;
  void               *p;
    
  if ( returned_type ) {
    *returned_type   = itemtyp;
  }
  errcode = DM_check_access(context, object, field, DM_DEFAULT_ACCESS);
  if ( errcode != DME_SUCCESS ) {
    return errcode;
  } else {
    DM_Object* proxy = DM_ObjectToProxy(context, object);

    if ( DM_IsA(context, proxy, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, proxy);

      /* The object is remote. */
      if ( DM_HasOption(exit, DM_ENTRY_IS_UNINITIALIZED) ) {
        /* When the object is uninitialized, always fetch a fresh copy. */
        DM_send_object_fetch_request(context, &object, exit);
      } else if ( DM_HasOption(field, DM_FIELD_IS_LOCAL) ) {
        /* NOTHING: a local field may be read without precaution. */
      } else if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
        /* Check if the replica is up to date. */
        DM_ensure_up_to_date(context, object);
      } else if ( DM_HasOption(field, DM_FIELD_IS_VOLATILE) ) {
        /* Fetch a fresh copy of the object. */
        DM_send_object_fetch_request(context, &object, exit);
      } else {
        /* NOTHING: an immutable field may be read without precaution. */
      }
    } else { 
      /* NOTHING: the object is local and may be read. However there
       * is a special case: Sites may be usable (and read) or not
       * usable (and must be filled before being read). */
      if ( DM_IsA(context, object, Site) ) {
        DM_DeclareNarrowedObject(site, Site, object);

        DM_ensure_site_usable(context, site);
      }
    }
    t = DM_compute_offset(context, object, field);
    p = DM_Displace(object, t);
    if ( itemtyp->index == DM_reference_index ) {
      DM_EnsurePresent(context, p);

      DM_DebugAtLevel(1,
        /* No External instance should ever be seen by the user. */
        if ( DM_IsA(context, *((DM_Object**)p), External) ) {
          DM_Abort(context, DME_INTERNAL_ERROR);
        }
);

    }
    /* Read the field now! */
    memcpy(returned_value, p, itemtyp->size);
    return DME_SUCCESS;
  }
}

DM_ErrorCode
DMeroon_field_value (DM_Context*     context,
                     void            *returned_value,
                     DM_DMeroonType* *returned_type,
                     DM_Object*      object,
                     DM_Field*       field ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(field)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_field_value(context, 
                                            returned_value,
                                            returned_type,
                                            object,
                                            field ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}


/* RESOURCE DMeroon_indexed_field_value
 * Reads the content of an indexed field of an Object. 
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_indexed_field_value (DM_Context*     context,
                           void            *returned_value,
                           DM_DMeroonType* *returned_type,
                           DM_Object*      object,
                           DM_Field*       field,
                           DM_nat3         index ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_offset          last_offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *itemtyp  = 
    implemented_types->item[DM_OffsetToItemTypeIndex(last_offset)];
  DM_DMeroonType     *reptyp  = 
    implemented_types->item[DM_OffsetToRepetitionTypeIndex(last_offset)];
  DM_displacement    t;
  DM_nat3            rep;
  DM_ErrorCode       errcode;
  void               *p;
    
  if ( returned_type ) {
    *returned_type   = itemtyp;
  }
  errcode = DM_check_access(context, object, field, DM_INDEXED_ACCESS);
  if ( errcode != DME_SUCCESS ) {
    return errcode;
  } else {
    DM_Object* proxy = DM_ObjectToProxy(context, object);

    if ( DM_IsA(context, proxy, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, proxy);

      /* The object is remote. */
      if ( DM_HasOption(exit, DM_ENTRY_IS_UNINITIALIZED) ) {
        /* When the object is uninitialized, always fetch a fresh copy. */
        DM_send_object_fetch_request(context, &object, exit);
      } else if ( DM_HasOption(field, DM_FIELD_IS_LOCAL) ) {
        /* NOTHING: a local field may be read without precaution. */
      } else if ( DM_HasOption(field, DM_FIELD_IS_MUTABLE) ) {
        /* Check if the replica is up to date. */
        DM_ensure_up_to_date(context, object);
      } else if ( DM_HasOption(field, DM_FIELD_IS_VOLATILE) ) {
        /* Fetch a fresh copy of the object. */
        DM_send_object_fetch_request(context, &object, exit);
      } else {
        /* NOTHING: an immutable field may be read without precaution. */
      }
    } else { 
      /* NOTHING: the object is local and may be read. However there
       * is a special case: Sites may be usable (and read) or not
       * usable (and must be filled before being read). */
      if ( DM_IsA(context, object, Site) ) {
        DM_DeclareNarrowedObject(site, Site, object);

        DM_ensure_site_usable(context, site);
      }
    }
    /* Check the index to be in the correct range. */
    t   = DM_compute_offset(context, object, field);
    p = DM_Displace(object, t);
    rep = DM_read_repetition_factor(context, p, reptyp->index);
    if ( index >= rep ) {
      return DME_BAD_INDEX;
    } else {
      /* Read the field now! */
      t = DM_OffsetExtend(t, reptyp);
      t = DM_OffsetPad(t, itemtyp);
      t += index * itemtyp->repsize;
      p = DM_Displace(object, t);
      if ( itemtyp->index == DM_reference_index ) {
        DM_EnsurePresent(context, p);

        DM_DebugAtLevel(1,
          /* Allow Requests to be output but not Entry nor Exit items. */
          if (    DM_IsA(context, *((DM_Object**)p), External) 
               && ! DM_IsA(context, *((DM_Object**)p), Request) ) {
            DM_Abort(context, DME_INTERNAL_ERROR);
          }
);

      }
      memcpy(returned_value, p, itemtyp->size);
      return DME_SUCCESS;
    }
  }
}


DM_ErrorCode
DMeroon_indexed_field_value (DM_Context*     context,
                             void            *returned_value,
                             DM_DMeroonType* *returned_type,
                             DM_Object*      object,
                             DM_Field*       field,
                             DM_nat3         index ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(field)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_indexed_field_value(context, 
                                                    returned_value,
                                                    returned_type,
                                                    object,
                                                    field,
                                                    index ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* RESOURCE DMeroon_set_field_value
 * Changes the content of a field of an Object. It is not useful
 * to return the type of the values since one must know it before in order
 * to give a value. This primitive returns the value that was overwritten.
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_set_field_value (DM_Context*     context,
                       void            *returned_value,
                       DM_Object*      object,
                       DM_Field*       field,
                       void*           value ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_offset          last_offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *itemtyp  = 
    implemented_types->item[DM_OffsetToItemTypeIndex(last_offset)];
  DM_displacement    t;
  DM_ErrorCode       errcode;
  void               *p;
    
  errcode = DM_check_access(context, object, field, DM_WRITE_ACCESS);
  if ( errcode != DME_SUCCESS ) {
    return errcode;
  } else {
    DM_Object *proxy = DM_ObjectToProxy(context, object);

    if ( DM_IsA(context, proxy, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, proxy);
      DM_ObjectMutationRequest *request = 
        DM_Allocate0(context, ObjectMutationRequest);
      DM_Class *class = DM_ObjectToClass(context, object);
      DM_nat3  fn = DM_class_indexed_field_length(context, class);
      DM_nat3  i;

      request->options    = DM_REQUEST_DEFAULT_OPTIONS;
      DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
      request->requesting = DM_TheCurrentSite();
      request->requested  = exit->owner;
      /* Create a shortcutable reference. */
      request->object     = (DM_Object*)exit;
      /* Find the rank of the field within the class. */
      for ( i=0 ; i<fn ; i++ ) {
        if ( field == DM_class_field(context, class, i) ) {
          request->field_index = i;
          break;
        }
      }
      request->value = DM_reify_value(context, itemtyp, value);
      DM_SetSharable(context, request);
      DM_externalize(context, (DM_Object*)request);
      /* Encode the Request in the appropriate OutBuffer */
      DM_SendRequest(context, request, exit->owner);
      /* Wait for the Answer to come back.  */
      { DM_ObjectMutationAnswer *answer;

        DM_wait_until_answer(context, 
                             DM_TheCurrentSite()->abnormal_duration,
                             (DM_Answer**)&answer,
                             (DM_Request*)request);

        if ( returned_value ) {
          memcpy(returned_value, answer->value, itemtyp->size);
        }
        return DME_SUCCESS;
      }
    } else { 
      /* The object is local and may be mutated. However there is a
       * special case: Sites may be usable or not. But since Sites are
       * immutable, this is not a problem. */
      t = DM_compute_offset(context, object, field);
      p = DM_Displace(object, t);
      if ( itemtyp->index == DM_reference_index ) {
        DM_EnsurePresent(context, p);

        DM_DebugAtLevel(1,
          if ( DM_IsA(context, *((DM_Object**)p), External) ) {
            DM_Abort(context, DME_INTERNAL_ERROR);
          }
);

        /* Verify the class of the value to be written. */
        if (    (*((DM_Object**)value) == DM_NULL) 
             || DM_is_a(context, *((DM_Object**)value), field->referring) ) {
          /* Correct class */
        } else {
          return DME_BAD_CLASS;
        }
      }
      if ( returned_value ) {
        memcpy(returned_value, p, itemtyp->size);
      }
      /* then mutate it! */
      memcpy(p, value, itemtyp->size);
      DM_increment_clock_of(context, object);
      return DME_SUCCESS;
    }
  }
}

DM_ErrorCode
DMeroon_set_field_value (DM_Context*     context,
                         void            *returned_value,
                         DM_Object*      object,
                         DM_Field*       field,
                         void*           value ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(field)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_set_field_value(context,
                                                returned_value,
                                                object,
                                                field,
                                                value ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* RESOURCE DMeroon_set_indexed_field_value
 * Writes the content of an indexed field of an Object. 
 *
 * This primitive may block.
 */

static DM_ErrorCode
DM_do_set_indexed_field_value (DM_Context*     context,
                               void            *returned_value,
                               DM_Object*      object,
                               DM_Field*       field,
                               DM_nat3         index,
                               void*           value ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_offset          last_offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *itemtyp  = 
    implemented_types->item[DM_OffsetToItemTypeIndex(last_offset)];
  DM_DMeroonType     *reptyp  = 
    implemented_types->item[DM_OffsetToRepetitionTypeIndex(last_offset)];
  DM_displacement    t;
  DM_nat3            rep;
  DM_ErrorCode       errcode;
  void               *p;
    
  errcode = DM_check_access(context, object, field, 
                            DM_WRITE_ACCESS|DM_INDEXED_ACCESS);
  if ( errcode != DME_SUCCESS ) {
    return errcode;
  } else {
    DM_Object *proxy = DM_ObjectToProxy(context, object);

    if ( DM_IsA(context, proxy, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, proxy);
      DM_Class *class = DM_ObjectToClass(context, object);
      DM_ObjectIndexedMutationRequest *request = 
        DM_Allocate0(context, ObjectIndexedMutationRequest);
      DM_nat3 fn = DM_class_indexed_field_length(context, class);
      DM_nat3 i;

      request->options    = DM_REQUEST_DEFAULT_OPTIONS;
      DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
      request->requesting = DM_TheCurrentSite();
      request->requested  = exit->owner;
      /* Create a shortcutable reference. */
      request->object     = (DM_Object*)exit;
      /* Find the rank of the field within the class. */
      for ( i=0 ; i<fn ; i++ ) {
        if ( field == DM_class_field(context, class, i) ) {
          request->field_index = i;
          break;
        }
      }
      request->value = DM_reify_value(context, itemtyp, value);
      request->index = index;
      DM_SetSharable(context, request);
      DM_externalize(context, (DM_Object*)request);
      /* Encode the Request in the appropriate OutBuffer */
      DM_SendRequest(context, request, exit->owner);
      /* Wait for the Answer to come back. */
      { DM_ObjectMutationAnswer *answer;

        DM_wait_until_answer(context, 
                             DM_TheCurrentSite()->abnormal_duration,
                             (DM_Answer**)&answer,
                             (DM_Request*)request);
        if ( returned_value ) {
          memcpy(returned_value, answer->value, itemtyp->size);
        }
        return DME_SUCCESS;
      }
    }
    /* Check the index to be in the correct range. */
    t   = DM_compute_offset(context, object, field);
    rep = DM_read_repetition_factor(
             context, DM_Displace(object, t), 
             DM_OffsetToRepetitionTypeIndex(last_offset) );
    if ( index >= rep ) {
      return DME_BAD_INDEX;
    } else {
      /* Read the field now! */
      t = DM_OffsetExtend(t, reptyp);
      t = DM_OffsetPad(t, itemtyp);
      t += index * itemtyp->repsize;
      p = DM_Displace(object, t);
      if ( itemtyp->index == DM_reference_index ) {
        DM_EnsurePresent(context, p);

        DM_DebugAtLevel(1,
          if ( DM_IsA(context, *((DM_Object**)p), External) ) {
            DM_Abort(context, DME_INTERNAL_ERROR);
          }
);

        /* Verify the class of the value to be written. */
        if (    (*((DM_Object**)value) == DM_NULL) 
             || DM_is_a(context, *((DM_Object**)value), field->referring) ) {
          /* Correct class */
        } else {
          return DME_BAD_CLASS;
        }
      }
      if ( returned_value ) {
        memcpy(returned_value, p, itemtyp->size);
      }
      /* then mutate it! */
      memcpy(p, value, itemtyp->size);
      DM_increment_clock_of(context, object);
      return DME_SUCCESS;
    }
  }
}

DM_ErrorCode
DMeroon_set_indexed_field_value (DM_Context*     context,
                                 void            *returned_value,
                                 DM_Object*      object,
                                 DM_Field*       field,
                                 DM_nat3         index,
                                 void*           value ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(field)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_set_indexed_field_value(context, 
                                                        returned_value,
                                                        object,
                                                        field,
                                                        index,
                                                        value ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* RESOURCE DMeroon_field_types
 * This primitive takes a Field and returns its type as well as the type
 * of its repetition factor if any. It is possible for this function
 * to give NULL for the returned locations.
 */

DM_ErrorCode
DMeroon_field_types (DM_Context*     context,
                     DM_DMeroonType* *returned_item_type,
                     DM_DMeroonType* *returned_repetition_type,
                     DM_Field*       field ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(field)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  { 
    DM_offset      last_offset = field->offset[field->offset_length-1];
    DM_DMeroonType *itemtyp    = 
      implemented_types->item[DM_OffsetToItemTypeIndex(last_offset)];
    DM_DMeroonType *reptyp     = 
      implemented_types->item[DM_OffsetToRepetitionTypeIndex(last_offset)];

    if ( returned_item_type) 
      *returned_item_type       = itemtyp;
    if ( returned_repetition_type )
      *returned_repetition_type = reptyp;
  }
  return DME_SUCCESS;
}

/* RESOURCE DMeroon_indexed_field_length
 * Reads the length of an indexed field of an Object. 
 *
 * This is a non-blocking primitive.
 */

static DM_ErrorCode
DM_do_indexed_field_length (DM_Context*     context,
                    void            *returned_length,
                    DM_DMeroonType* *returned_type,
                    DM_Object*      object,
                    DM_Field*       field ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_offset          last_offset = field->offset[field->offset_length-1];
  DM_DMeroonType     *reptyp  = 
    implemented_types->item[DM_OffsetToRepetitionTypeIndex(last_offset)];
  DM_displacement    t;
  DM_ErrorCode       errcode;
    
  if ( returned_type ) {
    *returned_type   = reptyp;
  }
  errcode = DM_check_access(context, object, field, DM_LENGTH_ACCESS);
  if ( errcode != DME_SUCCESS ) {
    return errcode;
  } else {
    /* A length of an indexed field is a structural property which 
     * is always present if the object is present. */
    t = DM_compute_offset(context, object, field);
    memcpy(returned_length, DM_Displace(object, t), reptyp->size);
    return DME_SUCCESS;
  }
}  

DM_ErrorCode
DMeroon_indexed_field_length (DM_Context*     context,
                              void            *returned_length,
                              DM_DMeroonType* *returned_type,
                              DM_Object*      object,
                              DM_Field*       field ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object))
          || !(DM_IsObject(field)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, field, Field)) ) {
      return DME_NOT_FIELD;
    }
  }

  DM_ProtectWithContext(context,
       context->errcode = DM_do_indexed_field_length(context, 
                                                     returned_length,
                                                     returned_type,
                                                     object,
                                                     field ),
       /* Nothing even if something went wrong. */
);
  return context->errcode;
}

/* }}}
 * {{{ Allocation
 
 * RESOURCE DMeroon_allocate
 * Allocate an instance of a class with the given sizes. All sizes are
 * expressed in terms of nat3 natural numbers (even when the size is
 * a nat4 number). 

 * The interface does not use the standarg (or vararg) mechanism but
 * the more usable argc, argv interface. The reason is that it is not
 * possible to build a va_list portably.
 */

DM_ErrorCode
DMeroon_do_allocate (DM_Context* context,
                     DM_Object*  *returned_object,
                     DM_Class*   class,
                     DM_nat3     number_of_sizes,
                     DM_nat3     sizes[] ) 
{
  DM_Object *o;
  DM_nat3   fn;

  if ( DM_is_class_usable(context, class) ) {
    fn = DM_class_indexed_field_length(context, class);

    if ( fn == 0 ) {
      /* DM_allocate_bare_instance should abort in case of problem so we
       * can safely continue. */
      o = DM_allocate_bare_instance(context, class, 0);
      DM_SetClassOf(context, o, class);
      DM_SetProxyOf(context, o, DM_TheConstant(uninitialized));

    } else { 
      /* This is not a class without fields. We must then compute its size. */
      DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
      DM_Field          *field = DM_class_field(context, class, fn-1);
      DM_nat3           on     = field->offset_length;
      DM_displacement   t      = 0;
      DM_offset         offset;
      DM_nat1           itemtypindex, reptypindex;
      DM_DMeroonType    *itemtyp, *reptyp;
      DM_nat3           i, rep;
      void              *p;
    
      offset       = field->offset[on-1];
      reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
      /* If the last field is regular there must be exactly `on-1'
       * sizes given to DM_allocate, if the last field is indexed then
       * there should be `on' sizes present. */
      if ( reptypindex == DM_nought_index ) {
        if ( number_of_sizes != on-1 ) 
          return DME_INCORRECT_NUMBER_OF_SIZES;
      } else {
        if ( number_of_sizes != on ) 
          return DME_INCORRECT_NUMBER_OF_SIZES;
      }
    
      /* Compute the size of the instance to allocate */
      for ( i=0 ; i<on ; i++ ) {
        offset       = field->offset[i];
        reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
        reptyp       = implemented_types->item[reptypindex];
        itemtypindex = DM_OffsetToItemTypeIndex(offset);
        itemtyp      = implemented_types->item[itemtypindex];

        t += DM_OffsetToTranslation(offset);
        if ( reptypindex == DM_nought_index ) {
          /* i == on-1 since only the last offset may be non indexed. */
          /* Useless: t = DM_OffsetPad(t, itemtyp); */
          t = DM_OffsetExtend(t, itemtyp);
        } else {
          /* Useless: t = DM_OffsetPad(t, reptyp); */
          t = DM_OffsetExtend(t, reptyp);
          t = DM_OffsetPad(t, itemtyp);
          rep = sizes[i];
          /* Check the size of the repetition factor. */
          switch ( reptypindex ) {
          case DM_nat1_index : {
            if ( rep >= DM_NAT1_MAXIMUM ) {
              return DME_TOO_BIG_NAT1_SIZE;
            }
          }
          case DM_nat2_index : {
            if ( rep >= DM_NAT2_MAXIMUM ) {
              return DME_TOO_BIG_NAT2_SIZE;
            }
          }
          default: {
            /* NOTHING to check since nat3 fits in nat3 and nat4 is
             * reduced to the range of nat3. */
          }
          }
          t += rep * itemtyp->repsize;
          /* Realign after any non final indexed field. */
          if ( i<on-1 ) t = DM_OffsetRealign(t);
        }
      }
    
      /* DM_allocate_bare_instance should abort in case of problem so we
       * can safely continue. */
      o = DM_allocate_bare_instance(context, class, t);
      DM_SetClassOf(context, o, class);
      DM_SetProxyOf(context, o, DM_TheConstant(uninitialized));

      /* Skeletize the freshly allocated instance. */
      t = 0;
      for ( i=0 ; i<on ; i++ ) {
        offset       = field->offset[i];
        reptypindex  = DM_OffsetToRepetitionTypeIndex(offset);
        reptyp       = implemented_types->item[reptypindex];
        itemtypindex = DM_OffsetToItemTypeIndex(offset);
        itemtyp      = implemented_types->item[itemtypindex];
      
        t += DM_OffsetToTranslation(offset);
        if ( reptypindex != DM_nought_index ) {
          /* Useless: t = DM_OffsetPad(t, reptyp); */
          p = DM_Displace(o, t);
          rep = sizes[i];
          switch (reptypindex) {
          case DM_nat1_index : {
            DM_Content(DM_nat1, p) = rep;
            break;
          }
          case DM_nat2_index : {
            DM_Content(DM_nat2, p) = rep;
            break;
          }
          case DM_nat3_index : {
            DM_Content(DM_nat3, p) = rep;
            break;
          }
          case DM_nat4_index : {
            DM_Content(DM_nat4, p) = rep;
            break;
          }
          }
          t = DM_OffsetExtend(t, reptyp);
          t = DM_OffsetPad(t, itemtyp);
          t += rep * itemtyp->repsize;
          /* Realign after any non final indexed field. */
          if ( i<on-1 ) t = DM_OffsetRealign(t);
        }
      }
    }
    *returned_object = o;
    return DME_SUCCESS;
  } else {
    /* The class is unusable so we cannot allocate something with it. */
    return DME_UNUSABLE_CLASS;
  }
}

DM_ErrorCode
DMeroon_allocate (DM_Context* context,
                  DM_Object*  *returned_object,
                  DM_Class*   class,
                  DM_nat3     number_of_sizes,
                  DM_nat3     sizes[] ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(class)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_CLASS) ) {
    if ( !(DM_IsA(context, class, Class)) ) {
      return DME_NOT_CLASS;
    }
  }
  if ( DM_HasOption(class, DM_CLASS_IS_NOT_INSTANTIABLE) ) {
    return DME_NOT_UNINSTANTIABLE;
  }
  return DMeroon_do_allocate(context, 
                             returned_object,
                             class,
                             number_of_sizes,
                             sizes );
}

/* RESOURCE DMeroon_set_sharable
 * This function tells that an object is initialized ie may be shared over
 * the network.
 *
 * This primitive does not block.
 */

DM_ErrorCode
DMeroon_set_sharable (DM_Context* context,
                      DM_Object*  object ) {

  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  { DM_Object *proxy = DM_ObjectToProxy(context, object);
    
    if ( proxy == DM_TheConstant(uninitialized) ) {
      /* The object is local and uninitialized. */
      DM_SetSharable(context, object);
      return DME_SUCCESS;
    } else if ( proxy == DM_no_proxy_value ) {
      /* The object was already meant to be sharable. */
      return DME_SUCCESS;
    } else if ( proxy == DM_TheConstant(true) ) {
      /* Problem: the object was already specified as copyable. */
      return DME_ALREADY_COPYABLE;
    } else if ( DM_IsA(context, proxy, Entry) ) {
      if ( DM_IsA(context, proxy, Exit) ) {
        /* The object is remote. Don't touch it! */
        return DME_REMOTE_OBJECT;
      } else {
        /* The object is local and already externalized. Check that the
         * mode is compatible.  */
        DM_DeclareNarrowedObject(entry, Entry, proxy);
        
        if ( DM_HasOption(entry, DM_ENTRY_IS_COPYABLE) ) {
          /* Problem: the object was already specified as copyable. */
          return DME_ALREADY_COPYABLE;
        } else {
          DM_ClearOption(entry, DM_ENTRY_IS_UNINITIALIZED);
          return DME_SUCCESS;
        }
      }
    } else {
      /* This should never appear. NOT TRUE since one may initialize a
       * remote uninitialized object.                        FUTURE */
      DM_Abort(context, DME_BAD_PROXY);
      return DME_BAD_PROXY;
    }
  }
}

/* RESOURCE DMeroon_set_copyable
 * This function tells that an object is initialized ie may be copied over
 * the network.
 *
 * This primitive does not block.
 */

DM_ErrorCode
DMeroon_set_copyable (DM_Context* context,
                      DM_Object*  object ) {

  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  { DM_Object *proxy = DM_ObjectToProxy(context, object);
    
    if ( proxy == DM_TheConstant(uninitialized) ) {
      /* The object is local and uninitialized. */
      DM_SetCopyable(context, object);
      return DME_SUCCESS;
    } else if ( proxy == DM_no_proxy_value ) {
      /* Problem: the object was already specified as sharable. */
      return DME_ALREADY_SHARABLE;
    } else if ( proxy == DM_TheConstant(true) ) {
      /* The object was already meant to be copyable. */
      return DME_SUCCESS;
    } else if ( DM_IsA(context, proxy, Entry) ) {
      if ( DM_IsA(context, proxy, Exit) ) {
        /* The object is remote. Don't touch it! */
        return DME_REMOTE_OBJECT;
      } else {
        /* The object is local and already externalized. Check that the
         * mode is compatible.  */
        DM_DeclareNarrowedObject(entry, Entry, proxy);
        
        if ( DM_HasOption(entry, DM_ENTRY_IS_COPYABLE) ) {
          DM_ClearOption(entry, DM_ENTRY_IS_UNINITIALIZED);
          return DME_SUCCESS;
        } else {
          /* Problem: the object was already specified as sharable. */
          return DME_ALREADY_SHARABLE;
        }
      }
    } else {
      /* This should never appear. */
      DM_Abort(context, DME_BAD_PROXY);
      return DME_BAD_PROXY;
    }
  }
}

/* RESOURCE DMeroon_set_initialized
 * This function tells that an object is initialized ie may be copied
 * or shared over the network depending on its class.
 *
 * This primitive does not block.
 */

DM_ErrorCode
DMeroon_set_initialized (DM_Context* context,
                         DM_Object*  object ) 
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  { 
    DM_Class *class = DM_ObjectToClass(context, object);
    
    if ( DM_HasOption(class, DM_CLASS_IS_COPYABLE) ) {
      return DMeroon_set_copyable(context, object);
    } else {
      return DMeroon_set_sharable(context, object);
    }
  }
}

/* RESOURCE DMeroon_is_initialized
 * This function tells whether an object is initialized.
 *
 * This primitive does not block.
 */

DM_ErrorCode
DM_do_is_initialized (DM_Context*  context,
                      DM_nat3      *returned_boolean,
                      DM_Object*   object )
{
  DM_Object *proxy = DM_ObjectToProxy(context, object);

  if ( DM_IsA(context, proxy, Entry) ) {
    /* The object is remote or at least semi-externalized. */
    DM_DeclareNarrowedObject(entry, Entry, proxy);

    if ( DM_HasOption(entry, DM_ENTRY_IS_UNINITIALIZED) ) {
      *returned_boolean = DM_FALSE;
    } else if ( DM_HasOption(entry, DM_ENTRY_IS_COPYABLE) ) {
      *returned_boolean = DM_OBJECT_IS_COPYABLE;
    } else {
      *returned_boolean = DM_OBJECT_IS_SHARABLE;
    }
  } else {
    /* The object is local. */
    if ( proxy == DM_TheConstant(uninitialized) ) {
      /* The object is not yet initialized. */
      *returned_boolean = DM_FALSE;
    } else if ( proxy == DM_TheConstant(true) ) {
      *returned_boolean = DM_OBJECT_IS_COPYABLE;
    } else {
      *returned_boolean = DM_OBJECT_IS_SHARABLE;
    }
  }
  return DME_SUCCESS;
}

DM_ErrorCode
DMeroon_is_initialized (DM_Context*  context,
                        DM_nat3      *returned_boolean,
                        DM_Object*   object )
{
  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && !(DM_IsObject(object)) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }
  return DM_do_is_initialized(context, returned_boolean, object);
}

/* }}}
 * {{{ URLscript

 * RESOURCE DMeroon_interpret
 * The idea is to let users run their own urlscript programs. The
 * interest is to let them specify sequence of operations as an
 * urlscript program and to let the DMeroon sites evaluate this program
 * anywhere. A combined access such as o.x.y may then be performed on the
 * site of the o.x, this intermediate object do not need to be on the 
 * requesting site. 
 *           This interface is not well thought                  FUTURE
 *           The behavior of remote evaluation neither is        FUTURE
 */

DM_ErrorCode
DMeroon_interpret (DM_Context*  context,
                   DM_UrlState* *urlstate,
                   char*        urlprogram,
                   DM_nat3      nobjects,
                   DM_Object*   object, ... )
{
  DM_UrlState     *state = DM_Allocate0(context, UrlState);
  DM_UrlEnv       *env   = DM_Allocate1(context, UrlEnv, 2);
  DM_BoundedStack *stack = DM_Allocate1(context, BoundedStack, 8);
  va_list         objects;

  state->previous = DM_NULL;
  state->script   = DM_cstring_to_string(context, urlprogram, 0);
  state->pc       = 0;
  state->stack    = stack;
  va_start(objects, object);
  while ( nobjects-- > 0 ) {
    DM_Object *o = va_arg(objects, DM_Object*);
    
    /* context maybe NULL. */
    if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
         && ( !(DM_IsObject(o)) ) ) {
      return DME_NOT_DMEROON_INSTANCE;
    }
    DM_push_object_in_bounded_stack(context, o, &(state->stack));
  }
  va_end(objects);
  /* Protect the standard predefined environment (so user definition
   * do not pollute it). */
  env->hasher    = DM_TheCurrentSite()->urlenv->hasher;
  env->extractor = DM_TheCurrentSite()->urlenv->extractor;
  env->next = (DM_Hashtable*)(DM_TheCurrentSite()->urlenv);
  env->size = 0;
  state->environment = env;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetCopyable(context, state);
    DM_SetCopyable(context, stack);
    DM_SetCopyable(context, env);
  }
  /* NULL means no printing during the evaluation. */
  DM_ProtectWithContext(context,
       *urlstate = DM_interpret_url_script(context, state, NULL),
       /* Nothing even if something went wrong. */
);
  /* Handle the possible results. */
  (*urlstate)->errcode = context->errcode;
  return context->errcode;
}

/* }}}
 * {{{ Comparison
 */

DM_ErrorCode                                 /* BOGUS FOR NOW */
DMeroon_eq (DM_Context*  context,
            DM_c_boolean *returned_boolean,
            DM_Object*   object1,
            DM_Object*   object2 ) {

  /* context maybe NULL. */
  if ( (context) && DM_HasOption(context, DM_CONTEXT_CHECK_DMEROON)
      && (   !(DM_IsObject(object1))
          || !(DM_IsObject(object2)) ) ) {
    return DME_NOT_DMEROON_INSTANCE;
  }

  *returned_boolean = DM_FALSE;
  if ( object1 == object2 ) {
    *returned_boolean = DM_TRUE;
    return DME_SUCCESS;
  } else {
    
    return DME_NOT_YET_IMPLEMENTED;
  }
}

/* }}}
 * {{{ Errors
 * Escape to a safer place. Also record the error code and the content
 * of the errno variable (even if that content is not meaningfull across
 * different platforms). This function should never return! Code that
 * uses DM_Abort() do not expect this expression to return.
 *
 * Should probably not be there.
 */

void 
DM_abort (DM_Context    *context,
          DM_ErrorCode  code,
          char*         file,
          unsigned long line ) {
  if ( (context) && (context->continuation) ) {
    jmp_buf *exit = context->continuation;

    context->errcode   = code;
    context->file      = file;
    context->line      = line;
    context->errnumber = DM_ErrnoValue();
    DM_DebugAtLevel(0,
      fprintf(DM_FindErrorStream(context), 
              "\n!?!?!? Escaping from DMeroon error %lu at %s:%lu\n", 
              (unsigned long)code, file, line);
);
    longjmp(*exit, code); 
  } else {
    /* No context present: emergency exit! Report the error and stop
     * abruptly. Try to find a printable port to output the error
     * message. */
    if ( ! DM_debug_stream ) {
      DM_debug_stream = stderr;
    }
    fprintf(DM_debug_stream,
            "\n!?!?!? DMeroon emergency error %lu (errno=%lu) at %s:%lu\n",
            (unsigned long)code, (unsigned long)DM_ErrnoValue(), file, line);
    fflush(DM_debug_stream);
    exit(5);
  }
}

/* This function is called when control reaches a point that should
 * not be executed. This is only useful in DM_DEBUG mode. */

void
DM_barf_if_reached (DM_Context* context)
{
  fprintf(stderr, "\n\n\t\tDead code reached.\n");
  exit(DME_DEAD_CODE_REACHED);
}


/* }}}
 * {{{ Initialization
 * Since the new API is the culmination of DMeroon, its initializer will 
 * initialize all the other modules composing DMeroon. Note, this is an
 * internal entry point used by the binding language. The user of DMeroon
 * just invokes the DMeroon_initialize function.

 * NOTE: add an argc,argv argument for special initialization (like
 * port number to use, etc. )                            FUTURE
 */

void 
DM_newapi_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmnewapi);

  /* Initialize all modules (there is a specific order).
   * Currently, comm needs site (for the sites hashtable).
   * No initialization appears in the generated dmstruct.c file.
   */

  /* check for DM_0X. */
  DM_print_initialize(context);
  /* Adapt DMeroonTypes instances wrt C types, check their index. */
  DM_types_initialize(context);
  /* Set the rank of predefined classes. */
  DM_kernel_initialize(context);
  /* nothing */
  DM_hash_initialize(context);
  /* nothing */
  DM_extr_initialize(context);
  /* Initialize the Site instance. */
  DM_site_initialize(context);
  /* initialize communication resources, create Server. */
  DM_net_initialize(context);
  /* Initialize default protocol in site and original repository. */
  DM_prtcl_initialize(context);
  /* Initialize hashtables in site. */
  DM_comm_initialize(context);
  DM_rdobj_initialize(context);
  DM_wrobj_initialize(context);
  DM_req_initialize(context);
  DM_sched_initialize(context);
  DM_lib_initialize(context);
  DM_debug_initialize(context);
  DM_html_initialize(context);
  DM_reqst_initialize(context);
  DM_act_initialize(context);
  DM_gif_initialize(context);
}
  
/* }}}
 * end of dmnewapi.c */
