/* $Id: dmrdobj.c,v 1.88 1998/06/01 19:30:44 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

#include "dmeroon.h"

/* DOCUMENTATION dmrdobj.c
 * These functions allow to decode a stream of bytes to form a DMeroon
 * object. They match their equivalent encoding functions in the
 * dmwrobj.c file.
 *
 * The main functions are:
 * 
 *   DM_read_object
 *         It decodes a predefined, special or regular object.
 *         It is built as an interpreter executing a stream of bytecodes
 *         with intersparsed data.
 *
 *     DM_read_exit
 *         It decodes a EMITTER-REFERENCE or REMOTE-REFERENCE.
 *
 *     DM_read_object_size_and_allocate
 *         It decodes the sizes of the indexed fields so it can allocate
 *         the intended object.
 *
 *       DM_read_object_content
 *         It decodes the content of the fields of an already allocated object.
 */

DM_DefineVersionString(dmrdobj,
  "$Id: dmrdobj.c,v 1.88 1998/06/01 19:30:44 queinnec Exp $");

/* Reserved for recursion. */

static DM_Object*
DM_do_read_object (DM_Context  *context,
                   DM_InBuffer *inbuffer);

/* RESOURCE DM_read_object_content
 * Fill an already allocated object unmarshaling its content from a
 * bytestream. Since the class of the object is known, the stream of
 * bytes is decoded to fill every field in turn. The entire content is
 * unmarshaled except fields with the :local property which are never
 * transmitted nor initialized on the receiving site.
 */

static void
DM_read_object_content (DM_Context  *context,
                        DM_InBuffer *inbuffer,
                        DM_Object   *o) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class          *class  = DM_ObjectToClass(context, o);
  DM_nat3           fn      = DM_class_indexed_field_length(context, class);
  DM_translation    t       = 0;
  DM_Field          *field;
  DM_nat3           on;
  DM_offset         last_offset;
  DM_nat1           itemtypindex, reptypindex;
  DM_DMeroonType    *itemtyp, *reptyp;
  DM_nat3           i, j, rep;
  void              *p;

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), "<<ReadContent(%s) ", 
            DM_predefined_name(context, (DM_Object*)class) );
);

  for ( i=0 ; i<fn ; i++ ) {
    field        = DM_class_field(context, class, i); /* FUTURE: improve! */
    on           = field->offset_length;
    last_offset  = field->offset[on-1];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp      = implemented_types->item[itemtypindex];

    if ( DM_HasOption(field, DM_FIELD_IS_LOCAL) ) { 

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), "<%s.%s :local SKIPPED>",  
                DM_predefined_name(context, (DM_Object*)class),
                DM_predefined_name(context, (DM_Object*)field) );
);

      /* The field is local: just skip it. */
      if ( reptypindex == DM_nought_index ) {
        /* This field is not indexed. */
        t = DM_OffsetPad(t, itemtyp);
        t = DM_OffsetExtend(t, itemtyp);
      } else {
        t   = DM_OffsetPad(t, reptyp);
        p   = DM_Displace(o, t);
        rep = DM_read_repetition_factor(context, p, reptypindex);
        t   = DM_OffsetExtend(t, reptyp);
        t   = DM_OffsetPad(t, itemtyp);
        t += rep * itemtyp->repsize;  
        /* Realign after any non final indexed field. */
        t = DM_OffsetRealign(t);      
      }
    } else {
      /* The field is not local, so it has to be unmarshaled. */

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), "<%s.%s=",
                DM_predefined_name(context, (DM_Object*)class),
                DM_predefined_name(context, (DM_Object*)field) );
);

      if ( reptypindex == DM_nought_index ) {
        /* This field is not indexed. */
        t = DM_OffsetPad(t, itemtyp);
        p = DM_Displace(o, t);
        if (   (itemtypindex == DM_data_index)
            || (itemtypindex == DM_code_index) ) {
          DM_receive_raw_field(context, itemtyp, p, inbuffer);
        } else if ( itemtypindex == DM_reference_index ) {
          DM_Content(DM_reference, p) = DM_do_read_object(context, inbuffer);
        } else {
          (itemtyp->decoder)(context, inbuffer, p);
        }
        t = DM_OffsetExtend(t, itemtyp);
      } else {
        /* This field is indexed. */
        t   = DM_OffsetPad(t, reptyp);
        p   = DM_Displace(o, t);
        rep = DM_read_repetition_factor(context, p, reptypindex);
        t   = DM_OffsetExtend(t, reptyp);
        t   = DM_OffsetPad(t, itemtyp);
        for ( j=0 ; j<rep ; j++ ) {
          p = DM_Displace(o, t);
          if (   (itemtypindex == DM_data_index)
              || (itemtypindex == DM_code_index) ) {
            DM_receive_raw_field(context, itemtyp, p, inbuffer);
          } else if ( itemtypindex == DM_reference_index ) {
            DM_Content(DM_reference, p) = DM_do_read_object(context, inbuffer);
          } else {
            (itemtyp->decoder)(context, inbuffer, p);
          }
          t += itemtyp->repsize;
        }
        /* Realign after any non final indexed field. */
        t = DM_OffsetRealign(t);
      }

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), "%s.%s>>", 
                DM_predefined_name(context, (DM_Object*)class),
                DM_predefined_name(context, (DM_Object*)field));
);

    }
  }

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), " EndReadContent(%s)>>",
            DM_predefined_name(context, (DM_Object*)class) );
);

}

/* RESOURCE DM_read_object_size_and_allocate
 * According to the class, read the sizes needed to allocate the 
 * object, allocate it then return it. The allocated object will 
 * filled later.
 *
 * RESTRICTION: No more than DM_MAX_INDEXED_FIELDS indexed fields.
 */

static DM_Object*
DM_read_object_size_and_allocate (DM_Context  *context,
                                  DM_InBuffer *inbuffer,
                                  DM_Class    *class ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3           fn          = DM_class_indexed_field_length(context, class);
  DM_Field          *field;
  DM_nat3           on;
  DM_offset         last_offset;
  DM_nat1           reptypindex;
  DM_DMeroonType    *reptyp;
  DM_nat3           i;
  DM_nat3           size_index = 0;
  DM_nat3           sizes[DM_MAX_INDEXED_FIELDS];
  DM_Object         *o;

  for ( i=0 ; i<fn ; i++ ) {
    field       = DM_class_field(context, class, i);
    on          = field->offset_length;
    last_offset = field->offset[on-1];
    reptypindex = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp      = implemented_types->item[reptypindex];

    if ( reptypindex != DM_nought_index ) {
      switch ( reptypindex ) {
      case DM_nat1_index: {
        DM_nat1 sz;

        (reptyp->decoder)(context, inbuffer, &sz);
        sizes[size_index++] = DM_Content(DM_nat1, &sz);
        break;
      }
      case DM_nat2_index: {
        DM_nat2 sz;

        (reptyp->decoder)(context, inbuffer, &sz);
        sizes[size_index++] = DM_Content(DM_nat2, &sz);
        break;
      }
      case DM_nat3_index: {
        DM_nat3 sz;

        (reptyp->decoder)(context, inbuffer, &sz);
        sizes[size_index++] = DM_Content(DM_nat3, &sz);
        break;
      }
      case DM_nat4_index: {
        DM_nat4 sz;

        (reptyp->decoder)(context, inbuffer, &sz);
        if ( sz <= DM_NAT3_MAXIMUM ) {
          sizes[size_index++] = (DM_nat3)DM_Content(DM_nat4, &sz);
        } else {
          DM_Abort(context, DME_TOO_LARGE_INDEX);
        }
        break;
      }
      default: 
        DM_Abort(context, DME_BAD_INDEX);
      }
      if ( size_index >= DM_MAX_INDEXED_FIELDS )
        DM_Abort(context, DME_TOO_MUCH_INDEXES);
    }
  }

  o = DM_allocate_instance(context, class, size_index, sizes);
  return o;
}

/* RESOURCE DM_read_reference
 * Read an Exit item from an InBuffer.
 * It also tries to shortcut this Exit item into a present object if possible.
 * It appears as a function since this is quite the same code for
 * DM_REMOTE_REFERENCE_COMMMAND and for DM_EMITTER_REFERENCE_COMMAND.
 */

static DM_Object* 
DM_read_reference (DM_Context  *context,
                   DM_InBuffer *inbuffer,
                   DM_nat1     command ) 
{
  DM_Object *o;
  DM_Port   *port  = inbuffer->port;
  DM_Exit   *exit = DM_Allocate0(context, Exit);

  DM_DebugAtLevel(9,
    switch ( command ) {
    case DM_REMOTE_REFERENCE_COMMAND: {
      fprintf(DM_FindErrorStream(context), "<<GeneralReference ");
      break;
    }
    case DM_EMITTER_REFERENCE_COMMAND: {
      fprintf(DM_FindErrorStream(context), "<<ReferenceOwnedByEmitter ");
      break;
    }
    default: {
      DM_Abort(context, DME_WRONG_COMMAND);
    }
    }
);

    exit->counter  = 0;
    exit->object   = DM_NULL;
    DM_receive_key(context, inbuffer, &(exit->number));
    DM_receive_external_options(context, inbuffer, &(exit->options));
    DM_ClearOption(exit, DM_EXTERNAL_IS_EXTERNALIZED);
    if ( DM_IsA(context, port, Connection) ) {
      exit->from     = ((DM_Connection*)port)->site;
    } else {
      /* Not possible to receive such a command from a port that is
       * not a Connection. */
      DM_Abort(context, DME_NOT_CONNECTION);
    }
    /* Decode the class of the remotely pointed object. */
    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Class) ) {
      DM_DeclareNarrowedObject(class, Class, o);

      /* Give a chance to the class to be usable without wait. */
      exit->class = class;
      (void) DM_is_class_usable(context, class);
    } else if ( DM_IsA(context, o, Exit) ) {
      /* This is an Exit item to a remote class. Trust it! */
      DM_DeclareNarrowedObject(ex, Exit, o);

      exit->class = (DM_Class*)ex;
      /* It is possible here to try to prefetch the class. */
      if ( DM_HasOption(port, DM_PORT_WITHOUT_PREFETCHING) ) {
        /* Don't prefetch things. */
      } else {

        DM_DebugAtLevel(7,
          fprintf(DM_FindErrorStream(context), 
                  "Prefetch class of " DM_0X "%p\n", ex);
);

        DM_try_prefetch_exit(context, ex);
      }
    } else {
      /* Not a class nor an Exit item: surely an error. */
      DM_Abort(context, DME_NOT_CLASS);
    }

    /* Even if immutable, decode a clock (NULL if immutable). */
    o = DM_do_read_object(context, inbuffer);
    if ( o == DM_NULL ) {
      /* No clock, the class should be immutable then. */

      DM_DebugAtLevel(15,
        if ( DM_IsA(context, exit->class, Class) ) {
          if ( DM_HasOption(exit->class, DM_CLASS_IS_MUTABLE) ) {
            DM_Abort(context, DME_MUTABLE_CLASS_WITHOUT_CLOCK);
          } else {
            /* NOTHING, this is ok. */
          }
        } else {
          /* NOTHING: The class is not present so we cannot check. */
        }
);

    } else if ( DM_IsA(context, o, Clock) ) {
      exit->clock = (DM_Clock*)o;
    } else if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(ex, Exit, o);
      
      /* Trust o to be a remote clock ! */
      exit->clock = (DM_Clock*)ex;
      if ( DM_HasOption(port, DM_PORT_WITHOUT_PREFETCHING) ) {
        /* don't prefetch things. */
      } else {
        /* Prefetch the Clock to ease the reception of future
           DM_UPDATE_CLOCK_COMMANDs */

        DM_DebugAtLevel(7,
          fprintf(DM_FindErrorStream(context), 
                  "Prefetch clock of " DM_0X "%p\n", ex);
);

        DM_try_prefetch_exit(context, ex);
      }
    } else {
      /* Not NULL, not a Clock, not an Exit: surely an error. */
      DM_Abort(context, DME_NOT_CLOCK);
    }
    exit->time     = 0;
    /* This is where the difference between the two commands is: */
    switch ( command ) {
    case DM_REMOTE_REFERENCE_COMMAND: {
      /* Read a site.  */
      o = DM_do_read_object(context, inbuffer);
      if ( DM_IsA(context, o, Site) ) {
        exit->owner = (DM_Site*)o;
      } else {
        DM_Abort(context, DME_NOT_SITE);
      }
      DM_receive_key(context, inbuffer, &(exit->number));
      break;
    }
    case DM_EMITTER_REFERENCE_COMMAND: {
      /* The owner are the from value. */
      exit->owner = exit->from;
      break;
    }
    default: {
      DM_Abort(context, DME_WRONG_COMMAND);
    }
    }

    /* Check if a similar Exit item already exists. */
    { 
      DM_Object *prev = DM_register_exit(context, exit);
      
      if ( DM_IsA(context, prev, Exit) ) {
        DM_DeclareNarrowedObject(prevexit, Exit, prev);

        if ( prevexit != exit ) {
          /* There is another similar Exit item, Improve the old one,
           * if possible, on its class and clock fields. Then discard
           * the new one: don't forget to send a decrement message. */
          DM_ShortCutReference(context, &(prevexit->class));
          DM_ShortCutReference(context, &(exit->class));
          if ( DM_IsA(context, prevexit->class, Exit) ) {
            if ( DM_IsA(context, exit->class, Class) ) {
              prevexit->class = exit->class;
            }
          }
          DM_ShortCutReference(context, &(prevexit->clock));
          DM_ShortCutReference(context, &(exit->clock));
          if ( DM_IsA(context, prevexit->clock, Exit) ) {
            if ( DM_IsA(context, exit->clock, Clock) ) {
              prevexit->clock = exit->clock;
            }
          }
          /* Discard the new Exit item. DM_remove_exit allows
           * registered exit items to be removed from the exits
           * hashtable, this is not the case of 'exit' since it was
           * not registered. We are also sure that the
           * DM_EXIT_DECREMENT_SENT option is not set on 'exit'. */

          DM_DebugAtLevel(2,
            fprintf(DM_FindErrorStream(context),
                    "GC: Freeing " DM_0X "%p\n", exit);
);

          exit->next = exit->from->freed;
          exit->from->freed = exit;
          /* Substitute the old exit item to the new one. */
          exit = prevexit;
        } else {
          /* NOTHING: we never heard before of this Exit and it is
           * already registered. */
        }
      } else {
        /* Incoherent. */
        DM_Abort(context, DME_INTERNAL_ERROR);
      }
    }
    /* Always try to return a present object rather than an Exit item. */
    if ( exit->object ) {
      o = exit->object;
    } else {
      o = (DM_Object*)exit;
    }

    DM_DebugAtLevel(7,
    switch ( command ) {
    case DM_REMOTE_REFERENCE_COMMAND: {
      fprintf(DM_FindErrorStream(context), " GeneralReference>>");
      break;
    }
    case DM_EMITTER_REFERENCE_COMMAND: {
      fprintf(DM_FindErrorStream(context), " ReferenceOwnedByEmitter>>");
      break;
    }
    default: {
      DM_Abort(context, DME_WRONG_COMMAND);
    }
    }
);

    return o;
}

/* RESOURCE DM_do_read_object
 * Read an object from an Inbuffer.
 * Normally the inbuffer is sufficiently big to hold the entire object;
 * nevertheless to prevent problems, the inbuffer is still checked to 
 * be big enough when extracting characters. This function returns a
 * present object or an Exit item.

 * There are some goodies: 
 *     # starts a comment finishing with the next CR or LF
 *     whitespaces are ignored and so are CR and LF.
 * This should allow to have bytecoded modules using the #! convention.
 * FUTURE: allow modules to use uuencoded bytecode directly.
 */

static DM_Object*
DM_do_read_object (DM_Context  *context,
                   DM_InBuffer *inbuffer) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_PredefinedClassesVector *predefined_classes = DM_ThePredefinedClasses();
  DM_Port           *port  = inbuffer->port;
  DM_Object         *o;
  DM_Class          *class;
  DM_Field          *field;
  DM_Site           *site;
  DM_Type           *type;
  DM_nat1           command, index, index2;
  enum {handle, skip} mode = handle;

 read_command:

  if ( !(DM_CanReceiveBytes(context, inbuffer, 1)) )
    DM_Abort(context, DME_MISSING_BYTES);

  command = DM_ReceiveOneByte(context, inbuffer);
  if ( mode == skip ) {
    if ( (command == DM_CR_COMMAND) || (command == DM_LF_COMMAND) ) {
      mode = handle;
    }
    goto read_command;
  }
  switch ( command ) {
    
  case DM_NIL_COMMAND:
    return DM_TheConstant(nil);
  case DM_TRUE_COMMAND:
    return DM_TheConstant(true);
  case DM_FALSE_COMMAND:
    return DM_TheConstant(false);
  case DM_UNINITIALIZED_COMMAND:
    return DM_TheConstant(uninitialized);
  case DM_NULL_COMMAND:
    return DM_NULL;

  /* Ignore these commandes. */
  case DM_CR_COMMAND: 
  case DM_LF_COMMAND: 
  case DM_BLANK_COMMAND: {
    goto read_command;
  }

  case DM_SHARP_COMMAND: {
    /* Turn to skip mode and ignore other bytes until end of line. */
    mode = skip;
    goto read_command;
  }

  /* Reset cache. */

  case DM_RESET_CACHE_COMMAND: {

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<ResetCache>>");
);

    DM_reset_cache(context, inbuffer->cache);
    goto read_command;
  }


  /* General stack-oriented commands: reset, push, pop, dup, swap */

  case DM_RESET_STACK_COMMAND: {

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<ResetStack>>");
);

    DM_reset_stack(context, inbuffer->stack);
    goto read_command;
  }

  case DM_PUSH_STACK_COMMAND: {
    /* Read an object, cache it and return it. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Push ");
);

    o = DM_read_object(context, inbuffer);
    DM_push_object_in_bounded_stack(context, o, &(inbuffer->stack));

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Pushed>>");
);

    return o;
  }

  case DM_POP_STACK_COMMAND: {
    /* Pop the cache. */
    o = DM_pop_from_bounded_stack(context, inbuffer->stack);
 
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Popped " DM_0X "%p>>", o);
);

    return o;
  }

  case DM_SWAP_STACK_COMMAND: {
    /* Swap first two elements of the cache. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Swap ");
);

    DM_swap_top_of_bounded_stack(context, inbuffer->stack);
     
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Swapped>>");
);

    goto read_command;

  }

  case DM_DUP_STACK_COMMAND: {
    /* Duplicate the top of the stack. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Dup ");
);

    DM_dup_top_of_bounded_stack(context, &(inbuffer->stack));
     
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Dupped>>");
);

    goto read_command;

  }

  case DM_DECREMENT_COMMAND: {
    /* Decrement a reference counter then continue to decode the object.
     * This is a dangerous primitive if one can invent keys. */
    DM_Entry *entry;
    DM_key   number;

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Decrement ");
);

    DM_receive_key(context, inbuffer, &number);
    entry = DM_extract_entry(context, number);
    if ( entry ) {
      DM_decrement_counter(context, entry);
    } else { 
      /* Ignore that command unless in debugging mode. */

      DM_DebugAtLevel(3,
        fprintf(DM_FindErrorStream(context), "Lost Entry");
        DM_Abort(context, DME_LOST_ENTRY);
);

    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Decrement>>");
);

    goto read_command;
  }

  case DM_EMITTING_SITE_COMMAND: {
    /* This command encodes the emitting site. This saves some bytes. */
    if ( DM_IsA(context, port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, port);

      if ( connection->site ) {

        DM_DebugAtLevel(7,
         fprintf(DM_FindErrorStream(context), "<<EmittingSite>>");
);

        return (DM_Object*)(connection->site);
      } else {
        DM_Abort(context, DME_BAD_SITE);
        return DM_NULL;
      }
    } else {
      DM_Abort(context, DME_NOT_CONNECTION);
      return DM_NULL;
    }
  }

  case DM_RECEIVING_SITE_COMMAND: {
    /* This command encodes the receiving site. This saves some bytes. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<ReceivingSite>>");
);

    return (DM_Object*)DM_TheCurrentSite();
  }

  case DM_LIMIT_COMMAND: {
    /* This command is followed by a nat3 and an object. It returns
     * this object but ensures that the encoding of this object is
     * less than nat3 bytes. After reading the object and if less than
     * nat3 bytes were read, additional bytes are skipped. */
    DM_nat3   current_index = inbuffer->index;
    DM_nat3   limit;
    DM_nat3   height;
    volatile DM_Object *o = DM_NULL;
    
    height = inbuffer->stack->index;
    DM_receive_nat3(context, inbuffer, &limit);

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), 
              "<<Limit(%lu) ", (unsigned long)limit);
);

    DM_ProtectWithFreshContext(context, newcontext,
           o = DM_do_read_object(newcontext, inbuffer),
       /* If something went wrong: */
       o = DM_NULL;
                           );

    DM_DebugAtLevel(7,
      if ( inbuffer->index != (current_index + limit) ) {
        fprintf(DM_FindErrorStream(context), "LIMIT mismatch %lu<>%lu\n", 
                (unsigned long)(inbuffer->index), 
                (unsigned long)(current_index + limit) );
      }
      fprintf(DM_FindErrorStream(context), " Limit>>");
);

    /* Reset the index (believe the limit not the current index). */
    inbuffer->index = current_index + limit;
    /* Reset the height of the stack. */
    DM_adjust_stack(context, inbuffer->stack, height);
    return (DM_Object*)o;
  }

  case DM_RECEIVE_ANSWER_COMMAND:
  case DM_RECEIVE_REQUEST_COMMAND: {
    /* This command reads a Request or an Answer. It will not be
     * handled on the fly since if it were to require a message from
     * the same emitting site, this would dead-lock since the decoding
     * of the current message is suspended meanwhile thus making the
     * decoding of following messages blocked. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<ReceiveRequest/Answer ");
);

    o = DM_do_read_object(context, inbuffer);

    DM_DebugAtLevel(17,
      /* Log requests and answers for debug. */
      DM_memorize_received_object(context, o);
);

    /* Accumulate the Request or Answer. */
    DM_enqueue_received_request_or_answer(context, o);

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " ReceiveRequest/Answer>>");
);

    return o;
  }
      
  case DM_PROG2_COMMAND: {
    /* This command reads two objects and return the second one.
     * For instance you may emit an object, it will appear in the cache
     * (both receiving and emitting caches) and be used in the second
     * object. This trick allows to pre-send objects.
     */
    DM_Object *o1;
    DM_Object *o2;

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Prog2 ");
);

    o1 = DM_do_read_object(context, inbuffer);
    o2 = DM_do_read_object(context, inbuffer);

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Prog2>>");
);

    return o2;
  }

  case DM_PROG1_COMMAND: {
    /* This command reads two objects and return the first one.
     */
    DM_Object *o1;
    DM_Object *o2;

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Prog1 ");
);

    o1 = DM_do_read_object(context, inbuffer);
    o2 = DM_do_read_object(context, inbuffer);

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Prog1>>");
);

    return o1;
  }

  case DM_SET_SHARABLE_COMMAND: {
    /* This command reads an object and makes it sharable. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<SetSharable ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, o);

      /* PROBLEM: what if the exit means that the remote object was
         exported as copyable ??? */
      DM_ClearOption(exit, DM_ENTRY_IS_UNINITIALIZED);
      DM_ClearOption(exit, DM_ENTRY_IS_COPYABLE);
    } else {
      DMeroon_set_sharable(context, o);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " SetSharable>>");
);

    return o;
  }

  case DM_SET_COPYABLE_COMMAND: {
    /* This command reads an object and makes it copyable. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<SetCopyable ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, o);

      /* PROBLEM: what if the exit means that the remote object was
         exported as sharable ??? */
      DM_ClearOption(exit, DM_ENTRY_IS_UNINITIALIZED);
      DM_SetOption(exit, DM_ENTRY_IS_COPYABLE);
    } else {
      DMeroon_set_copyable(context, o);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " SetCopyable>>");
);

    return o;
  }

  case DM_ALLOCATE_COMMAND: {
    /* This command allocates a local object. 
     * INVARIANT: The class of the object should
     * necessarily be known before, otherwise the incoming object
     * cannot be decoded. At last the class must be received just
     * before the object itself. 
     */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Allocate ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Class) ) {
      DM_DeclareNarrowedObject(class, Class, o);

      if ( DM_is_class_usable(context, class) ) {
        /* Here we return a present object with a present class. */
        o = DM_read_object_size_and_allocate(context, inbuffer, class);
      } else {
        DM_Abort(context, DME_UNUSABLE_CLASS);
      }
    } else {
      /* The class is not present. */
      DM_Abort(context, DME_NOT_CLASS);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Allocated>>");
);

    return o;
  }

  case DM_FILL_COMMAND: {
    /* Fill an object. It must not be an Exit item.  This is a
     * dangerous primitive since it will fill the object obtained as
     * first argument. 
     
     * NOTE: we may limit FILL to only fill ALLOCATED objects with a 
     * special tag ??? FUTURE ???
     */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Fill ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Exit) ) {
      DM_Abort(context, DME_CANNOT_FILL);
    } else {
      DM_read_object_content(context, inbuffer, o);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Filled>>");
);

    return o;
  }

  case DM_BIND_COMMAND: {
    /* Receive an object and an exit then associate them. This is
     * useful to pre-send an object.  Return the object. */
    DM_Object* e;

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Cache ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Entry) ) {
      DM_Abort(context, DME_NOT_REGULAR_OBJECT);
    }
    e = DM_do_read_object(context, inbuffer);
    /* e may be an Entry, an Exit or a regular object already
     * associated to an Entry or Exit item. */
    if ( DM_IsA(context, e, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, e);
      
      if ( exit->object ) {
        /* Already an associated object !? */
        if ( o == exit->object ) {
          /* These are the same, the command is useless but harmless. */
        } else {
          /* Cannot associate an object and an Exit item cached with
           * another different object. */
          DM_Abort(context, DME_INCORRECT_CACHE_ARGUMENTS);
        }
      } else {
        /* Associate the Exit item and the object.  Verify that the
        * object has no associated Entry/Exit (remember, in the next
        * test, that Exit is a subclass of Entry). */
        DM_Object *proxy = DM_ObjectToProxy(context, o);

        if ( DM_IsA(context, proxy, Entry) ) {
          /* This would be very weird. */
          DM_Abort(context, DME_INTERNAL_ERROR);
        } else {
          exit->object = o;
          DM_SetProxyOf(context, o, exit);
          /* Also verify that the classes are the same. (what if the
             class of the Exit was a superclass of the exact class of
             the remote object !? Bad idea, we can no longer unmarshal
             the content of the object if we want to update it.)  
             FUTURE: should shortcut exit->class to be a real class ???  
             */
          if ( exit->class != DM_ObjectToClass(context, o) ) {
            DM_Abort(context, DME_INCORRECT_CLASSES);
          }
        }
      }
    } else if ( DM_IsA(context, e, Entry) ) {
      /* Must not associate (from the outside) an Entry and an object. */
      DM_Abort(context, DME_INCORRECT_CACHE_ARGUMENTS);
    } else {
      /* e should be an Exit item but might have been coerced into an
       * already cached object (due to the automatic dereference of
       * Exit/Entry items). Same verifications as above. */
      if ( e == o ) {
        /* Already associated: NOTHING to do. */
      } else {
        /* Even if e has an Exit or Entry as proxy, it is not compatible. */
        DM_Abort(context, DME_INCORRECT_CACHE_ARGUMENTS);
      }
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Cached>>");
);

    return o;
  }
        

  case DM_RECEIVER_REFERENCE_COMMAND: {
    /* Receive a reference onto an object owned (or reexported) by the
     * receiving site. This is used to shorten messages if detected by
     * the emitting site. This command returns of course an Object or
     * an Exit item. */
    DM_Entry *entry;
    DM_key   number;
    
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<ReferenceOwnedByReceiver ");
);

    DM_receive_key(context, inbuffer, &number);
    entry = DM_extract_entry(context, number);
    if ( entry ) {
      o = entry->object;
    } else {
      DM_Abort(context, DME_LOST_ENTRY);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " ReferenceOwnedByReceiver>>");
);

    return o;
  }

  case DM_EMITTER_REFERENCE_COMMAND: 
    /* Receive a reference onto an object owned by the emitting
     * site. This command may return an Exit item if the remote object
     * is not yet locally cached. INVARIANT: the class (resp clock)
     * field of the Exit item should hold a Class (resp NULL or Clock)
     * or an Exit item to them.  */

  case DM_REMOTE_REFERENCE_COMMAND: {
    /* Receive a reference onto an object not belonging to the
     * emitting site (and probably not belonging to the receiving site
     * due to the availability of immediate_references). This command
     * may return an Exit item if the remote object is not yet locally
     * cached.  */
    return DM_read_reference(context, inbuffer, command);
  }

  case DM_SITE_COMMAND: {
    /* This command is used to cite (or create) a site. It always
     * return a Site instance even if the site was not already
     * known. This command is followed by the main characteristics of
     * a site: ip+port numbers, creation date and length of name if
     * the site object has to be allocated. */
    DM_nat3 site_ip;
    DM_nat2 site_port;
    DM_date site_date;
    DM_nat1 sitename_length;
    
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Site ");
);

    DM_receive_nat3(context, inbuffer, &site_ip);
    DM_receive_nat2(context, inbuffer, &site_port);
    DM_receive_date(context, inbuffer, &site_date);
    DM_receive_nat1(context, inbuffer, &sitename_length);
    site = DM_extract_site(context, site_ip, site_port);
    if ( site ) {
      /* A site on the same machine and port is already known. When a
       * date is specified (ie not null), that means that we want a
       * particular site. Recall that 0 is 1/1/1970 and no DMeroon was
       * living at that time! So 0 means "don't care".  */
      if ( site->name_length != sitename_length ) {
        DM_Abort(context, DME_INCOMPATIBLE_SITENAME_LENGTH);
      }
      if ( site->date && site_date && site->date != site_date ) {
        DM_Abort(context, DME_INCOMPATIBLE_SITE_DATES);
      }
    } else {
      /* This is a site never seen before. */
      site          = DM_Allocate1(context, Site, sitename_length);
      site->address = htonl(site_ip);
      site->port    = htons(site_port);
      site->date    = site_date;
      site->options = DM_SITE_DEFAULT_OPTIONS;
      DM_ClearOption(site, DM_SITE_IS_USABLE);
      /* CAUTION: This option means that a registered remote site may
       * be not associated with an Exit item. This is why there is
       * this option and the DM_ensure_site_usable function.  */
      DM_register_site(context, site);
      /* Record from where we heard of that site. The 'route' field is
       * used to send messages to that site if no direct connection is
       * opened towards it. Recall that an unknown site has no
       * associated exit item. */
      if ( DM_IsA(context, port, Connection) ) {
        DM_DeclareNarrowedObject(connection, Connection, port);

        if ( connection->site ) {
          site->route = connection->site;
        }
        site->password = connection->password;
      } else {
        DM_Abort(context, DME_NOT_CONNECTION);
      }
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Site>>");
);

    return (DM_Object*)site;
  }

  case DM_DEFINE_EMITTER_COMMAND: {
    /* This command is used to register a Site instance. It is followed
     * by a Site and an ownerid. It links the Site and the Exit and
     * registers them. It also initializes the current connection. 
     * This command is only used when booting up a connection. */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<RegisterEmittingSite ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Site) ) {
      /* Since this is the emitter site, this cannot be the current site. */
      DM_DeclareNarrowedObject(site, Site, o);
      DM_Object *proxy = DM_ObjectToProxy(context, site);
      DM_Exit   *exit  = DM_Allocate0(context, Exit);
      
      exit->counter = 0;
      exit->options = DM_EXTERNAL_DEFAULT_OPTIONS;
      exit->object  = o; /* ie site */
      exit->clock   = DM_NULL;
      exit->class   = DM_ReferenceToClass(Site);
      exit->time    = 0;
      DM_receive_key(context, inbuffer, &(exit->number));
      exit->owner = site;
      exit->from = site;
      if ( DM_IsA(context, proxy, Exit) ) {
        /* The site was already associated to an Exit item. Check! */
        DM_DeclareNarrowedObject(p, Exit, proxy);

        if ( ! DM_EqualKey(p->number, exit->number) )
          DM_Abort(context, DME_INCOMPATIBLE_KEY);
        /* forget about exit since the remote site already has an
         * associated proxy. Since exit was not registered it is safe
         * to forget it, but we give it an option so the GC will not
         * send a decrment message. */
        DM_SetOption(exit, DM_EXIT_DECREMENT_SENT);
      } else {
        /* The site had no exit, give it the current one. */
        DM_SetProxyOf(context, site, exit);
        DM_register_exit(context, exit);
        if ( DM_HasOption(DM_TheCurrentSite(), DM_SITE_WANTS_FACTS) ) {
          /* Report a fact if required by the site. */
          DM_Fact *fact = DM_Allocate1(context, Fact, 1);
      
          fact->errcode = DME_GOT_SITE;
          fact->item[0] = (DM_Object*)site;
          if (    context 
               && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
            DM_SetCopyable(context, fact);
          }
          DM_enqueue_received_object(context, (DM_Object*)fact);      
        }
      }
    } else {
      /* Error: this is not a site. */
      DM_Abort(context, DME_NOT_A_SITE);
    }
    site = (DM_Site*)o;
    /* Initialize the emitting site of the current connection. */
    if ( DM_IsA(context, port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, port);

      if ( connection->site ) {
        /* The current connection has already its 'connection' field
         * filled. Check! */
        if ( connection->site != (DM_Site*)o ) {
          DM_Abort(context, DME_INCOMPATIBLE_SITE);
        }
      } else {
        /* Record the site at the other end of the current connection. */
        connection->site = (DM_Site*)o;
      }
      /* Initialize the connection to use to send message to site. */
      site->connection = connection;
      site->password = connection->password;
    } else {
      DM_Abort(context, DME_NOT_CONNECTION);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " EmittingSiteRegistered>>");
);

    return o;
  }

  case DM_UPDATE_CLOCK_COMMAND: {
    /* Be notified of a value a Clock had. Update the locally cached
     * clock if time has evolved. These commands normally prefix the
     * sent object. */
    DM_Site  *site  = DM_TheCurrentSite();
    DM_nat4  newtime;

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<UpdateClock ");
);

    o = DM_do_read_object(context, inbuffer);
    DM_receive_nat4(context, inbuffer, &newtime);
    if ( DM_IsA(context, o, Clock) ) {
      DM_DeclareNarrowedObject(clock, Clock, o);

      if ( clock->time < newtime ) {
        /* Update perception of that clock. */
        clock->time = newtime;
        site->clocker++;
        clock->clocker = site->clocker;

        DM_DebugAtLevel(0,
          fprintf(DM_FindErrorStream(context), "Ah! Clock " DM_0X "%p is now 0x", clock);
          DM_nat4_to_string(context, 
                            clock->time,
                            (DM_SupportPointer)context->errstream);
          fprintf(DM_FindErrorStream(context), " \n");
);

      } else {
        /* We already know that the clock is beyond that time, so ignore it. */
      }
    } else if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, o);

      if ( DM_HasOption(port, DM_PORT_WITHOUT_PREFETCHING) ) {
        /* don't prefetch things. */
      } else {
        /* The Clock is not locally cached.  Ignore the message for now
         * but ask for its content with a non-blocking request. */

        DM_DebugAtLevel(0,
          fprintf(DM_FindErrorStream(context), 
                  "Prefetch clock " DM_0X "%p\n", exit);
);

        DM_try_prefetch_exit(context, exit);
      }

      DM_DebugAtLevel(0,
        fprintf(DM_FindErrorStream(context), 
                "Ignored remote clock " DM_0X "%p\n", exit);
);

    } else {
      DM_Abort(context, DME_NOT_CLOCK);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " UpdateClock>>");
);

    goto read_command;
  }

  case DM_FIELD_COMMAND: {
    /* Read a Field
     * Part of the offsets (translations) depend on the current site.
     * they must be recomputed. The simplest is to make the class 
     * unusable so its layout will be recomputed. Another solution would be 
     * to modify DM_update_object to take into account offsets better, to
     * respect the translation part which is local from the reptyp and 
     * itemtyp parts which are ubiquitous.       FUTURE */

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Field ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Field) ) {
      DM_DeclareNarrowedObject(field, Field, o);

      DM_ShortCutReference(context, &(field->introducing));
      if ( DM_IsA(context, field->introducing, Class) ) {
        /* Invalidate the class */
        DM_ClearOption(field->introducing, DM_CLASS_IS_USABLE);
      } else {
        /* NOTHING: The introducing class is not present. */
      }
    } else if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(ex, Exit, o);

      if ( DM_HasOption(port, DM_PORT_WITHOUT_PREFETCHING) ) {
        /* don't prefetch things. */
      } else {

        DM_DebugAtLevel(7,
          fprintf(DM_FindErrorStream(context), 
                  "Prefetch field " DM_0X "%p\n", ex);
);

        DM_try_prefetch_exit(context, ex);
      }
      /* Caution: Here we return an Exit item. */
      o = (DM_Object*)ex;
    } else {
      /* Not a Field, not an Exit item, surely an error. */
      DM_Abort(context, DME_NOT_FIELD);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Field>>");
);

    return o;
  }

  case DM_CLASS_COMMAND: {
    /* Read a Class (and its options).
     * Don't recompute layout since some fields may be not present.
     * In fact, layout will be recomputed in DM_ensure_class_usable
     * when really needed. This command may return an Exit item. 
     * NOTE WHEN DEBUGGING: the read class is local since the request
     * asking for it is not handled yet.
     */
    
    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<Class ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Class) ) {
      /* NOTE: Currently, the class is always reinitialized even if
       * it is perfectly known. This is because the sender always reset
       * the usability of the class.                             FUTURE 
       * SHOULD TEST when the class is coming from two different sites.... */
      /* Insert the class itself as its first superclass. */
      DM_DeclareNarrowedObject(class, Class, o);

      class->super[class->super_length-1] = class;
      /* Give a chance to the class to be usable. */
      (void) DM_is_class_usable(context, class);

    } else if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(ex, Exit, o);

      if ( DM_HasOption(port, DM_PORT_WITHOUT_PREFETCHING) ) {
        /* don't prefetch things. */
      } else {

        DM_DebugAtLevel(7,
          fprintf(DM_FindErrorStream(context),
                  "Prefetch class " DM_0X "%p\n", ex);
);

        DM_try_prefetch_exit(context, ex);
      }
      /* Caution: Here we return an Exit item. */
      o = (DM_Object*)ex;
    } else {
      /* Not a Class, not an Exit item, surely an error. */
      DM_Abort(context, DME_NOT_CLASS);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " Class>>");
);

    return o;
  }

  case DM_SYMBOL_COMMAND: {
    /* Read a Symbol then intern it in the local oblist. Return the
     * interned symbol (which may be different if the oblist already
     * contains it.
     */

    DM_DebugAtLevel(0,
      fprintf(DM_FindErrorStream(context), "<<Symbol= ");
);

    o = DM_do_read_object(context, inbuffer);
    if ( DM_IsA(context, o, Symbol) ) {
      o = (DM_Object*)DM_intern_symbol(context, (DM_Symbol*)o);
    } else {
      DM_Abort(context, DME_NOT_SYMBOL);
    }

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "EndSymbol>>");
);

    return o;
  }

  default: {
    /* All the following commands take at least one byte as
     * argument. Factorize this code waiting for this byte to be
     * present. */
    if ( !(DM_CanReceiveBytes(context, inbuffer, 1)) )
      DM_Abort(context, DME_MISSING_BYTES);
    
    index = DM_ReceiveOneByte(context, inbuffer);
    switch ( command ) {
      
    case DM_PREDEFINED_CLASS_COMMAND: {
      class = predefined_classes->item[index];

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<PredefinedClass=%s>>", 
              DM_predefined_name(context, (DM_Object*)class) );
);

      return (DM_Object*)class;
    }
      
    case DM_PREDEFINED_TYPE_COMMAND: {

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<PredefinedType= ");
);

      type = (DM_Type*)implemented_types->item[index];

      DM_DebugAtLevel(7,
        fprintf(DM_FindErrorStream(context), "%s>>", 
                DM_predefined_name(context, (DM_Object*)type) );
);

      return (DM_Object*)type;
    }

    case DM_REFER_WITH_NAT1_COMMAND: {
      /* Extract a cached object. The index is a nat1 number. */

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<GetCache[%d] ", index);
);

      o = DM_extract_from_cache(context, inbuffer->cache, index);

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " GetCache>>");
);

      return o;
    }
      
    case DM_INSERT_WITH_NAT1_COMMAND: {
      /* Replace a cached object. The index is a nat1 number. */

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<SetCache[%d] ", index);
);

      o = DM_do_read_object(context, inbuffer);    
      inbuffer->cache->item[index] = o;

      DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " SetCache>>");
);

      return o;
    }

    default: {
      /* All the following commands take at least a second byte as
       * argument. Factorize this code waiting for this byte to be
       * present. */
      if ( !(DM_CanReceiveBytes(context, inbuffer, 1)) )
        DM_Abort(context, DME_MISSING_BYTES);
      
      index2 = DM_ReceiveOneByte(context, inbuffer);
      switch ( command ) {
        
      case DM_PREDEFINED_FIELD_COMMAND: {
        class = predefined_classes->item[index];

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<PredefinedField= ");
);

        field = DM_class_field(context, class, index2);

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "%s.%s>>",
              DM_predefined_name(context, (DM_Object*)class),
              DM_predefined_name(context, (DM_Object*)field) );
);

        return (DM_Object*)field;
      }
      
      case DM_REFER_WITH_NAT2_COMMAND: {
        DM_nat2 i = 256*index + index2;

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<GetCache[%d] ", i);
);

        o = DM_extract_from_cache(context, inbuffer->cache, i);

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " GetCache>>");
);

        return o; 
      }

      case DM_INSERT_WITH_NAT2_COMMAND: {
        DM_nat2 i = 256*index + index2;

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), "<<SetCache[%d] ", i);
);

        o = DM_do_read_object(context, inbuffer);
        inbuffer->cache->item[i] = o;

        DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), " SetCache>>");
);

        return o;
      }

      default: 
        DM_Abort(context, DME_UNMARSHALING_ERROR);
        return DM_NULL;
      } /* end switch with two byte arguments */
    }   /* end of default clause for one byte argument */
    }   /* end switch with one byte argument */
  }     /* end of default clause for zero byte argument */
  }     /* end switch with zero byte */
}

/* RESOURCE DM_read_object
 * This is the entry point of that file. It unmarshals an object.
 */

DM_Object*
DM_read_object (DM_Context  *context,
                DM_InBuffer *inbuffer) 
{
  return DM_do_read_object(context, inbuffer);
}

/*
 * Initialize the module and check some properties.
 */

void
DM_rdobj_initialize (DM_Context *context)
{
  DM_LinkVersionStringIntoSite(dmrdobj);

  /* Site must be immutable and site must be a direct instance of Site. */
  { DM_Class *siteclass = DM_ReferenceToClass(Site);
    DM_Site  *site  = DM_TheCurrentSite();
    
    if ( DM_HasOption(siteclass, DM_CLASS_IS_MUTABLE) )
      DM_Abort(context, DME_BASIC_CLASS_MUTABLE);
    if ( DM_ObjectToClass(context, site) != siteclass )
      DM_Abort(context, DME_NOT_DIRECT_INSTANCE);
  }
}

/* end of dmrdobj.c */
