/* $Id: dmreq.c,v 1.68 1998/06/10 16:31:30 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 
 * DOCUMENTATION dmreq.c
 * This file implements the DMeroon requests and their answers. It is
 * very important that these functions never block! In effect, as soon
 * as a Request or an Answer is unmarshaled, it is processed
 * immediately; therefore the decoding/processing of the other
 * messages that are already present in the same Connection (as well
 * as the future messages that may arrive) are suspended. So a Request
 * cannot wait for an Answer on the same Connection it arrives itself 
 * otherwise we will have a deadlock.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmreq,
  "$Id: dmreq.c,v 1.68 1998/06/10 16:31:30 queinnec Exp $");

#if defined(DM_DEBUG) && (DM_DEBUG>6)

/* {{{ Utility
 *
 * This is where Requests/Answers are stored for debug purpose. This
 * dictionary is initialized in DM_req_initialize().
 */

DM_BoundedStack* DM_lastreqans = DM_NULL;

/* RESOURCE DM_memorize_received_object
 * Memorize an object somewhere where it can be inspected by a debugger.
 * It actually records all incoming objects.
 */

void
DM_memorize_received_object (DM_Context *context,
                             DM_Object  *o) {
  DM_Class *class = DM_ObjectToClass(context, o);

  if ( DM_IsA(context, o, External) ) {
    DM_DeclareNarrowedObject(external, External, o);

    fprintf(DM_FindErrorStream(context), "Received %s",
            DM_predefined_name(context, (DM_Object*)class) );
    DM_key_print(context, &(external->number), context->errstream);
    fprintf(DM_FindErrorStream(context), "\n");
  }
  DM_push_object_in_bounded_stack(context, o, &DM_lastreqans);
}

/* RESOURCE DM_memorize_sent_object
 * Memorize an object somewhere where it can be inspected by a debugger.
 * It actually records all outgoing objects.
 * They are actually stored in the information field of the current site.
 */

void
DM_memorize_sent_object (DM_Context *context,
                         DM_Object  *o) {
  DM_Class *class = DM_ObjectToClass(context, o);

  if ( DM_IsA(context, o, External) ) {
    DM_External *external = (DM_External*)o;

    fprintf(DM_FindErrorStream(context), "Sending %s",
            DM_predefined_name(context, (DM_Object*)class) );
    DM_key_print(context, &(external->number), context->errstream);
    fprintf(DM_FindErrorStream(context), "\n");
  }
  DM_push_object_in_bounded_stack(context, o, &DM_lastreqans);
}

#endif

/* }}}
 * {{{ Requests

 * RESOURCE DM_handle_site_connection_request
 * When receiving a SiteConnectionRequest, the current site decodes
 * the requesting site object and answers by giving its site to the
 * requesting (remote) site. It also toggles the used Connection in
 * DMeroon inner protocol mode. This function is special since it is
 * used to initialize the Connection. It can also be used after to
 * reinitialize the exchange of site objects.

 * This function answers the Request emitted by DM_site_handshake.
 */

void
DM_handle_site_connection_request (DM_Context               *context,
                                   DM_SiteConnectionRequest *request) {
  DM_Site       *requesting  = request->requesting;
  /* should verify that connection and outbuffer are usable.   FUTURE */
  DM_Connection *connection  = requesting->connection;
  DM_OutBuffer  *outbuffer   = connection->emission;
  DM_SiteConnectionAnswer *answer = 
    DM_Allocate0(context, SiteConnectionAnswer);
  DM_Site       *site        = DM_TheCurrentSite();
  DM_Class      *class       = DM_ReferenceToClass(SiteConnectionAnswer);
  DM_nat3       site_ip      = ntohl(site->address);
  DM_nat2       site_portnum = ntohs(site->port);
  DM_date       site_date    = site->date;
  DM_Entry      *site_entry;

  /* Normally DM_DEFINE_EMITTER_COMMAND already did part of the work.
   * There is only the current site to send back with the same
   * precautions since the Connection is bootstrapped.  
   * FUTURE: no need to allocate the Answer instance, just encode 
   * it on the fly. */
  answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
  answer->requesting = site;
  answer->requested  = request->requesting;
  answer->request    = (DM_Request*)request;
  DM_SetCopyable(context, answer);
  /* Toggle the Connection to the DMeroon protocol mode if not yet done. */
  { char line[100+DMEROON_PASSWORD_LENGTH];
    sprintf(line, "DMeroon%s " DM_RELEASE_LEVEL "\r\n", DM_password);
    DM_print_char_star(context, line, 0, outbuffer);
  }
  /* Reset local cache/stack. The other one will be reset when entering the
   * DMeroon inner protocol mode at the other end of the Connection. */
  DM_reset_stack(context, outbuffer->stack);
  DM_reset_cache(context, outbuffer->cache);
  outbuffer = DM_start_message(context, outbuffer, DM_FALSE);
  /* Transmit a copy of the SiteConnectionAnswer instance. */
  DM_print_byte(context, DM_RECEIVE_ANSWER_COMMAND, outbuffer);
  DM_print_byte(context, DM_SET_COPYABLE_COMMAND, outbuffer);
  DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
  DM_print_byte(context, DM_ALLOCATE_COMMAND, outbuffer);
  DM_print_byte(context, DM_PREDEFINED_CLASS_COMMAND, outbuffer);
  DM_print_byte(context, DM_PredefinedClassNumber(class), outbuffer);
  /* Transmit answer->options. */
  DM_send_nat3(context, outbuffer, &(answer->options));
  /* Transmit answer->requesting (with a special command code). */
  DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
  DM_print_byte(context, DM_DEFINE_EMITTER_COMMAND, outbuffer);
  DM_print_byte(context, DM_SITE_COMMAND, outbuffer);
  /* Send the port which is served (not the one used by the Connection)
   * since this is used to hash the site on the receiving site. */
  DM_send_nat3(context, outbuffer, &site_ip);
  DM_send_nat2(context, outbuffer, &site_portnum);
  DM_send_date(context, outbuffer, &site_date);
  /* Send the size of the name of the Site. */
  DM_send_nat1(context, outbuffer, &(site->name_length));
  /* Send the key of the current site. */
  site_entry = DM_externalize(context, (DM_Object*)site);
  DM_increment_counter(context, (DM_External*)site_entry);
  DM_send_key(context, outbuffer, &(site_entry->number));
  /* And send the content of the site. */
  DM_write_object_content(context, (DM_Object*)site, outbuffer);
  /* Transmit answer->requested (only a reference onto it: that
   * reference will be converted into the other site object on this
   * other site. */
  DM_write_object_reference(context, 
                            (DM_Object*)request->requesting,
                            outbuffer);
  /* Transmit answer->request (only a reference onto it). */
  DM_write_object_reference(context, 
                            (DM_Object*)answer->request,
                            outbuffer);
  DM_finish_message(context, outbuffer, answer->requested);

  DM_DebugAtLevel(0,
    DM_memorize_sent_object(context, (DM_Object*)answer);
);
}

/* RESOURCE DM_handle_object_send_request
 * Receive the reference of an object. No answer to provide. Just
 * call the appropriate handler on it (by default, this is 
 * DM_accumulate_object which enqueues the object).
 */

void
DM_handle_object_send_request (DM_Context           *context,
                               DM_ObjectSendRequest *request) {
  /* No problem if request->object is an Exit since we just write it
   * in a reference field. But we try to prefetch the object
   * immediately. */

  DM_DebugAtLevel(6,
      fprintf(DM_FindErrorStream(context),
              "Received object " DM_0X "%p\n",
              request->object );
);

  if ( DM_IsA(context, request->object, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, request->object);

    DM_DebugAtLevel(7,
      fprintf(DM_FindErrorStream(context), 
              "Prefetch object whose exit item is " DM_0X "%p\n",
              exit );
);

    DM_try_prefetch_exit(context, exit);
  }
  { /* Handle the incoming object. */
    void (*handler)(DM_Context*, DM_Object*) = 
      (void (*)(DM_Context*, DM_Object*))(DM_TheCurrentSite()->handler);
    
    handler(context, request->object);
  }
}

/* RESOURCE DM_handle_object_fetch_request
 * This request asks for an object. Its class (and clock if mutable) are
 * already known from the caller site.
 */

void
DM_handle_object_fetch_request (DM_Context            *context,
                                DM_ObjectFetchRequest *request) {
  DM_ShortCutReference(context,  &(request->object));
  if ( DM_IsA(context, request->object, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, request->object);

    /* The object is not local, forward the request to the owner. */
    DM_SendRequest(context, request, exit->owner);
    return;
  } else {
    /* The object is local and has necessarily an associated Entry. */
    DM_Object *o = request->object;
    DM_DeclareNarrowedObject(entry, Entry, DM_ObjectToProxy(context, o));
    DM_ObjectFetchAnswer *answer = DM_Allocate0(context, ObjectFetchAnswer);

    /* Date the Answer with its clock if mutable or with the current
     * value of ticker otherwise. This is needed so object_fetch_request
     * for volatile fields may be ordered. */
    if (   DM_HasOption(entry, DM_ENTRY_IS_UNINITIALIZED) 
        || ! (entry->clock) ) {
      /* The object is uninitialized or has no associated Clock. */
      answer->time = DM_TheCurrentSite()->ticker++;
    } else {
      answer->time = entry->clock->time;
    }
    DM_SetCopyable(context, answer);
    answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
    answer->requesting = DM_TheCurrentSite();
    answer->requested  = request->requesting;
    answer->request    = (DM_Request*)request;

    /* Awful and expensive hack to use DM_clone. */
    answer->object = DM_clone(context, o);
    /* Also send these options if they change. */ 
    answer->object_options = entry->options;
    /* End of awful hack, 
     * make this object copyable just to send its content. */
    DM_SetCopyable(context, answer->object);

    DM_DebugAtLevel(0,
      DM_memorize_sent_object(context, (DM_Object*)answer);
);

    DM_SendAnswer(context, answer, request->requesting);
  }
}

/* RESOURCE DM_handle_object_mutation_request
 * Receive an update request, perform it and return the previous value of 
 * the mutated field. Don't return the freshly updated object so the
 * remote user may batch up a few mutation requests before asking to
 * refresh the object.
 */

void
DM_handle_object_mutation_request (DM_Context               *context,
                                   DM_ObjectMutationRequest *request) {
  DM_ShortCutReference(context,  &(request->object));
  if ( DM_IsA(context, request->object, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, request->object);

    DM_SendRequest(context, request, exit->owner);
    return;
  } else {
    DM_Object     *o            = request->object;
    DM_Class      *class        = DM_ObjectToClass(context, o);
    DM_ObjectMutationAnswer *answer = 
      DM_Allocate0(context, ObjectMutationAnswer);
    DM_Field *field = DM_class_field(context, class, request->field_index);

    DM_SetCopyable(context, answer);
    answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
    answer->requesting = DM_TheCurrentSite();
    answer->requested  = request->requesting;
    answer->request    = (DM_Request*)request;
    answer->value      = 
      DM_patch_from_reified_value(context, o, field, request->value);
    DM_SetCopyable(context, answer->value);
    DM_SendAnswer(context, answer, request->requesting);

    DM_DebugAtLevel(0,
        DM_memorize_sent_object(context, (DM_Object*)answer);
);

  }
}

void
DM_handle_object_indexed_mutation_request 
               (DM_Context                      *context,
                DM_ObjectIndexedMutationRequest *request) {
  DM_ShortCutReference(context,  &(request->object));
  if ( DM_IsA(context, request->object, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, request->object);

    DM_SendRequest(context, request, exit->owner);
    return;
  } else {
    DM_Object     *o            = request->object;
    DM_Class      *class        = DM_ObjectToClass(context, o);
    /* No need for an ObjectIndexedMutationAnswer. */
    DM_ObjectMutationAnswer *answer = 
      DM_Allocate0(context, ObjectMutationAnswer);
    DM_Field *field = DM_class_field(context, class, request->field_index);

    DM_SetCopyable(context, answer);
    answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
    answer->requesting = DM_TheCurrentSite();
    answer->requested  = request->requesting;
    answer->request    = (DM_Request*)request;
    answer->value      = 
      DM_indexed_patch_from_reified_value(context, 
                                          o, 
                                          field,
                                          request->index,
                                          request->value );
    DM_SetCopyable(context, answer->value);
    DM_SendAnswer(context, answer, request->requesting);

    DM_DebugAtLevel(0,
        DM_memorize_sent_object(context, (DM_Object*)answer);
);

  }
}

/* RESOURCE DM_handle_owner_change_request
 * The current site must become the former owner of the associated object.
 * FUTURE: Might be interesting to change the site and the clock altogether.
 */

void                                  /* BOGUS */
DM_handle_owner_change_request (DM_Context            *context,
                                DM_OwnerChangeRequest *request) {
  DM_ShortCutReference(context,  &(request->object));
  if ( DM_IsA(context, request->object, Exit) ) {
    /* The object has already been migrated, forward the request. */
    DM_DeclareNarrowedObject(exit, Exit, request->object);

    DM_SendRequest(context, request, exit->owner);
  } else {
    DM_Entry *entry = (DM_Entry*)(DM_ObjectToProxy(context, request->object));
    DM_Exit  *exit  = DM_Allocate0(context, Exit);
    DM_OwnerChangeAnswer *answer = DM_Allocate0(context, OwnerChangeAnswer);
    
    DM_SetCopyable(context, answer);
    exit->number  = request->entry_num;
    exit->counter = entry->counter;
    exit->object  = entry->object;
    exit->clock   = DM_NULL;
    exit->class   = DM_ObjectToClass(context, request->object);
    exit->time    = 0; 
    exit->from    = request->requesting;
    exit->owner   = exit->from;
    DM_remove_entry(context, (DM_External*)entry);
    DM_register_exit(context, exit);
    answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
    answer->requesting = DM_TheCurrentSite();
    answer->requested  = request->requesting;
    answer->request    = (DM_Request*)request;
    answer->object     = request->object;
    DM_SendAnswer(context, answer, answer->requesting);
  }
}

/* RESOURCE DM_handle_remote_owner_change_request
 * The current site must become the new owner of the associated object. 
 */

void                                        /* BOGUS */
DM_handle_remote_owner_change_request (DM_Context                  *context,
                                       DM_RemoteOwnerChangeRequest *request) {
  DM_ShortCutReference(context,  &(request->object));
  if ( DM_IsA(context, request->object, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, request->object);
    DM_OwnerChangeRequest *ocrequest = 
      DM_Allocate0(context, OwnerChangeRequest);
    DM_EntryWithDelayedRequests *entry = 
      DM_Allocate0(context, EntryWithDelayedRequests);

    DM_SetSharable(context, ocrequest);
    entry->counter    = 0;
    entry->object     = (DM_Object*)exit;
    entry->clock      = DM_NULL;
    entry->requests   = DM_NULL;
    DM_register_entry(context, (DM_External*)entry);
    ocrequest->options    = DM_REQUEST_DEFAULT_OPTIONS;
    ocrequest->requesting = DM_TheCurrentSite();
    ocrequest->requested  = exit->owner;
    ocrequest->object     = (DM_Object*)exit;
    ocrequest->entry_num  = entry->number;
    DM_externalize(context, (DM_Object*)ocrequest);
    DM_SendRequest(context, ocrequest, exit->owner);
  } else {
    /* The object is already owned by the current site. */
  }
}

/* RESOURCE DM_do_receive_request
 * Process a received Request, choose the appropriate method to answer it.
 */

static void 
DM_do_receive_request (DM_Context *context,
                       DM_Request *request) {
  DM_Class *request_class = DM_ObjectToClass(context, request);
  
  if ( DM_IsA(context, request_class, RequestMetaClass) ) {
    DM_DeclareNarrowedObject(rc, RequestMetaClass, request_class);
    DM_Field *f = DM_ReferenceToFieldOfClass(handler, RequestMetaClass);
    DM_displacement t = DM_compute_offset(context, (DM_Object*)rc, f);
    void *p = DM_Displace(rc, t);
    DM_handler_code hc = DM_Content(DM_handler_code, p);
    
    hc(context, request);
  } else {
    DM_Abort(context, DME_UNKNOWN_REQUEST);
  }
}

/* RESOURCE DM_report_failure
 * Interrrupt the current message if any and report a failure.
 */

static void
DM_report_failure (DM_Context *context,
                   DM_Object  *o, 
                   DM_Site    *site ) {
  DM_Connection *connection = DM_which_connection(context, site);
  DM_OutBuffer  *outbuffer;

  DM_DebugAtLevel(5,
    fprintf(DM_FindErrorStream(context), "Reporting failure.\n");
);

  if ( DM_IsA(context, connection, Connection) ) {
    /* There is a working Connection towards the site. */
    outbuffer = connection->emission;
    if ( outbuffer->mark ) {
      /* A message is already started, cancel it!  
       * First find the first buffer. */
      DM_Marker *marker = outbuffer->mark;

      while ( marker->next ) {
        marker = marker->next;
      }
      outbuffer = (DM_OutBuffer*)(marker->buffer);
      DM_cancel_message(context, outbuffer);
      outbuffer = DM_start_message(context, outbuffer, DM_FALSE);
      DM_print_byte(context, DM_RESET_STACK_COMMAND, outbuffer);
      DM_print_byte(context, DM_RESET_CACHE_COMMAND, outbuffer);
      DM_print_byte(context, DM_RECEIVE_ANSWER_COMMAND, outbuffer);
      DM_write_object(context, o, outbuffer);
      DM_finish_message(context, outbuffer, site);
    } else {
      DM_SendAnswer(context, o, site);
    }    
  } else {
    /* No Connection to reach this Site. */
    DM_Abort(context, DME_UNREACHABLE_SITE);
  }
}

/* RESOURCE DM_receive_request
 * Handle a request. This function is invoked on every received Request 
 * prefixed with the DM_RECEIVE_REQUEST command. 
 * If the Request is not addressed to the current site, forward it in the
 * right direction towards the requested site.
 */

void
DM_receive_request (DM_Context *context,
                    DM_Request *request) {

  DM_DebugAtLevel(7,
    DM_Class *c = DM_ObjectToClass(context, request);
    fprintf(DM_FindErrorStream(context), 
            "Processing %s " DM_0X "%p\n", 
            DM_predefined_name(context, (DM_Object*)c), request);
);

  /* Always increment the ticker */
  DM_TheCurrentSite()->ticker++;

  if ( request->requested != DM_TheCurrentSite() ) {
    /* This Request is not addressed to the current site, forward it
     * towards the requested site. Don't register it since we cannot
     * ensure that we will receive the Answer. */
    DM_SendRequest(context, request, request->requested);

  } else {
    /* Protect the execution of this request with a new context.
     * This will allow to trap and return failure codes in case of
     * problems. */
    volatile DM_ErrorCode errcode = DME_SUCCESS;

    DM_ProtectWithFreshContext(context, newcontext,
            DM_do_receive_request(newcontext, request),
       /* If something went wrong. */
       errcode = newcontext->errcode
                             );

    if (   (errcode != DME_SUCCESS)
        && DM_HasOption(request, DM_REQUEST_WANTS_ANSWER) ) {
      /* Return the failure code to the remote invoker.  Since the
       * preceding error might have occurred while writing a message,
       * there's a problem.  */
      DM_FailureAnswer *answer = DM_Allocate0(context, FailureAnswer);

      DM_SetCopyable(context, answer);
      answer->options    = DM_REQUEST_DEFAULT_OPTIONS;
      answer->requesting = DM_TheCurrentSite();
      answer->requested  = request->requesting;
      answer->request    = (DM_Request*)request;
      answer->errcode    = errcode;
      DM_report_failure(context, (DM_Object*)answer, request->requesting);
    }
  }
}

/* }}}
 * {{{ Answers

 * RESOURCE DM_handle_site_connection_answer
 * That answer carries the site object that was requested. Nearly all the 
 * work is done when decoding that Answer (using the DM_DEFINE_EMITTER_COMMAND),
 * it only suffices to store the requested site in the recorded Connection.
 */

void
DM_handle_site_connection_answer (DM_Context              *context,
                                  DM_SiteConnectionAnswer *answer) {
}

/* RESOURCE DM_handle_object_fetch_answer
 * Handle an answer to a request asking for an object. Depending on the 
 * state of the local exit, the object may be overwritten if its content
 * is newer.

 * NOTE: very ugly and expensive. Should overwrite the object to
 * refresh directly from DM_read_object with another command. But what
 * if answer must be forwarded ? 

 * If an uninitialized object becomes sharable (and therefore is
 * monitored by a clock) it must be possible to compare its old
 * exit->time with the new value of its clock. This is why
 * DM_increment_clock uses the ticker to reset its clock.
 */

void
DM_handle_object_fetch_answer (DM_Context           *context,
                               DM_ObjectFetchAnswer *answer) {
  DM_DeclareNarrowedObject(request, ObjectFetchRequest, answer->request);
  DM_Exit *exit;

  /* If the Request was displayed (in html), this reference may have
   * been shortcut. */
  if ( DM_IsA(context, request->object, Exit) ) {
    exit = (DM_Exit*)(request->object);
  } else {
    exit = (DM_Exit*)(DM_ObjectToProxy(context, request->object));
  }
  if ( exit->object ) {
    /* There is already a cached object associated to the Exit item.
     * Refresh it if the new copy is newer. */
    DM_Object *o = exit->object;

    exit->options = answer->object_options;
    DM_ClearOption(exit, DM_EXIT_IS_PREFETCHED);
    if ( answer->time > exit->time ) {
      /* Update the object! Cannot write simply:
             memcpy(o, answer->object, DM_size(context, o));
       * Since that would alter the local fields. */
      DM_update_object(context, o, answer->object);
      exit->time = answer->time;
    } else {
      /* The received copy is obsolete wrt the local cache: discard
       * it and keep the currently cached object. */
    }
  } else {
    /* The Exit item has no cached replica. */
    exit->options = answer->object_options;
    DM_ClearOption(exit, DM_EXIT_IS_PREFETCHED);
    exit->object = answer->object;
    exit->time   = answer->time;
    DM_SetProxyOf(context, answer->object, exit);
  }

  /* Even if the current site is not the target site, it takes benefit
   * of the Answer to update its local cache. After that it forwards
   * the Answer. */
  if ( answer->requested != DM_TheCurrentSite() ) {
    /* Forward the Answer to the requested site. */
    DM_SendAnswer(context, answer, answer->requested);
  }
}

/* RESOURCE DM_handle_object_mutation_answer
 * Nothing to do and do the same thing for 
 * DM_handle_object_indexed_mutation_answer.
 */

void
DM_handle_object_mutation_answer (DM_Context              *context,
                                  DM_ObjectMutationAnswer *answer) {
  if ( answer->requested != DM_TheCurrentSite() ) {
    /* Forward the Answer to the requested site. */
    DM_SendAnswer(context, answer, answer->requested);
  } else {
    /* NOTHING. */
  }
}

/* RESOURCE DM_handle_owner_change_answer
 * 
 */

void
DM_handle_owner_change_answer (DM_Context           *context,
                               DM_OwnerChangeAnswer *answer) {
  DM_DeclareNarrowedObject(request, OwnerChangeRequest, answer->request);

  request = request;
}

/* RESOURCE DM_handle_remote_owner_change_answer
 * Maybe change the local Exit to refer to the new owner ???
 */

void
DM_handle_remote_owner_change_answer (DM_Context                 *context,
                                      DM_RemoteOwnerChangeAnswer *answer) {
}

/* RESOURCE DM_handle_failure_answer
 * Propagate a remote failure code.
 */

void
DM_handle_failure_answer (DM_Context       *context,
                          DM_FailureAnswer *answer ) {
}

/* RESOURCE DM_receive_answer 
 * This function processes an Answer and handles it according to its
 * type. After being processed, answers may be forwarded to the
 * requested site: they are also processed since it is wasteful not to
 * profit of the information in the Answer. 
 */

void
DM_receive_answer (DM_Context *context,
                   DM_Answer  *answer) {

  DM_DebugAtLevel(7,
    DM_Class *c = DM_ObjectToClass(context, answer);
    fprintf(DM_FindErrorStream(context), 
            "Processing %s " DM_0X "%p\n", 
            DM_predefined_name(context, (DM_Object*)c), answer);
);

  /* Always increment the ticker */
  DM_TheCurrentSite()->ticker++;

  if ( answer->requested != DM_TheCurrentSite() ) {
    /* Forward this Answer. */
    DM_SendAnswer(context, answer, answer->requested);
  } else {
    DM_Class *answer_class = DM_ObjectToClass(context, answer);
  
    if ( DM_IsA(context, answer_class, RequestMetaClass) ) {
      DM_DeclareNarrowedObject(ac, RequestMetaClass, answer_class);
      DM_Field *f = DM_ReferenceToFieldOfClass(handler, RequestMetaClass);
      DM_displacement t = DM_compute_offset(context, (DM_Object*)ac, f);
      void *p = DM_Displace(ac, t);
      DM_handler_code hc = DM_Content(DM_handler_code, p);
    
      hc(context, answer);
    } else {
      DM_Abort(context, DME_UNEXPECTED_ANSWER);
    }
    /* In all cases, signal that the Answer has been handled.  In
     * fact, this is only useful for Answers that are waited for.
     * (Prefetch requests do not wait for their answers). */
    DM_signal_answer(context, answer);
  }
}

/* }}}
 * Initialize the module.
 */

void
DM_req_initialize (DM_Context *context)
{
  DM_LinkVersionStringIntoSite(dmreq);

  DM_DebugAtLevel(9,
    /* Allocate a BoundedStack to hold the most recent Requests/Answers.
    * This is useful when debugging (no extractor to read it more easily). */
    { DM_BoundedStack *bs = DM_Allocate1(context, BoundedStack, 2);

      bs->index   = 0;
      DM_lastreqans  = bs;
    }
);

}

/* end of dmreq.c */
