/* $Id: dmreqst.c,v 1.87 1998/05/30 20:30:10 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * This file defines the main functions dealing with remote objects.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmreqst,
  "$Id: dmreqst.c,v 1.87 1998/05/30 20:30:10 queinnec Exp $");

/* {{{ Utilities
 
 * RESOURCE DM_shortcut_reference
 * Shortcut a reference but only if possible that is if it references an
 * Exit item with a locally cached object.
 *
 *.FIGURE shortcut of a reference
 an Object                     
 +--------+                     an Exit
 | field: |                     \------/ <----------------------+
 |    ------------+------------> \  --------+                   |
 |        |       !               \  /      |                   |
 +--------+       !                \/       |                   |
                  !                         |                   |
                  ! shortcut                V another Object    |
                  !==============>      +--------+              |
                                        |proxy: ----------------+
                                        |        |
                                        |        |
                                        +--------+
 *.ENDFIGURE
 */

void
DM_shortcut_reference (DM_Context   *context,
                       DM_reference *location ) {
  if ( *location ) {
    DM_Object *o = *location;

    if ( DM_IsA(context, o, Exit) ) {
      DM_DeclareNarrowedObject(exit, Exit, o);

      if ( exit->object ) {
        /* The Exit item is locally cached, then shortcut the reference. */
        *location = exit->object;

        DM_DebugAtLevel(8,
          fprintf(DM_FindErrorStream(context), 
                  "Shortcut: substituting " DM_0X "%p by " DM_0X "%p\n", 
                  o, exit->object );
);

      } else {
        /* NOTHING: The Exit item is not locally cached. */
      }
    } else {
      /* NOTHING: Already a pointer onto a present object. */
    }
  } else {
    /* NOTHING: the location contains DM_NULL. */
  }
}

/* DOCUMENTATION DM_longcut_reference
 * The gc may perform DM_longcut_reference ie transform a pointer onto a
 * present object into a pointer onto an Exit item. This may also be done
 * when migrating objects. Care must be taken with DMeroon invariants.
 */

void
DM_longcut_reference (DM_Context   *context,
                      DM_reference *location ) {
  if ( *location ) {
    DM_Object *o = *location;

    if ( DM_IsA(context, o, Exit) ) {
      /* NOTHING: the target of the pointer is already an Exit item. */
    } else {
      /* The object o is present and is not an Exit item. */
      DM_Object *proxy = DM_ObjectToProxy(context, o);
    
      if ( DM_IsA(context, proxy, Exit) ) {
        /* o is a remote object then longcut the reference. */
        DM_DeclareNarrowedObject(exit, Exit, proxy);

        DM_DebugAtLevel(8,
          fprintf(DM_FindErrorStream(context), 
                  "Longcut: substituting " DM_0X "%p by " DM_0X "%p\n", 
                  o, exit );
);

        *location = exit;
      } else {
        /* NOTHING: Not a remote object. */
      }
    }
  } else {
    /* NOTHING: the location contains DM_NULL */
  }
}

/* RESOURCE DM_ensure_class_usable
 * Ensure that a Class may be used to decode an object. That means that
 * the class, all its supers and all its fields must be present. A special
 * bit DM_CLASS_IS_USABLE in the class_options records this state. Whenever
 * a new replica of that class (or of one of its field) the usability bit
 * should be reset.
 *
 * This function may block (hence its "ensure_..." name).
 */

void
DM_ensure_class_usable (DM_Context *context,
                        DM_Class   *class ) {

  if ( DM_HasOption(class, DM_CLASS_IS_USABLE) ) {
    /* NOTHING: the Class is already usable. */
  } else {
    DM_nat3 nfields = DM_class_indexed_field_length(context, class);
    DM_nat3 i;

    for ( i=0 ; i<class->super_length-1 ; i++ ) {
      DM_EnsurePresent(context, &(class->super[i]));
      DM_ensure_class_usable(context, class->super[i]);
    }
    class->super[class->super_length-1] = class;
    for ( i=0 ; i<nfields ; i++ ) {
      DM_Field *field;
      DM_Field **field_address = DM_class_field_address(context, class, i);

      DM_EnsurePresent(context, field_address);
      field = *field_address;
      DM_EnsurePresent(context, &(field->introducing));
      DM_EnsurePresent(context, &(field->referring));
    }
    DM_compute_layout(context, class);
    /* Mark that the class is usable now (for allocation, field extraction). */
    DM_SetOption(class, DM_CLASS_IS_USABLE);
  }
}

/* RESOURCE DM_ensure_present
 * Given the address of a DMeroon reference, ensure that the referenced 
 * object is present (ie local or locally cached). Caution: present does
 * not mean up to date! 
 *
 * This primitive may block as attested by its "ensure_..." name.
 */

void
DM_ensure_present (DM_Context *context,
                   DM_Object* *location) {
  DM_Object* o = *location;

  if ( o ) {
    if ( DM_IsA(context, o, Exit) ) {
      /* The object is a remote object. */
      DM_DeclareNarrowedObject(exit, Exit, o);
    
      if ( exit->object ) {
        /* The object is already present, shortcut the reference. */
        *location = exit->object;
      } else {
        /* The object is not locally cached. */
        DM_EnsurePresent(context, &(exit->class));
        DM_ensure_class_usable(context, exit->class);
        DM_EnsurePresent(context, &(exit->clock));
        /* Wait for the object to be fetched and cached locally. */
        DM_send_object_fetch_request(context, location, exit);
      }
    } else {
      /* NOTHING: The referenced object is already present. */
    }
  } else {
    /* NOTHING: the reference is NULL. */
  }
}

/* RESOURCE DM_ensure_site_usable
 * Ensure that the content of a Site instance is present. 
 * When allocated (with DM_SITE_COMMAND) a site only contains its ip, port
 * date and sitenamelength fields. Normally sites propagate sites and 
 * associate them with Exit items then fills them. Meanwhile, they may
 * be in an intermediate state.
 *
 * This primitive may block.
 */

void
DM_ensure_site_usable (DM_Context *context,
                       DM_Site    *site)
{
  if ( DM_HasOption(site, DM_SITE_IS_USABLE) ) {
    /* NOTHING: The content of the site is already there. */
  } else {
    /* If the site is not usable, then it has no associated exit item.
     * Ask for its content from its route.  */
    DM_ObjectFetchRequest *request = DM_Allocate0(context, ObjectFetchRequest);

    request->options = DM_REQUEST_DEFAULT_OPTIONS;
    DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
    request->requesting = DM_TheCurrentSite();
    if ( site->route ) {
      request->requested = site->route;
      request->object    = (DM_Object*)site;
      /* Encode the Request in the appropriate OutBuffer */
      DM_SendRequest(context, request, site->route);
      {
        DM_Answer *answer;
        /* Wait for the Answer to come back. */
        DM_wait_until_answer(context,
                             DM_TheCurrentSite()->abnormal_duration,
                             &answer,
                             (DM_Request*)request);
      }
      /* Normally here, the site is present and completely filled. */
      DM_SetOption(site, DM_SITE_IS_USABLE);
    } else {
      /* No route then open a direct connection. */
      DM_reconnect(context, site);
      DM_Abort(context, DME_NOT_YET_IMPLEMENTED);
    }
  }
}

/* RESOURCE DM_reconnect 
 * Open or reopen a connection to a given site.
 * Return the connection if success, DM_NULL in case of failure.
 * If the connection is successful, then the site handshake will have
 * refilled the site object.
 */

DM_Connection*
DM_reconnect (DM_Context* context,
              DM_Site*    site ) 
{
  DM_nat3 port_number = ntohs(site->port);
  char*   hostname    = site->name;
  DM_Connection* connection = site->connection;

  if ( connection ) {
    /* already a connection. */
    if ( DM_HasOption(connection, DM_PORT_IS_CLOSED) ) {
      /* A Connection once existed but is now closed, reopen it. */
      site->connection = DM_NULL;
      goto open_new_connection;
    } else {
      return connection;
    }
  } else {
    /* No connection then open one. */
  open_new_connection: {
    char *pw = NULL;

    if ( site->password ) {
      pw = site->password->character;
    }
    connection = DM_create_connection(context, hostname, port_number, pw);
    if ( connection ) {
      if ( DM_site_handshake(context, connection) == DME_SUCCESS ) {
        return connection;
      } else {
        /* A connection was open, but the handshake of sites went wrong. */
        (void) DM_finish_port(context, (DM_Port*)connection);
        site->connection = DM_NULL;
        return DM_NULL;
      }
    } else {
      /* Not possible to connect to that host:port, abandon! */
      return DM_NULL;
    }
  }
  }
}

/* RESOURCE DM_disconnect
 * Close a direct connection to a site. 
 */

void
DM_disconnect (DM_Context* context,
               DM_Site*    site ) 
{
  DM_Connection *connection = site->connection;
  
  if ( connection ) {
    /* If a Connection exists then the Site object is known. */
    if ( DM_finish_port(context, (DM_Port*)connection) ) {
      site->connection = DM_NULL;
    } else {
      DM_Abort(context, DME_FAILED_CLOSE);
    }
  } else {
    /* No connection to close (or the site is unknown). */
  }
}

/* RESOURCE DM_ensure_up_to_date
 * Ensure that a remote object locally cached is uptodate with respect to 
 * the current local perception of the value of its associated (if mutable
 * of course). 
 *
 * This primitive may block as attested by its "ensure_..." name.
 */

void
DM_ensure_up_to_date (DM_Context *context,
                      DM_Object  *o ) {
  DM_Object *proxy = DM_ObjectToProxy(context, o);

  if ( DM_IsA(context, proxy, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, proxy);
    DM_Clock  *clock = exit->clock;
    
    if ( DM_IsA(context, clock, Clock)  ) {
      if ( clock->time != exit->time ) {
        DM_send_object_fetch_request(context, &o, exit);
      } else {
        /* NOTHING: the object is up to date. */
      }
    } else {
      /* NOTHING: The object has no associate clock, it is immutable. */
    }
  } else {
    /* NOTHING: The object is not remote, so it is always up to date. */
  }
}

/* RESOURCE DM_is_class_usable
 * Check whether a class is usable without blocking. Try to prefetch missing
 * parts of the class. This primitive may set the DM_CLASS_IS_USABLE option
 * if the class is found to be usable. 
 */

DM_c_boolean
DM_is_class_usable (DM_Context *context,
                    DM_Class   *class) 
{
  if ( DM_HasOption(class, DM_CLASS_IS_USABLE) ) {
    return DM_TRUE;
  } else {
    DM_nat3 nfields = DM_class_indexed_field_length(context, class);
    DM_nat1 usable = 1;
    DM_nat3 i;
    
    /* Check all super-classes, they must be present. */
    for ( i=0 ; i<class->super_length-1 ; i++ ) {
      DM_ShortCutReference(context, &(class->super[i]));
      if ( DM_IsA(context, class->super[i], Class) ) {
        if ( ! DM_is_class_usable(context, class->super[i]) ) {
          usable = 0;
        }
      } else {
        usable = 0;
        if ( DM_IsA(context, class->super[i], Exit) ) {
          DM_TryPrefetchAddress(context, &(class->super[i]));
        } else {
          /* NULL here means that we are currently unmarshaling the
           * class. This is too early to check whether it is usable. */
        }
      }
    }
    class->super[class->super_length-1] = class;

    /* Check all fields, they must be present. */
    for ( i=0 ; i<nfields ; i++ ) {
      DM_Field **field_address = DM_class_field_address(context, class, i);
      DM_Field *field;

      DM_ShortCutReference(context, field_address);
      field = *field_address;
      if ( DM_IsA(context, field, Field) ) {
        /* Shortcut reference fields within Fields (introducing,
         * referring) and test if the expected classes are present. */
        DM_ShortCutReference(context, &(field->introducing));
        if ( DM_IsA(context, field->introducing, Class) ) {
          /* Nothing to do. */
        } else {
          usable = 0;
          if ( DM_IsA(context, field->introducing, Exit) ) {
            DM_TryPrefetchAddress(context, &(field->introducing));
          }
        }
        /* The 'referring' field of a field is meaningful only if the
         * field has type reference. 
         * FUTURE: would be simpler to test if referring is DM_NULL if
         * not needed.                         FUTURE */
        {
          DM_offset last_offset  = field->offset[field->offset_length-1];
          DM_nat1   itemtypindex = DM_OffsetToItemTypeIndex(last_offset);

          if ( itemtypindex == DM_reference_index ) {
            DM_ShortCutReference(context, &(field->referring));
            if ( DM_IsA(context, field->referring, Class) ) {
              /* Nothing to do. */
            } else {
              usable = 0;
              if ( DM_IsA(context, field->referring, Exit) ) {
                DM_TryPrefetchAddress(context, &(field->referring));
              } else {
              }
            }
          }
        }
      } else {
        usable = 0;
        if ( DM_IsA(context, field, Exit) ) {
          DM_TryPrefetchAddress(context, field_address);
        } else {
          /* NULL means that we are currently unmarshaling the
           * class. This is too early to check whether it is usable. */
        }
      }
    }
    if ( usable ) {
      /* All necessary fields are present, the class is usable. */
      DM_compute_layout(context, class);
      /* Mark that the class is usable now. */
      DM_SetOption(class, DM_CLASS_IS_USABLE);

      DM_DebugAtLevel(5,
        fprintf(DM_FindErrorStream(context), 
                "Class " DM_0X "%p is usable now.\n", class);
);

      return DM_TRUE;
    } else {
      return DM_FALSE;
    }
  }
}

/* RESOURCE DM_site_handshake
 * This the DMeroon handshake protocol between two sites. They exchange 
 * their Site instances to initialize the communication. This is mandatory
 * when a Connection is opened to verify the handshaking sites.
 *
 * This function is answered by DM_handle_site_connection_request.
 */

DM_ErrorCode
DM_site_handshake (DM_Context    *context,
                   DM_Connection *connection) {
  DM_Site      *site = DM_TheCurrentSite();
  DM_SiteConnectionRequest *request = 
    DM_Allocate0(context, SiteConnectionRequest);
  DM_OutBuffer *outbuffer       = connection->emission;
  DM_Class     *class           = DM_ReferenceToClass(SiteConnectionRequest);
  DM_nat3      site_ip          = ntohl(site->address);
  DM_nat2      site_portnum     = ntohs(site->port);
  DM_date      site_date        = site->date;
  DM_Entry     *site_entry;

  request->options    = DM_REQUEST_DEFAULT_OPTIONS;
  DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
  request->requesting = site;
  /* The requested site is unknown since the goal of that Request is
   * to obtain a remote reference onto it. */
  request->requested  = DM_NULL;
  /* Toggle the Connection into DMeroon inner protocol.
   * Don't forget to send the password. */
  DM_print_char_star(context, "DMeroon", 0, outbuffer);
  if ( connection->password ) {
    DM_print_string(context, connection->password, outbuffer);
  } else {
    DM_print_char_star(context, DMEROON_ADDITIONAL_PASSWORD, 0, outbuffer);
  }
  DM_print_char_star(context, " " DM_RELEASE_LEVEL "\r\n", 0, outbuffer);
  /* Register and encode the Request. This is very special since we
   * are bootstrapping the Connection. The encoding is done by hand
   * (we do not use DM_send_request) but it corresponds to a
   * mechanically decodable encoding.
   */
  DM_SetSharable(context, request);
  DM_externalize(context, (DM_Object*)request);
  /* Reset local cache, the DMeroon protocol will reset the cache at
   * the other end. */
  DM_reset_stack(context, outbuffer->stack);
  DM_reset_cache(context, outbuffer->cache);
  outbuffer = DM_start_message(context, outbuffer, DM_FALSE);
  /* Transmits a copy of the SiteConnectionRequest instance encoded by hand. */
  DM_print_byte(context, DM_RECEIVE_REQUEST_COMMAND, outbuffer);
  DM_print_byte(context, DM_BIND_COMMAND, outbuffer);
  DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
  /* Allocate a SiteConnectionRequest */
  DM_print_byte(context, DM_ALLOCATE_COMMAND, outbuffer);
  DM_print_byte(context, DM_PREDEFINED_CLASS_COMMAND, outbuffer);
  DM_print_byte(context, DM_PredefinedClassNumber(class), outbuffer);
  /* Transmit request->options. */
  DM_send_nat3(context, outbuffer, &(request->options));
  /* Transmit request->requesting. This is the current site ie a
  * complex object to encode while bootstrapping. */
  DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
  DM_print_byte(context, DM_DEFINE_EMITTER_COMMAND, outbuffer);
  DM_print_byte(context, DM_SITE_COMMAND, outbuffer);
  /* Send the port which is served (not the one used by the Connection)
   * since this is used to hash the site on the receiving site. */
  DM_send_nat3(context, outbuffer, &site_ip);
  DM_send_nat2(context, outbuffer, &site_portnum);
  DM_send_date(context, outbuffer, &site_date);
  /* Send the size of the name of the Site. */
  DM_send_nat1(context, outbuffer, &(site->name_length));
  /* Send the key of the current site. */
  site_entry = DM_externalize(context, (DM_Object*)site);
  DM_increment_counter(context, (DM_External*)site_entry);
  DM_send_key(context, outbuffer, &(site_entry->number));
  /* And send the content of the site. */
  DM_write_object_content(context, (DM_Object*)site, outbuffer);
  /* Transmit request->requested. We don't know it so we emit a
   * command to find it on the other end. This should be easy since
   * this can only be the receiving site itself. */
  DM_print_byte(context, DM_RECEIVING_SITE_COMMAND, outbuffer);
  /* Send now the Exit item that corresponds to the request. */
  DM_write_object_reference(context, (DM_Object*)request, outbuffer);
  DM_finish_message(context, outbuffer, request->requested);

  DM_DebugAtLevel(0,
    DM_memorize_sent_object(context, (DM_Object*)request);
);

  /* Wait for the Answer to come back, if it does not come after some
   * time then close the Connection. */
  { volatile DM_Answer *a;

    DM_ProtectWithFreshContext(context, newcontext,
           (void) DM_wait_until_answer(newcontext, 
                                       DM_TheCurrentSite()->abnormal_duration,
                                       (DM_Answer**)&a,
                                       (DM_Request*)request),
       /* If something went wrong. */
       a = DM_NULL
                             );

    if ( a ) {
      return DME_SUCCESS;
    } else {

      DM_DebugAtLevel(5,
        fprintf(DM_FindErrorStream(context), 
                "No site handshake, abandon!\n");
);

      /* Connection inherits from Port. */
      DM_finish_port(context, (DM_Port*)connection);
      return DME_UNHANDSHAKABLE_SITE;
    }
  }
}

/* RESOURCE DM_which_connection
 * Given a site, return the Connection leading (more or less directly) to it
 * or DM_NULL if the site is unreachable. When a Connection is open towards
 * a Site, it appears in the 'connection' field of the Site. Otherwise use
 * the 'route' field of the site. This 'route' site will forward the message.
 */

DM_Connection*
DM_which_connection (DM_Context *context,
                     DM_Site    *site) 
{
  DM_Connection     *connection;

  if ( DM_IsA(context, site->connection, Connection) ) {
    connection = site->connection;
    if ( DM_HasOption(connection, DM_PORT_IS_CLOSED) ) {
      /* A Connection once existed but is now closed. Use the route. */
      site->connection = DM_NULL;
      goto use_route;
    } else {
      /* The Connection is open, keeps it. May be check that there
       * is still a peer ? But no need to test if it is dried since that
       * limits reading the Connection, not writing to it. */
      return connection;
    }
  } else {
  use_route:
    /* There is no Connection currently open towards the Site. Use the
     * route field if it exists. Don't cycle! */
    if ( (site->route) && (site->route != site) ) {

      DM_DebugAtLevel(6,
        fprintf(DM_FindErrorStream(context), 
                "Forwarding message to " DM_0X "%p via " DM_0X "%p\n", 
                site, site->route );
);

      return DM_which_connection(context, site->route);
    } else {
      /* There is no route so open a direct connection. */
      return DM_reconnect(context, site);
    }
  }
}

/* }}}
 * {{{ Send/receive objects

 * RESOURCE DM_send_object_fetch_request
 * Does the real work of fetching a remote object whose exit item is
 * present.  If the location is not DM_NULL then it acts as a blocking
 * primitive that waits for the Answer to fill the location, otherwise
 * it is a non-blocking primitive (used to prefetch objects). The
 * fetched object is returned in location.
 *
 * NOTE: Until the Connection is initialized (and SiteConnectionRequest or
 * SiteConnectionAnswer are exchanged), no request is sent.
 */

void
DM_send_object_fetch_request (DM_Context   *context,
                              DM_Object*   *location,
                              DM_Exit      *exit ) {
  DM_ObjectFetchRequest *request = DM_Allocate0(context, ObjectFetchRequest);

  request->options    = DM_REQUEST_DEFAULT_OPTIONS;
  if ( location ) {
    DM_SetOption(request, DM_REQUEST_WANTS_ANSWER);
  }
  request->requesting = DM_TheCurrentSite();
  request->requested  = exit->owner;
  /* Create a shortcutable reference. */
  request->object       = (DM_Object*)exit;
  /* Encode the Request in the appropriate OutBuffer */
  DM_SendRequest(context, request, exit->owner);
  if ( location ) {
    /* Wait for the Answer to come back. */
    DM_Answer *answer;

    DM_wait_until_answer(context, 
                         DM_TheCurrentSite()->abnormal_duration,
                         &answer,
                         (DM_Request*)request);
    /* Normally here, the object is present unless a timeout occurs (but
     * at that time, abort was called and control is no longer
     * here. This may also depend on the protocol for object migration
     * when e->owner is no longer the owner.  FUTURE  */

    DM_DebugAtLevel(2,
      if ( ! exit->object ) {
        DM_Abort(context, DME_INTERNAL_ERROR);
      }
);

    *location = exit->object;
  } else {
    /* NOTHING: Don't wait for the answer. */
  }
}

/* RESOURCE DM_try_prefetch_exit
 * Try if possible to prefetch an Object. Send an object fetch request
 * but do not wait at all. Could be run as a dead-end thread.
 */

void
DM_try_prefetch_exit (DM_Context *context,
                      DM_Exit    *exit ) 
{
  if ( DM_HasNotOption(exit, DM_EXIT_IS_PREFETCHED) ) {
    DM_try_prefetch_address(context, (DM_Object**)(&(exit->class)));
    DM_try_prefetch_address(context, (DM_Object**)(&(exit->clock)));
    if (    DM_IsA(context, exit->class, Class) 
         && DM_is_class_usable(context, exit->class) ) {
      DM_SetOption(exit, DM_EXIT_IS_PREFETCHED);
      DM_DebugAtLevel(17,
        fprintf(DM_FindErrorStream(context), 
                "Prefetching Exit " DM_0X "%p\n", exit);
);
      DM_send_object_fetch_request(context, DM_NULL, exit);
    }
  }
}

/* RESOURCE DM_try_prefetch_address
 * Try if possible to prefetch the content of a location. Send an object 
 * fetch request but do not wait at all. Could be run as a dead-end thread.
 */

void
DM_try_prefetch_address (DM_Context *context,
                         DM_Object* *address ) 
{
  DM_Object *o;

  DM_ShortCutReference(context, address);
  o = *address;
  if ( DM_IsA(context, o, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, o);

    DM_try_prefetch_exit(context, exit);
  } else {
    /* NOTHING: o is or NULL or a present object. */
  }
}                       

/* RESOURCE DM_send_object_send_request
 * Send the reference of an object to a remote site. No answer is expected.

 * NOTE: This is a Request and not a simple use of DM_RECEIVE_OBJECT
 * (as before) since such a Request may be forwarded from site to
 * site. Could be better encoded if the requested is the receiver
 * site or maybe decide to emit an answer ?                           FUTURE
 */

void
DM_send_object_send_request (DM_Context *context, 
                             DM_Object  *o, 
                             DM_Site    *site) {
  DM_ObjectSendRequest *request = DM_Allocate0(context, ObjectSendRequest);
  
  request->options    = DM_REQUEST_DEFAULT_OPTIONS;
  request->requesting = DM_TheCurrentSite();
  request->requested  = site;
  request->object     = o;
  DM_SendRequest(context, request, site);
}

/* RESOURCE DM_send_request
 * Encode a Request. A Request is a shared object so we transmit its content 
 * as well as its key. 
 */

DM_ErrorCode
DM_send_request (DM_Context *context,
                 DM_Object  *o,
                 DM_Site    *site ) {
  DM_Connection *connection = DM_which_connection(context, site);
  DM_OutBuffer  *outbuffer;

  DM_DebugAtLevel(10,
     DM_Class *c = DM_ObjectToClass(context, o);
     fprintf(DM_FindErrorStream(context), 
             "Sending %s " DM_0X "%p\n",
             DM_predefined_name(context, (DM_Object*)c), o);
);

  if ( DM_IsA(context, connection, Connection) ) {
    /* There is a working Connection towards the site. */
    outbuffer = connection->emission;
    /* Externalize the request as sharable. */
    DM_SetSharable(context, o);
    DM_externalize(context, (DM_Object*)o);
    /* Tells DM_start_message that it should accompany the message
     * with the information that must be propagated. */
    outbuffer = DM_start_message(context, outbuffer, DM_TRUE);
    DM_print_byte(context, DM_RECEIVE_REQUEST_COMMAND, outbuffer);
    DM_print_byte(context, DM_BIND_COMMAND, outbuffer);
    DM_write_object(context, o, outbuffer);
    DM_write_object_reference(context, o, outbuffer);
    DM_finish_message(context, outbuffer, site);
    return DME_SUCCESS;
  } else {
    /* No Connection to reach this Site. */
    DM_Abort(context, DME_UNREACHABLE_SITE);
    return DME_UNREACHABLE_SITE;
  }
}

/* RESOURCE DM_send_answer
 * Encode an Answer. An answer is a copyable object so we simply
 * transmit its content. An answer has a field that designates the
 * Request to which it is the answer, this field must be preserved of
 * course so the receiving site may pair the Answer with its
 * associated Request. Normally, when the Answer is reclaimed, a
 * decrement will be to the Request. 
 */

DM_ErrorCode
DM_send_answer (DM_Context *context,
                DM_Object  *o,
                DM_Site    *site ) {
  DM_Connection *connection = DM_which_connection(context, site);
  DM_OutBuffer  *outbuffer;

  DM_DebugAtLevel(10,
     DM_Class *c = DM_ObjectToClass(context, o);
     fprintf(DM_FindErrorStream(context), 
             "Sending %s " DM_0X "%p\n",
             DM_predefined_name(context, (DM_Object*)c), o);
);

  if ( DM_IsA(context, connection, Connection) ) {
    /* There is a working Connection towards the site. */
    outbuffer = connection->emission;
    /* Make the Answer copyable. */
    DM_SetCopyable(context, o);
    /* Tells DM_start_message that it should accompany the message
     * with the information that must be propagated. */
    outbuffer = DM_start_message(context, outbuffer, DM_TRUE);
    DM_print_byte(context, DM_RECEIVE_ANSWER_COMMAND, outbuffer);
    DM_write_object(context, o, outbuffer);
    DM_finish_message(context, outbuffer, site);
    return DME_SUCCESS;
  } else {
    /* No Connection to reach this Site. */
    DM_Abort(context, DME_UNREACHABLE_SITE);
    return DME_UNREACHABLE_SITE;
  }
}

/* }}}
 * {{{ Initialization
 */

void
DM_reqst_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmreqst);
}

/* }}}
 * end of dmreqst.c */
