/* $Id: dmsched.c,v 1.51 1998/11/27 18:05:42 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmsched.c
 * This file contains the DMeroon event loop. DMeroon is operated as
 * if in its own thread. That is, when using DMeroon API, either the
 * operation is immediate, either the current thread is paused, a
 * Request is built and handled by the DMeroon thread. When an Answer
 * comes back, the thread will be resumed with the requested value.

 * The most important function is 
 *   DM_work_once
 *     		It scans all fd (files descriptors), record their state,
 * 		and runs the appropriate actions if a port is ready and
 *              is a DMeroon port. (It may also signal ready fd that are
 *              watched for). DMeroon ports are handled with:
 *       DM_analyze_all_ports
 *         	which in fact calls the appropriate function:
 *         DM_analyze_readable_port
 *         DM_analyze_writable_port

 * DM_work_once is very low level, there are some interesting
 * functions to use it:
 *      DM_work_at_most
 *      DM_work_at_least
 
 * To be independent of the thread system, this file should only be used
 * through the DM_wait_* or DM_signal_* functions. That is,
 *      DM_wait_until_object_reception
 *          	when waiting on DM_receive
 *	DM_wait_until_answer
 *		when waiting for an answer of a request
 *	DM_wait_until_event
 *   		waits for the next something
 *      DM_wait_until_duration
 *              waits for at least duration
 *      DM_wait_until_fd_ready
 *              when waiting for a file descriptor to become ready.
 * The end of these waits is signaled by the according signal functions ie
 *      DM_signal_answer
 *      DM_signal_object_reception
 *      DM_signal_event
 *      DM_signal_end_of_duration 
 *      DM_signal_fd_ready
 * Normally, signal functions never return.
 */

#include "dmeroon.h"

DM_DefineVersionString(dmsched,
  "$Id: dmsched.c,v 1.51 1998/11/27 18:05:42 queinnec Exp $");

/* {{{ Port functions

 * RESOURCE DM_analyze_readable_port
 * Analyze a Port to check whether there is some associated work that
 * can be performed.  Returns DM_TRUE if something interesting was
 * performed. This function returns as soon as one interesting thing
 * was performed, other works may still be do-able.
 */

static DM_c_boolean
DM_analyze_readable_port (DM_Context *context,
                          DM_Port    *port) {
  DM_Connection *connection;
  DM_OutBuffer  *outbuffer;
  DM_InBuffer   *inbuffer;

  /* SECTION 1: Take benefit of a Connection that is dried ie a
   * Connection that is to be closed soon. To be dried means that it
   * can no longer be filled but (1) it may still contain bytes to be
   * read and handled and (2) it may still be written on and, of
   * course, must be flushed. */

  if (   DM_HasOption(port, DM_PORT_TO_READ)
      && DM_HasOption(port, DM_PORT_IS_DRIED) ) {

    if ( DM_IsA(context, port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, port);

      inbuffer   = connection->reception;
      if ( inbuffer->maximum > inbuffer->index ) {
        /* There are some characters that remain to be handled. */
        if (   ( inbuffer->maximum == (inbuffer->index + 1) ) 
            && DM_HasOption(port, DM_PORT_RUNS_DMEROON) ) {
          /* An LF remains unread, this is normal in DMeroon inner
             protocol. Just swallow it. */
          inbuffer->index++;
          return DM_TRUE;
        } else {
          /* Try to handle all the remaining bytes.  Don't care of the
           * result, imperatively empty the InBuffer.  */
          DM_handle_readable_connection(context, connection);
          inbuffer->index = inbuffer->maximum;
          return DM_TRUE;
        }
      } else {
        /* No remaining characters in the InBuffer. */
        outbuffer  = connection->emission;
        /* Check whether the last OutBuffer is empty. Don't close
         * the port if it is not the case. */
        if (   (connection->emission == connection->emitqueue)
            && (outbuffer->maximum == outbuffer->index) ) {
          /* The Connection is dried, does not contain any more bytes
           * to read or write, so close it and unlink it from the
           * site that used it. */
          if ( DM_finish_port(context, port) ) {
            if ( connection->site ) {
              connection->site->connection = DM_NULL;
            }
            return DM_TRUE;
          } else {
            /* Could not close the port ??? */
            return DM_FALSE;
          }
        } else {
          /* The last OutBuffer is not empty. Wait for the Port to
           * become writable again so these buffers can be flushed. */
          return DM_FALSE;
        }
      }
    }

#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
    /* An InputPort is used to read from a file in Scheme. */
    else if ( DM_IsA(context, port, InputPort) ) {
      return DM_finish_port(context, port);
    }
#endif /* defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT */

    else {
      DM_MayAbort(context, DME_BAD_PORT);
      return DM_FALSE;
    }
  } /* We're now done with dried incoming ports. */
        
  /* SECTION 2: Take benefit of a Connection that is readable.
   * It cannot be dried here. */

  if (   DM_HasOption(port, DM_PORT_IS_READABLE)
      && DM_HasOption(port, DM_PORT_TO_READ) ) {
    DM_ClearOption(port, DM_PORT_IS_READABLE);
    
    /* Generic dispatch. */

    if ( DM_IsA(context, port, Server) ) {
      DM_DeclareNarrowedObject(server, Server, port);

      connection = DM_accept_connection(context, server);
      if ( connection ) {
        return DM_TRUE;
      } else {
        return DM_FALSE;
      }
    }

    else if ( DM_IsA(context, port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, port);

      /* Try to fill the input buffer. */
      if ( DM_fill_buffer(context, connection->reception) ) {
        return DM_handle_readable_connection(context, connection);
      } else {
        /* Something happened while filling the buffer. What ??? */
        return DM_TRUE;
      }
    }

#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
    else if ( DM_IsA(context, port, BootPort) ) {
      DM_DeclareNarrowedObject(bootport, BootPort, port);
      DM_Site *site = DM_TheCurrentSite();

      inbuffer  = (DM_InBuffer*)(bootport->buffer);
      /* Try to fill the BootPort buffer. */
      DM_fill_buffer(context, inbuffer);
      /* Handle the BootPort as if a Connection, handle all
       * present messages. */
      while ( (DM_handle_protocol(context, inbuffer, site->protocol))
             == DM_TRUE ) {
      }
      return DM_TRUE;
    } 
#endif /* defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT */
       
    else {
      DM_MayAbort(context, DME_BAD_PORT);
      return DM_FALSE;
    }
  } /* We're now done with readable incoming port. */
  
  /* SECTION 3: Nothing performable. */
  return DM_FALSE;
}

/* RESOURCE DM_analyze_writable_port
 * Analyze a Port to check whether there is some associated work that
 * can be performed.  Returns DM_TRUE if something interesting was
 * performed. This function returns as soon as one interesting thing
 * was performed, other works may still be do-able.
 */

static DM_c_boolean
DM_analyze_writable_port (DM_Context *context,
                          DM_Port    *port) {
  DM_c_boolean  r = DM_FALSE;

  /* SECTION 1: Take benefit of a Connection that is writable. Try to
   * flush waiting OutBuffers, keep Connection alive, etc. */

  if (   DM_HasOption(port, DM_PORT_IS_WRITABLE)
      && DM_HasOption(port, DM_PORT_TO_WRITE) ) {
    if ( DM_HasOption(port, DM_PORT_IS_PEERLESS) ) {
      /* Port is peerless, useless to flush it: skip to next section. */
    } else {
      DM_ClearOption(port, DM_PORT_IS_WRITABLE);

      /* Analyse the class of the port to deduce the appropriate
       * action to perform on it. This is a generic dispatch. */

      /* Try to flush the output buffer or to ping the other end. */
      if ( DM_IsA(context, port, Connection) ) {
        DM_DeclareNarrowedObject(connection, Connection, port);

        r = DM_handle_writable_connection(context, connection);
        if ( r ) {
          /* Something happened so return immediately. */
          return r;
        } else {
          /* NOTHING: Skip to next section. */
        }
      } 

#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
      /* Try to flush an OutputPort. This is used to write onto a file. */
      else if ( DM_IsA(context, port, OutputPort) ) {
        DM_DeclareNarrowedObject(outputport, OutputPort, port);
        DM_OutBuffer *outbuffer = (DM_OutBuffer*)(outputport->buffer);

        DM_flush_buffer(context, outbuffer);
        return DM_TRUE;
      }
#endif /* defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT */

      else {
        DM_MayAbort(context, DME_BAD_PORT);
        return DM_FALSE;
      }
    }
  } /* We're now done with outgoing writable port. */
  
  /* SECTION 2: Nothing performable. */
  return DM_FALSE;
}

/* DOCUMENTATION DM_analyze_all_ports
 * Examine every port in turn starting from the last one that was
 * previously examined (its index is recorded in
 * DM_last_selected_port). If one port has some work to perform,
 * perform that work and return DM_TRUE. If all ports are workless
 * then return DM_FALSE.
 *
 * DM_analyze_all_ports(0) means: continue to scan ports from the last
 * scanned one. DM_analyze_all_ports(DM_max_port_number) means: restart
 * scanning from Port at position 0.
 */

static volatile unsigned DM_last_selected_port = FD_SETSIZE;

static DM_c_boolean 
DM_analyze_all_ports (DM_Context *context,
                      DM_nat3    delta) {
  DM_ports_vector *active_ports = 
    (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));
  DM_Port      *port;
  DM_c_boolean result = DM_FALSE;
  DM_c_boolean r;
  /* Copy DM_max_port_number since it may evolve as Connections arrive. */
  DM_nat3 max = DM_max_port_number;
  DM_nat3 i;

  DM_last_selected_port += delta;
  if ( DM_last_selected_port > max ) {
    DM_last_selected_port = 0;
    /* This means that we are right after a select(2) invokation.  We
     * then try to flush all writable Ports. */
    for ( i=0 ; i<=max ; i++ ) {
      port = active_ports->port[i];
      if ( DM_IsA(context, port, Port) ) {
        r = DM_analyze_writable_port(context, port);
        if ( r ) { 
          result = r;
          return DM_TRUE;
        }
      }
    }
  }
  
  /* Second: Try to read any readable Port and return as soon as
   * something has been or may be done. */
  while ( DM_last_selected_port <= max ) {
    port = active_ports->port[DM_last_selected_port];
    DM_last_selected_port++;
    if ( DM_IsA(context, port, Port) ) {
      if ( DM_analyze_readable_port(context, port) ) {
        /* exit from DM_analyze_all_ports since something happened. */
        return DM_TRUE;
      } else {
        continue;
      }
    } else {
      /* This is not a Port but the fd was looked for.
       * See dmports.c file (for Bigloo) for example. */
      i = DM_last_selected_port - 1;
      if ( FD_ISSET(i, &DM_last_ports_to_read) ) {
        DM_signal_fd_ready (context, i);
        return DM_TRUE;
      }
    }
  }
  /* Nothing happened on readable ports (but something might have
   * happened on writable ports. We reset DM_last_selected_port to
   * force flushing the writable ports again. */
  DM_last_selected_port = FD_SETSIZE;
  return result;
}

/* RESOURCE DM_accumulate_object
 * This is the default site->handler. It simply enqueues incoming objects.
 * Their arrival is signaled by the server loop.
 */

void
DM_accumulate_object (DM_Context *context,
                      DM_Object  *object )
{
  DM_enqueue_received_object(context, object);
}

/* RESOURCE DM_work_once
 * This function manages the event loop of DMeroon. It watches Port
 * connections and sets their options to their exact status (writable,
 * readable ...). Non null parameters will limit this function to a
 * given number of seconds and microseconds.  The function returns as
 * soon as one event is handled, it may thus return sooner. It returns
 * DM_FALSE if and only if nothing happens during the given duration.
 */

static DM_c_boolean
DM_work_once (DM_Context *context,
              DM_nat3    sec,
              DM_nat3    usec ) 
{
  DM_ports_vector *active_ports = 
    (DM_ports_vector*)(DM_ReferenceTo(DM_active_ports));
  DM_Port         *port;
  DM_nat3         i;
  int             nfound;
  int             nselect = 0;
  struct timeval  tv;

  /* Finish scanning the watched ports and exit if some work is
   * performed. Normally this will take care of all writable/readable
   * Connections. If nothing can be done then select(2) will be called
   * again to refresh the state of Connections. */
  if ( DM_analyze_all_ports(context, 0) ) {
    return DM_TRUE;
  }

  /* Get a fresh look of readable/writable Ports that is, prepare to
   * make a call to select(2). The result of select(2) is held in
   * these variables. */
  DM_last_ports_to_read  = DM_current_ports_to_read;
  DM_last_ports_to_write = DM_current_ports_to_write;
  tv.tv_sec  = 0;
  tv.tv_usec = 0;

 retry_select:
  DM_ResetErrno();

  DM_DebugAtLevel(2,
    fprintf(DM_FindErrorStream(context), "Selecting for %lus %luus on [ ", 
            (long unsigned)tv.tv_sec, 
            (long unsigned)tv.tv_usec );
    for ( i=0 ; i<FD_SETSIZE ; i++ ) {
      if ( FD_ISSET(i, &DM_last_ports_to_read) ) {
        fprintf(DM_FindErrorStream(context), "<%lu ", (unsigned long)i);
      }
      if ( FD_ISSET(i, &DM_last_ports_to_write) ) {
        fprintf(DM_FindErrorStream(context), ">%lu ", (unsigned long)i);
      }
    }
    fprintf(DM_FindErrorStream(context), "]\n");
);

  if ( (nfound = select(DM_max_port_number+1,
                        &DM_last_ports_to_read,
                        &DM_last_ports_to_write,
                        (fd_set*)NULL,
                        &tv )) < 0 ) {
    /* This means that the select was interrupted by ^C or something
     * similar, just return False in that case rather than raising an
     * exception. */
    if ( DM_IsErrnoEqualTo(DM_EINTR) ) 
      return DM_FALSE;
    DM_Abort(context, DME_FAILED_SELECT);
  }
  /* The select(2) operation was successful. 
   * Normally, at most two select(2) are performed. */
  nselect++;

  DM_DebugAtLevel(10,
    fprintf(DM_FindErrorStream(context), 
            "Select returns %d, readable=[ ", nfound);
    for ( i=0 ; i<FD_SETSIZE ; i++ ) {
      if ( FD_ISSET(i, &DM_last_ports_to_read) ) {
        fprintf(DM_FindErrorStream(context), "%lu ", (unsigned long)i);
      }
    }
    fprintf(DM_FindErrorStream(context), "], writable=[ ");
    for ( i=0 ; i<FD_SETSIZE ; i++ ) {
      if ( FD_ISSET(i, &DM_last_ports_to_write) ) {
        fprintf(DM_FindErrorStream(context), "%lu ", (unsigned long)i);
      }
    }
    fprintf(DM_FindErrorStream(context), "]\n");
);

  /* Record within Ports, which ones are readable/writable. This
   * is for the sake of reflection as well as for the loop that
   * exploits the characteristics of Ports. */
  for ( i=0 ; i<=DM_max_port_number ; i++ ) {
    port = active_ports->port[i];
    if ( DM_IsA(context, port, Port) ) {
      if ( FD_ISSET(i, &DM_last_ports_to_read) ) {
        DM_SetOption(port, DM_PORT_IS_READABLE);
      } else {
        DM_ClearOption(port, DM_PORT_IS_READABLE);
      }
      if ( FD_ISSET(i, &DM_last_ports_to_write) ) {
        DM_SetOption(port, DM_PORT_IS_WRITABLE);
      } else {
        DM_ClearOption(port, DM_PORT_IS_WRITABLE);
      }
    }
  }

  /* At least one port seems readable or writable then analyze the
   * resulting ports starting from Port 0. */
  if ( nfound > 0 ) {
    if ( DM_analyze_all_ports(context, DM_max_port_number) ) {
      return DM_TRUE;
    }
  } 

  if ( nselect < 2 ) {
    /* Nothing could be done. Reselect only on readable ports with the
     * full time now. */
    DM_last_ports_to_read  = DM_current_ports_to_read;
    FD_ZERO(&DM_last_ports_to_write);
    tv.tv_sec  = sec;
    tv.tv_usec = usec;
    goto retry_select;
  } else {
    /* Nothing happens while selecting. */

    DM_DebugAtLevel(9,
      fprintf(DM_FindErrorStream(context),
              "Nothing happened during DM_work_once.\n");
);

    return DM_FALSE;
  }
}

/* }}} 
 * {{{ Work functions

 * DOCUMENTATION event loop management
 * These are the functions that manage the event loop. The basic loop
 * management is done by DM_work_once that scans file descriptors and
 * return a boolean telling if something happened or not. 

 * RESOURCE DM_work_at_most
 * This function manages the event loop of DMeroon but it ensures that
 * the longest atomic wait is limited by site->cyclic_duration. This
 * allows from time to time to maintain Connections in a live state,
 * to flush buffers etc.  This function may return before if
 * something arrived, whatever it is.
 */

static DM_c_boolean
DM_work_at_most (DM_Context *context,
                 DM_nat3    sec )
{
  DM_date end_date = DM_get_date(context) + sec;
  DM_nat3 cycle    = DM_TheCurrentSite()->cyclic_duration;
  DM_int3 remain   = sec;

  while ( remain > 0 ) {
    if ( remain < cycle ) {
      return DM_work_once(context, remain, 0);
    } else {
      if ( DM_work_once(context, cycle, 0) ) {
        return DM_TRUE;
      } else {
        remain = end_date - DM_get_date(context);
      }
    }
  }
  return DM_FALSE;
}

/* RESOURCE DM_work_at_least
 * This function ensures that DMeroon will serve requests for at least
 * a given duration. Contrarily to DM_work_at_most which stops when
 * something arrived, DM_work_at_least serves until exhaustion of the
 * given duration. Nevertheless, it returns true if something happened.
 */

DM_c_boolean
DM_work_at_least (DM_Context* context,
                  DM_nat3     duration ) 
{
  DM_date end_date    = DM_get_date(context) + duration;
  DM_int3 remain      = duration;
  DM_c_boolean result = DM_FALSE;

  while ( remain > 0 ) {
    if ( DM_work_at_most(context, remain) ) {
      result = DM_TRUE;
    } else {
      /* Nothing: continue to wait. */
    }
    remain = end_date - DM_get_date(context);
  }
  return result;
}

/* }}}
 * {{{ Thread system interface

 * DOCUMENTATION Thread system interface
 * There are three possible threads interface. Or you don't have
 * threads and you are in a sequential setting (DM_MONOTHREAD_CONTEXT)
 * or you have a threads system (DM_MULTITHREAD_CONTEXT) or you have
 * coroutines (DM_COROUTINE_CONTEXT). Waiting is signaled by a DM_wait*
 * function, the wait is ended by a DM_signal* associated function.
 */

#if defined(DM_MONOTHREAD_CONTEXT)

/*    {{{ Sequential context
 * DOCUMENTATION Sequential context
 * In mono-thread context, Requests are pushed in a stack and Answers
 * are waited for in that order (last-in, first-out) that is the
 * Request at the top will be popped only if its Answer come
 * back. This may induce (later) other Requests to be popped off the
 * stack if their Answers were already handled. 
*/

#define DM_CoRoutineSuspend(context, waiter) \
  DM_monothread_suspend(context, waiter)

#define DM_CoRoutineResume(context, waiter, errcode) \
  DM_monothread_resume(context, waiter, errcode)

#define DM_CoRoutineCreate(context, expression, date) \
  DM_BarfIfReached(context)

#define DM_CoRoutineEnd(context) \
  DM_BarfIfReached(context)

#define DM_InsertTimedWaiter(context, w) \
  { DM_Site   *site = DM_TheCurrentSite();     \
    DM_Waiter *_waiter = w;                     \
    _waiter->next = site->waiter;                \
    site->waiter = _waiter;                       \
  }

/* RESOURCE DM_monothread_suspend
 * Enters a new server loop.  The Waiter is already pushed in the list 
 * of pending calls (see DM_InsertTimedWaiter right above).
 */

DM_ErrorCode
DM_monothread_suspend (DM_Context *context,
                       DM_Waiter  *waiter ) 
{
  jmp_buf _exit;

  if ( setjmp(_exit) ) {
    /* Someone calls DM_monothread_resume */

    DM_DebugAtLevel(9,
      fprintf(DM_FindErrorStream(context),
              "Bask to suspend, old Waiter is " DM_0X "%p\n",
              waiter );
);

    return waiter->errcode;
  } else {
    waiter->thread = &_exit;

    DM_DebugAtLevel(9,
      fprintf(DM_FindErrorStream(context),
              "Suspend, new Waiter is " DM_0X "%p\n",
              waiter );
);

    DM_ProtectWithFreshContext(context, ctxt, {
           DM_monothread_loop(context);
           }, 
           /* In case of errors. */
           return ctxt->errcode;
                             );
    /* Should never be reached. */
    DM_BarfIfReached(context);
    return DME_DEAD_CODE_REACHED;
  }
}

/* RESOURCE DM_monothread_resume
 * This is to signal that we may resume a waiting call but only if this
 * was the most recent pending call (otherwise, to longjmp to an older
 * pending call would remove from the stack an interesting but still
 * waiting call).
 */

void
DM_monothread_resume (DM_Context   *context,
                      DM_Waiter    *waiter,
                      DM_ErrorCode errcode )
{
  DM_Site *site = DM_TheCurrentSite();

  waiter->errcode = errcode;
  DM_SetOption(waiter, DM_WAITER_DONE);
  if ( waiter == site->waiter ) {

    DM_DebugAtLevel(9,
      DM_Class *c = DM_ObjectToClass(context, waiter);
      fprintf(DM_FindErrorStream(context),
              "Resuming %s " DM_0X "%p\n",
              DM_predefined_name(context, (DM_Object*)c), waiter );
);

    site->waiter = waiter->next;
    longjmp(*((jmp_buf*)(waiter->thread)), 1);
  } else {

  DM_DebugAtLevel(9,
    DM_Class* c = DM_ObjectToClass(context, waiter);
    fprintf(DM_FindErrorStream(context),
            "Cannot resume %s " DM_0X "%p\n",
             DM_predefined_name(context, (DM_Object*)c), waiter );
);

    DM_ProtectWithFreshContext(context, ctxt, {
           DM_monothread_loop(context);
           }, 
           /* In case of errors. */
           DM_Abort(context, ctxt->errcode);
                             );
  }
}

/* RESOURCE DM_monothread_loop
 * The basic server loop for the monothread context. This function is
 * only invoked from DM_monothread_{resume,suspend} so there is always
 * a Waiter (ie a pending call). It may force a timeout for this waiter
 * if nothing happened for 'period' seconds.
 */

DM_ErrorCode
DM_monothread_loop (DM_Context *context)
{
  DM_int3   duration;
  DM_date   current_date;
  DM_date   calm;
  DM_Site   *site = DM_TheCurrentSite();

  /* At `calm' time and if nothing happened, exit the loop. */
  current_date = DM_get_date(context);
  calm = current_date + site->period;

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "Entering server loop, current date is %lu\n",
            (unsigned long) current_date);
);

  /* Now we enter the real loop. Always reset current date. */
  while ( (current_date = DM_get_date(context)) <= calm ) {

    /* Check if there are pending calls waiting to be resumed. 
     * If there is one, restart it. */
    if ( site->waiter && DM_HasOption(site->waiter, DM_WAITER_DONE) ) {
      DM_monothread_resume(context, site->waiter, site->waiter->errcode);
      /* Should never be reached. */
      DM_BarfIfReached(context);
    }

    /* Analyze whether there are some answers or requests to be handled. */
    {
      DM_Object *ra = DM_remove_last_queuelink(context, &(site->arrived));
      
      if ( ra ) {
        if ( DM_IsA(context, ra, Answer) ) {
          DM_receive_answer(context, (DM_Answer*)ra);
        } else if ( DM_IsA(context, ra, Request) ) {
          DM_receive_request(context, (DM_Request*)ra);
        } else {
          DM_Abort(context, DME_NOT_REQUEST_NOR_ANSWER);
        }
        /* Handle all requests/answers before the sequel. */
        continue;
      }
    }

    /* Analyse whether there are receivable objects. */
    if ( DM_TheCurrentSite()->objects ) {
      DM_signal_object_reception(context);
      /* Returns here, if no one wants to receive an object. */
    }

    /* duration is positive here but to ensure that ping-ing other
     * sites is done, duration should not exceed cyclic_duration. */
    if ( duration > DM_TheCurrentSite()->cyclic_duration ) {
      duration = DM_TheCurrentSite()->cyclic_duration;
    }
    
    if ( DM_work_once(context, duration, 0) ) {
      /* Something happened, extend calm date. */
      current_date = DM_get_date(context);
      calm = current_date + site->period;

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), 
                "Extending calm period to %lu (current time is %lu)\n",
                (unsigned long)calm,
                (unsigned long) current_date );
);

      continue;
    } else {
      /* Nothing happened. */

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), 
                "Nothing worth, current time is %lu\n",
                (unsigned long) current_date );
);
      
      continue;
    }
  }
  /* Nothing happened for `period' seconds, exits current Waiter. */

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "Calm period exceeded (Current date is %lu, calm is %lu)\n",
            (unsigned long) current_date,
            (unsigned long) calm );
);

  DM_monothread_resume(context, site->waiter, DME_TIMEOUT);
  /* Should never be reached. */
  DM_BarfIfReached(context);
  return DME_DEAD_CODE_REACHED;
}

#endif /* !defined(DM_MONOTHREAD_CONTEXT) */
#if defined(DM_COROUTINE_CONTEXT)

/*    }}}
 *    {{{ Coroutine context
 * DOCUMENTATION Coroutine context
 * When callcc is available from C (or any other coroutine system), we
 * may use it. This has the interest that only one thread is running.
 * Passing control is explicit between coroutines. 

 * RESOURCE DM_insert_waiter
 * Insert a Waiter in the list of waiters. This list is ordered by date of
 * time-out, the first one being the more urgent.
 */

#define DM_InsertTimedWaiter(context, _w) \
  DM_insert_waiter(context,                \
                   (DM_Waiter*)(_w),        \
                   &(DM_TheCurrentSite()->waiter) )

static void
DM_insert_waiter (DM_Context *context,
                  DM_Waiter  *waiter,
                  DM_Waiter* *location )
{
  while ( *location ) {
    if ( (*location)->end_date < waiter->end_date ) {
      /* Don't insert here, skip the current Waiter. */
      location = &((*location)->next);
    } else {
      /* Insert right before the next Waiter. */
      waiter->next = *location;
      *location = waiter;
      return;
    }
  }
  /* All present Waiters (if any) are more urgent, insert at end. */
  waiter->next = DM_NULL;
  *location = waiter;
  return;
}

/* Only the Bigloo binding implements the DM_COROUTINE_CONTEXT.
 * Normally, only the two next functions provided by the Scheme library
 * are essential for that binding. */

#ifdef DM_SCHEME_IS_BIGLOO   

extern obj_t call_cc (obj_t);
extern obj_t apply_continuation (obj_t, obj_t);

/* RESOURCE DMbgl_return_to_server
 * This variable will hold a Bigloo closure that wraps a call to 
 * DMbgl_escape_to_server: 
 *        (lambda (errcode) (DM-escape-to-server errcode)).
 * Since it is difficult to create this closure from C, it is created
 * from Scheme in main1.bgl and given to C via DMbgl_register_return_to_server
 * where it is stored in the DMbgl_return_to_server variable. 
 *
 * Normally, DMbgl_register_return_to_server is called only once.
 */

static obj_t DMbgl_return_to_server = NULL;

void
DMbgl_register_return_to_server (obj_t proc)
{

#if defined(DM_DEBUG)
  static int called_once = 0;
  if ( called_once == 0 ) {
    called_once = 1;
  } else {
    DM_BarfIfReached(DM_NULL);
  }
#endif /# DM_DEBUG */

  DMbgl_return_to_server = proc;
}

/* DOCUMENTATION Coroutine macros
 * These four macros encapsulate for C the use of coroutines.
 * one to create them, one to suspend them, one to resume them
 * and one to let them die. 

 * DOCUMENTATION DM_last_waiter 
 * This is a conduit variable between DM_CoRoutineSuspend
 * and DMbgl_escape_to_server. It is only used temporarily in
 * a coroutine context.
 */

static DM_Waiter *DM_last_waiter = DM_NULL;

/* RESOURCE DM_CoRoutineSuspend
 * Capture the current continuation into a Scheme continuation and gives it
 * to the DMbgl_escape_to_server function. When resumed this continuation
 * will continue execution with the instruction that follows the 
 * DM_CoRoutineSuspend instruction. The previous DM_last_waiter global variable
 * is used to communication with the DMbgl_escape_to_server function below.
 * Normally, the idiom to use this macro is:
 *            errcode = DM_CoRoutineSuspend(...)
 */

#define DM_CoRoutineSuspend(context, waiter) \
  ( DM_last_waiter = (DM_Waiter*)waiter, \
    (DM_nat3)(DM_FromSupportFixnum(call_cc(DMbgl_return_to_server))) )

/* RESOURCE DM_CoRoutineResume
 * Resume a coroutine with a given errorcode.
 */

#define DM_CoRoutineResume(context, waiter, errcode) \
    DM_DebugAtLevel(9, { \
      DM_Class *c = DM_ObjectToClass(DM_NULL, waiter); \
      fprintf(DM_debug_stream, "Resuming %s at %p with %lu\n", \
              DM_predefined_name(DM_NULL, (DM_Object*)c), \
              waiter, (unsigned long)errcode); \
    } \
  ); \
  DM_apply_continuation(context, waiter->thread, DM_ToSupportFixnum(errcode))

/* RESOURCE DM_CoRoutineCreate
 * Create a new coroutine. It will be run at date (zero means as soon as
 * possible). When the coroutine is finished, its descriptor will 
 * disappear and control will be given back to the DMbgl_escape_to_server.
 */

#define DM_CoRoutineCreate(context, expression, date) \
  { DM_Waiter* _waiter = DM_Allocate0(context, EvaluationWaiter); \
    _waiter->end_date = date; \
    DM_InsertTimedWaiter(context, _waiter); \
    /* Ignore the DME_TIMEOUT error code: */ \
    (void) DM_CoRoutineSuspend(context, _waiter); \
    expression; \
    _waiter->thread = NULL; \
    DM_CoRoutineEnd(context); \
  }

/* RESOURCE DM_CoRoutineEnd
 * Return to the scheduler. 
 */

#define DM_CoRoutineEnd(context) \
  ( DMbgl_escape_to_server(NULL) )

/* DOCUMENTATION DM_apply_continuation 
 * Invoke a Scheme continuation. This function just converts a DMeroon
 * signature into a Scheme invocation. 
 */

obj_t
DM_apply_continuation (DM_Context *context, void* k, void* value) 
{
  if ( k ) {
    return apply_continuation((obj_t)k, value);
  } else {
    DM_Abort(context, DME_WRONG_CONTINUATION);
    /* Dead code here. */
    return BNIL;
  }
}

/* RESOURCE DMbgl_handle_incoming_object
 * This is the interface function between DMeroon and Bigloo. Any time 
 * an object arrives, it is submitted to this function which will check 
 * if the Bigloo user has specified how to handle it. By default, the
 * object is enqueued in site->objects.
 */

void
DMbgl_handle_incoming_object (DM_Context *context,
                              DM_Object  *o ) 
{
  if ( DM_HANDLE_OBJECT_IN_BIGLOO == BFALSE ) {
    DM_enqueue_received_object(context, o);
    /* Don't signal object reception, this is done by the server loop. */
  } else {
    DM_CoRoutineCreate(context, 
                       apply(DM_HANDLE_OBJECT_IN_BIGLOO, 
                             MAKE_PAIR(DM_ToSupport(o), BNIL) ),
                       0 ); 
  }
}

/* RESOURCE DM_start_server
 * This simple function is used from Bigloo to start the server. It should
 * be called once only. It is called from Bigloo/main1.bgl (DM-start-server)
 * which is defined (in Bigloo/dm.bgl) to be, in fact, DMbgl_start_server.
 * Returning from this function exits the Bigloo+DMeroon executable.
 */

void
DMbgl_start_server ()
{
  DM_ProtectWithFreshContext(NULL, ctxt, {
                (void) DM_basic_server_loop(ctxt);
                return;
                },
     /* In case of errors. */
     return;
);
}

#endif /* DM_SCHEME_IS_BIGLOO */

/* DOCUMENTATION
 * DM_server_jmp is the jmpbuf that allows to return the main server loop.
 * DM_server_waiter holds the Scheme continuation that allows to return 
 * the main server loop.
 */

#if defined(DM_MONOTHREAD_CONTEXT)
static jmp_buf      DM_server_jmp;
#endif

#if defined(DM_COROUTINE_CONTEXT)
static DM_Waiter *DM_server_waiter = DM_NULL;
#endif

/* RESOURCE DMbgl_escape_to_server
 * This function receives the continuation right after its allocation via
 * call/cc. The current waiter to fill is stored in DM_last_waiter. Then
 * it escapes to the basic server loop. Called with null, it simply 
 * escapes to the basic server loop.
 *
 * There is a special case, when DM_last_waiter is DM_NULL, to record
 * how to return to the basic server loop.
 */

DM_SupportPointer
DMbgl_escape_to_server (DM_SupportPointer wrapped_continuation)
{
  DM_DebugAtLevel(9,
    fprintf(DM_debug_stream, "Passing through DMbgl_escape_to_server\n");
);

  if ( DM_last_waiter ) {
    if ( wrapped_continuation ) {
      /* Store the received continuation in the current waiter. */
      DM_last_waiter->thread = wrapped_continuation;
    } else {
      DM_last_waiter->thread = NULL;
    }

    DM_DebugAtLevel(9, {
      DM_Class *c = DM_ObjectToClass(DM_NULL, DM_last_waiter);
      fprintf(DM_debug_stream, "%s %s at %p\n",
              (wrapped_continuation)?"Suspending":"Ending",
              DM_predefined_name(DM_NULL, (DM_Object*)c),
              DM_last_waiter);
    }
);

    /* Return inside the basic server loop. */

#if defined(DM_MONOTHREAD_CONTEXT)
    longjmp(DM_server_jmp, 1);
#endif
#if defined(DM_COROUTINE_CONTEXT)
    if ( DM_server_waiter->thread ) {
      DM_CoRoutineResume(DM_NULL, DM_server_waiter, 1);
    } else {
      DM_Abort(DM_NULL, DME_BASIC_SERVER_LOOP_UNINITIALIZED);
    }
#endif

  } else {

    DM_DebugAtLevel(9,
      fprintf(DM_debug_stream, "set up basic server loop Waiter\n");
);

    if ( DM_server_waiter->thread ) {
      DM_Abort(DM_NULL, DME_BASIC_SERVER_LOOP_REINITIALIZED);
    } else {
      /* When DM_last_waiter is DM_NULL, we are recording a continuation
       * that allows to return to the basic server loop. Return to it
       * immediately with 0 as value (as setjmp does). */
      DM_server_waiter->thread = wrapped_continuation;
      DM_CoRoutineResume(DM_NULL, DM_server_waiter, 0);
    }
  }
  /* Never reached */
  DM_BarfIfReached(DM_NULL);
  return DM_NULL;
}

/* RESOURCE DM_shutdown
 * This function shuts down a DMeroon server. It has to finish ports etc.
 */

DM_ErrorCode
DM_shutdown (DM_Context *context)
{

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), "Shutdown server...\n");
);

  /* Many things to finish cleanly.....                 FUTURE
   * For instance, close the Connections. */

  exit(EXIT_SUCCESS);
}

/* RESOURCE DM_basic_server_loop
 * This is the single server loop (at the bottom of the stack). It exits
 * when nothing happened for period. This function must be called as soon
 * as possible (ie the lowest possible in the stack) and must not be called 
 * again.
 * 
 * ATTENTION: call/cc copies stacks so stack-allocated shared mutable
 * variables may be duplicated.
 */

DM_ErrorCode
DM_basic_server_loop (DM_Context *context)
{
  DM_int3 duration;
  static DM_date      current_date;
  static DM_date      calm;
  static DM_c_boolean server_started = DM_FALSE;
  static DM_Waiter    *waiter = DM_NULL;
  DM_Site             *site = DM_TheCurrentSite();

  if ( server_started ) {
    /* Should never call again this function. */
    DM_Abort(context, DME_BASIC_LOOP_RESTARTED);
  } else {
    /* NOTHING: continue */
  }

  { /* Record the continuation of the basic server loop. This will be
     * needed by DMbgl_escape_to_server to return here.  NOTE: This
     * programmation is more portable than before since it does not
     * mix continuations with longjmp something that is highly
     * difficult in a world with dynamic-wind (as shown by Bigloo
     * 1.9c). */
#if defined(DM_MONOTHREAD_CONTEXT)
    int status = setjmp(DM_server_jmp);
#endif
#if defined(DM_COROUTINE_CONTEXT)
    /* This call with a DM_NULL second argument should be the only one. */
    DM_nat3 status = DM_CoRoutineSuspend(context, DM_NULL);
#endif

    if ( status ) {
      /* Here should come back DMbgl_escape_to_server then skip the next
       * section where initial threads are started: they must be started
       * only once! */
      goto enter_loop;
    } else {
      /* setjmp/DM_CoRoutineResume returns here. */
    }
  }

  /* Before, we enter the loop, we may start some threads. */
  if ( ! server_started ) {
    server_started = DM_TRUE;
    
#if defined(DM_SCHEME_IS_BIGLOO)
    { extern obj_t repl___eval();

    /* In Bigloo, every incoming object is given, in a fresh
     * coroutine, to a function that handles it. */
    (void) DMeroon_install_receiver
              (context, DM_CodeOf(DMbgl_handle_incoming_object));

    /* repl will block at the first read (from stdin or other) which
     * will call DM_wait_until_fd_ready which comes back (in this
     * function, some lines above) via longjmp/CoRoutineResume.  0
     * means as soon as possible. Immediately shutdowns the server
     * when stdin is empty. */
    DM_CoRoutineCreate(context, 
                       { repl___eval();
                         (void) DM_shutdown(context); 
                       },
                       0 );
    DM_BarfIfReached(context);
    }
#endif /* DM_SCHEME_IS_BIGLOO */
    
    /* Start other concurrent threads here ...   */
  }

enter_loop:
  /* At `calm' time and if nothing happened, exit the loop. */
  current_date = DM_get_date(context);
  calm = current_date + site->period;

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "Back to basic server loop. Current date is %lu\n",
            (unsigned long) current_date);
);

  /* Now we enter the real loop. Always reset current date. */
  while ( (current_date = DM_get_date(context)) <= calm ) {

    /* Analyze whether there are some answers or requests to be handled. */
    {
      DM_Object *ra = DM_remove_last_queuelink(context, &(site->arrived));
      
      if ( ra ) {
        if ( DM_IsA(context, ra, Answer) ) {
          DM_receive_answer(context, (DM_Answer*)ra);
        } else if ( DM_IsA(context, ra, Request) ) {
          DM_receive_request(context, (DM_Request*)ra);
        } else {
          DM_Abort(context, DME_NOT_REQUEST_NOR_ANSWER);
        }
        /* Handle all requests/answers before the sequel. */
        continue;
      }
    }

    /* Analyse whether there are receivable objects. This may be speed
       up if maintaining the number of Waiters on DM-receive.  FUTURE */
    if ( DM_TheCurrentSite()->objects ) {
      DM_signal_object_reception(context);
      /* If there were some objects but no threads to receive them, then
       * we come back here and continue in sequence. */
    }

    /* Analyze pending time-outs, signal all of them before the sequel. */
    if ( site->waiter ) {
      waiter = site->waiter;
      duration = waiter->end_date - DM_get_date(context);
      if ( duration <= 0 ) {
        DM_signal_end_of_duration(context, waiter);
        DM_BarfIfReached(context);
      } else {
        
        DM_DebugAtLevel(9,
          { int n = 0;
            DM_Waiter* w = site->waiter;
            void *class = DM_ObjectToClass(context, w);

            while ( w ) {
              n++;
              w = w->next;
            }
            fprintf(DM_FindErrorStream(context), 
                    "First %s is " DM_0X "%p with end_date %lu.\n"
                    "Current date is %lu, %d other Waiter(s) queued.\n",
                    DM_predefined_name(context, class),
                    waiter, 
                    (unsigned long)(site->waiter->end_date),
                    (unsigned long) current_date,
                    n-1 );
          }
);

      }
    } else {
      waiter = DM_NULL;
      duration = calm - DM_get_date(context);
    }
    /* duration is positive here but to ensure that ping-ing other
     * sites is done, duration should not exceed cyclic_duration. */
    if ( duration > DM_TheCurrentSite()->cyclic_duration ) {
      duration = DM_TheCurrentSite()->cyclic_duration;
    }
    
    if ( DM_work_once(context, duration, 0) ) {
      /* Something happened, reset calm date. */
      current_date = DM_get_date(context);
      calm = current_date + site->period;

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), 
                "Extending calm period to %lu\n",
                (unsigned long)calm );
);

      continue;
    } else {
      /* Nothing happened. */
      continue;
    }
  }
  /* Nothing happened for `period' seconds, exits. Make possible now
   * to restart the basic loop. */

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "Calm period exceeded (Current date is %lu, calm is %lu)\n",
            (unsigned long) current_date,
            (unsigned long) calm );
);

  server_started = DM_FALSE;
  return DME_SUCCESS;
}

#endif /* !defined(DM_COROUTINE_CONTEXT) */
#if defined(DM_MULTITHREAD_CONTEXT)

/*    }}}
 *    {{{ Multi-thread context
 * DOCUMENTATION multi-thread context
 * In a multi-thread context, as soon as an Answer comes back, the
 * thread that was suspended while waiting for that Answer is resumed
 * in parallel with the others. Care must be taken in the API not to
 * have synchronization problems (FUTURE).
 */


DM_ErrorCode
DM_wait_until_answer (DM_Context *context,
                      DM_nat3    duration,
                      DM_Answer* *location,
                      DM_Request *request )
{
}

void
DM_signal_answer (DM_Context *context,
                  DM_Answer  *answer ) 
{
}

DM_ErrorCode
DM_wait_until_object_reception (DM_Context* context,
                                DM_nat3     duration,
                                DM_Object*  *location ) 
{
}

void
DM_signal_object_reception (DM_Context *context) 
{
}

DM_ErrorCode
DM_wait_until_duration (DM_Context *context,
                        DM_nat3    duration ) 
{
}

void
DM_signal_end_of_duration (DM_Context *context,
                           DM_Waiter  *waiter )
{
}

DM_ErrorCode
DM_wait_until_fd_ready (DM_Context *context,
                        DM_nat3    fd) 
{
}

void
DM_signal_fd_ready (DM_Context *context,
                    DM_nat3    fd) 
{
}

#endif /* DM_MULTITHREAD_CONTEXT */

#if defined(DM_MONOTHREAD_CONTEXT) || defined(DM_COROUTINE_CONTEXT) 

/* RESOURCE DM_wait_until_answer
 * Waits until the Answer for a request comes in. Stores the Answer in the 
 * given location.
 */

DM_ErrorCode
DM_wait_until_answer (DM_Context *context,
                      DM_nat3    duration,
                      DM_Answer* *location,
                      DM_Request *request )
{
  DM_ErrorCode errcode;
  DM_AnswerWaiter *waiter = DM_Allocate0(context, AnswerWaiter);

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "wait_until_answer to " DM_0X "%p, Waiter is " DM_0X "%p\n", 
            request, waiter );
);

  waiter->end_date = DM_get_date(context) + duration;
  waiter->request  = request;
  waiter->answer   = location;
  DM_InsertTimedWaiter(context, (DM_Waiter*)waiter);
  errcode = DM_CoRoutineSuspend(context, (DM_Waiter*)waiter);
  return errcode;
}

void
DM_signal_answer (DM_Context *context,
                  DM_Answer  *answer ) 
{
  DM_Waiter* *location = &(DM_TheCurrentSite()->waiter);
  DM_Request *request;

  DM_ShortCutReference(context, &(answer->request));
  request = answer->request;

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "signal_answer " DM_0X "%p to request " DM_0X "%p\n",
            answer, request );
);

  while ( *location ) {
    if ( DM_IsA(context, *location, AnswerWaiter) ) {
      DM_DeclareNarrowedObject(waiter, AnswerWaiter, *location);

      if ( waiter->request == request ) {
        *((DM_Answer**)(waiter->answer)) = answer;

#if defined(DM_COROUTINE_CONTEXT)
        /* Remove the waiter in the coroutine context only. In the
         * monothread context, DM_monothread_resume will do it. */
        *location = waiter->next;
        waiter->next = DM_NULL;
#endif /* !defined(DM_COROUTINE_CONTEXT) */

        DM_CoRoutineResume(context, (DM_Waiter*)waiter, DME_SUCCESS);
        DM_BarfIfReached(context);
      } else {
        location = &((*location)->next);
      }
    } else {
      location = &((*location)->next);
    }
  }

  /* This means that no thread was waiting for this answer. This is
   * the case for pre-fetch requests. */
  return;
}

/* RESOURCE DM_wait_until_object_reception
 * There is no need to have a special list of threads waiting for objects.
 */

DM_ErrorCode
DM_wait_until_object_reception (DM_Context* context,
                                DM_nat3     duration,
                                DM_Object*  *location ) 
{
  DM_ErrorCode errcode;
  DM_ReceptionWaiter *waiter = DM_Allocate0(context, ReceptionWaiter);
  
  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "wait_until_object_reception, Waiter is " DM_0X "%p\n", 
            waiter );
);

  waiter->end_date = DM_get_date(context) + duration;
  DM_InsertTimedWaiter(context, (DM_Waiter*)waiter);
  errcode = DM_CoRoutineSuspend(context, (DM_Waiter*)waiter);
  return errcode;
}

void
DM_signal_object_reception (DM_Context *context) 
{
  DM_Waiter* *location = &(DM_TheCurrentSite()->waiter);

  while ( *location ) {
    if ( DM_IsA(context, *location, ReceptionWaiter) ) {
      DM_DeclareNarrowedObject(waiter, ReceptionWaiter, *location);

      DM_DebugAtLevel(9,
        fprintf(DM_FindErrorStream(context), 
                "signal_object_reception to Waiter " DM_0X "%p\n",
                waiter );
);

#if defined(DM_COROUTINE_CONTEXT)
      /* Remove the waiter in the coroutine context only. In the
       * monothread context, DM_monothread_resume will do it. */
      *location = waiter->next;
      waiter->next = DM_NULL;
#endif /* !defined(DM_COROUTINE_CONTEXT) */

      DM_CoRoutineResume(context, (DM_Waiter*)waiter, DME_SUCCESS);
      DM_BarfIfReached(context);
    } else {
      location = &((*location)->next);
    }
  }
  /* No threads are waiting on a DM_receive: that's OK: leave the
   * object enqueued. The next iteration of the server loop will see
   * if there are some Waiters for it.                FUTURE to improve. */
  return;
}

DM_ErrorCode
DM_wait_until_duration (DM_Context *context,
                        DM_nat3    duration ) 
{
  DM_ErrorCode errcode;
  DM_Waiter *waiter = DM_Allocate0(context, DurationWaiter);

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "wait_until_duration, Waiter is " DM_0X "%p\n", 
            waiter );
);

  waiter->end_date = DM_get_date(context) + duration;
  DM_InsertTimedWaiter(context, waiter);
  errcode = DM_CoRoutineSuspend(context, waiter);
  return errcode;
}

void
DM_signal_end_of_duration (DM_Context *context,
                           DM_Waiter  *waiter )
{
  DM_Waiter* *location = &(DM_TheCurrentSite()->waiter);

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "signal_end_of_duration to " DM_0X "%p\n",
            waiter );
);

  if ( waiter ) {
    while ( *location ) {
      DM_Waiter *w = *location;
        
      if ( waiter == (DM_Waiter*)w ) {
#if defined(DM_COROUTINE_CONTEXT)
        *location = waiter->next;
        waiter->next = DM_NULL;
#endif /* !defined(DM_COROUTINE_CONTEXT) */

        DM_CoRoutineResume(context, waiter, DME_TIMEOUT);
        DM_BarfIfReached(context);
      } else {
        location = &((*location)->next);
      }
    }
    DM_Abort(context, DME_MISSING_WAITER);
  } else {
    /* NOTHING: No Waiter is interested by this time-out. */
  }
}

DM_ErrorCode
DM_wait_until_fd_ready (DM_Context *context,
                        DM_nat3    duration,
                        DM_nat3    fd) 
{
  DM_ErrorCode errcode;
  DM_FileWaiter *waiter = DM_Allocate0(context, FileWaiter);

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "wait_until_fd_ready %lu, Waiter is " DM_0X "%p\n", 
            (unsigned long)fd, waiter );
);

  waiter->end_date = DM_get_date(context) + duration;
  waiter->fd       = fd;
  DM_activate_port(context, fd, NULL, DM_PORT_TO_READ);
  DM_InsertTimedWaiter(context, (DM_Waiter*)waiter);
  errcode = DM_CoRoutineSuspend(context, (DM_Waiter*)waiter);
  return errcode;
}

void
DM_signal_fd_ready (DM_Context *context,
                    DM_nat3    fd) 
{
  DM_Waiter* *location = &(DM_TheCurrentSite()->waiter);

  DM_DebugAtLevel(9,
    fprintf(DM_FindErrorStream(context), 
            "signal_fd_ready %lu\n", (unsigned long)fd);
);

  DM_inactivate_port(context, fd);
  while ( *location ) {
    if ( DM_IsA(context, *location, FileWaiter) ) {
      DM_DeclareNarrowedObject(waiter, FileWaiter, *location);

      if ( waiter->fd == fd ) {
#if defined(DM_COROUTINE_CONTEXT)
        *location = waiter->next;
        waiter->next = DM_NULL;
#endif /* !defined(DM_COROUTINE_CONTEXT) */

        DM_CoRoutineResume(context, (DM_Waiter*)waiter, DME_SUCCESS);
        DM_BarfIfReached(context);
      } else {
        location = &((*location)->next);
      }
    } else {
      location = &((*location)->next);
    }
  }

  DM_Abort(context, DME_NO_FILE_WAITER);
}

#endif /* defined(DM_MONOTHREAD_CONTEXT) || defined(DM_COROUTINE_CONTEXT) */

/*     }}}
 * }}}
 * Initializes the module.
 */

void
DM_sched_initialize (DM_Context *context) 
{
  char    *periodstr = getenv("DMEROON_SERVING_PERIOD");
  DM_Site *site = DM_TheCurrentSite();

  DM_LinkVersionStringIntoSite(dmsched);

  if ( periodstr ) {
    site->period = DM_string_to_nat3(NULL, periodstr, NULL);
    if ( site->period == 0 ) {
      DM_Abort(context, DME_PERIOD_IS_ZERO);
    }
  } else {
    /* Already statically done:
       site->period = DM_DMEROON_FAVORITE_PERIOD;
     */
  }

#if defined(DM_COROUTINE_CONTEXT)
  /* Initialize the DM_server_waiter variable. */
  DM_server_waiter = DM_Allocate0(context, Waiter);
#endif

}

/* end of dmsched.c */
