/* $Id: dmscm.c,v 1.37 1998/06/10 16:31:30 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmscm.c
 * This is the Scheme generic runtime interface with DMeroon.
 * It defines functions taking Scheme values (prefixed with "wrapped-")
 * and produces Scheme values. It uses a series of macros (with the
 * word "Support" in their names) characterizing the Scheme implementation.
 * Of course, another language, different from Scheme, may be used
 * provided the necessary macros are specified.
 *
 * It supposes that the dmXXX.c file also brings a definition for 
 * DM_allocate_bare_instance and DM_raise_exception. 
 *
 * The dmscm.h associated file defines the Scheme signatures for these
 * functions. These Sexpressions are extracted and converted into the
 * appropriate Scheme code to provide the foreign function interface.

 * NOTE: Errors are reported here with DM_Abort. Normally this
 * should raise an exception to be caught by the support language.
 */

#include "dmeroon.h"

DM_DefineStringWithCstring(dmscm_version,
  "$Id: dmscm.c,v 1.37 1998/06/10 16:31:30 queinnec Exp $");

/* All invokations to the C API functions require a context. Errors are
 * still reported with DM_abort except at the interface with Scheme
 * where an exception is raised (if possible). 

 * NOTE: Correct class is always checked, errors are immediately signalled.
 * DMeroon membership is always checked so avoid to repeat this test in
 * dmnewapi.c.
 */

/* DOCUMENTATION
 * Any function called from Scheme, should first test the validity of 
 * its arguments and raise an exception (with DM_RaiseException) in case
 * of problems, then, after unwrapping arguments to their C equivalent, 
 * it should call the API function, test the errorcode (with
 * DM_CheckReturnedStatus) and return a Scheme value. Nearly all of these
 * functions are defined in this file, in Bigloo there is also dmbgl.c
 * and the DMbgl_allocate function.

 * RESOURCE DM_CheckReturnedStatus
 * This macro may be used to check the result of a call to the API.
 * Rather than longjmp-ing to the most recent context, it raises an
 * exception for the underlying language.
 */

#define DM_CheckReturnedStatus(context, errcode) \
  if ( (errcode) != DME_SUCCESS ) { \
    DM_RaiseException(context, errcode); \
  }

/* RESOURCE DM_RaiseException
 * Raise an exception for the underlying language.
 */

#define DM_RaiseException(context, errcode) \
  DM_raise_exception(context, errcode, __FILE__, __LINE__)

/* 
 * Initialization 
 */

DM_SupportPointer
DMscm_initialize () {
  DM_DefineStaticContext(DMscm_context);

  DMscm_context->errcode = DMeroon_initialize(DMscm_context, DM_NULL);
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(DM_TheCurrentSite());
}
    
/* {{{ Predicate

 * RESOURCE DM_is_object
 * This predicate returns true if given a pointer onto a DMeroon instance.
 */

DM_SupportPointer
DMscm_is_object (DM_SupportPointer wrapped_p) {
  if ( DM_IsDMeroonInstance(wrapped_p) ) {
    return DM_SupportTrue;
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_is_a (DM_SupportPointer wrapped_object,
            DM_SupportPointer wrapped_class ) {
  DM_c_boolean result;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_class)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = DMeroon_is_a(DMscm_context,
                                        &result,
                                        DM_FromSupport(wrapped_object),
                                        DM_FromSupport(wrapped_class) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( result ) {
    return DM_SupportTrue;
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_is_a_subclass (DM_SupportPointer wrapped_subclass,
                     DM_SupportPointer wrapped_class ) {
  DM_c_boolean result;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_subclass))
      || !(DM_IsDMeroonInstance(wrapped_class)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_is_a_subclass(DMscm_context,
                          &result,
                          DM_FromSupport(wrapped_subclass),
                          DM_FromSupport(wrapped_class) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( result ) {
    return DM_SupportTrue;
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_class_of (DM_SupportPointer wrapped_object) {
  DM_Class     *class;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_class_of(DMscm_context,
                     &class,
                     DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(class);
}

DM_SupportPointer
DMscm_site_of (DM_SupportPointer wrapped_object) {
  DM_Site      *site;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_site_of(DMscm_context,
                    &site,
                    DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(site);
}

DM_SupportPointer
DMscm_clock_of (DM_SupportPointer wrapped_object) {
  DM_Clock     *clock;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_clock_of(DMscm_context,
                     &clock,
                     DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( clock ) {
    return DM_ToSupport(clock);
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_receive (DM_SupportPointer wrapped_duration) {
  DM_Object    *object;
  DM_nat3      duration;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsSupportFixnum(wrapped_duration)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  duration = DM_FromSupportFixnum(wrapped_duration);
  DMscm_context->errcode = DMeroon_receive(DMscm_context,
                                           &object,
                                           duration );
  if ( DMscm_context->errcode == DME_TIMEOUT ) {
    return DM_SupportFalse;
  }

#ifdef DM_SCHEME_IS_BIGLOO

  if ( DMscm_context->errcode == DME_CANNOT_RECEIVE ) {
    /* There is a function (site->handler) that receives all incoming
     * objects. If DM-handle-object is #f then the user does not
     * specify a Bigloo function to handle objects so there is no
     * error at all, objects are enqueued and we try to fetch them
     * there. */
    if ( DM_HANDLE_OBJECT_IN_BIGLOO == BFALSE ) {
      DM_Site *site = DM_TheCurrentSite();
      DMscm_context->errcode = 
        DM_wait_until_object_reception(DMscm_context, 
                                       duration,
                                       (DM_Object**)(&(site->objects)) );

      if ( DMscm_context->errcode == DME_SUCCESS ) {
        /* An object was received, handle it. */
        object = DM_remove_last_queuelink(DMscm_context, &(site->objects));
        DM_ensure_present(DMscm_context, &object);
        return DM_ToSupport(object);
      } else if ( DMscm_context->errcode == DME_TIMEOUT ) {
        return DM_SupportFalse;
      } else {
        /* NOTHING: report the errorcode */
      }
    } else {
      /* NOTHING: report the errcode. */
    }
  }

#endif /* DM_SCHEME_IS_BIGLOO */

  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(object);
}

DM_SupportPointer
DMscm_send (DM_SupportPointer wrapped_object,
            DM_SupportPointer wrapped_site ) {
  DM_DefineStaticContext(DMscm_context);
  
  if (   !(DM_IsDMeroonInstance(wrapped_object)) 
      || !(DM_IsDMeroonInstance(wrapped_site)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_send(DMscm_context,
                 DM_FromSupport(wrapped_object),
                 DM_FromSupport(wrapped_site) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return wrapped_object;
}

DM_SupportPointer
DMscm_connect (DM_SupportPointer wrapped_hostname,
               DM_SupportPointer wrapped_portnumber,
               DM_SupportPointer wrapped_password ) {
  DM_Site      *site;
  DM_DefineStaticContext(DMscm_context);
  
  if ( !(DM_IsSupportString(wrapped_hostname)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_STRING);
  }
  if ( !(DM_IsSupportFixnum(wrapped_portnumber)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  DMscm_context->errcode = 
    DMeroon_connect(DMscm_context,
                    &site,
                    DM_FromSupportString(wrapped_hostname),
                    (DM_nat2)DM_FromSupportFixnum(wrapped_portnumber),
                    DM_FromSupportString(wrapped_password) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(site);
}

DM_SupportPointer
DMscm_create_subclass (DM_SupportPointer wrapped_superclass,
                       DM_SupportPointer wrapped_classname,
                       DM_SupportPointer wrapped_classoptions,
                       DM_SupportPointer wrapped_fieldname,
                       DM_SupportPointer wrapped_fieldoptions,
                       DM_SupportPointer wrapped_reptyp,
                       DM_SupportPointer wrapped_itemtyp,
                       DM_SupportPointer wrapped_referred_class ) {
  DM_Class     *class;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_superclass)) 
      || !(DM_IsDMeroonInstance(wrapped_reptyp)) 
      || !(DM_IsDMeroonInstance(wrapped_itemtyp)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  if (   !(DM_IsSupportString(wrapped_classname)) 
      || !(DM_IsSupportString(wrapped_fieldname)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_STRING);
  }
  if (   !(DM_IsSupportFixnum(wrapped_classoptions)) 
      || !(DM_IsSupportFixnum(wrapped_fieldoptions)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  DMscm_context->errcode =
    DMeroon_create_subclass(DMscm_context,
                            &class,
                            DM_FromSupport(wrapped_superclass),
                            DM_FromSupportString(wrapped_classname),
                            DM_FromSupportFixnum(wrapped_classoptions),
                            DM_FromSupportString(wrapped_fieldname),
                            DM_FromSupportFixnum(wrapped_fieldoptions),
                            DM_FromSupport(wrapped_reptyp),
                            DM_FromSupport(wrapped_itemtyp),
                            DM_FromSupport(wrapped_referred_class) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(class);
}

DM_SupportPointer
DMscm_smash_class (DM_SupportPointer wrapped_class,
                   DM_SupportPointer wrapped_classname,
                   DM_SupportPointer wrapped_classoptions,
                   DM_SupportPointer wrapped_level ) {
  DM_Class     *class;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_class)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  if (   !(DM_IsSupportString(wrapped_classname)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_STRING);
  }
  if (   !(DM_IsSupportFixnum(wrapped_classoptions)) 
      || !(DM_IsSupportFixnum(wrapped_level)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  DMscm_context->errcode =
    DMeroon_smash_class(DMscm_context,
                        &class,
                        DM_FromSupport(wrapped_class),
                        DM_FromSupportString(wrapped_classname),
                        DM_FromSupportFixnum(wrapped_classoptions),
                        DM_FromSupportFixnum(wrapped_level) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(class);
}

DM_SupportPointer
DMscm_set_site_of (DM_SupportPointer wrapped_object,
                   DM_SupportPointer wrapped_site ) {
  DM_Site      *site;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_site)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_set_site_of(DMscm_context,
                        &site,
                        DM_FromSupport(wrapped_object),
                        DM_FromSupport(wrapped_site) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( site ) {
    return DM_ToSupport(site);
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_set_clock_of (DM_SupportPointer wrapped_object,
                    DM_SupportPointer wrapped_clock ) {
  DM_Clock     *clock;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_clock)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_set_clock_of(DMscm_context,
                         &clock,
                         DM_FromSupport(wrapped_object),
                         DM_FromSupport(wrapped_clock) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( clock ) {
    return DM_ToSupport(clock);
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_set_sharable (DM_SupportPointer wrapped_object) {
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_set_sharable(DMscm_context,
                         DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return wrapped_object;
}

DM_SupportPointer
DMscm_set_copyable (DM_SupportPointer wrapped_object) {
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_set_copyable(DMscm_context,
                         DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return wrapped_object;
}

DM_SupportPointer
DMscm_set_initialized (DM_SupportPointer wrapped_object) {
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_set_initialized(DMscm_context,
                            DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return wrapped_object;
}

DM_SupportPointer
DMscm_is_initialized (DM_SupportPointer wrapped_object) {
  DM_nat3 result;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_object)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_is_initialized(DMscm_context,
                           &result,
                           DM_FromSupport(wrapped_object) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( result ) {
    return DM_ToSupportFixnum(result);
  } else {
    return DM_SupportFalse;
  }
}

DM_SupportPointer
DMscm_eq (DM_SupportPointer wrapped_object1,
          DM_SupportPointer wrapped_object2 ) {
  DM_c_boolean boolean;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object1))
      || !(DM_IsDMeroonInstance(wrapped_object2)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
 DMeroon_eq(DMscm_context,
            &boolean,
            DM_FromSupport(wrapped_object1),
            DM_FromSupport(wrapped_object2) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  if ( boolean ) {
    return DM_SupportTrue;
  } else {
    return DM_SupportFalse;
  }
}

/* }}}
 *    {{{ Data conversions
 
 * RESOURCE DMscm_to_support
 * This function converts a DMeroon slot-value into a Support value.
 * FUTURE: should test when nat3 = supportfixnum. 
 */

static DM_SupportPointer
DMscm_to_support (DM_Context     *context,
                  DM_DMeroonType *type,
                  void           *thing ) {
  switch ( type->index ) {
  case DM_nat1_index: {
    return DM_ToSupportFixnum(DM_Content(DM_nat1, thing));
  }
  case DM_nat2_index: {
    return DM_ToSupportFixnum(DM_Content(DM_nat2, thing));
  }
  case DM_nat3_index:
  case DM_offset_index:
  case DM_date_index:
  case DM_class_options_index:
  case DM_field_options_index:
  case DM_port_options_index:
  case DM_site_options_index:
  case DM_external_options_index:
  case DM_context_options_index:
  case DM_file_options_index:
  case DM_request_options_index: {
    return DM_ToSupportFixnum(DM_Content(DM_nat3, thing));
  }
  case DM_nat4_index: {

#ifdef DM_HAS_LONG_LONG_ENTITIES
    DM_nat4 the_nat4 = DM_Content(DM_nat4, thing);
    return DM_ToSupportFixnum(the_nat4);
#else
    DM_nat4 the_nat4 = DM_Content(DM_nat4, thing);
    if ( the_nat4._high ) {
      DM_Abort(context, DME_BAD_SUPPORT_REPRESENTATION);
    } else {
      return DM_ToSupportFixnum(the_nat4._low);
    }
#endif

  }
  case DM_reference_index: {
    DM_Object *o = DM_Content(DM_reference, thing);

    /* Avoid returning the NULL pointer, convert it into #f. */
    if ( o ) {
      return DM_ToSupport(o);
    } else {
      return DM_SupportFalse;
    }
  }
  case DM_data_index: {
    return DM_Content(DM_data, thing);
  }
  case DM_int1_index: {
    return DM_ToSupportFixnum(DM_Content(DM_int1, thing));
  }
  case DM_int2_index: {
    return DM_ToSupportFixnum(DM_Content(DM_int2, thing));
  }
  case DM_int3_index: {
    return DM_ToSupportFixnum(DM_Content(DM_int3, thing));
  }
  case DM_int4_index: {
    return DM_ToSupportFixnum(DM_Content(DM_int4, thing));
  }
  case DM_char1_index: {
    return DM_ToSupportCharacter(DM_Content(DM_char1, thing));
  }
  case DM_netnat2_index: {
    return DM_ToSupportFixnum(ntohs(DM_Content(DM_netnat2, thing)));
  }
  case DM_netnat3_index: {
    return DM_ToSupportFixnum(ntohl(DM_Content(DM_netnat3, thing)));
  }
  case DM_float1_index: {
    return DM_ToSupportFloat(DM_Content(DM_float1, thing));
  }
  case DM_float2_index: {
    return DM_ToSupportDouble(DM_Content(DM_float2, thing));
  }
  case DM_float3_index: {
    DM_DMeroonType *type = DM_ReferenceToDMeroonType(float3);

    if ( type->size ) {
      return DM_ToSupportLongDouble(DM_Content(DM_float3, thing));
    } else {
      DM_Abort(context, DME_FLOAT3_NOT_SUPPORTED);
    }
  }
  default: {
    DM_Abort(context, DME_BAD_SUPPORT_REPRESENTATION);
  }
  }
  /* NOT REACHED Normally. */
  return DM_SupportFalse;
}

/* RESOURCE DMscm_from_support
 * Converts a Scheme value into a DMeroon slot-value.
 */ 

static void
DMscm_from_support (DM_Context        *context,
                    DM_DMeroonType    *type,
                    void              *thing,
                    DM_SupportPointer wrapped_value )
{
  switch ( type->index ) {
  case DM_nat1_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (0 <= value) && (value <= DM_NAT1_MAXIMUM) ) {
        DM_Content(DM_nat1, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_NAT1_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_NAT1_FIXNUM);
    }
    return;
  }
  case DM_nat2_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (0 <= value) && (value <= DM_NAT2_MAXIMUM) ) {
        DM_Content(DM_nat2, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_NAT2_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_NAT2_FIXNUM);
    }
    return;
  }
  case DM_nat3_index:
  case DM_offset_index:
  case DM_date_index:
  case DM_class_options_index:
  case DM_field_options_index:
  case DM_port_options_index:
  case DM_site_options_index:
  case DM_external_options_index:
  case DM_context_options_index: 
  case DM_file_options_index:
  case DM_request_options_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (0 <= value) && (value <= DM_NAT3_MAXIMUM) ) {
        DM_Content(DM_nat3, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_NAT3_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_NAT3_FIXNUM);
    }
    return;
  }
  case DM_nat4_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);

#ifdef DM_HAS_LONG_LONG_ENTITIES
      if ( (0 <= value) && (value <= DM_NAT4_MAXIMUM) ) {
        DM_Content(DM_nat4, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_NAT4_FIXNUM);
      }
#else
      if ( (0 <= value) && (value <= DM_NAT3_MAXIMUM) ) {
        DM_nat4 the_nat4 = thing;
        the_nat4.low  = value;
        the_nat4.high = 0;
      } else {
        DM_RaiseException(context, DME_NOT_HANDLABLE_NAT4_FIXNUM);
      }
#endif

    } else {
      DM_RaiseException(context, DME_NOT_NAT4_FIXNUM);
    }
    return;
  }
  case DM_reference_index: {
    if ( DM_IsDMeroonInstance(wrapped_value) ) {
      DM_Content(DM_reference, thing) = DM_FromSupport(wrapped_value);
    } else if ( wrapped_value == DM_SupportFalse ) {
      /* From Scheme, #f is coerced into DM_NULL. */
      DM_Content(DM_reference, thing) = DM_NULL;
    } else {
      DM_RaiseException(context, DME_NOT_DMEROON_INSTANCE);
    }
    return;
  }
  case DM_data_index: {
    DM_Content(DM_data, thing) = wrapped_value;
    return;
  }
  case DM_code_index: 
  case DM_handler_code_index: {
    DM_Content(DM_code, thing) = DM_CodeOf(wrapped_value);
    return;
  }
  case DM_int1_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (DM_INT1_MINIMUM <= value) && (value <= DM_INT1_MAXIMUM) ) {
        DM_Content(DM_int1, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_INT1_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_INT1_FIXNUM);
    }
    return;
  }
  case DM_int2_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (DM_INT2_MINIMUM <= value) && (value <= DM_INT2_MAXIMUM) ) {
        DM_Content(DM_int2, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_INT2_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_INT2_FIXNUM);
    }
    return;
  }
  case DM_int3_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (DM_INT3_MINIMUM <= value) && (value <= DM_INT3_MAXIMUM) ) {
        DM_Content(DM_int3, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_INT3_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_INT3_FIXNUM);
    }
    return;
  }
  case DM_int4_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (DM_INT4_MINIMUM <= value) && (value <= DM_INT4_MAXIMUM) ) {
        DM_Content(DM_int4, thing) = value;
      } else {
        DM_RaiseException(context, DME_NOT_INT4_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_INT4_FIXNUM);
    }
    return;
  }
  case DM_char1_index: {
    if ( DM_IsSupportCharacter(wrapped_value) ) {
      DM_Content(DM_char1, thing) = DM_FromSupportCharacter(wrapped_value);
    } else {
      DM_RaiseException(context, DME_NOT_SUPPORT_CHARACTER);
    }
    return;
  }
  case DM_netnat2_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (0 <= value) && (value <= DM_NAT2_MAXIMUM) ) {
        DM_Content(DM_netnat2, thing) = htons(value);
      } else {
        DM_RaiseException(context, DME_NOT_NETNAT2_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_NETNAT2_FIXNUM);
    }
    return;
  }
  case DM_netnat3_index: {
    if ( DM_IsSupportFixnum(wrapped_value) ) {
      DM_SupportFixnum value = DM_FromSupportFixnum(wrapped_value);
      if ( (0 <= value) && (value <= DM_NAT3_MAXIMUM) ) {
        DM_Content(DM_netnat3, thing) = htonl(value);
      } else {
        DM_RaiseException(context, DME_NOT_NETNAT3_FIXNUM);
      }
    } else {
      DM_RaiseException(context, DME_NOT_NETNAT3_FIXNUM);
    }
    return;
  }
  case DM_float1_index: {
    if ( DM_IsSupportFloat(wrapped_value) ) {
      DM_Content(DM_float1, thing) = DM_FromSupportFloat(wrapped_value);
    } else {
      DM_RaiseException(context, DME_NOT_SUPPORT_FLOAT1);
    }
    return;
  }
  case DM_float2_index: {
    if ( DM_IsSupportDouble(wrapped_value) ) {
      DM_Content(DM_float2, thing) = DM_FromSupportDouble(wrapped_value);
    } else {
      DM_RaiseException(context, DME_NOT_SUPPORT_FLOAT2);
    }
    return;
  }
  case DM_float3_index: {
    DM_DMeroonType* type = DM_ReferenceToDMeroonType(float3);

    if ( type->size ) {
      if ( DM_IsSupportLongDouble(wrapped_value) ) {
        DM_Content(DM_float3, thing) = DM_FromSupportLongDouble(wrapped_value);
      } else {
        DM_RaiseException(context, DME_NOT_SUPPORT_FLOAT3);
      }
    } else {
      DM_RaiseException(context, DME_FLOAT3_NOT_SUPPORTED);
    }
    return;
  }
  default: {
    DM_RaiseException(context, DME_UNKNOWN_TYPE);
  }
  }
  /* NOT REACHED normally. */
  DM_RaiseException(context, DME_BAD_SUPPORT_REPRESENTATION);
  return;
}

/* }}}
 *    {{{ Field access
 */

DM_SupportPointer
DMscm_field_value (DM_SupportPointer wrapped_object,
                   DM_SupportPointer wrapped_field ) {
  DM_DMeroonType  *type;
  DM_largest_type thing;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_field_value(DMscm_context,
                        &thing,
                        &type,
                        DM_FromSupport(wrapped_object),
                        DM_FromSupport(wrapped_field) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DMscm_to_support(DMscm_context, type, &thing);
}

DM_SupportPointer
DMscm_indexed_field_value (DM_SupportPointer wrapped_object,
                           DM_SupportPointer wrapped_field,
                           DM_SupportPointer wrapped_index ) {
  DM_DMeroonType  *type;
  DM_largest_type thing;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  if ( !(DM_IsSupportFixnum(wrapped_index)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  DMscm_context->errcode = 
    DMeroon_indexed_field_value(DMscm_context,
                                &thing,
                                &type,
                                DM_FromSupport(wrapped_object),
                                DM_FromSupport(wrapped_field),
                                DM_FromSupportFixnum(wrapped_index) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DMscm_to_support(DMscm_context, type, &thing);
}

DM_SupportPointer
DMscm_set_field_value (DM_SupportPointer wrapped_object,
                       DM_SupportPointer wrapped_field,
                       DM_SupportPointer wrapped_value ) {
  DM_DMeroonType  *type;
  DM_largest_type thing1;
  DM_largest_type thing2;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_field_types(DMscm_context,
                        &type,
                        DM_NULL,
                        DM_FromSupport(wrapped_field) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  DMscm_from_support(DMscm_context, type, &thing2, wrapped_value);
  DMscm_context->errcode = 
    DMeroon_set_field_value(DMscm_context,
                            &thing1,
                            DM_FromSupport(wrapped_object),
                            DM_FromSupport(wrapped_field),
                            &thing2);
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DMscm_to_support(DMscm_context, type, &thing1);
}

DM_SupportPointer
DMscm_set_indexed_field_value (DM_SupportPointer wrapped_object,
                               DM_SupportPointer wrapped_field,
                               DM_SupportPointer wrapped_index,
                               DM_SupportPointer wrapped_value ) {
  DM_DMeroonType  *type;
  DM_largest_type thing1;
  DM_largest_type thing2;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  if ( !(DM_IsSupportFixnum(wrapped_index)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  DMscm_context->errcode =
    DMeroon_field_types(DMscm_context,
                        &type,
                        DM_NULL,
                        DM_FromSupport(wrapped_field) );
  DMscm_from_support(DMscm_context, type, &thing2, wrapped_value);
  DMscm_context->errcode =
    DMeroon_set_indexed_field_value(DMscm_context,
                                    &thing1,
                                    DM_FromSupport(wrapped_object),
                                    DM_FromSupport(wrapped_field),
                                    DM_FromSupportFixnum(wrapped_index),
                                    &thing2);
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DMscm_to_support(DMscm_context, type, &thing1);
}

DM_SupportPointer
DMscm_indexed_field_length (DM_SupportPointer wrapped_object,
                            DM_SupportPointer wrapped_field ) {
  DM_DMeroonType  *type;
  DM_largest_type thing;
  DM_DefineStaticContext(DMscm_context);

  if (   !(DM_IsDMeroonInstance(wrapped_object))
      || !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_indexed_field_length(DMscm_context,
                                 &thing,
                                 &type,
                                 DM_FromSupport(wrapped_object),
                                 DM_FromSupport(wrapped_field) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DMscm_to_support(DMscm_context, type, &thing);
}

DM_SupportPointer
DMscm_field_item_type (DM_SupportPointer wrapped_field) 
{
  DM_DMeroonType *itemtyp;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode =
    DMeroon_field_types(DMscm_context,
                        &itemtyp,
                        DM_NULL,
                        DM_FromSupport(wrapped_field) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(itemtyp);
}

DM_SupportPointer
DMscm_field_repetition_type (DM_SupportPointer wrapped_field) 
{
  DM_DMeroonType *reptyp;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsDMeroonInstance(wrapped_field)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  DMscm_context->errcode = 
    DMeroon_field_types(DMscm_context,
                        DM_NULL,
                        &reptyp,
                        DM_FromSupport(wrapped_field) );
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_ToSupport(reptyp);
}

/* RESOURCE DMscm_serve
 * Serve for at least duration seconds. Return #t only if
 * something was served.
 */

DM_SupportPointer
DMscm_serve (DM_SupportPointer wrapped_duration) {
  DM_nat3      duration;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsSupportFixnum(wrapped_duration)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  duration  = DM_FromSupportFixnum(wrapped_duration);
  DMscm_context->errcode = DMeroon_serve(DMscm_context, duration);
  DM_CheckReturnedStatus(DMscm_context, DMscm_context->errcode);
  return DM_SupportDone;
}

/* RESOURCE DMscm_set_site_debug_level
 * Change the debug level of the current site. this function can be invoked
 * from Scheme but is currently not documented.
 */

DM_SupportPointer
DMscm_set_site_debug_level (DM_SupportPointer wrapped_newlevel) 
{
  DM_nat3      newlevel;
  DM_nat3      oldlevel;
  DM_DefineStaticContext(DMscm_context);

  if ( !(DM_IsSupportFixnum(wrapped_newlevel)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_SUPPORT_FIXNUM);
  }
  oldlevel = DM_TheCurrentSite()->debug_level;
  newlevel = DM_FromSupportFixnum(wrapped_newlevel);

#ifdef DM_DEBUG
  /* Modify only in debug mode. */
  DM_TheCurrentSite()->debug_level = newlevel;
#endif /* DM_DEBUG */

  return DM_ToSupportFixnum(oldlevel);
}

/* CAUTION: this function is for debug purpose only. Users must not
 * use it since it breaks all DMeroon encapsulation.
 */

DM_SupportPointer
DMscm_proxy_of (DM_SupportPointer wrapped_o) 
{
  DM_DefineStaticContext(DMscm_context);

#if defined(DM_DEBUG)
  if ( !(DM_IsDMeroonInstance(wrapped_o)) ) {
    DM_RaiseException(DMscm_context, DME_NOT_DMEROON_INSTANCE);
  }
  return DM_ToSupport(DM_ObjectToProxy(DMscm_context, 
                                       DM_FromSupport(wrapped_o) ));
#else
  DM_RaiseException(DMscm_context, DME_NON_DEBUG_MODE);
  return wrapped_o;
#endif

}

/* 
 *                           Initialization.
 * This function has to be called explicitly by the dmXXX.c module.
 * It checks some assertions to foresee problems with the support
 * language. 
 */

void 
DM_scm_initialize (DM_Context* context) {
  
  /* Just to avoid a warning. */
  DM_dmscm_version_string_object = DM_dmscm_version_string_object;

  /* Check various things concerning the port.  */

  DM_DebugAtLevel(0,

    /* Is DM_SupportAddress big enough to hold a pointer?  */

    if ( sizeof(DM_SupportAddress) < sizeof(DM_reference) )
      DM_Abort(context, DME_INCORRECT_SUPPORTADDRESS);

    /* The beginning of a DMeroon instance should be at an address
     * respecting the most stringent alignment.  */

    { DM_SupportAddress q  = (DM_SupportAddress)(DM_ReferenceToClass(Class));
      DM_Object         *r = (DM_Object*)(DM_FromC(&DM_Class_class_object));

      if ( (q) != (DM_OffsetRealign(q)) ) {
        DM_Abort(context, DME_PREFIX_ALIGNMENT_PROBLEM);
      }
      if ( r != ((DM_Object*)q) ) {
        DM_Abort(context, DME_TOC_FROMC_PROBLEM);
      }
    }
);

}

/* }}}
 * end of dmscm.c */
