/* $Id: dmsizes.h,v 1.34 1998/06/10 16:31:30 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

#ifndef DM_SIZES_H
#define DM_SIZES_H

/* DOCUMENTATION dmsizes.h
 * DMeroon is parameterized by the machine on which it runs as seen by
 * the C compiler which is used (this influences sizes and alignment
 * constraints) and also by the language with which it is operated:
 * Bigloo, C, etc.)
 *
 * The following definitions are checked while initializing DMeroon if
 * compiled with the DEBUG option (-DDEBUG=11). In case of problem, all
 * the sizes and alignment constraints of any C or DMeroon types will
 * be printed to help you to figure out what is wrong.
 *
 * DM_MOST_STRINGENT_TYPE is the type with the most stringent
 * alignment constraint. It is usually double float which corresponds
 * to an 8byte chunk which is generally aligned on a 4- or 8-byte
 * address.
 *
 * DM_MOST_STRINGENT_TYPE_OBJECT is the same word with "_object" suffix.
 * I could not succeed in generating it automatically.
 *
 * DM_MOST_STRINGENT_ALIGNMENT_LOG is usually 2 or 3 depending on the most
 * stringent alignment respectively 4 or 8.
 *
 * DM_HAS_LONG_LONG_ENTITIES is defined to be 1 if the hardware and the
 * compiler allows long long (unsigned or not) integers. Gcc can as some other
 * compilers. This is not ISO-C at all but convenient. It must be undefined 
 * otherwise.
 *
 * DM_HAS_UNICODE_ENTITIES is defined to be 1 when unicode characters are
 * implemented. It must be undefined otherwise.
 *
 */

#define DM_MOST_STRINGENT_TYPE            DM_float2
#define DM_MOST_STRINGENT_TYPE_OBJECT     DM_float2_type_object

/* Intel cpu seems to require things to be aligned on quad-words
 * whereas gcc on most machines aligns to oct-words (which is
 * over-constraining). */

#if defined(i386)
#define DM_MOST_STRINGENT_ALIGNMENT_LOG   2
#else
#define DM_MOST_STRINGENT_ALIGNMENT_LOG   3
#endif /* defined(i386) */

/* Gnu C implements long long entities. */

#ifdef __GNUC__
#define DM_HAS_LONG_LONG_ENTITIES         1
#else
#undef DM_HAS_LONG_LONG_ENTITIES
#endif /* __GNUC__ */

/* One day I should enquire about Unicode.                    FUTURE */

#undef DM_HAS_UNICODE_ENTITIES

/* RESOURCE DM_ROOM_FOR_REALIGNMENT
 * This constant is needed when allocating an instance of a stringent
 * class.  Its value is determined by the interaction of malloc, the most
 * stringent alignment (as used by C) and the hardware alignment
 * constraint (not so often the same). For instance on Sony, GC_malloc
 * is 4aligned, gcc uses 8alignment so 4 bytes is sufficient for
 * DM_ROOM_FOR_REALIGNMENT. See DM_allocate_bare_instance C function.
 */

#define DM_ROOM_FOR_REALIGNMENT (1<<(DM_MOST_STRINGENT_ALIGNMENT_LOG/2))

/* RESOURCE DM_SupportAddress
 * This is the type to represent addresses. This should work for all machines
 * since this is the biggest integral type.

 * NOTE: Define it to be a long rather than a DM_nat3 which, on
 * DEC alpha, is not big enough to hold an address. 
 */

#define DM_SupportAddress  unsigned long int

/******************************************************************
 *                    BASIC FIELD TYPES
 * DOCUMENTATION
 * This part depends on the implementation. The implementation must provide
 * C types that are sufficiently large to hold DMeroon values ie:
 * DM_nat1 range is 0 .. 255 ie one byte
 * DM_nat2 range is two bytes
 * DM_nat3 range is four bytes
 * DM_nat4 range is eight bytes (when present)
 * DM_int1 range is -128 .. 127 ie one byte
 * DM_int2 range is two bytes
 * DM_int3 range is four bytes
 * DM_int4 range is eight bytes (when present)
 * DM_float1 range is four bytes 
 * DM_float2 range is eight bytes
 * DM_float3 range is sixteen(?) bytes (when present)
 *
 * DM_reference and DM_data correspond to regular pointers while
 * DM_code corresponds to pointers to C functions. A DM_reference refers
 * to a DMeroon object while a DM_data refers to some data depending
 * on the support language.
 *
 * It is possible not to have all these types defined but to help you
 * to find what the current sizes are as well as if your C compiler
 * supports 8 or 16 bytes entities, you can invoke
 * DM_print_types(NULL). Remember also that this is checked.
 *
 * On many machines long int are similar to int, so to get nat4 use
 * the non ISO long long int type (usually offered by gcc).
 */

typedef void 			DM_nought;
typedef unsigned char 		DM_nat1;
typedef unsigned short int 	DM_nat2;
typedef unsigned int 		DM_nat3;
#ifdef DM_HAS_LONG_LONG_ENTITIES
typedef unsigned long long int	DM_nat4;
#else
typedef struct DM_nat4 {
  DM_nat3 _high;
  DM_nat3 _low;
} DM_nat4;
#endif
typedef void 			*DM_reference;
typedef void 			*DM_data;
typedef DM_reference		(*DM_code)();
typedef signed char 		DM_int1;
typedef short int 		DM_int2;
typedef int 			DM_int3;
#ifdef DM_HAS_LONG_LONG_ENTITIES
typedef long long int 		DM_int4;
#else
typedef struct DM_int4 {
  DM_int3 _high;
  DM_nat3 _low;
} DM_int4;
#endif
typedef char 			DM_char1;
#ifdef DM_HAS_UNICODE_ENTITIES
typedef wchar_t 		DM_unicode;
#endif
typedef float 			DM_float1;
typedef double 			DM_float2;
/* sizeof(long double) = 8 on sun4m, 16 on solaris, 
 * 10 on FreeBSD, 12 on Windows...
 * IF the size of float3 is not 16, then the corresponding DMeroon type
 * will be automatically invalidated. */
typedef long double		DM_float3;

/* This is the type of C functions that handle requests or answers. 
 * Unfortunately, some types are not yet defined. */
#define DM_ErrorCode 	DM_nat3
#define DM_Context   	void
#define DM_Request   	void
typedef DM_ErrorCode	(*DM_handler_code)(DM_Context*, DM_Request*);
#undef DM_Request
#undef DM_Context
#undef DM_ErrorCode

/******************************************************************
 * These definitions depend on the previous ones but are automatic and
 * should not be changed. They appear here rather than in dmeroon.h
 * since they are needed in dmbgl.h and others.
 */

/*
 * A DM_offset encodes three things.
 * To compact it, it is encoded as a nat3 and not as:
 *
 * typedef struct DM_offset {
 *   DM_translation translation;
 *   DM_nat1        repetition;
 *   DM_nat1        item;
 *} DM_offset;
 */

typedef DM_nat3 DM_offset;

/* 
 * DM_translation represents the first subfield of a DM_offset. Since
 * computation on translation must be less restricted, we use the
 * DM_displacement type.
 */

typedef DM_nat2 DM_translation;

typedef DM_nat3 DM_displacement;

/* 
 * Network-ordered unsigned integers have fixed sizes.
 */

typedef DM_nat2 DM_netnat2;
typedef DM_nat3 DM_netnat3;

/* 
 * time are encoded as nat3 and so are ClassOptions, FieldOptions.
 */

typedef DM_nat3 DM_date;
typedef DM_nat3 DM_class_options;
typedef DM_nat3 DM_field_options;
typedef DM_nat3 DM_port_options;
typedef DM_nat3 DM_site_options;
typedef DM_nat3 DM_external_options;
typedef DM_nat3 DM_context_options;
typedef DM_nat3 DM_file_options;
typedef DM_nat3 DM_request_options;

/* DOCUMENTATION Key
 * Key is the type for object unique id.
 * At the beginning, this was a long long ie 8 random bytes. To ensure
 * unicity of oid all around the earth that is, to never reuse the same
 * oid again I added the current date and the ip/portnum of the current
 * server to the oid thus making it 12 bytes long. But to prevent easy
 * oid forgeing, I also added 4 random bytes. This amounts to 16 bytes.
 */

typedef struct DM_key {
  DM_nat2 _idx;
  DM_nat2 _port;
  DM_nat3 _ip;
  DM_nat3 _date;
  DM_nat3 _rand;
} DM_key;

/* We need in some places (dmscm.c mainly) to know the maximal amount
 * of bytes a DMeroon type takes.  */

typedef union DM_largest_type_union {
  DM_nat1 	nat1;
  DM_nat2 	nat2;
  DM_nat3 	nat3;
  DM_nat4 	nat4;
  DM_reference 	reference;
  DM_data	data;
  DM_code	code;
  DM_int1	int1;
  DM_int2	int2;
  DM_int3	int3;
  DM_int4	int4;
  DM_char1      char1;
  DM_nat2 	netnat2;
  DM_nat3 	netnat3;
  DM_float1	float1;
  DM_float2	float2;
  DM_float3	float3;
  DM_offset	offset;
  DM_key	key;
} DM_largest_type;

/******************************************************************
 * These are the names of the underlying C types.
 */

typedef char 			DM_C_char;
typedef signed char 		DM_C_signed_char;
typedef short int 		DM_C_short_int;
typedef int 			DM_C_int;
typedef long int 		DM_C_long_int;
typedef unsigned char 		DM_C_unsigned_char;
typedef unsigned short int 	DM_C_unsigned_short_int;
typedef unsigned int 		DM_C_unsigned_int;
typedef unsigned long int 	DM_C_unsigned_long_int;
typedef float 			DM_C_float;
typedef double 			DM_C_double;
typedef long double 		DM_C_long_double;
typedef void* 			DM_C_void_star;
typedef void                    (*DM_C_function_star)();

#ifdef DM_HAS_LONG_LONG_ENTITIES
typedef unsigned long long int	DM_C_unsigned_long_long_int;
typedef long long int		DM_C_long_long_int;
#endif

#endif /* DM_SIZES_H */

/* end of dmsizes.h */
