;;; $Id: dmstruct.dm,v 1.103 1998/08/23 08:54:18 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; DOCUMENTATION dmstruct.dm
;;; This file defines the predefined classes needed by DMeroon and
;;; described in C source files.  It has to be expanded by a
;;; Scheme-based (case-sensitive) script into a .c and .h files.  It
;;; is easier to define classes, instances and external signatures
;;; with Sexpressions.
;;;
;;; The syntax of that file is compatible with the Meroon Object
;;; system for Scheme.

;;;==================================================================
;;; Classes for dmkernel

;;; Constants are nearly pure objects except that I record the name of
;;; their constant within them to help printing them.

(define-class Constant Object
  ((name	:type char1 :indexing-type nat1)) )

(define-class Type Object
  ((index 	:type nat1)
   (alignment	:type nat1 :local)
   (size	:type nat1 :local)
   (repsize	:type nat1 :local)
   (printer	:type code :local)
   ) :uninstantiable )

(define-class CType Type
  ((dmeroon_type :type reference :local :refer DMeroonType)
   (name 	:type char1 :indexing-type nat1)
   ) :uninstantiable )

(define-class DMeroonType Type
  ((encoder	:type code :local)
   (decoder	:type code :local)
   (dmeroon_class :type reference :refer Class)
   (c_type	:type reference :local :refer CType)
   (name 	:type char1 :indexing-type nat1)
   ) :uninstantiable )

(define-class Class Object
  ((options	:type class_options)
   (super	:type reference :indexing-type nat3 :refer Class)
   (field	:type reference :indexing-type nat3 :refer Field)
   (name 	:type char1 :indexing-type nat1)
   ) :immotile )

(define-class Field Object
  ((options	:type field_options)
   (introducing	:type reference :refer Class)
   (referring	:type reference :refer Class)
   ;; The 'offset' field is not local since part of it must be marshaled.
   (offset	:type offset :indexing-type nat3)
   (name 	:type char1 :indexing-type nat1)
   ) )

;;; Request classes have this class as metaclass, it includes a pointer
;;; to the C function that handles their instances.

(define-class RequestMetaClass Class
  ((handler	:type handler_code :local))
  :immotile )

(define-class String Object
  ((character	:type char1 :indexing-type nat3)
   ) )

(define-class MutableVector Object
  ((item	:type reference :indexing-type nat3 :mutable :refer Object)
   ) )

(define-class ImmutableVector Object
  ((item	:type reference :indexing-type nat3 :refer Object)
   ) )

;;;==================================================================
;;; Classes for dmcomm

;;; DOCUMENTATION External
;;; This is the mother class for all externalized DMeroon objects that is
;;; objects onto which remote pointers may exist. They have a reference 
;;; counter ('counter') and an external identity: an oid ('number' field).
;;;
;;; External objects are used for Requests to be answered and for Entry.

(define-class External Object
  ((number	:type key)
   (counter	:type nat3 :volatile :local)
   (options	:type external_options)
   ) :immotile )

;;; DOCUMENTATION Entry
;;; An Entry allows a DMeroon 'object' to be known from remote sites. If
;;; the object is mutable then the 'clock' field holds its monitoring clock;
;;; it holds DM_nil otherwise. An object which has an Entry (not an Exit)
;;; as proxy is local and remotely pointed. An object which has DM_nil as
;;; proxy is local and not remotely pointed.
;;;
;;;.FIGURE Entry item
;;;                                 an Object
;;;      an Entry                   +----------+
;;;       ---------                 |  class:  |
;;;       \       / <----------------- proxy:  |
;;;        \object: --------------> +----------+
;;;         \   /                   |          |
;;;          \ /                    ............
;;;
;;;.ENDFIGURE

(define-class Entry External
  ((object	:type reference :refer Object)
   (clock	:type reference :volatile :refer Clock)
   ) :immotile )

;;; DOCUMENTATION EntryWithDelayedRequests
;;; This is a variant of Entry that accumulates requests (without serving 
;;; them) until it can be transformed into a real Entry. It is used in the
;;; migration protocol.

(define-class EntryWithDelayedRequests Entry
  ((requests	:type reference :refer QueueLink)
   ) :immotile )

;;; DOCUMENTATION EntryWithClass
;;; This corresponds to a kind of Entry that may refer to an object
;;; that is not a DMeroon object (a C object for instance). The class
;;; must be recorded so it may be exported. This is dangerous and 
;;; currently not used in DMeroon (it might be used in some FUTURE).

(define-class EntryWithClass Entry
  ((class	:type reference                :refer Class)
   ) )

;;; DOCUMENTATION Exit
;;; An Exit item allows to refer to a remote DMeroon object. The object
;;; was externalized as 'number' on site 'from' and this is the site from
;;; which the object came. Its accurate identity is 'number' from 'owner'
;;; (but this information might be false if the object migrated). If 
;;; locally cached, the object's cache is referred with 'object' and if
;;; mutable, the date when it was cached is in 'time' according to 'clock'.

;;; The Exit item may exist with different states. 
;;; The minimal state corresponds to 'number', 'owner' and 'from' fields,
;;; then the  'class' (to decode the bytestream),
;;; then its 'clock' (if the class is mutable),
;;; then its content.
;;; The full state is when the object is locally cached. At that time, 
;;; the cache just needs to be refreshed from time to time.
;;;
;;; INVARIANT: For any DMeroon object:       o._proxy.object == o

;;; FUTURE: reorder these fields to avoid padding.

(define-class Exit EntryWithClass
  ((time	:type nat4      :volatile)
   (from	:type reference                :refer Site)
   (owner	:type reference :volatile      :refer Site)
   (next        :type reference :local :secret :refer Exit)
   ) :immotile )

;;; DOCUMENTATION Clock
;;; A clock monitors the mutations of a set of objects. Note that even if
;;; incremented (by the implementation), a Clock is immutable. This avoids
;;; giving Clocks to Clocks.

(define-class Clock Object
  ((time	:type nat4 :volatile)
   (clocker	:type nat4 :local :volatile)
   ) :immotile )

;;; DOCUMENTATION Hashtable
;;; A hashtable with linear hashing (so there is only one contiguous block
;;; of information to deal with). The 'hasher' function computes the hash
;;; of an item. 'size' records the number of items that are present in the
;;; hashtable. The number of possible elements is the length of the indexed
;;; 'item' field. 'extractor' is NULL or a function to be applied on all
;;; values that are to be extracted from the hashtable (this is often used
;;; to hold the DM_fake_pointer function that prevents the Boehm's GC to 
;;; recognize and keep objects).

(define-class Hashtable Object
  ((hasher	:type code :local)
   (extractor	:type code :local)
   (next	:type reference :refer Hashtable)
   (size	:type nat3 :volatile)
   (item	:type reference :indexing-type nat3 :volatile :refer Object)
  ) )

;;; DOCUMENTATION BoundedStack
;;; This class is used to implement a Bounded stack.
;;; 'index' designates the slot to be filled next so 0 means that the 
;;; stack is empty. Since such an object is only used within the 
;;; implementation to hold internal values, the item field is considered
;;; secret.

(define-class BoundedStack Object
  ((index :type nat3 :volatile)
   (item  :type reference :indexing-type nat3 :volatile :secret :refer Object)
  ) )

;;;==================================================================
;;; Classes for dmextr
;;; ImmutablePair may be used to represent Scheme pairs (but this imposes
;;; an ImmutableLink to refer to an Object instead of an ImmutableLink).

(define-class ImmutableLink Object
  ((rest	:type reference :refer Object)
   ) )

(define-class ImmutablePair ImmutableLink
  ((value	:type reference :refer Object)
   ) )

;;; When an object is read from a Connection, it is enqueued in the
;;; 'objects' field of the current site. The queue is managed in
;;; reverse order, older objects at the end. The queue uses specific
;;; immutable pairs whose cdr is volatile.

(define-class QueueLink Object
  ((rest	:type reference :volatile :refer QueueLink)
   (value	:type reference :refer Object)
   ) )

;;; The Symbol class is immutable. Its instances should be ubiquitous
;;; (and for that are kept in local hahtables). What would mean a
;;; mutable subclass of Symbol ?

(define-class Symbol String
  ( ) :unsubclassable )

(define-class Boolean Object
  () )

(define-class True Boolean
  () )

(define-class False Boolean
  () )

(define-class ImmutableByteVector Object
  ((item	:type nat1 :indexing-type nat3)
   ) )

;;; This is the class for nil.

(define-class Null Object
  () )

;;; All files are identified by such an object (initialized by a
;;; Revision string). They are linked to form a list of version
;;; numbers.

(define-class VersionString Object
  ((next :type reference :refer VersionString)
   (name :type char1 :indexing-type nat1)
   ) )

;;;==================================================================
;;; Classes for dmnet

;;; DOCUMENTATION Port
;;; A Port is an abstract class with the following fields:
;;;   'number' is the associated file descriptor when active.
;;;   'retries' is a not yet used field.
;;;   'options' is a sequence of 32 bits encoding the state of the port
;;;             all errors, problems are reported through this field.

(define-class Port Object
  ((number	:type int2 :volatile)
   (retries	:type nat2)
   (options	:type port_options :volatile)
   ) )

;;; DOCUMENTATION Tcp
;;; A Tcp essentially inherits from Port and adds a sockaddr structure
;;; to it. The four following fields must exactly mimic the sockaddr
;;; structure of the current CPU.

(define-class Tcp Port
  ((family	:type nat2)
   (port	:type netnat2)
   (address	:type netnat3)
   (zero	:type nat4 :local :volatile)
   (last_read	:type date :volatile)
   (last_write	:type date :volatile)
   ) :immotile )

;;; DOCUMENTATION Connection
;;; A Connection corresponds to what is necessary to exchange DMeroon messages
;;; through a TCP/IP channel. 
;;;   'reception' holds the current buffer where messages are received.
;;;   'emission' holds the current buffer where messages to be emitted are
;;;              written. When this buffer is full, a new buffer is allocated
;;;              and chained through it (via its 'next' field).
;;;   'emitqueue' holds the first OutBuffer not yet sent.
;;;   'clocker' tells which clocks to send.
;;;   'site' is a reference to the remote site.
;;;   'protocol' is the list of protocols to use when receiving message headers.

(define-class Connection Tcp
  ((reception	:type reference :volatile :refer InBuffer)
   (emission	:type reference :volatile :refer OutBuffer)
   (emitqueue	:type reference :volatile :refer OutBuffer)
   (clocker	:type nat4 :volatile :local)
   (site	:type reference :refer Site)
   (protocol	:type reference :mutable :refer Protocol)
   (handler	:type code :volatile)
   (password    :type reference :local :volatile :secret :refer String)
   ) :immotile )

;;; DOCUMENTATION Server
;;; A Server is a kind of Tcp that remembers the number of incoming
;;; connections in its 'connected' field and the 'backlog' parameter
;;; (usually always 5 for OS reasons) needed by sockets.

(define-class Server Tcp
  ((connected	:type nat4 :volatile)
   (backlog	:type nat2)
   ) :immotile )

;;; DOCUMENTATION Buffer
;;; A Buffer has the following fields:
;;;    'port' refers to the Connection that may fill (or flush) the buffer.
;;;    'cache' is a MutableVector to cache marshaled objects.
;;;    'stack' is a BoundedStack for marshaled objects.
;;;    'index' refers to the next character to read/write from/in the buffer.
;;;            When 'index' is equal to 'maximum', then the buffer is exhausted
;;;            and must be filled/flushed again. 
;;;    'maximum' refers to the first meaningless character of the buffer
;;;              This character must not be read. 'maximum' is at least 0
;;;              and at most equal to 'byte_length' the number of characters
;;;              the buffer can hold. When flushing a buffer, 'maximum'
;;;              refers to the first unflushed character.
;;;    'byte' the actual zone to hold characters.
;;;
;;; This is an InBuffer.
;;;.FIGURE InBuffer
;;;        +-------------------------------+
;;;   byte |m|e|s|s|a|g|e|?|?|?|?|?|?|?|?|?|
;;;        +-------------------------------+
;;;         0     ^       ^                 ^
;;;               index   maximum           byte_length
;;;.ENDFIGURE
;;;
;;; And this is an OutBuffer.
;;;.FIGURE OutBuffer
;;;        +-------------------------------+
;;;   byte |m|e|s|s|a|g|e|?|?|?|?|?|?|?|?|?|
;;;        +-------------------------------+
;;;         0     ^       ^                 ^
;;;               maximum index             byte_length
;;;.ENDFIGURE

(define-class Buffer Object
  ((port	:type reference :refer Port)
   (cache	:type reference :local :mutable :refer MutableVector)
   (stack	:type reference :local :mutable :refer BoundedStack)
   (mark	:type reference :volatile :refer Marker)
   (index	:type nat3 :mutable)
   (maximum	:type nat3 :mutable)
   (amount	:type nat4 :volatile)
   (next	:type reference :volatile :refer Buffer)
   (byte	:type nat1 :indexing-type nat3 :mutable)
   ) :immotile )

;;; An InBuffer does not use the 'next' field.

(define-class InBuffer Buffer
  () :immotile )

;;; DOCUMENTATION 
;;; An OutBuffer is the head of a linked list of OutBuffers (linked via the 
;;; 'next' field). 

(define-class OutBuffer Buffer
  () :immotile )

;;; DOCUMENTATION Marker
;;; A marker designates a byte in a Buffer. They are used to record positions
;;; or to compute lengths in linked buffers.

(define-class Marker Object
  ((next	:type reference :refer Marker)
   (buffer	:type reference :refer Buffer)
   (index	:type nat3) ) )

;;; DOCUMENTATION Protocol
;;; A Protocol instance defines a linked list of keyword/handler to
;;; decode an incoming packet of information. When a line ending with
;;; a NL is present in an InBuffer and if the keyword matches its
;;; beginning then the handler is triggered to decode the buffer
;;; otherwise the next protocol is tried. The list ends with DM_NULL.
;;; The name field includes a final '\0'.

(define-class Protocol Object
  ((next	:type reference :refer Protocol)
   (handler	:type code :local)
   (name	:type char1 :indexing-type nat1)
   ) )

;;;==================================================================
;;; Classes for dmsite

;;; DOCUMENTATION Site
;;; This huge class describes a Site instance.
;;;  'address' is the IP number of the machine that runs this Server.
;;;  'port'    is the portnumber to use to connect to this Server. (this 
;;;            is not the port really used for the outgoing connection which
;;;            may be different).
;;;  'date'    is the date when the Server started. (this date is
;;;            to be understood relatively to the machine running the
;;;            DMeroon Server. It may also be used as a key to identify
;;;            this Server.
;;;  'name' the official hostname where DMeroon is run.
;;;  'options' is reserved for future use.
;;;  'version' is a float number holding the current DMeroon version number.
;;;          It is actually the only float number of this implementation.
;;;  'support' is a DMeroon string holding the name of the support language.
;;;  'information' holds the head of the multi-level Alist where information
;;;          are published similarly to a ftp site. This Alist is implemented
;;;          by Dictionaries.
;;;  'processid' is the process id of the current process.
;;;          [Does this has a meaning on non-Un*x OS ?]
;;;  'clock' is a reference onto the general (default) Clock of the site.
;;;  'clocker'
;;;  'connection', the connection through which this site may be reached.
;;;          The Connection may be open or closed or even DM_nil.
;;;  'amount' is the total number of bytes sent to the site.
;;;  'exits' the hashtable of all Exit items.
;;;  'entries' the hashtable of all Entry items.
;;;  'ports' the vector of all active Ports.
;;;  'objects' the queue of objects to be sent to that site. For the current
;;;         site, this means the queue of objects to be received (in reverse 
;;;         order).
;;;  'hosts' is the hashtable of all known sites.
;;;  'oblist' the Hashtable of symbols known by this site.
;;;  'protocol' the Protocols recognized by this DMeroon Server.
;;;  'arrived' is the queue of incoming Requests/Answers to be processed.
;;;  'requests' is the stack of pending requests. This is needed when using
;;;         DMeroon in a mono-thread setting.
;;;  'freed' is the chained list of needless Exit items.
;;;         --- The two next fields describe the support language binding.
;;;  'prefixsize' is the size of the support language and DMeroon prefixes.
;;;  'moststringent' refers to the most stringent type.
;;;
;;; RESTRICTION: Site must be an immutable class. 
;;; RESTRICTION: The current site must be a direct instance of Site.
;;; These restrictions are currently required for the establishment of
;;; a Connection.

(define-class Site Object
  ((address	      :type netnat3)
   (port	      :type netnat2)
   (date	      :type date)
   (options	      :type site_options :volatile)
   (processid	      :type nat2)
   (version	      :type float1)
   (versions	      :type reference :refer VersionString)
   (clock	      :type reference :refer Clock)
   (clocker	      :type nat4   :volatile :local)
   (ticker	      :type nat4   :volatile)
   (cyclic_duration   :type nat3   :volatile)
   (normal_duration   :type nat3   :volatile)
   (abnormal_duration :type nat3   :volatile)
   (period            :type nat3   :volatile)
   (debug_level       :type nat3   :volatile :local :secret)
   (compress_threshold  :type nat3 :volatile :local)
   (context        :type reference           :volatile      :refer Context)
   (support	   :type reference                          :refer String)
   (information	   :type reference                          :refer Dictionary)
   (connection	   :type reference :volatile :local         :refer Connection)
   (route	   :type reference :volatile :local         :refer Site)
   (amount	   :type nat4      :volatile)
   (exits	   :type reference :volatile :local :secret :refer Hashtable)
   (entries	   :type reference :volatile :local :secret :refer Hashtable)
   (ports	   :type reference           :local :secret :refer Object)
   (objects	   :type reference :volatile :local :secret :refer QueueLink)
   (handler        :type code      :volatile :local)
   (sites	   :type reference :volatile :local :secret :refer Hashtable)
   (oblist	   :type reference :volatile :local :secret :refer Hashtable)
   (protocol	   :type reference                          :refer Protocol)
   (urlenv	   :type reference :volatile :local         :refer UrlEnv)
   (arrived	   :type reference :volatile :local :secret :refer QueueLink)
   (threads	   :type data                :local :secret)
   (waiter	   :type reference :volatile :local :secret :refer Waiter)
   (freed	   :type reference           :local :secret :refer Exit)
   (prefixsize	   :type nat3)
   (moststringent  :type reference                          :refer DMeroonType)
   (password       :type reference :volatile :local :secret :refer String)
   (name	   :type char1 :indexing-type nat1)
   ) :immotile :uninstantiable )

;;; DOCUMENTATION Dictionary
;;; Information is published via the 'information' field of instances of Site
;;; and structured along Dictionaries a kind of Alist. Note that the two
;;; first fields are mutable.
;;;  'rest' is a Dictionary or DM_nil.
;;;  'value' is a published value or a sub-Dictionary.
;;;  'name' is the name of the published value or sub-Dictionary.
 ;; FUTURE order of fields ? 'value' mutable or volatile ???

(define-class Dictionary Object
  ((rest	:type reference :mutable :refer Dictionary)
   (value	:type reference :mutable :refer Object)
   (name	:type char1 :indexing-type nat1)
   ) :immotile )

;;;==================================================================
;;; All the classes for the DMeroon types.

(define-class Value Object ())

(define-class Number Value ())

(define-class ExactNumber Number ())

(define-class SignedNumber ExactNumber ())

(define-class Int1 SignedNumber
  ((value	:type int1))
  :copyable )

(define-class Int2 SignedNumber
  ((value	:type int2))
  :copyable )

(define-class Int3 SignedNumber
  ((value	:type int3))
  :copyable )

(define-class Int4 SignedNumber
  ((value	:type int4))
  :copyable )

(define-class UnsignedNumber ExactNumber ())

(define-class Nat1 UnsignedNumber
  ((value	:type nat1))
  :copyable )

(define-class Nat2 UnsignedNumber
  ((value	:type nat2))
  :copyable )

(define-class Nat3 UnsignedNumber
  ((value	:type nat3))
  :copyable )

(define-class Nat4 UnsignedNumber
  ((value	:type nat4))
  :copyable )

(define-class InexactNumber Number ())

(define-class Float1 InexactNumber
  ((value	:type float1))
  :copyable )

(define-class Float2 InexactNumber
  ((value	:type float2))
  :copyable )

(define-class Float3 InexactNumber
  ((value	:type float3))
  :copyable )

;;; other classes.

(define-class Char1 Value
  ((value	:type char1))
  :copyable )

(define-class NetNat2 Value
  ((value	:type netnat2))
  :copyable )

(define-class NetNat3 Value
  ((value	:type netnat3))
  :copyable )

(define-class Data Value
  ((value	:type data))
  :copyable )

(define-class Code Value
  ((value	:type code))
  :copyable )

(define-class Reference Value
  ((value	:type reference :refer Object))
  :copyable )

(define-class Date Value
  ((value	:type date))
  :copyable )

(define-class Key Value
  ((value	:type key))
  :copyable )

(define-class ClassOptions Value
  ((value	:type class_options))
  :copyable )

(define-class FieldOptions Value
  ((value	:type field_options))
  :copyable )

(define-class PortOptions Value
  ((value	:type port_options))
  :copyable )

(define-class SiteOptions Value
  ((value	:type site_options))
  :copyable )

(define-class ExternalOptions Value
  ((value	:type external_options))
  :copyable )

(define-class ContextOptions Value
  ((value	:type context_options))
  :copyable )

(define-class FileOptions Value
  ((value	:type file_options))
  :copyable )

(define-class RequestOptions Value
  ((value	:type request_options))
  :copyable )

(define-class HandlerCode Value
  ((value	:type handler_code))
  :copyable )

;;;==================================================================
;;; Request and Answer classes.

;;; DOCUMENTATION Request 

;;; A Request is a subclass of External since sent requests are kept
;;; as Entries until they get their associated answers. All
;;; specialized Requests are subclasses of Request. 
;;;
;;; Among the fields are:
;;;    'requesting' refers to the site waiting for a reply.
;;;    'requested' refers to the site to which the request is sent.

(define-class Request Object
  ( (options            :type request_options)
    (requesting		:type reference :refer Site)
    (requested		:type reference :refer Site)
    ) :metaclass RequestMetaClass )

(define-class SiteConnectionRequest Request
  ( ) :metaclass RequestMetaClass )

(define-class ObjectSendRequest Request
  ( (object		:type reference :refer Object)
    ) :metaclass RequestMetaClass )

(define-class ObjectFetchRequest Request
  ( (object		:type reference :refer Object)
    ) :metaclass RequestMetaClass )

(define-class SiteFetchRequest ObjectFetchRequest
  ( (site        :type reference :local :refer Site)
    ) :metaclass RequestMetaClass )  

;;; This class asks for the mutation of a remote object.

(define-class ObjectMutationRequest Request
  ( (object		:type reference :refer Object)
    (value		:type reference :refer Value :presend)
    (field_index	:type nat3)
    ) :metaclass RequestMetaClass )

(define-class ObjectIndexedMutationRequest ObjectMutationRequest
  ( (index		:type nat3)
    ) :metaclass RequestMetaClass )

(define-class OwnerChangeRequest Request
  ( (object	:type reference :refer Object)
    (entry_num	:type key)
    ) :metaclass RequestMetaClass )

(define-class RemoteOwnerChangeRequest Request
  ( (object	:type reference :refer Object)
    ) :metaclass RequestMetaClass )

;;; Ask a remote site to run an UrlState. 
;;; The output field will be fixed when a remote printing facility is present.

(define-class UrlStateMigrationRequest Request
  ( (urlstate	:type reference :refer UrlState)
    (output	:type reference :refer Object)  ; TEMP ?
    ) :metaclass RequestMetaClass )

;;; DOCUMENTATION Answer
;;; When an Answer is received, its associated request is looked for and
;;; stored in the 'request' field. Normally Answer subclasses match their
;;; Request subclass counterparts. Attention, the 'requesting' and 
;;; 'requested' fields are swapped compared to the matching Request instance.
;;; 'requesting' means who sent the answer while 'requested' means who
;;; should receive it.

(define-class Answer Request
  ( (request	:type reference :refer Request)
    ) :metaclass RequestMetaClass )

;;; DOCUMENTATION SiteConnectionAnswer
;;; The Answer being a subclass of Request already contains the requesting
;;; Site which is the only information asked for by the associated Request.

(define-class SiteConnectionAnswer Answer 
  ()
  :metaclass RequestMetaClass )

(define-class ObjectFetchAnswer Answer
  ( (object		:type reference :refer Object :presend)
    (object_options	:type external_options )
    (time		:type nat4)
    ) :metaclass RequestMetaClass )

(define-class ObjectMutationAnswer Answer
  ( (value	:type reference :refer Value :presend)
    ) :metaclass RequestMetaClass )

(define-class OwnerChangeAnswer Answer
  ( (object	:type reference :refer Object) 
    ) :metaclass RequestMetaClass )

(define-class RemoteOwnerChangeAnswer Answer
  ( ) :metaclass RequestMetaClass )

(define-class FailureAnswer Answer
  ( (errcode	:type nat3)
    ) :metaclass RequestMetaClass )

(define-class UrlStateMigrationAnswer Answer
  ( (urlstate	:type reference :refer UrlState)
    ) :metaclass RequestMetaClass )

;;; All API functions take a Context as first argument.
;;; NOTE: Pay attention that errno are not portable from site to site.
;;;       It is nevertheless recorded to help debugging.
;;; Signalled by Y. Maheo: renamed errno into errnumber since errno is a 
;;; quasi reserved identifier.

(define-class Context Object
  ( (handler		:type code :local     :volatile)
    (options		:type context_options :volatile)
    (errcode		:type nat3            :volatile)
    (file		:type data            :volatile)
    (line		:type nat3            :volatile)
    (errnumber		:type nat3 :local     :volatile)
    (previous           :type reference                 :refer Context)
    (continuation	:type data :local     :volatile)
    (errstream		:type data :local     :volatile)
    ) )

;;; These are the classes to manage queues of waiting processes.

(define-class Waiter Object
  ( (next	:type reference :refer Waiter)
    (end_date	:type date)
    (errcode	:type nat3)
    (options	:type nat3)
    (thread	:type data) ; a non DMeroon entity
    ) )

(define-class FileWaiter Waiter
  ( (fd		:type int2) 
    ) )

(define-class ReceptionWaiter Waiter
  ( ) )

(define-class DurationWaiter Waiter
  ( ) )

(define-class AnswerWaiter Waiter
  ( (request	:type reference :refer Request)
    (answer	:type data)
    ) )

;;; This class represents a delayed computation that must start just
;;; after end_date. 

(define-class EvaluationWaiter Waiter
  () )

;;; Some events of DMeroon are reported to the user via Facts.

(define-class Fact Object
  ( (errcode	:type nat3)
    (item	:type reference :indexing-type nat1 :refer Object)
    ) )

;;; Support for distributed generic functions in Scheme (see the gendist paper)
;;;   arity is the number of arguments if fixed and minus it minus one if
;;;         variant. list has an arity of -1 while cons has an arity of 2.
;;;   profile is a series of 32 bits telling which variables are discriminant.
;;;         No more than 32 variables are allowed for a generic function.
;;;   class is a sequence of classes on which methods are defined.
;;;   method is a sequence of pointers onto methods (normal closures). 
;;;   defaulting is the default method. [Cannot use the word default in C!]
;;;   instruction is a series of numbers that encode the dispatching tree.
;;;   name is a label to identify the generic function.

(define-class Generic Object
  ((arity	:type int1)
   (profile	:type nat3)
   (defaulting  :type reference :refer Object)
   (class	:type reference :refer Class  :indexing-type nat3)
   (method	:type reference :refer Object :indexing-type nat3)
   (instruction :type int3                    :indexing-type nat3)
   (name        :type char1                   :indexing-type nat1)
   ) :copyable )

;;; Methods for this kind of generic functions may be regular Scheme closures.

(define-class SchemeMethod Data () :immotile)

;;; This is the class that encapsulates Scheme continuations.

(define-class SchemeContinuation Data () :immotile)

;;; This object serves to carry invokations across sites. This is not
;;; a DMeroon internal request but a request exchanged between Scheme
;;; sites.

(define-class Invokation Object
  ((function	 :type reference :refer Object)
   (continuation :type reference :refer Object)
   (argument     :type reference :refer Object :indexing-type nat1)
   ) :copyable )

;;; This object serves to represent a file. For instance a source file.

(define-class File Object
  ((options     :type file_options)
   (date        :type date)
   (name  	:type char1 :indexing-type nat2)
   (content     :type char1 :indexing-type nat3) ) )

;;; URLscript language
;;; A command is implemented by an Action that refers to a C function
;;; with signature DM_ErrCode (*action)(context, action, state,
;;; output_port, ...). The internal field is used to hold the definition of
;;; URLscript commands.

(define-class UrlAction Object
  ((behavior	:type code)
   (internal	:type reference :refer Object)
   (name	:type char1     :indexing-type nat1)
   (argument	:type reference :indexing-type nat1 :refer Class) )
  :immotile )

;;; The state of the UrlScript interpreter. 
;;;   previous is the control stack
;;;   script[pc] is the script to interpret
;;;   stack is the arguments stack
;;;   environment is the linked vocabularies of possible commands.

(define-class UrlState Object
  ((previous	:type reference :refer UrlState)
   (script	:type reference :refer String)
   (pc		:type nat3                          :volatile)
   (stack	:type reference :refer BoundedStack :volatile)
   (environment :type reference :refer UrlEnv       :volatile)
   (errcode     :type nat3                          :volatile) 
  ) )

;;; This is the environment, a linked list of vocabularies implemented
;;; as hashtables.

(define-class UrlEnv Hashtable ())

;;; Local Variables:
;;; mode: scheme
;;; End:

;;; end of dmstruct.dm
