/* $Id: dmwrobj.c,v 1.67 1998/04/07 16:27:17 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 
 * DOCUMENTATION dmwrobj.c
 * These functions allow to encode a DMeroon object into a stream of bytes.
 * The produced stream should be decodable by the dmrdobj.c file. 
 * This file is made of highly mutually recursive functions and has proven
 * to be very difficult to get right. 

 * DOCUMENTATION Sending/Migrating objects
 * One must not confuse sending (the reference of) an object and migrating
 * an object. Sending an object lets an other site receive a reference onto
 * an object. Migrating an object means to change its owner site. 
 */

#include "dmeroon.h"

DM_DefineVersionString(dmwrobj,
  "$Id: dmwrobj.c,v 1.67 1998/04/07 16:27:17 queinnec Exp $");

/* {{{ Cache management
 * Utility functions are in dmcache.c
 *
 * RESOURCE DM_already_sent_object
 * Look whether an object is in the cache and emit its reference when
 * present. This is used when sending cyclic objects. 
 * If not in cache, return DM_FALSE. 
 *            FUTURE May also search in the stack.      FUTURE
 */

static DM_c_boolean
DM_already_sent_object (DM_Context   *context,
                        DM_Object    *o, 
                        DM_OutBuffer *outbuffer) {
  DM_MutableVector *cache = outbuffer->cache;
  DM_nat3          index  = DM_is_in_cache(context, o, cache);

  if ( index>=cache->item_length ) {
    /* The object was not in the cache. */
    return DM_FALSE;
  }  
  if ( index <= 255 ) {
    DM_print_byte(context, DM_REFER_WITH_NAT1_COMMAND, outbuffer);
    DM_print_byte(context, index, outbuffer);
    return DM_TRUE;
  }
  if ( index <= 65535 ) {
    DM_print_byte(context, DM_REFER_WITH_NAT2_COMMAND, outbuffer);
    /* Should be DM_send_nat2               FUTURE */
    DM_print_char(context, index/256, outbuffer);
    DM_print_char(context, index%256, outbuffer);
    return DM_TRUE;
  }
  /* Assume the index to be a nat3. */
  DM_print_byte(context, DM_REFER_WITH_NAT3_COMMAND, outbuffer);
  DM_send_nat3(context, outbuffer, &index);
  return DM_TRUE;
} 

/* }}}
 * {{{ Encoding 
 * The main functions are:
 *   DM_write_object
 *                Encodes a predefined, special or regular object.
 *     DM_write_ubiquitous_object
 *                Encodes an object if special (otherwise return FALSE)
 *     DM_write_regular_object
 *                Encodes the content of a regular object (prefixed by class)
 *       DM_write_object_sizes
 *                Encodes the sizes of a regular object
 *       DM_write_object_content
 *                Encodes all the values of the fields of a regular object.
 *         DM_write_object_field_content (static)
 *                Just encodes a single (value of) a field.
 *            DM_write_object_reference
 *                Encode the content of a reference field.
 *          DM_write_exit_reference
 *                Encodes a reference onto an Exit item.
 *          DM_write_site_reference
 *                Encodes a reference onto a Site.
 */

/* RESOURCE DM_write_site_reference
 * Sites are special objects specially encoded. This function encodes any
 * reference to a site (usable or not). It does not encode the content of
 * a site however.
 */

static void
DM_write_site_reference (DM_Context   *context,
                         DM_Site      *site,
                         DM_OutBuffer *outbuffer )
{
  /* Is it the current site ? */
  if ( site == DM_TheCurrentSite() ) {
    DM_print_byte(context, DM_EMITTING_SITE_COMMAND, outbuffer);
    return;
  }
  
  /* Or is it the receiving site ? */
  if ( DM_IsA(context, outbuffer->port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, outbuffer->port);
    
    if ( site == connection->site ) {
      DM_print_byte(context, DM_RECEIVING_SITE_COMMAND, outbuffer);
      return;
    }
  } else {
    /* This is a remote site. Just give the minimum information.
     * The site will be received in the non-usable state so the
     * receiving site will have to fetch the content of the site. */
    DM_nat3  site_ip     = ntohl(site->address);
    DM_nat2  site_port   = ntohs(site->port);
    
    DM_print_byte(context, DM_SITE_COMMAND, outbuffer);
    DM_send_nat3(context, outbuffer, &site_ip);
    DM_send_nat2(context, outbuffer, &site_port);
    DM_send_date(context, outbuffer, &(site->date));
    DM_send_nat1(context, outbuffer, &(site->name_length));
  }
}

/* RESOURCE DM_write_ubiquitous_object
 * Special objects if ubiquitous are specially encoded. If the object to 
 * encode is special, encode it otherwise do nothing and return DM_FALSE.
 */

static DM_c_boolean
DM_write_ubiquitous_object (DM_Context   *context,
                            DM_Object    *o,
                            DM_OutBuffer *outbuffer) {
  DM_PredefinedClassesVector *predefined_classes = DM_ThePredefinedClasses();

  /* Is it the DM_NULL reference ? */
  if ( o == DM_NULL ) {
    DM_print_byte(context, DM_NULL_COMMAND, outbuffer);
    return DM_TRUE;
  }
    
  /* Is it a constant ? */
  if ( DM_IsA(context, o , Constant) ) {
    if ( o == DM_TheConstant(nil) ) {
      DM_print_byte(context, DM_NIL_COMMAND, outbuffer);
      return DM_TRUE;
    }
    if ( o == DM_TheConstant(true) ) {
      DM_print_byte(context, DM_TRUE_COMMAND, outbuffer);
      return DM_TRUE;
    }
    if ( o == DM_TheConstant(false) ) {
      DM_print_byte(context, DM_FALSE_COMMAND, outbuffer);
      return DM_TRUE; 
    }
    if ( o == DM_TheConstant(uninitialized) ) {
      DM_print_byte(context, DM_UNINITIALIZED_COMMAND, outbuffer);
      return DM_TRUE;
    }
  }

  if ( DM_IsA(context, o, Site) ) {
    DM_DeclareNarrowedObject(site, Site, o);

    /* Is it the current site ? */
    if ( site == DM_TheCurrentSite() ) {
      DM_print_byte(context, DM_EMITTING_SITE_COMMAND, outbuffer);
      return DM_TRUE;
    }
  
    /* Or is it the receiving site ? */
    if ( DM_IsA(context, outbuffer->port, Connection) ) {
      DM_DeclareNarrowedObject(connection, Connection, outbuffer->port);
    
      if ( site == connection->site ) {
        DM_print_byte(context, DM_RECEIVING_SITE_COMMAND, outbuffer);
        return DM_TRUE;
      }
    }

    /* This is a different site: it is not ubiquitous. */
    return DM_FALSE; 
  }

  /* Is it a Predefined Class ? */
  if ( DM_IsA(context, o, Class) ) {
    DM_DeclareNarrowedObject(class, Class, o);
    DM_nat1 cn = DM_PredefinedClassNumber(class);

    if ( class == predefined_classes->item[cn] ) {
      DM_print_byte(context, DM_PREDEFINED_CLASS_COMMAND, outbuffer);
      DM_send_nat1(context, outbuffer, &cn);
      return DM_TRUE;
    } else {
      /* Not a predefined class. */
      return DM_FALSE;
    }
  }

  /* Is it a DMeroon Type ? */
  if ( DM_IsA(context, o, DMeroonType) ) {
    DM_DeclareNarrowedObject(type, Type, o);

    DM_print_byte(context, DM_PREDEFINED_TYPE_COMMAND, outbuffer);
    DM_send_nat1(context, outbuffer, &(type->index));
    return DM_TRUE;
  }

  /* Is it the field of a Predefined class ? */
  if ( DM_IsA(context, o, Field) ) {
    DM_DeclareNarrowedObject(field, Field, o);
    DM_Class *class = field->introducing;
    DM_nat1  cn     = DM_PredefinedClassNumber(class);

    if ( class == predefined_classes->item[cn] ) {
      DM_nat3 fn = DM_class_indexed_field_length(context, class);
      DM_nat1 i;

      DM_print_byte(context, DM_PREDEFINED_FIELD_COMMAND, outbuffer);
      DM_send_nat1(context, outbuffer, &cn);
      for ( i=0 ; i<fn ; i++ ) {
        DM_Field *otherfield = DM_class_field(context, class, i);

        if ( field == otherfield ) {
          DM_send_nat1(context, outbuffer, &i);
          break;
        }
      }
      return DM_TRUE;
    } else {
      /* Not a predefined field. */
      return DM_FALSE;
    }
  }

  /* Handle Symbols. */
  if ( DM_IsADirect(context, o, Symbol) ) {
    DM_print_byte(context, DM_SYMBOL_COMMAND, outbuffer);
    DM_write_regular_object(context, o, outbuffer);
    return DM_TRUE;
  }

  /* Use the standard encoding. */
  return DM_FALSE;
}

/* RESOURCE DM_write_exit_reference
 * Write a reference onto a remote object. To avoid always sending a
 * big remote reference, we may use the fact that the remote site is
 * the one at the other end of the Connection and send the lighter
 * immediate reference command.
 */

void
DM_write_exit_reference (DM_Context   *context,
                         DM_Exit      *exit,
                         DM_OutBuffer *outbuffer) {
  DM_Port *port = outbuffer->port;

  /* This is a shortcut to use when the Exit to send is owned or
   * externalized by the receiving site. Just send the external
   * number, this is sufficient to retrieve all other data. */
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);

    if (    ( exit->owner == connection->site ) 
         || ( exit->from == connection->site ) ) {
      DM_print_byte(context, DM_RECEIVER_REFERENCE_COMMAND, outbuffer);
      DM_send_key(context, outbuffer, &(exit->number));
      return;
    }
  }

  /* It may appear that the Exit is reemitted, register that Exit with
   * the same external number if not yet done. This will make it appear
   * as an Entry. */
  if ( DM_HasNotOption(exit, DM_EXTERNAL_IS_EXTERNALIZED) ) {
    DM_register_entry(context, (DM_External*)exit);
  }

  DM_increment_counter(context, (DM_External*)exit);
  DM_print_byte(context, DM_REMOTE_REFERENCE_COMMAND, outbuffer);
  /* Send the key, the 'from' field is implicit by the Connection. */
  DM_send_key(context, outbuffer, &(exit->number));
  DM_send_external_options(context, outbuffer, &(exit->options));
  /* Pre-send some information (class and clock) to save future
   * messages. Pay attention to the fact that the object may not be
   * present. */
  DM_ShortCutReference(context,  &(exit->class));
  DM_write_object_reference(context, (DM_Object*)exit->class, outbuffer);
  /* Send the associated Clock (or null if immutable) */
  DM_ShortCutReference(context, &(exit->clock));
  DM_write_object_reference(context, (DM_Object*)exit->clock, outbuffer);
  /* Send 'owner' field. */
  DM_write_site_reference(context, exit->owner, outbuffer);
}

/* RESOURCE DM_write_object_reference
 * Encode a (remote) reference onto an object. That is don't encode it
 * but just send sufficient information to identify it. Since that object
 * may be the content of a field, it may be an instance of Exit.
 */

void
DM_write_object_reference (DM_Context   *context,
                           DM_Object    *o,
                           DM_OutBuffer *outbuffer) {
  /* Test special case that can be emitted in a few bytes (as well as
   * the DM_NULL reference). If these tests succeed, then the object
   * is emitted. */
  if ( DM_write_ubiquitous_object(context, o, outbuffer) ) {
    /* object was emitted while testing. */
    return;
  } else if ( DM_already_sent_object(context, o, outbuffer) ) {
    /* object was emitted while testing. */
    return;
  } else if ( DM_IsA(context, o, Site) ) {
    DM_write_site_reference(context, (DM_Site*)o, outbuffer);
    return;
  } else if ( DM_IsA(context, o, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, o);
    
    DM_write_exit_reference(context, exit, outbuffer);
    return;
  } else {
    DM_Object *proxy = DM_ObjectToProxy(context, o);

    if ( DM_IsA(context, proxy, Exit)  ) {
      DM_DeclareNarrowedObject(exit, Exit, proxy);

      DM_write_exit_reference(context, exit, outbuffer);
      return;
    } else {
      /* The object is local: send a reference onto an Entry.
       * First externalize it. */
      DM_Entry *entry = DM_externalize(context, o);
      DM_Clock *clock = entry->clock;
      DM_Class *class = DM_ObjectToClass(context, o);

      DM_increment_counter(context, (DM_External*)entry);
      DM_print_byte(context, DM_EMITTER_REFERENCE_COMMAND, outbuffer);
      DM_send_key(context, outbuffer, &(entry->number));
      DM_send_external_options(context, outbuffer, &(entry->options));
      /* Try to send the class if predefined. This will save a message. */
      DM_write_object_reference(context, (DM_Object*)class, outbuffer);
      /* Send a clock (NULL if immutable). Normally the Clock is local. */
      DM_write_object_reference(context, (DM_Object*)clock, outbuffer);
      return;
    }
  }
}

/* RESOURCE DM_write_object_field_content
 * Encode the content of a field depending on its type. Don't recurse
 * on reference fields (unless when pre-sendable).
 */

static void
DM_write_object_field_content (DM_Context   *context,
                               DM_Field     *field,
                               void         *p,
                               DM_OutBuffer *outbuffer ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_nat3   offset        = field->offset[field->offset_length - 1];
  DM_nat1   itemtypindex  = DM_OffsetToItemTypeIndex(offset);
  DM_DMeroonType *itemtyp = implemented_types->item[itemtypindex];

  if ( (   (itemtypindex == DM_data_index)
        || (itemtypindex == DM_code_index) )
       && DM_HasNotOption(context, DM_CONTEXT_DONT_MARSHAL_DATA_CODE_FIELD) ) {
    /* Send the content of a data or code field only if the user
     * specified so. This is only a best effort, nothing is warranted
     * at all (as these types of fields may be very different on the
     * receiving site. */
    DM_send_raw_field(context, itemtyp, p, outbuffer);

  } else if ( itemtypindex == DM_reference_index ) {
    /* Always try to shortcut references before encoding them. */
    DM_ShortCutReference(context, p);
    {
      DM_Object *o = DM_Content(DM_reference, p);
 
      if ( DM_HasOption(field, DM_FIELD_IS_PRE_SENDABLE) ) {
        /* This option forces the marshaling engine to encode the
         * content of the referenced object. This trick (hack?) is
         * only used for some internal DMeroon requests. */
        DM_write_object(context, o, outbuffer);
      } else {
        DM_write_object_reference(context, o, outbuffer);
      }
    }
  } else {
    (itemtyp->encoder)(context, outbuffer, p);
  }
}

/* RESOURCE DM_write_object_content
 * This function emits the content of all the fields of an object. This is
 * a simple iterator on every (non local) fields of an object.
 */

void
DM_write_object_content (DM_Context   *context,
                         DM_Object    *o,
                         DM_OutBuffer *outbuffer ) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class       *class = DM_ObjectToClass(context, o);
  DM_nat3        fn     = DM_class_indexed_field_length(context, class);
  DM_translation t      = 0;
  DM_Field       *field;
  DM_nat3        on;
  DM_offset      last_offset;
  DM_nat1        itemtypindex, reptypindex;
  DM_DMeroonType *itemtyp, *reptyp;
  DM_nat3        i, j, rep;
  void           *p;
  
  for ( i=0 ; i<fn ; i++ ) {
    field        = DM_class_field(context, class, i);
    on           = field->offset_length;
    last_offset  = field->offset[on-1];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp      = implemented_types->item[itemtypindex];

    if ( DM_HasOption(field, DM_FIELD_IS_LOCAL) ) {
      /* Don't encode :local fields. */
      if ( reptypindex == DM_nought_index ) {
        /* This field is not indexed. */
        t = DM_OffsetPad(t, itemtyp);
        t = DM_OffsetExtend(t, itemtyp);
      } else {
        t   = DM_OffsetPad(t, reptyp);
        p   = DM_Displace(o, t);
        rep = DM_read_repetition_factor(context, p, reptypindex);
        t   = DM_OffsetExtend(t, reptyp);
        t   = DM_OffsetPad(t, itemtyp);
        t += rep * itemtyp->repsize;  
        /* Realign after any non final indexed field. */
        t = DM_OffsetRealign(t);      
      }
    } else {
      /* Field is not local. */
      if ( reptypindex == DM_nought_index ) {
        /* This field is not indexed. */
        t = DM_OffsetPad(t, itemtyp);
        p = DM_Displace(o, t);
        DM_write_object_field_content(context, field, p, outbuffer);
        t = DM_OffsetExtend(t, itemtyp);
      } else {
        /* The field is indexed. */
        t   = DM_OffsetPad(t, reptyp);
        p   = DM_Displace(o, t);
        rep = DM_read_repetition_factor(context, p, reptypindex);
        t   = DM_OffsetExtend(t, reptyp);
        t   = DM_OffsetPad(t, itemtyp);

        for ( j=0 ; j<rep ; j++ ) {
          p = DM_Displace(o, t);
          DM_write_object_field_content(context, field, p, outbuffer);
          t += itemtyp->repsize;
        }
        /* Realign after any non final indexed field. */
        t = DM_OffsetRealign(t);
      }
    }
  }
}

/* RESOURCE DM_write_object_sizes
 * Encode the sizes of an object.
 */

void
DM_write_object_sizes (DM_Context   *context,
                       DM_Object    *o,
                       DM_OutBuffer *outbuffer) {
  DM_ImplementedTypesVector *implemented_types = DM_TheImplementedTypes();
  DM_Class          *class = DM_ObjectToClass(context, o);
  DM_nat3           fn     = DM_class_indexed_field_length(context, class);
  DM_translation    t      = 0;
  DM_Field          *field;
  DM_nat3           on;
  DM_offset         last_offset;
  DM_nat1           itemtypindex, reptypindex;
  DM_DMeroonType    *itemtyp, *reptyp;
  DM_nat3           i, rep;
  void              *p;

  /* Send the allocation sizes of the object. */
  for ( i=0 ; i<fn ; i++ ) {
    field        = DM_class_field(context, class, i);
    on           = field->offset_length;
    last_offset  = field->offset[on-1];
    reptypindex  = DM_OffsetToRepetitionTypeIndex(last_offset);
    reptyp       = implemented_types->item[reptypindex];
    itemtypindex = DM_OffsetToItemTypeIndex(last_offset);
    itemtyp      = implemented_types->item[itemtypindex];

    if ( reptypindex == DM_nought_index ) {
      /* This field is not indexed. */
      t = DM_OffsetPad(t, itemtyp);
      t = DM_OffsetExtend(t, itemtyp);
    } else {
      /* The field is indexed, send its size. */
      t   = DM_OffsetPad(t, reptyp);
      p   = DM_Displace(o, t);
      rep = DM_read_repetition_factor(context, p, reptypindex);
      (reptyp->encoder)(context, outbuffer, p);
      t   = DM_OffsetExtend(t, reptyp);
      t   = DM_OffsetPad(t, itemtyp);
      t  += rep * itemtyp->repsize;
      /* Realign after any non final indexed field. */
      t = DM_OffsetRealign(t);
    }
  }
}

/* RESOURCE DM_write_regular_object
 * Encode a general object (prefixed with its class and sizes).
 * This is the standard marshaling procedure.
 */

void
DM_write_regular_object (DM_Context   *context,
                         DM_Object    *o, 
                         DM_OutBuffer *outbuffer) 
{
  DM_Class  *class = DM_ObjectToClass(context, o);
  DM_Object *proxy = DM_ObjectToProxy(context, o);
  DM_Class *cls = DM_NULL;
  DM_class_options clop;

  if ( DM_IsA(context, proxy, Exit) ) {
    DM_DeclareNarrowedObject(exit, Exit, proxy);

    /* The object is not local, just emit a reference onto it.  */
    DM_write_exit_reference(context, exit, outbuffer);
  } else {
    /* Some objects require to be prefixed by a particular command.
     * These are Classes and Fields. Can it be dispensed ??? FUTURE */

    if ( DM_IsA(context, o, Class) ) {
      /* Special case for local non predefined classes.  Memorize the
       * options and reset class usability when marshaling. */
      cls = (DM_Class*)o;
      clop = cls->options;
      DM_print_byte(context, DM_CLASS_COMMAND, outbuffer);
      /* Remove the usability of the class. */
      DM_ClearOption(cls, DM_CLASS_IS_USABLE);

    } else if ( DM_IsA(context, o, Field) ) {
      /* Special case for local non predefined fields. */
      DM_print_byte(context, DM_FIELD_COMMAND, outbuffer);
    }

    /* The object is local. Push it before encoding it and pop it at
     * the end: this makes easier to encode cycles (such as the cycle
     * corresponding to a class since its last super class is
     * itself). */
    if ( DM_IsCopyable(context, o) ) {
      DM_print_byte(context, DM_SET_COPYABLE_COMMAND, outbuffer);
    } else if ( DM_IsSharable(context, o) ) {
      DM_print_byte(context, DM_SET_SHARABLE_COMMAND, outbuffer);
    }

    DM_print_byte(context, DM_PROG2_COMMAND, outbuffer);
    DM_print_byte(context, DM_FILL_COMMAND, outbuffer);
    DM_print_byte(context, DM_PUSH_STACK_COMMAND, outbuffer);
    DM_push_object_in_bounded_stack(context, o, &(outbuffer->stack));
    DM_print_byte(context, DM_ALLOCATE_COMMAND, outbuffer);
    DM_write_object_reference(context, (DM_Object*)class, outbuffer);
    DM_write_object_sizes(context, o, outbuffer);
    DM_write_object_content(context, o, outbuffer);
    DM_print_byte(context, DM_POP_STACK_COMMAND, outbuffer);
    (void) DM_pop_from_bounded_stack(context, outbuffer->stack);
    if ( cls ) {
      /* Restaure class options after marshaling a class. */
      cls->options = clop;
    }
  }
}

/* RESOURCE DM_write_object
 * This function encodes a DMeroon Object into a stream of
 * bytes. Therefore the decoder of that byte-stream will build an
 * equivalent object at the other end of a Connection. It is up to the
 * DMeroon requests and answers to ensure more coherency between
 * sharable objects. 
 */

void
DM_write_object (DM_Context   *context,
                 DM_Object    *o,
                 DM_OutBuffer *outbuffer) 
{
  /* Is it a predefined object (or the DM_NULL reference) ? */
  if ( DM_write_ubiquitous_object(context, o, outbuffer) ) {
    /* NOTHING, marshaling is already done with the test. */
  } else if ( DM_already_sent_object(context, o, outbuffer) ) {
    /* NOTHING, marshaling is already done with the test. */
  } else {
    /* Use the standard encoding. */
    DM_write_regular_object(context, o, outbuffer);
  }
}

/* }}}
 * {{{ Initialize the module.
 */

void
DM_wrobj_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmwrobj);
}

/* }}}
 *  end of dmwrobj.c */
