/* $Id: dmzlib.c,v 1.1 1998/05/06 20:03:17 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* DOCUMENTATION dmzlib.c
 * This file handles the compression primitives of DMeroon,
 * and the use of the zlib library.
 *
 */

#ifdef DM_USE_ZLIB

#include "dmeroon.h"
#include "dmzlib.h"

DM_DefineVersionString(dmzlib,
  "$Id: dmzlib.c,v 1.1 1998/05/06 20:03:17 queinnec Exp $");

/* {{{ zlib interface : compression/decompression primitives
   
 * RESOURCE DM_compress_message 
 
 * Compresses the source buffer(s) into the destination buffer.
 *
 *  Zlib returns Z_OK if success, Z_MEM_ERROR if there was not enough
 *  memory, Z_BUF_ERROR if there was not enough room in the output
 *  buffer, Z_STREAM_ERROR if the level parameter is invalid.
 
 *.FIGURE Structure of a compressed message
 * (SUGGESTION)
 *                  <........... XXXX ...........>
 *        +-----------------------------------------------+
 *   byte |   \n0002XXXXYYYY--------------------\n 
 *        +-----------------------------------------------+
 *              ^   ^   ^   ^                   ^
 *              |   |   |   |                   End of Message
 *              |   |   |   Beginning of compressed message
 *              |   |   Size of the uncompressed message (original size)
 *              |   Size of the compressed whole message
 *              Compressed message subcode

 *.ENDFIGURE

 * NOTE: 
 *  The code for DM_compress_message is based upon : compress.c --
 *  compress a memory buffer Copyright (C) 1995-1998 Jean-loup Gailly.
 *  For conditions of distribution and use, see copyright notice in
 *  zlib.h
 *   
 */


DM_nat3
DM_compress_message(DM_Context   *context,
		    DM_OutBuffer *compressed_buffer,
		    DM_Marker    *marker)
{
  z_stream stream;
  DM_nat3 err;
  DM_nat3 size;
  DM_nat3 orig_size, new_size;
  DM_Buffer *uncompressed_buffer;
  
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
    fprintf(DM_debug_stream, "*Z* Starting compression of a message... \n");
		  );
  /* Initialize the compression stream used by zlib. */
  uncompressed_buffer = marker->buffer;
  /* Skip the first 4 bytes (the size) as it is transmitted elsewhere */
  stream.next_in = 
    &(uncompressed_buffer->byte[(marker->index + 4)]);
  stream.avail_in = 
    (uncompressed_buffer->index) - ((marker->index) + 4);
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
    fprintf(DM_debug_stream, "*Z* Here is the message BEFORE compression : ");  
    DM_print_memory_slice(context, 
		&(uncompressed_buffer->byte[(marker->index +4)]), 
                (uncompressed_buffer->index) - ((marker->index)+4),
		"", "", 
		(DM_SupportPointer)(DM_debug_stream) );
		  );
  
  orig_size = stream.avail_in;

  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, "*Z* Compressing size : %d\n", 
			  stream.avail_in);
		  fprintf(DM_debug_stream,
			  "*Z* maximum : %d, byte_length : %d, index : %d\n",
			  compressed_buffer->maximum,
			  compressed_buffer->byte_length,
			  compressed_buffer->index);
		  );

  /* Except for the first 12 bytes (used as DMeroon internal
   * compressed message headers), and the last one (for the final
   * CR/LF), the whole new OutBuffer is available for the comressed
   * message.  
   */
  
  stream.next_out = 
    &(compressed_buffer->byte[12]);
  stream.avail_out = 
    (compressed_buffer->byte_length) - 13;
  
  /* Memory management functions for zlib */
  /* Memory allocation: */
  stream.zalloc = (alloc_func)DM_zalloc;
  /* Memory de-allocation: */
  stream.zfree = (free_func)DM_zfree;
  /* Opaque parameter transmitted as-is as the first parameter of
   * zalloc and zfree: */
  stream.opaque = (voidpf)context;
  
  /* Initialize compression primitive. */
  err = deflateInit(&stream, Z_DEFAULT_COMPRESSION);
  
if (err != Z_OK) {
  /* Initialization failed for some reason. */
    DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		    err = DM_zdebug_translate(err);
		    fprintf(DM_debug_stream, "*Z* deflateInit failed ! (return code : %d)\n", err);
		    DM_Abort(context, err);
		    );
    DM_Abort(context, DME_INTERNAL_ERROR);
}
  
  /* Actual compression begins here. */
  while (uncompressed_buffer->next) {
    /* This is not the last buffer containing the message to be
     * compressed. Don't flush the stream because this must be the
     * LAST operation done on it. 
     */
    err = deflate(&stream, Z_NO_FLUSH);
    if (err != Z_OK) {
      /* Some error occured during the compression. */
      DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		      err = DM_zdebug_translate(err);
		      fprintf(DM_debug_stream, "*Z* deflate failed ! (return code : %d)\n", err);
		      deflateEnd(&stream);
		      DM_Abort(context, err);
		      );
      deflateEnd(&stream);
      DM_Abort(context, DME_INTERNAL_ERROR);
    }
    
    /* Now process the next OutBuffer. */
    uncompressed_buffer = uncompressed_buffer->next;
    stream.next_in = 
      &(uncompressed_buffer->byte[0]);
    stream.avail_in = 
      (uncompressed_buffer->index);
    orig_size = orig_size + (uncompressed_buffer->index);

    DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		    fprintf(DM_debug_stream, "*Z* deflate once more for %d bytes\n",
			    (uncompressed_buffer->index));
		    );
  }
  
  /* This is the last deflate operation. */
  err = deflate(&stream, Z_FINISH);
  if (err != Z_STREAM_END) {
    /* Some error occured during the compression */
    DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		    err = DM_zdebug_translate(err);
		    fprintf(DM_debug_stream, 
			    "*Z* final deflate failed ! (return code : %d)\n", 
			    err);
		    deflateEnd(&stream);
		    DM_Abort(context, err);
		    );
    deflateEnd(&stream);
    DM_Abort(context, DME_INTERNAL_ERROR);
  }
  
  /* Prepare the header for the message in the compressed_buffer : put
   * the 0002 false size to make DMeroon switch to compressed mode, then
   * the size of the compressed message (including headers and CR/LF)
   * on the first 4 bytes, then the size of the original
   * (uncompressed) message, and last the final CR/LF.
   */
  size = new_size = (stream.total_out + 13);
  
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, "*Z* patching compression header (0002)...\n");
		  );
  
  compressed_buffer->byte[0] = 2;
  compressed_buffer->byte[1] = 0;
  compressed_buffer->byte[2] = 0;
  compressed_buffer->byte[3] = 0;
  
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, "*Z* now patching compressed size to %d...\n", 
			  size);
		  );
  
  compressed_buffer->byte[4] = size % 256;
  size /= 256;
  compressed_buffer->byte[5] = size % 256;
  size /= 256;
  compressed_buffer->byte[6] = size % 256;
  size /= 256;
  compressed_buffer->byte[7] = size % 256;
  
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, "*Z* then patching un-compressed size to %d...\n", 
			  orig_size);
		  );

  compressed_buffer->byte[8] = orig_size % 256;
  orig_size /= 256;
  compressed_buffer->byte[9] = orig_size % 256;
  orig_size /= 256;
  compressed_buffer->byte[10] = orig_size % 256;
  orig_size /= 256;
  compressed_buffer->byte[11] = orig_size % 256;
  
  /* Put the final \n */ 
  compressed_buffer->byte[new_size] = '\n';
    
  /* Set the informations needed by DMeroon about the
   * OutBuffer (index : position of the last CR/LF)
   */ 
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, 
			  "*Z* new index : %d\n", 
			  new_size);
		  );

  compressed_buffer->index = new_size + 1;
  
  /* End of compression : ask for de-allocation...
   * Here, it's the Garbage Collector that will take care of it
   */
  err = deflateEnd(&stream);
  if (err != Z_OK) {
    /* deflateEnd failed for some reason */
    DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		    err = DM_zdebug_translate(err);
		    fprintf(DM_debug_stream, "*Z* deflateEnd failed ! (return code : %d)\n", err);
		    DM_Abort(context, err);
		    return 0;
		    );
    DM_Abort(context, DME_INTERNAL_ERROR);
    return 0;
  }
  
  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, "*Z* Done.\n\n\n");
		  );

  return new_size;
} /* DM_compress_message() */

/* RESOURCE DM_uncompress_message
 *  

 * NOTE:
 *  The code for DM_uncompress_message is based upon :
 *  uncompress.c -- uncompress a memory buffer
 *  Copyright (C) 1995-1998 Jean-loup Gailly.
 *  For conditions of distribution and use, see copyright notice in zlib.h
 *
 */

DM_nat3
DM_uncompress_message(DM_Context  *context,
		      DM_InBuffer *compressed_buffer,
		      DM_InBuffer *uncompressed_buffer,
		      DM_nat3     lf_index)
{
  z_stream stream;
  DM_nat3
    err,
    size, 
    new_size;

DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		fprintf(DM_debug_stream, 
			"*Z* Starting uncompression of a message... ");
		);

  /* Initialize the compression stream used by zlib */
  stream.next_in = 
    &(compressed_buffer->byte[lf_index + 13]);
  stream.avail_in = 
    (compressed_buffer->index) - (lf_index + 13);

DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		fprintf(DM_debug_stream, "size : %d\n", stream.avail_in);
		fprintf(DM_debug_stream,
			"*Z* compressed_buffer->maximum : %d, length : %d, index : %d\n",
			compressed_buffer->maximum,
			compressed_buffer->byte_length,
			compressed_buffer->index);
		);

  stream.next_out = 
    &(uncompressed_buffer->byte[0]);
  stream.avail_out = 
    (uncompressed_buffer->byte_length) - 1;

  DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		  fprintf(DM_debug_stream, 
			  "*Z* Initializing memory management primitives.\n");
		  );
  
  /* Memory management functions for zlib */
  /* Memory allocation: */
  stream.zalloc = (alloc_func)DM_zalloc;
  /* Memory de-allocation: */
  stream.zfree = (free_func)DM_zfree;
  /* Opaque parameter transmitted as-is as the first parameter
   * of zalloc and zfree:
   */
  stream.opaque = (voidpf)context;
  
  /* Initializing uncompression primitive */
  err = inflateInit(&stream);
  if (err != Z_OK) {
    /* Initialization failed for some reason */  
    DM_DebugAtLevel(DM_CACHE_DEBUG_LEVEL,
		    err = DM_zdebug_translate(err);
		    fprintf(DM_debug_stream, 
			    "*Z* inflateInit failed ! (return code : %d)\n", 
			    err);
		    DM_Abort(context, err);
		    );
    DM_Abort(context, DME_INTERNAL_ERROR);
  }
  
  /* Actual uncompression begins here */
  err = inflate(&stream, Z_FINISH);
  if (err != Z_STREAM_END) {
    /* Some error occured during the uncompression */
    DM_DebugAtLevel(DM_CACHE_DEBUG,
		    err = DM_zdebug_translate(err);
		    fprintf(DM_debug_stream, 
			    "*Z* inflate failed ! (return code : %d)\n", 
			    err);
		    inflateEnd(&stream);
		    DM_Abort(context, err);
		    );
    inflateEnd(&stream);
    DM_Abort(context, DME_INTERNAL_ERROR);
  }
  
  new_size = size = (stream.total_out + 1);
  
DM_DebugAtLevel(DM_CACHE_DEBUG,
		fprintf(DM_debug_stream, "*Z* decompressed %d bytes.\n", 
			stream.total_out);
		fprintf(DM_debug_stream, "*Z* CR LF at : %d\n", 
			size);
		);

  uncompressed_buffer->byte[size] = '\n';
  
  uncompressed_buffer->maximum = size;
  uncompressed_buffer->index = 0;

DM_DebugAtLevel(DM_CACHE_DEBUG,
		 fprintf(DM_debug_stream, "*Z* Done.\nMessage AFTER uncompression is : \n");
		 DM_print_memory_slice(context, 
				       &(uncompressed_buffer->byte[0]), 
				       size, 
				       "", "", 
				       (DM_SupportPointer)(DM_debug_stream) );
		 );
  
  return size;
}  /* DM_uncompress_message () */

DM_nat3
DM_zdebug_translate(DM_nat3 error_code)
{
  switch(error_code)
    {
    case Z_OK:
      return DME_Z_OK;
    case Z_STREAM_END:
      return DME_Z_STREAM_END;
    case Z_NEED_DICT:
      return DME_Z_NEED_DICT;
    case Z_STREAM_ERROR:
      return DME_Z_STREAM_ERROR;
    case Z_DATA_ERROR:
      return DME_Z_DATA_ERROR;
    case Z_MEM_ERROR:
      return DME_Z_MEM_ERROR;
    case Z_BUF_ERROR:
      return DME_Z_BUF_ERROR;
    case Z_VERSION_ERROR:
      return DME_Z_VERSION_ERROR;
    default:
      return DME_Z_ERRNO;
    }
}

/* }}}
 * {{{ zlib interface : memory management primitives
 */  
   
/* RESOURCE DM_zalloc
 *
 */

DM_String *
DM_zalloc(DM_Context *context,
	  DM_nat3 p1,
	  DM_nat3 p2)
{
  DM_DebugAtLevel(DM_CACHE_DEBUG,
		  fprintf(DM_debug_stream, 
			  "*Z* call to zalloc, size : %d * %d = %d.\n", 
			  p1, 
			  p2, 
			  (p1 * p2));
		  );
  return DM_Allocate1(context, String, (p1 * p2));
  /* err ?? */
}

/* RESOURCE DM_zfree
 *
 */
    
void
DM_zfree(DM_Context *context,
	 DM_Object o)
{
  /* gc */
  DM_DebugAtLevel(DM_CACHE_DEBUG,
		  fprintf(DM_debug_stream, "*Z* call to zfree... (NOP)\n");
		  );
  return;
}

/* }}}
 * {{{ initialization
 */

/*
 * Initializes the module.
 */

void
DM_zlib_initialize (DM_Context *context) 
{
  DM_LinkVersionStringIntoSite(dmzlib);
}

/* }}} */

#endif /* DM_USE_ZLIB */

/* end of dmzlib.c */
