;;; $Id: dmeroonet.scm,v 1.9 1997/10/03 21:12:05 queinnec Exp $

;;; This file defines an object system on top of DMeroon. It provides
;;; the usual three defining macros: DM-define-class,
;;; DM-define-generic and DM-define-method. It has many intended
;;; similarities with MeroonV3 and meroonet.  The least I can say is
;;; that the current implementation does not favour execution speed.
;;; There are also some ugly things I want to change (naming classes
;;; rather than evaluating them, true support for multi-methods,
;;; testing remote invokation of methods ....).

;;; This file suppose define-dmeroonet-macro to be a macro to define macros.

;;; {{{ Utilities

;;; This variable holds the last class defined at macroexpand-time. It
;;; avoids to rebuild the class at evaluation-time if immediately
;;; after as under a toplevel loop. It is useful for separate
;;; compilation (see the DM-define-class macro).

(define *last-defined-class* #f)

;;; Take a list of symbols and return a symbol with the concatenated
;;; name.  Useful to create derived names.

(define (symbol-concatenate . names)
  (string->symbol 
   (apply string-append 
          (map (lambda (thing)
                 (cond 
                  ((symbol? thing) (symbol->string thing))
                  ((string? thing) thing)
                  (else #f) ) )
               names ))) )

;;; Check whether a field is indexed or not.

(define (Poly-Field? field)
  (not (eq? (DM-field-repetition-type field)
            DM-nought-type )) )

;;; Retrieve the superclass of a class.

(define (DM-Class-superclass class)
  (let ((depth (DM-indexed-field-length class DM-Class-super-field)))
    (DM-indexed-field-value class DM-Class-super-field (- depth 2)) ) )

;;; }}}
;;; {{{ Names to values
;;; These functions convert names into values such as classes, types ...

;;; Convert a name into a class. Not all classes are there, only the
;;; ones that are known ie defined in the same session. With an
;;; additional class, register it with its name.

(define find-class 
  (let ((*classes*
         (list (cons 'Object DM-Object-class)
               (cons 'Class  DM-Class-class)
               (cons 'Field  DM-Field-class)
               ) ) )
    (lambda (name . class)
      (if (pair? class)
          ;; register name -> class
          (set! *classes*
                (cons (cons name (car class)) *classes*) )
          (let ((p (assq name *classes*)))
            (if (pair? p)
                (cdr p)
                (dmeroonet-error "Unknown class" name) ) ) ) ) ) )

;;; Convert a name into a type. Not all types are present here but
;;; they could be since they are all statically known.

(define find-type 
  (let ((*types*
         (list (cons 'nought   DM-nought-type)
               (cons 'nat1     DM-nat1-type)
               (cons 'nat2     DM-nat2-type)
               (cons 'nat3     DM-nat3-type)
               (cons 'nat4     DM-nat4-type)
               (cons 'reference DM-reference-type)
               (cons 'data     DM-data-type)
               (cons 'code     DM-code-type)
               (cons 'int1     DM-int1-type)
               (cons 'int2     DM-int2-type)
               (cons 'int3     DM-int3-type)
               (cons 'int4     DM-int4-type)
               (cons 'char1    DM-char1-type)
               (cons 'netnat2  DM-netnat2-type)
               (cons 'netnat3  DM-netnat3-type)
               (cons 'float1   DM-float1-type)
               (cons 'float2   DM-float2-type)
               (cons 'float3   DM-float3-type) )))
    (lambda (name)
      (let ((p (assq name *types*)))
        (if (pair? p)
            (cdr p)
            (dmeroonet-error "Unknown type" name) ) ) ) ) )

;;; Associate functions and generic records.

(define find-generic
  (let ((*generics* (list)))
    (lambda (g . generic)
      (if (pair? generic)
          (set! *generics* 
                (cons (cons g (car generic)) *generics*) )
          (let ((p (assq g *generics*)))
            (if (pair? p)
                (cdr p)
                (dmeroonet-error "Unknown generic function" g) ) ) ) ) ) )

;;; }}}
;;; {{{ Class definition
;;; Syntax: (DM-define-class <name-of-the-class> <name-of-the-superclass>
;;;             ( <name-of-a-Mono-Field>          |
;;;               (= <name-of-a-Mono-Field>)      |
;;;               (* <name-of-a-Poly-Field>)      |
;;;               (<name-of-a-Field> <DMeroon field options>)
;;;               ... ) <DMeroon class options> )

(define-dmeroonet-macro (DM-define-class name supername 
                                         own-field-descriptions
                                         . class-options )
  (DM-process-define-class 
   name supername own-field-descriptions class-options ) )

(define (DM-process-define-class name supername 
                                 own-field-descriptions
                                 class-options )
  (set! *last-defined-class* #f)
  (let ((class (DM-register-class name 
                                  supername 
                                  own-field-descriptions
                                  class-options )))
    (set! *last-defined-class* class)
    `(begin
       (if (not *last-defined-class*)
         (set! *last-defined-class*
               (DM-register-class ',name 
                                  ',supername
                                  ',own-field-descriptions
                                  ',class-options ) ) )
      ,(DM-Class-generate-related-names class)
      ',name ) ) )

;;;    {{{ Naming global variables
;;; These functions define the names of the variables that will hold the
;;; values (mostly functions) that accompany a class creation.

(define (compute-class-variable-name cname)
  (symbol-concatenate cname '-class) )
(define (compute-predicate-variable-name cname)
  (symbol-concatenate cname '?) )
(define (compute-maker-variable-name cname)
  (symbol-concatenate 'make- cname) )
(define (compute-allocator-variable-name cname)
  (symbol-concatenate 'allocate- cname) )
(define (compute-reader-variable-name cname fname)
  (symbol-concatenate cname '- fname) )
(define (compute-writer-variable-name cname fname)
  (symbol-concatenate 'set- cname '- fname '!) )
(define (compute-lengther-variable-name cname fname)
  (symbol-concatenate cname '- fname '-length) )
;;; }}}

;;; Generate from a class all the definitions of the related functions.
;;; As in Meroon V3 the Class instance is stored in a global variable
;;; whose name is computed by compute-class-variable-name.

(define (DM-Class-generate-related-names class)
  (let* ((name (DM->string class DM-Class-name-field))
         (class-variable-name (compute-class-variable-name name))
         (predicate-name (compute-predicate-variable-name name))
         (maker-name (compute-maker-variable-name name))
         (allocator-name (compute-allocator-variable-name name)) )
    `(begin 
       (define ,class-variable-name *last-defined-class*)
       (define ,predicate-name (DM-make-predicate ,class-variable-name))
       (define ,maker-name (DM-make-maker ,class-variable-name))
       (define ,allocator-name (DM-make-allocator ,class-variable-name))
       ,@(let ((fn (DM-indexed-field-length class DM-Class-field-field))
               (accessors (list)) )
           (do ((i 0 (+ i 1)))
               ((>= i fn) (reverse! accessors))
             (let ((field (DM-indexed-field-value 
                           class DM-Class-field-field i )))
               (set! accessors
                     (cons (DM-Field-generate-related-names field class i)
                           accessors ) ) ) ) )
       #f ) ) )

;;;    {{{ Predicate
;;; Create from a class its appropriate recognizer.

(define (DM-make-predicate class)
  (lambda (o) (and (DM-is-object? o)
                   (DM-is-a? o class) )) )

;;; }}} 
;;;    {{{ Allocation
;;; Create from a class its appropriate allocator. Sizes must be a
;;; list of positive integers.

(define (DM-make-allocator class)
  (lambda sizes
    (apply DM-allocate class sizes) ) )

;;; Create from a class an appropriate maker. The strategy is to create the
;;; vector then to check if it has an appropriate skeleton. Improve
;;; the creation of instances without poly-fields ie do not cons.
;;; Restriction: Do not check if there are too much values given for
;;; initialization, they are ignored.

(define (DM-make-maker class)
  (lambda parms
    (sort-parameters
     class
     parms
     (lambda (sizes values)
       (define (pop-value!)
         ;; Don't check for car! sort-parameters ensure this is correct.
         (let ((v (car values)))
           (set! values (cdr values))
           v ) )
       ;; create the instance
       (let ((o (apply DM-allocate class sizes)))
         (DM-iterate o
                     (lambda (o field itemtyp)
                       (DM-set-field-value! o field (pop-value!)) )
                     (lambda (o field reptyp itemtyp)
                       #f )
                     (lambda (o field itemtyp j)
                       (DM-set-indexed-field-value! o field j (pop-value!)) ) )
         o ) ) ) ) )

;;; Sort the parameters to differentiate the sizes and the values.
;;; FUTURE: It is possible to do it without reverse and without consing.

(define (sort-parameters class parms k)
  (let ((fn (DM-indexed-field-length class DM-Class-field-field))
        (sizes '())
        (values '()) )
    (define (get-value!)
      (if (pair? parms)
          (let ((v (car parms)))
            (set! values (cons v values))
            (set! parms (cdr parms))
            v )
          (dmeroonet-error "Missing value") ) )
    (do ((i 0 (+ i 1)))
        ((>= i fn) (k (reverse! sizes) (reverse! values)))
      (if (Poly-Field? (DM-indexed-field-value class DM-Class-field-field i))
          ;; The field is indexed
          (if (pair? parms)
              (if (integer? (car parms))
                  (let ((j (car parms)))
                    (set! sizes (cons j sizes))
                    (set! parms (cdr parms)) ; remove size
                    (do ((k 0 (+ k 1)))
                        ((>= k j))
                      (get-value!) ) )
                  (dmeroonet-error "Not a legal size" (car parms)) )
              (dmeroonet-error "Missing size") )
          ;; The field is not indexed
          (get-value!) ) ) ) )

;;; }}}
;;;    {{{ Accessor
;;; Generate from a field all the definitions of the related accessors.
;;; An instance of Field refers to the class that introduced it. 

;;; Names are constructed with the current class rather than the
;;; defining class, that is why the class is also given to this
;;; function. The lengther is also generated in case of a Poly-Field.

(define (DM-Field-generate-related-names field class rank)
  (define (is-field-secret? field)
    (let ((options (DM-field-value field DM-Field-options-field)))
      (>= (modulo options 32) 16) ) )
  (define (is-field-mutable? field)
    (let ((options (DM-field-value field DM-Field-options-field)))
        (>= (modulo options 2) 1) ) )
  (let* ((fname (DM->string field DM-Field-name-field))
         (cname (DM->string class DM-Class-name-field))
         (cname-variable (compute-class-variable-name cname)) )
    `(begin 
       ,@(if (is-field-secret? field)
             `()
             `((define ,(compute-reader-variable-name cname fname)
                 (DM-make-reader 
                  (DM-indexed-field-value ,cname-variable 
                                          DM-Class-field-field
                                          ,rank ) ) )) )
       ,@(if (is-field-mutable? field)
             `((define ,(compute-writer-variable-name cname fname)
                 (DM-make-writer 
                  (DM-indexed-field-value ,cname-variable  
                                          DM-Class-field-field
                                          ,rank ) ) ))
             `() )
       ,@(if (Poly-Field? field)
             `((define ,(compute-lengther-variable-name cname fname)
                 (DM-make-lengther
                  (DM-indexed-field-value ,cname-variable  
                                          DM-Class-field-field
                                          ,rank ) ) ))
             '() ) ) ) )

;;; Create the definition of a field reader.

(define (DM-make-reader field)
  (if (Poly-Field? field)
      (lambda (o i)
        (DM-indexed-field-value o field i) )
      (lambda (o)
        (DM-field-value o field) ) ) )

;;; Create the definition of field writer for mutable fields only.

(define (DM-make-writer field)
  (if (Poly-Field? field)
      (lambda (o i v)
        (DM-set-indexed-field-value! o field i v) )
      (lambda (o v)
        (DM-set-field-value! o field v) ) ) )

;;; Create the definition of a lengther.

(define (DM-make-lengther field)
  (lambda (o)
    (DM-indexed-field-length o field) ) )

;;; }}}
;;;    {{{ Class initialization
;;; Create a class instance and graft it into the inheritance hierarchy.

(define (DM-register-class name supername own-field-descriptions class-options)
  (let subclass ((class (find-class supername))
                 (ofd   own-field-descriptions) )
    (if (pair? ofd)
        (cond 
         ((pair? (car ofd))
          (case (caar ofd)
            ((=) 
             (subclass (enrich class 
                            `(,(cadar ofd)
                              :mutable
                              :type data
                              :refer Object ) )
                    (cdr ofd) ) )
            ((*) 
             (subclass (enrich class 
                            `(,(cadar ofd)
                              :mutable
                              :type data
                              :indexing-type nat3
                              :refer Object ) )
                    (cdr ofd) ) )
            (else (subclass (enrich class (car ofd))
                            (cdr ofd) )) ) )
         ((symbol? (car ofd))
          (subclass (enrich class 
                            `(,(car ofd)
                              :mutable
                              :type data
                              :refer Object ) )
                    (cdr ofd) ) )
         (else (dmeroonet-error "Weird field description" (car ofd))) )
        ;; should take care of class-options
        (let ((c (DM-smash-class class 
                                 (symbol->string name)
                                 (decode-class-options class-options)
                                 (length own-field-descriptions) ) ))
          (find-class name c)
          c ) ) ) )

;;; Decode class options.

(define (decode-class-options class-options)
  (if (pair? class-options)
      (case (car class-options)
        ((:mutable)
         (+ 1 (decode-class-options (cdr class-options))) )
        ((:immotile)
         (+ 2 (decode-class-options (cdr class-options))) )
        ((:stringent)
         (+ 4 (decode-class-options (cdr class-options))) )
        ((:volatile)
         (+ 8 (decode-class-options (cdr class-options))) )
        ((:unsubclassable)
         (+ 64 (decode-class-options (cdr class-options))) ) 
        ((:copyable)
         (+ 128 (decode-class-options (cdr class-options))) ) 
        ((:uninstantiable)
         (+ 256 (decode-class-options (cdr class-options))) ) 
        (else (dmeroonet-error "Unknown class option"
                               (car class-options) )) )
      0 ) )

;;; Parse field description

(define (parse-field-description field-description k)
  (let ((field-name "")
        (options (list))
        (itemtyp #f)
        (reptyp  #f)
        (referred-class #f) )
    ;; Take care of the name of the field
    (cond ((string? (car field-description))
           (set! field-name (car field-description)) )
          ((symbol? (car field-description))
           (set! field-name (symbol->string (car field-description))) )
          (else (dmeroonet-error "Bad field name" (car field-description))) )
    ;; Process other field options
    (let parse ((fd (cdr field-description)))
      (if (pair? fd)
          (case (car fd)
            ((:mutable :volatile :local :presend :secret) 
             (set! options (cons (car fd) options))
             (parse (cdr fd)) )
            ((:type)
             (if itemtyp
                 (dmeroonet-error ":type redefined" fd)
                 (if (pair? (cdr fd))
                     (set! itemtyp (find-type (cadr fd)))
                     (dmeroonet-error "Missing type" fd) ) )
             (parse (cddr fd)) )
            ((:indexing-type)
             (if reptyp
                 (dmeroonet-error ":indexing-type redefined" fd)
                 (if (pair? (cdr fd))
                     (set! reptyp (find-type (cadr fd)))
                     (dmeroonet-error "Missing repetition type" fd) ) )
             (parse (cddr fd)) )
            ((:refer)
             (if referred-class
                 (dmeroonet-error ":type redefined" fd)
                 (if (pair? (cdr fd))
                     (set! referred-class (find-class (cadr fd)))
                     (dmeroonet-error "Missing class" fd) ) )
             (parse (cddr fd)) )
            (else (dmeroonet-error "Unknown field descriptor" (car fd))) )
          ;; Set default if unspecified
          (begin
            (if (not itemtyp)
                (set! itemtyp DM-reference-type) )
            (if (not reptyp)
                (set! reptyp DM-nought-type) )
            (if (not referred-class)
                (set! referred-class DM-Object-class) )
            (k field-name (decode-field-options (reverse! options) )
               itemtyp reptyp referred-class ) ) ) ) ) )

;;; Decode field description and return a subclass with that field.

(define (enrich class field-description)
  (parse-field-description
   field-description
   (lambda (field-name field-options itemtyp reptyp referred-class)
     (check-duplicated-name class field-name)
     (DM-create-subclass class "" 0 
                         field-name field-options
                         reptyp itemtyp 
                         referred-class ) ) ) )

;;; Check if a field-name was already used in a super-class.

(define (check-duplicated-name class field-name)
  (let ((n (DM-indexed-field-length class DM-Class-field-field)))
    (do ((i 0 (+ i 1)))
        ((>= i n))
      (let ((field (DM-indexed-field-value class DM-Class-field-field i)))
        (if (equal? field-name (DM->string field DM-Field-name-field))
            (dmeroonet-error "Duplicated field name" field-name) ) ) ) ) )

;;; Decode field options

(define (decode-field-options field-options)
  (if (pair? field-options)
      (case (car field-options)
        ((:mutable)
         (+ 1 (decode-field-options (cdr field-options))) )
        ((:volatile)
         (+ 2 (decode-field-options (cdr field-options))) )
        ((:local)
         (+ 4 (decode-field-options (cdr field-options))) )
        ((:presend)
         (+ 8 (decode-field-options (cdr field-options))) )
        ((:secret)
         (+ 16 (decode-field-options (cdr field-options))) )
        (else (dmeroonet-error "Unknown field option"
                               (car field-options) )) )
      0 ) )

;;; }}}
;;; }}}

;;; {{{ Generic functions
;;; Unlike MeroonV3, there is an explicit operator to invoke distributed
;;; generic functions: DM-invoke. You may wrap it inside a function but
;;; this is not automatically done. 

;;; Define a named generic function. 

(define-dmeroonet-macro (DM-define-generic call . body)
  `(define ,(car call)
     (DM-generic ,call . ,body) ) )

;;; DM-generic creates an anonymous generic function.

(define-dmeroonet-macro (DM-generic call . body)
  (DM-process-generic call body) )

(define (DM-process-generic call body)
  (parse-variable-specifications
   (cdr call)
   (lambda (discriminants variables)
     (check-discriminants! discriminants)
     `(DM-register-generic
       ',(car call)                     ; label
       ',(cdr call)                     ; signature
       (lambda (,(car call) . ,(flat-variables variables)) ; default-method
         ,(if (pair? body)
              `(begin . ,body)
              `(dmeroonet-error 
                "No method" ',(car call)
                . ,(flat-variables variables) ) ) ) ) ) ) )

;;; There should be at least one discriminant.

(define (check-discriminants! discriminants)
  (if (pair? discriminants)
      #t
      (dmeroonet-error 'DM-generic "No discriminant") ) )

;;; Parse a list of specifications, extract all the variables and the
;;; discriminating variable among them. 
;;; Restriction: Silently ignore if there is more than one
;;; discriminant variable ie multi-methods are not implemented.

(define (parse-variable-specifications specifications k)
  (if (pair? specifications)
      (parse-variable-specifications
       (cdr specifications)
       (lambda (discriminant variables)
         (if (pair? (car specifications))
             (k (car specifications) 
                (cons (caar specifications) variables) )
             (k discriminant (cons (car specifications) variables)) ) ) )
      (k #f specifications) ) )

;;; Flatten a list of variables to avoid a final dotted variable. This
;;; is necessary since all methods (even the default method) are
;;; represented by closures with a fix arity.

(define (flat-variables variables)
  (if (pair? variables)
      (cons (car variables) (flat-variables (cdr variables)))
      (if (null? variables) variables (list variables)) ) )

;;; Compute arity of a function (taking into account dotted variable).
;;; For example, cons will have arity 2, while list will have -1.

(define (compute-arity signature)
  (let count ((s signature)
              (arity 0) )
    (if (pair? s)
        (count (cdr s) (+ 1 arity))
        (if (null? s)
            arity
            (- (+ arity 1)) ) ) ) )

;;; Check arity and raise error if disagreement.

(define (check-arity! generic arguments)
  (let ((arity (DM-field-value generic DM-Generic-arity-field)))
    (if (if (>= arity 0)
            (= arity (length arguments))
            (>= (length arguments) (- (+ arity 1))) )
        #t
        (dmeroonet-error 'DM-invoke "Wrong arity for generic function"
                         generic arguments ) ) ) )

;;; Compute the profile telling which variables are discriminant.
;;; This profile is stored as a nat3 number: a 1 bit signals the
;;; corresponding variable to be discriminant.

(define (compute-profile signature)
  (if (pair? signature)
      (if (pair? (car signature))
          (+ 1 (* 2 (compute-profile (cdr signature))))
          (* 2 (compute-profile (cdr signature))) )
      0 ) )

;;; Extract discriminant values.

(define (extract-discriminants generic arguments)
  (let extract ((profile (DM-field-value generic DM-Generic-profile-field))
                (arguments arguments) )
    (let* ((nq (quotient profile 2))
           (others (if (= nq 0) ; no more discriminating variables.
                       (list)
                       (extract nq (cdr arguments)) ) ) )
      (if (= 1 (modulo profile 2))
        (cons (car arguments) others)
        others ) ) ) )

;;; Create and register an instance of Generic.
;;; FUTURE: store the defaulting method in the first slot of methods. FUTURE

(define (DM-register-generic generic-name signature default)
  (let* ((label (symbol->string generic-name))
         (label-length (string-length label))
         (arity (compute-arity signature))
         (profile (compute-profile signature))
         (g (DM-allocate DM-Generic-class
                         0 ; class
                         1 ; method
                         1 ; instruction
                         (+ 1 label-length) )) )
    (DM-set-field-value! g DM-Generic-arity-field arity)
    (DM-set-field-value! g DM-Generic-profile-field profile)
    ;; Only one method: the default one.
    (DM-set-indexed-field-value! 
     g DM-Generic-method-field 0 
     (let ((m (DM-allocate DM-SchemeMethod-class)))
       (DM-set-field-value! m DM-Data-value-field default)
       (DM-set-sharable! m)
       (DM-set-field-value! g DM-Generic-defaulting-field m)
       m ) )
    ;; Only one instruction to force the choice of the default method.
    (DM-set-indexed-field-value! g DM-Generic-instruction-field 0 0)
    ;; Store the label
    (do ((i 0 (+ i 1)))
        ((>= i label-length))
      (DM-set-indexed-field-value! 
       g DM-Generic-name-field i (string-ref label i) ) )
    (DM-set-indexed-field-value! 
     g DM-Generic-name-field label-length (integer->char 0) )
    ;; g is an immutable value but why not making it copyable (methods
    ;; are shared so they can keep their origin)?
    (DM-set-copyable! g)
    g ) )

;;; Calling a (generic) function. This is a kind of generic function
;;; discriminating on the nature of the function to invoke and
;;; choosing the appropriate way to do it.

(define (DM-invoke function . arguments)
  (internal-invoke function arguments) )

(define (internal-invoke function arguments)
  (define (listify args arity)
    (if (> arity 0)
        (if (pair? args)
            (cons (car args) (listify (cdr args) (- arity 1)))
            (dmeroonet-error 'DM-invoke "Missing arguments" 
                             function arguments ) )
        (list args) ) )
  (cond 
   ((DM-is-a? function DM-Generic-class)
    (check-arity! function arguments)
    (let* ((method      (method-lookup function arguments))
           (method-site (DM-site-of function)) )
      (if (eq? DM-site method-site)
          ;; apply locally 
          (let* ((closure (DM-field-value method DM-Data-value-field))
                 (arity   (DM-field-value function DM-Generic-arity-field)) )
            (case arity
              ((1) (closure function (car arguments)))
              ((2) (closure function (car arguments) (cadr arguments)))
              ((3) (closure function 
                       (car arguments) (cadr arguments) (caddr arguments) ))
              (else 
               (if (>= arity 0) 
                   (apply closure function arguments)
                   (apply closure function 
                          (listify arguments (- (+ arity 1))) ) ) ) ) )
          ;; transmit a remote invokation request. No need to wait or
          ;; serve, the continuation is kept in the request.
          (call/cc 
           (lambda (k)
             (DM-send method-site 
              (build-request method (cons function arguments) k) ) ) ) ) ) )
   ((DM-is-a? function DM-SchemeMethod-class)
    (let ((method-site (DM-site-of function)))
      (if (eq? DM-site method-site)
          ;; apply locally 
          (let ((closure (DM-field-value function DM-Data-value-field)))
            (apply closure arguments) )
          ;; transmit a remote invokation request. No need to wait or
          ;; serve, the continuation is kept in the request.
          (call/cc 
           (lambda (k)
             (DM-send method-site 
              (build-request function arguments k) ) ) ) ) ) )
   ((DM-is-a? function DM-SchemeContinuation-class)
    (if (= 1 (length arguments))
        ((DM-field-value function DM-Data-value-field)
         (car arguments) )
        (dmeroonet-error 'DM-invoke "Wrong arity for a Scheme continuation"
                         function arguments ) ) )
   (else
    (dmeroonet-error 'DM-invoke "Cannot invoke that sort of function"
                     function arguments ) ) ) )

;;; Unwrap an Invokation request and perform it.

(define (perform-Invokation invokation)
  (let* ((function (DM-field-value invokation DM-Invokation-function-field))
         (k        (DM-field-value invokation DM-Invokation-continuation-field))
         (args-length (DM-indexed-field-length 
                       invokation DM-Invokation-argument-field ))
         (arguments (let extract ((i 0))
                      (if (< i args-length)
                          (let ((e (DM-indexed-field-value
                                    invokation
                                    DM-Invokation-argument-field
                                    i )))
                            (cons (DM->Sexpr e) (extract (+ i 1))) )
                          (list) ) ))
         (result (internal-invoke function arguments)) )
    (if (DM-is-object? k)
        (DM-send (DM-site-of k) (build-request k (list result) #f))
        'nothing ) ) )        

;;; Build a request asking for a remote invokation.

(define (build-request method arguments k)
  (let* ((arguments-length (length arguments))
         (request (DM-allocate DM-Invokation-class arguments-length)) )
    ;; store the method to invoke
    (DM-set-field-value! 
     request DM-Invokation-function-field method )
    ;; copy the arguments
    (do ((i 0 (+ i 1))
         (arguments arguments (cdr arguments)) )
        ((>= i arguments-length))
      (DM-set-indexed-field-value!
       request DM-Invokation-argument-field i 
       (DM<-Sexpr (car arguments)) ) )
    ;; store the continuation
    (DM-set-field-value!
     request DM-Invokation-continuation-field 
     (let ((d (DM-allocate DM-Data-class)))
       (DM-set-field-value! d DM-Data-value-field k)
       (DM-set-sharable! d)
       d ) )
    (DM-set-copyable! request)
    request ) )

;;; Search the appropriate method.

(define (method-lookup generic arguments)
  (let* ((discriminants (extract-discriminants generic arguments))
         (ndiscr        (length discriminants)) )
    (if (= ndiscr 1)
        (let ((discriminant (car discriminants)))
          (if (DM-is-object? discriminant)
              (search-method-1 generic (DM-class-of discriminant))
              (DM-field-value generic DM-Generic-defaulting-field) ) )
        (dmeroonet-error 'DM-invoke "not yet implemented") ) ) )

(define (search-method-1 generic receiver-class)
;;(dm-describe generic); DEBUG
  (let execute ((pc 0))
    (let ((word (DM-indexed-field-value 
                 generic DM-Generic-instruction-field pc )))
(display `(search ,generic ,receiver-class ,word))(newline); DEBUG
      (if (>= word 0)
          (DM-indexed-field-value generic DM-Generic-method-field word)
          (let* ((next-word (DM-indexed-field-value
                             generic
                             DM-Generic-instruction-field
                             (+ pc 1) ))
                 (class (DM-indexed-field-value
                        generic 
                        DM-Generic-class-field 
                        next-word )) )
            (if (DM-is-a-subclass? receiver-class class)
                (execute (+ pc 2))
                (execute (+ pc (- word))) ) ) ) ) ) )

;;; }}}
;;; {{{ Method definition

;;; Methods are created via InstallableMethods that do not need to be
;;; DMeroon objects. FUTURE hide this vector into a lambda ???

(define (DM-make-InstallableMethod premethod signature class)
  (vector 'InstallableMethod premethod signature class) )
(define (InstallableMethod-premethod im)
  (vector-ref im 1) )
(define (InstallableMethod-signature im)
  (vector-ref im 2) )
(define (InstallableMethod-class im)
  (vector-ref im 3) )

;;; Define a method on a generic function. 

(define-dmeroonet-macro (DM-define-method call . body)
  `(set! ,(car call)
         (DM-add-methods ,(car call)
                         (DM-method ,call . ,body) ) ) )

;;; Define a method.
;;; FUTURE: In fact, one may allow to only change the non-discriminating
;;; variables in a call-next/former-method.

(define-dmeroonet-macro (DM-method call . body)
  (DM-process-method call body) )

(define (DM-process-method call body)
  (parse-variable-specifications
   (cdr call)
   (lambda (discriminant variables)
     (let ((og (gensym))(cl (gensym)))
       `(DM-make-InstallableMethod
         (lambda (,og ,cl)              ; pre-method
           (lambda (,(car call) . ,(flat-variables variables))
             ;(define (call-next-method)
             ;  (DM-invoke
             ;   (search-method-1 ,(car call) (DM-Class-superclass ,cl))
             ;   ,(car call) . ,(flat-variables variables) ) )
             ;(define (call-former-method . ,(flat-variables variables))
             ;  (DM-invoke
             ;   (search-method-1 ,og ,cl)
             ;   ,(car call) . ,(flat-variables variables) ) )
             (define (call-next-method)
               (DM-special-invoke ,(car call) (DM-Class-superclass ,cl)
                ,(car call) . ,(flat-variables variables) ) )
             (define (call-former-method . ,(flat-variables variables))
               (DM-special-invoke ,og ,cl
                ,(car call) . ,(flat-variables variables) ) )
             . ,body ) )
         ',(cdr call)                   ; signature
         ,(cadr discriminant) ) ) ) ) ) ; discriminating class

;;; This is not efficient, this is rather a hack to hide
;;; search-method-1 from the expansion of DM-method.             FUTURE

(define (DM-special-invoke og cl g . args)
  (apply DM-invoke (search-method-1 og cl) g args) )

;;; Add a number of methods altogether.

(define (DM-add-methods generic . installable-methods)
  (define (add-one-method g installable-method)
    (if (congruent? g (InstallableMethod-signature installable-method))
        (let ((class (InstallableMethod-class installable-method)))
          (insert-method g
                         class
                         ((InstallableMethod-premethod installable-method)
                          g class ) ) )
        (DM-error 'DM-add-methods "Non congruent signatures") ) )
  (let add ((g generic)(ims installable-methods))
    (if (pair? ims)
        (add (add-one-method g (car ims)) (cdr ims))
        g ) ) )

;;; Compare signatures of a generic and a method.

(define (congruent? generic signature)
  (and (= (compute-arity signature)
          (DM-field-value generic DM-Generic-arity-field) )
       (= (compute-profile signature) 
          (DM-field-value generic DM-Generic-profile-field) ) ) )

;;; Build a new generic function with an added method.
;;; Currently, all methods are kept even needless ones.    TO BE IMPROVED.
;;; To store the Object class is also useless.

(define (insert-method generic class closure)
  (define (get-instruction i)
    (DM-indexed-field-value generic DM-Generic-instruction-field i) )
  (define (get-class i)
    (DM-indexed-field-value generic DM-Generic-class-field i) )
  (define (get-tree pc)
    (let ((word (get-instruction pc)))
      (if (>= word 0)
          (list word)
          `(word ,(get-instruction (+ pc 1))
                ,@(get-tree (+ pc 2))
                ,@(get-tree (+ pc (- word))) ) ) ) )
  ;; These are the indexes where the new method and the new class will
  ;; be recorded within the new Generic instance.
  (define method-index 
    (DM-indexed-field-length generic DM-Generic-method-field) )
  (define class-index (DM-indexed-field-length generic DM-Generic-class-field))
  (define name-length (DM-indexed-field-length generic DM-Generic-name-field))
  (define (insert pc topclass)
;;(display `(insert ,pc ,topclass))(newline) ; DEBUG
    (let ((word (get-instruction pc)))
      (if (>= word 0)
          (if (eq? class topclass)
              `(,method-index)
              `(-3 ,class-index ,method-index ,word) )
          (let* ((next-word (get-instruction (+ pc 1)))
                 (cl        (get-class next-word)) )
            (cond
             ((DM-is-a-subclass? class cl)
              (let ((yes-tree (insert (+ pc 2) cl)))
                `(,(- (+ 2 (length yes-tree)))
                  ,next-word
                  ,@yes-tree 
                  ,@(get-tree (+ pc (- word))) ) ) )
             ((DM-is-a-subclass? cl class)
              (let* ((yes-tree (get-tree (+ pc 2)))
                     (no-tree  (get-tree (+ pc (- word))))
                     (new-yes-tree `(,(- (+ 2 (length yes-tree)))
                                     ,next-word
                                     ,@yes-tree
                                     ,method-index )) )
                `(,(- (+ 2 (length new-yes-tree)))
                  ,class-index
                  ,@new-yes-tree
                  ,@no-tree ) ) )
             (else (let ((yes-tree (get-tree (+ pc 2)))
                         (no-tree (insert (+ pc (- word)) topclass)) )
                     `(,word ,next-word ,@yes-tree ,@no-tree) )) ) ) ) ) )
  (let* ((new-instructions (insert 0 DM-Object-class))
         (g (DM-allocate DM-Generic-class
                         (+ 1 class-index)
                         (+ 1 method-index)
                         (length new-instructions)
                         (DM-indexed-field-length
                          generic DM-Generic-name-field ) ))
         (method (let ((m (DM-allocate DM-SchemeMethod-class)))
                   (DM-set-field-value! m DM-Data-value-field closure)
                   (DM-set-sharable! m)
                   m )) )
;;(display `(insert yields ,new-instructions))(newline) ; DEBUG
    (DM-set-field-value! g DM-Generic-arity-field 
                         (DM-field-value generic DM-Generic-arity-field) )
    (DM-set-field-value! g DM-Generic-profile-field 
                         (DM-field-value generic DM-Generic-profile-field) )
    (DM-set-field-value! g DM-Generic-defaulting-field 
                         (DM-field-value generic DM-Generic-defaulting-field) )
    (do ((i 0 (+ i 1)))
        ((>= i class-index))
      (DM-set-indexed-field-value!
       g DM-Generic-class-field i (get-class i)) )
    (DM-set-indexed-field-value!
     g DM-Generic-class-field class-index class )
    (do ((i 0 (+ i 1)))
        ((>= i method-index))
      (DM-set-indexed-field-value!
       g DM-Generic-method-field i
         (DM-indexed-field-value generic DM-Generic-method-field i) ) )
    (DM-set-indexed-field-value!
     g DM-Generic-method-field method-index method )
;;(DM-describe g)
    (do ((is new-instructions (cdr is))
         (pc 0 (+ pc 1)) )
        ((null? is))
      (DM-set-indexed-field-value! g DM-Generic-instruction-field pc (car is)) )
;;(DM-describe g)
    (do ((i 0 (+ i 1)))
        ((>= i name-length))
      (DM-set-indexed-field-value!
       g DM-Generic-name-field i
         (DM-indexed-field-value generic DM-Generic-name-field i) ) )
    (DM-set-copyable! g)
    g ) )


;;; }}}
;;; end of dmeroonet.scm
