;;; $Id: dmlib.scm,v 1.16 1998/03/19 19:03:43 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; This file defines a number of utility functions that may be used
;;; from Scheme. It might have been split into different files.

;;; {{{ Iteration

;;; RESOURCE DM-iterate
;;; Iterate on all the fields of an object from first to last fields.
;;;    On each regular field,               apply by-item,
;;;    on the length of each indexed field, apply by-length,
;;;    on each value of an indexed field,   apply by-indexed-item.
;;; Ignore secret fields that cannot be read.

(define (DM-iterate o by-item by-length by-indexed-item)
  (let* ((class         (DM-class-of o))
         (fields-number (DM-indexed-field-length class DM-Class-field-field)) )
    (define (is-field-secret? field)
      (let ((options (DM-field-value field DM-Field-options-field)))
        (>= (modulo options 32) 16) ) )
    (define (scan field)
      (let* ((reptyp  (DM-field-repetition-type field))
             (itemtyp (DM-field-item-type field)) )
        (if (not (is-field-secret? field))
            (if (eq? DM-nought-type reptyp)
                ;; Regular field
                (by-item o field itemtyp)
                ;; Indexed field
                (let ((rep (DM-indexed-field-length o field)))
                  (by-length o field reptyp itemtyp)
                  (do ((j 0 (+ j 1)))
                      ((>= j rep))
                    (by-indexed-item o field itemtyp j) ) ) ) ) ) )
    (do ((i 0 (+ i 1)))
        ((>= i fields-number))
      (scan (DM-indexed-field-value class DM-Class-field-field i)) ) ) )

;;; }}}
;;; {{{ Coercion

;;; RESOURCE DM->string
;;; Convert a String into a Scheme string.  With an additional Field,
;;; convert an indexed field of char1 into a Scheme string. Also remove
;;; trailing nul characters.

(define (DM->string o . field)
  (if (pair? field)
      (let ((field (car field)))
        (if (DM-is-a? field DM-Field-class)
            (let ((class (DM-field-value field DM-Field-introducing-field)))
              (if (DM-is-a? o class)
                  (if (eq? (DM-field-item-type field) DM-char1-type)
                      (let* ((len (let shorten ((len (DM-indexed-field-length 
                                                      o field )))
                                    (if (> len 0)
                                        (if (= 0 (char->integer
                                                  (DM-indexed-field-value
                                                   o field (- len 1) ) ) )
                                            (shorten (- len 1))
                                            len )
                                        len ) ))
                             (s   (make-string len)) )
                        (do ((i 0 (+ i 1)))
                            ((>= i len) s)
                          (string-set! s i (DM-indexed-field-value
                                            o field i )) ) )
                      (DM-error "Not an repetition of characters" field) )
                  (DM-error "Mismatch class/instance" o class) ) )
            (DM-error "Not a Field" field) ) )
      (if (DM-is-a? o DM-String-class)
          (let* ((len (let shorten ((len (DM-indexed-field-length 
                                          o DM-String-character-field )))
                        (if (> len 0)
                            (if (= 0 (char->integer
                                      (DM-indexed-field-value
                                       o 
                                       DM-String-character-field 
                                       (- len 1) ) ) )
                                (shorten (- len 1))
                                len )
                            len ) ))
                 (s   (make-string len)) )
            (do ((i 0 (+ i 1)))
                ((>= i len) s)
              (string-set! s i (DM-indexed-field-value 
                                o DM-String-character-field i )) ) )
          (DM-error "Not a DMeroon String" o) ) ) )
;;; Test: (DM->string DM-site DM-Site-name-field)
;;;       (DM->string (DM-field-value DM-site DM-Site-support-field))

;;; }}}
;;; {{{ Description

;;; RESOURCE DM-describe
;;; Describe a DMeroon object at a superficial (non recursive) level.
;;; Secret fields are not shown.

(define (DM-describe o . stream)
  (define (DM-do-describe o stream)
    (define (check-item o field itemtyp)
      (display (DM->string field DM-Field-name-field) stream)
      (display "(" stream)
      (display (DM->string (DM-field-item-type field) DM-DMeroonType-name-field)
               stream )
      (display "): " stream)
      (cond
       ((or (eq? itemtyp DM-code-type)
            (eq? itemtyp DM-key-type) )
        (display "non printable") )
       (else
        (write (DM-field-value o field) stream) ) )
      (newline stream) )
    (define (check-length o field reptyp itemtyp)
      (display (DM->string field DM-Field-name-field) stream)
      (display "(" stream)
      (display (DM-indexed-field-length o field) stream)
      (display " * " stream)
      (display (DM->string (DM-field-item-type field) DM-DMeroonType-name-field)
               stream )
      (display ") " stream)
      (newline stream) )
    (define (check-indexed-item o field itemtyp j)
      (display (DM->string field DM-Field-name-field) stream)
      (display "(" stream)
      (display (DM->string (DM-field-item-type field) DM-DMeroonType-name-field)
               stream )
      (display ")[" stream)
      (display j stream)
      (display "]: " stream)
      (cond
       ((or (eq? itemtyp DM-code-type)
            (eq? itemtyp DM-key-type) )
        (display "non printable") )
       (else
        (write (DM-indexed-field-value o field j) stream) ) )
      (newline stream) )
    (newline stream)
    (display "--- DMeroon instance " stream)
    (display o stream)
    (display " of class " stream)
    (display (DM->string (DM-class-of o) DM-Class-name-field) stream)
    (display "---" stream)
    (newline stream)
    (DM-iterate o check-item check-length check-indexed-item)
    (newline stream)
    o )
  (DM-do-describe 
   o (if (pair? stream) (car stream) (current-output-port)) ) )
;;; Test: (DM-describe DM-site)

;;; RESOURCE DM-describe-all
;;; Describe recursively an object and its offspring but it avoids to
;;; display secret fields. This function is analogous to the unveil 
;;; function from MeroonV3.

(define (DM-describe-all o . stream)
  (define (DM-do-describe-all o stream all indent)
    (define (left-margin indent)
      (do ((i 0 (+ i 1)))
          ((>= i indent))
        (display " " stream) ) )
    (define (do-scan o indent)
      (if o (scan o indent)) )
    (define (scan o indent)
      (define (check-item o field itemtyp)
        (left-margin indent)
        (display (DM->string field DM-Field-name-field) stream)
        (display "(" stream)
        (display (DM->string (DM-field-item-type field) 
                             DM-DMeroonType-name-field )
                 stream )
        (display "): " stream)
        (cond
         ((or (eq? itemtyp DM-code-type)
              (eq? itemtyp DM-key-type) )
          (display "non printable")
          (newline stream ) )
         (else (let ((v (DM-field-value o field)))
                 (write v stream)
                 (newline stream)
                 (if (eq? DM-reference-type itemtyp)
                     (do-scan v (+ indent 1)) ) )) ) )
      (define (check-length o field reptyp itemtyp)
        (left-margin indent)
        (display (DM->string field DM-Field-name-field) stream)
        (display "(" stream)
        (display (DM-indexed-field-length o field) stream)
        (display " * " stream)
        (display (DM->string (DM-field-item-type field)
                             DM-DMeroonType-name-field )
                 stream )
        (display ") " stream)
        (newline stream) )
      (define (check-indexed-item o field itemtyp j)
        (left-margin indent)
        (display (DM->string field DM-Field-name-field) stream)
        (display "(" stream)
        (display (DM->string (DM-field-item-type field)
                             DM-DMeroonType-name-field )
                 stream )
        (display ")[" stream)
        (display j stream)
        (display "]: " stream)
        (cond
         ((or (eq? itemtyp DM-code-type)
              (eq? itemtyp DM-key-type) )
          (display "non printable")
          (newline stream) )
         (else (let ((v (DM-indexed-field-value o field j)))
                 (write v stream)
                 (newline stream)
                 (if (eq? DM-reference-type itemtyp)
                     (do-scan v (+ indent 1)) ) )) ) )
      (left-margin indent)
      (display "--- DMeroon instance " stream)
      (display o stream)
      (display " of class " stream)
      (display (DM->string (DM-class-of o) DM-Class-name-field) stream)
      (cond 
       ((memq o all)
        (display "(already seen)" stream)
        (display "---" stream)
        (newline stream) )
       (else 
        ;; serve a little: (DM-serve 0) ; DEBUG
        (set! all (cons o all))
        (display "---" stream)
        (newline stream)
        (DM-iterate o check-item check-length check-indexed-item) ) )
      o )
    (scan o indent) )
  (DM-do-describe-all
   o (if (pair? stream) (car stream) (current-output-port)) (list) 0 ) )
;;; Test: (DM-describe-all DM-site)

;;; }}}
;;; {{{ Publishing
;;; We use a number of internal functions whose name start with DMinternal-.
;;; They are not to be used directly.
;;;     {{{ Internals 

;;; Search a dictionary with a specific name. Search only though the spine
;;; of these. Always return #f or a Dictionary.

(define (DMinternal-dict-find dict name)
  (if (and dict (DM-is-a? dict DM-Dictionary-class))
      (let ((dict-name (DM->string dict DM-Dictionary-name-field)))
        (if (equal? name dict-name)
            ;; returns the dictionary itself (not the value).
            dict
            (DMinternal-dict-find 
             (DM-field-value dict DM-Dictionary-rest-field) 
             name ) ) )
      #f ) )

;;; Find recursively a dictionary with a pathname.
;;; Always return #f or a Dictionary.

(define (DMinternal-dict-rec-find dict names)
  (if (pair? names)
      (let ((d (DMinternal-dict-find dict (car names))))
        (if d
            (if (pair? (cdr names))
                (DMinternal-dict-rec-find 
                 (DM-field-value d DM-Dictionary-value-field)
                 (cdr names) )
                d )
            d ) )
      dict ) )

;;; Create a Dictionary and initialize it.

(define (DMinternal-mk-dict rest value name)
  (let* ((namelen (string-length name))
         (d (DM-allocate DM-Dictionary-class (+ 1 namelen))) )
    (DM-set-field-value! d DM-Dictionary-rest-field rest)
    (DM-set-field-value! d DM-Dictionary-value-field value)
    (do ((i 0 (+ i 1)))
        ((>= i namelen))
      (DM-set-indexed-field-value! 
       d DM-Dictionary-name-field i (string-ref name i) ) )
    (DM-set-initialized! d) ) )

;;; Bind a name to a value. Insert the new created Dictionary in the given
;;; one. Try to do it physically.

(define (DMinternal-dict-adjoin! dict value name)
  (if (and dict (DM-is-a? dict DM-Dictionary-class))
      (let ((d (DMinternal-dict-find dict name)))
        (if d
            ;; We found a dictionary with the given name, patch it!
            (begin
              (DM-set-field-value! d DM-Dictionary-value-field value)
              dict )
            ;; There is no such dictionary, create one and insert right
            ;; after the top one.
            (let ((d (DMinternal-mk-dict 
                      (DM-field-value dict DM-Dictionary-rest-field)
                      value
                      name )))
              (DM-set-field-value! dict DM-Dictionary-rest-field d)
              dict ) ) )
      (DMinternal-mk-dict #f value name) ) )
        
;;; Bind a value to a pathname. Try to do it physically.

(define (DMinternal-dict-rec-adjoin! dict value names)
  (if (pair? names)
      (let ((d (DMinternal-dict-find dict (car names))))
        (if d
            (begin
              (DM-set-field-value!
               d
               DM-Dictionary-value-field
               (if (pair? (cdr names))
                   (DMinternal-dict-rec-adjoin! 
                    (DM-field-value d DM-Dictionary-value-field)
                    value 
                    (cdr names) )
                   value ) )
              dict )
            (DMinternal-dict-adjoin! 
             dict
             (if (pair? (cdr names))
                 (DMinternal-dict-rec-adjoin! DM-nil value (cdr names))
                 value )
             (car names) ) ) )
      dict ) )

;;;     }}}

;;; RESOURCE DM-publish!
;;; Publish a value under a given pathname. Create or patch the intermediate
;;; Dictionaries. 

(define (DM-publish! value . names)
  (let ((dict (DM-field-value DM-site DM-Site-information-field))
        (pathname (if (and (pair? names)
                           (equal? (car names) "user") )
                      names
                      (cons "user" names) )) )
    (DMinternal-dict-rec-adjoin! dict value pathname) ) )

;;; RESOURCE DM-retrieve
;;; Retrieve the value associated to a pathname or #f it no such value 
;;; exists.

(define (DM-retrieve . names)
  (let* ((dict (DM-field-value DM-site DM-Site-information-field))
         (pathname (if (and (pair? names)
                            (equal? (car names) "user") )
                       names
                       (cons "user" names) ))
         (d (DMinternal-dict-rec-find dict pathname)) )
    (if d
        (DM-field-value d DM-Dictionary-value-field)
        #f ) ) )

;;; }}}
;;; {{{ Conversion

;;; RESOURCE DM<-sexpr
;;; Converts a Scheme Sexpression into a DMeroon object.

(define (DM<-sexpr e . handler)
  (define (default-handler e) 
    (DM-error "Cannot convert to DMeroon" e) )
  (let ((handler (if (pair? handler) (car handler) default-handler)))
    (cond
     ((DM-is-object? e) e)
     ((null? e) DM-nil)
     ((boolean? e) (if e DM-true DM-false))
     ((symbol? e)
      (let* ((str (symbol->string e))
             (len (string-length str))
             (s (DM-allocate DM-Symbol-class (+ len 1))) )
        (do ((i 0 (+ i 1)))
            ((>= i len))
          (DM-set-indexed-field-value!
           s DM-String-character-field i (string-ref str i) ) )
        (DM-set-indexed-field-value!
         s DM-String-character-field len (integer->char 0) )
        s ) )
     ((char? e)
      (let ((c (DM-allocate DM-Char1-class)))
        (DM-set-field-value! c DM-Char1-value-field e)
        c ) )
     ((vector? e)
      (let* ((len (vector-length e))
             (v (DM-allocate DM-ImmutableVector-class len)) )
        (do ((i 0 (+ i 1)))
            ((>= i len))
          (DM-set-indexed-field-value!
           v DM-ImmutableVector-item-field i 
           (DM<-sexpr (vector-ref e i) handler) ) )
        v ) )
     ((pair? e)
      (let ((p (DM-allocate DM-ImmutablePair-class)))
        (DM-set-field-value! p DM-ImmutablePair-value-field
                             (DM<-sexpr (car e) handler) )
        (DM-set-field-value! p DM-ImmutableLink-rest-field
                             (DM<-sexpr (cdr e) handler) )
        p ) )
     ((string? e) 
      (let* ((len (string-length e))
             (str (DM-allocate DM-String-class (+ 1 (string-length e)))) )
        (do ((i 0 (+ i 1)))
            ((>= i len))
          (DM-set-indexed-field-value!
           str DM-String-character-field i (string-ref e i) ) )
        (DM-set-indexed-field-value!
         str DM-String-character-field len (integer->char 0) )
        str ) )
     ((integer? e) 
      (let ((i (DM-allocate DM-Int3-class)))
        (DM-set-field-value! i DM-Int3-value-field e)
        i ) )
     ((real? e)
      (let ((r (DM-allocate DM-Float2-class)))
        (DM-set-field-value! r DM-Float2-value-field e)
        r ) )
     (else (handler e)) ) ) )

;;; RESOURCE DM->sexpr
;;; Converts a DMeroon object into a Scheme Sexpression.

(define (DM->sexpr o . handler)
  (define (default-handler o)
    (DM-error "Cannot convert from DMeroon" o) )
  (let ((handler (if (pair? handler) (car handler) default-handler)))
    (cond
     ((not (DM-is-object? o)) o)
     ((eq? o DM-nil) '())
     ((eq? o DM-true) #t)
     ((eq? o DM-false) #f)
     ((DM-is-a? o DM-Symbol-class) 
      (string->symbol (DM->string o DM-String-character-field)) )
     ((DM-is-a? o DM-Char1-class)
      (DM-field-value o DM-Char1-value-field) )
     ((DM-is-a? o DM-ImmutableVector-class)
      (let* ((len (DM-indexed-field-length o DM-ImmutableVector-item-field))
             (v   (make-vector len)) )
        (do ((i 0 (+ i 1)))
            ((>= i len) v)
          (let ((value (DM-indexed-field-value 
                        o DM-ImmutableVector-item-field i )))
            (vector-set! v i (DM->sexpr value handler)) ) ) ) )
     ((DM-is-a? o DM-ImmutablePair-class)
      (cons 
       (DM->sexpr (DM-field-value o DM-ImmutablePair-value-field))
       (DM->sexpr (DM-field-value o DM-ImmutableLink-rest-field)) ) )
     ;; works for Int*, Nat* or Float*.
     ((DM-is-a? o DM-Number-class)
      (let* ((class  (DM-class-of o))
             (field  (DM-indexed-field-value class DM-Class-field-field 0)) )
        (DM-field-value o field) ) )
     ((DM-is-a? o DM-String-class) (DM->string o))
     (else (handler o)) ) ) )

;;; }}}
;;; end of dmlib.scm
