;;; $Id: sigtools.scm,v 1.4 1997/03/17 19:59:36 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; This file contains a series of functions to help writing sig2XXX
;;; compilers. Such a compiler takes a sequence of signatures and
;;; convert them into C or Scheme code to make the described resources
;;; (C variables or functions) appear as Scheme variables (or
;;; functions). 

;;; First, all *.[ch] files are read and interfaces declarations are
;;; extracted (by a Perl script) then they are read by
;;; read-all-signatures and appropriately handled.

;;; There are three sorts of signatures:
;;;   ("type" "variable")
;;;             Tells that variable is of type.
;;;   ("type" ("function" "types" ...))
;;;             Tells that function takes types as arguments and returns type.
;;;   ("type" ("macro" "function" "types" ...))
;;;             Tells that the C macro named function takes types as arguments
;;;             and returns type. MACRO is a keyword.

;;; Types that appear in the previous signatures may be 
;;;     C types such as int, double etc.
;;;     DMeroon classes if their name is capitalized such as Class, Tcp etc.
;;;     DMeroon types such as nat1, int2, float3 etc.

;;; To avoid the use of a case-sensitive Scheme, signatures are collected
;;; as Sexp whose symbols are turned into strings which are turned back
;;; into symbols with the appropriate name.

;;;{{{ Reporting errors

;;; All errors are mapped onto the sigtools-error function.

(define (sigtools-error message . culprits)
  (display `(*** *** *** ERROR *** ***))
  (newline)
  (display `(,message . ,culprits))
  (newline)
  (display `(*** *** *** ERROR *** ***))
  (newline)
  (car "Force an error portably (:)") )

;;;}}}
;;;{{{ Reading signatures from a file. (read here is case-sensitive).

(define (read-signatures-file sigfile)
  (define (read-all-signatures in)
    (define (symbolify e)
      (if (pair? e)
          (cons (symbolify (car e)) (symbolify (cdr e)))
          (if (null? e) '() (symbol-concatenate e)) ) )
    (let loop ()
      (let ((sig (read-cs in)))
        (if (eof-object? sig)
            '()
            (cons (symbolify sig)
                  (loop) ) ) ) ) )
  (call-with-input-file sigfile
    read-all-signatures ) )

;;;}}}
;;;{{{ Utilities

(define (flat l)
  (define (flatten l r)
    (if (pair? l)
        (flatten (car l) (flatten (cdr l) r))
        (if (null? l)
            r
            (cons l r) ) ) )
  (flatten l '()) )

;;; Forge a new symbol, the name of which is the concatenation of all
;;; these things. Does not seem worth memoizing this function.

(define (symbol-concatenate . names)
  (string->symbol
   (apply string-append
          (map (lambda (thing)
                 (cond ((symbol? thing)
                        (symbol->string thing) )
                       ((string? thing) thing)
                       ((number? thing)
                        (number->string thing) )
                       (else (sigtools-error
                              "non coercible to String"
                              thing )) ) )
               names ) ) ) )

;;; Produces a list of variables whose length is the same as a given
;;; list.  It produces a Scheme variable list which might be dotted if
;;; required. Since sigtypes may be names of C types and since C types
;;; are not legal names of variables, these names are suffixed.

(define (convert-sigtypes-to-names sigtypes)
  (define index 0)
  (define (convert-single-variable sigtype)
    (set! index (+ index 1))
    (symbol-concatenate sigtype "_" index) )
  (let loop ((sigtypes sigtypes))
    (if (pair? sigtypes)
        (cons (convert-single-variable (car sigtypes))
              (loop (cdr sigtypes)) )
        (if (null? sigtypes)
            '()
            (convert-single-variable sigtypes) ) ) ) )

;;;}}}
;;;{{{ Type converter

;;; Converts a type appearing in a signature into the appropriate type
;;; of the support language. 
;;; DMeroon classes are turned into the anything type,
;;; other types are translated following the types-mapping variable.

;;; This is the map from signature types to support language types.
;;; It should at least define the following types: string, char, fixnum,
;;; floatnum, anything, boolean, nat1, nat2, nat3, int1, int2, int3.

(define types-mapping
  '() )

;;; These other types are defined as aliases.
;;; Don't forget to call that function at initialization.

(define (complete-types-mapping)
  (let adjoin ((others (list (cons (string->symbol "netnat2")
                                   (string->symbol "nat2") )
                             (cons (string->symbol "netnat3")
                                   (string->symbol "nat3") )
                             (cons (string->symbol "offset")
                                   (string->symbol "nat3") )
                             (cons (string->symbol "index")
                                   (string->symbol "nat1") )
                             (cons (string->symbol "options")
                                   (string->symbol "nat3") )
                             (cons (string->symbol "translation")
                                   (string->symbol "nat2") )
                             (cons (string->symbol "displacement")
                                   (string->symbol "nat3") )
                             (cons (string->symbol "date")
                                   (string->symbol "nat3") )
                             (cons (string->symbol "void")
                                   (string->symbol "anything") )
                             (cons (string->symbol "stream")
                                   (string->symbol "anything") ) )))
    (if (pair? others)
        (let ((type (cdr (assq (cdar others) types-mapping))))
          (set! types-mapping
                (cons (cons (caar others) type)
                      types-mapping ) )
          (adjoin (cdr others)) ) ) ) )

(define (DMeroon-class-type? type)
  (char-upper-case? (string-ref (symbol->string type) 0)) )

;;; If typename starts with a capital then it designates a DMeroon class.
;;; Otherwise it must be one of the types listed in types-mapping.

(define (convert-type type)
  (if (DMeroon-class-type? type)
      (cdr (assoc (string->symbol "anything") types-mapping))
      (if types-mapping
          (let ((v (assoc type types-mapping)))
            (if (pair? v)
                (cdr v)
                (sigtools-error "No equivalent type" type) ) )
          (sigtools-error "No types mapping defined") ) ) )
;;; Test: (convert-type "Object")
;;;       (convert-type "foo")

;;; Convert a (possibly dotted) list of types.

(define (convert-list-of-types sigtypes)
  (if (pair? sigtypes)
      (cons (convert-type (car sigtypes))
            (convert-list-of-types (cdr sigtypes)) )
      (if (null? sigtypes) 
          '()
          (convert-type sigtypes) ) ) )

;;;}}}
;;;{{{ Generation utilities

(define (display-list-with comma e* out)
  (display-list-with-individual-prefix " " comma e* out) )

;;; Prints (a b c) into (T a, T b, T c)
;;; The prefix and the comma are programmable.

(define (display-list-with-individual-prefix prefix comma e* out)
  (define (iterate e*)
    (if (pair? e*)
        (begin (display prefix out)
               (display (car e*) out)
               (if (pair? (cdr e*))
                   (begin (display comma out)
                          (iterate (cdr e*)) ) ) ) ) )
  (display "(" out)
  (iterate e*)
  (display ")" out) )
;;; Test: (display-list-with-individual-prefix "T " ", " '(a b c) stdout)

;;; For debug purpose, it is good to know what is compiled therefore
;;; outputs as a comment the current signature.

(define (produce-c-comment sig out)
  (display "/* " out)
  (display sig out)
  (display " */" out)
  (newline out) )

(define (produce-c-eof out)
  (display "/* end. */" out)
  (newline out) )

(define (produce-scheme-comment sig out)
  (newline out)
  (display ";;; " out)
  (display sig out)
  (newline out) )

(define (produce-scheme-eof out)
  (newline out)
  (display ";;; end." out)
  (newline out) )

;;;}}}
;;;{{{ Convert C identifiers into support identifiers

;;; Convert a C name into a Scheme identifier unless the translation is
;;; registered in the special-names list.
;;; The displacement mainly converts _ into - essentially.

(define special-names 
  '( ) )

(define (convert-name name)
  (let ((v (assoc name special-names)))
    (if (pair? v)
        (cdr v) 
        (let* ((name   (symbol->string name))
               (len    (string-length name))
               (result (make-string len)) )
          (do ((i 0 (+ i 1)))
              ((>= i len) (string->symbol result))
            (string-set! 
             result 
             i
             (let ((ch (string-ref name i)))
               (cond ((char=? ch #\_) #\-)
                     (else ch) ) ) ) ) ) ) ) )
;;; Test: (convert-name "fOO")
;;;       (convert-name "DM_a_b")

;;; Similar conversion for macro names.

(define convert-macro-name convert-name)

;;;}}}
;;;{{{ 
;;;}}}

;;; end of sigtools.scm
