;;; $Id: dmeroonscript.el,v 1.8 1998/07/07 08:24:10 queinnec Exp $
;;; ///////////////////////////////////////////////////////////////////////
;;;                    //          //
;;;                   // DMeroon  //
;;;                  //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;; ///////////////////////////////////////////////////////////////////////

;;; DOCUMENTATION dmeroonscript.el
;;; This is the Emacs mode for Dmeroonscript programs. It allows to indent
;;; Dmeroonscript programs, to send them to a DMeroon server and to obtain
;;; the result.  It uses internally the w3 packages.
;;;
;;; Insert into your .emacs prologue
;;;    (setq auto-mode-alist
;;;          (cons '("\\.dms$"      . dmeroonscript-mode) auto-mode-alist) )
;;;    (autoload 'dmeroonscript-mode "dmeroonscript"
;;;              "Major mode to edit Dmeroonscript programs" t )
;;; After that, opening a file with extension .dms will set up the 
;;; dmeroonscript mode. M-C-x will send the content of the current buffer
;;; and open a new buffer with the result displayed with w3.

;;; You may fetch the w3 package, by William Perry, from:
;;;               ftp://ftp.cs.indiana.edu/pub/elisp/w3/

(require 'w3)
(require 'url-http)

;;; {{{ Customizable variables

(defvar dmeroonscript-server-host "localhost"
  "The host name of the DMeroon server to ask." )

(defvar dmeroonscript-server-port 56423
  "The port number of the DMeroon server to ask." )

;;; }}}
;;; {{{ Internal variables

(defvar dmeroonscript-mode nil
  "Boolean telling if Dmeroonscript mode is on." )

(defvar dmeroonscript-mode-map (copy-keymap text-mode-map)
  "Dmeroonscript mode map." )

;;; }}}
;;; {{{ Functions

(defun dmeroonscript-mode (&optional arg)
  "Major mode to edit Dmeroonscript files (text-based mode).

Customizable variables:
     dmeroonscript-server-host, dmeroonscript-server-port.

Additional command:
  \\[dmeroonscript-send] submits the current buffer to the default 
  DMeroon server, renders the resulting output in another buffer.
"
  (interactive)
  (setq dmeroonscript-mode t)
  (use-local-map dmeroonscript-mode-map)
  (define-key dmeroonscript-mode-map "\M-\C-x" 'dmeroonscript-send)
  (setq major-mode 'dmeroonscript-mode)
  (setq mode-name (concat "Dmeroonscript(" 
                          dmeroonscript-server-host
                          ":"
                          dmeroonscript-server-port
                          ")" ))
  ;; refresh modeline.
  (force-mode-line-update)
  (run-hooks 'dmeroonscript-mode-hook) )

(defun dmeroonscript-send ()
  "Send the current buffer as an Dmeroonscript request.
The targetted DMeroon server is specified by the dmeroonscript-server-host
and dmeroonscript-server-port variables."
  ;;; This code is snarfed from w3 code.
  (interactive)
  ;; Don't offer to save buffer. This is more interactive.
  (let ((buffer (get-buffer-create "DMeroon output")))
    (url-lazy-message "Contacting %s:%s" 
                      dmeroonscript-server-host
                      dmeroonscript-server-port )
    (let ((process (open-network-stream 
                    "DMeroon-stream"
                    url-working-buffer
                    dmeroonscript-server-host
                    dmeroonscript-server-port )) )
      (if (not (processp process))
          (url-sentinel url-working-buffer nil)
        (progn
          (url-process-put process 'url nil)
          (set-process-sentinel process 'ignore)
          (process-kill-without-query process)
          (message "Sending Dmeroonscript program...")
          (process-send-string process "Dmeroonscript\r\n")
          (process-send-string 
           process (buffer-substring (point-min) (point-max)) )
          (process-send-string process "\0\0\r\n")
          (url-lazy-message "Request sent, waiting for response...")
          (setq url-current-content-length nil)
          (make-local-variable 'after-change-functions)
          (add-hook 'after-change-functions 'url-after-change-function)
          (if url-be-asynchronous
              (set-process-sentinel process 'url-sentinel)
            (unwind-protect
                (save-excursion
                  (set-buffer url-working-buffer)
                  (erase-buffer)
                  (while (memq (url-process-status process)
                               '(run open))
                    (url-accept-process-output process)))
              (condition-case ()
                  (url-kill-process process)
                (error nil))))
          (if url-be-asynchronous
              nil
            (message "Retrieval complete.")
            (remove-hook 'after-change-functions
                         'url-after-change-function)))))
    (pop-to-buffer url-working-buffer)
    ;; render the buffer with w3
    (w3-preview-this-buffer)
    't ) )

;;; }}}
;;; end of dmeroonscript.el
