;;; $Id: javacomp.scm,v 1.5 1997/12/09 19:47:20 queinnec Exp $
;;; By Christian.Queinnec@inria.fr (http://pauillac.inria.fr/~queinnec/)
;;; This file is part of a Scheme->Java compiler derived from the Scheme->C
;;; compiler described in the LiSP book. 

;;; This file defines functions to compile a whole file into a .class file.

;;; Exportations:
;;; end of exportations.

;;;==================================================================
;;; A class for compilation commands.

(define-class Compilation Object
  ( source      ; a filename containing Scheme code
    verbose?    ; a boolean (tells if the compiler is verbose or not)
    ) )

(define-class Java-Compilation Compilation
  ( javafile    ; the file where to put Java code
    calljavac?  ; call the java compiler
    ) )

;;; This function may act as an application entry point with options
;;; being the list of options given on the Un*x command.  Compile a
;;; file to Java then a compiled class.
;;;    cxla file.scm 

(define (compiler-main options)
  (monitor compiler-main-handler
    (let ((cmd (make-Java-Compilation #f #f #f #f)))
      (scan-options! cmd options)
      (when (Compilation-verbose? cmd)
        (display "Reading source code...")(newline) )
      (let ((e `(begin . ,(file->list (Compilation-source cmd)))))
        (when (Compilation-verbose? cmd)
          (display "Expanding Scheme code...")(newline) )
        (set! e (walk e enriched-icslas-expansion-environment))
        (when (Compilation-verbose? cmd)
          (display "Generating Java code...")(newline) )
        (if (Java-Compilation-javafile cmd)
            ;; If there is a target file, generate java code into it.
            (begin
              (call-with-output-file (Java-Compilation-javafile cmd)
                (lambda (out) (compile->Java e (predefined-global-env) out)) )
              (when (Java-Compilation-calljavac? cmd)
                (when (Compilation-verbose? cmd)
                  (display "Calling external Java compiler...")(newline) )
                (call-external-compiler cmd) )
              #t )
            ;; For Jaja, generate java code into a string.
            (let ((out (open-output-string)))
              (compile->Java e (predefined-global-env) out)
              (let ((string (close-output-port out)))
                (when (Java-Compilation-calljavac? cmd)
                  (when (Compilation-verbose? cmd)
                    (display "Calling internal Java compiler...")(newline) )
                  (call-internal-compiler string) )
                string ) ) ) ) ) ) )

(define-generic (compiler-main-handler (exc) c?)
  (diagnose (list "No method for " (Class-name (object->class exc))
                  "Details on" exc )
            #f ) )

(define-method (compiler-main-handler (exc Compilation-Exception) c?)
  (flush-output-port stdout)
  (display "   *** *** *** " stderr)
  (display (Compilation-Exception-phase exc) stderr)
  (display " Exception *** *** ***" stderr)
  (newline stderr)
  (display (Compilation-Exception-message exc) stderr)
  (newline stderr)
  (for-each (lambda (culprit) 
              (show culprit stderr)
              (newline stderr) )
            (Compilation-Exception-culprits exc) )
  (call-next-method) )

;;; Just prints the warning and resume the diagnoser.

(define-method (compiler-main-handler (exc Compilation-Warning) c?)
  (flush-output-port stdout)
  (display (Compilation-Warning-phase exc) stderr)
  (display " Warning: " stderr)
  (display (Compilation-Warning-message exc) stderr)
  (newline stderr)
  (for-each (lambda (hint) 
              (show hint stderr)
              (newline stderr) )
            (Compilation-Warning-culprits exc) ) )

(define-method (compiler-main-handler (exc Exception) c?)
  (flush-output-port stdout)
  (display "!!!??? an exception occurred ???!!!" stderr)
  (newline stderr)
  (display exc stderr)
  (call-next-method) )

;;;==================================================================
;;; Show some specific objects.

(define-method (show (o Variable) . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (format stream "#<Var:~A>" (Variable-name o)) ) )

;;;==================================================================
;;; How to invoke the Java compiler.

(define-generic (call-external-compiler (cmd)))

;;; The Java compiler has to be parameterized with these variables:

(define Java-compiler "javac ")

;;; Just check that this variable is defined.

TARGET

(define Java-flags 
  (string-append " -g -classpath ${CLASSPATH}:"
                 TARGET ":. " ) )

(define-method (call-external-compiler (cmd Java-Compilation))
  (let* ((status 0)
         (cd-cmd (string-append 
                  "cd " 
                  (extract-directory (Java-Compilation-javafile cmd)) ) )
         (java-cmd (string-append Java-compiler
                                  Java-flags
                                  (extract-basefilename
                                   (Java-Compilation-javafile cmd) ) ))
         (complete-shell-command (string-append cd-cmd " ; " java-cmd)) )
    (set! status (system complete-shell-command))
    (unless (= (modulo status 256) 0)
      (compiler-main-error "Java compilation problem" status) )
    cmd ) )

(define compiler-main-error
  (create-compilation-exception-raiser 
   "Code generation" ) )

;;; Process the options.

(define (scan-options! cmd options)
  (define generatefile? #t)
  (set-Java-Compilation-calljavac?! cmd #t)
  (let scan ((options options))
    (if (pair? options)
        (cond ((equal? (car options) "-v")
               (set-Compilation-verbose?! cmd #t)
               (scan (cdr options)) )
              ((equal? (car options) "-nofile")
               (set! generatefile? #f)
               (scan (cdr options)) )
              ((equal? (car options) "-nojavac")
               (set-Java-Compilation-calljavac?! cmd #f)
               (scan (cdr options)) )
              (else (if (Compilation-source cmd)
                        (compiler-main-error "Too much file.scm" options)
                        (begin (set-Compilation-source! 
                                cmd (car options) )
                               (scan (cdr options)) ) )) ) ) )
  (unless (Compilation-source cmd)
    (compiler-main-error "No file to compile" cmd) )
  (when (and generatefile? (not (Java-Compilation-javafile cmd)))
    (set-Java-Compilation-javafile! 
     cmd (compute-default-generated-file cmd) ) )
  cmd )

;;; Compute by default the name of the file to be produced.

(define (compute-default-generated-file cmd)
  (let ((file (Compilation-source cmd)))
    (string-append (extract-filename-sans-extension file)
                   ".java" ) ) )

;;; If in Jaja (ie in Java) call directly the compiler.
;;; To use this option, say -nofile when calling compiler-main.

(define (call-internal-compiler string)
  string ) 

;;; end of javacomp.scm
