;;; $Id: javagen.scm,v 1.8 1997/12/09 19:47:20 queinnec Exp $
;;; By Christian.Queinnec@inria.fr (http://pauillac.inria.fr/~queinnec/)
;;; This file is part of a Scheme->Java compiler derived from the Scheme->C
;;; compiler described in the LiSP book. 

;;; Compilation towards Java.

;;; Enrich the class of Linearized-Flattened-Program to hold the
;;; original code to compile.

(define-class Java-Program Linearized-Flattened-Program
  ( original       ; original expression
    globalenv      ; the global environment that was used.
    ) )

;;;==================================================================
;;; This function performs the Java code generation. It takes a Sexp to
;;; compile, the current global environment and the output-port to write.
;;; Note that the expression to compile is supposed to already be expanded.

(define (compile->Java e g out)
  (monitor compile->java-handler
      (let ((prg (prepare-expression e g)))
        (->java prg out #f) ) ) )

;;; This is the general generic function to convert a program onto a
;;; port towards a Java class file.

(define-generic (->java (e) out tail?))

(define-method (->java (prg Java-Program) out tail?)
    (generate-header             prg out)
    (generate-global-environment prg out)
    (generate-quotations         prg out)
    (generate-main               prg out)
    (generate-functions          prg out)
    (generate-trailer            prg out)
    prg )

;;; Transform an Sexpression to a Java-Program.

(define (prepare-expression e g)
  (let ((o (objectify e (make-Local-Environment '()) g)))
    ;; Alphatize variable names
    (set! o (alphatize! o))
    ;; A-normalization
    (set! o (a-normalize o))
    ;; Add the Box transformation to remove mutable local variables:
    (set! o (insert-box! o))
    ;; Figure which global variables are surely initialized
    (set! o (initialization-analyze! o))
    ;; Closurize the whole program
    (set! o (make-Regular-Application
             (make-No-Argument)
             (make-Function '() o) ))
    ;; Lambda-lift in place (turn Function into Flat-Function)
    (set! o (lift! o))
    ;; Migrate quotations and functions towards the highest level.
    (set! o (extract-things! o))
    ;; linearize quotations
    (set! o (linearize-quotations! o))
    ;; simplify Sequences with a reference as first term
    (set! o (simplify-sequence! o))
    ;;  Return the result of all these transformations after recording
    ;;  the initial parameters of this compilation.
    (make-Java-Program
     (Flattened-Program-body o)
     (Flattened-Program-quotations o)
     (Flattened-Program-definitions o)
     (Linearized-Flattened-Program-linearizations o)
     (Linearized-Flattened-Program-initialization o)
     e
     g ) ) )

;;;==================================================================
;;; This is the error handler used while compiling. It catches the
;;; Unknown-Variable exception to define it on the fly but all other
;;; exceptions are re-raised.

(define-generic (compile->java-handler (exc) c?)
  (diagnose exc c?) )

;;; Define new global variables on the fly.

(define-method (compile->java-handler (exc Unknown-Variable) c?)
  (let ((v (make-Global-Variable (Unknown-Variable-name exc) #f))
        (g (Unknown-Variable-g exc)) )
    (set-Global-Environment-global!
     g (cons v (Global-Environment-Global g)) )
    (make-Global-Reference v) ) )

;;; Takes care of the exceptions raised by the initialization analysis.
;;; This means that the variable is by default not initialized.

(define-method (compile->java-handler (exc Uninitialized-Variable) c?)
  #f )

;;;==================================================================
;;;              Generation

;;; File header and trailer are simple.

(define (generate-header prg out)
  (let ((e (Java-Program-original prg)))
    (format out "// Scheme to Java $Revision: 1.8 $ ~%~%")
    (format out "/* Original program: ~%")
    (pp e out) ; DEBUG
    (format out " */~%")
    ;; Import Scheme runtime classes.
    (format out "~%
import SchemeRT.*; ~%
public class Main extends Scheme {~%") ) )
  
(define (generate-trailer prg out)
  (format out "~%// end.~%") )

;;;==================================================================
;;; Generate global environment. For all Global-Variables declare
;;; static variable in the Main class. This variable will hold a Box
;;; to hold its value. The reason is that there is no expression in
;;; Java to perform an assignment so I use a method for that.

;;; gv* is a list of instances of Global-Variable.

(define (generate-global-environment prg out)
  (let* ((globalenv (Java-Program-globalenv prg))
         (gv*       (Global-Environment-global globalenv)) )
    (when (pair? gv*)
      (format out "~%    // Global variables ~%")
      (for-each (lambda (gv) (generate-global-variable out gv))
                gv* )) ) )

(define (generate-global-variable out gv)
  (let ((name (Global-Variable-name gv)))
    (format out "    static SchemeValue ~A = new SchemeNamedBox(\"~A\"); ~%"
            (IdScheme->JavaId name) name ) ) )

;;; Convert a Scheme name to a Java name. All predefined names will be
;;; inserted in the following map so they can be uniformly named. The
;;; map is an Alist ( (symbol . "Cname") ... ).

(define Scheme->java-names-mapping 
  '( (*        . "times")
     (<        . "lessp")
     (pair?    . "consp")
     (set-cdr! . "rplacd")
     ) )

(set! Scheme->java-names-mapping 
  '( (*        . "times")
     (+        . "plus")
     (-        . "minus")
     (/        . "quotient")
     (>        . "gtp")
     (>=       . "gep")
     (<        . "ltp")
     (<=       . "lep")
     (=        . "eqnp")
     (eq?      . "eqp")
     (pair?    . "pairp")
     (symbol?  . "symbolp")
     (set-car! . "rplaca")
     (set-cdr! . "rplacd")
     ) )

(define (IdScheme->JavaId name)
  (let ((v (assq name Scheme->java-names-mapping)))
    (if (pair? v) (cdr v)
        (let ((str (string-lowercase (symbol->string name))))
          (let retry ((Javaname (compute-Javaname str)))
            (if (Javaname-clash? Javaname Scheme->java-names-mapping)
                (retry (compute-another-Javaname str))
                (begin (set! Scheme->java-names-mapping
                             (cons (cons name Javaname) 
                                   Scheme->java-names-mapping ) )
                       Javaname ) ) ) ) ) ) )

(define (Javaname-clash? Javaname mapping)
  (let check ((mapping mapping))
    (and (pair? mapping)
         (or (string=? Javaname (cdr (car mapping)))
             (check (cdr mapping)) ) ) ) )

;;; These functions compute a Java name for a symbol. Scheme symbols
;;; cannot be transleted into a name containing an isolated underscore
;;; so all these names will be used for Java generation purposes.

(define compute-another-Javaname 
  (let ((counter 1))
    (lambda (str)
      (set! counter (+ 1 counter))
      (compute-Javaname (format #f "~A_~A" str counter)) ) ) )

(define (compute-Javaname str)
  (define (mapcan f l)
    (if (pair? l)
        (append (f (car l)) (mapcan f (cdr l)))
        '() ) )
  (define (convert-char char)
    (case char
      ((#\-)             '(#\_))
      ((#\?)             '(#\p))
      ((#\!)             '(#\i))
      ((#\<)             '(#\l))
      ((#\>)             '(#\g))
      ((#\=)             '(#\e))
      ((#\/ #\* #\:)     '())
      (else              (list char)) ) )
  (let ((Javaname (mapcan convert-char (string->list str))))
    (if (pair? Javaname) (list->string Javaname) "weird") ) )

;;;==================================================================
;;; To ease writing printing expressions.

(define-icslas-macro (between-parentheses out . body)
  (let ((g (gensym)))
    `(let ((,g ,out))
       (format ,g "(")
       (begin . ,body)
       (format ,g ")") ) ) )

(define-icslas-macro (within-block out . body)
  (let ((g (gensym)))
    `(let ((,g ,out))
       (format ,g " { ~%")
       (begin . ,body)
       (format ,g "~%}~%") ) ) )

(define-icslas-macro (with-possible-return tail? out . body)
  (let ((t (gensym))
        (g (gensym)) )
    `(let ((,t ,tail?)
           (,g ,out) )
       (if ,t (format ,g "return "))
       (begin . ,body)
       (if ,t (format ,g ";")) ) ) )

;;;==================================================================
;;; Generate static shared quotations.

(define (generate-quotations prg out)
  (let ((lqv* (Linearized-Flattened-Program-linearizations prg)))
    (when (pair? lqv*)
      (format out "~%    // Quotations ~%")
      (for-each (lambda (lqv) (->java lqv out #f))
                lqv* )
      ;; since lqv* is a pair, the initialization is not #f
      (generate-quotations-aliases 
       (Linearized-Flattened-Program-initialization prg)
       out ) ) ) )

(define-generic (generate-quotations-aliases (form) out))

(define-method (generate-quotations-aliases (form Sequence) out)
  (generate-quotations-aliases (Sequence-first form) out)
  (generate-quotations-aliases (Sequence-last form) out) )

(define-method (generate-quotations-aliases (form Global-Assignment) out)
  (let ((qv  (Global-Assignment-variable form))
        (lqv (Reference-variable (Global-Assignment-form form))) )
    (format out "    protected static SchemeValue ")
    ;; not declared with the name under which it will be referenced.
    ;; so don't use this: (variable->java qv out)
    (format out "quote~A" (Variable-name qv))
    (format out " = ")
    (variable->java lqv out)
    (format out ";~%") ) )

(define-method (->java (qv Symbol-Quotation-Variable) out tail?)
  (let ((strqv (Symbol-Quotation-Variable-string qv)))
    (format out "    protected private static SchemeValue ")
    (variable->java qv out)
    (format out " = SchemeSymbol.create(")
    (variable->java strqv out)
    (format out "); ~%") ) )

(define-method (->java (qv Cons-Quotation-Variable) out tail?)
  (let ((aqv (Cons-Quotation-Variable-car qv))
        (dqv (Cons-Quotation-Variable-cdr qv)) )
    (format out "    protected private static SchemeValue ")
    (variable->java qv out)
    (format out " = SchemeProcedure.cons(")
    (variable->java aqv out)
    (format out ", ")
    (variable->java dqv out)
    (format out "); ~%") ) )

(define-method (->java (qv Immediate-Quotation-Variable) out tail?)
  (let ((value (Quotation-Variable-value qv)))
    (format out "    protected private static SchemeValue ")
    (variable->java qv out)
    (cond ((null? value)
           (format out " = NIL; ~%") )
          ((boolean? value)
           (format out " = ~A; ~%" (if value "TRUE" "FALSE") value) )
          ((integer? value)
           (format out " = SchemeFixnum.create(~A); ~%" value) )
          ((string? value)
           (format out " = new SchemeString(")
           (chop-string-for-java value out)
           (format out "); ~%") )
          ((char? value)
           (format out " = SchemeCharacter.create(~A); ~%" 
                   (char->integer value) ) )
          (else (compiler-error "Unknown quotation value" value)) ) ) )

(define (chop-string-for-java string out)
  (write-char #\" out)
  (let ((len (string-length string)))
    (do ((i 0 (+ i 1)))
        ((>= i len))
      (let ((ch (string-ref string i)))
        (if (char=? ch #\newline) 
            (format out "\\n")
            (write-char ch out) ) ) ) )
  (write-char #\" out) )

;;; RESTRICTION: strings in Java cannot span multiple lines. That's
;;; why the previous generation is insufficient. However, this feature
;;; is required to compile the compiler. So this is fixed by the
;;; chop-string-for-java function.

(define-method (->java (qv Vector-Quotation-Variable) out tail?)
  (let ((terms (Vector-Quotation-Variable-terms qv)))
    (format out "    protected private static SchemeValue ")
    (variable->java qv out)
    (format out " = new SchemeVector(~A);~%    static" (length terms))
    (within-block out
      (let loop ((terms terms)
                 (i 0) )
        (when (pair? terms)
          (format out "    ((SchemeVector)")
          (variable->java qv out)
          (format out ").item[~A] = " i)
          (variable->java (car terms) out)
          (format out ";~%")
          (loop (cdr terms) (+ i 1)) ) ) ) ) )

;;;==================================================================
;;; Convert a variable to Java

(define-generic (variable->java (e Variable) out))

(define-method (variable->java (e Renamed-Local-Variable) out)
  (call-next-method)
  (format out "_~A" (Renamed-Local-Variable-index e)) )

(define-method (variable->java (e Variable) out)
  (let* ((name (Variable-name e))
         (str  (IdScheme->JavaId name)) )
    (format out "~A" str) ) )

(define-method (variable->java (e Global-Variable) out)
  (let* ((name (Variable-name e))
         (str  (IdScheme->JavaId name)) )
    (format out "((SchemeBox)Main.~A)" str) ) )

;;; Quotations are defined under a name and used with a qualified one.

(define-method (variable->java (e Quotation-Variable) out)
  (format out "Main.quote~A" (Variable-name e)) )

(define-method (variable->java (e Linearized-Quotation-Variable) out)
  (format out "thing~A" (Variable-name e)) )

;;;==================================================================
;;; Convert an expression to Java

(define-method (->java (e Reference) out tail?)
  (with-possible-return tail? out
    (reference->java (Reference-variable e) out) ) )

(define-generic (reference->java (e Variable) out)
  (variable->java e out) )

(define-method (reference->java (gv Global-Variable) out)
  (let* ((name (Variable-name gv))
         (str  (IdScheme->JavaId name)) )
    (variable->java gv out)
    (format out ".getBoxContent()") ) )

(define-method (->java (e Global-Assignment) out tail?)
  (with-possible-return tail? out
    (variable->java (Global-Assignment-variable e) out)
    (format out ".setBoxContent")
    (between-parentheses out
      (->java (Global-Assignment-form e) out #f) ) ) )

(define-method (->java (e Box-Read) out tail?)
  (with-possible-return tail? out
    (format out "((SchemeBox)")
    (->java (Box-Read-reference e) out #f)
    (format out ").getBoxContent()") ) )

(define-method (->java (e Box-Write) out tail?)
  (with-possible-return tail? out
    (format out "((SchemeBox)")
    (->java (Box-Write-reference e) out #f)
    (format out ").setBoxContent")
    (between-parentheses out
      (->java (Box-Write-form e) out #f) ) ) )

(define-method (->java (e Box-Creation) out tail?)
  (variable->java (Box-Creation-variable e) out)
  (format out " = new SchemeBox(")
  (variable->java (Box-Creation-variable e) out)
  (format out ");~%")
  (->java (Box-Creation-body e) out tail?) )

;;;  distinguer qqs  cas comme  eq?, ou pair?

(define-method (->java (e Alternative) out tail?)
  (format out "if ")
  (boolean->java (Alternative-condition e) out)
  (within-block out
    (->java (Alternative-consequent e) out tail?) )
  (format out " else ")
  (within-block out
    (->java (Alternative-alternant e) out tail?) ) )

(define-generic (boolean->java (e) out)
  (between-parentheses out
    (->java e out #f)
    (format out " != Scheme.FALSE") ) )

;;; After the A-normalization, followed by the analysis that simplifies
;;; Sequences, the first term of a sequence can only be a reference
;;; to a variable (or a Box-reading) that is not surely initialized.
;;; NOTE: a reference to a variable is not a legal instruction in Java
;;; so we add this noop method.

(define-method (->java (e Sequence) out tail?)
  (->java (Sequence-first e) out #f)
  (format out ";~%")
  (->java (Sequence-last e) out tail?) )

;;; NOTE: an expression such as { exp, exp, ... } to create and
;;; initialize an array is not a legal expression in Java. Annoying!

;;; RESTRICTION: the compiler only supports functions with less than 
;;; ten arguments.

(define-method (->java (e Regular-Application) out tail?)
  (let ((n (number-of (Regular-Application-arguments e))))
    (with-possible-return tail? out
      (cond
       ((< n 11)
        (format out "((SchemeProcedure)")
        (->java (Regular-Application-function e) out #f)
        (format out ").invoke")
        (between-parentheses out
          (->java (Regular-Application-arguments e) out #f) ) )
       (else 
        (compiler-error "No more than 10 variables for Functions") ) ) ) ) )

(define-method (->java (e Arguments) out tail?)
  (->java (Arguments-first e) out #f)
  (if (Arguments? (Arguments-others e))
      (begin (format out ", ")
             (->java (Arguments-others e) out #f) ) ) )

(define-method (->java (e No-Argument) out tail?)
  #t )

(define-method (->java (e Fix-Let) out tail?)
  (within-block
    (local-bindings->java (Fix-Let-variables e) (Fix-Let-arguments e) out)
    (->java (Fix-Let-body e) out tail?) ) )

(define-generic (local-bindings->java variables (arguments) out))

(define-method (local-bindings->java variables (e Arguments) out)
  (format out "SchemeValue ")
  (variable->java (car variables) out)
  (format out " = ")
  (->java (Arguments-first e) out #f)
  (format out ";~%")
  (local-bindings->java (cdr variables) (Arguments-others e) out) )

(define-method (local-bindings->java variables (e No-Argument) out)
  #f )

(define-method (->java (e Predefined-Application) out tail?)
  (with-possible-return tail? out
    ((Functional-Description-for-Java-generator
      (Predefined-Variable-description
       (Predefined-Application-variable e) ) ) e out ) ) )

;;; This generates a direct call to the method of the SchemeProcedure
;;; class. This is the fastest way to call it apart inlining the
;;; function.

(define (make-predefined-application-generator javaname)
  (lambda (e out)
    (format out "~A" javaname)
    (between-parentheses out
      (->java (Predefined-Application-arguments e) out #f) ) ) )

;;; For n-ary predefined primitives, this one will use the dispatching
;;; invoker on the SchemeSubrN object which is the value of the
;;; eponymous variable of the Scheme class. 

(define (make-predefined-nary-application-generator javaname)
  (lambda (e out)
    (format out "Scheme.~A.invoke" javaname)
    (between-parentheses out
      (->java (Predefined-Application-arguments e) out #f) ) ) )

(define-method (->java (e Closure-Creation) out tail?)
  (with-possible-return tail? out
    (format out "new Closure~A" 
            (Function-Definition-index
             (Closure-Creation-definition e) ) )
    (between-parentheses out
      (->java (Closure-Creation-free e) out #f) ) ) )

(define-method (->java (e No-Free) out tail?)
  #t )

(define-method (->java (e Free-Environment) out tail?)
  (->java (Free-Environment-first e) out #f)
  (if (Free-Environment? (Free-Environment-others e))
      (begin 
        (format out ", ")
        (->java (Free-Environment-others e) out #f) ) ) )

;;; NOTE: should abstract over these iterators (as in lisp2tex).

;;; TEMP: monitor is compiled as begin.                      FUTURE

(define-method (->java (e Monitoring) out tail?)
  (->java (Monitoring-body e) out #f) )

;;;==================================================================
;;; Function definitions

(define (generate-functions prg out)
  (let ((definitions (Flattened-Program-definitions prg)))
    (format out "~%// Closures ~%")
    (for-each (lambda (def)
                (generate-closure-class def out) )
              definitions ) ) )

(define (generate-closure-class definition out)
  (format out "final class Closure~A extends "
          (Function-Definition-index definition) )
  ;; Functions with dotted variable inherits from SchemeProcedureN instead.
  (if (function-is-dotted? definition)
      (format out "SchemeProcedureN")
      (format out "SchemeProcedure") )
  (within-block out
    (generate-closure-free-variables definition out)
    (generate-closure-constructor definition out)
    (generate-closure-behavior definition out) ) )

(define (function-is-dotted? definition)
  (let loop ((vars (Function-Definition-variables definition)))
    (if (pair? vars)
        (if (Local-Variable-dotted? (car vars))
            #t
            (loop (cdr vars)) )
        #f ) ) )

(define (generate-closure-free-variables definition out)
  (format out "~%    // Free variables~%")
  (for-each (lambda (fv)
              (format out "    SchemeValue ")
              (variable->java fv out)
              (format out ";~%") )
            (Function-Definition-free definition) ) )

(define (generate-closure-constructor definition out)
  (format out "~%    // Constructor~%    public Closure~A "
          (Function-Definition-index definition) )
  (between-parentheses out
    (let loop ((fv* (Function-Definition-free definition))
               (i   0) )
      (when (pair? fv*)
        (format out "SchemeValue f~A" i)
        (when (pair? (cdr fv*))
          (format out ", ")
          (loop (cdr fv*) (+ i 1)) ) ) ) )
  (within-block out
    (let loop ((fv* (Function-Definition-free definition))
               (i   0) )
      (when (pair? fv*)
        (format out "        ")
        (variable->java (car fv*) out)
        (format out " = f~A;~%" i)
        (loop (cdr fv*) (+ i 1)) ) ) ) )

(define (generate-closure-behavior definition out)
  (format out "~%    // Behavior~%    public SchemeValue invoke ")
  (if (function-is-dotted? definition)
      (generated-dotted-closure-behavior definition out)
      (if (< (length (Function-Definition-variables definition)) 4)
          (generate-regular-closure-behavior definition out)
          (generate-big-closure-behavior definition out) ) ) )

(define (generate-regular-closure-behavior definition out)
  (let ((variables (Function-Definition-variables definition)))
    (between-parentheses out
      (let loop ((variables variables))
        (when (pair? variables)
          (format out "SchemeValue ")
          (variable->java (car variables) out)
          (when (pair? (cdr variables))
            (format out ", ")
            (loop (cdr variables)) ) ) ) )
    (within-block out
      (->java (Function-Definition-body definition) out #t) ) ) )

(define (generate-big-closure-behavior definition out)
  (let ((variables (Function-Definition-variables definition)))
    (format out "(SchemeValue arguments[])")
    (within-block out
      (format out "if (arguments.length != ~A) {
throw new RuntimeException(\"Wrong arity\");
}; ~%" (length variables) )
      (let loop ((variables variables)
                 (i 0) )
        (when (pair? variables)
          (format out "SchemeValue ")
          (variable->java (car variables) out)
          (format out " = arguments[~A];~%" i)
          (loop (cdr variables) (+ i 1)) ) )
      (->java (Function-Definition-body definition) out #t) ) ) )

(define (generated-dotted-closure-behavior definition out)
  (let ((variables (Function-Definition-variables definition)))
    (format out "(SchemeValue arguments[])")
    (within-block out
      (format out "if (arguments.length < ~A-1) {
throw new RuntimeException(\"Wrong arity\");
}; ~%" (length variables) )
      (let loop ((variables variables)
                 (i 0) )
        (when (pair? variables)
          (format out "SchemeValue ")
          (variable->java (car variables) out)
          (if (Local-Variable-dotted? (car variables))
              (format out " = this.listify(arguments, ~A);~%" i)
              (format out " = arguments[~A];~%" i) )
          (loop (cdr variables) (+ i 1)) ) )
      (->java (Function-Definition-body definition) out #t) ) ) )

;;;==================================================================
;;; Main procedure. We choose to print the result. This is easier for debug
;;; but might have been conditionalized by an option on the command line
;;; starting the Main.class with java.

(define (generate-main prg out)
  ;; The Main class.
  (format out "")
  (let ((form (Flattened-Program-body prg)))
    (format out "~%
    // Initialization
    public static void main (String argv[]) {
        try {
            SchemeValue r = " )
    (->java form out #f)
    (format out ";
            r.write();
            System.out.print(\"\\n\");
        } catch (Exception e) {
            System.out.print(\"Uncaught Exception: \" + e.getMessage() + \"\\n\");
            System.exit(1);
        }
        System.exit(0);
    }
}~%" ) ) )

;;; end of javagen.scm
