;;; $Id: read.scm,v 1.7 1998/06/10 16:31:32 queinnec Exp $

;;;                  A little reader
;;; This code is based on the reader of LiSP2TeX: 
;;; The only primitives that are needed to run this file are
;;; read-char and peek-char. Both manages characters.

;;; This file contains a Scheme Reader which I hope is conformant to 
;;; Scheme Standard (but for float and complex numbers). By default,
;;; this reader is case-sensitive and named read-cs.

;;;==================================================================
;;; Needs: peek-char, read-char, eof-object?
;;;   FUTURE make-Reader-exception ???              FUTURE

;;; Exports variable: read-cs

;;;==================================================================
;;; The final fuction, the sole entry point. Not to be confused with read.
;;; Note that it is unary and always takes an input port as argument.

(define (read-cs . in)
  (let ((in (if (pair? in) (car in) (current-input-port))))
    (skip-and-read-object (peek-char in) in) ) )

;;; You may change this function if you prefer to have a
;;; case-insensitive read function. 

(define preferred-case (lambda (ch) ch))

;;;==================================================================
;;; Some error messages when reading unparsable things.

(define too-much-closing-bracket
  "Superfluous closing bracket" )
(define premature-eof-in-string
  "Stream ends up while reading a string" )
(define premature-eof-in-symbol
  "Stream ends up while reading a symbol" )
(define premature-eof-in-barred-symbol
  "Stream ends up while reading a barred symbol" )
(define premature-eof-in-list
  "Stream ends up while reading a list" )
(define incorrect-dots
  "Two contiguous dots" )
(define too-long-dotted-list
  "Incorrect dotted (too long) list" )
(define too-short-dotted-list
  "Incorrect dotted (too short) list" )
(define premature-eof-in-sharp
  "Stream ends up while reading a sharp character" )
(define incorrect-character-name
  "Incorrect name of character" )
(define invalid-dispatch-character
  "Invalid dispatch character" )
(define premature-eof-in-vector
  "Stream ends up while reading a vector" )
(define dot-in-vector
  "Vectors cannot be dotted" )
(define premature-eof-in-number
  "Stream ends up while reading a number" )
(define incorrect-figure-in-number
  "Incorrect figure while reading number" )

;; Named characters that are recognized. Case is not sigificant.

(define named-characters
  '( ("space"   . 32)
     ("tab"     . 9)
     ("newline" . 10)
     ;; addition to R4RS:
     ("return"  . 13)
     ("bell"    . 7)
     ("delete"  . 127)
     ) )

;;;==================================================================
;;; The functions read-<something> all have a similar structure:
;;; they take a character (or eof) as first argument, the
;;; stream from which they read and they return whatever 
;;; object is read. 
;;; They are usually programmed as a test against eof followed
;;; by a big case expression driven by the current character.

;;; Whenever an end of file is found, the reader tries to finish
;;; its work and usually return this end-of-file or report an error
;;; if the read object is not complete.

;;; Symbols and strings are built by first transfering characters to a
;;; common zone. This zone is implemented as a non-extensible string
;;; associated to an index (symbols and strings are therefore limited
;;; to 1024 characters whereas in ~/scm/LiSP strings were automatically
;;; extended).

(define *stream-to-zone-alist* (list))

;;; Create the zone if not yet done. This is a quick and dirty hack since
;;; zones are not freed.

(define (reset-zone! stream)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (set-car! (cdr zone) 0)
        (set! *stream-to-zone-alist*
              (cons (cons stream (cons 0 (make-string 1024)))
                    *stream-to-zone-alist* ) ) ) ) )

;;; Return the ith character accumulated in the zone.

(define (zone-read-char stream i)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (string-ref (cddr zone) i)
        (error "No such zone") ) ) )

;;; Check if the zone has less characters than i.

(define (zone-end? stream i)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (<= (cadr zone) i)
        (error "No such zone") ) ) )

;;; Add a character to an extensible string. If the string within the
;;; extensible string is not so big as to contain this new character
;;; then extend it automatically.

(define (increment-string! char stream)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (let ((i (cadr zone)))
          (string-set! (cddr zone) i char)
          (set-car! (cdr zone) (+ i 1))
          char )
        (error "No such zone") ) ) )

;;; Add a character to a symbol under construction.
;;; The boolean `respect' if true imposes that the character will not be
;;; turned to upper or lower case.

(define (increment-symbol! char stream respect)
  (increment-string! (respect char) stream) )

;;; Takes an extensible string and turns its content into a symbol.

(define (finalize-symbol stream)
  (string->symbol (finalize-string stream)) )

(define (finalize-string stream)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (substring (cddr zone) 0 (cadr zone))
        (error "No such zone") ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; pipe two operations in one: consume the head of the stream 
;;; and peek the next character. At any time the first character of 
;;; stream is the poken char.

(define (read-then-peek-char stream)
  (read-char stream)
  (peek-char stream) )
    
;;; Two operations in one: skip whitespaces then read an object. 

(define (skip-and-read-object char stream)
  (read-object (skip-until-object char stream) stream) )

;;; skip whitespace and comments until finding a character that
;;; begins an expression, return that very character.

(define (skip-until-object char stream)
  (if (eof-object? char)
      char
      (case char
        ((#\space #\newline #\tab)      ; #\tab is not in R3RS
         (skip-until-object (read-then-peek-char stream) stream) )
        ((#\;)
         (read-after-end-of-line 
          (read-then-peek-char stream) stream skip-until-object ) )
        (else char) ) ) )
 
;;; Read an object, the first character of which is char.

(define (read-object char stream)
  (if (eof-object? char)
      char
      (case char
        ((#\')
         (let ((o (skip-and-read-object 
                   (read-then-peek-char stream) stream )))
           (if (eof-object? o) o `(quote ,o)) ) )
        ((#\`)
         (let ((o (skip-and-read-object 
                   (read-then-peek-char stream) stream )))
           (if (eof-object? o) o (list 'quasiquote o)) ) )
        ((#\,)
         (let ((ch (read-then-peek-char stream)))
           (case ch
             ((#\@)                     ; ,@ must be contiguous
              (let ((o (skip-and-read-object (read-then-peek-char stream) 
                                             stream )))
                (if (eof-object? o) o 
                    (list 'unquote-splicing o) ) ) )
             (else 
              (let ((o (skip-and-read-object ch stream)))
                (if (eof-object? o) o 
                    (list 'unquote o )) ) ) ) ) )
        ((#\()
         (skip-and-read-list (read-then-peek-char stream) stream) )
        ((#\))
         ;; consume the offending parenthesis and report an error.
         (read-char stream)		
         (read-error stream too-much-closing-bracket) )
        ((#\")
         (reset-zone! stream)
         (read-rope (read-then-peek-char stream) 
                    stream ) )
        ((#\#)
         (read-sharp-object (read-then-peek-char stream) stream) )
        ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
         (reset-zone! stream)
         (read-number-or-symbol char stream #f #t) )
        ((#\+ #\-)
         (reset-zone! stream)
         (increment-string! char stream)
         (read-number-or-symbol 
          (read-then-peek-char stream) 
          stream 
          char 
          #f ) )
        (else
         (reset-zone! stream)
         (read-symbol char stream) ) ) ) )

;;; skip characters until the end of the line then resume the
;;; fn function on the next character of the next line.

(define (read-after-end-of-line char stream fn)
  (if (eof-object? char)
      char
      (case char
        ((#\newline)
         (fn (read-then-peek-char stream) stream) )
        (else (read-after-end-of-line 
               (read-then-peek-char stream) stream fn )) ) ) )

;;; Read a string. Characters are accumulated in the common area
;;; using the increment-string! function.  When the string is
;;; finished the common area is finalized to return a string.

(define (read-rope char stream)
  (if (eof-object? char)
      (read-error stream premature-eof-in-string)
      (case char
        ((#\") (read-char stream)
               (finalize-string stream) )
        ((#\\)
         (read-char stream)             ; consume the backslash
         (let ((ch (read-char stream)))
           (if (eof-object? ch)
               (read-error stream premature-eof-in-string) )
           (increment-string! ch stream)
           (read-rope (peek-char stream) stream) ) )
        (else (increment-string! char stream)
              (read-rope 
               (read-then-peek-char stream) stream ) ) ) ) )

;;; Read a symbol (backslash and bar are allowed).
;;; Note that dot and quote do not finish a symbol.

(define (read-symbol char stream)
  (if (eof-object? char)
      (finalize-symbol stream)
      (case char
        ((#\space #\newline #\tab #\( #\) #\; #\") ; #\tab is not in R3RS
         (finalize-symbol stream) )
        ((#\\)
         (read-char stream)             ; consume the backslash
         (let ((ch (read-char stream)))
           (if (eof-object? ch)
               (read-error stream premature-eof-in-symbol) )
           ;; after a backslash the case is respected.
           (increment-symbol! ch stream (lambda (x) x))
           (read-symbol (peek-char stream) stream) ) )
        ((#\|)
         (read-barred-symbol (read-then-peek-char stream) 
                             stream ) )
        (else 
         (increment-symbol! char stream preferred-case)
         (read-symbol
          (read-then-peek-char stream) stream ) ) ) ) )

;;; Read the part of a symbol which is between bars.

(define (read-barred-symbol char stream)
  (if (eof-object? char)
      (read-error stream premature-eof-in-barred-symbol)
      (case char
        ((#\|) 
         (let ((newchar (read-then-peek-char stream)))
           (if (eof-object? newchar)
               (read-error stream premature-eof-in-barred-symbol) )
           (case newchar
             ((#\|)
              (increment-symbol! newchar stream (lambda (x) x))
              (read-barred-symbol (read-then-peek-char stream)
                                  stream ) )
             (else (read-symbol newchar stream)) ) ) )
        ((#\\)
         (read-char stream)             ; consume the backslash
         (let ((ch (read-char stream)))
           (if (eof-object? ch)
               (read-error stream premature-eof-in-symbol) )
           ;; after a backslash the case is respected.
           (increment-symbol! ch stream (lambda (x) x))
           (read-barred-symbol 
            (peek-char stream) stream) ) )
        (else (increment-symbol! char stream preferred-case)
              (read-barred-symbol
               (read-then-peek-char stream) stream ) ) ) ) )

;;; skip whitespaces to read the content of a list

(define (skip-and-read-list char stream)
  (if (eof-object? char)
      (read-error stream premature-eof-in-list)
      (case char
        ((#\space #\newline #\tab)
         (skip-and-read-list (read-then-peek-char stream) stream) )
        ((#\;)
         (read-after-end-of-line
          (read-then-peek-char stream) stream skip-and-read-list ) )
        (else (read-list char stream)) ) ) )
    
;;; Reads a list

(define (read-list char stream)
  (case char
    ((#\))
     (read-char stream)                 ; consume the close bracket
     '() )
    ((#\.)
     (let ((ch (read-then-peek-char stream)))
       (if (eof-object? ch)
           (read-error stream premature-eof-in-list)
           (case ch
             ((#\.)
              (let ((other-ch (read-then-peek-char stream)))
                (if (eof-object? other-ch)
                    (read-error stream premature-eof-in-list)
                    (case other-ch
                      ((#\.)
                       (read-char stream) ; consumes third dot
                       (reset-zone! stream)
                       (increment-symbol! ch stream preferred-case)
                       (increment-symbol! ch stream preferred-case)
                       (increment-symbol! ch stream preferred-case)
                       (let* ((s (finalize-symbol stream))
                              (d (skip-and-read-list (peek-char stream)
                                                      stream )) )
                         (if (eof-object? d)
                             d
                             (cons s d) ) ) )
                      (else (read-error stream incorrect-dots)) ) ) ) )
             (else
              (read-end-of-list 
               (skip-and-read-list ch stream) 
               stream ) ) ) ) ) )
    (else 
     (let ((a (skip-and-read-object char stream)))
       (if (eof-object? a)
           a
           (let ((d (skip-and-read-list (peek-char stream) stream)))
             (if (eof-object? d)
                 d
                 (cons a d) ) ) ) ) ) ) )

;;; Check if what was read is a regular end of list ie is a list
;;; with exactly one term otherwise report an error.

(define (read-end-of-list expressions stream)
  (if (pair? expressions)
      (if (pair? (cdr expressions))
          (read-error stream too-long-dotted-list expressions)
          (car expressions) )
      (read-error stream too-short-dotted-list expressions) ) )

;;; Read a sharp notation

(define (read-sharp-object char stream)
  (case char
    ((#\\)
     (read-until-delimiter
      (read-then-peek-char stream)
      stream
      (reset-common-zone)
      (lambda (chars delimiter)
        (case (string-length chars)
          ((0) (read-char stream))      ; consume the delimiter
          ((1) (string-ref chars 0))
          (else 
           (let ((value (let search ((nmchs named-characters))
                          (if (pair? nmchs)
                              (if (string-ci=? (caar nmchs) chars)
                                  (car nmchs)
                                  (search (cdr nmchs)) )
                              #f ) )))
             (if value 
                 (integer->char (cdr value))
                 (read-error 
                  stream incorrect-character-name chars ) ) )) ) ) ) )
    ((#\t #\T) (read-char stream) #t)
    ((#\f #\F) (read-char stream) #f)
    ((#\b #\B)
     (read-number (read-then-peek-char stream) stream 2 0) )
    ((#\o #\O)
     (read-number (read-then-peek-char stream) stream 8 0) )
    ((#\d #\D)
     (read-number (read-then-peek-char stream) stream 10 0) )
    ((#\x #\X)
     (read-number (read-then-peek-char stream) stream 16 0) )
    ((#\()
     (let ((items
            (skip-and-read-vector (read-then-peek-char stream) stream) ))
       (apply vector items) ) )
    (else (read-char stream)
          (read-error stream invalid-dispatch-character char) ) ) )

;;; Returns a string of all characters until a delimiter.
;;; Useful to read a named character.

(define (read-until-delimiter char stream cont)
  (if (eof-object? char)
      (read-error stream premature-eof-in-sharp)
      (case char
        ((#\( #\) #\; #\" #\space #\newline #\tab)
         (cont (finalize-string stream) char) )
        (else (increment-string! char stream)
              (read-until-delimiter
               (read-then-peek-char stream)
               stream
               cont )) ) ) )
           
;;; Skip whitespaces to read the content of a vector.

(define (skip-and-read-vector char stream)
  (if (eof-object? char)
      (read-error stream premature-eof-in-vector)     
      (case char
        ((#\space #\newline #\tab)
         (skip-and-read-vector (read-then-peek-char stream) stream) )
        ((#\;)
         (read-after-end-of-line
          (read-then-peek-char stream) stream skip-and-read-vector ) )
        (else (read-vector char stream)) ) ) )

;;; read a vector.

(define (read-vector char stream)
  (case char
    ((#\))
     (read-char stream)
     '() )
    ((#\.)
     (read-char stream)                 ; consume the dot
     (read-error stream dot-in-vector) )
    (else 
     (let ((item (skip-and-read-object char stream)))
       (cons item (skip-and-read-vector (peek-char stream) 
                                        stream )) ) ) ) )

;;; Read a number.

(define (read-number char stream base result)
  (if (eof-object? char)
      (or result (read-error stream premature-eof-in-number))
      (let ((figure (convert-figure char base 0)))
        (if figure
            (read-number (read-then-peek-char stream) 
                         stream base (+ figure (* base result)) )
            ;; numbers are finished by the first non-figure character 
            ;; This is questionable !?
            result ) ) ) )

;;; Convert a string to a number

(define (convert-figure char base result)
  (if (< result base)
      (if (char-ci=? char (string-ref "0123456789ABCDEF" result))
          result
          (convert-figure char base (+ 1 result)) )
      #f ) )

;;; It is not yet known if we are reading a number or a symbol.
;;; sign is #f or #\+ or #\- if the sign has been read
;;; number? is #t or #f if at least one figure has been read (useful 
;;; to read + as a symbol and +0 as zero).

(define (read-number-or-symbol char stream sign number?)
  (case char
    ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
     (increment-string! char stream)
     (read-number-or-symbol 
      (read-then-peek-char stream) stream sign #t ) )
    ((#\space #\newline #\tab #\( #\) #\;)
     (if number?
         (process-read-number stream sign)
         (read-symbol char stream) ) )
    (else
     (read-symbol char stream) ) ) )

;;; Convert a zone to a number.

(define (process-read-number stream sign)
  (let ((zone (assq stream *stream-to-zone-alist*)))
    (if (pair? zone)
        (do ((i      (if sign 1 0) (+ i 1))
             (char   'wait)
             (result 0) )
            ((zone-end? stream i)
             (case sign
               ((#\-) (- result))
               (else  result) ) )
          (set! char (zone-read-char stream i))
          (set! result (+ (convert-figure char 10 0) (* 10 result))) )
        (error "No such zone") ) ) )

;;; The entry point for the reader.
;;; If kind is 'read then read and return the read object.
;;; If kind is 'skip then skip whitespaces until an object starts.
;;; stream is the stream to read,

(define (standard-reader kind stream)
  (case kind
    ((read) (skip-and-read-object (peek-char stream) stream))
    ((skip) (skip-until-object (peek-char stream) stream))
    (else   (read-error stream "No such operation" kind)) ) )

;;; end of reader.scm
