;;; $Id: clexpand.scm,v 1.7 1997/11/25 19:40:52 queinnec Exp $

;;; Macros are difficult.
;;; This file provides a "neutral expander" which removes user macros
;;; but leaves untouched the other syntaxes such as let, let*, letrec,
;;; do and others. However it removes backquoted forms.  This
;;; macroexpander uses a classical walking model but expanded forms
;;; are not reexpanded unless the macro calls itself the macroexpander
;;; on its result. Note: this file does not use macros at all, only regular
;;; Scheme syntaxes (define, let ...).

;;; This file only defines the basic walking model. Macros to define
;;; macros and special macros are defined in the clmacros.scm
;;; file. This walking model assumes the walked expression to be
;;; syntactically correct.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; The basic algorithm to code-walk expressions.  W represents the
;;; walking environment. It is a semi-predicate that takes an
;;; expression and tells whether or not there is a specialized
;;; code-walker for it. It returns this appropriate code-walker to
;;; mean True. A code-walker is a function with signature (lambda (e
;;; w) e) where e is an expression and w a walking environment.

(define (walk e w)
  ;;(display `("about to expand" ,e) stderr)(newline stderr)
  (let ((walker (w e)))
    (if (procedure? walker)
        (walker e w)
        (if (pair? e)
            (walk-combination e w)
            e ) ) ) )

;;; A walking environment is a semi-predicate that takes a Scheme
;;; expression and returns the specialized code-walker for that
;;; expression. It is the duty of walk to invoke this code-walker on
;;; this expression (giving the walking environment itself as second
;;; argument as in EPS). Therefore a code-walker has to recursively
;;; walk the sub-expressions. The following walking environment only
;;; knows the regular Scheme syntaxes. It returns False if no code-walker
;;; is found so it is possible to inspect the walking environment to know
;;; if a macro is defined or not (this is a reflective capability).

(define (neutral-walking-env e)
  (and (pair? e)
       (case (car e)
         ((quote)      walk-quote)
         ((if)         walk-if)
         ((begin)      walk-begin)
         ((or)         walk-or)
         ((and)        walk-and)
         ((set!)       walk-set!)
         ((lambda)     walk-lambda)
         ((quasiquote) walk-quasiquote)
         ((let)        walk-let)
         ((let*)       walk-let*)
         ((letrec)     walk-letrec)
         ((define)     walk-define)
         ((cond)       walk-cond)
         ((delay)      walk-delay)
         ((case)       walk-case)
         ((do)         walk-do)
         ;; No specialized code-walker for e.
         (else         #f) ) ) )

;;; Extend a walking environment with a new macro. Keyword is the name
;;; of the new macro and expander is its (normal) expander. This
;;; latter has to be wrapped into a suitable form that restart the
;;; walking process on the expanded forms. This is the traditional
;;; macro expansion where the expanded result is reexpanded again.

(define (extend-with-repetitive-macro w keyword expander)
  (define (wrapped-expander ee ww)
    (walk (apply expander ww (cdr ee)) ww) )
  (lambda (e)
    (if (and (pair? e) (eq? (car e) keyword))
        wrapped-expander
        (w e) ) ) )

(define (extend-with-one-shot-macro w keyword expander)
  (define (wrapped-expander ee ww)
    (apply expander ww (cdr ee)) )
  (lambda (e)
    (if (and (pair? e) (eq? (car e) keyword))
        wrapped-expander
        (w e) ) ) )

;;; Extend a walking environment so that local variables (whether read
;;; or written) mask macros. The list of variables may be dotted so
;;; flatten it before using it. This function returns a walking
;;; environment that recognizes references or assignments to variables
;;; and leave them as they are. 

(define (extend-with-variables w variables)
  (define (flat spec)
    (cond ((null? spec) spec)
          ((symbol? spec) (list spec))
          ((pair? spec) (cons (car spec) (flat (cdr spec))))
          (else (walk-error "incorrect list of variables" spec)) ) )
  (let ((variables (flat variables)))
    (lambda (e)
      (cond ((and (symbol? e)
                  (memq e variables) )
             walk-leave-as-it-is )
            ((and (pair? e)
                  (eq? (car e) 'set!)
                  (pair? (cdr e))
                  (memq (cadr e) variables) )
             (lambda (ee ww)
               `(,(car ee) ,(cadr ee) ,(walk (caddr ee) ww)) ) )
            (else (w e)) ) ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; The various specialized code-walkers for standard Scheme.
;;; Different names are used for different syntaxes.
;;; All these code-walkers assume the walked code to be syntactically correct.

;;; Stop walking a quotation.

(define (walk-leave-as-it-is e w) e)

(define walk-quote walk-leave-as-it-is)

;;; Recursively walk a combination. Respect the last cdr.  Perform
;;; from left to right (so macros can be defined first and used right
;;; after).

(define (walk* e* w)
  (if (pair? e*)
      (let ((r (walk (car e*) w)))
        (cons r (walk* (cdr e*) w)) )
      e* ) )

(define walk-combination walk*)

;;; Recursively walk a special form but its first term (its name)
;;; This is used for a number of special forms.

(define (walk-special-form e w)
  (cons (car e) (walk* (cdr e) w)) )

(define walk-if walk-special-form)
(define walk-begin walk-special-form)
(define walk-or walk-special-form)
(define walk-and walk-special-form)

;;; Delays

(define (walk-delay e w)
  `(,(car e) ,(walk (cadr e) w)) )

;;; lambda is a binding form and bindings may shadow macros.  Pay
;;; attention to local define forms that may happen in a lambda form.

(define (walk-lambda e w)
  `(,(car e) ,(cadr e) 
       ,@(walk-definitions-and-terms 
          (cddr e) (extend-with-variables w (cadr e)) ) ) )

;;; Walk a series of expressions maybe preceded by a series of local
;;; definitions. This is used in lambda and let, let*, letrec forms.
;;; First, collect every local definitions to know the set of mutually
;;; local variables to extend the walking environment.

(define (walk-definitions-and-terms e* w)
  (let scan ((e* e*)
             (rev-definitions '()) )
    (if (and (pair? e*)
             (pair? (car e*))
             (eq? (caar e*) 'define) )
        (scan (cdr e*) (cons (car e*) rev-definitions))
        (let ((w (extend-with-variables
                  w (map (lambda (def) 
                           (find-defined-variable (cadr def)) )
                         rev-definitions ) )))
          (append (map (lambda (def) (walk-define def w))
                       (reverse rev-definitions) )
                  (walk* e* w) ) ) ) ) )

;;; Find the name that is defined in the second term of a define form.
;;; This is to suppport the (define (foo x) ...) syntax.

(define (find-defined-variable spec)
  (cond ((symbol? spec) spec)
        ((pair? spec) (find-defined-variable (car spec)))
        (else (walk-error "Incorrect defined name" spec)) ) )

(define (walk-define e w)
  (cond ((pair? (cadr e))
         `(,(car e) ,(cadr e)
             ,@(walk-definitions-and-terms 
                (cddr e) (extend-with-variables w (cadr e)) ) ) )
        ((symbol? (cadr e))
         `(,(car e) ,(cadr e)
              ,(walk (caddr e) w) ) )
        (else (walk-error "Incorrect define form" e)) ) )

;;; Walk an assignment. Observe in extend-with-variables that
;;; assignments may have specialized code-walkers. This will be needed
;;; for the with-fields macro of Meroon.

(define (walk-set! e w)
  `(,(car e) ,(cadr e) ,(walk (caddr e) w)) )

;;; Walk a let form. Pay attention to named let forms.

(define (walk-let e w)
  (if (symbol? (cadr e))
      `(,(car e) ,(cadr e)
         ,(let ((w (extend-with-variables w (list (cadr e)))))
            (map (lambda (binding) (walk-binding binding w))
                 (caddr e) ) )
         ,@(walk-definitions-and-terms 
            (cdddr e) (extend-with-variables w (map car (caddr e))) ) )
      `(,(car e)
        ,(map (lambda (binding) (walk-binding binding w))
              (cadr e) )
        ,@(walk-definitions-and-terms 
           (cddr e) (extend-with-variables w (map car (cadr e))) ) ) ) )

(define (walk-binding var+form w)
  (if (and (pair? var+form)
           (pair? (cdr var+form))
           (null? (cddr var+form)) )
      `(,(car var+form) ,(walk (cadr var+form) w))
      (walk-error "Incorrect local variables" var+form) ) )

;;; Walk a let* form.

(define (walk-let* e w)
  `(,(car e)
    ,@(let scan ((bindings (cadr e))
                 (new-bindings '())
                 (w w) )
        (if (pair? bindings)
            (scan (cdr bindings)
                  (append new-bindings (list (walk-binding (car bindings) w)))
                  (extend-with-variables w (list (caar bindings))) )
            (cons new-bindings
                  (walk-definitions-and-terms (cddr e) w) ) ) ) ) )

;;; Walk a letrec form.

(define (walk-letrec e w)
  (let ((w (extend-with-variables w (map car (cadr e)))))
    `(,(car e) 
      ,(map (lambda (binding) (walk-binding binding w))
            (cadr e) )
      ,@(walk-definitions-and-terms (cddr e) w) ) ) )

;;; Walk a case form.

(define (walk-case e w)
  `(,(car e) ,(walk (cadr e) w)
      ,@(map (lambda (clause)
               `(,(car clause) ,@(walk* (cdr clause) w)) )
             (cddr e) ) ) )

;;; Walk a cond form.

(define (walk-cond e w)
  `(,(car e) ,@(map (lambda (clause) (walk* clause w))
                    (cdr e) ) ) )

;;; Walk a do form.

(define (walk-do e w)
  `(,(car e)
    ,(map (lambda (locals)
            `(,(car locals) ,@(walk* (cdr locals) w)) )
          (cadr e) )
    ,@(walk* (cddr e) (extend-with-variables w (map car (cadr e)))) ) )

;;; Walk a quasiquotation.  Pay attention to the citation of unquote
;;; and unquote-splicing (ie do not replace the (or eq eq) form by a
;;; memq equivalent one) since this poses printing problems for many
;;; implementations. The succession . , at the end of a quasiquotation
;;; may pose problems in some implementations. Normally Meroon sources
;;; use ,@ instead.

(define (walk-quasiquote e w)
  (define (walk-quasiquotation-list e w)
    (if (pair? e)
        (if (or (eq? (car e) 'unquote)
                (eq? (car e) 'unquote-splicing) )
            `(,(car e) ,(walk (cadr e) w))
            (cons (walk-quasiquote (car e) w)
                  (walk-quasiquotation-list (cdr e) w) ))
        e ) )
  (if (pair? e)
      (if (or (eq? (car e) 'unquote)
              (eq? (car e) 'unquote-splicing) )
          `(,(car e) ,(walk (cadr e) w))
          (walk-quasiquotation-list e w) )
      e ) )

;;; Test:
;;; (walk 'a neutral-walking-env)
;;; (walk '(foo a 34) neutral-walking-env)
;;; (walk '(if a 34) neutral-walking-env)
;;; (walk '`(if ,a ,@b c) neutral-walking-env)
;;; (walk '`(if ,a ,@b . ,c) neutral-walking-env)
;;; (walk '(define (a) 33) neutral-walking-env)
;;; (walk '(define (a) (define b 4) 33) neutral-walking-env)
;;; (walk '(cond (a 1) (b 2 3)) neutral-walking-env)
;;; (walk '(case a ((b) 1) (else 2)) neutral-walking-env)

;;; end of clexpand.scm
