;;; $Id: clmacros.scm,v 1.11 1997/12/09 19:47:27 queinnec Exp $

;;; This file defines the additional macros needed by the clexpand.scm
;;; walking model. It customizes this expansion engine.
;;; In fact there are two customizations here:
;;; -- One is for the expansion of all Scheme derived syntaxes (it removes
;;;   let, letrec ... in favor of simple expressions).
;;; -- The other is for boot purpose, it only removes icslas macros (such
;;;   as when, unless but leaves all Scheme syntaxes as they stand so a
;;;   following compiler can take full benefit of a real Scheme program.
;;; You may use either macro environment through:
;;;         enriched-icslas-expansion-environment
;;;      or icslas-expansion-environment

;;;==================================================================
;;; Icslas define macros with define-icslas-macro.
;;; The definition of this macro is expanded in the current 
;;; macroexpansion environment.

(define (walk-define-icslas-macro e w)
  (let* ((call      (cadr e))
         (name      (car call))
         (variables (cdr call))
         (w         (extend-with-variables w variables))
         (body      (walk* (cddr e) w)) )
    ;; Build and evaluate the expander on the fly.    
    (let* ((expander `(lambda (macroexpand . ,variables) . ,body))
           (expanded-expander (walk expander w)) )
      (set! icslas-expansion-environment
            (extend-with-repetitive-macro
             icslas-expansion-environment 
             name 
             (expand-eval expanded-expander) ) )
      ;; Nothing is left after this definition.
      `(quote ,name) ) ) )

;;; This is the icslas expansion environment for bootstrap.
;;; Only one special macro is added to the initial walking environment:
;;; the define-icslas-macro which allows to define new macros on the fly.

(define icslas-expansion-environment 
  (lambda (e)
    (if (pair? e)
        (case (car e)
          ((define-icslas-macro) walk-define-icslas-macro)
          (else (neutral-walking-env e)) )
        (neutral-walking-env e) ) ) )

;;;==================================================================
;;; This is another walking environment that expands full Scheme into the
;;; essential Scheme. This global mutable variable will be enriched 
;;; with new expanders.

(define enriched-icslas-expansion-environment
  icslas-expansion-environment )

;;; REMARK: If you want to expand this code with itself you need to
;;; incorporate define-icslas-expander into
;;; enriched-icslas-expansion-environment and not in
;;; icslas-expansion-environment. Therefore you need to redefine:
;;; (define enriched-icslas-expansion-environment
;;;    (lambda (e) (icslas-expansion-environment e)) )
  
;;; This is a macro to add new macros on the previous environment.
;;; NOTE: When you define a macro with that form, it will be considered
;;; as an active macro only after the resulting code is evaluated. These
;;; macros are not defined on the fly (as with define-icslas-macro which
;;; therefore requires expand-eval).

(define-icslas-macro (define-icslas-expander call . body)
  (let ((name      (car call))
        (variables (cdr call)) )
    `(set! enriched-icslas-expansion-environment
           (extend-with-repetitive-macro 
            enriched-icslas-expansion-environment
            ',name
            (lambda (macroexpand . ,variables)
              ;;(display (cons "expanding" ',name) stderr)(newline stderr)
              . ,body ) ) ) ) )

;;; This string serves as a sentinel. If your target language offers
;;; the define-icslas-expander macro then you may start here.

"Expanders section"

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Other expanders for derived expression types (DP1178 on Scheme)
;;; NOTE: in the expanded expression might appear
;;;  (begin)
;;;  (uninitialized-let (variables...) body)
;;; That must be handled appropriately by the compiler (or additional macros).

;;; For now, since uninitialized-let is not yet in the compiler
;;; expand it into a let.                                       FUTURE

(define-icslas-expander (uninitialized-let variables . body)
  `(let ,(map (lambda (v) `(,v 543))
              variables )
     . ,body ) )

;;; This string serves as a sentinel. If your target language provides
;;; an uninitialized-let special form then you do not need the
;;; previous definition and you can skip the beginning of this file
;;; until this sentinel. If you want to expand files for a dialect
;;; that offers uninitialized-let then you should define a walker for
;;; this special form. Something like:
;;;  (set! enriched-icslas-expansion-environment
;;;        (extend-with-one-shot-macro
;;;         enriched-icslas-expansion-environment
;;;         'uninitialized-let
;;;         (lambda (macroexpand variables . body)
;;;           `(uninitialized-let ,variables 
;;;               . ,(walk* body macroexpand) ) ) ) )

"Expanders section after uninitialized-let"

;;; The lambda walker must handle its body but the result must not be
;;; rewalked.

(set! enriched-icslas-expansion-environment
      (extend-with-one-shot-macro 
       enriched-icslas-expansion-environment
       'lambda
       (lambda (macroexpand variables . body)
         `(lambda ,variables
            ,(walk (analyze-body body)
                   (extend-with-variables macroexpand variables) ) ) ) ) )

(define-icslas-expander (let* binds . body)
  (if (pair? binds)
      `(let (,(car binds))
         (let* ,(cdr binds) . ,body) )
      (analyze-body body) ) )

;;; Handle-sequence tries to simplify sequence of forms. It recognizes
;;; (begin form) and may also be applied on an empty sequence.

(define (handle-sequence body)
  (if (pair? body)
      (if (null? (cdr body))
          (car body)
          `(begin . ,body) )
      icslas-unspecified ) )

(define-icslas-expander (let binds . body)
  (if (symbol? binds)
      ;; Named let
      (let ((name binds)
            (binds (car body))
            (body (cdr body)) )
        (sort-bindings
         binds
         (lambda (inited uninited)
           (if (pair? uninited)
             (expand-error 
              "Named let is incompatible of uninitialized bindings" ) )
           `((letrec ((,name (lambda ,(map car binds)
                               . ,body )))
               ,name )
             . ,(map cadr binds) ) ) ) )
      ;; Anonymous let
      (sort-bindings 
       binds
       (lambda (inited uninited)
         (let* ((body (analyze-body body))
                (body (if (pair? uninited)
                          `(uninitialized-let ,uninited ,body)
                          body )) )
           (if (pair? inited)
               `((lambda ,(map car inited) ,body)
                 . ,(map cadr inited) )
               body ) ) ) ) ) )
             
;;; Sort the bindings of a let into two classes: these that are initialized
;;; and the others. This allows to not use the (too long) word
;;; uninitialized-let when it is not required: just write let!

(define (sort-bindings bindings k)
  (let scan ((bindings bindings)
             (inited '())
             (uninited '()) )
    (if (pair? bindings)
        (if (pair? (car bindings))
            (scan (cdr bindings) (cons (car bindings) inited) uninited)
            (scan (cdr bindings) inited (cons (car bindings) uninited)) )
        (k (reverse! inited)
           (reverse! uninited) ) ) ) )

(define (analyze-body body)
  (collect-definitions 
   body
   (lambda (definitions body)
     (if (null? definitions)
         (handle-sequence body)
         (let* ((assignments 
                 (map (lambda (definition)
                        (analyze-definition (cadr definition)
                                            (cddr definition) ) )
                      definitions ) )
                (names (map cadr assignments)) )
           `(letrec ,(map cdr assignments)
              ,(handle-sequence body) ) ) ) ) ) )

(define-icslas-expander (letrec binds . body)
  (if (pair? binds)
      `(uninitialized-let
        ,(map (lambda (bind) 
                (if (pair? bind) (car bind) bind) )
              binds )
        ,@(map (lambda (bind) 
                 (if (pair? bind) 
                     `(set! . ,bind)
                     `#t ) )
               binds )
        ,(analyze-body body) )
      (analyze-body body) ) )

;;; Define forms are turned into regular assignment since it is
;;; sufficient in Icslas to mention a global variable for it to exist
;;; (with an uninitialized status).

(define-icslas-expander (define call . body)
  (analyze-definition call body) )

(define (analyze-definition call body)
  (if (pair? call)
      `(set! ,(car call)
             (lambda ,(cdr call)
               ,(analyze-body body) ) )
      (if (and (pair? body)
               (null? (cdr body)) )
          `(set! ,call ,(car body))
          (expand-error 'incorrect-definition call) ) ) )

(define (collect-definitions e* k)
  (if (and (pair? e*) (pair? (car e*)) (eq? (caar e*) 'define))
      (collect-definitions 
       (cdr e*)
       (lambda (bindings body)
         (k (cons (car e*) bindings) body) ) )
      (k '() e*) ) )

(define-icslas-expander (and . arguments)
  (if (pair? arguments)
      (if (pair? (cdr arguments))
          `(if ,(car arguments)
               (and . ,(cdr arguments))
               #f )
          (car arguments) )
      '#t ) )

(define-icslas-expander (or . arguments)
  (if (pair? arguments)
      (if (pair? (cdr arguments))
          (let ((g (gensym)))
            `((lambda (,g)
                (if ,g ,g (or . ,(cdr arguments))) )
              ,(car arguments) ) )
          (car arguments) )
      '#f ) )

(define-icslas-expander (cond . arguments)
  (if (pair? arguments)
      (if (pair? (car arguments))
          (if (eq? (car (car arguments)) 'else)
              (if (null? (cdr arguments))
                  (handle-sequence (cdr (car arguments)))
                  (expand-error 'cond-else-not-last) )
              `(if ,(car (car arguments))
                   ,(handle-sequence (cdr (car arguments)))
                   (cond . ,(cdr arguments)) ) )
          (expand-error 'cond-1st-clause) )
      icslas-unspecified ) )

(define-icslas-expander (begin0 first . others)
  (let ((g (gensym)))
    `((lambda (,g) 
        ,(handle-sequence others)
        ,g )
      ,first ) ) )

;;; This is used when no value is particularly interesting to return
;;; in the expansion. The value of icslas-unspecified must be a legal
;;; Icslas expression that can be legally evaluated in the Icslas
;;; runtime environment.

(define icslas-unspecified '(quote 813))
(set! icslas-unspecified icslas-unspecified)

(define-icslas-expander (case exp . clauses)
  (let ((g (gensym)))
    (define (generate clauses)
      (if (pair? clauses)
          (if (pair? (car clauses))
              (if (eq? (caar clauses) 'else)
                  (if (null? (cdr clauses))
                      (handle-sequence (cdar clauses))
                      (expand-error 'case-else-not-last clauses) )
                  (if (pair? (caar clauses))
                      (if (null? (cdaar clauses))
                          `(if (eqv? ,g ',(caaar clauses))
                               ,(handle-sequence (cdar clauses))
                               ,(generate (cdr clauses)) )
                          `(if (memq ,g ',(caar clauses))
                               ,(handle-sequence (cdar clauses))
                               ,(generate (cdr clauses)) ) )
                      (expand-error 'case-incorrect-clause) ) )
              (expand-error 'case-incorrect-clause) )
          icslas-unspecified ) )
    `((lambda (,g) ,(generate clauses)) ,exp) ) )

(define-icslas-expander (do bindings test . body)
  (define (get list n default)
    (if (> n 0)
        (if (pair? list)
            (get (cdr list) (- n 1) default)
            (expand-error 'incorrect-do-form list) )
        (if (pair? list) (car list) (default)) ) )
  (let ((g (gensym)))
    `(letrec ((,g (lambda ,(map car bindings)
                    (if ,(car test)
                        ,(handle-sequence (cdr test))
                        (begin ,@body
                               (,g . ,(map (lambda (bind)
                                             (get bind 2 (lambda () 
                                                           (car bind) )) )
                                           bindings )) ) ) )))
                                                     
       (,g . ,(map (lambda (bind)
                     (get bind 1 (lambda () 
                                   (expand-error
                                    'missing-init-in-do bind ) )) )
                   bindings )) ) ) )

;;; This quasiquotation is restricted to a single level.
;;; FUTURE: to be extended.

(define-icslas-expander (quasiquote exp)
  (define (quasiquote-walk e)
    (if (pair? e)
        (if (eq? (car e) 'unquote)
            (if (pair? (cdr e))
                (if (null? (cddr e))
                    (cadr e)
                    (expand-error 'incorrect-unquote-form e) )
                (expand-error 'incorrect-unquote-form e) )
            (if (eq? (car e) 'quasiquote)
                (expand-error "No embedded quasiquotation" e);; TEMP
                (quasiquote-pair-walk e) ) )
        (list 'quote e) ) )
  (define (quasiquote-pair-walk e)
    (if (pair? (car e))
        (if (eq? (car (car e)) 'unquote-splicing)
            (if (pair? (cdr (car e)))
                (if (null? (cddr (car e)))
                    (list 'append
                          (cadr (car e))
                          (quasiquote-walk (cdr e)) )
                    (expand-error 'incorrect-unquote-splicing-form (car e)) )
                (expand-error 'incorrect-unquote-splicing-form (car e)) )
            (list 'cons
                  (quasiquote-walk (car e))
                  (quasiquote-walk (cdr e)) ) )
        (list 'cons
              (list 'quote (car e))
              (quasiquote-walk (cdr e)) ) ) )
  (quasiquote-walk exp) )

;;; end of clmacros.scm
