;;; $Id: expand.scm,v 1.2 1996/08/16 17:12:18 queinnec Exp $

;;; This files defines the macroexpansion algorithm.
;;; FUTURE: should be improved rather than being hacked from XLA/expand.scm.
;;; Syntax errors are not harmonious (should use match-lambda) and
;;; should be better reported.

;;; Exportation: 
;;;                (icslas-expand Sexp)
;;; End of exportations.
;;; Importations:
;;;    check.scm   (variable-list? e)
;;;    utils.scm   (expand-error ...)
;;;                (expand-eval Sexp)
;;; End of importations.

;;;                   Expansion Passing Style
;;; After R. Kent Dybvig, Daniel P. Friedman and Christopher T. Haynes
;;; Lisp and Symbolic Computation, 1, 53-75 (1988),
;;; Kluwer Academic Publishers.

;;; This file uses a local macro to define the predefined expanders
;;; for Icslas. New expanders may be defined by define-expander (see
;;; below). These expanders are added to the global variable
;;; icslas-macroexpand that defines the set of macros that apply to
;;; Icslas programs (not Icslas sources). 

;;; Note: We use here an eta-conversion around icslas-macroexpand to
;;; allow it to be dynamically enriched (see the define-macro for
;;; Icslas defined at the end of this file). Otherwise, Icslas macros
;;; may only be used the next time the icslas-expand function is
;;; invoked. This trick allows to correctly evaluate something like:
;;;  (begin (define-macro (foo...)...) ... (foo...) ...).

(define (icslas-expand form)
  (macro-walk form (lambda (e m) (icslas-macroexpand e m))) )

(define (macro-walk exp macroexpand)
  (macroexpand exp macroexpand) )

;;; Macro-walk extended to a list of terms, except that it preserves
;;; the last cdr of a list. Force the expansion order from left to
;;; right, this is needed to recognize programs as (begin
;;; (define-macro ...) form) and allow the form to use the macro just
;;; defined.

(define (macro-walk* exps macroexpand)
  (if (pair? exps)
      (let* ((a (macro-walk (car exps) macroexpand))
             (d (macro-walk* (cdr exps) macroexpand)) )
        (cons a d) )
    ;; Keep the last cdr of expressions to be macroexpanded.
    exps ) )

;;; The initial macroexpand. It just walks the expression and tries
;;; to macroexpand anything it recognizes. Note that no syntax is
;;; checked since it is possible to specify macros that overload
;;; the native syntax.

(define (initial-macroexpand exp macroexpand)
  (if (atom? exp) exp
      (case (car exp)
        ((quote)  exp)
        ((lambda) (expand-lambda exp macroexpand))
        (else     (macro-walk* exp macroexpand)) ) ) )

;;; Lambda is a specific cases since local variables shadow 
;;; macros. But then it is necessary to recognize the variables so we
;;; must check a little the syntax of variable-list.

(define (expand-lambda exp macroexpand)
  (if (and (pair? (cdr exp))
           (variable-list? (cadr exp)) )
      `(,(car exp) ,(cadr exp)
            . ,(macro-walk* 
                (cddr exp)
                (macro-variables-extend (cadr exp) macroexpand) ) )
      exp ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Two different extenders of syntax are provided. The first one
;;; recognizes forms with a given keyword in functional position, the
;;; other is triggered on single atoms.

;;; Add a new association between a key and an expander. If the
;;; key appears in the car of a form, the expander will be invoked.

(define (macro-form-extend macroexpand key expander)
  (lambda (exp2 macroexpand2)
      (if (atom? exp2) (macroexpand exp2 macroexpand2)
          (if (eq? (car exp2) key)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) ) ) ) )

;;; Add a common expander to many keys.

(define (macro-form-extend* macroexpand keys expander)
  (lambda (exp2 macroexpand2)
      (if (atom? exp2) (macroexpand exp2 macroexpand2)
          (if (member (car exp2) keys)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) ) ) ) )

;;; It is also possible to associate an expander to a single atom,
;;; something reminiscent of symbol-macrolet.

(define (macro-atom-extend macroexpand key expander)
  (lambda (exp2 macroexpand2)
    (if (atom? exp2) 
        (if (eq? exp2 key)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) )
        (macroexpand exp2 macroexpand2) ) ) )

;;; Add a common expander to many atoms

(define (macro-atom-extend* macroexpand keys expander)
  (lambda (exp2 macroexpand2)
    (if (atom? exp2) 
        (if (member exp2 keys)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) )
        (macroexpand exp2 macroexpand2) ) ) )

;;; Enrich a macroexpander so that variables shadow macros whatever
;;; their position are (functional or parametric). Check however if the
;;; list of variables is syntactically correct otherwise don't shadow.

(define (macro-variables-extend variables macroexpand)
  (define (check-variable-list o)
    (or (null? o)
        (symbol? o)
        (and (pair? o)
             (symbol? (car o))
             (check-variable-list (cdr o)) ) ) )
  (if (check-variable-list variables)
      (let ((variables (flat variables)))
        (macro-atom-extend* 
         (macro-form-extend* 
          macroexpand
          variables
          (lambda (exp2 macroexpand2)
            `(,(car exp2) . ,(macro-walk* (cdr exp2) macroexpand2)) ) )
         variables
         (lambda (exp2 macroexpand2)
           exp2 ) ) )
      macroexpand ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Define the macroexpander defining the macros of Icslas. It will be
;;; enriched by define-expander below.

(define icslas-macroexpand initial-macroexpand)

;;; Defines an expander and associate it under a name.  macroexpand
;;; may be used in the body of expanders.  The expanded result is not
;;; automatically expanded again, so macros defined with
;;; define-icslas-expander have to invoke macroexpand recursively.

(define-icslas-macro (define-icslas-expander call . body)
  (let ((name      (car call))
        (variables (cdr call))
        (callvar   (gensym)) )
    `(set! icslas-macroexpand
           (macro-form-extend 
            icslas-macroexpand
            ',name
            (lambda (,callvar macroexpand)
              (apply (lambda ,variables . ,body) (cdr ,callvar)) ) ) ) ) )

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Other expanders for derived expression types (DP1178 on Scheme)

(define-icslas-expander (let* binds . body)
  (macro-walk (if (pair? binds)
                  `(let (,(car binds))
                     (let* ,(cdr binds) . ,body) )
                  (handle-sequence body) )
              macroexpand ) )

(define (handle-sequence body)
  (if (pair? body)
      (if (null? (cdr body))
          (car body)
          `(begin . ,body) )
      `(begin) ) )

(define-icslas-expander (let binds . body)
  (if (symbol? binds)
      ;; Named let
      (macro-walk 
       (let ((name binds)
             (binds (car body))
             (body (cdr body)) )
         (sort-bindings
          binds
          (lambda (inited uninited)
            (when (pair? uninited)
                  (expand-error 
                   "Named let is incompatible of uninitialized bindings" ) )
            `((letrec ((,name (lambda ,(map car binds)
                                . ,body )))
                ,name )
              . ,(map cadr binds) ) ) ) )
       macroexpand )
      ;; Anonymous let
      (sort-bindings 
       binds
       (lambda (inited uninited)
         (let* ((body (analyze-body body macroexpand))
                (body (if (pair? uninited)
                          `(uninitialized-let ,uninited ,body)
                          body )) )
           (if (pair? inited)
               `((lambda ,(map car inited) ,body)
                 . ,(macro-walk* (map cadr inited) macroexpand) )
               body ) ) ) ) ) )
             
;;; Sort the bindings of a let into two classes: these that are initialized
;;; and the others. This allows to not use the (too long) word
;;; uninitialized-let: just write let!

(define (sort-bindings bindings k)
  (let scan ((bindings bindings)
             (inited '())
             (uninited '()) )
    (if (pair? bindings)
        (if (pair? (car bindings))
            (scan (cdr bindings) (cons (car bindings) inited) uninited)
            (scan (cdr bindings) inited (cons (car bindings) uninited)) )
        (k (reverse! inited)
           (reverse! uninited) ) ) ) )

(define (analyze-body body macroexpand)
  (collect-definitions 
   body
   (lambda (definitions body)
     (macro-walk 
      (if (null? definitions)
          (handle-sequence body)
          (let* ((assignments 
                  (map (lambda (definition)
                         (analyze-definition (cadr definition)
                                             (cddr definition)
                                             macroexpand ) )
                       definitions ) )
                 (names (map cadr assignments)) )
            `(letrec ,(map cdr assignments)
               ,(handle-sequence body) ) ) )
      macroexpand ) ) ) )

(define-icslas-expander (letrec binds . body)
  (if (pair? binds)
      `(uninitialized-let
        ,(map (lambda (bind) 
                (if (pair? bind) (car bind) bind) )
              binds )
        ,@(map (lambda (bind) 
                 (if (pair? bind) 
                     (macro-walk `(set! . ,bind) macroexpand)
                     `#t ) )
               binds )
        ,(analyze-body body macroexpand) )
      (analyze-body body macroexpand) ) )

(define-icslas-expander (define call . body)
  (analyze-definition call body macroexpand) )

(define (analyze-definition call body macroexpand)
  (if (pair? call)
      `(set! ,(car call)
             (lambda ,(cdr call)
               ,(analyze-body body macroexpand) ) )
      (if (and (pair? body)
               (null? (cdr body)) )
          `(set! ,call 
                 ,(macro-walk (car body) macroexpand))
          (expand-error 'incorrect-definition call) ) ) )

(define-icslas-expander (lambda variables . body)
  `(lambda ,variables 
     ,(analyze-body body (macro-variables-extend variables macroexpand)) ) )

(define (collect-definitions e* k)
  (if (and (pair? e*) (pair? (car e*)) (eq? (caar e*) 'define))
      (collect-definitions 
       (cdr e*)
       (lambda (bindings body)
         (k (cons (car e*) bindings) body) ) )
      (k '() e*) ) )

(define-icslas-expander (and . arguments)
  (if (pair? arguments)
      (macro-walk (if (pair? (cdr arguments))
                      `(if ,(car arguments)
                           (and . ,(cdr arguments))
                           #f )
                      (car arguments) )
                  macroexpand )
      '#t ) )

(define-icslas-expander (or . arguments)
  (if (pair? arguments)
      (macro-walk (if (pair? (cdr arguments))
                      (let ((g (gensym)))
                        `((lambda (,g)
                            (if ,g ,g (or . ,(cdr arguments))) )
                          ,(car arguments) ) )
                      (car arguments) )
                  macroexpand )
      '#f ) )

(define-icslas-expander (cond . arguments)
  (macro-walk
   (if (pair? arguments)
       (if (pair? (car arguments))
           (if (eq? (car (car arguments)) 'else)
               (if (null? (cdr arguments))
                   (handle-sequence (cdr (car arguments)))
                   (expand-error 'cond-else-not-last) )
               `(if ,(car (car arguments))
                    ,(handle-sequence (cdr (car arguments)))
                    (cond . ,(cdr arguments)) ) )
           (expand-error 'cond-1st-clause) )
       icslas-unspecified )
   macroexpand ) )

(define-icslas-expander (begin0 first . others)
  (macro-walk (let ((g (gensym)))
                `((lambda (,g) 
                    ,(handle-sequence others)
                    ,g )
                  ,first ) )
              macroexpand ) )

;;; This is used when no value is particularly interesting to return
;;; in the expansion. The value of icslas-unspecified must be a legal
;;; Icslas expression that can be legally evaluated in the Icslas
;;; runtime environment.

(define icslas-unspecified '(quote 813))

(define-icslas-expander (case exp . clauses)
  (let ((g (gensym)))
    (define (generate clauses)
      (if (pair? clauses)
          (if (pair? (car clauses))
              (if (eq? (caar clauses) 'else)
                  (if (null? (cdr clauses))
                      (handle-sequence (cdar clauses))
                      (expand-error 'case-else-not-last clauses) )
                  `(if (memq ,g ',(caar clauses))
                       ,(handle-sequence (cdar clauses))
                       ,(generate (cdr clauses)) ) )
              (expand-error 'case-incorrect-clause) )
          icslas-unspecified ) )
    (macro-walk `((lambda (,g) ,(generate clauses)) ,exp)
                macroexpand ) ) )

(define-icslas-expander (do bindings test . body)
  (define (get list n default)
    (if (> n 0)
        (if (pair? list)
            (get (cdr list) (- n 1) default)
            (expand-error 'incorrect-do-form list) )
        (if (pair? list) (car list) (default)) ) )
  (let ((g (gensym)))
    (macro-walk
     `(letrec ((,g (lambda ,(map car bindings)
                     (if ,(car test)
                         ,(handle-sequence (cdr test))
                         (begin (begin . ,body)
                                (,g . ,(map (lambda (bind)
                                              (get bind 2 (lambda () 
                                                            (car bind) )) )
                                            bindings )) ) ) )))
                                                     
        (,g . ,(map (lambda (bind)
                      (get bind 1 (lambda () 
                                    (expand-error
                                     'missing-init-in-do bind ) )) )
                    bindings )) )
     macroexpand ) ) )

;;; This quasiquotation is restricted to a single level.
;;; FUTURE: to be extended.

(define-icslas-expander (quasiquote exp)
  (define (quasiquote-walk e)
    (if (pair? e)
        (if (eq? (car e) 'unquote)
            (if (pair? (cdr e))
                (if (null? (cddr e))
                    (cadr e)
                    (expand-error 'incorrect-unquote-form e) )
                (expand-error 'incorrect-unquote-form e) )
            (if (eq? (car e) 'quasiquote)
                (expand-error "No embedded quasiquotation" e);; TEMP
                (quasiquote-pair-walk e) ) )
        (list 'quote e) ) )
  (define (quasiquote-pair-walk e)
    (if (pair? (car e))
        (if (eq? (car (car e)) 'unquote-splicing)
            (if (pair? (cdr (car e)))
                (if (null? (cddr (car e)))
                    (list 'append
                          (cadr (car e))
                          (quasiquote-walk (cdr e)) )
                    (expand-error 'incorrect-unquote-splicing-form (car e)) )
                (expand-error 'incorrect-unquote-splicing-form (car e)) )
            (list 'cons
                  (quasiquote-walk (car e))
                  (quasiquote-walk (cdr e)) ) )
        (list 'cons
              (list 'quote (car e))
              (quasiquote-walk (cdr e)) ) ) )
  (macro-walk (quasiquote-walk exp) macroexpand) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; These are not defined in Scheme R4RS but useful:

(define-icslas-expander (when condition . body)
  (macro-walk `(if ,condition ,(handle-sequence body) ,icslas-unspecified)
              macroexpand ) )

(define-icslas-expander (unless condition . body)
  (macro-walk `(if ,condition ,icslas-unspecified ,(handle-sequence body))
              macroexpand ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Now the difficult part: how to let the user add its own macros
;;; when writing Icslas code. A new macro is offered with the
;;; following semantics (or restrictions): It evaluates its body in 
;;; the expansion environment. To add a new macro, just enrich the
;;; current value of the icslas-macroexpand global variable with the 
;;; macro-*-extend[*] functions.

(define-icslas-expander (within-macro-world . body)
  (expand-eval `(begin . ,body)) )

;;; If you submit this expression to Icslas, wrapped inside a
;;; with-macro-world form then you will gain a macro definer under
;;; Icslas. Expressions of that file are all considered as wrapped
;;; inside a within-macro-world so the previously described facility
;;; is already usable from Icslas toplevel.

;(within-macro-world
  (set! icslas-macroexpand
        (macro-form-extend 
         icslas-macroexpand
         'define-macro
         (lambda (e m)
           (let* ((call      (cadr e))
                  (body      (cddr e))
                  (name      (car call))
                  (variables (cdr call)) )
             (set! icslas-macroexpand
                   (macro-form-extend
                    icslas-macroexpand
                    name
                    (expand-eval `(lambda (ee mm)
                                    (mm (apply (lambda ,variables . ,body)
                                               (cdr ee) )
                                        mm ) ) ) ) )
             icslas-unspecified ) ) ) )
;  )

;;;        (define-macro (macro-name macro-variables) body...)
;;; defines a macro that can be used under the Icslas toplevel. Error
;;; handling while macroexpanding is very minimal.  The language in
;;; which the macro is written is the underlying Scheme and not
;;; Icslas! See some tests in the icslas.tst file.

;;; end of expand.scm
