/* $Id: dmoscheme.c,v 1.20 1998/04/13 20:31:56 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* This file defines the interface with Anselm Baird-Smith's OScheme.
 * It is made of two parts: the first one includes an automatically
 * generated file that converts exported C macros from DMeroon into
 * regular C functions that can be called from OScheme. The second
 * part is the initialization of this module providing all the
 * symbols needed by OScheme to invoke the DMeroon functions.
 */

#include "dmeroon.h"

#if defined(DM_ICSLAS_IS_PRESENT) && DM_ICSLAS_IS_PRESENT
#include "xla.h"
#endif

DM_DefineStringWithCstring(dmoscheme_version,
  "$Id: dmoscheme.c,v 1.20 1998/04/13 20:31:56 queinnec Exp $");

/******************************************************************
 * All DMeroon objects appear as OScheme instances of class
 * OSDMeroonClass.  To be available for statically allocated objects,
 * this class must exist statically. The following structure 
 * corresponds to OsClassDeclare("DMeroon", ...). It will be patched
 * later (see OS_DMeroon_initialize).
 */

OsClassRec OSDMeroonClass_object = {
  NULL,                         /* should be OsMetaClass ! (see below) */
  "DMeroon",                    /* name */
  sizeof(DMeroon_object),       /* instance size */
  {                             /* method array */
    OsSelInvalidSelector,       /* EQUAL */
    OsSelInvalidSelector,       /* HASH */
    OsSelInvalidSelector,       /* DISPLAY */
    OsSelInvalidSelector,       /* WRITE */
    OsSelInvalidSelector,       /* FINALIZE */
    OsSelInvalidSelector        /* COMPILE */
  }
};

/* NOTE: OScheme does not allocate statically constants such as #t or #f. */

/******************************************************************
 * Allocation.  This is an unsafe allocator since the higher levels
 * ensures that <class> is a Class and <size> is a natural number. The
 * size of the DMeroon and Bigloo prefixes is automatically added to
 * the previous size.

 SIGNATURE:
 (Object (DM_allocate_bare_instance Class nat3 Object))

 NOTE: On Sony, Boehm's GC_malloc return addresses that are multiple
 of 4 whereas, alignment of the most stringent type (as handled by
 gcc) is 8.  For C compatibility since the kernel classes are defined
 in C and must be readable by DMeroon, DM_allocate_bare_instance must
 provide addresses that are multiple of 8. This would be a waste of
 room if performed everywhere, so I optimize this as follows. Such a
 stringent alignment is only needed when there are fields of size 8 or
 at least two indexed fields. So we correct the address returned by
 malloc for these cases. This information is coded in the
 DM_CLASS_IS_STRINGENT bit.

 */

DM_SupportPointer 
DM_allocate_bare_instance (DM_SupportPointer wrapped_class, 
                           DM_nat3           size,
                           DM_SupportPointer wrapped_reference ) {
  DM_Class          *class = DM_FromSupport(wrapped_class);
  DM_SupportPointer wrapped_o;
  
  if ( DM_HasOption(class, DM_CLASS_IS_STRINGENT) ) {
    size += DM_ROOM_FOR_REALIGNMENT;
    wrapped_o = OsGCMalloc (DM_PrefixSize() + size);
    if ( wrapped_o == NULL )
      return DM_Abort(context, DME_ALLOCATION_PROBLEM);
    wrapped_o = (DM_SupportPointer)
      DM_OffsetRealign((DM_SupportAddress)wrapped_o);
  } else {
    wrapped_o = OsGCMalloc (DM_PrefixSize() + size);
    if ( wrapped_o == NULL )
      return DM_Abort(context, DME_ALLOCATION_PROBLEM);
  }
  
  /* Structure the object for OScheme. */
  wrapped_o->os_class = DM_HEADER;

  /* Initialize DMeroon header, its class and proxy, etc. */
  return DM_structure_bare_instance(wrapped_o, 
                                    wrapped_class, 
                                    size,
                                    wrapped_reference );
}

/*
 * This is the higher level function used directly from OScheme to
 * allocate an instance. It is very specific since it takes a variable
 * number of arguments in OScheme and a fix one in C.

 SIGNATURE:
 (Object (DM_allocate Class . anything))

 * FUTURE: This binding is costly in memory, an argc++argv styled
 * interface would be preferrable.
 */

DM_SupportPointer 
DM_allocate (unsigned int argc, OsObject *argv) {
  DM_nat2 i;

  if ( argc < 1 ) {
    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                           , OsIntegerMake (1)
                           , OsIntegerMake (argc));
  }
  { DM_SupportPointer wrapped_class = argv[0];
    DM_nat3 sizes[DM_MAX_INDEXED_FIELDS];

    if ( argc > DM_MAX_INDEXED_FIELDS )
      DM_Abort(context, DME_STATIC_SIZE_MISFIT);
    for ( i=1 ; i<argc ; i++ ) {
      if ( ! OsIsInteger( argv[i] ) )
        DM_Abort(context, DME_NOT_NUMBER);
      sizes[i-1] = OsIntegerValue( argv[i] );
    }
    return DM_allocate_instance(wrapped_class, argc-1, sizes);
  }
}

/******************************************************************
 *                    Cooperation with GC
 *
 * OScheme uses Boehm's GC.
 */

/*
 * When Exit instances are to be recycled, a message must be sent to the
 * remotely referenced Entry. This requires cooperation with the GC.
 * Fortunately, Boehm GC provides this ability.
 */

void
DM_exit_finalizer (void *wrapped_exit, void *client_data) {
  /* Remove the Exit instance from the Exits hashtable */
  DM_remove_exit(wrapped_exit);
}

DM_SupportVoid 
DM_set_exit_finalization (DM_SupportPointer wrapped_exit) {
  GC_register_finalizer(wrapped_exit, DM_exit_finalizer, NULL, NULL, NULL);
  return DM_DONE;
}

/******************************************************************
 * These functions belong to Oscheme and know how to print a string or
 * a fixnum on a variety of output ports: file-port or Tcp of DMeroon.
 * When writing on a Tcp, the len argument if non zero is taken for
 * granted: len characters will be output even if some null chars
 * appear in string. There is an additional case when debugging where
 * the function accepts the port to be stdout or stderr.

 SIGNATURES:
 (boolean (DM_print_char_star string nat3 anything))

 */

DM_SupportBoolean
DM_print_char_star (char              *string, 
                    DM_nat3           len,
                    DM_SupportPointer wrapped_port) 
{
  int n = len ? len : strlen(string);

  if ( wrapped_port == NULL ) return;

  DM_DebugAtLevel(0,
    if (    (wrapped_port == (DM_SupportPointer)stdout)
         || (wrapped_port == (DM_SupportPointer)stderr) ) {
      FILE *stream = (FILE*) wrapped_port;
      
      fprintf(stream, "%*s", n, string);
      fflush(stream);
      return DM_TRUE;
    }
                  );

  if ( OsIsOutputPort( wrapped_port ) ) {
    OsSendDisplay(OsStringMakeFromBytes(n, string), wrapped_port);
    return DM_TRUE;
  }

  { DM_OutBuffer *outbuffer = DM_FromSupport(wrapped_port);
  if ( DM_IsA(outbuffer, OutBuffer) ) {
    DM_print_char_star_in_outbuffer(string, len, wrapped_port);
    return DM_TRUE;
  }
  }

  return DM_Abort(context, DME_UNPRINTABLE_OBJECT);
}

/******************************************************************
 * The error function. 

 SIGNATURE:
 (anything (DM_raise_exception nat3 string string nat3))

 */

static OsError OsE_DMeroon_error;

DM_SupportPointer
DM_raise_exception (DM_nat3 code, 
                    char    *message, 
                    char    *file, 
                    DM_nat3 line) {
  (void) OsErrorTrigger (1, OsE_DMeroon_error, 
                         OsIntegerMake(code),
                         OsStringMake(file),
                         OsIntegerMake(line),
                         OsStringMake(message) );
  /* NOT REACHED */
  return DM_DONE;
}

/******************************************************************
 * The following file defines functions wrapping DMeroon functions and
 * macros. These functions check their argument and translate values
 * from DMeroon and OScheme.
 */

#include "dmoscf.c"

/******************************************************************
 * This is the procedures to initialize DMeroon with OScheme.  The
 * first initializes DMeroon innards the second initializes the
 * OScheme interface and DMeroon. Any DMeroon resource is turned into
 * a symbol incorporated to the OScheme interpreter.

 SIGNATURE:
 (Site (DM_initialize))

 */

DM_SupportPointer
DM_initialize () {

  /* Only initialize once! */
  static DM_nat1 already_done = 0;
  if ( already_done ) return DM_ToSupport(DM_ReferenceTo(DM_site));
  already_done = 1;

  /* Initialize first the generic code and checks its binding. */
  DM_bind_initialize();

  /* Advise Boehm GC that DMeroon uses interior pointers.
   * Also advise that we may correct misalignment (see allocate_bare_instance).
   */
  GC_register_displacement(DM_PrefixSize());
  GC_register_displacement((1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);
  GC_register_displacement(DM_PrefixSize()
                           + (1<<DM_MOST_STRINGENT_ALIGNMENT_LOG)/2);

  /* Initialize the various packages (in order).  */
  DM_api_initialize();
  
  /* Return the site object. */
  return DM_ToSupport(DM_ReferenceTo(DM_site));
}

DM_SupportVoid
Os_DMeroon_initialize (OsInterp interp) {

  /* Avoid a warning telling that this variable is useless. */
  DM_dmoscheme_version_string_object = DM_dmoscheme_version_string_object;

  /* Patch the class of OSDMeroonClass to become OsMetaClass as well
   * as its selectors which cannot be statically reached
   * (os_default_display is compiled static).                 HACK!
   */
  { OsClass c = OsClassDeclare(((OsClass)&OSDMeroonClass_object)->name,
                               ((OsClass)&OSDMeroonClass_object)->isize);
    memcpy(&OSDMeroonClass_object, c, sizeof(OsClassRec));
  }

  /* Create the error object used to report DMeroon anomalies. */
  OsE_DMeroon_error = OsErrorMake(interp,
                                  "error:DMeroon",
                                  "DMeroon anomaly %0 at %1:%2\n%3",
                                  4,
                                  OSE_NORMAL);

  /* Initialize DMeroon per se. Don't need to keep the returned Site
   * instance, it will be made visible by dmoscm.c below. */
  DM_initialize();

  /* This is the automatically generated file that contains the
   * definitions of all the symbols that are needed to interface
   * DMeroon resources from OScheme.
   */
#  include "dmoscm.c"
  
  /* These are done by hand since these numbers are defined as C macros */
  OsAtomMake (interp, "DM-nought-index", OsIntegerMake(DM_nought_index));
  OsAtomMake (interp, "DM-maximum-types-number", 
              OsIntegerMake(DM_MAXIMUM_TYPES_NUMBER) );
  OsAtomMake (interp, "DM-most-stringent-type", 
              &DM_MOST_STRINGENT_TYPE_OBJECT );        

  return DM_DONE;
}

/* end of dmoscheme.c */
