/* $Id: dmoscheme.h,v 1.14 1995/12/15 09:02:11 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

#ifndef DMeroon_OScheme_H
#define DMeroon_OScheme_H

/* This is the OScheme header file for DMeroon.  It defines how
 * DMeroon objects are represented, accessed, managed, etc.  It only
 * contains macros that will be used to instantiate the dm.h header
 * files (which of course must be included after).
 *
 * DMeroon objects have the prefix OScheme expects. These objects belong
 * (for OScheme) to the DMeroon class.
 *
 * This is the precise layout for DMeroon instances as seen from OScheme:
 *
 * typedef struct DMeroon_object {
 *   OsClass       os_class;        <-- OScheme prefix      <== OScheme address
 *   padding
 *   DM_reference _class;           <-- DMeroon prefix
 *   proxy information;
 *   blah blah blah;                <-- DMeroon content    <== DMeroon reference
 * } DMeroon_object;
 *
 * Pay attention that pointers in OScheme designates the word containing the
 * OScheme header while DMeroon references designate the first word of the
 * content of a Dmeroon object. The macros ToSupport and FromSupport make
 * the appropriate conversions.
 */

#define DM_SUPPORT_LANGUAGE "OScheme"

/* Since this is the binding for OScheme, include OScheme header. */

#include "oscheme.h"

/* 
 * This is the header DMeroon objects have for Oscheme.
 */

#define OSDMeroonClass ((OsClass) &OSDMeroonClass_object)
#define DM_HEADER      OSDMeroonClass

/*
 * This is the macro to recognizes a DMeroon instance from an OScheme value.
 */

#define DM_IsDMeroonInstance(wrapped_o) \
  (((OsObject) (wrapped_o))->os_class == OSDMeroonClass)

/*
 * Here are the various kinds of values handled by OScheme.
 * DM_SupportBoolean is only for advertizing predicates.
 * DM_SupportVoid means that the returned value is unspecified.
 */

#define DM_SupportPointer  OsObject
#define DM_SupportFixnum   int
#define DM_SupportFloatnum double
#define DM_SupportBoolean  DM_SupportPointer
#define DM_SupportAddress  long
#define DM_SupportVoid     DM_SupportPointer

/*
 * This is how OScheme values are recognized.
 */

#define DM_IsSupportFixnum(wrapped_o) \
  OsIsInteger(wrapped_o)
#define DM_IsSupportCharacter(wrapped_o) \
  OsIsChar(wrapped_o)
#define DM_IsSupportFloat(wrapped_o) \
  OsIsFloat(wrapped_o)
#define DM_IsSupportDouble(wrapped_o) \
  DM_IsSupportFloat(wrapped_o)
#define DM_IsSupportLongDouble(wrapped_o) \
  DM_IsSupportDouble(wrapped_o)

/*
 * This is how OScheme values are built
 */

#define DM_ToSupport(dmeroon) \
  ((OsObject) (((((DMeroon_object*) (dmeroon))) - 1)) )
#define DM_ToSupportFixnum(int) \
  ((OsObject) (OsIntegerMake((DM_SupportFixnum)int)))
#define DM_ToSupportCharacter(char) \
  ((OsObject) (OsCharMake(char)))
#define DM_ToSupportFloat(float) \
  DM_ToSupportDouble((double)float)
#define DM_ToSupportDouble(double) \
  ((OsObject) (OsFloatMake((DM_SupportFloatnum)double)))
#define DM_ToSupportLongDouble(longdouble) \
  DM_ToSupportDouble((double)longdouble)

/*
 * This is how OScheme values are decoded.
 */

#define DM_FromSupport(oscheme) \
  ((void*)((((DMeroon_object*) (oscheme))) + 1))
#define DM_FromSupportFixnum(wrapped_int) \
  OsIntegerValue(wrapped_int)
#define DM_FromSupportCharacter(wrapped_char) \
  (((OsChar)(wrapped_char))->c)
#define DM_FromSupportFloat(wrapped_float) \
OsFloatValue(wrapped_float)
#define DM_FromSupportDouble(wrapped_double) \
  DM_FromSupportFloat(wrapped_double)
#define DM_FromSupportLongDouble(wrapped_longdouble) \
  DM_FromSupportDouble(wrapped_longdouble)

/*
 * DM_FALSE and DM_TRUE designates the OScheme representations of #f and #t.
 */

#define DM_FALSE OsA_False
#define DM_TRUE  OsA_True

/*
 * When a C procedure has nothing better to return, it returns DM_DONE.
 * Generally such a procedure has type DM_SupportVoid corresponding to
 * C void. The C void type should be avoided since it makes invocation
 * of procedures harder: it is better to always return a value.
 */

#define DM_DONE DM_TRUE

/*
 * The OScheme header occupies one word for its instantiation link.
 * Pad it so the total header is four words. (for padding, we store the
 * size of the instance).
 */

#define DM_SupportPrefix() \
  OsClass        _os_class; \
  long            _padding;

#define DM_SupportPrefixInitialization(type, varname) \
  DM_HEADER, \
  sizeof(type),

#undef DM_SUPPORT_FIXNUM_MAXIMUM
#undef DM_SUPPORT_FIXNUM_MINIMUM

/* 
 * Since cpp (even that of gcc) does not handle long long entities, these
 * two macros are defined as 
 *     #define ULONG_LONG_MAX (LONG_LONG_MAX * 2ULL + 1)
 * No computation at cpp-time can be done on them. Moreover && does not
 * seem to have a short-circuit semantics at cpp-time. Therefore I remove
 * these definitions.
 */

#undef ULONG_LONG_MAX
#undef LONG_LONG_MAX

/******************************************************************
 *                      FUNCTIONS of dmoscheme.c
 */

extern OsClassRec OSDMeroonClass_object;

extern DM_SupportPointer 
DM_allocate_bare_instance (DM_SupportPointer wrapped_class, 
                           DM_nat3           size,
                           DM_SupportPointer wrapped_reference);

extern void
DM_finalize_exit (void *wrapped_exit, void* client_data);

extern DM_SupportVoid
DM_set_exit_finalization (DM_SupportPointer wrapped_exit);

extern DM_SupportPointer
DM_raise_exception (DM_nat3 code, 
                    char    *message, 
                    char    *file, 
                    DM_nat3 line);

extern DM_SupportBoolean
DM_print_char_star (char              *string, 
                    DM_nat3           len,
                    DM_SupportPointer wrapped_port);

extern DM_SupportPointer
DM_initialize (void);

extern DM_SupportVoid
Os_DMeroon_initialize (OsInterp interp);

#endif /* DMeroon_OScheme_H */

/* end of dmosc.h */
