;;; $Id: sig2osc.scm,v 1.6 1996/07/01 17:58:31 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; This file converts signatures into appropriate interfaces for
;;; foreign languages such as module for OScheme.

;;; Don't forget to load the appropriate utilities from sigtools.scm
;;; (load "sigtools.scm")

;;;{{{ Types mapping

(set! types-mapping
  '(;; Remove this case since many languages have problems with that 
    ;; return type:   (void         . void)  !!!
    ;; Dont confuse string and String!
    (string      . OsObject)
    (char        . OsObject)
    (fixnum      . OsObject)
    (floatnum    . OsObject)
    (anything    . OsObject)
    ;; these are the booleans of Bigloo
    (boolean     . OsObject)
    (char1       . OsObject)
    (float1      . OsObject)
    (float2      . OsObject)
    (nat1        . OsObject)
    (nat2        . OsObject)
    (nat3        . OsObject)
    (int1        . OsObject)
    (int2        . OsObject)
    (int3        . OsObject)
    ) )

;;; Remove signatures containing a nat4, an int4, a float3 ...

(define (remove-signatures sigs)
  (if (pair? sigs)
      (if (or (appear? 'int4 (car sigs))
              (appear? 'nat4 (car sigs))
              (appear? 'code (car sigs))
              (appear? 'float3 (car sigs)) )
          (begin 
            (display `(removing ,(car sigs)))(newline)
            (remove-signatures (cdr sigs)) )
          (cons (car sigs) (remove-signatures (cdr sigs))) )
      '() ) )

(define (appear? a e)
  (or (eq? a e)
      (and (pair? e)
           (or (appear? a (car e))
               (appear? a (cdr e)) ) ) ) )

;;;}}}
;;;{{{ Main function

;;; Read signatures, produce two C file to be compiled as an extension
;;; of OScheme. The first will define new symbols bound to C functions
;;; or C variables, the second introduces C functions (checking
;;; arguments) wrapping C macros or C functions.

(define (prepare-c-4oscheme sigfile cfile mfile)
  (complete-types-mapping)
  (let ((signatures (remove-signatures (read-signatures-file sigfile))))
    ;; Declare the C functions for the interpreter.
    (call-with-output-file cfile
      (lambda (out)
        (produce-c-comment "Automatically generated (Don't edit!)" out)
        (for-each (lambda (sig)
                    (produce-c-registration sig out)
                    (newline out) )
                  signatures )
        (produce-c-eof out) ) )
    ;; Wrap C functions or macros inside C functions checking their args.
    (call-with-output-file mfile
      (lambda (out)
        (produce-c-comment "Automatically generated (Don't edit!)" out)
        (for-each (lambda (sig)
                    (produce-wrapper sig out)
                    (newline out) )
                  signatures )
        (produce-c-eof out) ) )
    'done ) )

;;;}}}
;;;{{{ Produces C code to register a C variable or function within OScheme.

(define (produce-c-registration sig out)
  (if (pair? (cadr sig))
      (if (equal? 'macro (car (cadr sig)))
          (produce-c-macro-registration sig out)
          (produce-c-function-registration sig out) )
      (produce-c-constant-registration sig out) ) )

(define (produce-c-function-registration sig out)
  (display "  OsSecSubrMake(interp, \"" out)
  (display (convert-name (car (cadr sig))) out)
  (display "\", DMOS_" out)
  (display (car (cadr sig)) out)
  (display ", " out)
  (display (produce-length (cdr (cadr sig))) out)
  (display " , OS_SECURITY_LOWEST);" out) )

(define (produce-length types)
  (let count ((types  types)
              (result 0) )
    (if (pair? types)
        (count (cdr types) (+ 1 result))
        (if (null? types) 
            result
            "OS_NARY" ) ) ) )      

(define (produce-c-constant-registration sig out)
  (display "  OsAtomMake(interp, \"" out)
  (display (convert-name (cadr sig)) out)
  (display "\", " out)
  (if (DMeroon-class-type? (car sig))
      (begin
        (display "&" out)
        (display (cadr sig) out)
        (display "_object" out) )
      (case (car sig)
        ((nat1 nat2 nat3 int1 int2 int3 netnat2 natnat3
          translation displacement options index offset 
          fixnum )
         (display "DM_ToSupportFixnum(" out)
         (display (cadr sig) out)
         (display ")" out) )
        ((floatnum )
         (display "DM_ToSupportFloat(" out)
         (display (cadr sig) out)
         (display ")" out) )
        (else (sigtools-error "Cannot wrap" (car sig))) ) )
  (display ");" out) )

;;; Macros are handled similarly to functions after stripping the
;;; 'macro' marker.

(define (produce-c-macro-registration sig out)
  (produce-c-function-registration `(,(car sig) ,(cdr (cadr sig))) out) )

;;;}}}
;;;{{{ Wrappers

;;; Macros cannot be bound as they are. A C function wrapping them is
;;; created and this C function is bound to Oscheme. These functions
;;; have the DMOS_ prefix. C functions are also wrapped with a DMOS_
;;; prefixed function that tests the correction of arguments.

(define (produce-wrapper sig out)
  (if (pair? (cadr sig))
      (if (equal? 'macro (car (cadr sig)))
          (produce-macro-wrapper sig out)
          (produce-function-wrapper sig out) )
      (produce-constant-wrapper sig out) ) )

;;; Variables are ignored.

(define produce-constant-wrapper produce-c-comment)

;;; Macros are handled similarly to functions.

(define (produce-macro-wrapper sig out)
  (produce-function-wrapper `(,(car sig) ,(cdr (cadr sig))) out) )

;;; In the case of nary functions, just generate a #define to signify
;;; that DMOS_DM_fun is similar to DM_fun since DM_fun already has the
;;; appropriate argc,argv interface OScheme expects.

(define (produce-function-wrapper sig out)
  (if (null? (cdr (last-pair (cadr sig))))
      (produce-regular-function-wrapper sig out)
      (produce-nary-function-wrapper sig out) ) )

(define (produce-nary-function-wrapper sig out)
  (let* ((name      (car (cadr sig))))
    (display "#define DMOS_" out)
    (display name out)
    (display " " out)
    (display name out) ) )

(define (produce-regular-function-wrapper sig out)
  (let* ((name      (car (cadr sig)))
         (types     (cdr (cadr sig)))
         (variables (convert-sigtypes-to-names types)) )
    (display (convert-type (car sig)) out)
    (display " DMOS_" out)
    (display name out)
    (display-list-with-individual-prefix "OsObject " ", " variables out)
    (display " {" out)
    (newline out)
    ;;   test class membership 
    (produce-checking-code name types variables out)
    ;; Proceed to the real treatment now.
    (produce-return-code (car sig) name types variables out)
    ;; finish the definition
    (display ";" out)
    (newline out)
    (display "}" out) ) )

;;; Produce the code to check the arguments.
;;; name is the C function to call, types the sigtypes of the arguments,
;;; variables the names of the variables (non dotted list).

(define (produce-checking-code name types variables out)
  (if (pair? types)
      (begin
        (produce-variable-checking-code (car types) (car variables) out)
        (produce-checking-code name (cdr types) (cdr variables) out) ) ) )

(define (produce-variable-checking-code type var out)
  (let ((class-name (symbol->string type)))
    (if (DMeroon-class-type? type)
        (begin
          (display "  OsMustBeOfClass(" out)
          (display var out)
          (display ", OSDMeroonClass);" out)
          (newline out)
          (if (not (string=? class-name "Object"))
              (begin
                (display "  if ( ! (DM_IsA(DM_FromSupport(" out)
                (display var out)
                (display "), " out)
                (display type out)
                (display ")) )" out)
                (newline out)
                (display "    DM_Abort(NULL, DME_WRONG_INSTANCE);" out)
                (newline out) ) ) )
        (case type
          ((nat1 nat2 nat3 int1 int2 int3 netnat2 natnat3
                 translation displacement options index offset date
                 fixnum )
           (display "  OsMustBeInteger(" out)
           (display var out)
           (display ");" out)
           (newline out) )
          ((floatnum float1 float2 float3)
           (display "  OsMustBeFloat(" out)
           (display var out)
           (display ");" out)
           (newline out) )
          ((string)
           (display "  OsMustBeString(" out)
           (display var out)
           (display ");" out)
           (newline out) )
          ((char char1)
           (display "  OsMustBeChar(" out)
           (display var out)
           (display ");" out)
           (newline out) )
          ((anything stream void)
           #f )
          (else (sigtools-error "Cannot check" type)) ) ) ) )

;;; Produce the return expression, calls the function, decodes its
;;; arguments and encodes the result.

(define (produce-return-code return-type name types variables out)
  (display "  return " out)
  ;; convert the result to OScheme
  (if (DMeroon-class-type? return-type)
      (display "(" out)
      (case return-type
        ((nat1 nat2 nat3 int1 int2 int3 netnat2 netnat3
          translation displacement options index offset date
          fixnum )
         (display "DM_ToSupportFixnum(" out) )
        ((char char1)
         (display "DM_ToSupportCharacter(" out) )
        ((floatnum float1)
         (display "DM_ToSupportFloat(" out) )
        ((floatnum float2)
         (display "DM_ToSupportDouble(" out) )
        ((boolean anything stream void)
         (display "(" out) )
        ;; otherwise no coercion on the returned value
        (else (sigtools-error "Cannot coerce back" return-type)) ) )
  ;; output the function (or macro) name
  (display name out)
  (display "(" out)
  ;; convert arguments from OScheme now:
  (let loop ((types types)
             (variables variables) )
    (if (pair? types)
        (begin 
          (produce-argument (car types) (car variables) out)
          (if (pair? (cdr types))
              (begin (display ", " out)
                     (loop (cdr types) (cdr variables)) ) ) ) ) )              
  (display "))" out) )

(define (produce-argument type variable out)
  (if (DMeroon-class-type? type)
      (display variable out)
      (case type
        ((nat1 nat2 nat3 int1 int2 int3 netnat2 netnat3
          translation displacement options index offset date
          fixnum )
         (display "DM_FromSupportFixnum(" out)
         (display variable out)
         (display ")" out) )
        ((floatnum float1 float2 float3)
         (display "DM_FromSupportFloat(" out)
         (display variable out)
         (display ")" out) )
        ((string)
         (display "OsStringBytes(" out)
         (display variable out)
         (display ")" out) )
        ((char char1)
         (display "DM_FromSupportCharacter(" out)
         (display variable out)
         (display ")" out) )
        ((anything stream void)
         (display variable out) )
        (else (sigtools-error "Cannot decode" type)) ) ) )

;;;}}}

;;; end of sig2osc.scm
