;;; $Id: deval.scm,v 1.9 1994/10/24 19:36:16 queinnec Exp $

;;; This file defines an evaluator for Tree-Code written in direct
;;; style.  It does not offer concurrency at all and is just used to
;;; define the Scheme kernel of the Icslas language as well as the
;;; basic methods on Tree-Code nodes.

;;; How to use it is defined in the TreeCode/Test/devrun.scm file.

;;;==================================================================
;;; TC is the program to interpret, R the local environment.  No need
;;; to have the explicit global environment, it is patched inside the
;;; Tree-Code. 

(define-generic (evaluate (tc) r))

;;; Check Global-Mutable variables for initialized-ness. This is
;;; faster to test the content of the `value' field rather than adding
;;; an `initialized?' boolean field.

(define-handy-method (evaluate (tc RT-Global-Box) r)
  (let ((v (RT-Box-value tc)))
    (if (eq? v RT-uninitialized)
        (evaluate-error "Uninitialized variable" tc)
        v ) ) )

;;; A predefined global variable is always initialized.

(define-handy-method (evaluate (tc RT-Predefined-Box) r)
  (RT-Box-value tc) )

(define-handy-method (evaluate (tc TC-Constant) r)
  value )

(define-handy-method (evaluate (tc TC-Alternative) r)
  (if (not (eq? RT-f (evaluate condition r)))
      (evaluate consequent r)
      (evaluate alternant r) ) )

(define-handy-method (evaluate (tc TC-Sequence) r)
  (evaluate first r)
  (evaluate last r) )

(define-handy-method (evaluate (tc TC-Global-Assignment) r)
  (let ((newv (evaluate form r)))
    (assign! (TC-Global-Assignment-global tc) newv) ) )

(define-generic (assign! (o RT-Box) newv))

(define-method (assign! (o RT-General-Box) newv)
  (if (RT-General-Box-mutable? o)
      (let ((oldv (RT-Box-value o)))
        (set-RT-Box-value! o newv)
        (if (eq? oldv RT-uninitialized) newv oldv) )
      (evaluate-error "Immutable global" o newv) ) )

(define-method (assign! (o RT-Mutable-Box) newv)
  (let ((oldv (RT-Box-value o)))
    (set-RT-Box-value! o newv)
    (if (eq? oldv RT-uninitialized) newv oldv) ) )

;;;==================================================================
;;; Variables. 
;;; There are many cases. The first two methods arer the most general ones,
;;; the others are specialized ones.

(define-handy-method (evaluate (tc TC-Immediate-Reference) r)
  (RT-Environment-value r offset) )

(define-handy-method (evaluate (tc TC-Deep-Reference) r)
  (let deepen ((depth depth)
               (r r) )
    (if (= depth 0)
        (RT-Environment-value r offset)
        (deepen (- depth 1) (RT-Environment-next r)) ) ) )

(define-method (evaluate (tc TC-Immediate-Reference0) r)
  (RT-Environment-value r 0) )

(define-method (evaluate (tc TC-Immediate-Reference1) r)
  (RT-Environment-value r 1) )

(define-method (evaluate (tc TC-Immediate-Reference2) r)
  (RT-Environment-value r 2) )

(define-method (evaluate (tc TC-Immediate-Reference3) r)
  (RT-Environment-value r 3) )

(define-handy-method (evaluate (tc TC-Deep1-Reference) r)
  (RT-Environment-value (RT-Environment-next r) offset) )

(define-method (evaluate (tc TC-Deep1-Reference0) r)
  (RT-Environment-value (RT-Environment-next r) 0) )

(define-method (evaluate (tc TC-Deep1-Reference1) r)
  (RT-Environment-value (RT-Environment-next r) 1) )

(define-method (evaluate (tc TC-Deep1-Reference2) r)
  (RT-Environment-value (RT-Environment-next r) 2) )

(define-method (evaluate (tc TC-Deep1-Reference3) r)
  (RT-Environment-value (RT-Environment-next r) 3) )

(define-handy-method (evaluate (tc TC-Free-Deep-Reference) r)
  (let deepen ((depth depth)
               (r r) )
    (if (= depth 0)
        (RT-Flat-Invokable-value r offset)
        (deepen (- depth 1) (RT-Environment-next r)) ) ) )

(define-method (evaluate (tc TC-Free-Deep1-Reference0) r)
  (RT-Flat-Invokable-value (RT-Environment-next r) 0) )

(define-method (evaluate (tc TC-Free-Deep1-Reference1) r)
  (RT-Flat-Invokable-value (RT-Environment-next r) 1) )

(define-method (evaluate (tc TC-Free-Deep1-Reference2) r)
  (RT-Flat-Invokable-value (RT-Environment-next r) 2) )

(define-method (evaluate (tc TC-Free-Deep1-Reference3) r)
  (RT-Flat-Invokable-value (RT-Environment-next r) 3) )

;;;==================================================================

(define-handy-method (evaluate (tc TC-Fix-Closure-Creation) r)
  (make-RT-Fix-Closure
   (TC-Closure-Creation-coded-arity tc)
   (TC-Closure-Creation-body tc)
   r ) )

(define-handy-method (evaluate (tc TC-Nary-Closure-Creation) r)
  (make-RT-Nary-Closure
   (TC-Closure-Creation-coded-arity tc)
   (TC-Closure-Creation-body tc)
   r ) )

(define-handy-method (evaluate (tc TC-Fix-Combinator-Creation) r)
  (make-RT-Fix-Combinator
   (TC-Closure-Creation-coded-arity tc)
   (TC-Closure-Creation-body tc) ) )

(define-handy-method (evaluate (tc TC-Nary-Combinator-Creation) r)
  (make-RT-Nary-Combinator
   (TC-Closure-Creation-coded-arity tc)
   (TC-Closure-Creation-body tc) ) )

;;; specialize these ones ? or replaces closed variables by boxes in the body ?
(define-method (evaluate (tc TC-Fix-Flat-Closure-Creation) r)
  (let ((fun (instantiate RT-Fix-Flat-Closure
               :coded-arity  (TC-Closure-Creation-coded-arity tc)
               :body         (TC-Closure-Creation-body tc)
               :value-length (TC-Flat-Closure-Creation-size tc) )))
    fun ) )
    
(define-method (evaluate (tc TC-Nary-Flat-Closure-Creation) r)
  (let ((fun (instantiate RT-Nary-Flat-Closure
               :coded-arity  (TC-Closure-Creation-coded-arity tc)
               :body         (TC-Closure-Creation-body tc)
               :value-length (TC-Flat-Closure-Creation-size tc) )))
    fun ) )

(define-method (evaluate (tc TC-Capture) r)
  (let ((fun (evaluate (TC-Capturer-others tc) r)))
    (set-RT-Flat-Invokable-value! 
     fun (TC-Capture-rank tc) (evaluate (TC-Capturer-first tc) r) )
    fun ) )

(define-method (evaluate (tc TC-Capture0) r)
  (let ((fun (evaluate (TC-Capturer-others tc) r)))
    (set-RT-Flat-Invokable-value! fun 0 (evaluate (TC-Capturer-first tc) r))
    fun ) )

(define-method (evaluate (tc TC-Capture1) r)
  (let ((fun (evaluate (TC-Capturer-others tc) r)))
    (set-RT-Flat-Invokable-value! fun 1 (evaluate (TC-Capturer-first tc) r))
    fun ) )

(define-method (evaluate (tc TC-Capture2) r)
  (let ((fun (evaluate (TC-Capturer-others tc) r)))
    (set-RT-Flat-Invokable-value! fun 2 (evaluate (TC-Capturer-first tc) r))
    fun ) )

(define-method (evaluate (tc TC-Capture3) r)
  (let ((fun (evaluate (TC-Capturer-others tc) r)))
    (set-RT-Flat-Invokable-value! fun 3 (evaluate (TC-Capturer-first tc) r))
    fun ) )

;;; Use a let to force the evaluation order.

(define-handy-method (evaluate (tc TC-Regular-Application) r)
  (let ((rtf (evaluate function r)))
    (invoke rtf (evaluate arguments r)) ) )

(define-handy-method (evaluate (tc TC-Predefined-Application) r)
  (let ((rtf (evaluate function r)))
    (invoke rtf (evaluate arguments r)) ) )

(define-handy-method (evaluate (tc TC-Predefined0-Application) r)
  (address) )

(define-handy-method (evaluate (tc TC-Predefined1-Application) r)
  (let ((frame (evaluate arguments r)))
    (address (RT-Environment-value frame 0)) ) )
          
(define-handy-method (evaluate (tc TC-Predefined2-Application) r)
  (let ((frame (evaluate arguments r)))
    (address (RT-Environment-value frame 0)
             (RT-Environment-value frame 1) ) ) )
          
(define-handy-method (evaluate (tc TC-Predefined3-Application) r)
  (let ((frame (evaluate arguments r)))
    (address (RT-Environment-value frame 0)
             (RT-Environment-value frame 1)
             (RT-Environment-value frame 2) ) ) )

(define-handy-method (evaluate (tc TC-Fix-Let) r)
  (let ((rtb (evaluate arguments r)))
    (set-RT-Environment-next! rtb r)
    (evaluate body rtb) ) )

(define-handy-method (evaluate (tc TC-Arguments) r)
  (let* ((v   (evaluate first r))
         (rtb (evaluate others r)) )
    (set-RT-Environment-value! rtb rank v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Arguments0) r)
  (let* ((v   (evaluate first r))
         (rtb (evaluate others r)) )
    (set-RT-Environment-value! rtb 0 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Arguments1) r)
  (let* ((v   (evaluate first r))
         (rtb (evaluate others r)) )
    (set-RT-Environment-value! rtb 1 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Arguments2) r)
  (let* ((v   (evaluate first r))
         (rtb (evaluate others r)) )
    (set-RT-Environment-value! rtb 2 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Arguments3) r)
  (let* ((v   (evaluate first r))
         (rtb (evaluate others r)) )
    (set-RT-Environment-value! rtb 3 v)
    rtb ) )

;;;==================================================================
;;; Activation frames allocation.
;;; Allocate one more field for use by nary-closures. Don't forget
;;; that the number of arguments is therefore one less the size of the
;;; allocation frame.

(define-handy-method (evaluate (tc TC-Last-Argument) r)
  (let* ((v   (evaluate first r))
         (rtb (instantiate RT-Environment
                :value-length (+ size 1) ) ) )
    (set-RT-Environment-value! rtb size v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Last-Argument0) r)
  (let* ((v   (evaluate first r))
         (rtb (instantiate RT-Environment
                :value-length 2 ) ) )
    (set-RT-Environment-value! rtb 0 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Last-Argument1) r)
  (let* ((v   (evaluate first r))
         (rtb (instantiate RT-Environment
                :value-length 3 ) ) )
    (set-RT-Environment-value! rtb 1 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Last-Argument2) r)
  (let* ((v   (evaluate first r))
         (rtb (instantiate RT-Environment
                :value-length 4 ) ) )
    (set-RT-Environment-value! rtb 2 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Last-Argument3) r)
  (let* ((v   (evaluate first r))
         (rtb (instantiate RT-Environment
                :value-length 5 ) ) )
    (set-RT-Environment-value! rtb 3 v)
    rtb ) )

(define-handy-method (evaluate (tc TC-Allocate-Activation-Frame) r)
  (instantiate RT-Environment
    :value-length (+ size 1) ) )

;;; Needless due to Last-Argument nodes.

;(define-handy-method (evaluate (tc TC-Allocate-Frame0) r)
;  (instantiate RT-Environment
;    :value-length 1 ) )

;(define-handy-method (evaluate (tc TC-Allocate-Frame1) r)
;  (instantiate RT-Environment
;    :value-length 2 ) )

;(define-handy-method (evaluate (tc TC-Allocate-Frame2) r)
;  (instantiate RT-Environment
;    :value-length 3 ) )

;(define-handy-method (evaluate (tc TC-Allocate-Frame3) r)
;  (instantiate RT-Environment
;    :value-length 4 ) )

;;;==================================================================
;;; Invokation mechanism.

(define-generic (invoke (fun) frame))

(define-method (invoke (fun RT-Primitive0) frame)
  (if (= (RT-Environment-value-length frame) 1)
      ((RT-Primitive-address fun))
      (evaluate-error "Expected arity" 0 fun frame) ) )

(define-method (invoke (fun RT-Primitive1) frame)
  (if (= (RT-Environment-value-length frame) 2)
      ((RT-Primitive-address fun)
       (RT-Environment-value frame 0) )
      (evaluate-error "Expected arity" 1 fun frame) ) )

(define-method (invoke (fun RT-Primitive2) frame)
  (if (= (RT-Environment-value-length frame) 3)
      ((RT-Primitive-address fun)
       (RT-Environment-value frame 0)
       (RT-Environment-value frame 1) )
      (evaluate-error "Expected arity" 2 fun frame) ) )

(define-method (invoke (fun RT-Primitive3) frame)
  (if (= (RT-Environment-value-length frame) 4)
      ((RT-Primitive-address fun)
       (RT-Environment-value frame 0)
       (RT-Environment-value frame 1)
       (RT-Environment-value frame 2) )
      (evaluate-error "Expected arity" 3 fun frame) ) )

(define-method (invoke (fun RT-Primitive-1) frame)
  ((RT-Primitive-address fun) frame) )

(define-method (invoke (fun RT-Primitive-3) frame)
  (if (>= (RT-Environment-value-length frame) 3)
      ((RT-Primitive-address fun) frame)
      (evaluate-error "Expected arity" 2 fun frame) ) )

(define (coerce2RT-bool o)
  (if o RT-T RT-f) )

;;; To specialize RT-Closure allows to have specialized arity checks.

(define-method (invoke (fun RT-Fix-Closure) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity (RT-Invokable-coded-arity fun)) )
    (if (= arity argsnum)
        (begin
          (set-RT-Environment-next! frame (RT-Closure-closed fun))
          (evaluate (RT-Invokable-body fun) frame) )
        (evaluate-error "Incorrect arity" fun frame) ) ) )

(define-method (invoke (fun RT-Nary-Closure) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity   (- (- (RT-Invokable-coded-arity fun)) 1)) )
    (if (>= argsnum arity)
        (begin
          (set-RT-Environment-next! frame (RT-Closure-closed fun))
          (listify! frame arity)
          (evaluate (RT-Invokable-body fun) frame) )
        (evaluate-error "Incorrect arity" fun frame) ) ) )

(define-method (invoke (fun RT-Fix-Flat-Closure) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity   (RT-Invokable-coded-arity fun)) )
    (if (= arity argsnum)
        (begin
          (set-RT-Environment-next! frame fun)
          (evaluate (RT-Invokable-body fun) frame) )
        (evaluate-error "Incorrect arity" fun frame) ) ) )

(define-method (invoke (fun RT-Nary-Flat-Closure) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity   (- (- (RT-Invokable-coded-arity fun)) 1)) )
    (if (>= argsnum arity)
        (begin
          (set-RT-Environment-next! frame fun)
          (evaluate (RT-Invokable-body fun) frame) )
        (evaluate-error "Incorrect arity" fun frame) ) ) )

(define-method (invoke (fun RT-Fix-Combinator) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity (RT-Invokable-coded-arity fun)) )
    (if (= arity argsnum)
        (evaluate (RT-Invokable-body fun) frame)
        (evaluate-error "Incorrect arity" fun frame) ) ) )

(define-method (invoke (fun RT-Nary-Combinator) frame)
  (let ((argsnum (- (RT-Environment-value-length frame) 1))
        (arity   (- (- (RT-Invokable-coded-arity fun)) 1)) )
    (if (>= argsnum arity)
        (begin
          (listify! frame arity)
          (evaluate (RT-Invokable-body fun) frame) )
        (evaluate-error "Incorrect arity" fun frame) ) ) )

;;; gather all supplementary arguments in a list and store it in its
;;; right place in the activation frame.

(define (listify! frame arity)
  (let loop ((index (- (RT-Environment-value-length frame) 1))
             (result '()) )
    (if (= arity index)
        (set-RT-Environment-value! frame index result)
        (loop (- index 1)
              (cons (RT-Environment-value frame (- index 1))
                    result ) ) ) ) )

(define-method (invoke (fun RT-Continuation) frame)
  (if (= (RT-Environment-value-length frame) 2)
      ((RT-Continuation-stack fun) (RT-Environment-value frame 0))
      (evaluate-error "Incorrect arity for continuation" fun frame) ) )

;;;==================================================================
;;; Special functions
;;; They are special since they take a variable number of arguments, or 
;;; they receive functions that they must invoke.

(define (icslas-call/cc fun)
  (call/cc (lambda (stack)
             (let ((frame (instantiate RT-Environment
                            :value (make-RT-Continuation stack) 
                                   '() )))      ;; for variadic
               (invoke fun frame) ) )) )

(define (icslas-list frame)
  (let ((len (- (RT-Environment-value-length frame) 1)))
    (let accum ((i 0))
      (if (>= i len)
          '()
          (cons (RT-Environment-value frame i) (accum (+ i 1))) ) ) ) )

(define (icslas-apply frame)
  (let* ((proc (RT-Environment-value frame 0))
         (last-arg-index (- (RT-Environment-value-length frame) 2))
         (last-arg (RT-Environment-value frame last-arg-index))
         (size (+ last-arg-index (length last-arg)))
         (newframe (instantiate RT-Environment 
                     :value-length size )) )
    (do ((i 1 (+ i 1)))
        ((= i last-arg-index))
      (set-RT-Environment-value!
       newframe (- i 1) (RT-Environment-value frame i) ) )
    (do ((i (- last-arg-index 1) (+ i 1))
         (last-arg last-arg (cdr last-arg)) )
        ((atom? last-arg)
         (if (null? last-arg)
             (invoke proc newframe)
             (evaluate-error "Incorrect arguments list" last-arg) ) )
      (set-RT-Environment-value! newframe i (car last-arg)) ) ) )

;;;==================================================================
;;; Value is not important for this instruction. It just puts a value
;;; of the current activation frame in a box since it represents a
;;; mutable variable. Since during these Box-Creations, the frame is
;;; not shared it is safe to mutate it.

(define-handy-method (evaluate (tc TC-Box-Creation) r)
  (let* ((v   (RT-Environment-value r offset))
         (box (make-RT-Mutable-Box v)) )
    (set-RT-Environment-value! r offset box)
    (evaluate body r) ) )

;;; Perform a series of Box-creations. The preceding method now appears
;;; as a specialized version of the next one. It might be useful to 
;;; allocate a single multibox                                  FUTURE

(define-method (evaluate (tc TC-Boxes-Creation) r)
  (let ((len (TC-Boxes-Creation-offset-length tc)))
    (do ((i 0 (+ i 1)))
        ((>= i len) (evaluate (TC-Boxes-Creation-body tc) r))
      (let ((offset (TC-Boxes-Creation-offset tc i)))
        (set-RT-Environment-value! 
         r offset (make-RT-Mutable-Box 
                   (RT-Environment-value r offset) ) ) ) ) ) )

(define-handy-method (evaluate (tc TC-Box-Read) r)
  (RT-Mutable-Box-value (evaluate reference r)) )

(define-handy-method (evaluate (tc TC-Box-Write) r)
  (let* ((box  (evaluate reference r))
         (newv (evaluate form r)) )
    (assign! box newv) ) )

;;; end of deval.scm
