;;; $Id: tree.scm,v 1.8 1994/10/24 13:26:24 queinnec Exp $

;;; This file defines the classes needed to represent compiled
;;; programs in tree-code format. They all derive from Tree-Code and
;;; have a name beginning with TC-. Many classes only exist for
;;; optimization purposes ie compactify current cases, spare room in
;;; objects, lessen memory fetches etc.

(define-class Tree-Code Object () :virtual)

;;; A Tree-Code object is a tree that contains boxes for global
;;; variables which are thus linked to the program (ie present in the
;;; tree-code structure). This makes migration of Tree-Code easier
;;; since only the needed global variables migrate since their
;;; location (their associated box) is present in the code that has to
;;; be migrated. 

;;;==================================================================
;;; Reference to variables. 
;;; There are a lot of specialized classes to access variables.
;;; The runtime environment is represented by linked frames (instances of 
;;; RT-Environment). 

(define-class TC-Reference Tree-Code () :virtual)

;;; Access a variable that is in the first frame of the environment.

(define-class TC-Immediate-Reference TC-Reference
  ( offset      ; a natural number
    ) )

;;; Specialize the previous case for the first offsets.

(define-class TC-Immediate-Reference0 TC-Reference ())
(define-class TC-Immediate-Reference1 TC-Reference ())
(define-class TC-Immediate-Reference2 TC-Reference ())
(define-class TC-Immediate-Reference3 TC-Reference ())

;;; Access a variable that is not in the first frame of the
;;; environment but in a deeper frame.

(define-class TC-Deep-Reference TC-Reference
  ( offset       ; a natural number
    depth        ; a natural number
    ) )

;;; Specialize the previous case for depth one. This is often the
;;; case for a lambda with a local let forms. Do we gain to specialize
;;; wrt depth 2 ? 

(define-class TC-Deep1-Reference TC-Reference
  ( offset       ; a natural number
    ) )

;;; Specialize the previous case for small offsets.

(define-class TC-Deep1-Reference0 TC-Reference ())
(define-class TC-Deep1-Reference1 TC-Reference ())
(define-class TC-Deep1-Reference2 TC-Reference ())
(define-class TC-Deep1-Reference3 TC-Reference ())

;;; Access to a global variable when assigned. In fact the box of the
;;; global variable is present in the global field of that class.

(define-class TC-Global-Assignment Tree-Code
  ( global        ; a global box
    form          ; a tree-code
    ) )

;;; access to a closed variable. Closed variables may appear in a
;;; closure (ie an instance of RT-Flat-Invokable) and therefore must
;;; be accessed specially since the frame that contains them is not a
;;; RT-Environment but a RT-Flat-Invokable. Generally the closure
;;; appear as the end of the linked list of RT-Environment. The
;;; compiler may choose to use flat closures or to capture the current
;;; local environment (see treefy.scm).

(define-class TC-Free-Deep-Reference TC-Deep-Reference ())

;;; specialize the previous case for depth one.

(define-class TC-Free-Deep1-Reference0 TC-Deep1-Reference0 ())
(define-class TC-Free-Deep1-Reference1 TC-Deep1-Reference0 ())
(define-class TC-Free-Deep1-Reference2 TC-Deep1-Reference0 ())
(define-class TC-Free-Deep1-Reference3 TC-Deep1-Reference0 ())

;;;==================================================================
;;; Other classes

;;; A virtual class for Closure creation. The coded-arity is positive
;;; if the arity is positive and negative for dotted lambdas.
;;; Specialized classes exist for these two types of functions.

(define-class TC-Closure-Creation Tree-Code
  ( coded-arity      ; an integer
    body             ; a tree-code
   ) :virtual )

(define-class TC-Fix-Closure-Creation TC-Closure-Creation ())
(define-class TC-Nary-Closure-Creation TC-Closure-Creation ())

;;; Combinators characterize closures that close nothing.

(define-class TC-Fix-Combinator-Creation TC-Closure-Creation ())
(define-class TC-Nary-Combinator-Creation TC-Closure-Creation ())

;;; size gives the number of closed variables ie the size of the flat
;;; closure to allocate.

(define-class TC-Flat-Closure-Creation TC-Closure-Creation 
  ( size             ; a natural 
    ) :virtual )

(define-class TC-Fix-Flat-Closure-Creation TC-Flat-Closure-Creation ())
(define-class TC-Nary-Flat-Closure-Creation TC-Flat-Closure-Creation ())

;;; Alternative, sequence. 

(define-class TC-Alternative Tree-Code
  ( condition
    consequent
    alternant
    ) )

(define-class TC-Sequence Tree-Code
  ( first
    last
    ) )

;;; Quotations values are inserted directly in the Tree-Code (as quote
;;; forms in interpreters directly interpreting Sexp).

(define-class TC-Constant Tree-Code
  ( value
    ) )

;;; Applications.

(define-class TC-Invokation Tree-Code
  ( function           ; a tree-code 
    arguments          ; a tree-code
    ) :virtual )

(define-class TC-Regular-Application TC-Invokation ())

(define-class TC-Primitive-Application Tree-Code 
  ( address            ; the address of the primitive
    offset             ; the offset of the primitive to identify it.
    ) : virtual )

(define-class TC-Predefined-Invokation TC-Primitive-Application
  ( arguments          ; a tree-code
    ) )

;;; Specialized cases for application with 1,2,3 or 4 arguments.

(define-class TC-Predefined0-Application TC-Primitive-Application ())
(define-class TC-Predefined1-Application TC-Predefined-Invokation ())
(define-class TC-Predefined2-Application TC-Predefined-Invokation ())
(define-class TC-Predefined3-Application TC-Predefined-Invokation ())

(define-class TC-Predefined-Application TC-Invokation ())

;;; The `arguments' field return a RT-Environment that will be used to
;;; evaluate the `body' field.

(define-class TC-Fix-Let Tree-Code
  ( arguments         ; a tree-code
    body              ; a tree-code
    ) )

;;; This is the general allocator of activation frames. Specialized
;;; allocators once existed but are no longer generated by the
;;; compiler towards TreeCode. 

(define-class TC-Allocate-Frame Tree-Code () :virtual)

(define-class TC-Allocate-Activation-Frame TC-Allocate-Frame
  ( size            ; natural (size of the activation frame to allocate)
    ) )

;(define-class TC-Allocate-Frame0 TC-Allocate-Frame ())
;(define-class TC-Allocate-Frame1 TC-Allocate-Frame ())
;(define-class TC-Allocate-Frame2 TC-Allocate-Frame ())
;(define-class TC-Allocate-Frame3 TC-Allocate-Frame ())

(define-class TC-One-Argument Tree-Code
  ( first         ; a tree-code
    ) :virtual )

;;; This class generalizes TC-Allocate-Activation-Frame. It has to
;;; evaluate the value of the last argument of the application, then
;;; to allocate a frame and to store the value in it. 

(define-class TC-Last-Argument TC-One-Argument
  ( size
    ) )

;;; specialized cases do exist for typical number of arguments.

(define-class TC-Last-Argument0 TC-One-Argument ())
(define-class TC-Last-Argument1 TC-One-Argument ())
(define-class TC-Last-Argument2 TC-One-Argument ())
(define-class TC-Last-Argument3 TC-One-Argument ())

;;; Application arguments are stored in a RT-Environment that will be
;;; given to the invoked function and which serves as an activation
;;; frame. They have to compute the `first' field and to memorize its
;;; value, to compute the `others' field which returns a frame and to
;;; store the memorized value at its correct rank in the activation
;;; frame. 

(define-class TC-Some-Arguments TC-One-Argument
  ( others         ; a tree-code
    ) :virtual )

(define-class TC-Arguments0 TC-Some-Arguments ())
(define-class TC-Arguments1 TC-Some-Arguments ())
(define-class TC-Arguments2 TC-Some-Arguments ())
(define-class TC-Arguments3 TC-Some-Arguments ())

(define-class TC-Arguments TC-Some-Arguments
  ( rank           ; a natural  number
    ) )

;;; This is similar to TC-Some-Arguments except that these operations
;;; capture variables to be stored in a flat-closure. The `first'
;;; field is computed, its value is memorized, the `others' field is
;;; computed and returns a flat-closure which is filled with the
;;; memorized value at its correct rank.

(define-class TC-Capturer Tree-Code
  ( first          ; a tree-code
    others         ; a tree-code
    ) :virtual )

;;; specialized case for the previous class for typical ranks

(define-class TC-Capture0 TC-Capturer ())
(define-class TC-Capture1 TC-Capturer ())
(define-class TC-Capture2 TC-Capturer ())
(define-class TC-Capture3 TC-Capturer ())

(define-class TC-Capture TC-Capturer
  ( rank            ; a natural number
    ) )

;;;==================================================================
;;; Box-Creation is always applied on a (shallow, immediate) local
;;; variable. Its offset is sufficient to designate this variable.
;;; Boxes can be created one at a time or in a bunch.

(define-class TC-Box-Creation Tree-Code
  ( body           ; a tree-code
    offset         ; the offset of the variable to box
    ) )

(define-class TC-Boxes-Creation Tree-Code
  ( body           ; a tree-code
    (* offset)     ; the offsets of the variables that must be boxed
    ) )

;;; Boxes are obtained  through references to variables.

(define-class TC-Box-Read Tree-Code
  ( reference      ; a reference
    ) )

(define-class TC-Box-Write Tree-Code
  ( reference      ; a reference
    form           ; a tree-code
    ) )

;;; end of tree.scm
