;;; $Id: treefy.scm,v 1.10 1994/11/02 17:33:43 queinnec Exp $

;;; This file defines the compilation of Program instances to
;;; tree-code. The goal is to have the fastest interpretation, the
;;; minimum to migrate and still the information needed to allow for a
;;; user-friendly debugger.

;;; Exportations:
;;;                 (prepare Preparation)
;;; End of exportations.
;;; Importations:
;;;      ???        (Tree-Code-compile-error ...)
;;; End of importations.

;;; To prepare an expression into TreeCode means that it is first turned
;;; into a Program, this Program is then boxified, lifted and finally
;;; converted into TreeCode.

(define-class TreeCode-Preparation Preparation ())

(define-method (prepare (p TreeCode-Preparation))
  (let* ((g (Preparation-globalenv p))
         (e (Preparation-original p))
         (o (objectify e (make-Local-Environment '()) g)) )
    ;; Add the Box transformation to remove mutable local variables:
    (set! o (insert-box! o))
    ;; Lambda-lift in place (turn With-Temp-Function into Flat-Function)
    (set! o (lift! o))
    ;; record the result
    (set-Preparation-prepared! p (->Tree-Code o '() g))
    ;; return the preparation object itself
    p ) )

;;;==================================================================
;;; The function that converts a Program into Tree-Code. 
;;; G is the Global-Environment, R is a list of Local-Variables.

(define-generic (->Tree-Code (o Program) r g))

;;; To ease the interpretation of global or predefined variables, the
;;; global boxes that are associated to them are directly plugged into
;;; the Tree-Code expression. This makes the tree-code non reentrant
;;; since it is statically tied with the global locations. (This is similar
;;; to symbols (appearing in a Sexp) when symbols are equipped with a Cval).

;;;  +------------+     +------------+            +---------------+
;;;  | Predef Ref |     | Predef Var |            | PredefVar Box |
;;;  |     -----------> |            |   becomes  |   <value>     |
;;;  +------------+     +------------+            +---------------+

(define-method (->Tree-Code (o Predefined-Reference) r g)
  (let* ((pv   (Predefined-Reference-variable o))
         (gbox (TC-Predefined-Variable-box pv)) )
    gbox ) )

(define-method (->Tree-Code (o Global-Reference) r g)
  (let* ((gv   (Reference-variable o))
         (gbox (TC-Global-Variable-box gv)) )
    gbox ) )

(define-method (->Tree-Code (o Global-Assignment) r g)
  (let* ((gv   (Global-Assignment-variable o))
         (gbox (TC-Global-Variable-box gv)) )
    (make-TC-Global-Assignment
     gbox
     (->Tree-Code (Global-Assignment-form o) r g) ) ) )

;;;==================================================================
;;; These methods just copy recursively Program classes into their
;;; equivalent Tree-Code classes. Constants are explicit ie built in
;;; the store and just cited.

(define-method (->Tree-Code (o Alternative) r g)
  (make-TC-Alternative 
   (->Tree-Code (Alternative-condition o) r g)
   (->Tree-Code (Alternative-consequent o) r g)
   (->Tree-Code (Alternative-alternant o) r g) ) )

(define-method (->Tree-Code (o Sequence) r g)
  (make-TC-Sequence
   (->Tree-Code (Sequence-first o) r g)
   (->Tree-Code (Sequence-last o) r g) ) )

(define-method (->Tree-Code (o Constant) r g)
  (make-TC-Constant (Constant-value o)) )

;;;==================================================================
;;; Applications

(define-method (->Tree-Code (o Regular-Application) r g)
  (make-TC-Regular-Application
   (->Tree-Code (Regular-Application-function o) r g)
   (->Tree-Code-arguments (number-of (Regular-Application-arguments o))
                          (Regular-Application-arguments o) r g 0 ) ) )

;;; Directly insert the address of the primitive when the primitive
;;; has a fix arity. Also insert the offset to identify the primitive.
;;; Otherwise pass through the predefined box.

(define-method (->Tree-Code (o Predefined-Application) r g)
  (let* ((pv      (Predefined-Application-variable o))
         (gbox    (TC-Predefined-Variable-box pv))
         (value   (RT-Global-Box-value gbox))  ;; MOCHE melange TC/RT
         (address (RT-Primitive-address value))
         (offset  (TC-Predefined-Variable-offset pv))
         (args    (Predefined-Application-arguments o))
         (tcargs  (->Tree-Code-arguments (number-of args) args r g 0)) )
    (cond 
     ((RT-Primitive? value)                  ; dispatch MOCHE
      (cond ((RT-Primitive0? value)
             (make-TC-Predefined0-Application address offset) )
            ((RT-Primitive1? value)
             (make-TC-Predefined1-Application address offset tcargs) )
            ((RT-Primitive2? value)
             (make-TC-Predefined2-Application address offset tcargs) )
             ((RT-Primitive3? value)
              (make-TC-Predefined3-Application address offset tcargs) )
             (else (make-TC-Predefined-Application gbox tcargs)) ) )
     (else (Tree-Code-error "Unknown predefined" o r g)) ) ) )

(define-method (->Tree-Code (o Fix-Let) r g)
  (let ((args (Fix-Let-arguments o)))
    (make-TC-Fix-Let
     (->Tree-Code-arguments (number-of args) args r g 0)
     (->Tree-Code (Fix-Let-body o)
                  (cons (Fix-Let-variables o) r)
                  g ) ) ) )

;;; This function converts a sequence of arguments into a sequence of
;;; Tree-Code computations. The last node allocates the activation
;;; frame (a RT-Environment here), the previous nodes store the values
;;; they computed in the activation frame which is finally given to
;;; the invoked function.

;;; To give an intuition, if we have (f a b) then the tree-code will look like:
;;;    +-------------+      +------------+     +----------+
;;;    | RegularAppl |      | Arguments0 |     | LastArg1 |
;;;    |      f      |      |     a      |     |    b     |
;;;    |      ----------->  |     ---------->  +----------+
;;;    +-------------+      +------------+

(define-generic (->Tree-Code-arguments size (args) r g rank))

(define-method (->Tree-Code-arguments size (args Arguments) r g rank)
  (let ((first  (->Tree-Code (Arguments-first args) r g)))
    (if (= size (+ rank 1))
        ;; This is the last argument
        (case rank
          ((0)  (make-TC-Last-Argument0 first))
          ((1)  (make-TC-Last-Argument1 first))
          ((2)  (make-TC-Last-Argument2 first))
          ((3)  (make-TC-Last-Argument3 first))
          (else (let ((others (->Tree-Code-arguments 
                               size (Arguments-others args) r g (+ rank 1) )))
                  (make-TC-Arguments first others rank) )) )
        ;; this is not the last argument
        (let ((others (->Tree-Code-arguments size (Arguments-others args) 
                                             r g (+ rank 1) )) )
          (case rank
            ((0)  (make-TC-Arguments0 first others))
            ((1)  (make-TC-Arguments1 first others))
            ((2)  (make-TC-Arguments2 first others))
            ((3)  (make-TC-Arguments3 first others))
            (else (make-TC-Arguments first others rank)) ) ) ) ) )

;;; Allocation of small activation frames is no longer necessary since
;;; these cases are recognized by the TC-Last-Argument nodes.

(define-method (->Tree-Code-arguments size (args No-Argument) r g rank)
  (case size
    ;((0)  (make-TC-Allocate-Frame0))
    ;((1)  (make-TC-Allocate-Frame1))
    ;((2)  (make-TC-Allocate-Frame2))
    ;((3)  (make-TC-Allocate-Frame3))
    (else (make-TC-Allocate-Activation-Frame size)) ) )

;;;==================================================================
;;; Functions 

;;; A function accepting only a fix number of arguments has this
;;; number as arity, otherwise it has the opposite minus one (a
;;; negative number) as arity. for instance arity(cons) = 2,
;;; arity(list)=-1 since it takes at least 0 arguments, finally
;;; arity(display)=-2 since it takes at least one argument (also at
;;; most 2 but this is not checked in the invokation protocol).

;;; Distinguish between Fix- and Nary- closure. This helps to invoke
;;; the closures that will be created.

(define-method (->Tree-Code (o Flat-Function) r g)
  (let* ((coded-arity (encode-arity (Flat-Function-variables o)))
         (free (Flat-Function-free o))
         (nfree (number-of free)) )
    (if (> nfree 0)
        (->Tree-Code-Free nfree free r g 0 o)
        ;; No free variables
        ((if (>= coded-arity 0) 
             make-TC-Fix-Combinator-Creation
             make-TC-Nary-Combinator-Creation )
         coded-arity
         (->Tree-Code (Flat-Function-body o)
                      (cons (Flat-Function-variables o) r)
                      g ) ) ) ) )

(define-generic (->Tree-Code-free size (free) r g rank flat))

(define-method (->Tree-Code-free size (free No-Free) r g rank o)
  (let ((coded-arity (encode-arity (Flat-Function-variables o)))
        (free (Flat-Function-free o)) )
    ((if (>= coded-arity 0) 
         make-TC-Fix-Flat-Closure-Creation
         make-TC-Nary-Flat-Closure-Creation )
     coded-arity
     (->Tree-Code (Flat-Function-body o)
                  (list (Flat-Function-variables o)
                        (let free->var ((free free))
                          (if (Free-Environment? free)
                              (cons (Reference-variable
                                     (Free-Environment-first free) )
                                    (free->var
                                     (Free-Environment-others free) ) )
                              '()) ) )
                  g )
     size ) ) )

(define-method (->Tree-Code-free size (free Free-Environment) r g rank o)
  (let ((first (->Tree-Code (Free-Environment-first free) r g))
        (others (->Tree-Code-free
                 size (Free-Environment-others free) r g (+ rank 1) o )))
    (case rank
      ((0)  (make-TC-Capture0 first others))
      ((1)  (make-TC-Capture1 first others))
      ((2)  (make-TC-Capture2 first others))
      ((3)  (make-TC-Capture3 first others))
      (else (make-TC-Capture first others rank)) ) ) )

;;; Encode the arity (see discussion above).

(define (encode-arity variables)
  (let count ((r variables)
              (arity 0) )
    (if (pair? r)
        (if (Local-Variable-dotted? (car r))
            (- (- arity) 1)
            (count (cdr r) (+ arity 1)) )
        arity ) ) )

;;;==================================================================
;;; Variables
;;; Compute lexical indexes to access variables in the environment.
;;; The environment is actually a list of linked RT-Environments that 
;;; define all local variables (from Lambda and Let forms). Optimize
;;; some cases when variables are shallow or the first ones.

;;; Return the offset of variable among variables or #f if absent.

(define (compute-lexical-offset variable variables)
  (let scan ((offset 0)
             (variables variables) )
    (if (pair? variables)
        (if (eq? (car variables) variable)
            offset
            (scan (+ 1 offset) (cdr variables)) )
        #f ) ) )

;;; Return where is variable among the list of list of variables R.

(define (compute-depth-and-offset variable r)
  (let scan ((depth  0)
             (offset 0)
             (r      r) )
    (if (pair? r)
        (let ((n (compute-lexical-offset variable (car r))))
          (if n (cons depth n)
              (scan (+ 1 depth) 0 (cdr r)) ) )
        #f ) ) )

(define-method (->Tree-Code (o Local-Reference) r g)
  (let* ((variable     (Local-Reference-variable o))
         (depth+offset (compute-depth-and-offset variable r)) )
    (if depth+offset
        (let ((depth  (car depth+offset))
              (offset (cdr depth+offset)) )
          (case depth
            ((0) (case offset
                   ((0)  (make-TC-Immediate-Reference0))
                   ((1)  (make-TC-Immediate-Reference1))
                   ((2)  (make-TC-Immediate-Reference2))
                   ((3)  (make-TC-Immediate-Reference3))
                   (else (make-TC-Immediate-Reference offset)) ))
            ((1) (case offset
                   ((0)  (make-TC-Deep1-Reference0))
                   ((1)  (make-TC-Deep1-Reference1))
                   ((2)  (make-TC-Deep1-Reference2))
                   ((3)  (make-TC-Deep1-Reference3))
                   (else (make-TC-Deep1-Reference offset)) ))
            (else (make-TC-Deep-Reference offset depth)) ) )
        (Tree-Code-error "Lost variable" o) ) ) )

(define-method (->Tree-Code (o Free-Local-Reference) r g)
  (let* ((variable     (Local-Reference-variable o))
         (depth+offset (compute-depth-and-offset variable r)) )
    (if depth+offset
        (let ((depth  (car depth+offset))
              (offset (cdr depth+offset)) )
          (case depth
            ((1) (case offset
                   ((0)  (make-TC-Free-Deep1-Reference0))
                   ((1)  (make-TC-Free-Deep1-Reference1))
                   ((2)  (make-TC-Free-Deep1-Reference2))
                   ((3)  (make-TC-Free-Deep1-Reference3))
                   (else (make-TC-Free-Deep1-Reference offset)) ))
            (else (make-TC-Free-Deep-Reference offset depth)) ) )
        (Tree-Code-error "Lost variable" o) ) ) )

;;;==================================================================
;;; Box handling

;;; Due to the box-transformation, box-creation only assign to shallow
;;; variables. Gather box creations into a single TC node rather than
;;; a sequence of Box-Creation (but adopt that latter when there is
;;; only one box to allocate).

;(define-method (->Tree-Code (o Box-Creation) r g)
;  (make-TC-Box-Creation
;   (->Tree-Code (Box-Creation-body o) r g)
;   (compute-lexical-offset (Box-Creation-variable o) (car r)) ) )

;;; Gather all box-creations to perform in a single node.

(define-method (->Tree-Code (o Box-Creation) r g)
  (let gather ((body    (Box-Creation-body o))
               (offsets (list (compute-lexical-offset
                               (Box-Creation-variable o) (car r) ))) )
    (if (Box-Creation? body)
        (gather (Box-Creation-body body)
                (cons (compute-lexical-offset 
                       (Box-Creation-variable body) (car r) )
                      offsets ) )
        (let ((len (length offsets))
              (tcbody (->Tree-Code body r g)) )
          (if (= len 1)
              (make-TC-Box-Creation tcbody (car offsets))
              (apply make-TC-Boxes-Creation
                     tcbody len offsets ) ) ) ) ) )

;;; FUTURE: specialize those two nodes for shallow boxes ?
;;; It is not necessarily a good idea if they are remote ?

(define-method (->Tree-Code (o Box-Read) r g)
  (make-TC-Box-Read
   (->Tree-Code (Box-Read-reference o) r g) ) )

(define-method (->Tree-Code (o Box-Write) r g)
  (make-TC-Box-Write
   (->Tree-Code (Box-Write-reference o) r g)
   (->Tree-Code (Box-Write-form o) r g) ) )

;;; end of treefy.scm
