;;; $Id: icslas.scm,v 1.44 1997/12/09 19:47:17 queinnec Exp $

;;; This function defines what are the sources of Icslas and how to
;;; process them. The single function defined here is:
;;;      (process-icslas-sources features load-function)
;;; where features define what you want while load-function will be
;;; invoked on all the necessary filenames.

;;; NOTE: Pay attention to avoid macros (such as when or unless) in
;;; this file since they are not yet present (they're defined in
;;; Bind/macros.scm). In conclusion stay to pure Scheme.

;;; Source files are processed according to the features that are
;;; present in the features variable. Lots of features are possible
;;; depending on what you want: an icslas interpreter, a compiled
;;; icslas interpreter, a compiler towards C etc. All files should
;;; have the .scm extension. Read the process-icslas-sources to
;;; determine the features you want.

;;; Here are the main features:
;;;   debug to load various utilities useful for debug.
;;;   QDeval to load an interpreter that directly interprets the abstract
;;;          syntax tree (ast).
;;;   TreeCode to load the TreeCode compiler.
;;;   deval to load an interpreter (in direct style) to interpret TreeCode.
;;;   keval to load an interpreter (in CPS) to interpret TreeCode.
;;;   C-generation to load a compiler towards C.
;;;   Java-generation to load a compiler towards Java.
;;;   ByteCode for ByteCode compilation.
;;;   Bigloo if using Bigloo.
;;;   Scheme2C if using Scheme2C.

(define (process-icslas-sources features fn)
  ;; Features are symbols (or strings or numbers) composed with
  ;; connectors such as and, or or not.
  (define (feature-present? condition)
    (if (pair? condition)
        (case (car condition)
          ((and)
           (if (pair? (cdr condition))
               (if (pair? (cddr condition))
                   (and (feature-present? (cadr condition))
                        (feature-present? (cons 'and (cddr condition))) )
                   (feature-present? (cadr condition)) )
               #f ) )
          ((or)
           (if (pair? (cdr condition))
               (if (pair? (cddr condition))
                   (or (feature-present? (cadr condition))
                       (feature-present? (cons 'or (cddr condition))) )
                   (feature-present? (cadr condition)) )
               #t ) )
          ((not) (not (feature-present? (cadr condition))))
          (else (member condition features)) )
        (member condition features) ) )

  ;; Extract the implementation as specified by the features.
  (define (determine-used-scheme features)
    (cond
     ((feature-present? 'bigloo)    'Bigloo)
     ((feature-present? 'bigloo++)  'Bigloo++)
     ((feature-present? 'oscheme)   'OScheme)
     ((feature-present? 'sci)       'Scheme2C)
     ((feature-present? 'scm)       'Scm)
     ((feature-present? 'gsi)       'Gambit)
     ((feature-present? 'Jaja)      'Jaja)
     ;; others to be done ...                          FUTURE
     (else (runxla-error "Cannot determine which Scheme you want to use")) ) )

  ;; To help you, meta-features exist that set up the appropriate set
  ;; of specific features so you only need to specify one of them.
  ;; The direct-mode meta-feature does not modify the given features.
  (cond 

   ((feature-present? 'direct-mode)
    ;; Leave features as they are
    ;; makes the macro if-icslas-feature aware of these features.
    (set! *icslas-features* features) )

   ((feature-present? 'compiler-to-c)
    (set! features
          `(,(determine-used-scheme features)
            (not AddLibrary) gcc (not cps)
            Utilities FrameWork
            ;; Bigloo++ already contains MeroonV3 otherwise only load
            ;; Meroonet which is a very reduced version of MeroonV3.
            ,@(if (feature-present? 'bigloo++) 
                  '() '(Meroonet) )
            TraditionalMacros (not EPS) 
            FrontEnd 
            (not QDeval) (not TreeCode)
            (not ByteCode) C-generation 
            ;; Transmit the debug feature.
            ,@(if (feature-present? 'debug) '(debug) '())
            ) )
    ;; makes the macro if-icslas-feature aware of these features.
    (set! *icslas-features* features) )

   ((feature-present? 'compiler-to-java)
    (set! features
          `(,(determine-used-scheme features)
            (not AddLibrary) 
            Utilities FrameWork
            ,@(if (feature-present? 'bigloo++) 
                  '() '(Meroonet) )
            ,@(if (feature-present? 'jaja)
                  '() '(TraditionalMacros (not EPS)) )
            FrontEnd 
            (not QDeval) (not TreeCode)
            (not ByteCode) Java-generation 
            ;; Transmit the debug feature.
            ,@(if (feature-present? 'debug) '(debug) '())
            ) )
    ;; makes the macro if-icslas-feature aware of these features.
    (set! *icslas-features* features) )
   ;; others to be done ...                              FUTURE
   (else #f) )

  ;; This is the Scheme library of predefined but non primitive
  ;; functions.  They require Meroon-like features and a number of
  ;; global variables surrounded with < and >. This is needed only
  ;; when bootstrapping the system.
  (if (feature-present? '(or AddLibrary))
      (begin 
        ;; predefined predicates and accessors (based on [D]meroon[|et])
        (fn "Library/libscm0")
        ;; predefined primitives written in pure Scheme
        (fn "Library/libscm1")
        ;; the set of printing functions
        (fn "Library/print")
        ;; the set of reading functions
        (fn "Library/read") ) )

  ;; If you use Bigloo to run these files then add Bigloo as feature.
  (if (feature-present? '(or Bigloo Bigloo++))
      (begin 
        ;; These are the features defining Bigloo as well as two macros
        ;; define-icslas-macro and monitor (to simulate Icslas).
        (fn "Bigloo/macbgl")
        ;; These are the support routines for monitor.
        (fn "Bigloo/rtbgl") ) )

  ;; You may alternatively use OScheme to run these files.
  (if (feature-present? 'OScheme)
      (begin
        ;; Features for OScheme, how to define macros and monitor.
        (fn "OScheme/macosc")
        ;; support routines
        (fn "OScheme/rtosc") ) )

  ;; You may also use Scheme->C.
  (if (feature-present? '(or Scheme2C))
      (begin
        ;; Features for Scheme2C, how to define macros and monitor.
        (fn "Scheme2C/macs2c")
        ;; support routines
        (fn "Scheme2C/rts2c") ) )

  ;; You may also use Gambit
  (if (feature-present? '(or Gambit))
      (begin
        ;; Features for Gambit, how to define macros and monitor.
        (fn "Gambit/macgam")
        ;; support routines
        (fn "Gambit/rtgam") ) )

  ;; You may use also Scm.
  (if (feature-present? '(or scm))
      (begin
        ;; Features for Scm, how to define macros and monitor.
        (fn "Scm/macscm")
        ;; support routines
        (fn "Scm/rtscm") ) )

  ;; The Icslas package uses the 'monitor' special form to control
  ;; exceptions. It is simulated in Scheme with a macro and some
  ;; support functions.
  (if (not (feature-present? '(or Icslas)))
      (begin
        ;; Load the monitor simulation
        (fn "Bind/monitor") ) )

  ;; Load the macros and other utilities used by the sources of Icslas.
  (if (feature-present? 'Utilities)
      (begin
        ;; macros.scm defines if-icslas-features as well as other macros
        ;; (such as when and unless) needed by the rest of the source files.
        (fn "Bind/macros")
        ;; function that some implementations lack (such as flat, atom?).
        (fn "Bind/utils")
        ;; functions to manage filenames and files.
        (fn "Bootstrap/filutl")
        ;; a case-sensitive reader (assigned to the %read variable).
        ;;(fn "Tmp/read") 
        ) )

  ;; Load an object system.
  ;; Meroonet is a fat-free Meroon.
  (if (feature-present? 'Meroonet)
      (fn "Bind/meroonet") )
  ;; DMeroonet is Meroonet with DMeroon objects.     NO LONGER VALID
  (if (feature-present? 'DMeroonet)
      (begin 
        ;; Generic functions etc.
        (fn "Bind/DMeroonet") ) )
  ;; The DMeroon memory model.                       NO LONGER VALID
  (if (feature-present? 'DMeroon-level1)
      (begin
        (fn "DMeroon/rt")
        (fn "DMeroon/lib")
        (if (feature-present? 'debug)
            (fn "DMeroon/debug") )
        (fn "DMeroon/tcp")
        (fn "DMeroon/mkclass")
        (fn "DMeroon/fill")
        (fn "DMeroon/highrt")
        (if (feature-present? 'DMeroon-level2)
            (begin
              ;; additional utilities
              (fn "DMeroon/migr")
              (fn "DMeroon/fetch")
              (fn "DMeroon/sched")
              (fn "DMeroon/dict")
              (fn "DMeroon/c")
              (fn "DMeroon/html")
              (fn "DMeroon/scmlib") ) ) ) )
  ;; Otherwise Meroon is assumed to be present.

  ;; Load the basic framework and utilities for all compilers.
  (if (feature-present? 'FrameWork)
      (begin
        ;; framework for handling compilation of expressions
        ;; NOTE: does not seem to be useful ???
        (fn "Bind/prepare")
        ;; exception classes to report errors or warnings
        (fn "Bind/except")
        ;; glue.scm provides all the exception signallers.
        (fn "Bind/glue")
        ) )

  ;; The macroexpansion algorithm (actually a brand of EPS). This is
  ;; the macroexpansion that will be used to expand the expressions to
  ;; be compiled. The sources of the compiler itself is expanded with
  ;; the above-specified macros.
  (if (feature-present? 'EPS)
      (begin (fn "Macros/expand") ) )
  ;; Or traditional Lisp-style macros.
  (if (feature-present? 'TraditionalMacros)
      (begin (fn "Macros/clexpand")
             (fn "Macros/clmacros") ) )

  ;; Load the common front-end of the compilers.
  (if (feature-present? 'FrontEnd)
      (begin
        ;; The classes of objectified programs
        (fn "Ast/progr")
        ;; Local and global environments for compilation
        (fn "Ast/predenv")
        ;; The functions that check the syntax of Icslas programs
        (fn "Ast/check")
        ;; The objectifier (translates a Sexp into a Program instance)
        (fn "Ast/objify")

        ;; The code-walker used for the next analyses.
        (if (feature-present? '(or ByteCode C-generation Java-generation))
            (fn "Analyzes/walk") )
        ;; The alphatization transformation
        (if (feature-present? '(or TreeCode C-generation Java-generation))
            (fn "Analyzes/alphat") )
        ;; The a-normalization
        (if (feature-present? '(or C-generation Java-generation))
            (fn "Analyzes/anorm") )
        ;; The box transformation for mutable local variables
        (if (feature-present? '(or ByteCode C-generation Java-generation))
            (fn "Analyzes/box") )
        ;; The initialization analysis
        (if (feature-present? '(or ByteCode C-generation Java-generation))
            (fn "Analyzes/inian") )
        ;; collect local variables
        (if (feature-present? '(or TreeCode C-generation Java-generation))
            (fn "Analyzes/colltemp") )
        ;; Compute free variables
        (if (feature-present? '(or TreeCode C-generation Java-generation))
            (fn "Analyzes/lift") )
        ;; Migrate quotations and functions to the toplevel
        (if (feature-present? '(or TreeCode C-generation Java-generation))
            (fn "Analyzes/extract") )
        ;; linearize quotations
        (if (feature-present? '(or TreeCode C-generation Java-generation))
            (fn "Analyzes/quotes") )
        ;; Simplify sequences with a reference as first term
        (if (feature-present? '(or Java-generation))
            (fn "Analyzes/rmseq") )
        ) )
  
  ;; Global-variables extractor (for the Ramses subsystem).
  (if (feature-present? '(or Ramses))
      (begin (fn "Ramses/derstand")
             (fn "Ramses/r4rsenv") ) )

  ;; test-driver: needed by all tests.
  (if (feature-present? '(or debug QDeval))
      (fn "Others/tester") )

  ;; QDeval interprets directly the AST.
  (if (feature-present? 'QDeval)
      (fn "Ast/Test/qdeval") )

  ;; Compilation towards TreeCode (which is interpreted by a Tree Interpreter).
  (if (feature-present? 'TreeCode)
      (begin
        ;; The classes of compiled programs (TreeCode)
        (fn "TreeCode/tree")
        ;; The classes of the Runtime entities
        (fn "TreeCode/rtobj")
        ;; The compiler itself (from Programs to TreeCode)
        (fn "TreeCode/treefy")
        (if (feature-present? 'deval)
            (begin
              ;; A first interpreter (in direct style). t does not offer
              ;; //ism
              (fn "TreeCode/deval")
              ;; the runtime environment needed by this interpreter:
              (fn "TreeCode/genenv")
              ;; some utilities needed for tests
              (fn "TreeCode/see") 
              ;; How to run or test 
              (if (feature-present? 'debug)
                  (fn "TreeCode/Test/devrun") ) ) ) ) )

  ;; Compilation towards bytecodes.
  (if (feature-present? 'ByteCode)
      (begin 
        ;; The definition of bytecodes.
        (fn "ByteCode/BCcode")
        ;; this is the compiler towards bytes.
        (fn "ByteCode/BCgen")
        ;; Compilation driver and utilities.
        (fn "ByteCode/BCdrive")
        ;; This defines the predefined environment.
        (fn "ByteCode/BCpred")
        ;; this is what is needed to test this compiler with a bytecode
        ;; machine simulated in Scheme.
        (cond
         ((feature-present? 'debug)
          ;; These are the classes defining the runtime entities
          (fn "ByteCode/BCobj")
          ;; This describes the ByteCode machine
          (fn "ByteCode/BCinst")
          ;; This describes bytecoded object files
          (fn "ByteCode/BCappl")
          ;; This is the initial global predefined environment
          (fn "ByteCode/BCenv")
          ;; These are tests.
          (fn "ByteCode/BCtest") )
         ((feature-present? 'DMeroon)
          ;; convert a ByteCoded module in a DMeroon object or into a file.
          (fn "ByteCode/BC2dm")
          ) ) ) )

  ;; Additional packages for code generation
  (if (feature-present? '(or C-generation Java-generation))
      (begin 
        ;; The pp file is necessary for the format function. So if you
        ;; lack format, also include pp (even if you have it).
        (if (feature-present? '(or Bigloo Bigloo++ OScheme Gambit Scm Jaja))
            (begin 
              (fn "Others/format")
              (fn "Others/pp") ) ) ) )

  ;; Compilation towards C
  (if (feature-present? 'C-generation)
      (begin
        ;; C generators
        (fn "C/Cgen/Cgen")
        ;; C global environment
        (fn "C/Cgen/Cenv")
        ;; C overall compiler
        (fn "C/Cgen/Ccomp")
        ;; test this C generator
        (if (feature-present? 'debug)
            (fn "C/Cgen/Ctest") ) ) )

  ;; Compilation towards Java
  (if (feature-present? 'Java-generation)
      (begin
        ;; Java generators
        (fn "Java/Javagen/javagen")
        ;; Java global environment
        (fn "Java/Javagen/javaenv")
        ;; Java overall compiler
        (fn "Java/Javagen/javacomp")
        ;; test this Java generator
        (if (feature-present? 'debug)
            (fn "Java/Javagen/javatest") ) ) )

  "Icslas sources processed." )

;;; end of icslas.scm
