/* $Id: dmprtcl.c,v 1.40 1998/06/01 19:30:44 queinnec Exp $ 
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////

 * DOCUMENTATION dmprtcl.c
 * This file defines the protocols that may be used to exploit a Connection.
 * Basically, when a Connection is opened, the first characters allow to
 * determine what protocol (http, DMeroon, ...) it requires. Then the
 * appropriate handler is invoked on it. In case of errors, the Connection
 * will be closed. 
 */

#include "dmeroon.h"

DM_DefineVersionString(dmprtcl,
  "$Id: dmprtcl.c,v 1.40 1998/06/01 19:30:44 queinnec Exp $");

/* Handlers have this signature that deserves a macro to be written.
 */

#define DM_DeclareHandlerFunction(name, value) \
  DM_c_boolean (*name)(DM_Context*,DM_InBuffer*) = \
    (DM_c_boolean (*)(DM_Context*,DM_InBuffer*))(value)

/* RESOURCE DM_handle_readable_connection
 * This is a simple interface to DM_handle_protocol. It must handle
 * all possible messages already present in a Connection. It returns
 * DM_FALSE if no message is present. Handling a Connection is
 * protected with a new Context, in case of error, the Connection
 * will be closed. 
 */

DM_c_boolean
DM_handle_readable_connection (DM_Context*    context,
                               DM_Connection* connection) {
  volatile DM_InBuffer *inbuffer = connection->reception;
  DM_c_boolean result;

  DM_ProtectWithFreshContext(context, newcontext, 
          result = DM_handle_protocol(newcontext, 
                                      (DM_InBuffer*)inbuffer, 
                                      connection->protocol ),
      /* Something went wrong, close immediately the Connection. */
      DM_SetOption(connection, DM_PORT_IS_DRIED);
      result = DM_FALSE;
);
  /* Empty the InBuffer. */
  if ( DM_HasOption(connection, DM_PORT_IS_DRIED) ) {
    inbuffer->index = inbuffer->maximum;
  }
  return result;
}

/* {{{ Determine protocol

 * RESOURCE DM_handle_protocol
 * To allow messages emitted by a telnet client, they have a header
 * (a word followed by any information and finished by CR-LF). The word
 * indicates how to handle the message. Among words are "?", "error" and
 * "DMeroon". This function returns DM_FALSE if no message is handled.

 * When a Port is found readable, the following function tries to
 * handle the  first message that is contained in the associated
 * InBuffer. First, we check if the header is complete (has a newline
 * in the buffer). If this is not the case, then the buffer is rearranged.

 * This function is called from DM_handle_readable_connection which
 * protects it with a fresh context.
 */

DM_c_boolean
DM_handle_protocol (DM_Context*  context,
                    DM_InBuffer* inbuffer,
                    DM_Protocol* protocol) {
  DM_Port     *port = inbuffer->port;
  DM_Protocol *prtcl;
  DM_nat3     lf_index, start_index, left;

  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);

    /* Special case when a Port runs an already known protocol: apply
     * it immediately.  This is used by the DMeroon inner protocol
     * mode. If the handler is NULL then determine which protocol to
     * use. */
    if ( connection->handler ) {
      DM_DeclareHandlerFunction(handler, connection->handler);

      return (*handler)(context, inbuffer);
    }
    
    /* Is a header completely present ? Just look if there is a
     * complete line ie a LF character in the buffer: refer to that LF
     * character with lf_index. Just check for LF since Unix uses LF
     * ie newline and DOS CR-LF in that order. */
    start_index = inbuffer->index;
    left        = inbuffer->maximum - start_index;
    lf_index    = start_index;
    while ( lf_index < inbuffer->maximum ) {
      if ( inbuffer->byte[lf_index] == '\n' ) {
        goto header_present;
      } else {
        lf_index++;
      }
    }
    /* Here, the header is absent or incomplete, the LF character is
     * missing, then rearrange the buffer (to make room) and return
     * DM_FALSE meaning that the Connection will have to wait for
     * additional characters. */
      
    DM_DebugAtLevel(5,
      if ( left > 0 )
        fprintf(DM_FindErrorStream(context), 
                "Incomplete message header in " DM_0X "%p\n", 
                inbuffer);
);

    DM_rearrange_inbuffer(context, inbuffer, 0);
    return DM_FALSE;
      
    /* There is a message header ie at least one line, lookup for a
     * matching protocol. The linked list of protocols ends with
     * NULL. */
  header_present:
    prtcl = protocol;
      
  analyze_protocol:
    if ( prtcl ) {
      /* This length always includes the final \0 character. */
      int prtcl_name_length = prtcl->name_length;

      /* Always successful if the name is reduced to "\0". Consider a
       * zero-length protocol name to be equivalent to "\0". Use
       * strncmp to exclude the final \0 from being compared. */
      if ( ( prtcl_name_length <= 1 )
          || ( !strncmp( &(inbuffer->byte[start_index]), 
                        prtcl->name,
                        prtcl_name_length-1 ) ) ) {
        DM_DeclareHandlerFunction(handler, prtcl->handler);
          
        DM_DebugAtLevel(5,
          if ( prtcl_name_length >= 1 ) {
            fprintf(DM_FindErrorStream(context), 
                    "Protocol: \"%s\"\n", prtcl->name);
          } else {
            fprintf(DM_FindErrorStream(context), 
                    "Protocol: \"\\0\"\n");
          }
);

#if !defined(DMEROON_IS_MUTE)
        /* Report some statistical information */
        { struct in_addr ia;

          if ( DM_IsA(context, port, Tcp) ) {
            DM_Tcp *tcp = (DM_Tcp*)port;
            char *protocol_name = prtcl->name;

            ia.s_addr = tcp->address;
            if ( prtcl_name_length < 1 ) {
              protocol_name = "\\0";
            }
            fprintf(DM_FindErrorStream(context),
                    "**Connection from %s uses protocol \"%s\"\n", 
                    inet_ntoa(ia), protocol_name );
          } else {
            /* not a Tcp, ignore it! */
          }
        }
#endif
      
        /* Record the protocol and invoke it. */
        connection->handler = DM_CodeOf(handler);
        return ((*handler)(context, inbuffer));
      } else {
        /* Try the next possible protocol. */
        prtcl = prtcl->next;
        goto analyze_protocol;
      }
    } else {
      /* Default case: no more protocols to try. The first word of the
       * header is not recognized as a known protocol, close the Port
       * immediately! */

      DM_DebugAtLevel(5,
        /* Change LF into \0 to be sure that the strind ends.  We do not
         * worry for the patched buffer since we are about to close the
         * connection and free associated resources. */
        inbuffer->byte[lf_index] = '\0';
        fprintf(DM_FindErrorStream(context), "Unknown protocol %s\n",
                &(inbuffer->byte[start_index]) );
);
      
      inbuffer->index = inbuffer->maximum;
      DM_SetOption(port, DM_PORT_IS_DRIED);
      return DM_FALSE;
    }
  } else {
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* }}} 
 * {{{ Protocols

 * DOCUMENTATION 
 * The predefined protocol handlers.
 * They all have a similar signature: they receive the InBuffer (the
 * beginning of the header is indicated by InBuffer->index). It is the
 * duty of these handlers to consume the header and the associated
 * body of the message if either. If this is not possible (because,
 * for instance the message is incomplete) then the handler should
 * return DM_FALSE. This means that the handler will wait for some new
 * characters to come before retrying to handle the request.
 */

/* 
 * The 'help' protocol simply answers the identification of the site,
 * then close the connection. An alias for 'help' is '?'.
 */

static DM_c_boolean
DM_help_protocol_handler (DM_Context  *context,
                          DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;
  
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    DM_Site *site      = DM_TheCurrentSite();
    char    line[100];

    sprintf(line, "DMeroon@%s:%lu\r\n",
            site->name,
            (long unsigned)(ntohs(site->port)) );
    DM_print_char_star_in_outbuffer(context, line, 0, connection->emission);
    /* Dry the Connection so it will be closed later. */
    inbuffer->index = inbuffer->maximum;
    DM_SetOption(connection, DM_PORT_IS_DRIED);
    return DM_TRUE;
  } else {
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/*
 * The 'error' protocol just prints on the debug stream its parameters
 * that is the rest of the line. It does not answer anything and it
 * does not close the connection. [FUTURE? is it useful ?]
 */

static DM_c_boolean
DM_error_protocol_handler (DM_Context  *context,
                           DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;
  DM_nat3 index = inbuffer->index;
  DM_nat3 lf_index = index;

  /* Look for the next LF. DM_handle_protocol ensures it exists. */
  while ( inbuffer->byte[lf_index] != '\n' ) 
    lf_index++;
  /* Delimit the string to be printed to make %s works. */
  inbuffer->byte[lf_index] = '\0';
  fprintf(DM_FindErrorStream(context), "%s\n", &(inbuffer->byte[index]));
  inbuffer->index = lf_index+1;		/* remove the message. */
  /* forget the current protocol. */
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);

    connection->handler = DM_NULL;
  }
  return DM_TRUE;
}

/* 
 * The '#' protocol ignores the header. It is mainly used to ignore
 * comments in files piped through a Connection. We also ignore lines
 * starting with a whitespace or empty lines ie lines simply
 * consisting of LF or CR-LF characters.  Caution: an empty line also
 * marks a new message in DMeroon inner protocol mode.
 */

static DM_c_boolean
DM_sharp_protocol_handler (DM_Context  *context,
                           DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;
  DM_nat3 index = inbuffer->index;
  DM_nat3 lf_index = index;

  /* Look for the next LF. DM_handle_protocol ensures it exists. */
  while ( inbuffer->byte[lf_index] != '\n' ) 
    lf_index++;
  inbuffer->index = lf_index+1;		/* remove the message. */
  /* forget the current protocol. */
  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);

    connection->handler = DM_NULL;
  }
  return DM_TRUE;
}

/* RESOURCE DM_decode_percent
 * Transform in place a string with %ff encoded characters. The encoding
 * is used in the http protocol. start_index refers to the first 
 * character to handle, end_index is the first not to be handled.
 */

void
DM_decode_percent (DM_Context  *context,
                   char        *start,
                   char        *end )
{ 
  char* from = start;
  char* to   = from;

  while ( from < end ) {
    if ( *from == '%' ) {
      int  c  = (from++, 0);
      char c1 = *from++;
      char c2 = *from++;

      if ( ('0' <= c1) && (c1 <= '9') ) {
        c = c1 - '0';
      } else if ( ('a' <= c1) && (c1 <= 'f') ) {
        c = c1 - 'a' + 10;
      } else if ( ('A' <= c1) && (c1 <= 'F') ) {
        c = c1 - 'A' + 10;
      } else {
        *to++ = '%';
        *to++ = c1;
        *to++ = c2;
        continue;
      }
      c <<= 4;  /* for c = c<<4;     Fun! isn't it ? */
      if ( ('0' <= c2) && (c2 <= '9') ) {
        c += c2 - '0';
      } else if ( ('a' <= c2) && (c2 <= 'f') ) {
        c += c2 - 'a' + 10;
      } else if ( ('A' <= c2) && (c2 <= 'F') ) {
        c += c2 - 'A' + 10;
      } else {
        *to++ = '%';
        *to++ = c1;
        *to++ = c2;
        continue;
      }
      *to++ = (char)c;
    } else {
      if ( from == to ) {
        /* Nothing to move. */
        to++; from++;
      } else {
        *to++ = *from++;
      }
    }
  }
  while ( to < end ) {
    *to++ = *end;
  }
}

/* DOCUMENTATION http get protocol
 * The 'get' protocol is the html protocol.
 * So far, these requests look like:
 *      GET /path/name [ HTTPidentification ]
 *      [sometimes followed by many additional lines]
 *      CR LF CR LF
 * We just encode DMeroon requests in the /path/name.
 */

static DM_c_boolean
DM_get_protocol_handler (DM_Context  *context,
                         DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;

  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    DM_nat3 request_index = inbuffer->index;
    volatile DM_nat3 start_index;
    volatile DM_OutBuffer *outbuffer = connection->emission;
  
    /* skip 'get' word. */
    while ( isalnum(inbuffer->byte[request_index]) )
      request_index++;
    /* skip spaces and tabs. */
    while ( DM_strchr(" \t", inbuffer->byte[request_index]) )
      request_index++;
    start_index = request_index;
    /* Identify the end of the request. Most clients suffix it with
     * something like HTTP/1.0. They also add numerous lines. Convert
     * the first blank or LF into a \0 so we just keep the pathname.  */
    while ( !(DM_strchr(" \n\r\t", inbuffer->byte[request_index])) )
      request_index++;
    inbuffer->byte[request_index] = '\0';

    /* Now convert all %ff trigrams into the correct letters. */
    DM_decode_percent (context, 
                       (char*)(&(inbuffer->byte[start_index])),
                       (char*)(&(inbuffer->byte[request_index])) );
  
    DM_ProtectWithFreshContext(context, newcontext,
       DM_answer_html_request(newcontext, 
                              (char*)(&(inbuffer->byte[start_index])),
                              DM_NULL,
                              (DM_OutBuffer*)outbuffer ),
       /* Something went wrong. */
       goto dry_connection
                             );
    /* In all cases, clean up the Connection after evaluating the
     * script.  Remove the current message and skip the rest of the
     * buffer. Moreover, since an HTML client waits for the end of the
     * connection, dry the Connection so it will be closed later by
     * DM_analyze_readable_port. */
  dry_connection:
    inbuffer->index = inbuffer->maximum;
    DM_SetOption(connection, DM_PORT_IS_DRIED);
    return DM_TRUE;
  } else {
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* RESOURCE DM_post_protocol_handler
 * Decode a post request to mutate a DMeroon object. 
 * So far, these requests look like:
 *       POST /path/name [ http identification]
 *       [ lots of User-Agent, Accept ... directives ... ]
 *       Content-Length: <size of name=value& pairs>
 *       [ ... lots of other directives ... ]
 *       CR LF CR LF
 *       name=value&name=value
 * 
 * The basic idea is to convert the C string "name=value&..." into
 * a String that will be pushed at the bottom of the stack of the 
 * UrlScript interpreter (so the modify command may find it there).
 */

static DM_c_boolean
DM_post_protocol_handler (DM_Context  *context,
                          DM_InBuffer *inbuffer ) {
  DM_Port   *port = inbuffer->port;
  DM_String *nvpairs;

  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    DM_nat3 request_index = inbuffer->index;
    DM_nat3 url_start_index;
    DM_nat3 header_end_index;
    DM_nat3 content_length = 0;

    /* skip 'POST' word. */
    while ( isalnum(inbuffer->byte[request_index]) )
      request_index++;
    /* skip spaces and tabs. */
    while ( DM_strchr(" \t", inbuffer->byte[request_index]) )
      request_index++;
    /* Mark the first valid char of the url. */
    url_start_index = request_index;
    /* Identify the end of the POST header that is, CR LF CR LF. 
     * Trespassing the end of the buffer means that the request is
     * not yet entirely present. */
    while ( 1 ) {
      if ( request_index+3 < inbuffer->maximum ) {
        if ( !strncmp(&(inbuffer->byte[request_index]), "\r\n\r\n", 4) ) {
          /* Found the end of the header of the POST request.*/
          header_end_index = request_index;
          break;
        } else {
          request_index++;
        }
      } else {
        /* The POST request is not complete yet. */
        return DM_FALSE;
      }
    }
    /* Search now for the Content-Length directive within the header. */
    request_index = url_start_index;
    while ( 1 ) {

#define CONTENT_LENGTH "Content-Length:"
#define CONTENT_LENGTH_SIZE (sizeof(CONTENT_LENGTH)-1)

      if ( (request_index+CONTENT_LENGTH_SIZE) < inbuffer->maximum ) {
        if ( !strncasecmp(&(inbuffer->byte[request_index]), 
                          CONTENT_LENGTH,
                          CONTENT_LENGTH_SIZE ) ) {
          /* Found the Content-Length directive. */
          content_length = 
            DM_string_to_nat3(context, 
                              &(inbuffer->byte[request_index+CONTENT_LENGTH_SIZE]),
                              NULL);
          break;
        } else {
          request_index++;
        }
      } else {
        /* No Content-Length directive. Refuse the POST request. */
        goto dry_connection;
      }
    }
    if ( inbuffer->maximum < (header_end_index + 4 + content_length) ) {
      /* Request is not yet complete. */
      return DM_FALSE;
    }

    /* The request is now complete. Finally the request will look as if
     * it was written as:
     *     "name1=value&name2=value2" url \0.
     */
    request_index = url_start_index;
    /* Find the end of the url. */
    while ( !DM_strchr(" \n\r\t", inbuffer->byte[request_index]) )
      request_index++;
    inbuffer->byte[request_index] = '\0';

    /* Convert the "name=value&..." into a String (after decoding it). */
    nvpairs = DM_Allocate1(context, String, content_length+1);
      
    DM_decode_percent(context,
                      (char*)(&(inbuffer->byte[header_end_index + 4])), 
                      (char*)(&(inbuffer->byte[header_end_index 
                                              + 4 + content_length])) );
    memcpy(&(nvpairs->character[0]), 
           &(inbuffer->byte[header_end_index + 4]),
           content_length );
    DM_SetCopyable(context, nvpairs);
    DM_ProtectWithFreshContext(context, newcontext,
         DM_answer_html_request(newcontext, 
                                (char*)(&(inbuffer->byte[url_start_index])),
                                (DM_Object*)nvpairs,
                                connection->emission ),
       /* Something went wrong. */
       goto dry_connection
                             );
    /* In all cases, clean up the Connection after evaluating the
     * script.  Remove the current message and skip the rest of the
     * buffer. Moreover, since an HTML client waits for the end of the
     * connection, dry the Connection so it will be closed later by
     * DM_analyze_readable_port. */
  dry_connection:
    inbuffer->index = inbuffer->maximum;
    DM_SetOption(connection, DM_PORT_IS_DRIED);
    return DM_TRUE;
  } else {
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* DOCUMENTATION URL scripting protocol
 * Interpret a program written in the URL language but not reduced to a 
 * single line. The program is accumulated until complete then it is
 * interpreted. The program is determined to be complete when the last
 * characters read are \0\0\r\n. See DMeroon/Commands/urlscript.prl 
 * that sends such scripts.
 */

static DM_c_boolean
DM_url_script_protocol_handler (DM_Context  *context,
                                DM_InBuffer *inbuffer ) {
  DM_Port *port = inbuffer->port;
  DM_nat3 index = inbuffer->index;
  DM_nat3 lf_index = index;

  /* Look for the next LF. DM_handle_protocol ensures it exists. */
  while ( inbuffer->byte[lf_index] != '\n' ) 
    lf_index++;

  if ( DM_IsA(context, port, Connection) ) {
    DM_DeclareNarrowedObject(connection, Connection, port);
    char          *q          = (char*)(&(inbuffer->byte[inbuffer->maximum]));
    volatile DM_OutBuffer  *outbuffer = connection->emission;

    if (   (inbuffer->maximum > 4 )
        && ( *(--q) == '\n' )
        && ( *(--q) == '\r' )
        && ( *(--q) == '\0' )
        && ( *(--q) == '\0' ) ) {
      DM_ProtectWithFreshContext(context, newcontext,
            DM_answer_html_request(newcontext, 
                                   (char*)(&(inbuffer->byte[lf_index+1])),
                                   DM_NULL,
                                   (DM_OutBuffer*)outbuffer ),
         /* Something went wrong. */
         goto dry_connection;
                               );
      /* In all cases, clean up the Connection after evaluating the script. */
    dry_connection:
      inbuffer->index = inbuffer->maximum;
      /* Finish output */
      DM_print_char_star(context, "\r\nend of URLscript.\r\n", 
                         0, (DM_OutBuffer*)outbuffer );
      /* close the port so the client knows it is finished. */
      DM_SetOption(connection, DM_PORT_IS_DRIED);
      return DM_TRUE;
    } else {
      /* The program has not yet been entirely read. */
      return DM_FALSE;
    }
  } else {
    /* NOTHING: not a connection! */
    DM_Abort(context, DME_NOT_CONNECTION);
    return DM_FALSE;
  }
}

/* DOCUMENTATION DMeroon protocol
 * This is the 'DMeroon' protocol to unmarshal DMeroon instances.
 * The size of the message body includes the nat3 number that codes this size
 * excluding the final LF. Therefore size is at least 4 and the index is 
 * left on the final LF. The Port switches to inner DMeroon protocol mode.
 *
 *.FIGURE Message
 *        +-----------------------------------------------+
 *   byte |    header\n size content\n 
 *        +-----------------------------------------------+
 *         0   ^      ^      ^       ^    ^               ^
 *             index  |      |       |    maximum         byte_length
 *                    |      |       lf_index+size
 *                    |      lf_index+1+4
 *                    lf_index
 *.ENDFIGURE
 *
 * The inner DMeroon protocol avoids sending again and again the
 * "DMeroon" header. It simply corresponds to a sequence of
 * size--content messsages.  In this inner mode, the lf_index refers 
 * to a LF character (ignore index which is before the first time
 * the inner DMeroon protocol is entered).
 *
 * In fact, a size less than 4 allows to control the inner DMeroon
 * protocol.  0 corresponds to some padding just to keep the
 * connection alive (the index is left on the following NL). 1 means
 * that the size has not been set (probably DM_finish_message was not
 * called). 3 means exit from the inner DMeroon protocol, the index is
 * then left after the following NL.
 *
 *.FIGURE Inner DMeroon protocol control message
 *        +-----------------------------------------------+
 *   byte |   \n3000\n 
 *        +-----------------------------------------------+
 *             ^     ^ index (after)
 *             |
 *             lf_index (before)
 *.ENDFIGURE
 */

/* RESOURCE DM_dmeroon_protocol_read_object
 * Decode an object from a buffer and store it in a given address. Returns
 * DM_FALSE if unsuccessful ie the current message is incomplete. Do not
 * try to unmarshal an incomplete message! 
 *
 * A message in DMeroon inner protocol starts with a size (four bytes), 
 * followed by the content of the message ended with a LF. The 'index'
 * will be left on that LF character.
 */

static DM_c_boolean
DM_dmeroon_protocol_read_object (DM_Context  *context,
                                 DM_InBuffer *inbuffer,
                                 DM_nat3     lf_index,
                                 DM_Object*  *result) {
  DM_Port *port = inbuffer->port;
  DM_nat3 content_length = 0;
    
  /* If the size of the body of the message is not entirely present
   * then rearrange the buffer. */
  if ( (lf_index+4) >= inbuffer->maximum ) {
    DM_rearrange_inbuffer(context, inbuffer, 0);
    return DM_FALSE;
  }

  /* Read the size of the message body, do not use DM_receive_nat3
   * which increments the 'index' field of InBuffer.  */
  content_length = 
    (DM_nat3)(inbuffer->byte[lf_index+1])
    + 256 * ( (DM_nat3)(inbuffer->byte[lf_index+2])
            + 256 * ( (DM_nat3)(inbuffer->byte[lf_index+3])
                      + 256 * (DM_nat3)(inbuffer->byte[lf_index+4]) ) );

  /* Special cases where content_length is less than 4.  Among these
   * four cases: 0 is explicitly ignored (is just used for ping-ing a
   * site), while 3 exits from inner DMeroon protocol. In this case,
   * the final LF character is also swallowed so another protocol may
   * be used.  */
  switch (content_length) {
  case 0: {

    DM_DebugAtLevel(7,
      { if ( DM_IsA(context, port, Connection) ) {
          DM_DeclareNarrowedObject(connection, Connection, port);

          fprintf(DM_FindErrorStream(context), "Ping-ed by %s:%lu\n",
                  connection->site->name,
                  (long unsigned)ntohs(connection->site->port) );
        } 
      }
);

    /* To be ping-ed does not count as a real action so return false. */
    /* move the index to the next LF (if the buffer is large enough). */
    if ( (lf_index+5) < inbuffer->maximum ) {
      if ( inbuffer->byte[lf_index+5] != '\n' ) {
        DM_Abort(context, DME_NOT_AN_LF);
        return DM_FALSE;
      }
      inbuffer->index = lf_index+5;
      return DM_FALSE;
    } else {
      return DM_FALSE;
    }
  }
  case 1: {
    /* This is a sure sign that the size field was not set up because
     * I always put one before patching this number to hold the real
     * size of the message. */
    DM_Abort(context, DME_INCORRECT_MESSAGE_SIZE);
    return DM_FALSE;
  }
  case 2: {
    /* Currently unused. */
    DM_Abort(context, DME_UNUSED_PROTOCOL_SUBCODE);
    return DM_FALSE;
  }
  case 3: {
    /* swallow next LF if buffer is large enough. */
    if ( (lf_index+5) < inbuffer->maximum ) {
      if ( inbuffer->byte[lf_index+5] != '\n' ) {
        DM_Abort(context, DME_NOT_AN_LF);
        return DM_FALSE;
      }

      DM_DebugAtLevel(5,
        fprintf(DM_FindErrorStream(context), 
                "End of inner DMeroon protocol\n");
);

      DM_ClearOption(port, DM_PORT_RUNS_DMEROON);
      DM_reset_cache(context, inbuffer->cache);
      DM_reset_stack(context, inbuffer->stack);
      inbuffer->index = lf_index+6;
      /* forget the current protocol. */
      if ( DM_IsA(context, port, Connection) ) {
        DM_DeclareNarrowedObject(connection, Connection, port);
        
        connection->handler = DM_NULL;
      }
      return DM_TRUE;
    } else {
      return DM_FALSE;
    }
  }
  default: {
    break;
  }
  }

  /* Is the buffer big enough to contain the whole message? */
  if ( (lf_index+content_length) >= inbuffer->maximum ) {

    DM_DebugAtLevel(5,
      fprintf(DM_FindErrorStream(context), 
              "Message not complete\n");
);

    DM_DebugAtLevel(5,
      if ( inbuffer->byte[lf_index+content_length] != '\n' )
        DM_Abort(context, DME_INTERNAL_ERROR);
);

    /* Adjoin 512 bytes systematically for hysterisis. */
    DM_rearrange_inbuffer(context, 
                          inbuffer, 
                          lf_index+content_length+1+512 );
    return DM_FALSE;
  }

  /* The message is entirely present, remove the header, the size
   * and decodes the object and stores at the indicated place. 
   * The first time the DMeroon protocol is entered, reset input cache,
   * switches the Connection to the inner DMeroon protocol and turns off
   * prefetching. */
  if ( DM_HasNotOption(port, DM_PORT_RUNS_DMEROON) ) {
    DM_SetOption(port, DM_PORT_RUNS_DMEROON);
    DM_SetOption(port, DM_PORT_WITHOUT_PREFETCHING);
    DM_reset_stack(context, inbuffer->stack);
    DM_reset_cache(context, inbuffer->cache);
  }
  /* Advance index onto the first byte of the content of the message. 
   * This swallows the header the first time the protocol is entered */
  inbuffer->index = lf_index+5;
  /* NOTE: *result may be an Exit item without harm since it is only
   * written in a reference field of a present object. 
   * NOTE: No need to protect this instruction since it is already 
   * done by DM_handle_readable_connection. */
  *result = DM_read_object(context, inbuffer);

  /* Once the Connection is initialized then it is possible to
   * prefetch objects. This is not possible before since the
   * SiteConnection-Answer or -Request must be the first messages to
   * be exchanged and we don't want, within DM_read_object, that
   * prefetching be triggered. In fact (see DM_read_site_object), no
   * prefetch should take place while decoding a Site object. */
  DM_ClearOption(port, DM_PORT_WITHOUT_PREFETCHING);

  /* Normally, they should be the same. */
  if ( inbuffer->index != (lf_index+content_length) ) {
    /* Which one to trust if they disagree?  */
    DM_Abort(context, DME_ALIGNMENT_ERROR);
    return DM_FALSE;
  } else {
    return DM_TRUE;
  }
}

/* RESOURCE DM_dmeroon_protocol_handler
 * Handle the DMeroon protocol. Decode an object if the buffer holds it
 * entirely then enqueue it in the list of received objects in the current
 * site instance (in reverse order). Switch the Connection into DMeroon 
 * inner protocol mode.
 */

static DM_c_boolean
DM_dmeroon_protocol_handler (DM_Context  *context,
                             DM_InBuffer *inbuffer ) {
  DM_Object    *o;
  DM_c_boolean result = DM_FALSE;
  DM_c_boolean r;
  DM_nat3 index = inbuffer->index;
  DM_nat3 lf_index = index;

  /* Look for the next LF. DM_handle_protocol ensures it exists. 
  * In the inner DMeroon protocol, this is the first character. */
  while ( inbuffer->byte[lf_index] != '\n' ) {
    lf_index++;
    if ( lf_index >= inbuffer->maximum ) {
      DM_Abort(context, DME_INTERNAL_ERROR);
    }
  }
  inbuffer->index = lf_index;

  /* Decode an object and process it.  
   * NOTE: Do not toggle yet the Connection to the inner DMeroon
   * protocol mode, wait for the message to be complete (currently only
   * the header (beginning with DMeroon...) is complete).
   * NOTE: No need to protect that unmarshaling operation since if
   * something wrong occurs, we longjmp to handle_readable_connection
   * that will close the Connection.   */
  do {
    lf_index = inbuffer->index;
    r = DM_dmeroon_protocol_read_object(context, inbuffer, lf_index, &o);
    if ( r ) result = r;
  } while ( r );
  return result;
}

/* RESOURCE DM_add_protocol
 * Register a new protocol within the current Site.
 * This protocol is inserted in front of the present ones so it will be
 * tested first. It should ensure that all protocole.names are NUL-terminated.
 */

DM_Protocol*
DM_add_protocol (DM_Context   *context,
                 char         *string,
                 DM_c_boolean (*handler)(DM_Context*, DM_InBuffer*) ) {
  /* Include the final \0. */
  DM_nat3     string_length = strlen(string) + 1;
  DM_Protocol *prtcl        = DM_Allocate1(context, Protocol, string_length);
  
  prtcl->next    = DM_TheCurrentSite()->protocol;
  prtcl->handler = (DM_code)handler;
  strcpy(prtcl->name, string);
  DM_TheCurrentSite()->protocol = prtcl;
  if ( context && DM_HasNotOption(context, DM_CONTEXT_DONT_INITIALIZE) ) {
    DM_SetSharable(context, prtcl);
  }
  return prtcl;
}

/* RESOURCE
 * This is the password to use for all restricted protocols.
 * Yes I know it appears uncrypted in the core image of this process.
 */

char DM_password[DMEROON_PASSWORD_LENGTH+1];

/* RESOURCE DM_add_restricted_protocol
 * Register a new protocol within the current Site.
 * Its name will be sufffixed with the current password.
 */

DM_Protocol*
DM_add_restricted_protocol (DM_Context   *context,
                            char         *string,
                            DM_c_boolean (*handler)(DM_Context*, 
                                                    DM_InBuffer* ) ) {
# define pw_limit 255 /* Limit for full protocol names */
  char s[pw_limit];

  if ( strlen(string)+DMEROON_PASSWORD_LENGTH >= pw_limit )
    DM_Abort(context, DME_PROTOCOL_NAME_TOO_BIG);
  strcpy(s, string);
  strcat(&(s[strlen(s)]), DM_password);
  return DM_add_protocol(context, s, handler);
}

/* could add (as in PAWS) DM_add_regexp_protocol.        FUTURE
 * }}}
 * {{{ Initialize
 */

void
DM_prtcl_initialize (DM_Context *context) 
{
  char *rtpw;
  long l;
  
  DM_LinkVersionStringIntoSite(dmprtcl);

  /* Test if DM_ProtectWithFreshContext works. There is a problem with
   * that macro on sparc machines wrt automatic variables.  The
   * DM_prtcl_test procedure is defined in dmprint.c to avoid to be
   * inlined by a too clever compiler. */
  { DM_Object *o1 = DM_TheConstant(true);
    volatile DM_Object *o2 = DM_TheConstant(false);

    DM_ProtectWithFreshContext(context, newcontext,
            DM_prtcl_test(newcontext, o1, (DM_Object*)o2),
       /* Something went wrong but DM_prtcl_test already raise the error. */
       DM_BarfIfReached(context);
);
    /* Normal exit of DM_Protect... */
    DM_prtcl_test(context, o1, (DM_Object*)o2);
  }

  /* Register the available protocols in the Site object. Feel free
   * to add your own protocols with the DM_add_protocol function. */
  DM_add_protocol(context, "\n",    DM_sharp_protocol_handler);
  DM_add_protocol(context, "\r\n",  DM_sharp_protocol_handler);
  DM_add_protocol(context, "#",     DM_sharp_protocol_handler);
  DM_add_protocol(context, "help",  DM_help_protocol_handler);
  DM_add_protocol(context, "?",     DM_help_protocol_handler);
  DM_add_protocol(context, "error", DM_error_protocol_handler);
  DM_add_protocol(context, "get",   DM_get_protocol_handler);
  DM_add_protocol(context, "GET",   DM_get_protocol_handler);
  DM_add_protocol(context, "post",   DM_post_protocol_handler);
  DM_add_protocol(context, "POST",   DM_post_protocol_handler);
  /* Plan to add HEAD protocol ?              FUTURE */

  /* NOTE: If you want your DMeroon server to be more secure, you may
   * require an additional passwd that may be defined with the
   * DMEROON_ADDITIONAL_PASSWORD macro or in the DMEROON_PASSWORD
   * environment variable. */
  DM_password[0] = '\0';
  l = strlen(DMEROON_ADDITIONAL_PASSWORD);
  if ( l > 0 ) {
    if ( l >= DMEROON_PASSWORD_LENGTH )
      DM_Abort(context, DME_ADDITIONAL_PASSWORD_TOO_BIG);
    /* Insert compile-time password first. */
    strcat(DM_password, DMEROON_ADDITIONAL_PASSWORD);
  }
  if ( (rtpw = getenv("DMEROON_PASSWORD") ) ) {
    l = strlen(DM_password);
    if ( (l+strlen(rtpw)) >= DMEROON_PASSWORD_LENGTH )
      DM_Abort(context, DME_DMEROON_PASSWORD_TOO_BIG);
    /* Insert run-time password second. */
    strcat(&(DM_password[l]), rtpw);
  }

  DM_DebugAtLevel(0,
    if ( DM_password[0] ) {
      fprintf(DM_FindErrorStream(context), 
              "Password activated.\n");
    }
);

  /* DM_password now contains the password to use. Any protocol name
   * that follows is implicitly suffixed with the password. */

  DM_add_restricted_protocol(context, "URLscript", 
                             DM_url_script_protocol_handler);
  DM_add_restricted_protocol(context, "DMeroon",   
                             DM_dmeroon_protocol_handler);

}

/* }}}
 * end of dmprtcl.c */
